/*
 * Copyright (c) 1992,1993,1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	audiodecode.c - top-level control for audio decoding, independent
 *      of the audio type (mu-law, aiff, mpeg).
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/audiodecode.c,v 1.37 1995/08/18 00:32:22 russ Exp $
 */

#define NeedFunctionPrototypes 1

#include "vmpegdecomp.h"
#include "fileContent.h"
#include "mu-law.h"


#define NON_MPEG_SCHED_FREQ	(4)
#define MPEG_SCHED_FREQ		(32)
#define INITIAL_AUDIO		(MicroSecPerSec / 2)
#define NOSEEK_AUDIO		(-1)

void
CloseAudio(DisplayInfoPtr dpyInfo)
{
    /*
     * Don't close the file, just the device
     */
    if (dpyInfo->audioOpened) {
	(*dpyInfo->audioData.funcs->fileClose)(dpyInfo);
	DebugTask(N_AudioActivity, Msg("close au\n"));
    }
    dpyInfo->audioOpened = False;
    dpyInfo->audioEnabled = False;
}

InternalReasonPtr
OpenAudio(DisplayInfoPtr dpyInfo)
{
    InternalReasonPtr reason;
    AudioInfoPtr audioData = &dpyInfo->audioData;
    static InternalReason noFile = { "Can't open audio file\n", IR_BadIOFile};
    static InternalReason badInit = { "audio funcs not fully initialized!\n",
				  IR_InternalProblem };
    FileInfo fileInfo;

    /* have we been uninited? */
    if( ! IsInitialized(dpyInfo, SYM_AUDIOINPATH) ||
        dpyInfo->new.audioEnabled != True ) {
	dpyInfo->audioEnabled = False;
	return(NULL);
    }

    if (dpyInfo->reopenAudio) {
	CloseAudio(dpyInfo);
	dpyInfo->reopenAudio = False;
    } else {
	return(NULL);
    }
    audioData->seekPicture = -1;	/* force it different */
    
    if (reason = GetFileContentInfo(dpyInfo->audioInput,&fileInfo))
	return(reason);
    /* schedRate is advisory since the device routines control the schedule */
    switch(fileInfo.type) {
    case FT_AiffAudio:
    case FT_AifcAudio:
    case FT_SunAudio:
	audioData->funcs = &AnyAudioFuncs;
	audioData->schedRate = NON_MPEG_SCHED_FREQ;
	break;
    case FT_MpegAudio:
    case FT_MpegSystem:
	audioData->funcs = &MpegAudioFuncs;
	audioData->schedRate = MPEG_SCHED_FREQ;
	break;
    default:
	DebugTask(N_AudioActivity, Msg("open au: no matching types\n"));
	return(NULL);
    }

    if (audioData->funcs->fileOpen == NULL
     || audioData->funcs->filePreload == NULL
     || audioData->funcs->fileLoadNext == NULL
     || audioData->funcs->fileSetQuality == NULL
     || audioData->funcs->devInit == NULL
     || audioData->funcs->devOpen == NULL
     || audioData->funcs->devClose == NULL
     || audioData->funcs->devUpdate == NULL)
	return( &badInit );

    dpyInfo->audioEnabled = True;
    (*audioData->funcs->devInit)(dpyInfo);
    if(! (*audioData->funcs->fileOpen)(dpyInfo)) {
	DebugTask(N_AudioActivity,
	    Msg("open au: can't open %s\n", dpyInfo->audioInput));
    	return(&noFile);
    }
    dpyInfo->audioOpened = True;
    audioData->audioQuality = dpyInfo->audioQuality;
    
    return(NULL);
}

void
StopAudioPlayback(DisplayInfoPtr dpyInfo)
{
    DebugTask(N_AudioActivity, Msg("stop au\n"));

    if( dpyInfo->audioEnabled != True ) {
	return;
    }
    (*dpyInfo->audioData.funcs->devClose)(dpyInfo);
    dpyInfo->nextAudioTime = ~0;

    /*
     * do some audio preloading for the next time we play
     */
    PreloadAudio(dpyInfo);
}

void
SchedAudioPlayback(DisplayInfoPtr dpyInfo)
{
    ULong aTime;
    
    if( dpyInfo->audioEnabled == True ) {
	aTime = SchedulerStartDelay;
	if( dpyInfo->audioDelay > 0 ) {
	    aTime += dpyInfo->audioDelay;
	}
	dpyInfo->nextAudioTime = aTime;

	if(! (*dpyInfo->audioData.funcs->devOpen)(dpyInfo)) {
	    Msg("no audio device available\n");
	    dpyInfo->nextAudioTime = ~0;
	    return;
	}

	DebugTask(N_AudioActivity,
	    Msg("sched au: clk=%d, first audio @%d\n",
		dpyInfo->pSeq->seqClock, dpyInfo->nextAudioTime);
	)
    } else {
	dpyInfo->nextAudioTime = ~0;
    }
}

void
UpdateAudioPlayback(DisplayInfoPtr dpyInfo)
{
    AudioInfoPtr aData = &dpyInfo->audioData;
    ULong auOverhead = 0;
    DebugDeclare(
	ULong auWriteTime = 0;
	ULong auReadTime = 0;
    )

    dpyInfo->u.s.timeForAudio = False;

    if( dpyInfo->state == DisplayStopped || dpyInfo->audioEnabled != True ) {
	return;
    }
    DebugTask(N_AudioActivity, UpdateClock(dpyInfo->pSeq, NULL))
    if ((*aData->funcs->devUpdate)(dpyInfo)) {
	DebugTask(N_AudioActivity, UpdateClock(dpyInfo->pSeq, &auWriteTime))
	if (! (*aData->funcs->fileLoadNext)(dpyInfo)) {
	    dpyInfo->nextAudioTime = ~0;
	    return;
	}
	DebugTask(N_AudioActivity, UpdateClock(dpyInfo->pSeq, &auReadTime))
	UpdateClock(dpyInfo->pSeq, &auOverhead);

	DebugTask(N_AudioActivity,
	    auOverhead += auReadTime + auWriteTime;
	    Msg("upd au: clk=%d, ply=%d, overhead=(%d,%d)(%d%%) next au @%d\n",
	        dpyInfo->pSeq->seqClock,
	        aData->timeGather,
	        auWriteTime, auReadTime,
	        (long)(100 * auOverhead / aData->timeGather),
	        dpyInfo->nextAudioTime);
	)

	dpyInfo->nextAudioTime += aData->timeGather;
	dpyInfo->audioOverhead += auOverhead;
	dpyInfo->u.s.timeForAudio = False;	/* may have been set */

    } else {
	StopAudioPlayback(dpyInfo);
    }
}

void
PreloadAudio(DisplayInfoPtr dpyInfo)
{
    AudioInfoPtr audioData = &dpyInfo->audioData;
    Queue *dq = &PictBuffer.displayQueue;
    SequenceInfoPtr pSeq = dpyInfo->pSeq;
    PictureInfoPtr pic;
    long picNumber;
    DebugDeclare(char *errFunc = "PreloadAudio";)

    if( dpyInfo->audioEnabled != True ) {
	return;
    }

    if( QueueEmpty(dq)) {
	picNumber = 0;
    } else {
	pic = TailOfQueue(dq, PictureInfoPtr);
	picNumber = pic->picNumber;
    }

    /* 
     * We may have already done the necessary seeking
     */
    if (picNumber == audioData->seekPicture) {
	if (audioData->audioQuality != dpyInfo->audioQuality)
	    (*audioData->funcs->fileSetQuality)(dpyInfo);
	audioData->audioQuality = dpyInfo->audioQuality;
	return;
    }

    audioData->seekPicture = picNumber;
    audioData->timeSeek = picNumber * MicroSecPerSec
    			  / FrameRateTable[ pSeq->pictureRateCode ];
    DebugTask(N_AudioActivity,
	      Msg("%s: seek to frame %d\n", errFunc, picNumber)
    )
    audioData->timeGather = INITIAL_AUDIO;
    dpyInfo->audioEnabled = (*audioData->funcs->filePreload)(dpyInfo);
    audioData->timeSeek = -1;
}
