/*
 *      (c) Copyright 1992 North Valley Research, Inc.
 *              All Rights Reserved
 *
 *      Confidential -- All proprietary rights reserved.
 *      Licensed Material -- Property of North Valley Research, Inc.
 *
 *      This program is the property of North Valley Research, Inc. or others
 *      from whom North Valley Research has obtained a licensing right, and is
 *      considered by North Valley Research to be confidential.  It is protected
 *      by U.S. copyright law as an unpublished work and is
 *      furnished pursuant to a written license agreement.  It may
 *      not be used, copied or otherwise reproduced, or disclosed to
 *      others except in accordance with the terms and conditions of
 *      that agreement.
 *
 *      NAME
 *              convertrgb.c
 *
 *      DESCRIPTION
 *              Fast conversion routines for y/cr/cb -> rgb
 *
 *      REVISION
 *              $Header: /usr/local/src/nvrToMIT/RCS/convertrgb.c,v 4.2 1993/10/22 23:53:29 russ Exp $
 */


#define NeedFunctionPrototypes 1

#include "nvrTypes.h"
#include "general.h"
#include "bitstream.h"
#include "longjump.h"
#include "decoder.h"
#include "display.h"

void PrepareYCr();
double fabs();
double atof();

#ifdef LOOKUP_RGB

/*
 * This routine looks up eight rgb values from eight y values
 * that are arranged like:
 * -----------------
 * | 0a| 1a| 2b| 3b|  (yTop)
 * -----------------
 * | 4a| 5a| 6b| 7b|  (yBottom)
 * -----------------
 * and two pairs of cr/cb values that are arranged like:
 * -----------------
 * |cra|cba|crb|cbb|  (crCbIndex)
 * -----------------
 * and returns the four bottom rgb pixels in a long like this:
 * -----------------
 * | 0a| 1a| 2b| 3b|  (return value)
 * -----------------
 * and writes the four top rgb pixels in a long like this:
 * -----------------
 * | 4a| 5a| 6b| 7b|  (*pRGB)
 * -----------------
 * The yTop and yBottom values have already been shifted right one bit,
 * anded with 0x7f7f7f7f to remove the top bit of each byte, and had
 * the dither value added to them.  Since the dither value for G is
 * one bit lower than for R and B, we can simply add the dither value
 * again and it effectively gives us a dithered G.
 */
Inline static OctWord
Lookup232RGB(RGBLookup *rgbTable, ULong crCbIndex, ULong yTopG, ULong yBottomG,
	     ULong dithTop, ULong dithBottom)
{
    long y;
    ULong yTopRB, yBottomRB;
    UChar *rBase, *gBase, *bBase;
    OctWord final;
    ULong rgbTriple;

    rgbTriple = rgbTable->crCbToRGB[ crCbIndex & 0xffff ];
    rBase = rgbTable->lookup232.red + RIndexFromTriple(rgbTriple);
    gBase = rgbTable->lookup232.green + GIndexFromTriple(rgbTriple);
    bBase = rgbTable->lookup232.blue + BIndexFromTriple(rgbTriple);

    /*
     * Dither four pixels at once!
     */
    yTopG = RGBDitherAdd(yTopG, dithTop);
    yTopRB = RGBDitherAdd(yTopG, dithTop);
    yBottomG = RGBDitherAdd(yBottomG, dithBottom);
    yBottomRB = RGBDitherAdd(yBottomG, dithBottom);

    /*
     * compute the four "a" pixels that share the same cr/cb
     */
    y = yTopRB>>24;		/* pixel 0a */
    final.l.high = rBase[y];
    final.l.high |= bBase[y];
    final.l.high |= gBase[yTopG>>24];
    final.l.high <<= 8;

    y = (yTopRB>>16)&0xff;	/* pixel 1a */
    final.l.high |= rBase[y];
    final.l.high |= bBase[y];
    final.l.high |= gBase[(yTopG>>16)&0xff];
    final.l.high <<= 8;

    y = yBottomRB>>24;		/* pixel 4a */
    final.l.low = rBase[y];
    final.l.low |= bBase[y];
    final.l.low |= gBase[yBottomG>>24];
    final.l.low <<= 8;

    y = (yBottomRB>>16)&0xff;	/* pixel 5a */
    final.l.low |= rBase[y];
    final.l.low |= bBase[y];
    final.l.low |= gBase[(yBottomG>>16)&0xff];
    final.l.low <<= 8;

    rgbTriple = rgbTable->crCbToRGB[ crCbIndex >> 16 ];
    rBase = rgbTable->lookup232.red + RIndexFromTriple(rgbTriple);
    gBase = rgbTable->lookup232.green + GIndexFromTriple(rgbTriple);
    bBase = rgbTable->lookup232.blue + BIndexFromTriple(rgbTriple);

    y = (yTopRB>>8)&0xff;	/* pixel 2b */
    final.l.high |= rBase[y];
    final.l.high |= bBase[y];
    final.l.high |= gBase[(yTopG>>8)&0xff];
    final.l.high <<= 8;

    y = yTopRB&0xff;		/* pixel 3b */
    final.l.high |= rBase[y];
    final.l.high |= bBase[y];
    final.l.high |= gBase[yTopG&0xff];

    y = (yBottomRB>>8)&0xff;	/* pixel 6b */
    final.l.low |= rBase[y];
    final.l.low |= bBase[y];
    final.l.low |= gBase[(yBottomG>>8)&0xff];
    final.l.low <<= 8;

    y = yBottomRB&0xff;		/* pixel 7b */
    final.l.low |= rBase[y];
    final.l.low |= bBase[y];
    final.l.low |= gBase[yBottomG&0xff];

    return(final);
}

/*ARGSUSED*/
void Lookup232RGBFrame(UChar *pRgb, long rgbWidth, UChar *pY, UChar *pCb,
		       GeomRectangle *src, GeomRectangle *dst)

    /* UChar *pRgb;			input and output */
    /* GeomRectangle *src;		XXX ignored */
    /* GeomRectangle *dst;		x and y already compensated for */
{
    long width = dst->width;
    long height = dst->height;
    OctWord rgb;
    ULong yTop, yBottom;
    ULong crCbIndex;
    ULong dithTop, dithBottom;
    RGBLookup *rgbTable = &RGBTable;
    long w, h;
    long chromaAccelerator = MAX_X - (width>>1);
    long lumaAccelerator = 2*MAX_X - width;
    long rgbAccelerator = 2*rgbWidth - width;

    /*
     * The initial value of h, below, must be even, because macroblocks
     * are 16=4x4 pixels tall.
     */
    h = height>>1;
    width >>= 3;
    for (;;) {
	if (h & 1) {
	    dithTop = G232DitherBits(D00,D01,D02,D03);
	    dithBottom = G232DitherBits(D10,D11,D12,D13);
	} else {
	    dithTop = G232DitherBits(D20,D21,D22,D23);
	    dithBottom = G232DitherBits(D30,D31,D32,D33);
	}
	/*
	 * Each time through the loop does 8x2 pixels
	 * . . . . . . . .
	 * . . . . . . . .
	 * The initial value of w, below, must be even, because macroblocks
	 * are 16=8x2 pixels wide.
	 */
	for (w = width; --w >= 0;) {
	    /*
	     * first 4x2 pixels:
	     * x x x x . . . .
	     * x x x x . . . .
	     */
	    crCbIndex = CbToCrAddr(pCb)[0] << 24;
	    crCbIndex |= pCb[0] << 16;
	    crCbIndex |= CbToCrAddr(pCb)[1] << 8;
	    crCbIndex |= pCb[1];

	    yBottom = *ULongP(pY + MAX_X);
	    yBottom = (yBottom >> 1) & 0x7f7f7f7f;
	    yTop = *ULongP(pY)++;
	    yTop = (yTop >> 1) & 0x7f7f7f7f;

	    rgb = Lookup232RGB(rgbTable, crCbIndex, yTop,
			       yBottom, dithTop, dithBottom);
	    *ULongP(pRgb + rgbWidth) = rgb.l.low;
	    *ULongP(pRgb)++ = rgb.l.high;

	    /*
	     * second 4x2 pixels
	     * . . . . x x x x
	     * . . . . x x x x
	     */
	    crCbIndex = CbToCrAddr(pCb)[2] << 24;
	    crCbIndex |= pCb[2] << 16;
	    crCbIndex |= CbToCrAddr(pCb)[3] << 8;
	    crCbIndex |= pCb[3];
	    pCb += 4;

	    yBottom = *ULongP(pY + MAX_X);
	    yBottom = (yBottom >> 1) & 0x7f7f7f7f;
	    yTop = *ULongP(pY)++;
	    yTop = (yTop >> 1) & 0x7f7f7f7f;
	    rgb = Lookup232RGB(rgbTable, crCbIndex, yTop,
			       yBottom, dithTop, dithBottom);
	    *ULongP(pRgb + rgbWidth) = rgb.l.low;
	    *ULongP(pRgb)++ = rgb.l.high;
	}

	if (--h <= 0)
	    break;

	pRgb += rgbAccelerator;
	pY += lumaAccelerator;
	pCb += chromaAccelerator;
    }
}

/*
 * Like Lookup232RGB() this converts y/cr/cb to 2:3:2 rgb.  The 2x2 pixels are
 * doubled on input to produce a 4x4 output square.  Since we are
 * producing an output square that is isometric to the dither pattern,
 * the dither need not be arguments, and the output buffer can be
 * passed as an argument.
 *
 * The input pixels look like
 * ---------
 * | 0 | 1 |
 * ---------
 * | 2 | 3 |
 * ---------
 * These are arranged in yValues like
 * -----------------
 * | 2 | 3 | 0 | 1 |
 * -----------------
 * These are dithered in four passes (a, b, c, d) to produce a 4x4
 * output square that looks like:
 * -----------------				------------------
 * | 0a| 0b| 1a| 1b|	which suggests the	|D00|D01|D02|D03|
 * -----------------	dither patterns be	------------------
 * | 0c| 0d| 1c| 1d|	applied like -->	|D10|D11|D12|D13|
 * -----------------				------------------
 * | 2a| 2b| 3a| 3b|				|D20|D21|D22|D23|
 * -----------------				------------------
 * | 2c| 2d| 3c| 3d|				|D30|D31|D32|D33|
 * -----------------				------------------
 */
static void
Lookup232RGBHalf(RGBLookup *rgbTable, ULong crCbIndex, ULong yValues, UChar *pRgb,
		 ULong rgbLength)
{
    long y;
    ULong yGa, yRBa;
    ULong yGb, yRBb;
    ULong yGc, yRBc;
    ULong yGd, yRBd;
    ULong dith;
    UChar *rBase, *gBase, *bBase;
    ULong rgb;
    ULong rgbTriple;

    rgbTriple = rgbTable->crCbToRGB[ crCbIndex ];
    rBase = rgbTable->lookup232.red + RIndexFromTriple(rgbTriple);
    gBase = rgbTable->lookup232.green + GIndexFromTriple(rgbTriple);
    bBase = rgbTable->lookup232.blue + BIndexFromTriple(rgbTriple);

    /*
     * Dither four pixels at once!
     */
    dith = G232DitherBits(D22,D20,D02,D00);
    yGa = RGBDitherAdd(yValues, dith);
    yRBa = RGBDitherAdd(yGa, dith);

    dith = G232DitherBits(D23,D21,D03,D01);
    yGb = RGBDitherAdd(yValues, dith);
    yRBb = RGBDitherAdd(yGb, dith);

    dith = G232DitherBits(D32,D30,D12,D10);
    yGc = RGBDitherAdd(yValues, dith);
    yRBc = RGBDitherAdd(yGc, dith);

    dith = G232DitherBits(D33,D31,D13,D11);
    yGd = RGBDitherAdd(yValues, dith);
    yRBd = RGBDitherAdd(yGd, dith);

    /*
     * Compute:
     * -----------------
     * | 0a| 0b| 1a| 1b|
     * -----------------
     */
    y = (yRBa>>8) & 0xff;	/* pixel 0a */
    rgb  = rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[(yGa>>8) & 0xff];
    rgb <<= 8;

    y = (yRBb>>8) & 0xff;	/* pixel 0b */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[(yGb>>8) & 0xff];
    rgb <<= 8;

    y = yRBa & 0xff;		/* pixel 1a */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[yGa & 0xff];
    rgb <<= 8;

    y = yRBb & 0xff;		/* pixel 1b */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[yGb & 0xff];

    *ULongP(pRgb) = rgb;
    pRgb += rgbLength;

    /*
     * Compute:
     * -----------------
     * | 0c| 0d| 1c| 1d|
     * -----------------
     */
    y = (yRBc>>8) & 0xff;	/* pixel 0c */
    rgb = rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[(yGc>>8) & 0xff];
    rgb <<= 8;

    y = (yRBd>>8) & 0xff;	/* pixel 0d */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[(yGd>>8) & 0xff];
    rgb <<= 8;

    y = yRBc & 0xff;		/* pixel 1c */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[yGc & 0xff];
    rgb <<= 8;

    y = yRBd & 0xff;		/* pixel 1d */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[yGd & 0xff];

    *ULongP(pRgb) = rgb;
    pRgb += rgbLength;

    /*
     * Compute:
     * -----------------
     * | 2a| 2b| 3a| 3b|
     * -----------------
     */
    y = yRBa>>24;			/* pixel 2a */
    rgb  = rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[yGa>>24];
    rgb <<= 8;

    y = yRBb>>24;			/* pixel 2b */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[yGb>>24];
    rgb <<= 8;

    y = (yRBa>>16) & 0xff;		/* pixel 3a */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[(yGa>>16) & 0xff];
    rgb <<= 8;

    y = (yRBb>>16) & 0xff;		/* pixel 3b */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[(yGb>>16) & 0xff];

    *ULongP(pRgb) = rgb;
    pRgb += rgbLength;

    /*
     * Compute:
     * -----------------
     * | 2c| 2d| 3c| 3d|
     * -----------------
     */
    y = yRBc>>24;			/* pixel 2c */
    rgb  = rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[yGc>>24];
    rgb <<= 8;

    y = yRBd>>24;			/* pixel 2d */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[yGd>>24];
    rgb <<= 8;

    y = (yRBc>>16) & 0xff;		/* pixel 3c */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[(yGc>>16) & 0xff];
    rgb <<= 8;

    y = (yRBd>>16) & 0xff;		/* pixel 3d */
    rgb |= rBase[y];
    rgb |= bBase[y];
    rgb |= gBase[(yGd>>16) & 0xff];

    *ULongP(pRgb) = rgb;
}

/*ARGSUSED*/
void Lookup232RGBFrameDouble(UChar *pRgb, long rgbWidth, UChar *pY, UChar *pCb,
			     GeomRectangle *src, GeomRectangle *dst)

    /* UChar *pRgb;			input and output */
    /* GeomRectangle *src;		XXX ignored */
    /* GeomRectangle *dst;		x and y already compensated for */
{
    long width = dst->width;
    long height = dst->height;
    ULong yValues;
    ULong crCbIndex;
    RGBLookup *rgbTable = &RGBTable;
    long w, h;
    long chromaAccelerator = MAX_X - (width>>2);
    long lumaAccelerator = 2*MAX_X - (width>>1);
    long rgbAccelerator = 4*rgbWidth - width;

    h = height>>2;
    width >>= 2;
    for (;;) {
	/*
	 * Each time through the loop takes 2x2 input pixels and
	 * produces 4x4 output pixels.
	 * The initial value of w, below, must be even, because macroblocks
	 * are 16=8x2 pixels wide.
	 */
	for (w = width; --w >= 0;) {
	    /*
	     * first 2x2 pixels:
	     * x x . .
	     * x x . .
	     */
	    crCbIndex  = *CbToCrAddr(pCb) << 8;
	    crCbIndex |= *pCb++;

	    yValues = *UShortP(pY+MAX_X) << 16;
	    yValues |= *UShortP(pY)++;
	    yValues = (yValues >> 1) & 0x7f7f7f7f;

	    Lookup232RGBHalf(rgbTable, crCbIndex, yValues, pRgb, rgbWidth);
	    ULongP(pRgb)++;
	}

	if (--h <= 0)
	    break;

	pRgb += rgbAccelerator;
	pY += lumaAccelerator;
	pCb += chromaAccelerator;
    }
}

/*ARGSUSED*/
void Lookup332RGBFrame(UChar *pRgb, long rgbWidth, UChar *pY, UChar *pCb,
		       GeomRectangle *src, GeomRectangle *dst)

    /* UChar *pRgb;		input and output */
    /* GeomRectangle *src;		XXX ignored */
    /* GeomRectangle *dst;		x and y already compensated for */
{
    NormalOutput("output 332 frame unimplemented\n");
}

/*ARGSUSED*/
void Lookup332RGBFrameDouble(UChar *pRgb, long rgbWidth, UChar *pY, UChar *pCb,
			     GeomRectangle *src, GeomRectangle *dst)

    /* UChar *pRgb;		input and output */
    /* GeomRectangle *src;	XXX ignored */
    /* GeomRectangle *dst;		x and y already compensated for */
{
    NormalOutput("output 332 double size frame unimplemented\n");
}

/*ARGSUSED*/
void Lookup888RGBFrame(UChar *pRgb, long rgbWidth, UChar *pY, UChar *pCb,
		       GeomRectangle *src, GeomRectangle *dst)

    /* UChar *pRgb;		input and output */
    /* GeomRectangle *src;	XXX ignored */
    /* GeomRectangle *dst;	x and y already compensated for */
{
    NormalOutput("output 888 frame unimplemented\n");
}

/*ARGSUSED*/
void Lookup888RGBFrameDouble(UChar *pRgb, long rgbWidth, UChar *pY, UChar *pCb,
			     GeomRectangle *src, GeomRectangle *dst)

    /* UChar *pRgb;		input and output */
    /* GeomRectangle *src;	XXX ignored */
    /* GeomRectangle *dst;		x and y already compensated for */
{
    NormalOutput("output 888 double size frame unimplemented\n");
}

#endif /* LOOKUP_RGB */

#ifdef CALCULATE_RGB

#include "rgbdef.h"

void Calculate232RGBFrame(UChar *pRgb, UChar *pY, UChar *pCb, GeomRectangle *src,
			  GeomRectangle *dst)

    /* UChar *pRgb;		input and output */
    /* GeomRectangle *src;	XXX ignored */
    /* GeomRectangle *dst;	x and y already compensated for */
{
    long width = dst->width;
    long height = dst->height;
    int wb;	/* width in blocks (8) */
    int hhb;	/* height in half blocks (4) */
    int rgbWidth;
    int w;
    int accelToNextChromaRow;
    int accelToNextLumaRow;
    int accelToNextRgbRow;

    wb = width >> 3;
    hhb = height >> 2;

    /*
     * The pointers have to be bumped down one row plus what's left
     * at the end of the current row (usually nothing).
     */
    accelToNextChromaRow = MAX_X - (wb << 2);
    accelToNextLumaRow = (MAX_X << 1) - (wb << 3);
    accelToNextRgbRow = (rgbWidth << 1) - (wb << 3);

    LanguageAssert(((ULong)pY & 0x3) == 0
		&& ((ULong)pCb & 0x3) == 0
		&& ((ULong)pRgb & 0x3) == 0);

    for (;;) {
	/*
	 * we do this in pairs of rows because of the changing dither
	 * constants.  This is the even pair.
	 */
	w = wb;
	while (w-- > 0) {
	    YUVToRGB_8x2(pY, pCb, pRgb, rgbWidth,
			 D00,D01,D02,D03,
			 D10,D11,D12,D13);
	}

	/* full width means chroma is already in place */
	pY   += accelToNextLumaRow;
	if (wb != MAX_X>>3) {
	    pCb  += accelToNextChromaRow;
	}

	/*
	 * This is the odd pair.
	 */
	w = wb;
	while (w-- > 0) {
	    YUVToRGB_8x2(pY, pCb, pRgb, rgbWidth,
			 D20,D21,D22,D23,
			 D30,D31,D32,D33);
	}

	if (--hhb <= 0)
	    break;
	pY   += accelToNextLumaRow;
	if (wb != MAX_X>>3) {
	    pCb  += accelToNextChromaRow;
	}
    }
}

#endif /* CALCULATE_RGB */

void RealYUVtoRGB(long y, long cr, long cb, float *pRed, float *pGreen, float *pBlue)
{
    float red, green, blue;

    red = YCrCbToRed(y,cr,cb);
    green = YCrCbToGreen(y,cr,cb);
    blue = YCrCbToBlue(y,cr,cb);

    if (red < 0)
	red = 0;
    else if (red > 255)
	red = 255;
    if (green < 0)
	green = 0;
    else if (green > 255)
	green = 255;
    if (blue < 0)
	blue = 0;
    else if (blue > 255)
	blue = 255;
    *pRed = red;
    *pGreen = green;
    *pBlue = blue;
}
