/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	sgiaifc.c - sgi-specific audio routines for mpeg decoder
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/sgiaifc.c,v 1.12 1994/03/29 20:25:26 toddb Exp $
 */

#include <fcntl.h>
#include <audio.h>
#include <audiofile.h>
#include <sys/schedctl.h>

#define NeedFunctionPrototypes 1

#include "vmpegdecomp.h"

/*
 * local subroutines
 */
static void DefaultAFerror(long, const char *);
static void QuietAFerror(long, const char *);

/*
 * globals 
 */
extern char *ProgramName;

static int	total_frames;
static int	samp_count;
static int	frame_count;
static double	sec_count;
static int	samples_read;
static boolean	moreToDo;
static boolean	rude = 1;

static AFfilehandle audioFileFp;
static ALport	audioDevFp;
static char	*filename;		/* input file name */
static long	filefmt;		/* input file format */
static int	fd;			/* input file descriptor */
static int	bytesPerSample;		/* sample width */
static long	bits_per_samp;		/* sample resolution */
static int	samps_per_frame;	/* frame size */
static int	fileInputRate;		/* sample rate */
static int	samplesPerBuf;		/* samples / sample buffer */
static int	frames_per_buf;		/* frames / sample buffer */
static int	samp_wordsize;    	/* size of sample (AL_SAMPLE_*) */
static double	secs_per_frame;		/* time to play one audio frame */
static double	secs_per_buf;		/* time to play one sample buffer */
static char	*sampleBuffer = NULL;	/* sample buffer */
static int	bufferSize;		/* bytes / sample buffer */
static long	compression;		/* audio data compression type */
static double	realFileInputRate;	/* audio file sample rate */

/* ******************************************************************
 *  QuietAFerror:	silent application-defined error reporting routine
 *			for Audio File Library
 * ****************************************************************** */
static void
QuietAFerror(long code, const char *desc)
{
} /* --------------------- end QuietAFerror() --------------- */

/* ******************************************************************
 *  DefaultAFerror:	default application-defined error reporting routine
 *			for Audio Library
 * ****************************************************************** */
static void
DefaultAFerror(long code, const char *desc)
{
    switch (code)
    {
#if 0
     case AF_BAD_CODEC_LICENSE:
        Msg( "license unavailable for Aware %s decoder\n",
                compression==AF_COMPRESSION_AWARE_MULTIRATE?"MultiRate":"MPEG");
        break;
#endif
     default:
        Msg( "(Audio File Library error %d): %s\n",
                code, desc);
        break;
    }
} /* --------------------- end DefaultAFerror() --------------- */

/* ******************************************************************
 *  GetInputRate:	return Audio Hardware input sampling rate
 * ****************************************************************** */
static int
GetInputRate()
{
    long buf[6];

    buf[0] = AL_INPUT_RATE;
    buf[2] = AL_INPUT_SOURCE;
    buf[4] = AL_DIGITAL_INPUT_RATE;
    ALgetparams(AL_DEFAULT_DEVICE, buf, 6);

    /* we are clocked off digital input. find real input rate, if possible. */
    if	((buf[1] == AL_RATE_AES_1)||(buf[3] == AL_INPUT_DIGITAL)) {
	if (ALgetdefault(AL_DEFAULT_DEVICE, AL_DIGITAL_INPUT_RATE) >= 0) 
	    return (buf[5]);
    }
    /* input rate is Hz and we're using an analog input */
    else if (buf[1] > 0) 
	return (buf[1]);

    return (AL_RATE_UNDEFINED);
} /* --------------------- end GetInputRate() --------------- */

/* ******************************************************************
 *  GetOutputRate:	return Audio Hardware output sampling rate
 * ****************************************************************** */
static int
GetOutputRate()
{
    long buf[4];

    buf[0] = AL_OUTPUT_RATE;
    buf[2] = AL_DIGITAL_INPUT_RATE;
    ALgetparams(AL_DEFAULT_DEVICE, buf, 4);

    /* output rate is in Hz -- return it */
    if (buf[1] > 0) 
	return (buf[1]);

    /* output rate is logical -- track down what it means */
    if (buf[1] == AL_RATE_AES_1) {
	/* we are clocked off of digital input. find
	 * real input rate, if system supports this ability.
	 * Otherwise, return AL_RATE_UNDEFINED */
	if (ALgetdefault(AL_DEFAULT_DEVICE,AL_DIGITAL_INPUT_RATE) >= 0) 
	    return (buf[3]);
    } else if (buf[1] == AL_RATE_INPUTRATE) 
	return (GetInputRate());
    return (AL_RATE_UNDEFINED);
} /* --------------------- end GetOutputRate() --------------- */

static void
SetHighPriority()
{
    static boolean firstTime = True;
    extern int setreuid(int, int);
    extern int setregid(int, int);
    extern int geteuid(), getuid();
    extern int getegid(), getgid();

    if (firstTime) {
	firstTime = False;
    } else
	return;

    /*
     * swap effective & real userIDs.
     * this allows us to become super-user when we wish,  but run
     * as "joe user" for the rest of the time.
     */
    setreuid(geteuid(), getuid());
    setregid(getegid(), getgid());
    
    /* 
     * set non-degrading priority
     * use high priority if running as root
     * would be better to use DEADLINE scheduling, but, hey, it's a start
     */
    
    /*
     * swap permissions so that we become root for a moment.
     */
    setreuid(geteuid(), getuid());
    setregid(getegid(), getgid());
    
#   ifdef THIS_IS_BROKEN_REVISIT_AFTER_5_1
    if (schedctl(NDPRI, 0, NDPNORMMAX)!=-1) 
    {
        /* we are root, use real exploitation */
        schedctl(SLICE, 0, CLK_TCK);  
    }
    else
    {
        DebugTask(N_AudioActivity,
	    ProgMsg("Run as root for higher process priority\n")
	)
 
        /* find lowest user non-degrading priority, ...
         * this value is one more than ndpri_hilim which can be set with 
         * systune(3), but cannot be determined by a system call, hence
         * the for() loop.
         */
        for (i = NDPLOMAX+1; schedctl(NDPRI, 0, i) != -1; i--) ;
    }
#   else
    if (schedctl(NDPRI, 0, NDPNORMMAX)< 0) {
        DebugTask(N_AudioActivity,
	    ProgMsg( "Run as root for higher process priority\n")
	)
    }
#   endif
    
    /*
     * swap permissions back, now we're just "joe user"
     */
    setreuid(geteuid(), getuid());
    setregid(getegid(), getgid());
    
}

static void
CloseAiffFile(DisplayInfoPtr dpyInfo)
{
    if (audioFileFp) {
	AFclosefile(audioFileFp);
	audioFileFp = NULL;
    }
}

static boolean
OpenAiffFile(DisplayInfoPtr dpyInfo)
{
    long samp_type;
    long vers;
    int newBufferSize;
    char compressionname[10];
    
    SetHighPriority();
    CloseAiffFile(dpyInfo);

    ALseterrorhandler(0);
    AFseterrorhandler(DefaultAFerror);
    DebugTask(N_AudioActivity, AFseterrorhandler(QuietAFerror))
    
    filename = dpyInfo->audioInput;
    
    /*
     * obtain a unix file descriptor for the input file
     */
    if ((fd = open(filename, O_RDONLY)) < 0) {
	Error("failed to open file '%s' %s\n", filename);
	return(False);
    }
    /*
     * test the file descriptor to see whether we can attach an audio
     *    file handle to it
     */
    if (AFidentifyfd(fd) < 0) {
	ProgMsg("'%s' not an AIFF-C or AIFF file\n", filename);
	return(False);
    }
    /*
     * attach an audio file handle to the file descriptor
     */
    audioFileFp = AFopenfd(fd, "r", AF_NULL_FILESETUP);
    if (audioFileFp == AF_NULL_FILEHANDLE) {
	DebugTask(N_AudioActivity,
	    ProgMsg("failed to open file '%s'\n", filename)
	)
	return(False);
    }

    samps_per_frame   = AFgetchannels(audioFileFp, AF_DEFAULT_TRACK);
    realFileInputRate = AFgetrate(audioFileFp, AF_DEFAULT_TRACK);
    compression       = AFgetcompression(audioFileFp, AF_DEFAULT_TRACK);
    filefmt           = AFgetfilefmt(audioFileFp, &vers);
    
    fileInputRate = (long) realFileInputRate;
    total_frames    =  AFgetframecnt(audioFileFp, AF_DEFAULT_TRACK);
    moreToDo    = True;
    samp_count  = 0;
    frame_count = 0; 
    sec_count   = 0.0;
    samples_read = 0;
    
    AFgetsampfmt(audioFileFp, AF_DEFAULT_TRACK, &samp_type, &bits_per_samp);
    
    /*
     * decide what size blocks of samples we should read from the
     * input file and pass to ALwritesamps
     */
    if (bits_per_samp <= 8)
    {
        bytesPerSample = 1;
        samp_wordsize  = AL_SAMPLE_8;
    }
    else if (bits_per_samp <= 16)
    {
        bytesPerSample = 2;
        samp_wordsize  = AL_SAMPLE_16;
    }
    else if (bits_per_samp <= 24)
    {
        bytesPerSample = 4;
        samp_wordsize  = AL_SAMPLE_24;
    }
    else
    {
	DebugTask(N_AudioActivity,
            ProgMsg("%s: error can't play %d bit samples\n",
		filename, bits_per_samp)
        )
        return(False);
    }
    
    if ((samps_per_frame != 1)&&(samps_per_frame!= 2)&&(samps_per_frame!=4))
    {
	DebugTask(N_AudioActivity,
            ProgMsg("%s: error can't play %d channel sample data\n",
		filename, samps_per_frame)
        )
        return(False);
    }
    
    /*
     * make the buffer large enough to hold 1/2 sec of audio frames
     * we add one to fileInputRate before we divide in order to
     * correctly handle the 11025 case
     */
    secs_per_frame = 1.0 / ((double)fileInputRate);
    frames_per_buf = (fileInputRate+1)/2;
    
    samplesPerBuf = frames_per_buf * samps_per_frame;
    secs_per_buf  = secs_per_frame * frames_per_buf;
    newBufferSize = samplesPerBuf * bytesPerSample;
    
    if (newBufferSize > bufferSize) {
	if (sampleBuffer)
	    free(sampleBuffer);
	sampleBuffer = InsecureAlloc(newBufferSize);
	if (sampleBuffer == NULL) {
	    ProgMsg("Can't alloc %d-byte audio sample buffer\n", newBufferSize);
	    return(False);
	}
	bufferSize = newBufferSize;
    }
    
    switch (compression)
    {
    default:
    case AF_COMPRESSION_NONE: 
	strcpy(compressionname, "");
	break;
    case AF_COMPRESSION_G722:
	strcpy(compressionname, "G.722 -->");
	break;
    case AF_COMPRESSION_G711_ALAW:
	strcpy(compressionname, "A-law -->");
	break;
    case AF_COMPRESSION_G711_ULAW:
	strcpy(compressionname, "u-law -->");
	break;
#ifdef AWARE
    case AF_COMPRESSION_AWARE_MPEG:
	strcpy(compressionname, "Aware MPEG -->");
	break;
    case AF_COMPRESSION_AWARE_MULTIRATE:
	strcpy(compressionname, "Aware MultiRate -->");
	break;
#endif                          /*AWARE*/
    }
    DebugTask(N_AudioActivity,
	float filePlayingTime;
	filePlayingTime = (float)total_frames / (float)fileInputRate;
        ProgMsg("'%s':\n    %6.3f sec %7.1f Hz %6s %s %d-bit %s\n",
               filename,
               filePlayingTime,
               realFileInputRate,
               samps_per_frame == 1 ? "mono" : 
               (samps_per_frame == 2 ? "stereo" : "4-channel"),
               compressionname,
               bits_per_samp,
               filefmt == AF_FILE_AIFFC ? "AIFF-C" : "AIFF"
               );
    )
    return(True);
} /* --------------------- end main() --------------- */


static void
SgiAudioClose(DisplayInfoPtr dpyInfo)
{
    if (audioDevFp) {
	ALcloseport(audioDevFp);
	audioDevFp = NULL;
    }
}

/* ******************************************************************
 *  SgiAudioOpen():	initialize audio port and global state of IRIS Audio 
 *			Hardware
 * ****************************************************************** */

static boolean
SgiAudioOpen(DisplayInfoPtr dpyInfo)
{
    ALconfig audioDevFp_config;
    long pvbuf[4];
    long devOutputRate;
    int err;
    
    SgiAudioClose(dpyInfo);

    /*
     * need to determine whether audio is in use. if not, then we
     * can just go ahead and be "rude."
     */
    pvbuf[0] = AL_OUTPUT_COUNT;
    pvbuf[2] = AL_MONITOR_CTL;
    if (ALgetparams(AL_DEFAULT_DEVICE, pvbuf, 4) < 0) {
	if (oserror() == AL_BAD_DEVICE_ACCESS) {
	    ProgMsg("Can't play -- could not access audio hardware.\n");
	    return(False);
	}
    }
    
    
    if ((pvbuf[1] == 0) && (pvbuf[3] == AL_MONITOR_OFF)) {
        rude = 1;
    }
    
    /* 
     * determine the output rate for the audio hardware
     */
    devOutputRate = GetOutputRate();
    
    /*
     * if the rates are the same, all is well. if not, then we need to proceed
     * in a either a "rude" or "polite" manner.
     */
    
    if (devOutputRate != fileInputRate)
    {
        /*
         *  we set the rate to our value in Hz.  ALsetparams will set
	 *  the hardware to the closest rate to fileInputRate which the
	 *  hardware supports.
         */
	
        if (rude)
        {
            devOutputRate = fileInputRate;
            pvbuf[0] = AL_OUTPUT_RATE;
            pvbuf[1] = devOutputRate;
            ALsetparams(AL_DEFAULT_DEVICE, pvbuf, 2);
        }
        else                    /*polite*/
        {
	    DebugTask(N_AudioActivity,
		ProgMsg("'%s': Adjust output rate ", filename);
                if (devOutputRate < 0) 
		    Msg("[currently set to undetermined digital rate]\n");
                else 
		    Msg("[currently %d Hz]\n", devOutputRate);
            )
        }
    }

    /*
     * configure and open audio port
     */
    audioDevFp_config = ALnewconfig();
    ALsetwidth(audioDevFp_config, samp_wordsize);
    ALsetchannels(audioDevFp_config, samps_per_frame);
    
    /*
     * make the ring buffer large enough to hold 1 sec of audio samples
     */
    ALsetqueuesize(audioDevFp_config, samplesPerBuf*2);
    audioDevFp = ALopenport(ProgramName, "w", audioDevFp_config);
    if (audioDevFp == 0) {
	err = oserror();
	if (err == AL_BAD_NO_PORTS) {
	    ProgMsg("Can't play -- no audio ports available at the moment\n");
	} else if (err == AL_BAD_OUT_OF_MEM) {
	    ProgMsg("Can't play -- not enough memory to open audio port\n");
	}
	return(False);
    }

    DebugTask(N_AudioActivity,
        ProgMsg("audio dev: samp w=%d, samp/frame=%d, q sz=%d\n",
		samp_wordsize, samps_per_frame,
		ALgetqueuesize(audioDevFp_config));
    )
    
    /* Attempt to remove initial delay by writing 0 samples */
    ALwritesamps(audioDevFp, sampleBuffer, 0);
    return(True);
} /* --------------------- end SgiAudioOpen() --------------- */


/*
**      ->timeGather: Microseconds of audio requested to be decoded
**	and ready to play on "mark". On return ->timeGather is
**	filled with the Microseconds actually prepared.
**	0 => EOF, <0 => error.
*/

/*
** Globals:
**	samples_read - # samples read in current or past PreloadAiff()
*/
static boolean
PreloadAiff(DisplayInfoPtr dpyInfo)
{
    AudioInfoPtr aData = &dpyInfo->audioData;
    int num_bufs;
    int framesIn, fileOffset;
    int framesToPlay, samplesToPlay, sampleBytesToPlay;
    
    if( aData->timeSeek >= 0 ) {

	fileOffset = aData->timeSeek * (fileInputRate / (float)MicroSecPerSec);
	AFseekframe(audioFileFp, AF_DEFAULT_TRACK, fileOffset);
        DebugTask(N_AudioActivity,
	    ProgMsg("sgi preload au: seek to %.2fsecs (%d*%d=%d offset)\n",
		aData->timeSeek / (float)MicroSecPerSec,
		fileOffset, bytesPerSample,
		fileOffset * bytesPerSample);
	)
	aData->timeSeek = -1;
    } else if( samples_read > 0 ) {
	/* we're ahead of ourselves */
	return(False);
    }
    framesToPlay = aData->timeGather * (fileInputRate / (float)MicroSecPerSec);
    samplesToPlay = framesToPlay * samps_per_frame;
    sampleBytesToPlay = samplesToPlay * bytesPerSample;
    num_bufs = sampleBytesToPlay / bufferSize;
    
    if( sampleBytesToPlay % bufferSize ) {
	num_bufs++;
    }
    if( num_bufs > 1 ) {
	DebugTask(N_AudioActivity, ProgMsg("audio buffer overflow\n"))
	return(False);
    }
    
    /*
     * note that there may be some pad bytes following the valid samples -
     * for example, the sample data area may be padded so that the valid 
     * samples begin on a block boundary and the sample area ends on a block
     * boundary (where blocksize is specified by the user)
     */
    framesIn = AFreadframes(audioFileFp, AF_DEFAULT_TRACK, 
			       sampleBuffer, framesToPlay);
    if( framesIn < framesToPlay)
    {
	if (num_bufs > 0)
	{
	    DebugTask(N_AudioActivity,
		ProgMsg("short read for %s: expected %d frames, got %d\n",
		    filename, frames_per_buf, framesIn)
	    )
	}
	moreToDo = False;
    }
    samples_read = framesIn * samps_per_frame;
    
    samp_count  += samples_read;
    frame_count += framesIn;
    sec_count   += framesIn * secs_per_frame;
    
    DebugTask(N_AudioActivity,
	Msg("        %d sample frames  %6.3f secs\n", frame_count, sec_count) 
    )

    /*
     ** return the time delta to caller. >0 => underwrite
     **	                                     <0 => overwrite
     */
    aData->timeGather = (int)(1000000.0 * (samples_read / samps_per_frame) /
		       fileInputRate);
    return(True);
}

/*
** Globals used:
**	moreToDo - declares that a short read was encountered signifying EOF
**	samples_read - # of samples read in PreloadAiff()
*/
static boolean
SgiAudioUpdate(DisplayInfoPtr dpyInfo)
{
    if( samples_read > 0 ) {
	ALwritesamps(audioDevFp, sampleBuffer, samples_read);
    }
    samples_read = 0;
    return(moreToDo);
}

AudioFuncs AiffFuncs = {
    OpenAiffFile,
    CloseAiffFile,
    PreloadAiff,

    SgiAudioOpen,
    SgiAudioClose,
    SgiAudioUpdate
};
