/*
 * Copyright (c) 1992,1993,1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	util.c - various utilities for mpeg decoding
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/util.c,v 4.27 1995/08/18 03:42:09 toddb Exp $
 */

#define NeedFunctionPrototypes 1

#include "vmpegdecomp.h"
#include "huffman.h"
#include "motion.h"
#include "NvrLmFeatures.h"

static void InitDebug(SequenceInfoPtr, int, char **);

typedef union {
    UShort *shortValue;
    UChar *charValue;
} UValue, *UValuePtr;

static int HuffCodeMapCompare(HuffCodeMap *p1, HuffCodeMap *p2)
{
    return(p1->code - p2->code);
}

/*
 * copy global data into the sequence header & macroblock for optimized access.
 */
void
InitStructures(SequenceInfoPtr pSeq)
{
    memcpy(pSeq->macroBlk->zigZag, ZigZag, sizeof(ZigZag));
}

/*
 * prepare a table that can be addressed with MBAI_LOOKUP_BITS bits
 * to determine the bits consumed and the address increment.
 */
void InitLookupTable(HuffCodeMap initData[],
		     int nInit,
		     char *ptr,
		     int nLookupCells,
		     int nLookupBits,
		     int lookupSize)
{
    int i, j, next, start;
    ULong value;
    UValue lookupTable;

    qsort(initData, nInit, sizeof(HuffCodeMap), HuffCodeMapCompare);

    switch (lookupSize) {
    case sizeof(UChar):	lookupTable.charValue  = (UChar *) ptr; break;
    case sizeof(UShort):lookupTable.shortValue = (UShort *)ptr; break;
    }

    start = 0;
    value = 0;
    for (i = 0; i < nInit; i++) {
	next = initData[ i ].code;
	InitAssert(start >= 0);
	InitAssert(next == 0 || start < next);
	InitAssert(next < nLookupCells);
	for (j = start; j < next; j++) {
	    switch (lookupSize) {
	    case sizeof(UChar):	lookupTable.charValue[ j ]  = value; break;
	    case sizeof(UShort):lookupTable.shortValue[ j ] = value; break;
	    }
	}
	start = next;
	value = initData[ i ].value;
    }

    for (j = start; j < nLookupCells; j++) {
	switch (lookupSize) {
	case sizeof(UChar): lookupTable.charValue[ j ]  = value; break;
	case sizeof(UShort):lookupTable.shortValue[ j ] = value; break;
	}
    }
}

/*
 * create a table whose index is a macroblock address and where the indexed
 * value is a pixel offset into the y and cb buffers.  The cr offset
 * is handled in the MBCbToCrAddr() macro.  Note that the pixel offsets
 * that appear in the table must be multiplied by 8 (the height of
 * a chroma block) or 16 (luma block height).
 * The goals are:
 *  - keep the table small
 *  - make it usable for both luma and chroma offsets
 *  - perform the worst part of the arithmetic
 */
void InitMBAddressLookup(SequenceInfoPtr pSeq)
{
    MacroBlockInfoPtr pMb = pSeq->macroBlk;
    int mbAddr;

    /*
     * First, build the macroBlock-address-to-pixel-address map.
     */
    mbAddr = 0;
    for (mbAddr = 0; mbAddr < MAX_MB_PER_ROW*MAX_MB_PER_COL; mbAddr++) {
	/*
	 * Note that the pixel offset would be calculated by multiplying
	 * both parts of this sum by 8 or 16 (chroma or luma size).
	 */
	pMb->mbAddrMap[ mbAddr ] = (mbAddr/pSeq->mbWidth)*MAX_X
			    + mbAddr%pSeq->mbWidth;
    }
}

/*
 * prepare a table that can be addressed with MBAI_LOOKUP_BITS bits
 * to determine the bits consumed and the address increment.
 */
void InitMBAddressInc(void)
{
    InitLookupTable(MBAI_InitData, MBAI_INIT_CELLS,
		    (char *)HuffLookup.mbAddressInc,
		    MBAI_LOOKUP_CELLS, MBAI_LOOKUP_BITS,
		    sizeof(HuffLookup.mbAddressInc[0]));
}

/*
 * prepare a table that can be addressed with MBT_B_LOOKUP_BITS bits
 * to determine the bits consumed and the address increment.
 */
void InitMBType(void)
{
    /*
     * ... for B pictures...
     */
    InitLookupTable(MBT_B_InitData, MBT_B_INIT_CELLS,
		    (char *)HuffLookup.mb_B_Type,
		    MBT_B_LOOKUP_CELLS, MBT_B_LOOKUP_BITS,
		    sizeof(HuffLookup.mb_B_Type[0]));

    /*
     * ... for P pictures...
     */
    InitLookupTable(MBT_P_InitData, MBT_P_INIT_CELLS,
		    (char *)HuffLookup.mb_P_Type,
		    MBT_P_LOOKUP_CELLS, MBT_P_LOOKUP_BITS,
		    sizeof(HuffLookup.mb_P_Type[0]));
}

/*
 * prepare a table that can be addressed with MOTION_LOOKUP_BITS bits
 * to determine the bits consumed and the motion vector.
 */
void InitMotionVector(void)
{
    InitLookupTable(MotionVectorInitData, MOTION_INIT_CELLS,
		    (char *)HuffLookup.motionVector,
		    MOTION_LOOKUP_CELLS, MOTION_LOOKUP_BITS,
		    sizeof(HuffLookup.motionVector[0]));
}

/*
 * prepare a table that can be addressed with CODED_BLOCK_LOOKUP_BITS bits
 * to determine the bits consumed and the coded block value
 */
void InitCodedBlock(void)
{
    InitLookupTable(CodedBlockInitData, CODED_BLOCK_INIT_CELLS,
		    (char *)HuffLookup.codedBlock,
		    CODED_BLOCK_LOOKUP_CELLS, CODED_BLOCK_LOOKUP_BITS,
		    sizeof(HuffLookup.codedBlock[0]));
}

/*
 * prepare a tables that can be addressed with SIZE_LUMA_LOOKUP_CELLS bits
 * and SIZE_CHROMA_LOOKUP_CELLS bits to determine the bits consumed
 * and the size luma and chroma.
 */
void InitSizeYC(void)
{
    InitLookupTable(SizeLuminanceInitData, SIZE_YC_INIT_CELLS,
		    (char *)HuffLookup.sizeLuminance,
		    SIZE_LUMA_LOOKUP_CELLS, SIZE_LUMA_LOOKUP_BITS,
		    sizeof(HuffLookup.sizeLuminance[0]));
    InitLookupTable(SizeChrominanceInitData, SIZE_YC_INIT_CELLS,
		    (char *)HuffLookup.sizeChrominance,
		    SIZE_CHROMA_LOOKUP_CELLS, SIZE_CHROMA_LOOKUP_BITS,
		    sizeof(HuffLookup.sizeChrominance[0]));
}

/*
 * prepare a tables that can be addressed with DCT_COEFF1_LOOKUP_CELLS bits
 * and DCT_COEFF2_LOOKUP_CELLS bits to determine the bits consumed
 * and the dct coefficients.
 */
void InitDCTCoeff(void)
{
    InitLookupTable(DCTCoeff1InitData, DCT_COEFF1_INIT_CELLS,
		    (char *)HuffLookup.dctCoeff1,
		    DCT_COEFF1_LOOKUP_CELLS, DCT_COEFF1_LOOKUP_BITS,
		    sizeof(HuffLookup.dctCoeff1[0]));
    InitLookupTable(DCTCoeff2InitData, DCT_COEFF2_INIT_CELLS,
		    (char *)HuffLookup.dctCoeff2,
		    DCT_COEFF2_LOOKUP_CELLS, DCT_COEFF2_LOOKUP_BITS,
		    sizeof(HuffLookup.dctCoeff2[0]));
}

/*
 * Prepare tables indexed by signed numbers in the range -255 to 511 and
 * unsigned numbers in the range 0 to 511.  The indexed cell will
 * yield a value clamped to -128 to 127 and 0 to 256, respectively.
 * XXX we are assuming that 8-bits signed are enough to hold any
 * difference block generated
 */
void InitDCTClampLookup(void)
{
    long intValue;

    SignedClamp = (UChar *)SecureAlloc(SIGNED_CLAMP_SIZE);
    SignedClamp += 255;	/* this makes negative indices work */

    for (intValue = -255; intValue <= 256+255; intValue++) {
	if (intValue < 0)
	    SignedClamp[ intValue ] = 0; 
	else if (intValue > 255)
	    SignedClamp[ intValue ] = 255; 
	else
	    SignedClamp[ intValue ] = intValue; 
    }
}

static void LicenseExceptionCB(int condition, char *feature);

static FeatureName Features[] =
    { NVRLM_MPEGDECOMP, NVRLM_ALL_SOFTWARE };
static LmConfig LmConfigData = {
    Features,
    False,       /* heartbeat handled by HeartbeatTimer() */
    True,       /* shared */
    LicenseExceptionCB
};

/* ARGSUSED */
static void
LicenseExceptionCB(int condition,
                   char *feature)
{
    NvrLicenseError(condition, &LmConfigData, True);
}

void
AnnounceCopyright(SequenceInfoPtr pSeq)
{
    if (pSeq->pParse->outHandleIsTty) {
	NvrCopyrightMsg(  NvrCopyright
			| LmCopyright
			| TekCopyright
			| CopyrightToScreen,
			0);
    }
}

static int
HeartbeatTimer(void)
{
    NvrLmHeartBeat(LmConfigData.featureUsed);
    return(0);
}

static void
CheckLicense(SequenceInfoPtr pSeq)
{
    TimeSpec now;
    int lm_code;

    TimeNow(&now);
    lm_code = NvrGetLicense(&LmConfigData);
    if (lm_code != LM_OK
     || (TIME_TO_DIE && now.seconds > TIME_TO_DIE)) {
        NvrLicenseError(lm_code, &LmConfigData, False);

	/*
	 * Here we should fill main() with noop instructructions so that
	 * if someone finds the TimeBombDetonate() routine and disables
	 * it, its too late.  Instead, we try to subtly screw up the
	 * data in pSeq
	 */
	memcpy((char *)pSeq, ((char *)pSeq) + 1, sizeof(SequenceInfo)-1);
	longjmp(pSeq->licenseContext, 1);
    }

    if (TIME_TO_DIE == 0)
	SetFrameTimer(TICKS_PER_HEART_BEAT, HeartbeatTimer);
}

static void
AnnounceTimeBomb(void)
{
    TimeSpec now;

    if (TIME_TO_DIE == 0)
	return;
    TimeNow(&now);
    if (now.seconds < TIME_TO_DIE) {
	ProgMsg("This sofware will expire in %.1f days\n",
	    (TIME_TO_DIE-now.seconds)/(60*60*24.0));
    }
}

void
InitGeneral(SequenceInfoPtr pSeq, int argc, char **argv)
{
    AnnounceTimeBomb();
    InitDebug(pSeq, argc, argv);
    InitStructures(pSeq);
    InitMBAddressInc();
    InitMBType();
    InitMotionVector();
    InitCodedBlock();
    InitSizeYC();
    CheckLicense(pSeq);	/* random placement */
    InitDCTCoeff();
    InitDisplay(pSeq->dpyInfo);
    InitDCTClampLookup();
}

#ifdef DEBUG_ENABLED

boolean
VmpegDebugFormat(char *buf, int whatIsIt, ULong value)
{
    char *p;
    PicCodes picCodes;
    int labelLen;
    TimeCode tc;

    labelLen = strlen(buf);
    p = buf + labelLen;
    switch(whatIsIt) {
    case N_Timecode:
	tc.composite = value;
        sprintf(p, "%s (gop tc=%d/%02d:%02d:%02d)\n",
	    BitString(value, 25),
            tc.discrete.groupOfPictures,
            tc.discrete.minutes,
            tc.discrete.seconds,
            tc.discrete.pictures);
	return(True);
    case N_GopInfo:
        sprintf(p, "%s (closed gop=%d, broken link=%d)\n",
	    BitString(value, 2), value >> 1, value & 1);
	return(True);
    case N_PictureStart:
	picCodes.composite = value;
        sprintf(p, "%s (start code=0x%x)\n",
	    BitString(picCodes.s.code, 8), picCodes.s.code);
	p += strlen(p);
        sprintf(p, "%*s%s (timeref=%d)\n",
	    labelLen, "",
	    BitString(picCodes.s.timeRef, 10), picCodes.s.timeRef);
	p += strlen(p);
        sprintf(p, "%*s%s (type=%c)\n",
	    labelLen, "",
	    BitString(picCodes.s.type, 10), IPBDchars[ picCodes.s.type ]);

	return(True);

    default:
	return(False);
    }
}

static void
InitDebug(SequenceInfoPtr pSeq, int argc, char **argv)
{
    extern char *getenv(char *);
    char *logfile;

    SetDebugFileFlags(DebugFlags, DebugFlagDesc, "mpegdecomp.debug");

    if (logfile = getenv("_VMPEGDECOMP_LOGFILE"))
	(void)SetLogfile(logfile);

    DebugTask(N_CommandLine,
	int i;
	for (i = 0; i < argc; i++)
	    Msg("%s ", argv[i]);
	Msg("\n");
    )
}

void
PrintQMatrix(UChar *matrix)
{
    int i, j;

    for (i = 0; i < 8; i++) {
	for (j = 0; j < 8; j++)
	    NormalOutput("%3d", matrix[i*8+j]);
	NormalOutput("\n");
    }
}

void
PrintPix(UChar *pixels)
{
    int i, line, col;

    for (line = i = 0; i < 8; i++) {
	for (col = 0; col < 8; col++) {
	    NormalOutput("%s", col > 0 ? "," : "");
	    NormalOutput(HexPixels ?"%4x" : "%4d", pixels[ line + col ]);
	}
	NormalOutput("\n");
	line += MAX_X;
    }
}

void
PrintMbPix(MacroBlockInfoPtr pMb, int mbAddr, int mask)
{
    UChar *pixels;
    int i;
    register PictureInfo *ppic = pMb->pSeq->currentPict;

    mask &= (MB_Y0|MB_Y1|MB_Y2|MB_Y3|MB_CB|MB_CR
	    |MB_PIXELS0|MB_PIXELS1|MB_PIXELS2|MB_PIXELS3);
    for (i = 9; mask; i--) {
	switch ((1<<i) & mask) {
	case MB_Y0: pixels = &ppic->y[MBAddrToLuma(Y0, mbAddr)]; break;
	case MB_Y1: pixels = &ppic->y[MBAddrToLuma(Y1, mbAddr)]; break;
	case MB_Y2: pixels = &ppic->y[MBAddrToLuma(Y2, mbAddr)]; break;
	case MB_Y3: pixels = &ppic->y[MBAddrToLuma(Y3, mbAddr)]; break;
	case MB_CB: pixels = &ppic->cb[MBAddrToCb(mbAddr)]; break;
	case MB_CR: pixels = CbToCrAddr(&ppic->cb[MBAddrToCb(mbAddr)]); break;
	default: continue;
	}
	mask &= ~ (1<<i);

	NormalOutput("%s (0x%x):\n", BlockNames[i], pixels);
	PrintPix(pixels);
    }
}

#include "idct.h"

void
PrintMb(MacroBlockInfoPtr pMb,
	int mask,
	ULong cellsInBlock1,
	ULong cellsInBlock2)
{
    int bit;
    long *block;
    char *number;
    int i, j;

    mask &= (MB_Y0|MB_Y1|MB_Y2|MB_Y3|MB_CB|MB_CR);
    for (bit = 0; mask; bit++) {
	switch ((1<<bit) & mask) {
	case MB_Y0:
	case MB_Y1:
	case MB_Y2:
	case MB_Y3:
	case MB_CB:
	case MB_CR:
	    block = pMb->dctCoeff; break;
	default: continue;
	}
	mask &= ~ (1<<bit);

	NormalOutput("%s (coeff):\n", BlockNames[ bit ]);
	for (i = 0; i < 8; i++) {
	    for (j = 0; j < 8; j++) {
		if (cellsInBlock2 & (1<<j))
		    number =  FixedPointToA(block[ (i<<3) + j ], FIXED_PT_BITS);
		else
		    number =  FixedPointToA(0, FIXED_PT_BITS);
		NormalOutput("%s%7s", j > 0 ? "," : "", number);
	    }
	    ShiftCellMarks(cellsInBlock1, cellsInBlock2);
	    NormalOutput("\n");
	}
    }
}

# if defined(unix) || defined(_HPUX_SOURCE)
# include "process.h"

void
ShowPix(UChar *pix, int w, int h, int stride, char *name)
{
    Argvec args;
    OneWayPipe out;
    UChar *p;

    args.length = 0;
    args.path = "display";
    AddArgv(&args, "display", False);
    AddArgv(&args, "-title", False);
    AddArgv(&args, name, False);
    AddArgv(&args, "-geometry", False);
    AddFmtArgv(&args, "%dx%d", w, h);
    AddArgv(&args, "gray:-", False);

    out = PipeTo(&args, NULL);

    p = pix;
    while (h-- > 0) {
        (void) WriteData(out.outFd, p, w);
        p += stride;
    }
    CloseFile(out.outFd);
    FreeArgv(&args);
}
#endif /* unix || _HPUX_SOURCE */

char *MacroBlockTypeString(int flags)
{
    static char string[ 100 ];

    string[0] = '\0';
    if (flags & MBT_QUANT)	strcat(string, "quant,");
    if (flags & MBT_MFWD)	strcat(string, "fwd,");
    if (flags & MBT_MBAK)	strcat(string, "back,");
    if (flags & MBT_PATTERN)	strcat(string, "pattern,");
    if (flags & MBT_INTRA)	strcat(string, "intra,");
    string[ strlen(string)-1 ] = '\0'; /* remove trailing , */

    return (string);
}

/* Used for parameterized breakpoints */
void BreakHere(void) {}

#else /* DEBUG_ENABLED */

static void
InitDebug(SequenceInfoPtr pSeq, int argc, char **argv)
{
}

#endif /* DEBUG_ENABLED */
