/*
 * Copyright (c) 1993  North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research, Inc. has obtained a licensing right, and is
 * considered by North Valley Research, Inc. to be confidential.
 * It is protected by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 * 	NvrLm.c - platform independent license management stuff.
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/NvrLm.c,v 1.27 1994/12/17 13:03:10 russ Exp $
 */
#define NeedFunctionPrototypes 1

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <elm.h>
#include "process.h"
#include <signal.h>
#include <sys/signal.h>
#include "fileAccess.h"
#include "nvrAlloc.h"
#include "nvrString.h"
#include "osio.h"
#ifdef Solaris2
# include <sys/systeminfo.h>
#endif


/*
 *      EXTERNAL INCLUDES
 *              Include files that must be exported to any package or
 *              program using this package.
 */
#include <NvrLmFeatures.h>

/*
 *      GLOBALS
 */
int NvrLmInitialized = 0;

static char default_keydir[] = DEFAULT_KEYDIR;

static boolean
AnyKeyFound(char *keydir, LmConfigPtr config)
{
    int i;
    char keyFile[ MAXNAMLEN ];
    char hostname[ 256 ];
    char rightHost[ 256 ] = "<unknown>";
    struct elm_keyentry *keyentry;
    boolean keysExist = False;

#ifdef Solaris2
    if (sysinfo(SI_HOSTNAME, hostname, sizeof(hostname)) < 0)
#else
    if (gethostname(hostname, sizeof(hostname)) < 0)
#endif
	Fatal("Can't get the hostname for this machine!\n");

    for (i = 0; config->features[i][0]; i++) {
	sprintf(keyFile, "%s/%s", keydir, config->features[i]);
	if ((FileType(keyFile) & FILE_REGULAR) == 0)
	    continue;
	if ((keyentry = elm_getkeyfile(keyFile)) == NULL) {
	    ProgMsg("can't read key file \"%s\"\n", keyFile);
	    continue;
	}
	keysExist = True;
	for (i = 0; i < keyentry->numservers; i++) {
	    if (strcmp(keyentry->servers[i], hostname) == 0)
		return(True);
	    strcpy(rightHost, keyentry->servers[i]);
	}
    }
    if (keysExist) {
	ProgMsg("An appropriate key is present in %s,\n", keydir);
	ProgMsg("but the license server must run on \"%s\",\n", rightHost);
	ProgMsg("not here on \"%s\".  Please go to \"%s\" and\n",
	    hostname, rightHost);
	ProgMsg("type the command \"startnvrlmd\".  This will start the\n");
	ProgMsg("license server correctly\n");
	NvrLmExit(1);
    }
    return(False);
}

static char *
KeyFileList(LmConfigPtr config)
{
    int i;
    static char keyList[ MAXNAMLEN ];
    char *p;

    p = keyList;
    for (i = 0; config->features[i][0]; i++) {
        sprintf(p, "%s%s", i > 0 ? ", " : "", config->features[i]);
	p += strlen(p);
    }
    *p = '\0';
    return(keyList);
}

static char *
FindKeyDir(char *keydir, char *nvrhome, LmConfigPtr config)
{
    char newKeydir[ MAXNAMLEN ];
    extern int errno;

    errno = 0; /* suppress funny, residual errors from Error() */

    /* keydir...
     * Look for it in three places, in order:
     *   - $NVRKEYDIR
     *   - $NVRHOME/adm/keys
     *   - default_keydir
     */
    if (keydir && *keydir) {
	if (AnyKeyFound(keydir, config))
	    return(keydir);
	if (FileType(keydir) & FILE_DIRECTORY) {
	    ProgMsg("key directory $%s is present (%s), but\n",
		NVRKEYDIR_ENV, keydir);
	    ProgMsg("none of the keys (%s) are usable!\n", KeyFileList(config));
	} else {
	    ProgMsg("The key directory missing: $%s=%s\n",
		NVRKEYDIR_ENV, keydir);
	}
    }

    if (nvrhome && *nvrhome) {
	sprintf(newKeydir, "%s/adm/keys", nvrhome);
	if (AnyKeyFound(newKeydir, config))
	    return( CopyOfString(newKeydir) );
	if (FileType(newKeydir) & FILE_DIRECTORY) {
	    ProgMsg("The implied key directory $%s/adm/keys is present (%s)\n",
		NVRHOME_ENV, newKeydir);
	    ProgMsg("but the key for \"%s\" is not there!\n",
		KeyFileList(config));
	} else {
	    ProgMsg("The implied key directory is missing:\n");
	    ProgMsg("    $%s/adm/keys ($%s is %s)\n",
		NVRHOME_ENV, NVRHOME_ENV, nvrhome);
	}
    } else {
	ProgMsg("%s is not defined in your environment.\n", NVRHOME_ENV);
    }

    if (keydir == NULL)
	ProgMsg("%s is not defined in your environment.\n", NVRKEYDIR_ENV);
    Msg("\n");
    ProgMsg("When starting the license manager, we look for keys in\n");
    ProgMsg("$%s/. and then in $%s/adm/keys\n", NVRKEYDIR_ENV, NVRHOME_ENV);

    if (FileType(default_keydir) & FILE_DIRECTORY) {
	ProgMsg("falling back to default key directory: %s ...\n",
	    default_keydir);
	return(default_keydir);
    }

    ProgMsg("There doesn't seem to be an appropriate key directory for the\n");
    ProgMsg("license manager to use.\n");
    ProgMsg("Consult the Installation Guide and the nvrlmd man page.\n");
    NvrLmExit(1);

    /*NOTREACHED*/
    return(NULL);
}

static void
AddLogFile(Argvec *argv, char *nvrhome)
{
    char logfile[ MAXNAMLEN ];

    if (nvrhome == NULL || *nvrhome == '\0')
	return;
    
    sprintf(logfile, "%s/adm/log/nvrlmd.log", nvrhome);
    if (FileWritable(logfile) || FileCreatable(logfile)) {
	AddArgv(argv, "-l", False);
	AddArgv(argv, logfile, True);
	AddArgv(argv, "-m0.25m", False);
    }
}

static void
StartLicenseManager(char *keydir, char *nvrhome, LmConfigPtr config)
{
    Argvec argvec;

    argvec.length = 0;
    argvec.argv = NULL;
    AddArgv(&argvec, "nvrlmd", False);
    argvec.path = argvec.argv[0];
    AddArgv(&argvec, "-e", False);
    AddFmtArgv(&argvec, "%s",
	FindKeyDir(keydir, nvrhome, config));
    AddLogFile(&argvec, nvrhome);
    AddArgv(&argvec, "-s0", False);

    Execute(&argvec);

    FreeArgv(&argvec);
    sleep(1);
}

static void
LmQuit(void)
{
    NvrLmClose();
    exit(1);
}

static void
CloseLmOnExit(PrivPointer ignored)
{
    NvrLmClose();
}

/*
 * can't make this static because NvrLmXt.c uses it
 */
int
_NvrLmInit(LmConfigPtr config)
{
#define EDIT_HOST_OFFSET 6
    static char EditableServerName[ELM_HOST_LEN+EDIT_HOST_OFFSET+1] = 
    	"%EDIT%XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
    int defStrLen = 32;
    char servername[ELM_HOST_LEN+1];
    char keydir[ELM_KEYDIR_LEN+1];
    char *pServername;
    char *pKeydir;
    char *nvrhome;
    extern char *getenv(char *);
    int code, i;

    /* init */
    if (NvrLmInitialized)
	return(ELM_OK);

    pServername = getenv(NVRHOST_ENV);
    pKeydir = getenv(NVRKEYDIR_ENV);
    nvrhome = getenv(NVRHOME_ENV);

    /* elm server name */
    if (!pServername || (*pServername == '\0')) {
	for(i = 0; i < defStrLen; i++) {
	    if( EditableServerName[EDIT_HOST_OFFSET+i] != 'X' ) {
		break;
	    }
	}
	if( i < defStrLen ) {
	    strncpy(servername, &EditableServerName[EDIT_HOST_OFFSET],
		    ELM_HOST_LEN);
	} else {
	    servername[0] = '\0';
	}
    } else {
	strncpy(servername, pServername, ELM_HOST_LEN);
    }

    /* keydir */
    if (!pKeydir || (*pKeydir == '\0')) {
	keydir[0] = '\0';
    } else {
	strncpy(keydir, pKeydir, ELM_KEYDIR_LEN);
    }

    /*
     * First, set up a signal handler for interrupt so that
     * we can disconnect from the license manager.  This should
     * be called early in the program, so if someone else wants
     * to set up a signal handler, it will override ours.
     */
    if( signal(SIGTERM, LmQuit) == (VoidFuncPtr)-1 ) {
        Error("Couldn't catch signal %d\n", SIGTERM);
    }
    if( signal(SIGINT, LmQuit) == (VoidFuncPtr)-1 ) {
        Error("Couldn't catch signal %d\n", SIGINT);
    }

    /*
     * Register a fatal handler to close the license manager connection
     */
    RegisterFatalHandler(CloseLmOnExit, NULL);

    if ((code = elm_init(servername, NULL, keydir, NULL)) >= 0) {
	NvrLmInitialized = 1;
	return(code);
    }

    if (code == ELM_SRV_DOWN) {
	ProgMsg("The license manager is down... trying to start one...\n");
	StartLicenseManager(keydir, nvrhome, config);

	if ((code = elm_init(servername, NULL, keydir, NULL)) >= 0) {
	    NvrLmInitialized = 1;
	    return(code);
	}

	if (code == ELM_SRV_DOWN) {
	    ProgMsg("Still not available... ");
	    ProgMsg("Something seems to be wrong...\n");
	}
    }

    return(code);
#undef EDIT_HOST_OFFSET
}

static struct {
    ULong bit;
    char *archName;
} ArchStrings[] = {
    { LIC_SGI_ARCHITECTURE,  "Silicon Graphics" },
    { LIC_SOL1_ARCHITECTURE, "Sun Solaris 1" },
    { LIC_SOL2_ARCHITECTURE, "Sun Solaris 2" },
    { LIC_HP_ARCHITECTURE,   "Hewlett-Packard" },
};

static char *
BitsToArchNames(char *archList, ULong archBits)
{
    int i, bit, nNames;
    char *archName, *pArchList = archList;

    for (nNames = bit = 0; archBits; bit++) {
	if ((archBits & (1<<bit)) == 0)
	    continue;

	archName = "<unknown>";
	for (i = 0; i < NumberOfCells(ArchStrings); i++) {
	    if (ArchStrings[i].bit & (1<<bit)) {
		archName = ArchStrings[i].archName;
		break;
	    }
	}

	archBits &= ~(1<<bit);
	nNames++;
	sprintf(pArchList, "%s%s",
	    nNames > 1 ? ", " : "",
	    archName);
	pArchList += strlen(pArchList);
    }
    return(archList);
}


/************************************************************************
 *									*
 *			PUBLIC ROUTINES					*
 *									*
 ************************************************************************/

#ifndef LM_DISABLED

int
NvrGetLicense(LmConfigPtr config)
{
    int code, bestCode, i;
    char *feature;
    struct elm_info licInfo;
    ULong archLicensed, archBits;
    char archList[ 1024 ];

    if (!NvrLmInitialized) {
	/* init */
	if ((code = _NvrLmInit(config)) < 0) {
	    return(code);
	}

	/* start elm heartbeat */
	if (config->doHeartbeat) {
	    elm_heartbeat(BEAT, config->exceptionCB);
	}
    }

    bestCode = LM_NOKEY;
    archLicensed = 0;
    for (i = 0; config->features[i][0]; i++) {
	feature = config->featureUsed = config->features[i];

	/* get a license */
	code = elm_getlicense(feature,
		ELM_GETLIC | (config->shared ? ELM_SHARED : 0),
		0);
	if (code != LM_NOKEY && bestCode == LM_NOKEY)
	    bestCode = code;
	if (code != ELM_OK)
	    continue;

	/*
	 * Check that it covers this architecture
	 */
	code = elm_getinfo(&licInfo, feature, 0);
	if (code < 0) {
	    (void) elm_retlicense(feature, 0);
	    ProgMsg("can't read vendor data from feature \"%s\"!\n", feature);
	    bestCode = code;
	    continue;
	}

	if (licInfo.vendordata[0] == '\0') {
	    ProgMsg("License vendor data is empty!\n");
	    (void) elm_retlicense(feature, 0);
            bestCode = LM_NOKEY;
            continue;
	}

	archBits = atoi(licInfo.vendordata);
	if (archBits == 0 || (archBits & LIC_THIS_ARCHITECTURE))
	    break;	/* we are licensed */

	/*
	 * Not licensed.  Return the license and remember which architectures
	 * are able to use this feature.
	 */
	archLicensed |= archBits;
	(void) elm_retlicense(feature, 0);
	bestCode = LM_NOKEY;
    }

    if (bestCode != ELM_OK) {
	elm_bye();
	NvrLmInitialized = 0;
	if (archLicensed) {
	    ProgMsg("Licenses exist for %s.\n",
		BitsToArchNames(archList, archLicensed));
	    ProgMsg("But nothing is available for %s\n",
		BitsToArchNames(archList, LIC_THIS_ARCHITECTURE));
	}
    }

    return(bestCode);
}

void
NvrLmClose(void)
{
    if (NvrLmInitialized) {
	elm_bye();
    }
    NvrLmInitialized = 0;
}

void
NvrLmExit(int code)
{
    NvrLmClose();
    exit(code);
}

void
NvrLmHeartBeat(char *feature)
{
    (void) elm_alive(feature);
}

void
NvrLicenseError(int err_code, LmConfigPtr config, boolean die)
{
    extern int errno;
    int i;

    errno = 0;

    if (err_code == ELM_OK) {
	/*
	 * This is probably a call from the exception handler handed to
	 * elm_heartbeat().  Documentation says that this is non-fatal
	 * and that the server probably went down (a call to elm_getstatus()
	 * would tell us).  Fudge for now.
	 */
	ProgMsg("license server is down.  Continuing...\n");
	return;
    }

    ProgMsg("For feature(s):");
    for (i = 0; config->features[i][0]; i++) {
	Msg("%s %s", i > 0 ? "," : "", config->features[i]);
    }
    Msg("\n");

    
    switch (err_code) {
    case ELM_NO_LICENSES:
        /* No licenses currently available. */
        ProgMsg("Sorry, all licenses are in use... Try again later.\n");
        break;
    case ELM_NOKEY:
        /* No key file was found for the specified feature in the key */
        /* directory. */
        ProgMsg("Sorry, cannot find a license for this software.\n");
        break;
    case ELM_AUTH_FAIL:
        /* The feature's license is invalid. */
        ProgMsg("Sorry, invalid license.\n");
        break;
    case ELM_EXPIRED:
        /* License for requested feature has expired. */
        ProgMsg("Sorry, license has expired.\n");
        break;
    case ELM_WAIT:
        /* License server has just started. Try again after the startup */
        /* period. */
        ProgMsg("License server has just started.  Try again in 3 minutes.\n");
        break;
    case ELM_NOHOST:
        /* Hostname for license server is not valid. */
        ProgMsg("Sorry, hostname for license server is invalid.\n");
        break;
    case ELM_SRV_DOWN:
        /* License server is down. */
        ProgMsg("Sorry, license server is down.\n");
        break;
    case ELM_SENDFAIL:
        /* Unable to send message to license daemon. Check errno for */
        /* specific error condition. */
        ProgMsg("Sorry, unable to communicate with license server.\n");
        break;
    case ELM_INSUFFICIENT_SERVERS:
        /* Client running with redundant servers but there are not enough */
        /* servers active to validate license. */
        ProgMsg("The servers are running in \"redundant server mode\"\n");
	ProgMsg("There are not enough active servers to validate a license.\n");
        break;
    case ELM_INVALID_CLIENT:
        ProgMsg("Sorry, invalid license manager client.\n");
        break;
    case ELM_ERROR:
        /* System Error. */
        Error("Sorry, a system error was encountered with license server.\n");
        break;
    case ELM_SOCKFAIL:
        /* Unsuccessful connection -- socket open failed. Check errno for */
        /* specific error condition. */
        Error("Sorry, failed to open socket to license server.\n");
        break;
    default:
	ProgMsg("Unknown license failure code=%d\n", err_code);
	break;
    }
    if (die)
	exit(1);
}

#endif /* LM_DISABLED */
