/* Bennett Battaile, April-June 1995, under contract to North
 * Valley Research.  Support macros for the SubbandFilter_*
 * routines:
 *   figure_V_macro    does the V=N*S calculation;
 *   find_sum_macro    does windowing and scaling;
 *   put_sample_macro  puts the sample out.
 */


/* 
 * perform the DCT, step 2 from the MPEG audio spec flow chart.
 * 
 *
 * FloatType _input[32]   Only care about elements
 *                        0 .. _lasti (others are 0) 
 * FloatType _output[64]  This won't wrap around the buffer V; whew.
 * int       _lazy        if 1,  only care about _output[4n].
 * int       _cutoff      if _lasti > _cutoff,  compute using
 *                        auto-generated code; otherwise, plain
 *                        matrix multiplication is faster.
 * int       _lasti       i.e., lastSbIdx: _input[i] == 0 for i > _lasti 
 *
 * also, need to have FloatType defined.
 */
#define figure_V_macro( _input, _output, _lazy, _matrix, _lasti, _cutoff ) \
{                                                                          \
    if ( _lasti > _cutoff ) {                                              \
        figure_V_the_autocode_way( _input, _output, _lazy) ;               \
    } else {                                                               \
        int i;                                                             \
        int j;                                                             \
        FloatType sigma;                                                   \
        FloatType *outputBlock = _output;                                  \
        for ( i = 16; i <= 48; ( (_lazy)? (i+=4) : (i++) ) )               \
        {                                                                  \
            if ( i == 16 ) { outputBlock[i] = (FloatType)0.0; continue; }  \
            sigma = (FloatType)0.0;                                        \
            for ( j = 0; j <= _lasti; j++ )                                \
            {                                                              \
                sigma += _matrix[i][j] * _input[j];                        \
            }                                                              \
            outputBlock[i] = sigma;                                        \
        }                                                                  \
        for ( i = 0; i < 16;  ( (_lazy)? (i+=4) : (i++) ) )                \
        {                                                                  \
            outputBlock[i] = - outputBlock[32-i];                          \
        }                                                                  \
        for ( i = 48; i < 64; ( (_lazy)? (i+=4) : (i++) ) )                \
        {                                                                  \
            if ( i == 48 ) continue;                                       \
            outputBlock[i] = outputBlock[96-i];                            \
        }                                                                  \
    }                                                                      \
}





/*
 *  has to do U-building, windowing and making samples.
 *
 * FloatType _V[1024]   The working vector, reused from call to call
 * FloatType _D[512]    the windowing vector, constant
 * int       _sumno     the index of the sum; since the sum uses every
 *                      32nd element of each array, shift the bases
 *                      to save some computation.  0..31.
 * int       _offset    where the current base of V is (bufOffset).
 * FloatType _sum
 *
 * The loop unrolling here is a little convoluted, sorry.  It's
 * essentially the same as the NVR original code in structure,
 * with a couple of small rearrangements for efficiency.
 *
 */
 
 
  /* in braces, to allow local variable decls. */
#define find_sum_macro( _V, _D, _sumno, _offset, _sum ) \
{                                                       \
    FloatType *Vbase = &(_V[_sumno]);                   \
    FloatType *Dbase = &(_D[_sumno]);                   \
    _sum =                                              \
        Vbase[( 128*0 + 0 + _offset)&1023] * Dbase[ 64*0 + 0 ] + \
        Vbase[( 128*0 +96 + _offset)&1023] * Dbase[ 64*0 +32 ] + \
        Vbase[( 128*1 + 0 + _offset)&1023] * Dbase[ 64*1 + 0 ] + \
        Vbase[( 128*1 +96 + _offset)&1023] * Dbase[ 64*1 +32 ] + \
        Vbase[( 128*2 + 0 + _offset)&1023] * Dbase[ 64*2 + 0 ] + \
        Vbase[( 128*2 +96 + _offset)&1023] * Dbase[ 64*2 +32 ] + \
        Vbase[( 128*3 + 0 + _offset)&1023] * Dbase[ 64*3 + 0 ] + \
        Vbase[( 128*3 +96 + _offset)&1023] * Dbase[ 64*3 +32 ] + \
        Vbase[( 128*4 + 0 + _offset)&1023] * Dbase[ 64*4 + 0 ] + \
        Vbase[( 128*4 +96 + _offset)&1023] * Dbase[ 64*4 +32 ] + \
        Vbase[( 128*5 + 0 + _offset)&1023] * Dbase[ 64*5 + 0 ] + \
        Vbase[( 128*5 +96 + _offset)&1023] * Dbase[ 64*5 +32 ] + \
        Vbase[( 128*6 + 0 + _offset)&1023] * Dbase[ 64*6 + 0 ] + \
        Vbase[( 128*6 +96 + _offset)&1023] * Dbase[ 64*6 +32 ] + \
        Vbase[( 128*7 + 0 + _offset)&1023] * Dbase[ 64*7 + 0 ] + \
        Vbase[( 128*7 +96 + _offset)&1023] * Dbase[ 64*7 +32 ];  \
}





/*
 * take a number and make a sample out of it.  Ship it out of the calling
 * routine.
 *
 * FloatType   _input
 * short      *_output     (incremented after filling)
 * something  *_audioInfo
 * int         _lazy       0 or 1
 */
#define put_sample_macro( _input, _audioInfo, _output)		\
    {								\
	long sample = (long)(_input);				\
	ClampAudio(sample);					\
	*_output = (short)sample;				\
	_output++;						\
    }
    
