/*
 *	Copyright (c) 1992, North Valley Research, Inc.
 *	Confidential -- All proprietary rights reserved.
 *	Licensed Material -- Property of North Valley Research, Inc.
 *
 *	This program is the property of North Valley Research, Inc. or others
 *	from whom North Valley Research has obtained a licensing right, and is
 *	considered by North Valley Research to be confidential.  It is protected
 *	by U.S. copyright law as an unpublished work and is
 *	furnished pursuant to a written license agreement.  It may
 *	not be used, copied or otherwise reproduced, or disclosed to
 *	others except in accordance with the terms and conditions of
 *	that agreement.
 *
 *	NAME
 *		MAdecoder.c -- MPEG Audio Decoder
 *
 *	DESCRIPTION
 *		This module contains the core of the decoder ie all the
 *		computational routines. (Layer I and II only)
 *		Functions are common to both layer unless
 *		otherwise specified.
 *
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/MAdecoder.c,v 1.33 1995/08/16 20:14:48 russ Exp $
 *
 */

/*
 *      INTERNAL INCLUDES
 *              Include files for local use only, therefore, NOT exported
 *		to any package or program using this package.
 */
#include        "MAcommon.h"
#include        "nvrTimer.h"
#include	"fileContent.h"
#include	"nvrReason.h"
#include	"nvrBitTest.h"
#include	"mu-law.h"

/*
 * If samples are about evenly distributed above and below zero, then
 * the first test will fail about half the time and not more work is done.
 * the other half the time, we have to do another test.  This is about
 * 25% better than two tests all the time.  A small percentage of the
 * time, we will have to clamp and that requires another test.
 */
#define ClampAudio(_sample)						\
	if ((ULong)_sample >> 15) {					\
	    if ((ULong)_sample >> 15 != 0x1ffff) {			\
		/* Msg("sample %d clamped\n", _sample); */		\
		if (_sample < 0)					\
		    _sample = 0x8000;					\
		else							\
		    _sample = 0x7fff;					\
	    }								\
	}

#define BooleanTimerIncrementMacro(inc)					\
    (									\
	( ((inc).accumulator += (inc).remainder) >= (inc).denom )	\
	    ? ((inc).accumulator -= (inc).denom, True)   		\
	    : False						        \
    )

void
SetupSynFilter(MpegAudioInfoPtr audioInfo)
{
    BUTInitSynthesisFilter(0,32768.0, audioInfo->quarterSample);
    BUTInitSynthesisFilter(1,32768.0, audioInfo->quarterSample);
}



/************************************************************************
 *									*
 *			PRIVATE ROUTINES				*
 *									*
 ************************************************************************/

void
crcError(MpegAudioInfoPtr audioInfo,boolean stereo)
{
    audioInfo->crcFrameError = True;

    DebugTask(N_crc_check, Msg("crc check failed\n") );
}


/*
 *	NAME
 *		HeaderndErrorCheck -- Process frame header less syncword
 *
 *	SYNOPSIS
 */
static CacheReturnValue
HeaderAndErrorCheck(CacheArgsDecl, MpegAudioInfoPtr audioInfo)
{
    BitStreamPtr bitStream = audioInfo->bitStream;
    ULong composite, tmp;

    /*
     * move to byte alignment
     */
    _nCacheBits &= ~0x7;

    Read32Bits(bitStream, composite, N_syncword);
    while ( ((composite >> 20) & VAL_SYNCWORD) != VAL_SYNCWORD) {
        if (CacheAtEndOfStream(bitStream)) {
	    WhenDebugEnabled(
		ProgMsg("No more audio frames\nInput stream may be empty\n");
	    )
	    CacheReturn(bitStream);
	}
	composite <<= 8;
	ReadBits(bitStream, 8, tmp, N_syncword);
	composite |= tmp;
    }

    audioInfo->frameHeader.composite = composite;
    if (P_REDUNDANCY(audioInfo->frameHeader.s)) {
	ReadBits(bitStream, 16, audioInfo->crcCheck, N_crc_check);
    }

    CacheReturn(bitStream);
}

static void
FrameDiff(MpegAudioInfoPtr audioInfo, BitMaskPtr pDiffs)
{
    FrameHeader old, new;

    BM_ZERO(pDiffs);
    /*
     * There is no change for the first frame
     */
    if (audioInfo->frame == 0) {
	return;
    }

    new.composite = audioInfo->frameHeader.composite;
    old.composite = audioInfo->lastHeader.composite;

    if (new.s.ID != old.s.ID) {
	BM_SET(pDiffs,FHId);
    }
    if (new.s.layer != old.s.layer) {
	BM_SET(pDiffs,FHlayer);
    }
    if (new.s.protection_bit != old.s.protection_bit) {
	BM_SET(pDiffs,FHprotectionBit);
    }
    if (new.s.bitrate_index != old.s.bitrate_index) {
	BM_SET(pDiffs,FHbitrateIndex);
    }
    if (new.s.sampling_frequency != old.s.sampling_frequency) {
	BM_SET(pDiffs,FHsamplingFrequency);
    }
    if (new.s.padding_bit != old.s.padding_bit) {
	BM_SET(pDiffs,FHpaddingBit);
    }
    if (new.s.private_bit != old.s.private_bit) {
	BM_SET(pDiffs,FHprivateBit);
    }
    if (new.s.mode != old.s.mode) {
	BM_SET(pDiffs,FHmode);
    }
    if (new.s.mode_extension != old.s.mode_extension) {
	BM_SET(pDiffs,FHmodeExtension);
    }
    if (new.s.copyright != old.s.copyright) {
	BM_SET(pDiffs,FHcopyright);
    }
    if (new.s.original_home != old.s.original_home) {
	BM_SET(pDiffs,FHoriginalHome);
    }
    if (new.s.emphasis != old.s.emphasis) {
	BM_SET(pDiffs,FHemphasis);
    }
    return;
}

static void
ShowFrameChange(MpegAudioInfoPtr audioInfo, BitMaskPtr pDiffs)
{
    char msg[ 1024 ], *p = msg;
    FrameHeader old, new;

    new.composite = audioInfo->frameHeader.composite;
    old.composite = audioInfo->lastHeader.composite;

    DebugTask(N_AuSequenceInfo,
	FrameHeader f;

	f.composite = audioInfo->frameHeader.composite;
	Msg("frame %d:\n", audioInfo->frame);
	Msg("id lyr prot brate  freq pad priv mode ISbands copyr orig emph\n");
	Msg("%2d%3d%5d%6d%7d%3d%5d%5s%8s%6s%5s  %s\n",
	   f.s.ID,
	   4 - f.s.layer,
	   f.s.protection_bit,
	   MA_BitRateTable[ f.s.layer ][ f.s.bitrate_index ],
	   MA_SamplingFrequencyTable[ f.s.sampling_frequency ],
	   f.s.padding_bit,
	   f.s.private_bit,
	   AbbrevAudioModeNames[ f.s.mode ],
	   f.s.mode == MODE_JOINT_STEREO
		? JointStereoBandNames[ f.s.mode_extension ]
		: "     ",
	   f.s.copyright ? "yes" : "no",
	   f.s.original_home ? "yes" : "no",
	   EmphasisNames[ f.s.emphasis ]);
    )

    if (BM_ISSET(pDiffs,FHId)) {
	sprintf(p, "  id %d->%d\n", old.s.ID, new.s.ID);
	p += strlen(p);
    }
	if (BM_ISSET(pDiffs,FHlayer)) {
	sprintf(p, "  layer %d->%d\n", 4-old.s.layer, 4-new.s.layer);
	p += strlen(p);
    }
    if (BM_ISSET(pDiffs,FHprotectionBit)) {
	sprintf(p, "  protect %d->%d\n",
	    old.s.protection_bit, new.s.protection_bit);
	p += strlen(p);
    }
    if (BM_ISSET(pDiffs,FHbitrateIndex)) {
	sprintf(p, "  bitrate %d->%d kbit/sec\n",
	    MA_BitRateTable[ old.s.layer ][ old.s.bitrate_index ],
	    MA_BitRateTable[ new.s.layer ][ new.s.bitrate_index ]);
	p += strlen(p);
    }
    if (BM_ISSET(pDiffs,FHsamplingFrequency)) {
	sprintf(p, "  %d->%d kHz\n",
	    MA_SamplingFrequencyTable[ old.s.sampling_frequency ],
	    MA_SamplingFrequencyTable[ new.s.sampling_frequency ]);
	p += strlen(p);
    }
    if (BM_ISSET(pDiffs,FHmode)
	&& (new.s.mode == MODE_SINGLE_CHANNEL
	    || old.s.mode == MODE_SINGLE_CHANNEL)) {
	sprintf(p, "  mode %s", AbbrevAudioModeNames[ old.s.mode ]);
	p += strlen(p);
	if (old.s.mode == MODE_JOINT_STEREO) {
	    sprintf(p, "/%s", JointStereoBandNames[ old.s.mode_extension ]);
	    p += strlen(p);
	}
	sprintf(p, "->%s", AbbrevAudioModeNames[ new.s.mode ]);
	p += strlen(p);
	if (new.s.mode == MODE_JOINT_STEREO)
	    sprintf(p, "/%s\n", JointStereoBandNames[ new.s.mode_extension ]);
	else
	    sprintf(p, "\n");
	p += strlen(p);
    }
    if (BM_ISSET(pDiffs,FHemphasis)) {
	sprintf(p, "  %s->%s\n",
	    EmphasisNames[ old.s.emphasis ],
	    EmphasisNames[ new.s.emphasis ]);
	p += strlen(p);
    }

    DebugTask(N_AuSequenceInfo,
	if (BM_ISSET(pDiffs,FHpaddingBit)) {
	    sprintf(p, "  pad %s->%s\n",
		old.s.padding_bit ? "on" : "off",
		new.s.padding_bit ? "on" : "off");
	    p += strlen(p);
	}
	if (BM_ISSET(pDiffs,FHprivateBit)) {
	    sprintf(p, "  private %s->%s\n",
		old.s.private_bit ? "on" : "off",
		new.s.private_bit ? "on" : "off");
	    p += strlen(p);
	}
	if (BM_ISSET(pDiffs,FHcopyright)) {
	    sprintf(p, "  private %s->%s\n",
		old.s.copyright ? "on" : "off",
		new.s.copyright ? "on" : "off");
	    p += strlen(p);
	}
	if (BM_ISSET(pDiffs,FHoriginalHome)) {
	    sprintf(p, "  copyright %s->%s\n",
		old.s.original_home ? "on" : "off",
		new.s.original_home ? "on" : "off");
	    p += strlen(p);
	}
    )

    if (p != msg) {
	Msg("Warning: MPEG frame %d changed:\n%s", audioInfo->frame, msg);
    }
}

/*
 * FrameReset
 *
 *	SYNOPSIS
 */
static InternalReasonPtr
FrameReset(MpegAudioInfoPtr audioInfo)
{
    FrameInfoPtr	pContext = &audioInfo->frameInfo;
    int kBitsPerChannel, kBitrate;
    short i_sfreq;
    BitMask HeaderDiffs;
    int i;

    FrameDiff(audioInfo, &HeaderDiffs);
	   
    /*
     * Set Context
     */
    i = pContext->layer = audioInfo->frameHeader.s.layer;
    if (i == 1) {
	return( CreateReason("Layer III audio not implemented yet\n") );
    } else if (i == 0) {
	return( CreateReason("Invalid audio layer\n") );
    }

    i = pContext->sampling_frequency =
	audioInfo->frameHeader.s.sampling_frequency;
    if (i == 3) {
	return( CreateReason("Invalid sampling frequency\n") );
    }

    pContext->mode = audioInfo->frameHeader.s.mode;
    pContext->mode_extension = audioInfo->frameHeader.s.mode_extension;

    if (audioInfo->frame == 0 ||
	BM_ISSET(&HeaderDiffs,FHsamplingFrequency)) {
    }

    i = audioInfo->frameHeader.s.bitrate_index;
    if (i == 15) {
	return( CreateReason("Invalid bitrate index\n") );
    }
    i_sfreq = audioInfo->frameHeader.s.sampling_frequency;
    kBitrate = MA_BitRateTable[audioInfo->frameHeader.s.layer][i];

    switch (audioInfo->frameHeader.s.layer) {
      case LAYER_I:
	audioInfo->frameSize = 
	    12 * kBitrate * 1000 / MA_SamplingFrequencyTable[ i_sfreq ];
	switch (audioInfo->frameHeader.s.mode) {
	  case MODE_SINGLE_CHANNEL:
	    audioInfo->decodeProc = DecodeLayerI_SingleChannel;
	    break;
	  case MODE_STEREO:
	    /* fall through */
	  case MODE_DUAL_CHANNEL:
	    audioInfo->decodeProc = DecodeLayerI_Stereo;
	    break;
	  case MODE_JOINT_STEREO:
	    audioInfo->decodeProc = DecodeLayerI_IntensityStereo;
	    break;
	}
	break;
      case LAYER_II:
	/* test validity of mode and bitrate */
	if ( ( ((i >= 1 && i <=3) || i == 5) &&
	        audioInfo->frameHeader.s.mode != MODE_SINGLE_CHANNEL)
	    || (i >= 11 &&
		audioInfo->frameHeader.s.mode == MODE_SINGLE_CHANNEL)) {
	    return(
      CreateReason("Illegal combination of mode and bitrate in layer II\n")
		   );
	}
	audioInfo->frameSize =
	    144 * kBitrate * 1000 / MA_SamplingFrequencyTable[ i_sfreq ];
	if (audioInfo->frameHeader.s.mode != MODE_SINGLE_CHANNEL)
	    kBitsPerChannel = kBitrate / 2;
	else
	    kBitsPerChannel = kBitrate;

	if ((i_sfreq == INDEX_48000HZ && kBitsPerChannel >= 56) 
		|| (kBitsPerChannel >= 56 &&  kBitsPerChannel <= 80)) {
	    pContext->sblimit = 27;
	    pContext->nbal_table = II_Table0_nbal;
	    pContext->alloc_table = II_AllocTable0;
	} else if (i_sfreq != INDEX_48000HZ && kBitsPerChannel >= 96) {
	    pContext->sblimit = 30;
	    pContext->nbal_table = II_Table1_nbal;
	    pContext->alloc_table = II_AllocTable1;
	} else if (i_sfreq != INDEX_32000HZ && kBitsPerChannel <= 48) {
	    pContext->sblimit = 8;
	    pContext->nbal_table = II_Table2_nbal;
	    pContext->alloc_table = II_AllocTable2;
	} else {
	    pContext->sblimit = 12;
	    pContext->nbal_table = II_Table3_nbal;
	    pContext->alloc_table = II_AllocTable3;
	}
	switch (audioInfo->frameHeader.s.mode) {
	  case MODE_SINGLE_CHANNEL:
	    audioInfo->decodeProc = DecodeLayerII_SingleChannel;
	    break;
	  case MODE_STEREO:
	  case MODE_DUAL_CHANNEL:
	    audioInfo->decodeProc = DecodeLayerII_Stereo;
	    break;
	  case MODE_JOINT_STEREO:
	    audioInfo->decodeProc = DecodeLayerII_IntensityStereo;
	    break;
	}
	break;
    }

    ShowFrameChange(audioInfo, &HeaderDiffs);

    /*
     * Reset Accumulators
     */
    DebugTask(N_ResetSbSamples,
	memset((char*)audioInfo->u.monoSbSample,
	       0,
	       sizeof(audioInfo->u.monoSbSample));
    )
    audioInfo->nCrcErrors = 0;
    audioInfo->lastHeader.composite = audioInfo->frameHeader.composite;

    return(NULL);
}

/*
 *	NAME
 *		FrameCheck
 *
 *	SYNOPSIS
 */
static InternalReasonPtr
FrameCheck(MpegAudioInfoPtr audioInfo)
{
    InternalReasonPtr reason = NULL;

    /* check layer, bitrate_index, sampling_frequency, mode, & mode_extension */
    if (audioInfo->lastHeader.composite != audioInfo->frameHeader.composite) {
	reason = FrameReset(audioInfo);
    }
    return(reason);
}


/************************************************************************
 *									*
 *			PUBLIC ROUTINES					*
 *									*
 *		Publically accessible routines				*
 *									*
 ************************************************************************/

/*
 * Initialize mpeg decoding from beginning of audio file 
 */
boolean
NewMaSetupDecode(MpegAudioInfoPtr audioInfo, char *file, BitStreamPtr first)
{
    InternalReasonPtr irp;
    FileInfo fileInfo;

    InitLayerI_Tables(audioInfo);
    SetupSynFilter(audioInfo);
    InitBitStreamData(audioInfo->bitStream);
    audioInfo->bitStream->seqPriv = (PrivPointer)audioInfo;
    audioInfo->lastHeader.composite = 0;
    audioInfo->nFifoEntries = 0;

    irp = GetFileContentInfo(file, &fileInfo);
    if (irp) {
	TtyShowReason(irp);
	FreeReason(irp);
	return(False);
    }
    switch (fileInfo.type) {

    case FT_MpegAudio:
        irp = OpenBitStreamInput(audioInfo->bitStream, file);
        if (irp) {
	    TtyShowReason(irp);
	    FreeReason(irp);
	    return(False);
	}
        break;

    case FT_MpegSystem:
        irp = OpenSystemStreamInput(audioInfo->bitStream,
					 first, file, AudioBitstream);
        if (irp) {
	    TtyShowReason(irp);
	    FreeReason(irp);
	    return(False);
	}
        break;

    default:
	irp = CreateReason("Unrecognized audio file type\n");
	TtyShowReason(irp);
	FreeReason(irp);
        return(False);
    }

    return(True);
}

boolean
MAgatherSamples(MpegAudioInfoPtr audioInfo, long gather_samps)
{
    BitStreamPtr bitStream = audioInfo->bitStream;
    boolean status = True;
    BitByteOffset lastOffset;
    InternalReasonPtr reason;
    int frameSamples;
    CheckOutCache(bitStream);

    if( gather_samps > audioInfo->fifoSize ) {
	return(False);
    }
    DebugTask(N_Timer, StartTimer() )

    lastOffset = BitStreamOffset(bitStream, 0);
    while (!CacheAtEndOfStream(bitStream)
	   && audioInfo->nFifoEntries < gather_samps) {
	ShowMark(bitStream, N_AuFrameMark, ("# %d\n", audioInfo->frame));
	RestoreCacheAfterCall(bitStream,
	      HeaderAndErrorCheck(CacheArgs, audioInfo));
        if (CacheAtEndOfStream(bitStream)) {
	    Msg("premature EOF\n");
	    status = False;
	    goto out;
	}

	/*
	 * Check Header
	 */
	reason = FrameCheck(audioInfo);
	if (reason != NULL) {
	    TtyShowReason(reason);
	    FreeReason(reason);
	    status = False;
	    audioInfo->frame++;
	    goto out;
	}

	/*
	 * Process Audio Frame
	 */

	RestoreCacheAfterCall(bitStream,
	      (*audioInfo->decodeProc)(CacheArgs, audioInfo));
	if (audioInfo->crcFrameError) {
	    audioInfo->nCrcErrors++;

	    frameSamples =  audioInfo->frameInfo.mode == MODE_SINGLE_CHANNEL
			  ? NMonoSamplesPerFrame(audioInfo)
			  : NStereoSamplesPerFrame(audioInfo);
	    frameSamples *=   audioInfo->frameInfo.layer == LAYER_I
			    ? 12 : 12 * 3;
	    if ((audioInfo->nCrcErrors) == 2) {
		Msg("CrcErrors == 2\n");
		/* mute the frame */
		memset((char *)&audioInfo->fifo[audioInfo->nFifoEntries],
		       0, sizeof(short) * frameSamples);
	    }
	    
	    audioInfo->nFifoEntries += frameSamples;
	    audioInfo->crcFrameError = False;	/* we've seen it, so clear it */
	}
	audioInfo->frame++;

    }
    if (audioInfo->nFifoEntries == 0) {
	status = False;
    }

 out:


    DebugTask(N_Timer, StopTimer() )

    FlushCacheValues(bitStream);

    return(status);
}
