/*
 *	Copyright (c) 1992, North Valley Research, Inc.
 *	Confidential -- All proprietary rights reserved.
 *	Licensed Material -- Property of North Valley Research, Inc.
 *
 *	This program is the property of North Valley Research, Inc. or others
 *	from whom North Valley Research has obtained a licensing right, and is
 *	considered by North Valley Research to be confidential.  It is protected
 *	by U.S. copyright law as an unpublished work and is
 *	furnished pursuant to a written license agreement.  It may
 *	not be used, copied or otherwise reproduced, or disclosed to
 *	others except in accordance with the terms and conditions of
 *	that agreement.
 *
 *	NAME
 *		MAlayerI.c -- MPEG Audio Decoder
 *
 *	DESCRIPTION
 *		This module contains the core of the decoder ie all the
 *		computational routines. (Layer I and II only)
 *		Functions are common to both layer unless
 *		otherwise specified.
 *
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/MAlayerI.c,v 1.26 1995/08/16 20:14:48 russ Exp $
 *
 */

/*
 *      INTERNAL INCLUDES
 *              Include files for local use only, therefore, NOT exported
 *		to any package or program using this package.
 */
#define NeedFunctionPrototypes 1

#include        "nvrTypes.h"
#include        "nvrDebug.h"
#include        "bitstream.h"
#include        "longjump.h"
#include        "MAdebug.h"
#include        "MAcommon.h"


/************************************************************************
 *									*
 *			PUBLIC ROUTINES					*
 *									*
 *		Publically accessible routines				*
 *									*
 ************************************************************************/

/*
 *	NAME
 *		InitLayerI_Tables
 *
 *	SYNOPSIS
 */
void
InitLayerI_Tables(MpegAudioInfoPtr audioInfo)
{
    int nb;
    unsigned short sample;
    double fraction;
    FloatType *requantTable = audioInfo->I_RequantizationTable;

    for (nb = 2; nb < 16; nb++) {
        for (sample = 0; sample <= MaskOf(nb); sample++) {
            /* step 1 */
            if ((sample >> (nb-1)) & 1) {
                 fraction = 0.0;
            } else {
                fraction = -1.0;
            }

            /* step 2 */
            fraction +=
                (double) (sample & MaskOf(nb-1)) /
                              (double) (1L<<(nb-1));

            /* step 3 */
            fraction =
                 (double) (fraction+1.0/(double)(1L<<(nb-1))) *
                 (double) (1L<<nb) / (double) MaskOf(nb);


	    requantTable[ (1<<nb) - 4 + sample ] = fraction;
        }
    }
}

/*
 *	NAME
 *		I_CheckCRC
 *
 *	SYNOPSIS
 */
static UShort
I_CheckCRC(MpegAudioInfoPtr audioInfo, void *allocBits,
	   int stereo) /* no = 0; 1 = yes */
{
    register int	sb;
    UInt		mask;
    UShort		crc = ~0;
    int			*alloc_bits = (int *)allocBits;
    int			bound;

    UPDATE_CRC(mask, 4, audioInfo->frameHeader.s.bitrate_index, crc);
    UPDATE_CRC(mask, 2, audioInfo->frameHeader.s.sampling_frequency, crc);
    UPDATE_CRC(mask, 1, audioInfo->frameHeader.s.padding_bit, crc);
    UPDATE_CRC(mask, 1, audioInfo->frameHeader.s.private_bit, crc);
    UPDATE_CRC(mask, 2, audioInfo->frameHeader.s.mode, crc);
    UPDATE_CRC(mask, 2, audioInfo->frameHeader.s.mode_extension, crc);
    UPDATE_CRC(mask, 1, audioInfo->frameHeader.s.copyright, crc);
    UPDATE_CRC(mask, 1, audioInfo->frameHeader.s.original_home, crc);
    UPDATE_CRC(mask, 2, audioInfo->frameHeader.s.emphasis, crc);
    
    switch(audioInfo->frameHeader.s.mode) {
    case MODE_SINGLE_CHANNEL:
    case MODE_STEREO:
    case MODE_DUAL_CHANNEL:
	for (sb = 0; sb < (32 << stereo); sb++) {
	    UPDATE_CRC(mask, 4, alloc_bits[sb], crc);
	}
	break;
    case MODE_JOINT_STEREO:
	bound = MA_Bound[audioInfo->frameHeader.s.mode_extension];
	for (sb = 0; sb < (bound << 1); sb++) {
	    UPDATE_CRC(mask, 4, alloc_bits[sb], crc);
	}
	for (sb = (bound << 1); sb < (32 << 1); sb += 2) {
	    UPDATE_CRC(mask, 4, alloc_bits[sb], crc);
	}
    }
    return(crc);
}

/*
 *	NAME
 *		DecodeLayerI_SingleChannel
 *
 *	SYNOPSIS
 */
CacheReturnValue
DecodeLayerI_SingleChannel(CacheArgsDecl, MpegAudioInfoPtr audioInfo)
/*
 *	DESCRIPTION
 *		This function decodes audio data in Layer I, single
 *		channel format.
 *
 */
{
    BitStreamPtr	bitStream = audioInfo->bitStream;
    register int	sb, slot;
    unsigned int	nb;
    short		i_scale_factor[32];
    int 		alloc_bits[32];
    UShort		crc;
    UShort		inputSample;
    FloatType		requantized_sample[32];
    FloatType		*requantTable = audioInfo->I_RequantizationTable;

    /*
     * DECODE ALLOCATION
     */
    for (sb = 0; sb < 32; sb++) {
	ReadBits(bitStream, 4, alloc_bits[sb], N_allocation);
    }
    DebugTask(N_AllocationData,
	Msg("allocation:    ");
	for (sb = 0; sb < 32; sb++) {
	    if (alloc_bits[sb]) Msg("%2d", alloc_bits[sb]);
	    else                Msg(" .");
	}
	Msg("\n");
    )

    /*
     * DECODE SCALE FACTOR INDICES
     */
    for (sb = 0; sb < 32; sb++) {
	if (alloc_bits[sb]) {
	    ReadBits(bitStream, 6, i_scale_factor[sb], N_scale_factor);
	}
    }
    DebugTask(N_ScaleData,
	Msg("scale:         ");
	for (sb = 0; sb < 32; sb++) {
	    if (alloc_bits[sb]) Msg("%3d", i_scale_factor[sb]);
	    else                Msg("  .");
	}
	Msg("\n");
    )



    /*
     * CRC CHECK
     */
    if (P_REDUNDANCY(audioInfo->frameHeader.s)) {
	crc = I_CheckCRC(audioInfo, alloc_bits, 0);
	if (crc != audioInfo->crcCheck) {
	    crcError(audioInfo,False);
	    CacheReturn(bitStream);
	}
    }


    /*
     * DECODE SAMPLE
     */
    for (slot = 0; slot < 12; slot++) {
	/*
	 * BUFFER AND REQUANTIZE SAMPLES
	 */
	for (sb = 0; sb < 32; sb++) {
	    nb = alloc_bits[sb];
	    if (nb) {
		nb++;
		ReadBits(bitStream, nb, inputSample, N_sample);
		requantized_sample[sb] =
			requantTable[ (1<<nb) - 4 + inputSample ]
			* I_II_ScaleFactor[i_scale_factor[sb]];
	    } else {
		requantized_sample[sb] = 0.0;
	    }
	}

	DebugTask(N_SampleData,
	    Msg("slt%3d coef:", slot);
	    for (sb = 0; sb < 32; sb += 2) {
		if (alloc_bits[sb]) Msg("%6.1f", requantized_sample[sb]);
		else                Msg("    . ");
	    }
	    Msg("\n               ");
	    for (sb = 1; sb < 32; sb += 2) {
		if (alloc_bits[sb]) Msg("%6.1f", requantized_sample[sb]);
		else                Msg("    . ");
	    }
	    Msg("\n");
	)

        BUTMonoSubbandFilter(requantized_sample,
			     &audioInfo->fifo[ audioInfo->nFifoEntries ]);
        audioInfo->nFifoEntries += NMonoSamplesPerFrame(audioInfo);
    }
    CacheReturn(bitStream);
}


/*
 *	NAME
 *		DecodeLayerI_Stereo
 *
 *	SYNOPSIS
 */
/* ARGSUSED */
CacheReturnValue
DecodeLayerI_Stereo(CacheArgsDecl, MpegAudioInfoPtr audioInfo)
/*
 *	DESCRIPTION
 *		This function decodes audio data in Layer I, stereo
 *		and dual channel format.
 *
 *	NOTES
 *		Arrangement of arrays in this code differs from the MPEG
 *		specification and is more attuned to the way data is
 *		found in the bit stream.  For example, the MPEG spec
 *		declares allocation[ch][sb] ... however the channel data
 *		is interleaved per subband so a better organization is
 *		allocation[sb][ch], therefore we can just advance a pointer!
 *
 *	RETURNS
 */
{
    BitStreamPtr	bitStream = audioInfo->bitStream;
    register int	slot, sb;
    unsigned int	nb;
    short		i_scale_factor[32][2];
    int 		alloc_bits[32][2];
    UShort		crc;
    UShort		inputSample;
    FloatType		requantized_sample[2][32];
    FloatType		*requantTable = audioInfo->I_RequantizationTable;


    /*
     * DECODE ALLOCATION
     */
    for (sb = 0; sb < 32; sb++) {
	ReadBits(bitStream, 4, alloc_bits[sb][0], N_allocation);
	ReadBits(bitStream, 4, alloc_bits[sb][1], N_allocation);
    }
    DebugTask(N_AllocationData,
	register int	ch;
	Msg("allocation for ch 0\n");
	for (ch = 0; ch < 2; ch++) {
	    Msg("alloc ch%d:   ", ch);
	    for (sb = 0; sb < 32; sb++) {
		if (alloc_bits[sb][ch]) Msg("%2d", alloc_bits[sb][ch]);
		else                Msg(" .");
	    }
	}
	Msg("\n");
    )

    /*
     * DECODE SCALE FACTOR INDICES
     */
    for (sb = 0; sb < 32; sb++) {
	if (alloc_bits[sb][0]) {
	    ReadBits(bitStream, 6, i_scale_factor[sb][0], N_scale_factor);
	}
	if (alloc_bits[sb][1]) {
	    ReadBits(bitStream, 6, i_scale_factor[sb][1], N_scale_factor);
	}
    }


    /*
     * CRC CHECK
     */
    if (P_REDUNDANCY(audioInfo->frameHeader.s)) {
	crc = I_CheckCRC(audioInfo, alloc_bits, 1);
	if (crc != audioInfo->crcCheck) {
	    crcError(audioInfo, True);
	    CacheReturn(bitStream);
	}
    }


    /*
     * DECODE SAMPLE
     */
    for (slot = 0; slot < 12; slot++) {

	/*
	 * BUFFER AND REQUANTIZE SAMPLES
	 */
	for (sb = 0; sb < 32; sb++) {

	    if (alloc_bits[sb][0]) {
		nb = alloc_bits[sb][0] + 1;
		ReadBits(bitStream, nb, inputSample, N_sample);
		requantized_sample[0][sb] =
			requantTable[ (1<<nb) - 4 + inputSample ]
			* I_II_ScaleFactor[i_scale_factor[sb][0]];
	    } else {
		requantized_sample[0][sb] = 0.0;
	    }								

	    if (alloc_bits[sb][1]) {
		nb = alloc_bits[sb][1] + 1;
		ReadBits(bitStream, nb, inputSample, N_sample);
		requantized_sample[1][sb] =
			requantTable[ (1<<nb) - 4 + inputSample ]
			* I_II_ScaleFactor[i_scale_factor[sb][1]];
	    } else {
		requantized_sample[1][sb] = 0.0;
	    }								

	}



	if( audioInfo->monoSample ) {
            BUTStereoToMonoSubbandFilter(requantized_sample,
				 &audioInfo->fifo[ audioInfo->nFifoEntries ]);
        } else {
            BUTStereoSubbandFilter(requantized_sample,
				   &audioInfo->fifo[ audioInfo->nFifoEntries ]);
        }
        audioInfo->nFifoEntries += NStereoSamplesPerFrame(audioInfo);
    }

    CacheReturn(bitStream);
}

/*
 *	NAME
 *		DecodeLayerI_IntensityStereo
 *
 *	SYNOPSIS
 *
 */
/* ARGSUSED */
CacheReturnValue
DecodeLayerI_IntensityStereo(CacheArgsDecl, MpegAudioInfoPtr audioInfo)
/*
 *	DESCRIPTION
 *		This function decodes audio data in Layer I, intensity 
 *		stereo format.
 */
{
    BitStreamPtr	bitStream = audioInfo->bitStream;
    register int	slot, sb;
    int			bound;
    unsigned int	nb;
    short		i_scale_factor[32][2];
    int 		alloc_bits[32][2];
    UInt		crc;
    UShort		inputSample;
    FloatType		requantized_sample[2][32];
    FloatType		*requantTable = audioInfo->I_RequantizationTable;


    /*
     * DECODE ALLOCATION
     */
    bound = MA_Bound[audioInfo->frameHeader.s.mode_extension];
    for (sb = 0; sb < bound; sb++) {
	ReadBits(bitStream, 4, alloc_bits[sb][0], N_allocation);
	ReadBits(bitStream, 4, alloc_bits[sb][1], N_allocation);
    }
    for (sb = bound; sb < 32; sb++) {
	ReadBits(bitStream, 4, alloc_bits[sb][0], N_allocation);
	alloc_bits[sb][1] = alloc_bits[sb][0];
    }

    /*
     * DECODE SCALE FACTOR INDICES
     */
    for (sb = 0; sb < bound; sb++) {
	if (alloc_bits[sb][0]) {
	    ReadBits(bitStream, 6, i_scale_factor[sb][0], N_scale_factor);
	}
	if (alloc_bits[sb][1]) {
	    ReadBits(bitStream, 6, i_scale_factor[sb][1], N_scale_factor);
	}
    }
    for (sb = bound; sb < 32; sb++) {
	if (alloc_bits[sb][0]) {
	    ReadBits(bitStream, 6, i_scale_factor[sb][0], N_scale_factor);
	    ReadBits(bitStream, 6, i_scale_factor[sb][1], N_scale_factor);
	}
    }


    /*
     * CRC CHECK
     */
    if (P_REDUNDANCY(audioInfo->frameHeader.s)) {
	crc = I_CheckCRC(audioInfo, alloc_bits, 1);
	if (crc != audioInfo->crcCheck) {
	    crcError(audioInfo,True);
	    CacheReturn(bitStream);
	}
    }

    /*
     * DECODE SAMPLE
     */
    for (slot = 0; slot < 12; slot++) {
	/*
	 * BUFFER AND REQUANTIZE SAMPLES
	 */

	for (sb = 0; sb < bound ; sb++) {

	    if (alloc_bits[sb][0]) {
		nb = alloc_bits[sb][0] + 1;
		ReadBits(bitStream, nb, inputSample, N_sample);
		requantized_sample[0][sb] =
			requantTable[ (1<<nb) - 4 + inputSample ]
			* I_II_ScaleFactor[i_scale_factor[sb][0]];
	    } else {
		requantized_sample[0][sb] = 0.0;
	    }								

	    if (alloc_bits[sb][1]) {
		nb = alloc_bits[sb][1] + 1;
		ReadBits(bitStream, nb, inputSample, N_sample);
		requantized_sample[1][sb] =
			requantTable[ (1<<nb) - 4 + inputSample ]
			* I_II_ScaleFactor[i_scale_factor[sb][1]];
	    } else {
		requantized_sample[1][sb] = 0.0;
	    }								

	}
	for (sb = bound; sb < 32; sb++) {

	    if (alloc_bits[sb][0]) {
		nb = alloc_bits[sb][0] + 1;
		ReadBits(bitStream, nb, inputSample, N_sample);
		requantized_sample[0][sb] =
			requantTable[ (1<<nb) - 4 + inputSample ]
			* I_II_ScaleFactor[i_scale_factor[sb][0]];
		requantized_sample[1][sb] =
			requantTable[ (1<<nb) - 4 + inputSample ]
			* I_II_ScaleFactor[i_scale_factor[sb][1]];
	    } else {
		requantized_sample[0][sb] = 0.0;
		requantized_sample[1][sb] = 0.0;
	    }								

	}


	/*
	 * SYNTHESIS SUBBAND FILTER
	 */
	if( audioInfo->monoSample ) {
            BUTStereoToMonoSubbandFilter(requantized_sample,
				 &audioInfo->fifo[ audioInfo->nFifoEntries ]);
        } else {
            BUTStereoSubbandFilter(requantized_sample,
				   &audioInfo->fifo[ audioInfo->nFifoEntries ]);
        }
        audioInfo->nFifoEntries += NStereoSamplesPerFrame(audioInfo);
    }

    CacheReturn(bitStream);
}
