/*
 *   program to generate in-line computations for DCT calculation.
 *   Bennett Battaile, April-June 1995, under contract to North Valley
 *   Research.
 *   
 *   Uses a recursive algorithm to do the V = N*S matrix multiplication
 *   needed for MPEG audio decompression, and emits a string at each point
 *   it actually would do a computation.  Presto, branch-free C code,
 *   without ever having to try to unroll the recursion manually.
 *
 *
 *   Explanation of recursive algorithm
 *
 *      The full derivation won't fit in 80 column print, so I'm
 *   providing it elsewhere.  Quick summary:  The matrix N can be
 *   split into 3 horizontal bands (I'm considering N as 64 rows by
 *   32 columns, and S as a 32-element column vector).  The top and
 *   bottom 16 row chunks can be derived from the middle 32 rows with
 *   just sign changes, so the problem reduces to a 32x32 matrix
 *   multiplication.
 *      Now, the matrix chunk (32x32) can be decomposed into two
 *   16x16 matrices by (a) putting even rows in one matrix, E, and odd
 *   rows in another, O, and  (b) recognizing that each row in E is
 *   symmetrical and each row in O is negative-symmetrical, so that
 *   the input vector can be folded in half and applied to just the
 *   first 16 columns.  Voila:  E is now in the same form as the
 *   original, so the same procedure can be use recursively, and O
 *   is now in the form of step 8b in the IEEE paper, so the paper's
 *   recursive solution can be applied to it.
 *      Emphasize:  This is a recursive decomposition, which is written
 *   out in-line (N -> E -> EE -> EEE ...); at each step, a recursive
 *   routine (DCT) is called on one branch of the decomposition.
 *
 *   CAVEAT:
 *   The generated code assumes the definition of the symbol DCT_OUTPUT,
 *   an array into which the answers go.
 *   Also assumes the definition of OFFSET_I(i), a macro which translates
 *   a number i, 0..63, into the proper subscript for DCT_OUTPUT.
 *   Also assumes the definition of DCT_INPUT, which can be indexed 0..31.
 *   Also assumes the possible definition of STYLE_LAZY, which will
 *   indicate that only every 4th entry is to be calculated.
 *
 */
 
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "process.h"
#include "osio.h"
#ifndef M_PI
    #define M_PI 3.14159265358979323846
#endif
 
FILE *autocodeFP; // file to write generated code to.
 
enum {
    NoOutput,
    FigureVMacro,
    FigureVSubr
} outputType = NoOutput;
 
// rewriting the Microsoft CString class here, since Visual C++ seems
// to have a problem finding it when compiling for a QuickWin program.
// Grr.  (in the end, just as well -- makes the code self-contained,
// and so portable)
 
class CString
{
public:
    CString();
    CString( char *);
    void operator=( CString  );
    CString(const CString &);
    CString operator+(CString );
    ~CString();
    operator char*();
private:
    char *buf;
};
 
CString operator+(char *s1, CString s2)
{
    CString tmp(s1);
    CString answer;
    answer = tmp + s2;
    return answer;
}
 
CString::operator char *()
{
    return buf;
}
 
CString::CString()
{
    buf = new char[1];
    buf[0] = '\0';
}
CString::CString( char *s)
{
    buf = new char[ strlen(s) + 1 ];
    strcpy(buf, s);
}
 
CString::~CString()
{
    delete buf;
}
 
CString::CString(  const CString &cs )
{
    buf = new char[ strlen(cs.buf) + 1 ];
    strcpy(buf, cs.buf);
    
}
 
void CString::operator=(CString cs)
{
    buf = new char[ strlen(cs.buf) + 1 ];
    strcpy(buf, cs.buf);
}
 
CString CString::operator+ (CString cs)
{
    char *tmp = new char[ strlen(buf) + strlen(cs.buf) + 1 ];
    sprintf(tmp, "%s%s", buf, cs.buf);
    CString answer(tmp);
    return answer;
}
 
 
 
//  routine for generating new variable names.
CString gensym()
{
    static int count = 0;
    char tmp[10];
    sprintf(tmp, "xx%d", count);
    count++;
    return CString(tmp);
}
 
 
 
//  routine for introducing a new variable.  Typically the
//  input string is some arithmetic expression of other variable names.
//     -> Prints the assignment expression to the output file.
//
CString expression( CString input )
{
    CString answer = gensym();
    fprintf(autocodeFP, "    FloatType %s = %s;%s",
	(char *)answer, (char *) input,
	(outputType == FigureVMacro) ? " \\\n" : "\n");
    return answer;
}
 
 
 
 
 
 
///////////////////   now for the string-generating computations.
///////////////////   With the CString objects replaced by floats
///////////////////   and the arithmetic expressions replaced by the
///////////////////   operations, these will calculate the matrix
///////////////////   multiplication.  A handy check.
 
 
// DCT:  the full computation described in steps 17 to 19 of the
// IEEE paper.
//   B = answer vector.  (can't remember why I chose B.  Vanity?)
//   N = size of DCT
//   X = input vector
void DCT( CString *B, int N, CString *X )
{
    // work space vars -- assume 64 enough. only need N.
    CString XG[64];  // work space for recursion input
    CString XH[64];  // work space for recursion input 
    CString BG[64];  // work space for recursion result
    CString BH[64];  // work space for recursion result
    int i;
     
    // assert N is one of 2, 4, 8, or 16.
    
    if ( N == 2 )
    {
        //B[0] = C((1*0),4) * X[0] + C((1*1),4) * X[1];
        //B[1] = C((3*0),4) * X[0] + C((3*1),4) * X[1];
        
        // a little bit of streamlining here
        CString tmp = expression( "C16 * " + X[1] );
        //C(0,4) == 1
                
        B[0] =  expression(X[0] + " + " + tmp );
        B[1] =  expression(X[0] + " - " + tmp );
    } else
    {
        for ( i = 0; i+i < N; i++ )
        {
            XG[i] = X[2*i];
            if ( i == 0 )
            {
                XH[i] = X[2*i+1];
            } else
            {
                XH[i] =  expression(CString("(") + X[2*i-1]
						 + " + "
						 + X[2*i+1] 
						 + ")");
            }
        }
        DCT ( BG, N/2, XG );
        DCT ( BH, N/2, XH );
        for ( i = 0; i+i < N; i++ )
        {
            char tmp[50];
            sprintf(tmp, "INV_2C%02d * ", (2*i+1)*(32/N));
            CString expr = expression( CString(tmp) + BH[i] );
            
            B[i]     = expression(BG[i] + " + " + expr);            
            B[N-1-i] = expression(BG[i] + " - " + expr);
        }
    }
}
 
 
 
 
 
//  Here's the solution to the full matrix multiplication, V = N*S,
//  or rather, DCT_OUTPUT = N*DCT_INPUT.
//
//  NOTE!  some of the NVR routines, for efficiency, only calculate
//  every 4th term of the result vector.  Works fine for low sample
//  rates (by definition, I guess).  Provide an #ifdef in the output
//  code to switch to this.  It won't be 4 times as fast, since
//  the calculation of the rows is so interleaved, but it will be
//  moderately faster once the optimizer throws out any unused
//  upstream calculations.
void
BB2_SubbandFilter_Mono(
           CString S[32])
{
    int i, k;
    
    CString A[31]; // answer -- becomes V with minimal fuss
    // first branch
    CString E[16]; // input rows folded for even lines
    CString O[16]; // input rows folded for odd lines
    CString OX[16]; // input rows modified for DCT input as "X"
    CString OB[16]; // DCT result
    // second branch
    CString EE[8];
    CString EO[8];
    CString EOX[8];
    CString EOB[8];
    // third branch
    CString EEE[4];
    CString EEO[4];
    CString EEOX[4];
    CString EEOB[4];
    // fourth branch
    CString EEEE[2];
    CString EEEO[2];
    CString EEEOX[2];
    CString EEEOB[2];
    
    char *lineSuffix = (outputType == FigureVMacro) ? "  \\\n" : "\n";

    
    
    
    // Compute A[0..31], which is the negative of V[48..17].
    // (Later, we'll translate and make V[0..63].)
    //
    //  E[k] = folding of input to use on even numbered lines,
    //         which are symmetrical
    //  O[k] = folding of input to use on odd numbered lines,
    //         which are anti-symmetrical.
    // OX[k] = add neighboring elements of the input -- the trick
    //         used in steps 11-16 of the IEEE paper.
    //
    
 
    // first branch    
    for ( k = 0; k < 16; k++ )
    {
        //E[k] = S[k] + S[31-k];
        //O[k] = S[k] - S[31-k];
        //OX[k] =  ( k > 0 ) ? (O[k] + O[k-1]) : O[k];
        E[k] = expression( CString("(") + S[k] + " + " + S[31-k] + ")" );
        O[k] = expression( CString("(") + S[k] + " - " + S[31-k] + ")" );
        OX[k] = (k>0)
	      ?  expression(CString("(") + O[k] + " + " + O[k-1] + ")") : O[k];
    }
    DCT(OB, 16, OX);
 
    
    // second branch
    for ( k = 0; k < 8; k++ )
    {
        //EE[k] = E[k] + E[15-k];
        //EO[k] = E[k] - E[15-k];
        //EOX[k] =  ( k > 0 ) ? (EO[k] + EO[k-1]) : EO[k];
        EE[k] = expression(CString("(") + E[k] + " + " + E[15-k] + ")"  );
        EO[k] = expression(CString("(") + E[k] + " - " + E[15-k] + ")"  );
        EOX[k] = (k>0)
	       ? expression(CString("(") + EO[k] + " + " + EO[k-1] + ")")
	       : EO[k];
    }
    DCT(EOB, 8, EOX);
 
 
    // third branch
    for ( k = 0; k < 4; k++ )
    {
        //EEE[k] = EE[k] + EE[7-k];
        //EEO[k] = EE[k] - EE[7-k];
        //EEOX[k] =  ( k > 0 ) ? EEO[k] + EEO[k-1] : EEO[k];
        EEE[k] = expression(CString("(") + EE[k] + " + " + EE[7-k] + ")"  );
        EEO[k] = expression(CString("(") + EE[k] + " - " + EE[7-k] + ")"  );
        EEOX[k] = (k>0)
		?  expression(CString("(") + EEO[k] + " + " + EEO[k-1] + ")")
		: EEO[k];
    }
    DCT(EEOB, 4, EEOX);
 
 
    // fourth branch
    for ( k = 0; k < 2; k++ )
    {
        //EEEE[k] = EEE[k] + EEE[3-k];
        //EEEO[k] = EEE[k] - EEE[3-k];
        //EEEOX[k] =  ( k > 0 ) ? EEEO[k] + EEEO[k-1] : EEEO[k];
        EEEE[k] = expression(CString("(") + EEE[k] + " + " + EEE[3-k] + ")");
        EEEO[k] = expression(CString("(") + EEE[k] + " - " + EEE[3-k] + ")");
        EEEOX[k] = (k>0)
		 ?  expression(CString("(") + EEEO[k] + " + " + EEEO[k-1] + ")")
		 : EEEO[k];
    }
    DCT(EEEOB, 2, EEEOX);
 
    
    // and the tail
    //A[0] = /* C(0,64)=1 * */ (EEEE[0] + EEEE[1]);  
    //A[16] =   C(16,64)  *    (EEEE[0] - EEEE[1]);
    A[0]  = expression(                     EEEE[0] + " + " + EEEE[1]);
    A[16] = expression(CString("C16 * (") + EEEE[0] + " - " + EEEE[1] + ")");
 
 
    
    // reassemble from the branches
    for ( i = 0; i < 2; i++ )
    {
        //A[16*i+8] = (1.0/(2.0*C(16*i+8,64))) * EEEOB[i];
            char tmp[50];
            sprintf(tmp, "INV_2C%02d * ", 16*i+8);
            
        A[16*i+8] = expression(CString(tmp) + EEEOB[i]);
    }
    for ( i = 0; i < 4; i++ )
    {
        //A[8*i+4] = (1.0/(2.0*C(8*i+4,64))) * EEOB[i];
                    char tmp[50];
            sprintf(tmp, "INV_2C%02d * ", 8*i+4);
            
        A[8*i+4] = expression(CString(tmp) + EEOB[i]);
 
    }
    for ( i = 0; i < 8; i++ )
    {
        //A[4*i+2] = (1.0/(2.0*C(4*i+2,64))) * EOB[i];
                    char tmp[50];
            sprintf(tmp, "INV_2C%02d * ", 4*i+2);
            
        A[4*i+2] = expression(CString(tmp) + EOB[i]);
 
    }
    for ( i = 0; i < 16; i++ )
    {
        //A[2*i+1] = (1.0/(2.0*C(2*i+1,64))) * OB[i];
            char tmp[50];
            sprintf(tmp, "INV_2C%02d * ", 2*i+1);
            
        A[2*i+1] = expression(CString(tmp) + OB[i]);
    }
    
    // by now, all the calculations have been made and printed. 
    // now, print out the assignments into the result variables.
    // optimizations:
    //  - if(_lazy) for only using every 4th output variable
    //  - precompute the starting index of the result array; saves
    //    a few additions.
    //  - unroll the final assignments.  Mainly this is because
    //    in the _lazy case, compilers I've looked at aren't
    //    clever enough not to compute the unused values when there's
    //    a loop in the way.
    fprintf(autocodeFP, "    {%s", lineSuffix);
    fprintf(autocodeFP, "        FloatType *outputBase = _output;%s",
	lineSuffix);
    fprintf(autocodeFP, "        if ( _lazy ) {%s", lineSuffix);
    char *prefix = "            ";
    for ( i = 0; i < 64; i+= 4 )
    {
        if ( i < 16 )
        {
            fprintf(autocodeFP, "%soutputBase[%d] = %s;%s",
		prefix, i, (char*) A[16+i], lineSuffix);
        } else if ( i == 16 )
        {
            fprintf(autocodeFP, "%soutputBase[16] =(FloatType)0.0;%s",
		prefix, lineSuffix);
        } else if ( i <= 48 )
        {
            fprintf(autocodeFP, "%soutputBase[%d] = - %s;%s",
		prefix, i, (char*) A[48-i], lineSuffix );
        } else /* i > 48 */
        {
            fprintf(autocodeFP, "%soutputBase[%d] = - %s;%s",
		prefix, i, (char*) A[i-48], lineSuffix );
        }
    }
    
    fprintf(autocodeFP, "        } else {%s", lineSuffix);
    
    for ( i = 0; i < 64; i++ )
    {
        if ( i < 16 )
        {
            fprintf(autocodeFP, "%soutputBase[%d] = %s;%s",
		prefix, i, (char*) A[16+i], lineSuffix );
        } else if ( i == 16 )
        {
            fprintf(autocodeFP, "%soutputBase[16] = (FloatType)0.0;%s",
		prefix, lineSuffix);
        } else if ( i <= 48 )
        {
            fprintf(autocodeFP, "%soutputBase[%d] = - %s;%s",
		prefix, i, (char*) A[48-i], lineSuffix );
        } else /* i > 48 */
        {
            fprintf(autocodeFP, "%soutputBase[%d] = - %s;%s",
		prefix, i, (char*) A[i-48], lineSuffix );
        }
    }
    fprintf(autocodeFP, "        }%s", lineSuffix);
    fprintf(autocodeFP, "    }%s", lineSuffix);
    
}
 
 
static void
WriteAutoCodeDecl(FILE *fp, boolean prototype)
{
    if (prototype) {
	fprintf(fp, "CExternal void ");
	fprintf(fp, "figure_V_the_autocode_way(\n");
	fprintf(fp, "    FloatType *_input,\n");
	fprintf(fp, "    FloatType*_output,\n");
	fprintf(fp, "    int _lazy\n");
	fprintf(fp, ");\n");
    } else {
	fprintf(fp, "void\n");
	fprintf(fp, "figure_V_the_autocode_way(FloatType *_input,\n");
	fprintf(fp, "                          FloatType*_output,\n");
	fprintf(fp, "                          int _lazy)");
	fprintf(fp, "\n{\n");
    }
}
 
// put the #defines for the various cos(n) constants in
// their own file.  We know ahead of time which ones
// we'll need.
// format:  just a comment and some plain #defines.
void write_cosdefs()
{
    FILE *cosdefsFP = fopen("MAcosdefs.h", "w");
    fprintf(cosdefsFP, "/* This file generated automatically. */\n");
    fprintf(cosdefsFP, "/* Modify at your own risk. */\n");
    fprintf(cosdefsFP, "/* note! may need to use doubles instead */\n");
    fprintf(cosdefsFP, "/* of floats, to combat roundoff error.  */\n");
    fprintf(cosdefsFP, "/* INV_2Cnn means 1/( 2*cos(nn*pi/64)).  */\n");
    fprintf(cosdefsFP, "/* C16      means       cos(16*pi/64).   */\n");
    if (outputType == FigureVMacro)
	fprintf(cosdefsFP, "#include \"MAautocode.h\"\n");
    else
	WriteAutoCodeDecl(cosdefsFP, True);

    for ( int i = 0; i < 32; i++ )
    {
       fprintf(cosdefsFP, "#define INV_2C%02d ((float)%12.10g)\n",
                          i,
                          1.0/(2.0*cos( (M_PI * i)/64.0 )));
    }
    fprintf(cosdefsFP, "#define C16 ((float)%12.10g)\n",
                       cos(M_PI/4) );
    fclose(cosdefsFP);
}
 
 
 
#define UsageMacro(args)	Msg args; Usage();
 
char *ProgramName;

static void
Usage(void)
{
    NormalOutput("Usage: %s [macro|subr]\n", ProgramName);
    exit(1);
}

void
main(int argc, char **argv)
{
    ProgramName = *argv;

    for (argc--, argv++; argc; argc--, argv++) {
	if (strcmp(*argv, "macro") == 0)
	    outputType = FigureVMacro;
	else if (strcmp(*argv, "subr") == 0) {
	    outputType = FigureVSubr;
	} else {
	    UsageMacro(("unknown output type: %s\n", *argv));
	}
    }
    if (outputType == NoOutput) {
	UsageMacro(("unspecified output type\n"));
    }

    write_cosdefs();
    
 
 
    // format:
    //    comments
    //    #define figure_V_macro( _inputbase, _outputbase, _lazy) \ 
    //    { \  
    //        ... whole lot of float xxNN = computation ... \   
    //        { \       
    //            tmpvar = _outbase; \ 
    //            if ( _lazy ) { \  
    //                ... whole bunch of assignments ..  \  
    //            } else { \  
    //                ... whole bunch of assignments ... \   
    //            } \ 
    //        } \ 
    //     }
    //
    //  yes, it's a LOT of lines for a macro, but compilers seem
    //  to accept it.
    //
    if (outputType == FigureVMacro)
	autocodeFP = fopen("MAautocode.h", "w");
    else {
	autocodeFP = fopen("MAautocode.c", "w");
	fprintf(autocodeFP, "#include \"MAcommon.h\"\n");
	fprintf(autocodeFP, "#include \"MAcosdefs.h\"\n");
    }
 
    time_t ltime;
    time( &ltime );
    fprintf(autocodeFP, "/* This file generated automatically at */\n");
    fprintf(autocodeFP, "/*   %s\n */\n", ctime(&ltime));
    fprintf(autocodeFP, "/* Do not modify. */\n");
    fprintf(autocodeFP, "\n");
    fprintf(autocodeFP, "/* Assumes typedef for 'FloatType'. */\n");
    fprintf(autocodeFP, "/* Assumes #defines from cosdefs.h  */\n");
    fprintf(autocodeFP, "/* _input:   FloatType[32]          */\n");
    fprintf(autocodeFP, "/* _output:  FloatType[64]          */\n");
    fprintf(autocodeFP, "/* _lazy:    1 or 0 (will be compiled out). */\n");
    fprintf(autocodeFP, "\n");
    
    // start with a brace so that the declarations we put out
    // won't happen oops to follow non-declarations, giving a C
    // syntax error.

    if (outputType == FigureVMacro) {
	fprintf(autocodeFP, "#define figure_V_the_autocode_way" );
	fprintf(autocodeFP, "( _input, _output, _lazy ) \\\n" );
	fprintf(autocodeFP, "{       \\\n");
    } else {
	WriteAutoCodeDecl(autocodeFP, False);
    }
 
    CString S[32];   // S:  the input vector (as per MPEG spec doc)
    for ( int  i = 0; i < 32; i++ )
    {
        char tmp[50];
        sprintf(tmp, "_input[%d]", i);
        S[i] = tmp;
    }
 
    BB2_SubbandFilter_Mono(S );  // no output args; autocodeFP is the output.
 
    fprintf(autocodeFP, "}\n"); // no continuation 
    
    fclose(autocodeFP);

    exit (0);
}
