/*
 * Copyright (c) 1992, 1993, 1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	main.c - main for ampegdecomp
 *
 * DESCRIPTION
 *	decompresses an MPEG audio bitstream or system stream
 *	to raw PCM
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/main.c,v 1.13 1994/10/02 17:44:07 toddb Exp $
 */
#include <stdarg.h>
#include <stdio.h>
#include "MAcommon.h"
#include "fileContent.h"
#include "nvrImage.h"

#if DEBUG_ENABLED
unsigned long	DebugFlags[3];

static char *DebugNames[] = {
    "",
    "syncword",
    "ID",
    "layer",
    "protection_bit",
    "bitrate_index",
    "sampling_frequency",
    "padding_bit",
    "private_bit",
    "mode",
    "mode_extension",
    "copyright",
    "original_home",
    "emphasis",
    "crc_check",
    "allocation",
    "scale_factor",
    "sample",

    "sample data",
    "allocation data",
    "scale data",
    "timer",
    "granules",
    "reset sb samples",
    "audio sequence info",
    "audio frame",
};

static UChar MarkBits[] = {
    0,	/* no marks yet */
};

static UChar BitsBits[] = {
    N_syncword,
    N_ID,
    N_layer,
    N_protection_bit,
    N_bitrate_index,
    N_sampling_frequency,
    N_padding_bit,
    N_private_bit,
    N_mode,
    N_mode_extension,
    N_copyright,
    N_original_home,
    N_emphasis,
    N_crc_check,
    N_allocation,
    N_scale_factor,
    N_sample,
};

static FlagMacros DebugMacros[] = {
    { "marks", MarkBits, NumberOfCells(MarkBits) },
    { "bits", BitsBits, NumberOfCells(BitsBits) },
};

FlagLabels DebugInfo = {
    DebugNames,
    NumberOfCells(DebugNames),
    DebugMacros,
    NumberOfCells(DebugMacros),
};

#endif /* DEBUG_ENABLED */

#define DEFAULT_QUALITY          32

char *ProgramName;

static void
Usage(char *format, ...)
{
    char buf[ 1024 ];
    va_list ap;

    va_start(ap, format);
    vsprintf(buf, format, ap);
    va_end(ap);

    ProgMsg("%s", buf);
    Msg("Usage: %s [-8khz] [-ulaw] -aiff|-au mpegfile outfile\n\n",
	ProgramName);
    Msg("where the flags have the following meaning:\n");
    Msg("    -8khz    produce 8khz/16bit output\n");
    Msg("    -ulaw    produce 8khz/8bit/mu-law output\n");
    Msg("    -aiff    produce AIFF output\n");
    Msg("    -au      produce Sun Audio output\n");
    Msg("    -qN      set quality of output to N, where 1 <= N <= 31\n");
    Msg("The 'mpegfile' may be an mpeg audio stream or a full system\n");
    Msg("stream.  The 'outfile' is either an AIFF file or a Sun Audio file,\n");
    Msg("so -aiff or -au must be specified.  If -8khz or -ulaw are not\n");
    Msg("specified, then full bandwidth audio is produced (32kHz,\n");
    Msg("44.2kHz or 48kHz).\n\n");

    Msg("The -q flag with a low setting (e.g., 1-7) tends to speed up\n");
    Msg("processing at the expense of quality.  The -8khz or -ulaw\n");
    Msg("result in substantial processing speedup, but quality is lost.\n");
    exit(1);
}

/*
 * Standalone version of MPEG audio decoder.
 *
 * audio_decode <audioinput file> > <PCM audio output file>
 *
 * A raw audio bitstream is read and converted to 16-bit PCM samples (ie.
 * signed integers).  Messages on standard error will give the give the
 * number of samples per second and the number of channels.
 */

void
main(argc, argv)
char **argv;
{
    MpegAudioInfoPtr audioInfo;
    int aQuality = DEFAULT_QUALITY;
    FileInfo fileInfo;
    ImageReasonPtr irp;
    long nSamps = 0;
    int i, sampleShift;
    char *inputFile, *outputFile;
    boolean aiffOutput = False;
    boolean auOutput = False;

    ProgramName = *argv;

    /*
     * Note: when allocating fifo space, we must be careful.
     *	audioInfo->fifo[] by default is FIFO_SIZE which is enough to hold
     *	one frame of stereo LayerII audio (2*1152 samples).
     *  MAgatherSampes() operates in frame-size granularity and will decode
     *	to the next greater frame boundary.
     *	However if audioInfo->eightKOnly is set, may not evenly divide FIFO_SIZE,
     *	so we must ask for FIFO_SIZE/8 less samples than we have room for
     */
    i = sizeof(MpegAudioInfo) + sizeof(short) * FIFO_SIZE / 8;
    audioInfo = (MpegAudioInfoPtr) InsecureAlloc(i);
    if (audioInfo == NULL)
	Fatal("can't allocate space for mpeg audio\n");
    memset((char *)audioInfo, 0, i);

    inputFile = outputFile = NULL;
    for (argc--, argv++; argc; argc--, argv++) {
	if( strncmp(argv[0],"-q",2) == 0 ) {
	    aQuality = atoi(&argv[0][2]);
	    if( aQuality < -1 || aQuality > 32 ) {
		aQuality = DEFAULT_QUALITY;
	    }
	} else if( strcmp(argv[0],"-au") == 0 ) {
	    auOutput = True;
	} else if( strcmp(argv[0],"-aiff") == 0 ) {
	    aiffOutput = True;
	} else if( strcmp(argv[0],"-ulaw") == 0 ) {
	    audioInfo->eightKOnly = True;
	    audioInfo->ulawConversion = True;
	} else if( strcmp(argv[0],"-8khz") == 0 ) {
	    audioInfo->eightKOnly = True;
	} else if( argv[0][0] == '-' ) {
	    Usage("\"%s\" is an unrecognized flag\n", argv[0]);
	} else if (inputFile == NULL) {
	    inputFile = argv[0];
	} else if (outputFile == NULL) {
	    outputFile = argv[0];
	} else {
	    Usage("This looks like a third file name: \"%s\"\n", argv[0]);
	}
    }
    SetDebugFileFlags(DebugFlags, &DebugInfo, "ampegdecomp.debug");
    if (inputFile == NULL)
	Usage("Missing input and output file\n");
    if (outputFile == NULL)
	Usage("Missing output file\n");
    if (aiffOutput == auOutput)
	Usage("One, and only one of flags -aiff or -au is required\n");
    if (audioInfo->ulawConversion && aiffOutput)
	Usage("Sorry, can't put mu-law data in an aiff file.\n");

    audioInfo->quality = aQuality;
    audioInfo->bitStream = (BitStreamPtr)memalign(8, sizeof(BitStream));
    memset((char *)audioInfo->bitStream, 0, sizeof(BitStream));
    audioInfo->outputFd = OpenFileForWriting(outputFile);
    audioInfo->fifoSize = FIFO_SIZE + FIFO_SIZE / 8;
    if( ! NewMaSetupDecode(audioInfo, inputFile, NULL) ) {
	exit(1);
    }

    sampleShift = audioInfo->ulawConversion ? 0 : 1;
    while( MAgatherSamples(audioInfo, FIFO_SIZE) == True ) {
	(void) WriteData(audioInfo->outputFd,
			 (char *)audioInfo->fifo,
			 audioInfo->nFifoEntries << sampleShift);
	audioInfo->nPreviousFifoEntries = audioInfo->nFifoEntries;
	nSamps += audioInfo->nFifoEntries;
	audioInfo->nFifoEntries = 0;
    }
    exit(0);
}

