/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	macdevice.c - mac functions for playing audio
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/macdevice.c,v 1.5 1995/03/10 08:56:00 toddb Exp $
 *
 */

#include 	"nvrAudioI.h"
#include 	"nvrAlloc.h"
#include 	"macEvent.h"

#include <Sound.h>
#include <GestaltEqu.h>

#define kMiddleC	60	/* not in our version of Sound.h */

#ifdef THINK_C
enum {
  gestalt16BitSoundIO = 7,
  gestalt16BitAudioSupport = 12
};
#endif

extern int errno;

/* State Variables */
typedef union _soundHdrUnion {
    ExtSoundHeader ext;
    SoundHeader std;
} SoundHeaderUnion, *SoundHeaderUnionPtr;

typedef struct _soundCmdInfo SoundCmdInfo, *SoundCmdInfoPtr;

typedef struct _macDevState {
    SndChannelPtr chan;
    short encode;
    short requestsPending;
    SoundHeaderUnion proto;	/* prototype for sound header used in requests */
    SoundCmdInfoPtr active;	/* linked list of active buffers */
} MacDevState, *MacDevStatePtr;

struct _soundCmdInfo {
    MacDevStatePtr macState;
    boolean complete;
    long nSamples;		/* samples queued for this request */
    SoundCmdInfoPtr next;	/* used for linked list in active, above */
    SoundHeaderUnion hdr;
}; /* typedef: SoundCmdInfo, *SoundCmdInfoPtr */

static InternalReason AudioCmdRejected =
    { "Audio command rejected by sound manager.", IR_InternalProblem };
static InternalReason NoHardware =
    { "No audio hardware is available.", IR_BadHardware };
static InternalReason CantRead =
    { "Can't read mac device audio yet.\n", IR_NotImplemented };
static InternalReason CantWrite =
    { "Can't write mac device audio yet.\n", IR_NotImplemented };
static InternalReason CantOpen =
    { "Can't open mac device audio yet.\n", IR_NotImplemented };
static InternalReason CantInit =
    { "Can't init mac device audio yet.\n", IR_NotImplemented };
static InternalReason CantMode =
    { "Can't set mac device audio mode yet.\n", IR_NotImplemented };



static boolean
AudioError(AudioFilePtr auFile, char *string)
{
    SetAudioReason(auFile, CreateSysReason("%s", string));
    return(False);
}

static boolean
macAudioOpen(AudioFilePtr auFile)
{
    MacDevStatePtr macState;
    long attr;

    if (errno = Gestalt(gestaltSoundAttr, &attr))
        return( AudioError(auFile, "Can't get sound attributes") );

    if ((attr & gestaltSoundIOMgrPresent) == 0) {
        SetAudioReason(auFile, &NoHardware);
        return(False);
    }
    
    auFile->u.devFormats = AvailMono
			 | AvailUSign8BitPcm
			 | Avail22_254Khz;
    if (attr & gestaltStereoCapability)
        auFile->u.devFormats |= AvailStereo;

    /*
     * If 16-bit sound is present, assume we can do just about anything.
     */
    if (attr & gestalt16BitSoundIO)
        auFile->u.devFormats |= AvailUSign16BitPcm
                              | Avail48Khz
                              | Avail44_1Khz
                              | Avail32Khz
                              | Avail22_05Khz
                              | Avail8Khz
                              | AvailStereo;

    if (! AllocAudioState(auFile, sizeof(MacDevState)) )
        return(False);
    macState = (MacDevStatePtr) auFile->io->state;

    return(True);
}

static void
FreeBuffers(MacDevStatePtr macState, boolean forceComplete)
{
    SoundCmdInfoPtr active;
    SoundCmdInfoPtr this, next;
    
    active = nil;
    this = macState->active;
    while (this) {
        next = this->next;
        if (this->complete || forceComplete) {
            free((char *)this);
        } else {
            this->next = active;
            active = this;
        }
        this = next;
    }
    macState->active = active;
}

static void
macAudioClose(AudioFilePtr auFile)
{
    MacDevStatePtr macState = (MacDevStatePtr) auFile->io->state;
    
    while (macState->active) {
        FreeBuffers(macState, False);
        ProcessEvents();
    }
    if (macState->chan) {
        SndDisposeChannel(macState->chan, False);
        macState->chan = nil;
    }
}

static long
macAudioSamplesQueued(AudioFilePtr auFile)
{
    MacDevStatePtr macState = (MacDevStatePtr) auFile->io->state;
    SoundCmdInfoPtr this;
    long queued = 0;
    
    for (this = macState->active; this; this = this->next) {
        if (! this->complete)
            queued += this->nSamples;
    }
    return(queued);
}

static void
macAudioFlush(AudioFilePtr auFile)
{
    SndCommand cmd;
    MacDevStatePtr macState = (MacDevStatePtr) auFile->io->state;
    
    if (macState->active == nil)
    	return;
    /*
     * Stop the channel output
     */
    cmd.cmd = flushCmd;
    cmd.param1 = (short)0;
    cmd.param2 = (long)0;
    SndDoImmediate(macState->chan, &cmd);
    cmd.cmd = quietCmd;
    SndDoImmediate(macState->chan, &cmd);

    FreeBuffers(macState, True);
}

static boolean
BadRate(AudioFilePtr auFile)
{
    SetAudioReason(auFile,
                   CreateReason("Bad rate: %d kHz.\n", auFile->au.rate));
    return(False);
}

static boolean
BadDataForm(AudioFilePtr auFile)
{
    char *bits;

    if ((bits = EnumToName(auFile->au.dataForm, AudioBitNames)) == NULL)
        bits = "?pcm";
    SetAudioReason(auFile,
                   CreateReason("Bad data form: %s kHz.\n", bits));
    return(False);
}

static boolean
BadChannels(AudioFilePtr auFile)
{
    SetAudioReason(auFile,
                   CreateReason("Bad channel count: %d.\n", auFile->au.channels));
    return(False);
}

static pascal void
soundCallBackProc(SndChannelPtr chan, SndCommand *cmd)
{
    SoundCmdInfoPtr cmdInfo = (SoundCmdInfoPtr) cmd->param2;
    MacDevStatePtr macState = cmdInfo->macState;
    
    /*
     * Since we are called at interrupt level, we can't free
     * anything here.  Put it on the linked list to be freed later.
     */
    macState->requestsPending--;
    cmdInfo->complete = True;
}

static boolean
macAudioMode(AudioFilePtr auFile, long devCommand)
{
    SetAudioReason(auFile, &CantMode);
    return(False);
}

static boolean
macAudioInit(AudioFilePtr auFile)
{
    MacDevStatePtr macState = (MacDevStatePtr) auFile->io->state;
    static SndCallBackUPP soundCb;
    long init;
    Fixed sampleRate;
    short sampleSize;

    memset((char *)&macState->proto, 0, sizeof(macState->proto));

    switch (auFile->au.rate) {
    case Audio48Khz:     sampleRate = 48000 << 16; break;
    case Audio44_1Khz:   sampleRate = 44100 << 16; break;
    case Audio32Khz:     sampleRate = 32000 << 16; break;
    case Audio22_254Khz: sampleRate = rate22khz; break;
    case Audio22_05Khz:  sampleRate = 22050 << 16; break;
    case Audio16Khz:     sampleRate = 16000 << 16; break;
    case Audio11_025Khz: sampleRate = 11025 << 16; break;
    case Audio8Khz:      sampleRate =  8000 << 16; break;
    default:             return(BadRate(auFile));
    }
    
    switch (auFile->au.dataForm) {
    case Audio24BitPcm:      sampleSize = 24; break;
    case Audio16BitPcm:      sampleSize = 16; break;
    case AudioUSign16BitPcm: sampleSize = 16; break;
    case Audio8BitPcm:       sampleSize = 8; break;
    case AudioUSign8BitPcm:  sampleSize = 8; break;
    default:                 return(BadDataForm(auFile));
    }

    if (auFile->au.channels == 1 && auFile->au.dataForm == AudioUSign8BitPcm) {
        macState->encode = macState->proto.std.encode = stdSH;
        macState->proto.std.sampleRate = sampleRate;
        macState->proto.std.baseFrequency = kMiddleC;
    } else {
        macState->encode = macState->proto.ext.encode = extSH;
        macState->proto.ext.sampleSize = sampleSize;
        macState->proto.ext.numChannels = auFile->au.channels;
        macState->proto.ext.sampleRate = sampleRate;
        macState->proto.ext.baseFrequency = kMiddleC;
        ConvertToIeeeExtended((double)auFile->au.rate,
                              (char *)&macState->proto.ext.AIFFSampleRate);
    }

    init = 0;
    switch (auFile->au.channels) {
    case AudioMono:   init |= initMono; break;
    case AudioStereo: init |= initStereo; break;
    default:          return(BadChannels(auFile));
    }

    macState->chan = nil;
    if (soundCb == nil)
        soundCb = NewSndCallBackProc(soundCallBackProc);
    errno = SndNewChannel(&macState->chan, sampledSynth, init, soundCb);

    if (errno)
        return( AudioError(auFile, "Can't get sound channel.") );

    return(True);
}

static long
macAudioReadSamples(AudioFilePtr auFile, UChar *buffer, ULong nSamples)
{
    SetAudioReason(auFile, &CantRead);
    return(-1);
}

static long
macAudioWriteSamples(AudioFilePtr auFile, UChar *buffer, ULong nSamples)
{
    MacDevStatePtr macState = (MacDevStatePtr) auFile->io->state;
    SndCommand cmd;
    SoundCmdInfoPtr cmdInfo;
    long bufferSize;

    /*
     * Allocate new memory for the request.  These will be freed after
     * the request completes.
     */
    bufferSize = (nSamples << DataFormToShift[ auFile->au.dataForm ]);
    for (;;) {
        if (macState->active)
            FreeBuffers(macState, False);
        cmdInfo = (SoundCmdInfoPtr)
              InsecureAlloc(bufferSize + sizeof(SoundCmdInfo));
        if (cmdInfo == NULL) {
            if (macState->requestsPending == 0) {
                SetAudioReason(auFile, &IRBadAllocation);
                return (-1);
            }
            ProcessEvents();
        } else
            break;
    }

    memcpy((char *)&cmdInfo->hdr,
           (char *)&macState->proto,
           sizeof(SoundHeaderUnion));
    cmdInfo->macState = macState;
    cmdInfo->complete = False;
    cmdInfo->nSamples = nSamples;
    cmdInfo->next = macState->active;
    macState->active = cmdInfo;

    switch (macState->encode) {
    case stdSH:
        cmdInfo->hdr.std.length = nSamples;
        memcpy((char *)cmdInfo->hdr.std.sampleArea, buffer, bufferSize);
        break;
    case extSH:
        cmdInfo->hdr.ext.numFrames = nSamples >> (auFile->au.channels-1);
        memcpy((char *)cmdInfo->hdr.ext.sampleArea, buffer, bufferSize);
        break;
    }
    cmd.cmd = bufferCmd;
    cmd.param1 = 0;
    cmd.param2 = (long)&cmdInfo->hdr;
    if (errno = SndDoCommand(macState->chan, &cmd, False)) {
        free((char *)cmdInfo);
        return( AudioError(auFile, "Can't play audio.") );
    }
    
    cmd.cmd = callBackCmd;
    cmd.param1 = (short)0;
    cmd.param2 = (long)cmdInfo;
    if (errno = SndDoCommand(macState->chan, &cmd, False))
        return( AudioError(auFile, "Audio command rejected.") );
    macState->requestsPending++;
    return(nSamples);
}

/*		External Linkage				*/

AudioDevOpFunc MacDevFuncs = {
    macAudioOpen,
    macAudioInit,
    macAudioMode,
    macAudioReadSamples,
    macAudioWriteSamples,
    macAudioSamplesQueued,
    macAudioFlush,
    macAudioClose
};
