/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	mpegio.c - functions for reading/writing mpeg audio
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/mpegio.c,v 1.7 1995/08/04 23:46:43 toddb Exp $
 */
#define NeedFunctionPrototypes 1

#include "nvrAudioI.h"
#include "nvrAlloc.h"
#include "nvrString.h"
#include "fileAccess.h"
#include "MAcommon.h"

typedef struct _mpegState {
    MpegAudioInfo mdec;
    long unusedSamples;
} *MpegStatePtr;

static InternalReason CantDecode = {
    "Can't decode this mpeg file.\n", IR_InternalProblem
};
static InternalReason CantSeek = {
    "Can't seek on MPEG audio yet.\n", IR_NotImplemented
};
static InternalReason CantEncode = {
    "Can't encode MPEG audio yet.\n", IR_NotImplemented
};
 
static void
CloseMpegAudio(AudioFilePtr auFile)
{
    MpegStatePtr mpegState = (MpegStatePtr) auFile->io->state;

    CloseBitStream(mpegState->mdec.bitStream);
    free((char *)mpegState->mdec.bitStream->buffer);
}

static boolean
CreateMpegAudio(AudioFilePtr auFile)
{
    SetAudioReason(auFile, &CantEncode);
    return(False);
}

static boolean
OpenMpegAudio(AudioFilePtr auFile, AudioQualityPtr quality)
{
    MpegStatePtr mpegState;
    int size;
    double duration;
    long bitrate;

    /*
     * Note: when allocating fifo space, we must be careful.
     * mpegState->fifo[] by default is FIFO_SIZE which is enough to hold
     * one frame of stereo LayerII audio (2*1152 samples).
     * MAgatherSampes() operates in frame-size granularity and will decode
     * to the next greater frame boundary.  However if mpegState->eightKOnly
     * is set, the number of output samples may not evenly divide FIFO_SIZE,
     * so we must allocate FIFO_SIZE/8 more samples than we need.
     */
    size = sizeof(MpegAudioInfo)
	 + sizeof(short) * FIFO_SIZE / 8
	 + sizeof(BitStream);
    if (! AllocAudioState(auFile, size) )
	return(False);
    mpegState = (MpegStatePtr) auFile->io->state;
    mpegState->mdec.bitStream =
	(BitStreamPtr)((ULong)mpegState + (size - sizeof(BitStream)));

    if ( quality ) {
        if ( quality->decodeQuality > .75 ) {         /* full qual */
	    mpegState->mdec.quarterSample = False;
	    mpegState->mdec.monoSample = False;
        } else if ( quality->decodeQuality > .5  ) {  /* full mono*/
	    mpegState->mdec.quarterSample = False;
	    mpegState->mdec.monoSample = True;
        } else if ( quality->decodeQuality > .25 ) {  /* 1/4 stereo */
	    mpegState->mdec.quarterSample = True;
	    mpegState->mdec.monoSample = False;
        } else {                                      /* 1/4 mono */
	    mpegState->mdec.quarterSample = True;
	    mpegState->mdec.monoSample = True;
        }
    } else { /* assume quality is full on */
	mpegState->mdec.quarterSample = False;
	mpegState->mdec.monoSample = False;
    }

    mpegState->mdec.fifoSize = FIFO_SIZE + FIFO_SIZE / 8;

    if( ! NewMaSetupDecode(&mpegState->mdec, auFile->io->name, NULL) ) {
	/*
	 * The library should return us a reason, but doesn't.
	 */
	SetAudioReason(auFile, &CantDecode);
	CloseMpegAudio(auFile);
        return(False);
    }

    /*
     * Suck in the first bit so that we can know the audio parameters
     */
    if (! MAgatherSamples(&mpegState->mdec, FIFO_SIZE)) {
	SetAudioReason(auFile, &CantDecode);
	CloseMpegAudio(auFile);
        return(False);
    }
#if 0
    mpegState->mdec.nPreviousFifoEntries = mpegState->mdec.nFifoEntries;
#endif
    mpegState->unusedSamples = mpegState->mdec.nFifoEntries;
    mpegState->mdec.nFifoEntries = 0;


    auFile->au.dataForm = Audio16BitPcm;

    auFile->au.rate =
        MA_SamplingFrequencyTable[
      	    mpegState->mdec.frameHeader.s.sampling_frequency];
    if (mpegState->mdec.quarterSample)
	auFile->au.rate /= 4;

    if (mpegState->mdec.monoSample) {
	auFile->au.channels = AudioMono;
    } else {
        switch(mpegState->mdec.frameHeader.s.mode) {
        case MODE_STEREO:
        case MODE_JOINT_STEREO:
        case MODE_DUAL_CHANNEL:
            auFile->au.channels = AudioStereo;
            break;

        case MODE_SINGLE_CHANNEL:
            auFile->au.channels = AudioMono;
            break;

        default:
	    /* not possible because frameHeader.s.mode is only 2 bits */
	    SetAudioReason(auFile, &CantDecode);
            return(False);
        }
    }
    
    bitrate = MA_BitRateTable[ mpegState->mdec.frameHeader.s.layer ]
                             [ mpegState->mdec.frameHeader.s.bitrate_index ]
            * (1000 / 8);	/* its in kbits */
    duration = (double) FileSize(auFile->io->name) / bitrate;
    auFile->nSamples = duration
                     * auFile->au.rate
                     * auFile->au.channels;

    return(True);
}

static long
ReadMpegSamples(AudioFilePtr auFile, UChar *buffer, ULong nSamples)
{
    MpegStatePtr mpegState = (MpegStatePtr) auFile->io->state;
    long samplesNeeded = nSamples;
    long dataShift = DataFormToShift[ auFile->au.dataForm ];
    long samplesIn = mpegState->unusedSamples;
    long readyBytes = samplesIn << dataShift;
    char *pDst = (char *)buffer;
    char *pSrc;

    if (readyBytes) {
	pSrc = (char *)mpegState->mdec.fifo
	     + ((FIFO_SIZE<<dataShift) - readyBytes);
	if (readyBytes > (samplesNeeded << dataShift))
	    readyBytes = samplesNeeded << dataShift;
	memcpy(pDst, pSrc, readyBytes);
	pDst += readyBytes;
	samplesNeeded -= readyBytes >> dataShift;
	mpegState->unusedSamples -= readyBytes >> dataShift;
    }

    pSrc = (char *)mpegState->mdec.fifo;
    while (samplesNeeded) {
	if (! MAgatherSamples(&mpegState->mdec, FIFO_SIZE))
	    break;
	samplesIn = mpegState->mdec.nFifoEntries;
#if 0
        mpegState->mdec.nPreviousFifoEntries = mpegState->mdec.nFifoEntries;
#endif
        mpegState->mdec.nFifoEntries = 0;

	readyBytes = samplesIn << dataShift;
	if (readyBytes > (samplesNeeded << dataShift))
	    readyBytes = samplesNeeded << dataShift;
	memcpy(pDst, pSrc, readyBytes);
	samplesNeeded -= readyBytes >> dataShift;
	pDst += readyBytes;
    }
    mpegState->unusedSamples = samplesIn - (readyBytes >> dataShift);

    return(nSamples - samplesNeeded);
}

static long
WriteMpegSamples(AudioFilePtr auFile, UChar *buffer, ULong nSamples)
{
    SetAudioReason(auFile, &CantEncode);
    return(-1);
}

static InternalReasonPtr
SeekMpegSamples(AudioFilePtr auFile, ULong nSamples)
{
    return(&CantSeek);
}

AudioOpFunc AudioLibMpegFuncs = {
    OpenMpegAudio,
    CreateMpegAudio,
    ReadMpegSamples,
    WriteMpegSamples,
    SeekMpegSamples,
    CloseMpegAudio
};
