/*
 * Copyright (c) 1992,1993,1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *      nvrAudio.c - top-level control for audio decoding, independent
 *      of the audio type (mu-law, aiff, mpeg).
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/nvrAudio.c,v 1.33 1996/01/18 20:50:46 toddb Exp $
 */

#define NeedFunctionPrototypes 1

#include "osio.h"
#include "fileContent.h"
#include "fileAccess.h"
#include "nvrAudioI.h"
#include "nvrAlloc.h"
#include "nvrString.h"
#include "mu-law.h"

#include <stdarg.h>	/* needed for SetAudioWarning() */
#include <stdlib.h>	/* needed for abs() */
#include <ctype.h>

short DataFormToBytes[] = { 0, 4, 2, 2, 1, 1, 1 };
short DataFormToShift[] = { 0, 2, 1, 1, 0, 0, 0 };

static InternalReason ImproperIO =
    { "Wrong IO configuration for that operation\n", IR_BadValue };
static InternalReason CantOpen =
    { "Can't open input file\n", IR_BadIOFile };
static InternalReason CantCreate =
    { "Can't create output file\n", IR_BadIOFile };
static InternalReason UnknownInput =
    { "Don't know how to read that kind of input file\n", IR_BadIOFile };
static InternalReason UnknownOutput =
    { "Don't know how to make that kind of output file\n", IR_BadIOFile };
InternalReason CantSeekAudioDevice =
    { "Can't seek on an audio device\n", IR_BadIOFile };
static char *CddaSpecPrefix = "cdda:";

#ifdef sun
extern AudioDevOpFunc SunDevFuncs;
#define DfltDevOps	SunDevFuncs
#endif

#ifdef sgi
extern AudioDevOpFunc SgiDevFuncs;
#define DfltDevOps	SgiDevFuncs
#endif

#if defined(THINK_C) || defined(__MWERKS__)
extern AudioDevOpFunc MacDevFuncs;
#define DfltDevOps	MacDevFuncs
#endif

#ifdef __hpux
extern AudioDevOpFunc NoDevFuncs;
#define DfltDevOps	NoDevFuncs
#endif

static void FreeAudioFile(AudioFilePtr afp)
{
    if (afp->io->state)
	free((char *)afp->io->state);
    free((char *)afp);
}

long
StdAudioReadSamples(AudioFilePtr auFile, UChar *buffer, ULong nSamples)
{
    long nBytes;
    BufFileHndl fp = auFile->io->fp;
    long dataShift = DataFormToShift[ auFile->au.dataForm ];

    nBytes = BufferedReadData(fp, (char *)buffer, nSamples << dataShift);
    if (nBytes <= 0 )
        return(False); /* reached EOF */
    if (dataShift && (nBytes & ((1<<dataShift)-1)) ) {
        SetAudioWarning(auFile,
	    CreateReason("Warning: audio sample shift: asked %d, got %d\n",
			 nSamples << dataShift, nBytes)
	);
    }

    return(nBytes >> dataShift);
}

long
StdAudioWriteSamples(AudioFilePtr auFile, UChar *buffer, ULong nSamples)
{
    long nBytes;
    BufFileHndl fp = auFile->io->fp;
    long dataShift = DataFormToShift[ auFile->au.dataForm ];

    nBytes = BufferedWriteData(fp, (char *)buffer, nSamples << dataShift);
    if (nBytes < (nSamples << dataShift) ) {
	SetAudioReason(auFile, &IRCantWrite);
        return(-1); /* short write */
    }
    if (dataShift && (nBytes & ((1<<dataShift)-1)) ) {
        SetAudioWarning(auFile,
	    CreateReason("Warning: audio sample shift: asked %d, got %d\n",
			 nSamples << dataShift, nBytes)
	);
    }
    auFile->nSamples += nSamples;

    return(nBytes >> dataShift);
}

InternalReasonPtr
StdAudioSeekSamples(AudioFilePtr auFile, ULong nSamples)
{
    BufFileHndl fp = auFile->io->fp;
    long dataShift = DataFormToShift[ auFile->au.dataForm ];
    long offset = nSamples << dataShift;

    SeekBufferedFile(fp, offset + auFile->io->hdrSize, SK_Start);
    if (auFile->io->ioClass == AudioFileOutput)
        auFile->nSamples = nSamples;
    return(NULL);
}

static char *
ParseCddaTime(CddaTimePtr ct, char *ptr)
{
    if (isdigit(*ptr)) {
        ct->min = atoi(ptr);
        while (isdigit(*ptr))
            ptr++;
    }
    if (*ptr == ':')
	ptr++;

    if (isdigit(*ptr)) {
        ct->sec = atoi(ptr);
        while (isdigit(*ptr))
            ptr++;
    }
    if (*ptr == ':')
	ptr++;

    if (isdigit(*ptr)) {
        ct->frame = atoi(ptr);
        while (isdigit(*ptr))
            ptr++;
    }
    return(ptr);
}

/*
 * This function parses the cdda audio spec in auFile->io->name
 * of the form:
 *	cdda:[track #][+offset][,length]
 * The track number is one or more digits.  The offset and length
 * take the form
 *	min:sec:frame
 * where all fields must be present.
 */
static InternalReasonPtr
ParseCddaSpec(AudioFilePtr auFile)
{
    char *ptr = auFile->io->name + strlen(CddaSpecPrefix);
    CddaDescPtr cdda = auFile->io->cdda;

    /*
     * Get the optional track number
     */
    if (isdigit(*ptr)) {
	cdda->track = atoi(ptr);
	while (isdigit(*ptr))
	    ptr++;
    }

    /*
     * Get the optional start time
     */
    if (*ptr == '+') {
	if (NULL == (ptr = ParseCddaTime(&cdda->start, ptr+1))) {
	    return(BadValueReason("Bad CDDA offset specification: %s\n",
			auFile->io->name));
	}
    }

    /*
     * Get the optional length
     */
    if (*ptr == ',') {
	if (NULL == (ptr = ParseCddaTime(&cdda->length, ptr+1))) {
	    return(BadValueReason("Bad CDDA offset specification: %s\n",
			auFile->io->name));
	}
    }
    return (NULL);
}
    
AudioFilePtr
AllocAudioFile(char *path, AudioIoOpClass ioClass)
{
    AudioFilePtr auFile;
    int pathLength, allocSize;
    char *ptr;

    if (path == NULL)
	pathLength = 0;
    else
	pathLength = strlen(path)+1;
    /*
     * Allocate storage 
     */
    allocSize = sizeof(AudioFileRec) + sizeof(AudioIORec) + pathLength;
    if (ioClass == AudioCddaInput)
	allocSize += sizeof(CddaDesc);
    if (NULL == (ptr = InsecureAlloc(allocSize)))
	return( NULL );
    memset(ptr, 0, allocSize);

    auFile = (AudioFilePtr)ptr;
    ptr += sizeof(AudioFileRec);

    auFile->reason = NULL;
    auFile->io = (AudioIOPtr) ptr;
    ptr += sizeof(AudioIORec);

    auFile->io->ioClass = ioClass;
    if (ioClass == AudioCddaInput) {
	auFile->io->cdda = (CddaDescPtr)ptr;
	ptr += sizeof(CddaDesc);
    }

    if (pathLength) {
	auFile->io->name = ptr;
	strcpy(auFile->io->name, path);
    } else {
	auFile->io->name = "";
    }
    auFile->io->state = NULL;
    return( auFile );
}

AudioFilePtr
OpenAudioFile(char *path,
	      AudioQualityPtr quality,
	      InternalReasonPtr *pIrp)
{
    AudioFilePtr auFile;
    FileInfo fileInfo;
    AudioOpFuncPtr ops;
    BufFileHndl fp;
    AudioIoOpClass ioClass;

    if (path == NULL || ! FileReadable(path)) {
	if (strncmp(path, CddaSpecPrefix, strlen(CddaSpecPrefix)) == 0) {
	    ioClass = AudioCddaInput;
	    fileInfo.type = FT_Unknown;
	    ops = &AudioLibCddaFuncs;
	    fp = NULL;
	} else {
	    *pIrp = &CantOpen;
	    return(NULL);
	}
    } else {
	ioClass = AudioFileInput;

	if (NULL != (*pIrp = GetFileContentInfo(path, &fileInfo)))
	    return(NULL);
	switch(fileInfo.type) {
	case FT_QuickTime:
	    ops = &AudioLibQTFuncs;
	    break;
	case FT_AiffAudio:
	case FT_AifcAudio:
	    ops = &AudioLibAiffFuncs;
	    break;
	case FT_SunAudio:
	    ops = &AudioLibAuFuncs;
	    break;
	case FT_Jmovie:
	    ops = &AudioLibJmovieFuncs;
	    break;
	case FT_MpegAudio:
	case FT_MpegSystem:
	    ops = &AudioLibMpegFuncs;
	    break;
	case FT_Unknown:
	default:
	    *pIrp = &UnknownInput;
	    return(NULL);
	}

	fp = OpenBufferedFileForReading(path);
	if (fp == NULL) {
	    *pIrp = &CantOpen;
	    return(NULL);
	}
    }

    /*
     * Allocate storage 
     */
    if ((auFile = AllocAudioFile(path, ioClass)) == NULL) {
	*pIrp = &IRBadAllocation;
	if (fp)
	    CloseBufferedFile(fp);
	return(auFile);
    }
    auFile->u.fileType = fileInfo.type;
    auFile->io->input = True;
    auFile->io->u.fops = *ops;
    auFile->io->fp = fp;

    if (ioClass == AudioCddaInput
     && NULL != (*pIrp = ParseCddaSpec(auFile))) {
	if (fp)
	    CloseBufferedFile(fp);
	FreeAudioFile(auFile);
	return(NULL);
    }

    /*
     * Now call the file functions to really open the file
     * and fill in the auFile->au, and perhaps auFile->io->state.
     */
    if (! (*auFile->io->u.fops.openFile)(auFile, quality) ) {
	*pIrp = auFile->reason;
	if (fp)
	    CloseBufferedFile(fp);
	FreeAudioFile(auFile);
	return(NULL);
    }

    return(auFile);
}

AudioFilePtr
CreateAudioFile(char *path,
		AudioFileConfigPtr conf,
		InternalReasonPtr *pIrp)
{
    BufFileHndl fp;
    AudioFilePtr auFile;
    AudioOpFuncPtr ops;

    switch(conf->type) {
    case FT_QuickTime:
        ops = &AudioLibQTFuncs;
        break;
    case FT_AiffAudio:
    case FT_AifcAudio:
	ops = &AudioLibAiffFuncs;
	break;
    case FT_SunAudio:
	ops = &AudioLibAuFuncs;
	break;
    case FT_Jmovie:
	ops = &AudioLibJmovieFuncs;
	break;
    case FT_MpegAudio:
    case FT_MpegSystem:
	ops = &AudioLibMpegFuncs;
	break;
    default:
	*pIrp = &UnknownOutput;
	return(NULL);
    }

    if (path)
	fp = OpenBufferedFileForWriting(path);
    else
	fp = NULL;
    if (fp == NULL) {
	*pIrp = &CantCreate;
	return(NULL);
    }

    /*
     * Allocate storage 
     */
    if ((auFile = AllocAudioFile(path, AudioFileOutput)) == NULL) {
	*pIrp = &IRBadAllocation;
	if (fp)
	    CloseBufferedFile(fp);
	UnlinkPath(path, False);
	return(auFile);
    }
    auFile->u.fileType = conf->type;
    auFile->au = conf->au;
    auFile->io->input = False;
    auFile->io->u.fops = *ops;
    auFile->io->fp = fp;

    /*
     * Now call the file functions to really open the file
     * and fill in the auFile->au, and perhaps auFile->io->state.
     */
    if (! (*auFile->io->u.fops.createFile)(auFile) ) {
	*pIrp = auFile->reason;
	if (fp)
	    CloseBufferedFile(fp);
	FreeAudioFile(auFile);
	UnlinkPath(path, False);
	return(NULL);
    }

    return(auFile);
}

AudioFilePtr
OpenAudioDevice(char *path, InternalReasonPtr *pIrp)
{
    AudioFilePtr auFile;
    AudioDevOpFuncPtr ops = &DfltDevOps;

    /*
     * Allocate storage 
     */
    if ((auFile = AllocAudioFile(path, AudioDeviceOutput)) == NULL) {
	*pIrp = &IRBadAllocation;
	return(auFile);
    }
    auFile->u.devFormats = 0;
    memset((char *)&auFile->au, 0, sizeof(auFile->au));
    auFile->io->input = False;
    auFile->io->u.dops = *ops;
    auFile->io->fp = NULL;

    /*
     * Now call the dev functions to really open the dev
     */
    if (! (*auFile->io->u.dops.openDev)(auFile) ) {
	*pIrp = auFile->reason;
	FreeAudioFile(auFile);
	return(NULL);
    }

    return(auFile);
}

/*
 * XXX this should return an InternalReasonPtr
 */
boolean
InitAudioDevice(AudioFilePtr auFile, AudioFormPtr devAu)
{
    auFile->au = *devAu;
    return((*auFile->io->u.dops.initDev)(auFile));
}

/*
 * XXX this should return an InternalReasonPtr
 */
boolean
AudioDeviceMode(AudioFilePtr auFile, long devCommand)
{
    switch(devCommand) {
    case DevModeBlocking:
    case DevModeNonBlocking:
	return((*auFile->io->u.dops.modeDev)(auFile, devCommand));
    default:
	return(False);
    }

}

long
ReadAudioSamples(AudioFilePtr auFile, UChar *buffer, ULong nSamples)
{
    switch(auFile->io->ioClass) {
    case AudioFileInput:
    case AudioCddaInput:
	return( (*auFile->io->u.fops.readSamples)(auFile, buffer, nSamples) );
    case AudioDeviceInput:
	return( (*auFile->io->u.dops.readSamples)(auFile, buffer, nSamples) );
    default:
	SetAudioReason(auFile, &ImproperIO);
	return(-1);
    }
}

long
WriteAudioSamples(AudioFilePtr auFile, UChar *buffer, ULong nSamples)
{
    switch(auFile->io->ioClass) {
    case AudioFileOutput:
	return( (*auFile->io->u.fops.writeSamples)(auFile, buffer, nSamples) );
    case AudioDeviceOutput:
	return( (*auFile->io->u.dops.writeSamples)(auFile, buffer, nSamples) );
    default:
	SetAudioReason(auFile, &ImproperIO);
	return(-1);
    }
}

InternalReasonPtr
SeekAudioOffset(AudioFilePtr auFile, ULong nSamples)
{
    switch(auFile->io->ioClass) {
    case AudioFileInput:
    case AudioCddaInput:
	return( (*auFile->io->u.fops.seekSamples)(auFile, nSamples) );
    case AudioDeviceInput:
	return(&CantSeekAudioDevice);
    default:
	return( &ImproperIO );
    }
}

void
FlushAudio(AudioFilePtr auFile)
{
    switch(auFile->io->ioClass) {
    case AudioDeviceOutput:
    case AudioDeviceInput:
        (*auFile->io->u.dops.flushSamples)(auFile);
        break;
    default:
        break;
    }
}

long
AudioDelayMs(AudioFilePtr auFile)
{
    switch(auFile->io->ioClass) {
    case AudioDeviceOutput:
    case AudioDeviceInput:
        return( (*auFile->io->u.dops.delayMs)(auFile) );
    default:
        return(0);	/* zero delay for files */
    }
}

long
AudioSamplesQueued(AudioFilePtr auFile)
{
    switch(auFile->io->ioClass) {
    case AudioDeviceOutput:
    case AudioDeviceInput:
        return( (*auFile->io->u.dops.samplesQueued)(auFile) );
    default:
        return(0);
    }
}

void
CloseAudioFile(AudioFilePtr auFile)
{
    switch(auFile->io->ioClass) {
    case AudioDeviceOutput:
    case AudioDeviceInput:
	(*auFile->io->u.dops.closeDev)(auFile);
	break;
    case AudioFileOutput:
    case AudioFileInput:
    case AudioCddaInput:
	(*auFile->io->u.fops.closeFile)(auFile);
	if (auFile->io->fp)
	    CloseBufferedFile(auFile->io->fp);
	break;
    }
    FreeAudioFile(auFile);
}

NameEnumPair AudioRateNames[] = {
    { "48khz",     Audio48Khz },
    { "48000",     Audio48Khz },
    { "44.1khz",   Audio44_1Khz },
    { "44100",     Audio44_1Khz },
    { "32khz",     Audio32Khz },
    { "32000",     Audio32Khz },
    { "22.254khz", Audio22_254Khz },
    { "22254",     Audio22_254Khz },
    { "22.05khz",  Audio22_05Khz },
    { "22050",     Audio22_05Khz },
    { "16khz",     Audio16Khz },
    { "16000",     Audio16Khz },
    { "12khz",     Audio12Khz },
    { "12000",     Audio12Khz },
    { "11.025khz", Audio11_025Khz },
    { "11025",     Audio11_025Khz },
    { "8khz",      Audio8Khz },
    { "8000",      Audio8Khz },
    { NULL,        AudioRateUnknown }
};

NameEnumPair AudioBitNames[] = {
    { "24bit",          Audio24BitPcm },
    { "24",             Audio24BitPcm },
    { "16bit",          Audio16BitPcm },
    { "16",             Audio16BitPcm },
    { "16bit unsigned", AudioUSign16BitPcm },
    { "16u",            AudioUSign16BitPcm },
    { "8bit",           Audio8BitPcm },
    { "8",              Audio8BitPcm },
    { "8bit unsigned",  AudioUSign8BitPcm },
    { "8u",             AudioUSign8BitPcm },
    { "ulaw",           Audio8BitULaw },
    { "mulaw",          Audio8BitULaw },
    { NULL,             AudioDataUnknown }
};

NameEnumPair AudioChannelNames[] = {
    { "Mono",     AudioMono },
    { "1",        AudioMono },
    { "Stereo",   AudioStereo },
    { "2",        AudioStereo },
    { NULL,       AudioChannelsUnknown }
};

NameEnumPair AudioFileTypeNames[] = {
    { "au",            FT_SunAudio },
    { "aiff",          FT_AiffAudio },
    { "aifc",          FT_AifcAudio },
    { "mpeg (system)", FT_MpegSystem },
    { "mpeg (audio)",  FT_MpegAudio },
    { "raw",           FT_Unknown },
    { "jmovie",        FT_Jmovie },
    { "quicktime",     FT_QuickTime },
    { NULL,            FT_Unknown },
};

char *
AudioAttrName(AudioFormPtr au, char *buffer)
{
    char *rate, *channel, *bits;

    if ((rate = EnumToName(au->rate, AudioRateNames)) == NULL)
	rate = "?kHz";
    if ((channel = EnumToName(au->channels, AudioChannelNames)) == NULL)
	channel = "?channels";
    if ((bits = EnumToName(au->dataForm, AudioBitNames)) == NULL)
        bits = "?pcm";

    sprintf(buffer, "%s/%s/%s", bits, rate, channel);
    return(buffer);
}

static InternalReasonPtr
CantConvertReason(AudioFormPtr srcAu,
	          AudioFormPtr dstAu)
{
    char from[128], to[128];
    InternalReasonPtr reason;

    reason = CreateReason("Unimplemented conversion: from %s to %s\n",
			  AudioAttrName(srcAu, from),
			  AudioAttrName(dstAu, to));
    reason->hint = IR_BadValue;
    return(reason);
}

InternalReasonPtr
InitAudioConversion(AudioConversionPtr cvt,
		    AudioFormPtr srcAu,
		    AudioFormPtr dstAu)
{
    long conversion = 0;
    static TimerInc zeroCounter;
    TimerInc counter;
    InternalReasonPtr reason = NULL;

    cvt->srcAu = *srcAu;
    cvt->dstAu = *dstAu;
    if (cvt->buffer && cvt->bufferAllocated)
	free((char *)cvt->buffer);
    cvt->buffer = NULL;
    cvt->bufferLength = 0;
    cvt->bufferAllocated = False;

    counter = zeroCounter;

    switch( srcAu->channels ) {
    case AudioMono:   break;
    case AudioStereo: break;
    default:          return( CantConvertReason(srcAu, dstAu) );
    }

    switch( dstAu->channels ) {
    case AudioMono:   break;
    case AudioStereo: break;
    default:          return( CantConvertReason(srcAu, dstAu) );
    }

    if (srcAu->channels > dstAu->channels)
	conversion |= StereoToMono;
    else if (srcAu->channels < dstAu->channels)
	conversion |= MonoToStereo;
    else if (srcAu->channels == AudioStereo)
	conversion |= StereoToStereo;
    else
	conversion |= MonoToMono;

    if (srcAu->rate == dstAu->rate)
	conversion |= SameRate;
    else if (srcAu->rate < dstAu->rate) {
	reason = InitTimerInc(&counter, (double) srcAu->rate, dstAu->rate);
	conversion |= UpRate;
    } else {
	reason = InitTimerInc(&counter, (double) dstAu->rate, srcAu->rate);
	conversion |= DownRate;
    }
    if (reason)
	return(reason);

    switch(srcAu->dataForm) {
    default:
    case Audio24BitPcm:
	return( CantConvertReason(srcAu, dstAu) );
	break;
    case Audio16BitPcm:
	switch(dstAu->dataForm) {
	default: return( CantConvertReason(srcAu, dstAu) );
        case AudioUSign16BitPcm: conversion |= SignedToUnsigned; /* fall thru */
	case Audio16BitPcm: conversion |= Bits16To16; break;
        case AudioUSign8BitPcm:  conversion |= SignedToUnsigned; /* fall thru */
	case Audio8BitPcm:  conversion |= Bits16To8; break;
        case Audio24BitPcm:      return( CantConvertReason(srcAu, dstAu) );
	case Audio8BitULaw: conversion |= Bits16ToULaw; break;
	}
	break;
    case Audio8BitPcm:
	switch(dstAu->dataForm) {
	default: return( CantConvertReason(srcAu, dstAu) );
        case AudioUSign16BitPcm: conversion |= SignedToUnsigned; /* fall thru */
	case Audio16BitPcm: conversion |= Bits8To16; break;
        case AudioUSign8BitPcm:  conversion |= SignedToUnsigned; /* fall thru */
	case Audio8BitPcm:  conversion |= Bits8To8; break;
        case Audio24BitPcm:      return( CantConvertReason(srcAu, dstAu) );
	case Audio8BitULaw: conversion |= Bits8ToULaw; break;
	}
	break;
    case AudioUSign16BitPcm:
        switch(dstAu->dataForm) {
	default: return( CantConvertReason(srcAu, dstAu) );
        case AudioUSign16BitPcm: conversion |= SignedToUnsigned; /* fall thru */
        case Audio16BitPcm:      conversion |= Bits16To16; break;
        case AudioUSign8BitPcm:  conversion |= SignedToUnsigned; /* fall thru */
        case Audio8BitPcm:       conversion |= Bits16To8; break;
        case Audio24BitPcm:      return( CantConvertReason(srcAu, dstAu) );
        case Audio8BitULaw:      conversion |= Bits16ToULaw; break;
        }
        break;
    case AudioUSign8BitPcm:
	switch(dstAu->dataForm) {
	default: return( CantConvertReason(srcAu, dstAu) );
        case Audio16BitPcm:      conversion |= UnsignedToSigned; /* fall thru */
        case AudioUSign16BitPcm: conversion |= Bits8To16; break;
        case Audio8BitPcm:       conversion |= UnsignedToSigned; /* fall thru */
        case AudioUSign8BitPcm:  conversion |= Bits8To8; break;
	case Audio24BitPcm: return( CantConvertReason(srcAu, dstAu) );
        case Audio8BitULaw:      conversion |= Bits8ToULaw | UnsignedToSigned;
                                 break;
        }
        break;
    case Audio8BitULaw:
        switch(dstAu->dataForm) {
	default: return( CantConvertReason(srcAu, dstAu) );
        case AudioUSign16BitPcm: conversion |= SignedToUnsigned; /* fall thru */
	case Audio16BitPcm: conversion |= BitsULawTo16; break;
        case AudioUSign8BitPcm:  conversion |= SignedToUnsigned; /* fall thru */
	case Audio8BitPcm:  conversion |= BitsULawTo8; break;
        case Audio24BitPcm:      return( CantConvertReason(srcAu, dstAu) );
	case Audio8BitULaw: conversion |= BitsULawToULaw; break;
	}
	break;
    }

    switch(conversion) {
    case MonoToMono | SameRate | Bits8To8 | SameSign:
    case MonoToMono | SameRate | Bits16To16 | SameSign:
    case MonoToMono | SameRate | BitsULawToULaw | SameSign:
    case StereoToStereo | SameRate | Bits8To8 | SameSign:
    case StereoToStereo | SameRate | Bits16To16 | SameSign:
    case StereoToStereo | SameRate | BitsULawToULaw | SameSign:
	conversion = NoConversion;
	break;
    }

    cvt->convIndex = conversion;
    cvt->counter = counter;
    cvt->maxInSamples = 0;
    cvt->maxOutSamples = 0;
    cvt->skipNextSrcSamples = 0;
    return(NULL);
}

static AudioConversionPtr
CantConvert(AudioFormPtr srcAu,
	    AudioFormPtr dstAu,
	    InternalReasonPtr *pIrp)
{
    *pIrp = CantConvertReason(srcAu, dstAu);
    return(NULL);
}

AudioConversionPtr
CreateAudioConversion(AudioFormPtr srcAu,
		      AudioFormPtr dstAu,
		      InternalReasonPtr *pIrp)
{
    AudioConversionPtr cvt;

    cvt = (AudioConversionPtr)
	AllocNamedMem(sizeof(AudioConversion), "audio conversion buffer", pIrp);
    if (cvt == NULL)
	return(NULL);

    if (NULL != (*pIrp = InitAudioConversion(cvt, srcAu, dstAu))) {
	free((char *)cvt);
	return(NULL);
    }

    *pIrp = NULL;
    return(cvt);
}

long
ConvertNewAudioBuffer(AudioConversionPtr cvt,
		      UChar *buffer,
		      long nSamples,
		      InternalReasonPtr *pIrp)
{
    cvt->buffer = buffer;
    return( ConvertAudioBuffer(cvt, nSamples, pIrp) );
}

long
ConvertAudioBuffer(AudioConversionPtr cvt,
		   long nSamples,
		   InternalReasonPtr *pIrp)
{
    long conversion = cvt->convIndex;
    TimerIncPtr counter = &cvt->counter;
    AudioDataType from, to, last;
    long sample, srcDataShift, count;

    srcDataShift = DataFormToShift[ cvt->srcAu.dataForm ];

    if (nSamples > cvt->maxInSamples) {
	*pIrp =
	    CreateReason("%s: # of samples (%d) is larger than max (%d)\n",
		"ConvertAudioBuffer", nSamples, cvt->maxInSamples);
	return(-1);
    }

    /*
     * If we are doing uprate conversion, or the change in sample size
     * would cause the destination to overtake the source, we copy the
     * data to the top of the buffer.  This way conversion can still take
     * place left-to-right and the data being created will not
     * overrun the data being converted.
     */
    if (cvt->outIsBiggerThanIn) {
	to.anyType = cvt->buffer;
	from.anyType = cvt->buffer
		     + cvt->bufferLength
		     - (nSamples << srcDataShift);

	/*
	 * memcpy is not guaranteed to do overlapping copies and
	 * bcopy is not portable.  We do a backwards copy here.
	 */
	count = (nSamples << srcDataShift);
	while (--count >= 0)
	    from.anyType[ count ] = to.anyType[ count ];
    } else {
	to.anyType = from.anyType = cvt->buffer;
    }

    last.anyType = from.anyType + (nSamples << srcDataShift);

    /*
     * Since we like to work in signed arithmetic, convert here to
     * signed if appropriate.  Nothing complicated, because we
     * only convert pcm to pcm.  Note that conversion from signed
     * to unsigned is done after the rate conversion.
     */
        switch (cvt->srcAu.dataForm) {
        case AudioUSign16BitPcm:
            count = nSamples;
            while (--count >= 0)
                from.pcm16[ count ] = 0x8000 - (long) from.upcm16[ count ];
        if ((conversion & SignChangeMask) == SameSign)
            conversion |= SignedToUnsigned;
            break;
            
        case AudioUSign8BitPcm:
            count = nSamples;
            while (--count >= 0)
                from.pcm8[ count ] = 0x80 - (long) from.upcm8[ count ];
        if ((conversion & SignChangeMask) == SameSign)
            conversion |= SignedToUnsigned;
            break;
	default:
	    break; /* we don't care */
        }
    
    switch(conversion & (~SignChangeMask)) {
    /*
     * Mono to mono conversions
     */
    case MonoToMono | DownRate | Bits8ToULaw:
	from.pcm8 += cvt->skipNextSrcSamples;
	while (from.pcm8 < last.pcm8) {
	    *to.ulaw++ = PcmByteToMulaw(*from.pcm8);
	    from.pcm8 += TimerIncrementMacro(*counter);
	}
	nSamples = to.ulaw - cvt->buffer;
	cvt->skipNextSrcSamples = from.pcm8 - last.pcm8;
	break;

    case MonoToMono | DownRate | Bits8To16:
	from.pcm8 += cvt->skipNextSrcSamples;
	while (from.pcm8 < last.pcm8) {
	    *to.pcm16++ = *from.pcm8 << 8;
	    from.pcm8 += TimerIncrementMacro(*counter);
	}
	nSamples = (to.anyType - cvt->buffer) >> 1;
	cvt->skipNextSrcSamples = from.pcm8 - last.pcm8;
	break;

    case MonoToMono | DownRate | Bits16ToULaw:
	from.pcm16 += cvt->skipNextSrcSamples;
	while (from.pcm16 < last.pcm16) {
	    *to.ulaw++ = PcmShortToMulaw(*from.pcm16);
	    from.pcm16 += TimerIncrementMacro(*counter);
	}
	nSamples = to.ulaw - cvt->buffer;
	cvt->skipNextSrcSamples = from.pcm16 - last.pcm16;
	break;

    case MonoToMono | DownRate | Bits16To16:
	while (from.pcm16 < last.pcm16) {
	    *to.pcm16++ = *from.pcm16;
	    from.pcm16 += TimerIncrementMacro(*counter);
	}
	nSamples = (to.anyType - cvt->buffer) >> 1;
	break;

    case MonoToMono | SameRate | Bits16ToULaw:
	while (from.pcm16 < last.pcm16)
	    *to.ulaw++ = PcmShortToMulaw(*from.pcm16++);
	nSamples = to.ulaw - cvt->buffer;
	break;

    case MonoToMono | SameRate | Bits8To16:
	while (from.pcm8 < last.pcm8)
	    *to.pcm16++ = *from.pcm8++ << 8;
	nSamples = (to.anyType - cvt->buffer) >> 1;
	break;

    case MonoToMono | SameRate | BitsULawTo16:
	while (from.ulaw < last.ulaw)
	    *to.pcm16++ = MulawToPcmShort(*from.ulaw++);
	nSamples = (to.anyType - cvt->buffer) >> 1;
	break;

    case MonoToMono | UpRate | BitsULawTo16:
	while (from.ulaw < last.ulaw) {
	    sample = MulawToPcmShort(*from.ulaw++);
	    for (count = TimerIncrementMacro(*counter); count > 0; count--)
		*to.pcm16++ = sample;
	}
	nSamples = (to.anyType - cvt->buffer) >> 1;
	break;

    case MonoToMono | UpRate | Bits8To8:
        while (from.pcm8 < last.pcm8) {
	    sample = *from.pcm8;
	    for (count = TimerIncrementMacro(*counter); count > 0; count--)
		*to.pcm8++ = sample;
	    from.pcm8++;
	}
	nSamples = to.anyType - cvt->buffer;
	break;

    case MonoToMono | UpRate | Bits8To16:
        while (from.pcm8 < last.pcm8) {
	    sample = *from.pcm8 << 8;
	    for (count = TimerIncrementMacro(*counter); count > 0; count--)
		*to.pcm16++ = sample;
	    from.pcm8++;
	}
	nSamples = (to.anyType - cvt->buffer) >> 1;
	break;

    case MonoToMono | UpRate | Bits16To16:
	while (from.pcm16 < last.pcm16) {
	    sample = *from.pcm16;
	    for (count = TimerIncrementMacro(*counter); count > 0; count--)
		*to.pcm16++ = sample;
	    from.pcm16++;
	}
	nSamples = (to.anyType - cvt->buffer) >> 1;
	break;

    /*
     * Stereo to mono conversions
     */
    case StereoToMono | DownRate | Bits16To16:
	from.pcm16 += cvt->skipNextSrcSamples;
	while (from.pcm16 < last.pcm16) {
            sample = ((long)from.pcm16[0] + (long)from.pcm16[1]) >> 1;
	    *to.pcm16++ = sample;
	    from.pcm16 += TimerIncrementMacro(*counter) << 1;
	}
	nSamples = (to.anyType - cvt->buffer) >> 1;
	cvt->skipNextSrcSamples = from.pcm16 - last.pcm16;
	break;

    case StereoToMono | DownRate | Bits16ToULaw:
	from.pcm16 += cvt->skipNextSrcSamples;
	while (from.pcm16 < last.pcm16) {
            sample = ((long)from.pcm16[0] + (long)from.pcm16[1]) >> 1;
	    *to.ulaw++ = PcmShortToMulaw(sample);
	    from.pcm16 += TimerIncrementMacro(*counter) << 1;
	}
	nSamples = to.anyType - cvt->buffer;
	cvt->skipNextSrcSamples = from.pcm16 - last.pcm16;
	break;

    case StereoToMono | SameRate | Bits16To16:
	while (from.pcm16 < last.pcm16) {
            sample = ((long)from.pcm16[0] + (long)from.pcm16[1]) >> 1;
	    *to.pcm16++ = sample;
	    from.pcm16 += 2;
	}
	nSamples = (to.anyType - cvt->buffer) >> 1;
	break;

    case StereoToMono | SameRate | Bits8To8:
        while (from.pcm8 < last.pcm8) {
            sample = ((long)from.pcm8[0] + (long)from.pcm8[1]) >> 1;
            *to.pcm8++ = sample;
            from.pcm8 += 2;
        }
        nSamples = to.anyType - cvt->buffer;
        break;

    case StereoToMono | UpRate | Bits16To16:
	while (from.pcm16 < last.pcm16) {
            sample = ((long)from.pcm16[0] + (long) from.pcm16[1]) >> 1;
	    for (count = TimerIncrementMacro(*counter); count > 0; count--)
		*to.pcm16++ = sample;
	    from.pcm16 += 2;
	}
	nSamples = (to.anyType - cvt->buffer) >> 1;
	break;

    /*
     * Stereo to Stereo conversions
     */
    case StereoToStereo | DownRate | Bits16To16:
	from.pcm16 += cvt->skipNextSrcSamples;
	while (from.pcm16 < last.pcm16) {
	    *to.pcm16++ = from.pcm16[0];
	    *to.pcm16++ = from.pcm16[1];
	    from.pcm16 += TimerIncrementMacro(*counter) << 1;
	}
	nSamples = (to.anyType - cvt->buffer) >> 1;
	cvt->skipNextSrcSamples = from.pcm16 - last.pcm16;
	break;

    case StereoToStereo | UpRate | Bits16To16:
	while (from.pcm16 < last.pcm16) {
	    for (count = TimerIncrementMacro(*counter); count > 0; count--) {
		*to.pcm16++ = from.pcm16[0];
		*to.pcm16++ = from.pcm16[1];
	    }
	    from.pcm16 += 2;
	}
	nSamples = (to.anyType - cvt->buffer) >> 1;
	break;

    case CantDoConversion:
    default:
	(void) CantConvert(&cvt->srcAu, &cvt->dstAu, pIrp);
	nSamples = 0;
	break;
        
    case MonoToMono | SameRate | Bits8To8:
    case MonoToMono | SameRate | Bits16To16:
    case MonoToMono | SameRate | BitsULawToULaw:
    case StereoToStereo | SameRate | Bits8To8:
    case StereoToStereo | SameRate | Bits16To16:
    case StereoToStereo | SameRate | BitsULawToULaw:
        /*
         * Very special... nothing changes but the sign.  So just
         * do nothing!
         */
        break;
    }

    /*
     * Conversion from signed to unsigned
     */
    if ((conversion & SignChangeMask) == SignedToUnsigned) {
        to.anyType = cvt->buffer;
        switch (cvt->dstAu.dataForm) {
        case AudioUSign16BitPcm:
            count = nSamples;
            while (--count >= 0)
                to.upcm16[ count ] = 0x8000 + to.pcm16[ count ];
            break;
            
        case AudioUSign8BitPcm:
            count = nSamples;
            while (--count >= 0)
                to.upcm8[ count ] = 0x80 + to.pcm8[ count ];
            break;
	default:
	    break; /* we don't care */
        }
    }

    return(nSamples);
}

ULong
AudioRateToAvailBit(AudioSampleRate rate)
{
    switch (rate) {
    case Audio48Khz:     return Avail48Khz;
    case Audio44_1Khz:   return Avail44_1Khz;
    case Audio32Khz:     return Avail32Khz;
    case Audio22_254Khz: return Avail22_254Khz;
    case Audio22_05Khz:  return Avail22_05Khz;
    case Audio16Khz:     return Avail16Khz;
    case Audio11_025Khz: return Avail11_025Khz;
    case Audio8Khz:      return Avail8Khz;
    default: return 0;
    }
}

/*
 * pick a rate from among those available and closest to the source rate.
 * In general, "closest" should be below the source rate (to avoid aliasing),
 * but not too far below (to avoid quality loss).
 */
static AudioSampleRate
ChooseClosestRate(AudioFormPtr srcAu, long formats)
{
    int i;
    AudioSampleRate closestRate, rate;
    ULong distance, newDistance;
    static struct {
	AvailableAudioFormats availableRate;
	AudioSampleRate rate;
    } atorMap[] = {	/* "available rate" bit to "rate" mapping */
	{ Avail48Khz,		Audio48Khz },
	{ Avail44_1Khz,		Audio44_1Khz },
	{ Avail32Khz,		Audio32Khz },
	{ Avail22_254Khz,	Audio22_254Khz },
	{ Avail22_05Khz,	Audio22_05Khz },
	{ Avail16Khz,		Audio16Khz },
	{ Avail11_025Khz,	Audio11_025Khz },
	{ Avail8Khz,		Audio8Khz },
	{ Avail8BitULaw,	Audio8Khz }
    };

    if (formats & AudioRateToAvailBit(srcAu->rate))
        return( srcAu->rate );

    distance = ~0;
    closestRate = 0;
    for (i = 0; i < NumberOfCells(atorMap); i++) {
	if (formats & atorMap[ i ].availableRate) {
	    if (! closestRate) {
		closestRate = atorMap[ i ].rate;
		distance = abs(srcAu->rate - closestRate);
	    } else {
		rate = atorMap[ i ].rate;
		newDistance = abs(srcAu->rate - rate);
		/*
		 * We take the newly found rate if it is closer.
		 */
		if (newDistance < distance) {
		    distance = newDistance;
		    closestRate = rate;
		}
	    }
	}
    }
    return (closestRate);
}
    
InternalReasonPtr
ChooseAudioForm(AudioFormPtr srcAu,
		long formats,
		AudioFormPtr dstAu)
{
    boolean convertible = True;

    dstAu->rate = dstAu->channels = dstAu->dataForm = 0;

    switch (srcAu->channels) {
    case AudioMono:
	if (formats & AvailMono)
	    dstAu->channels = AudioMono;
	else
	    dstAu->channels = AudioStereo;
	break;

    case AudioStereo:
	if (formats & AvailStereo)
	    dstAu->channels = AudioStereo;
	else
	    dstAu->channels = AudioMono;
	break;

    default:
	convertible = False;
	break;
    }

    dstAu->rate = ChooseClosestRate(srcAu, formats);
    if (dstAu->rate == 0)
	convertible=False;

    switch(srcAu->dataForm) {
    case Audio24BitPcm:
	if (formats & Avail24BitPcm) {
	    dstAu->dataForm = Audio24BitPcm;
	    break;
	}
	/* fall through */
    case Audio16BitPcm:
	if (formats & Avail16BitPcm) {
	    dstAu->dataForm = Audio16BitPcm;
	    break;
	}
	/* fall through */
    case AudioUSign16BitPcm:
        if (formats & AvailUSign16BitPcm) {
            dstAu->dataForm = Audio16BitPcm;
            break;
        }
        /* fall through */
    case Audio8BitPcm:
	if (formats & Avail8BitPcm) {
	    dstAu->dataForm = Audio8BitPcm;
	    break;
	}
	/* fall through */
    case AudioUSign8BitPcm:
        if (formats & AvailUSign8BitPcm) {
            dstAu->dataForm = AudioUSign8BitPcm;
            break;
        }
        /* fall through */
    case Audio8BitULaw:
	if (formats & Avail8BitULaw) {
	    dstAu->dataForm = Audio8BitULaw;
	    break;
	}
	/* fall through */
    default:
	break;
    }

    if (dstAu->dataForm == 0) {
	if (formats & Audio16BitPcm)
            dstAu->dataForm = Audio16BitPcm;
        else if (formats & AvailUSign16BitPcm)
            dstAu->dataForm = AudioUSign16BitPcm;
        else if (formats & Avail8BitULaw)
            dstAu->dataForm = Audio8BitULaw; /* better than 8-bit pcm */
        else if (formats & Avail8BitPcm)
            dstAu->dataForm = Audio8BitPcm;
        else if (formats & AvailUSign8BitPcm)
            dstAu->dataForm = AudioUSign8BitPcm;
        else
	    convertible = False;
    }

    if (! convertible)
	return( CantConvertReason(srcAu, dstAu) );

    return( NULL );
}

/*
 * Allocate an audio buffer big enough to hold 'usec' microseconds in
 * either the source or destination form.  We allocate the larger of the two
 */
InternalReasonPtr
AllocMaxAudioBuffer(AudioConversionPtr cvt, long usec)
{
    long nDstBytes, nSrcBytes, nBytes;
    long srcDataShift, dstDataShift;
    double scale = (double)usec / MicroSecPerSec;

    srcDataShift = DataFormToShift[ cvt->srcAu.dataForm ];
    nSrcBytes = (long)(cvt->srcAu.rate * scale)
	      * (cvt->srcAu.channels << srcDataShift);

    dstDataShift = DataFormToShift[ cvt->dstAu.dataForm ];
    nDstBytes = (long)(cvt->dstAu.rate * scale)
	      * (cvt->dstAu.channels << dstDataShift);

    if (nSrcBytes >= nDstBytes) {
	nBytes = nSrcBytes;
    } else {
	nBytes = nDstBytes;
    }

    /*
     * Its up to the file reading routines to free this, so
     * we don't even check.
     */
    cvt->buffer = (UChar *)InsecureAlloc(nBytes);
    if (cvt->buffer == NULL)
	return( AllocReason("sound buffer", nBytes) );
    cvt->bufferAllocated = True;
    cvt->bufferLength = nBytes;
    cvt->maxInSamples = nSrcBytes >> srcDataShift;
    cvt->maxOutSamples = nDstBytes >> dstDataShift;
    cvt->outIsBiggerThanIn = (nDstBytes > nSrcBytes);

    return(NULL);
}

long
CalcAudioConversionSizes(AudioConversionPtr cvt, long nSrcSamples)
{
    long nDstBytes, nSrcBytes, nBytes;
    long srcDataShift, dstDataShift;
    double srcBytesPerSec, dstBytesPerSec;

    srcDataShift = DataFormToShift[ cvt->srcAu.dataForm ];
    nSrcBytes = nSrcSamples << srcDataShift;
    srcBytesPerSec = cvt->srcAu.rate * (cvt->srcAu.channels << srcDataShift);

    dstDataShift = DataFormToShift[ cvt->dstAu.dataForm ];
    dstBytesPerSec = cvt->dstAu.rate * (cvt->dstAu.channels << dstDataShift);

    nDstBytes = nSrcBytes * (dstBytesPerSec / srcBytesPerSec) + 0.5;

    if (nSrcBytes >= nDstBytes) {
	nBytes = nSrcBytes;
    } else {
	nBytes = nDstBytes;
    }

    cvt->bufferLength = nBytes;
    cvt->maxInSamples = nSrcBytes >> srcDataShift;
    cvt->maxOutSamples = nDstBytes >> dstDataShift;
    cvt->outIsBiggerThanIn = (nDstBytes > nSrcBytes);

    return(nBytes);
}

boolean
AllocAudioState(AudioFilePtr afp, long nBytes)
{
    PrivPointer state;

    state = InsecureAlloc(nBytes);
    if (state == NULL) {
        SetAudioReason(afp, &IRBadAllocation);
        return(False);
    }
    afp->io->state = state;
    memset((char *)state, 0, nBytes);
    return(True);
}
