/*
 * Copyright (c) 1992-1995 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	sgiCdda.c - sgi support for CDDA input (Compact Disk Digital Audio)
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/sgiCdda.c,v 1.3 1995/08/25 19:15:19 toddb Exp $
 */

#include "nvrAudioI.h"
#include "nvrAlloc.h"
#include "fileAccess.h"
#include "nvrDebug.h"

#include <sys/types.h>
#include <dmedia/cdaudio.h>
#include <errno.h>
#include <unistd.h>

static InternalReason CantSeek =
    { "Can't seek on CDDA yet\n", IR_NotImplemented };
static InternalReason CantWrite =
    { "Can't write CDDA\n", IR_BadIOFile };

enum {
    CdFramesPerSec = 75
};

typedef struct _cddaState {
    CDPLAYER *cd;
    CDPARSER *cdparser;
    long nFrames;		/* length of input in frames (1/75 sec) */
    long framesIn;		/* current number of frames read so far */
    CDFRAME *inbuf;		/* buffer used to read frames */
    long nBufFrames;		/* number of frames in inbuf */
    UChar *samplesBuffer;	/* buffer passed to ReadCddaSamples */
} CddaState, *CddaStatePtr;

static CDPLAYER *SgiCd;	/* XXX blech!!! a process can only open it once! */

static void
CloseCddaAudio(AudioFilePtr auFile)
{
    CddaStatePtr cddaState = (CddaStatePtr) auFile->io->state;

    if (cddaState->cdparser) {
	CDdeleteparser(cddaState->cdparser);
	cddaState->cdparser = NULL;
    }
    if (cddaState->cd) {
	/* CDclose(cddaState->cd); */
	cddaState->cd = NULL;
    }
}

static boolean
CreateCddaAudio(AudioFilePtr auFile)
{
    SetAudioReason(auFile, &CantWrite);
    return(False);
}

static void
CddaAudioCb(void *arg, CDDATATYPES type, short *audio)
{
    AudioFilePtr auFile = (AudioFilePtr)arg;
    CddaStatePtr cddaState = (CddaStatePtr) auFile->io->state;

    memcpy((char *)cddaState->samplesBuffer,
	   (char *)audio,
	   CDDA_DATASIZE);
    cddaState->samplesBuffer += CDDA_DATASIZE;
    cddaState->framesIn++;
}

static boolean
OpenCddaAudio(AudioFilePtr auFile, AudioQualityPtr quality)
{
    CddaStatePtr cddaState;
    CDTRACKINFO trkInfo;
    CDSTATUS cdstatus;
    CDPLAYER *cd;
    CddaDescPtr cdda = auFile->io->cdda;
    int nFrames, min, sec, frame, allocSize;
    long trackStart, trackLength, specifiedLength, specifiedOffset;
    long seekReturn, startFrame;
    InternalReasonPtr irp;
    DebugDeclare( long endFrame )

    if (SgiCd)	/* XXX Blech! A process can only open and close it once */
	cd = SgiCd;
    else
	SgiCd = cd = CDopen(0, "r");

    if (cd == NULL) {
	if (errno == EACCES) {
	    if (geteuid() == 0)
		irp = CreateSysReason("The CD-ROM drive is already in use.\n");
	    else
		irp = CreateSysReason("%s\n",
			    "You must be root to open the CD-ROM drive");
	} else {
	    irp = CreateSysReason("Couldn't open CD-ROM drive\n");
	}
	SetAudioReason(auFile, irp);
        return(False);
    }

    nFrames = CDbestreadsize(cd);
    allocSize = sizeof(*cddaState) + nFrames * sizeof(CDFRAME);

    if (! AllocAudioState(auFile, allocSize) )
        return(False);
    cddaState = (CddaStatePtr) auFile->io->state;

    cddaState->cd = cd;
    cddaState->nBufFrames = nFrames;
    cddaState->inbuf = (CDFRAME *)(cddaState + 1);

    /*
     * initialize our idea of the state of the CD-ROM
     */
    CDgetstatus(cddaState->cd, &cdstatus);

    /*
     * Make sure everything is kosher with the CD-ROM. Must
     * have an audio disc in before we proceed
     */
    switch (cdstatus.state) {
    case CD_NODISC:
        SetAudioReason(auFile,
	    BadFileReason("No disc in CD-ROM drive\n"));
	break;

    case CD_CDROM:
        SetAudioReason(auFile,
	    BadFileReason("Disc in CD-ROM drive is not an audio CD\n"));
	break;

    case CD_ERROR:
        SetAudioReason(auFile,
	    BadFileReason("Error reading CD-ROM drive\n"));
	break;

    case CD_READY:
	break;

    case CD_PLAYING:
    case CD_PAUSED:
    case CD_STILL:
    default:
	CDstop(cddaState->cd);
	break;
    }

    if (auFile->reason) {
	CloseCddaAudio(auFile);
        return(False);
    }

    if (cdda->track == 0) {
	cdda->track = cdstatus.first;
    } else if (cdda->track < cdstatus.first || cdda->track > cdstatus.last) {
	SetAudioReason(auFile,
            BadValueReason("bad specified track: %d; must be %d-%d\n",
		cdda->track, cdstatus.first, cdstatus.last));
	CloseCddaAudio(auFile);
        return(False);
    }

    if (CDgettrackinfo(cddaState->cd, cdda->track, &trkInfo) < 0) {
	CloseCddaAudio(auFile);
        return(False);
    }

    WhenDebugEnabled(
	Msg("cdda: track %d, start @%d:%d:%d for %d:%d:%d\n",
	    cdda->track,
	    trkInfo.start_min, trkInfo.start_sec, trkInfo.start_frame,
	    trkInfo.total_min, trkInfo.total_sec, trkInfo.total_frame);
    )

    /*
     * set the bounds of the track
     */
    trackLength = CdFramesPerSec*60*trkInfo.total_min
	        + CdFramesPerSec*trkInfo.total_sec
	        + trkInfo.total_frame;
    trackStart  = CdFramesPerSec*60*trkInfo.start_min
                + CdFramesPerSec*trkInfo.start_sec
                + trkInfo.start_frame;
    specifiedLength = CdFramesPerSec*60*cdda->length.min
		    + CdFramesPerSec*cdda->length.sec
		    + cdda->length.frame;
    if (specifiedLength == 0)
	cddaState->nFrames = trackLength;
    else
	cddaState->nFrames = specifiedLength;

    specifiedOffset  = CdFramesPerSec*60*cdda->start.min
		     + CdFramesPerSec*cdda->start.sec
		     + cdda->start.frame;
    if (specifiedOffset == 0) {
        min = trkInfo.start_min;
        sec = trkInfo.start_sec;
        frame = trkInfo.start_frame;
	startFrame = trackStart;
	WhenDebugEnabled( endFrame = trackStart + cddaState->nFrames; )
    } else {
	if (specifiedLength == 0) {
	    cddaState->nFrames = trackLength - specifiedOffset;
	    if (cddaState->nFrames <= 0) {
		SetAudioReason(auFile,
		    BadValueReason("%s %d: %d:%d:%d\n",
			"specified offset is longer than track",
			cdda->track, cdda->start.min,
			cdda->start.sec, cdda->start.frame));
		CloseCddaAudio(auFile);
		return(False);
	    }
	}
	specifiedOffset += trackStart;
	startFrame = specifiedOffset;
	WhenDebugEnabled( endFrame = specifiedOffset + cddaState->nFrames; )
	min = cdda->start.min = specifiedOffset / (CdFramesPerSec*60);
	sec = cdda->start.sec = (specifiedOffset % (CdFramesPerSec*60))
			      / CdFramesPerSec;
	frame = cdda->start.frame = specifiedOffset % CdFramesPerSec;
    }

    seekReturn = CDseek(cddaState->cd, min, sec, frame);
    if (seekReturn != startFrame - 150) {
        SetAudioReason(auFile,
	    CreateSysReason("CD-DA seek failed; wanted frame %d, got %d\n",
		startFrame, seekReturn));
	CloseCddaAudio(auFile);
        return(False);
    }

    WhenDebugEnabled(
	Msg("      read @%d:%d:%d to %d:%d:%d for %d:%d:%d%s\n",
	    min, sec, frame,
	    endFrame / (CdFramesPerSec*60),
	    (endFrame % (CdFramesPerSec*60)) / CdFramesPerSec,
	    endFrame % CdFramesPerSec,
	    cddaState->nFrames / (75*60),
	    (cddaState->nFrames % (75*60)) / 75,
	    cddaState->nFrames % 75,
	    specifiedLength ? " (specified)" : "");
    )

    cddaState->cdparser = CDcreateparser();
    if (cddaState->cdparser == NULL) {
        SetAudioReason(auFile,
	    AllocReason("SGI CDcreateparser()", sizeof(cddaState->cdparser)));
	CloseCddaAudio(auFile);
        return(False);
    }

    CDaddcallback(cddaState->cdparser, cd_audio,
		  (CDCALLBACKFUNC) CddaAudioCb, (void *)auFile);
    auFile->au. rate = Audio44_1Khz;
    auFile->au. channels = AudioStereo;
    auFile->au. dataForm = Audio16BitPcm;
    auFile->nSamples = cddaState->nFrames * CDDA_NUMSAMPLES;
    return(True);
}

static long
ReadCddaSamples(AudioFilePtr auFile, UChar *buffer, ULong nSamples)
{
    CddaStatePtr cddaState = (CddaStatePtr) auFile->io->state;
    long nFrames, framesIn, lastFrameWanted, entryFramesIn, limit;
    int i;
    CDFRAME *pFrame;
    long totalIn;

    entryFramesIn = cddaState->framesIn;
    limit = cddaState->nFrames - cddaState->framesIn;
    nFrames = (nSamples / CDDA_NUMSAMPLES);
    if (nFrames > limit)
	nFrames = limit;
    lastFrameWanted = cddaState->framesIn + nFrames;
    cddaState->samplesBuffer = buffer;

    while (cddaState->framesIn < lastFrameWanted) {
	nFrames = lastFrameWanted - cddaState->framesIn;
	if (nFrames > cddaState->nBufFrames)
	    nFrames = cddaState->nBufFrames;
	framesIn = CDreadda(cddaState->cd, cddaState->inbuf, nFrames);
        if (framesIn < 0) {
	    SetAudioReason(auFile,
		CreateSysReason("Error while reading CD-ROM drive\n"));
	    break;
        }

	/*
	 * samples are copied into buffer as a side effect of
	 * the cd_audio callback invoked from CDparseframe.
	 * Also cddaState->frameIn is updated.
	 */
	pFrame =  &cddaState->inbuf[0];
        for (i = 0; i < framesIn; i++, pFrame++) {

	    WhenDebugEnabled(
		static long lastFrame = -1000;
		static long lastMode;
		CddaTime cur;
		long curFrame;

		switch (pFrame->subcode.type) {
		case CDQ_MODE3:
		case CDQ_MODE2:
		    if (pFrame->subcode.type != lastMode) {
			lastMode = pFrame->subcode.type;
			Msg("    mode %d", pFrame->subcode.type);
		    }
		    break;

		case CDQ_MODE1:
		    cur.min = pFrame->subcode.data.mode1.atime.mhi * 10
			    + pFrame->subcode.data.mode1.atime.mlo;
		    cur.sec = pFrame->subcode.data.mode1.atime.shi * 10
			    + pFrame->subcode.data.mode1.atime.slo;
		    cur.frame = pFrame->subcode.data.mode1.atime.fhi * 10
			      + pFrame->subcode.data.mode1.atime.flo;
		    curFrame = cur.min * 60 * CdFramesPerSec
			     + cur.sec * CdFramesPerSec
			     + cur.frame;
		    if (curFrame >= lastFrame + CdFramesPerSec) {
			Msg("  %2d:%2d:%2d", cur.min, cur.sec, cur.frame);
			lastFrame = curFrame;
		    }
		    break;
		}
	    )

            CDparseframe(cddaState->cdparser, pFrame);
        }
    }

    totalIn = (cddaState->framesIn - entryFramesIn) * CDDA_NUMSAMPLES;
    if (totalIn == 0)
	CDstop(cddaState->cd);
    return(totalIn);
}

static long
WriteCddaSamples(AudioFilePtr auFile, UChar *buffer, ULong nSamples)
{
    SetAudioReason(auFile, &CantWrite);
    return(-1);
}

static InternalReasonPtr
SeekCddaSamples(AudioFilePtr auFile, ULong nSamples)
{
    return(&CantSeek);
}

AudioOpFunc AudioLibCddaFuncs = {
    OpenCddaAudio,
    CreateCddaAudio,
    ReadCddaSamples,
    WriteCddaSamples,
    SeekCddaSamples,
    CloseCddaAudio
};

