/*
 * Copyright (c) 1992, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	sunaudio.c
 *
 * DESCRIPTION
 *	Programming interface to SparcStation's audio device.
 *	OS-DEPENDENT
 *
 *	$Header: /usr/local/src/nvrToMIT/RCS/suncontrol.c,v 1.15 1994/12/17 13:01:36 russ Exp $
 *
 */

#ifdef sun
/*
 *      EXTERNAL INCLUDES
 *              Include files that must be exported to any package or
 *              program using this package.
 */
#define NeedFunctionPrototypes 1

#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <stropts.h>

#ifdef Solaris2
# include <sys/audioio.h>
# include <sys/file.h>
#else
# include <sun/audioio.h>
#endif

#include <audio_device.h>	/* in /usr/demo/SOUNDS/multimedia */
#include <libaudio.h>		/* in /usr/demo/SOUNDS/multimedia */

/*
 *      INTERNAL INCLUDES
 *              Include files for local use only, therefore, NOT exported
 *		to any package or program using this package.
 */
#include "sunaudio.h"
#include "nvrTypes.h"
#include "nvrString.h"
#include "nvrAlloc.h"
#include "osio.h"

/*
 *      EXTERNS
 *              External declarations required locally to this package
 *              that are not already declared in any of the included header
 *		files (external includes or internal includes).
 */


/*
 *      GLOBALS
 *              Variables declared in this package that are allowed
 *		to be used globally.
 */


/*
 *      LOCAL DEFINES
 *		#define declarations local to this package.
 */
#define AUDIO_IODEV     "/dev/audio"
#define AUDIO_CTLDEV    "/dev/audioctl"
#define MIN(x,y) ((x) > (y) ? (y) : (x))
#define GAIN_TO_FLOAT(value)						\
	((double)(value) / (double)AUDIO_MAX_GAIN)

/*
 *      LOCAL TYPEDEFS
 *              typedefs local to this package (for use with local vars).
 *
 */

typedef struct _AudioCallbackRec {
    unsigned		which_changes;
    AudioCallbackProc	callback_proc;
    void *		client_data;
    struct _AudioCallbackRec *next;
} AudioCallbackRec, *AudioCallbackList;


/*
 *      LOCALS VARIABLES
 *		Variables local to this package.
 */

static int		AudioFd = -1;
static int		AudioCtlFd = -1;
static int		AudioFdMode = 0;
static audio_info_t	AudioState;
static Audio_hdr	AudioPlayConfig;
static Audio_hdr	AudioRecordConfig;
static AudioCallbackList AudioCallbacks = (AudioCallbackList)NULL;



/************************************************************************
 *									*
 *			PRIVATE ROUTINES				*
 *									*
 *	Routines private to this package therefore declared static	*
 *									*
 ************************************************************************/

/*
 *	NAME
 *		HandleControlChanges
 *
 *	SYNOPSIS
 */
static void
HandleControlChanges(int fd, PrivPointer data)
/*
 *	DESCRIPTION
 *		Add a callback for the indicated audio changes.
 *
 *	RETURNS
 *		
 */
{
    AudioCallbackRec		*p_rec;
    static AudioCallbackStruct	cbs;
    unsigned			what_changed = 0;
    Audio_info			current_audio_state;

    BlockAsyncNotification(AudioCtlFd);

    if (!AudioGetState(&current_audio_state)) {
	/* can't get current audio state ... bail out */
	AllowAsyncNotification(AudioCtlFd);
	return;
    }

    if (current_audio_state.play.gain != AudioState.play.gain) {
	what_changed |= AUDIO_CHANGED_PLAY_GAIN;
	cbs.play_gain = GAIN_TO_FLOAT(current_audio_state.play.gain);
    }
    if (current_audio_state.play.pause != AudioState.play.pause) {
	what_changed |= AUDIO_CHANGED_PLAY_PAUSE;
	cbs.play_pause = current_audio_state.play.pause;
    }
    if (current_audio_state.play.port != AudioState.play.port) {
	what_changed |= AUDIO_CHANGED_PLAY_PORT;
	cbs.play_port = current_audio_state.play.port;
    }
    if (current_audio_state.record.gain != AudioState.record.gain) {
	what_changed |= AUDIO_CHANGED_RECORD_GAIN;
	cbs.record_gain = GAIN_TO_FLOAT(current_audio_state.record.gain);
    }
    if (current_audio_state.monitor_gain != AudioState.monitor_gain) {
	what_changed |= AUDIO_CHANGED_MONITOR_GAIN;
	cbs.monitor_gain = GAIN_TO_FLOAT(current_audio_state.monitor_gain);
    }

    /* update OUR LOCAL audio state info */
    memcpy((char *)&AudioState, (char *)&current_audio_state,
	    sizeof(Audio_info));

    for (p_rec = AudioCallbacks; p_rec != (AudioCallbackRec *)NULL;
	    p_rec = p_rec->next) {
	if (p_rec->which_changes & what_changed) {
	    cbs.changes = p_rec->which_changes & what_changed;
	    (*p_rec->callback_proc)(
	    p_rec->client_data, &cbs);
	}
    }

    AllowAsyncNotification(AudioCtlFd);
}


/************************************************************************
 *									*
 *			PUBLIC ROUTINES					*
 *									*
 *		Publically accessible routines				*
 *									*
 ************************************************************************/

/*
 *
 *	Interfaces for Audio CONTROL Device Interfaces
 *
 */



/*
 *	NAME
 *		AudioGetState
 *
 *	SYNOPSIS
 */
int
AudioGetState(ap)
    audio_info_t	*ap;
/*
 *	DESCRIPTION
 *		Obtains the current control parameters of the audio device.
 *
 *	RETURNS
 *		0 on error; otherwise non-zero.
 *
 */
{
    if (!AudioControlOpen(NULL)) {
	return(0);
    }

    audio_getinfo(AudioCtlFd, ap);

#ifdef OLD
    ap->play.gain = UNSCALE_GAIN(ap->play.gain);
    ap->record.gain = UNSCALE_GAIN(ap->record.gain);
    ap->monitor_gain = UNSCALE_GAIN(ap->monitor_gain);
#endif /* OLD */

    return(1);
}


/*
 *	NAME
 *		AudioSetMonitorVolume
 *
 *	SYNOPSIS
 */
void
AudioSetMonitorVolume(value)
    double value;
/*
 *	DESCRIPTION
 *		Sets the audio device's monitor volume.
 *		Valid values are any floating point values between 0.0 and 1.0.
 *
 *	RETURNS
 *		The new value used by the audio device.
 *
 *	NOTES
 *		Updating actual value is done via
 *		SIGIO ... see AudioAddControlChangeCallback().
 */
{
    double tmp = value;

    if (!AudioControlOpen(NULL)) {
	return;
    }

    audio_set_monitor_gain(AudioCtlFd, &tmp);

}


/*
 *	NAME
 *		AudioSetMonitorState
 *
 *	SYNOPSIS
 */
int
AudioSetMonitorState(control, ratio)
    int control;
    double ratio;
/*
 *	DESCRIPTION
 *		Sets the audio device's monitor volume
 *
 *	RETURNS
 *		1 on success, otherwise 0;
 *
 *	NOTES
 *		Updating actual set values are done via
 *		SIGIO ... see AudioAddControlChangeCallback().
 */
{
    static unsigned 	stored_monitor_gain;
    static double	tmp;

    if (!AudioControlOpen(NULL)) {
	return(0);
    }

    switch (control) {
      case PAUSE:
	/*
	 * really can't pause the monitor ... let's just fall through
	 * to mute.
	 */
      case MUTE:
	tmp = 0.0;
	stored_monitor_gain = AudioState.monitor_gain;
	audio_set_monitor_gain(AudioCtlFd, &tmp);
	break;
      case LOWER:
	stored_monitor_gain = AudioState.monitor_gain;
	tmp = GAIN_TO_FLOAT(stored_monitor_gain) * ratio;
	audio_set_monitor_gain(AudioCtlFd, &tmp);
	break;
      case RESUME:
	tmp = GAIN_TO_FLOAT(stored_monitor_gain);
	audio_set_monitor_gain(AudioCtlFd, &tmp);
	break;
      default:
	return(0);
    }
    return(1);
}


/*
 *	NAME
 *		AudioSetPlayVolume
 *
 *	SYNOPSIS
 */
void
AudioSetPlayVolume(value)
    double value;
/*
 *	DESCRIPTION
 *		Sets the audio device's play volume
 *		Valid values are any floating point values between 0.0 and 1.0.
 *
 *	RETURNS
 *		The new value used by the audio device.
 *
 *	NOTES
 *		Updating actual value is done via
 *		SIGIO ... see AudioAddControlChangeCallback().
 */
{
    double tmp = value;

    if (!AudioControlOpen(NULL)) {
	return;
    }

    audio_set_play_gain(AudioCtlFd, &tmp);
}


/*
 *	NAME
 *		AudioSetPlayState
 *
 *	SYNOPSIS
 */
int
AudioSetPlayState(control, ratio)
    int control;
    double ratio;
/*
 *	DESCRIPTION
 *		Sets the audio device's play state.  Ratio ignored unless
 *		control is LOWER.
 *
 *	RETURNS
 *		1 on success, otherwise 0;
 *
 *	NOTES
 *		Updating actual set values are done via
 *		SIGIO ... see AudioAddControlChangeCallback().
 */
{
    static unsigned 	stored_play_gain;
    static boolean	gain_stored = False;
    double		tmp;

    if (!AudioControlOpen(NULL)) {
	return(0);
    }

    switch (control) {
      case MUTE:
	stored_play_gain = AudioState.play.gain;
	gain_stored = True;
	tmp = 0.0;
	audio_set_play_gain(AudioCtlFd, &tmp);
	break;
      case LOWER:
	stored_play_gain = AudioState.play.gain;
	gain_stored = True;
	tmp = GAIN_TO_FLOAT(stored_play_gain) * ratio;
	audio_set_play_gain(AudioCtlFd, &tmp);
	break;
      case RESUME:
	if (gain_stored) {
	    tmp = GAIN_TO_FLOAT(stored_play_gain);
	    audio_set_play_gain(AudioCtlFd, &tmp);
	    gain_stored = False;
	}
	audio_resume_play(AudioCtlFd);
	break;
      case PAUSE:
	audio_pause_play(AudioCtlFd);
	break;
      case FLUSH:
	audio_flush_record(AudioCtlFd);
	break;
      default:
	return(0);
    }
    return(1);
}


/*
 *	NAME
 *		AudioSetRecordState
 *
 *	SYNOPSIS
 */
int
AudioSetRecordState(control, ratio)
    int control;
    double ratio;
/*
 *	DESCRIPTION
 *		Sets the audio device's play state.  Ratio ignored unless
 *		control is LOWER.
 *
 *	RETURNS
 *		1 on success, otherwise 0;
 *
 *	NOTES
 *		Updating actual set values are done via
 *		SIGIO ... see AudioAddControlChangeCallback().
 */
{
    static unsigned 	stored_record_gain;
    static boolean	gain_stored = False;
    double		tmp;

    if (!AudioControlOpen(NULL)) {
	return(0);
    }

    switch (control) {
      case MUTE:
	stored_record_gain = AudioState.record.gain;
	gain_stored = True;
	tmp = 0.0;
	audio_set_record_gain(AudioCtlFd, &tmp);
	break;
      case LOWER:
	stored_record_gain = AudioState.record.gain;
	gain_stored = True;
	tmp = GAIN_TO_FLOAT(stored_record_gain) * ratio;
	audio_set_record_gain(AudioCtlFd, &tmp);
	break;
      case RESUME:
	if (gain_stored) {
	    tmp = GAIN_TO_FLOAT(stored_record_gain);
	    audio_set_record_gain(AudioCtlFd, &tmp);
	    gain_stored = False;
	}
	audio_resume_record(AudioCtlFd);
	break;
      case PAUSE:
	audio_pause_record(AudioCtlFd);
	break;
      case FLUSH:
	audio_flush_record(AudioCtlFd);
	break;
      default:
	return(0);
    }
    return(1);
}


/*
 *	NAME
 *		AudioSetRecordVolume
 *
 *	SYNOPSIS
 */
void
AudioSetRecordVolume(value)
    double value;
/*
 *	DESCRIPTION
 *		Sets the audio device's record volume.
 *		Valid values are any floating point values between 0.0 and 1.0.
 *
 *	RETURNS
 *		The new value used by the audio device.
 *
 *	NOTES
 *		Updating actual value is done via
 *		SIGIO ... see AudioAddControlChangeCallback().
 */
{
    double tmp = value;

    if (!AudioControlOpen(NULL)) {
	return;
    }

    audio_set_record_gain(AudioCtlFd, &tmp);

}


/*
 *	NAME
 *		AudioControlOpen
 *
 *	SYNOPSIS
 */
int
AudioControlOpen(cbs)
    AudioCallbackStruct	*cbs;
/*
 *	DESCRIPTION
 *		Opens the audio control device for read and write permissions.
 *
 *	RETURNS
 *		-1 on error; otherwise the file descriptor
 *
 */
{
    if (AudioCtlFd > -1) {
	if (cbs != NULL)  {
	    AudioGetState(&AudioState);
	    cbs->play_gain = GAIN_TO_FLOAT(AudioState.play.gain);
	    cbs->play_pause = AudioState.play.pause;
	    cbs->play_port = AudioState.play.port;
	    cbs->record_gain = GAIN_TO_FLOAT(AudioState.record.gain);
	    cbs->monitor_gain = GAIN_TO_FLOAT(AudioState.monitor_gain);
	    cbs->changes = 0x1f;
	}
	return(AudioCtlFd);
    }

    if ((AudioCtlFd = open(AUDIO_CTLDEV, O_RDWR)) < 0) {
	return(-1);
    }

    if ((audio_get_play_config(AudioCtlFd, &AudioPlayConfig)
		!= AUDIO_SUCCESS) ||
	    (audio_get_record_config(AudioCtlFd, &AudioRecordConfig)
		!= AUDIO_SUCCESS)) {
	(void) close(AudioCtlFd);
	return(AudioCtlFd = -1);
    }

    AudioGetState(&AudioState);

    if (cbs != NULL)  {
	cbs->play_gain = GAIN_TO_FLOAT(AudioState.play.gain);
	cbs->play_pause = AudioState.play.pause;
	cbs->play_port = AudioState.play.port;
	cbs->record_gain = GAIN_TO_FLOAT(AudioState.record.gain);
	cbs->monitor_gain = GAIN_TO_FLOAT(AudioState.monitor_gain);
	cbs->changes = 0x1f;
    }

    AllocAsyncHandler(AudioCtlFd, HandleControlChanges, True, NULL);
    ioctl(AudioCtlFd, I_SETSIG, S_INPUT|S_OUTPUT|S_MSG);

    return(AudioCtlFd);

}



/*
 *	NAME
 *		AudioSetMaxGain
 *
 *	SYNOPSIS
 */
int
AudioSetSpeakerOutput()
/*
 *	DESCRIPTION
 *		Attempts to set output to the speaker
 *
 *	RETURNS
 *		1 on success; otherwise 0;
 */
{
    unsigned port = AUDIO_SPEAKER;
    
    if (!AudioControlOpen(NULL)) {
	return(0);
    }
    return (audio_set_play_port(AudioCtlFd, &port) == AUDIO_SPEAKER);
}


/*
 *	NAME
 *		AudioSetMaxGain
 *
 *	SYNOPSIS
 */
int
AudioSetHeadphoneOutput()
/*
 *	DESCRIPTION
 *		Close the audio i/o device.
 *
 *	RETURNS
 *		Attempts to set output to the headphone jack
 */
{
    unsigned port = AUDIO_HEADPHONE;
    
    if (!AudioControlOpen(NULL)) {
	return(0);
    }
    return (audio_set_play_port(AudioCtlFd, &port) == AUDIO_HEADPHONE);
}


/*
 *	NAME
 *		AudioAddControlChangeCallback
 *
 *	SYNOPSIS
 */
int
AudioAddControlChangeCallback(which_changes, callback_proc, client_data)
    unsigned which_changes;
    AudioCallbackProc callback_proc;
    void *client_data;
/*
 *	DESCRIPTION
 *		Add a callback for the indicated audio changes.
 *
 *	RETURNS
 *		
 */
{
    AudioCallbackRec *p_new;

    if (AudioCtlFd != -1) {
	BlockAsyncNotification(AudioCtlFd);
    }

    if (!(which_changes & 0xf)) {
	return(0);
    }

    p_new = (AudioCallbackRec *)malloc(sizeof(AudioCallbackRec));
    p_new->which_changes = which_changes & 0x1f;
    p_new->callback_proc = callback_proc;
    p_new->client_data = client_data;

    if (AudioCallbacks == NULL) {
	p_new->next = NULL;
    } else {
	p_new->next = AudioCallbacks;
    }
    AudioCallbacks = p_new;

    if (AudioCtlFd != -1) {
	AllowAsyncNotification(AudioCtlFd);
    }

    return(0);
}

/*
 *
 *	Interfaces for Audio I/O Device Interfaces
 *
 */


/*
 *	NAME
 *		AudioOpen
 *
 *	SYNOPSIS
 */
int
AudioOpen(mode, p_fd)
    int mode; /* PLAY or RECORD */
    int *p_fd;
/*
 *	DESCRIPTION
 *		Opens the audio i/o device for read or write permissions.
 *
 *	RETURNS
 *		AUDIO_SUCCESS	on successful open
 *		AUDIO_BUSY	if audio device busy
 *		AUDIO_ERROR	if error
 */
{
    int flag;

    *p_fd = -1;

    if (AudioFd > -1 ) {
	if (AudioFdMode == mode) {
	    *p_fd = AudioFd;
	    return(AUDIO_SUCCESS);
	} else {
	    return(AUDIO_ERROR);
	}
    }

    if ((*p_fd = open(AUDIO_IODEV,
	    (mode == RECORD ? O_RDONLY : O_WRONLY) | O_NDELAY)) < 0) {
#ifdef OLD
	if ((errno == EINTR) || (errno == EBUSY)) {
#else
	if (errno == EBUSY) {
#endif /* OLD */
	    /* device busy */
	    return(AUDIO_BUSY);
	} else {
	    /* error */
	    return(AUDIO_ERROR);
	}
    }

    if (mode == RECORD) {
	flag = fcntl(*p_fd, F_GETFL, 0) | FNDELAY;
	if ((fcntl(*p_fd, F_SETFL, flag) < 0) ||
	    (ioctl(*p_fd, I_SETSIG, S_INPUT|S_OUTPUT|S_MSG) < 0)) {
	    close(*p_fd);
	    *p_fd = -1;
	    return(AUDIO_ERROR);
	}
    }

    AudioFd = *p_fd;
    AudioFdMode = mode;
    return(AUDIO_SUCCESS);

}


/*
 *	NAME
 *		AudioFlushAndClose
 *
 *	SYNOPSIS
 */
void
AudioFlushAndClose(p_fd)
    int *p_fd;
/*
 *	DESCRIPTION
 *		Flush then close the audio i/o device.
 *
 *	RETURNS
 *		void
 */
{
    if (AudioFd < 0 ) {
	return;
    }

    switch (AudioFdMode) {
      case PLAY:
	audio_flush_play(AudioFd);
	break;
      case RECORD:
	audio_flush_record(AudioFd);
	break;
    }

    (void) close(AudioFd);
    *p_fd = AudioFd = -1;
    AudioFdMode = 0;
}


/*
 *	NAME
 *		AudioClose
 *
 *	SYNOPSIS
 */
void
AudioClose(p_fd)
    int *p_fd;
/*
 *	DESCRIPTION
 *		Close the audio i/o device.
 *
 *	RETURNS
 *		void
 */
{
    if (AudioFd < 0 ) {
	return;
    }

    (void) close(AudioFd);
    *p_fd = AudioFd = -1;
    AudioFdMode = 0;
}
#endif	/* sun */
