/*
 *	Copyright (c) 1992, North Valley Research, Inc.
 *	Confidential -- All proprietary rights reserved.
 *	Licensed Material -- Property of North Valley Research, Inc.
 *
 *	This program is the property of North Valley Research, Inc. or others
 *	from whom North Valley Research has obtained a licensing right, and is
 *	considered by North Valley Research to be confidential.  It is protected
 *	by U.S. copyright law as an unpublished work and is
 *	furnished pursuant to a written license agreement.  It may
 *	not be used, copied or otherwise reproduced, or disclosed to
 *	others except in accordance with the terms and conditions of
 *	that agreement.
 *
 *	NAME
 *		bitstream.c
 *
 *	DESCRIPTION
 *		Bitstream I/O from an elementary stream using the MPEG 1
 *		system decoder.
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/bitstream.c,v 1.31 1994/12/17 13:01:48 russ Exp $
 */
#define NeedFunctionPrototypes 1

#include "osio.h"
#include "nvrAlloc.h"
#include "nvrString.h"
#include "bitstream.h"
#include "mpegSysDecode.h"

/*
 * FillBitStream() is called from the macros in "bitstream.h" to get more data.
 * The request bit field extends beyond the current buffer.  The strategy
 * used here is to copy the bytes containing the partial field at the the of
 * the previous buffer to just before the start of the next buffer such that
 * the data will appear to be contiguous.  For the "raw" bitstream case,
 * there is no problem since there buffer is long word aligned and the next
 * read begins and buffer[1] always.  However in the case of the system
 * bitstream, the data can begin on an arbirary byte boundry and may be an
 * arbitrary number of bytes long.  The remaining discussion concerns the
 * system stream only.
 *
 * Another problem is ULong alignment:  The data returned from the system
 * layer may not be ULong aligned; and all of the bitstream macros require
 * ULong alignment.  Since copying is a bad idea, the strategy is to decrease
 * the size of the returned block of bytes by the size of the fractional ULong
 * and the hi-address end of the buffer.  This data is also copied into the
 * memory immediately before the next buffer.
 *
 * In the worst case, 7 bytes need to be copied from the end of the previous
 * buffer to the beginning to the new buffer.  Luckily, the minimum system
 * packet header (data prior to the first payload byte) is 7 bytes long
 * (startcode, length, next field code).
 *
 * Although FillBitStream() is called when an aligned ULong boundry is crossed,
 * when the data is returned, this crossing bitfield may now be totally
 * contained within an aligned ULong.  Since this does not easily "fit" into
 * the macros in "bitstream.h", FillBitSteam() will extract and explicitly
 * return the previously ULong crossing bitfield in ->retValue.  The return
 * Cache arguments (_nCacheBits & _pCacheBuf) must be updated in the case of
 * either a ReadBits() or a NextBits().  In the former case these return values
 * reflect the position following the crossing bitfield and nCacheBits is
 * negative.  In the latter case the return values denote the position before
 * the crossing bitfield and nCacheBits is >= 0.
 */

ULong *
FillBitStream(BitStream *buf, int nCacheBits, int nBits)
{
    ULong *pCacheBuf;
    int n, m, tmpbytes;
    ULong value;
    char *p, tmp[7];

    if (buf->system) {
	n = (nCacheBits < 0) ? (nBits + nCacheBits) : nCacheBits;
	m = (n + msk2(L_BYTE)) >> L_BYTE;
	tmpbytes = buf->leftover + m;

	if (!((0 < nBits) && (nBits <= 32)))
	    FailedAssert(n, buf->pbufLast, buf,
		"\"(0 < nBits) && (nBits <= 32)\"", __FILE__, __LINE__);
	if (!((0 <= n) && (n <= 32)))
	    FailedAssert(n, buf->pbufLast, buf,
		"\"(0 < n) && (n <= 32)\"", __FILE__, __LINE__);
	if (!(tmpbytes <= sizeof(tmp)))
	    FailedAssert(n, buf->pbufLast, buf,
		"\"tmpbytes < sizeof(tmp)\"", __FILE__, __LINE__);

	memcpy(tmp, (char *)buf->pbufLast + pwr2(L_LONG - L_BYTE) - m,
	    tmpbytes);
	n += buf->leftover << L_BYTE;
	if (buf->eof)
	    goto smalldata;
	buf->pos = buf->mstr.pos + buf->mstr.length;
	buf->mstr.buf = (UChar *)&buf->tmp[2];
	StreamRead(&buf->mstr);
	if (buf->mstr.length > 0) {
	    if ((buf->mstr.length + tmpbytes) <= sizeof(ULong)) {
		memcpy(&tmp[tmpbytes], buf->mstr.buf, buf->mstr.length);
		tmpbytes += buf->mstr.length;
		n += buf->mstr.length << L_BYTE;
		buf->mstr.buf -= buf->mstr.length;
		goto smalldata;
	    }
	    memcpy(buf->mstr.buf - tmpbytes, tmp, tmpbytes);
	    pCacheBuf = (ULong *)(((long)(buf->mstr.buf - tmpbytes)) &
		~msk2(L_LONG - L_BYTE));
	    n += ((-(long)buf->mstr.buf) & msk2(L_LONG - L_BYTE)) << L_BYTE;
	    buf->leftover = ((long)(buf->mstr.buf + buf->mstr.length)) &
		msk2(L_LONG - L_BYTE);
	    buf->pbufLast = (ULong *)(buf->mstr.buf + buf->mstr.length -
		buf->leftover - sizeof(ULong));
	} else {
	    buf->eof = True;
	    buf->mstr.pos = buf->pos;
smalldata:
	    memcpy(((char *)&buf->tmp[2]) - tmpbytes, tmp, tmpbytes);
	    buf->pbufLast = &buf->tmp[1];
	    buf->leftover = 0;
	    pCacheBuf = (ULong *)(((long)&buf->tmp[2] -
		tmpbytes) & ~msk2(L_LONG - L_BYTE));
	    if (n < nBits) {
		if (nCacheBits < 0) {
		    if (buf->eofok) {
			buf->eofok = False;
			buf->nCacheBits = 0;
			return(pCacheBuf);
		    }
		    (*buf->errorFunc)(buf, BitstreamEof);
		    Fatal("Unexpected EOF on %s\n", buf->file);
		}
		/*
 		 * Lookahead beyond EOF is OK, returns 0 bits after EOF.
		 * Make sure consume beyond EOF will re-call FillBitStream().
		 */
	    }
	}
	if ((n &= msk2(L_LONG)) == 0)
	    n = pwr2(L_LONG);
	if (nCacheBits < 0) {
	    if ((n -= nBits) >= 0) {		/* Consuming bits: ReadBits() */
		buf->nCacheBits = n;
		value = pCacheBuf[0] >> n;
	    } else {
		value = (pCacheBuf[0] << (-n)) |
			(pCacheBuf[1] >> (buf->nCacheBits = n + pwr2(L_LONG)));
		pCacheBuf++;
	    }
	} else {
	    buf->nCacheBits = n;		/* Lookahead only: NextBits() */
	    if ((n -= nBits) >= 0) {
		value = pCacheBuf[0] >> n;
	    } else {
		value = (pCacheBuf[0] << (-n)) |
			(pCacheBuf[1] >> (n + pwr2(L_LONG)));
	    }
	}
    } else {
	n = buf->bufSize;
	if (!buf->eof) {
	    buf->buffer[0] = *buf->pbufLast;     /* last unconsumed word */
	    for (p = (char *)&buf->buffer[1]; n > 0; n -= m) {
		if ((m = (*buf->readFunc)(buf->fd, p, n)) < 0)
		    Fatal("Read error @ %d\n", buf->pos);
		if (m == 0) {
		    buf->eof = True;
		    break;
		}
		p += m;
		buf->pos += m;
	    }
	    /* find the longword *after* the last byte */
	    if ((long)p & msk2(L_LONG - L_BYTE)) {
		n = pwr2(L_LONG - L_BYTE) - ((long)p & msk2(L_LONG - L_BYTE));
		memset(p, 0, n);
		p += n;
	    }
	    buf->pbufLast = ULongP(p) - 1;	/* the last longword */
	}
	if (buf->eof && (buf->bufSize - n) == 0) {
	    if (buf->eofok) {
		buf->eofok = False;
		buf->nCacheBits = 0;
		return(buf->pbufLast);
	    }
	    if (nCacheBits < 0) {
		(*buf->errorFunc)(buf, BitstreamEof);
		Fatal("Unexpected EOF on %s\n", buf->file);
	    }
	    /*
	     * Lookahead beyond EOF is OK, return 0 bits beyond EOF, however
	     * must re-call FillBitStream() if consuming bit beyond EOF.
	     */
	    buf->buffer[1] = 0;
	}
	if (nCacheBits < 0) {
	    n = nCacheBits;			/* Consuming bits: ReadBits() */
	    buf->nCacheBits = n + pwr2(L_LONG);
	    pCacheBuf = &buf->buffer[1];
	} else {
	    buf->nCacheBits = nCacheBits;	/* Lookahead only: NextBits() */
	    n = nCacheBits - nBits;
	    pCacheBuf = &buf->buffer[0];
	}
	if( n == -32 ) {
	    value = (buf->buffer[1] >> (pwr2(L_LONG) - 32));
	} else {
	    value = (buf->buffer[0] << (-n)) |
	            (buf->buffer[1] >> (n + pwr2(L_LONG)));
	}
    }

    if (nBits < 32)
	value &= msk2(nBits);
    buf->retValue = value;

    return(pCacheBuf);
}

static void
DefaultErrorFunction(BitStream *buf, BitstreamCondition condition)
{
    switch (condition) {
    case BitstreamError:
	Fatal("Exiting...\n");
	/* NOTERACHED */

    case BitstreamEof:
	Fatal("Unexpected EOF @ %d\n", buf->pos);
	/* NOTERACHED */
    }
}

static void
ResetBitStreamVaraibles(BitStreamPtr mpegBuf)
{
    mpegBuf->system = False;	/* will be changed by open routine */
    mpegBuf->eof = False;
    mpegBuf->eofok = False;
    mpegBuf->pos = 0;
    mpegBuf->tmp[2] = 0;
    mpegBuf->leftover = 0;
    mpegBuf->nCacheBits = 0;
    mpegBuf->pbuf = mpegBuf->pbufLast = NULL;
    mpegBuf->opened = False;
}

void
ResetBitstreamFuncs(BitStreamPtr mpegBuf)
{
#if 0
    extern int read(int, char *, int);
#endif /* 0 */

    mpegBuf->errorFunc = DefaultErrorFunction;
    mpegBuf->openFunc = OpenFileForReading;
    mpegBuf->readFunc = (BitStreamReadFunc) ReadData;
    mpegBuf->closeFunc = CloseFile;
    mpegBuf->seekFunc = (BitStreamSeekFunc) SeekFile;
}

void
InitBitStreamData(BitStreamPtr mpegBuf)
{
    ResetBitstreamFuncs(mpegBuf);
    mpegBuf->fd = -1;
    mpegBuf->file = NULL;
    mpegBuf->msys = NULL;
    mpegBuf->bufSize = DFLT_BITSTREAM_BUF_SIZE << (L_LONG - L_BYTE);
    mpegBuf->seqPriv = NULL;

    ResetBitStreamVaraibles(mpegBuf);
}

/*
 * OpenBitStreamInput() first check to see if there is a system stream open.
 * If so, it is used; otherwise the first 4 bytes of the named file are
 * examined to see if the file is a system stream (ie. begins with a pack
 * start code).  If so, the system layer is opened.  Otherwise, the file
 * is assumed to be a raw (video, audio or data) bitstream.
 */
InternalReasonPtr
OpenBitStreamInput(BitStreamPtr buf, char *file)
{
    ResetBitStreamVaraibles(buf);
    buf->file = file;
    buf->system = False;
    if ((buf->fd = (*buf->openFunc)(file)) < 0)
        return( CreateSysReason("can't open %s\n", file) );
    if (buf->buffer == NULL) {
	/*
	 * We allocate the requested size plus one longword.  Normally,
	 * sizeInLongs longwords are read into ->buffer.  The extra
	 * word is for when a NextBits request doesn't find enough
	 * bits in the last word of the buffer.  In this case ->pbuf[0]
	 * gets copied to ->buffer[0], and the next read goes into
	 * ->buffer[1];
	 */
	buf->buffer = (ULong *)InsecureAlloc(buf->bufSize + sizeof(ULong));
	if (buf->buffer == NULL)
	    return( &IRBadAllocation );
	memset((char *)buf->buffer, 0, buf->bufSize);   /* keep purify happy */
    }
    buf->pbuf = buf->pbufLast = buf->buffer;
    buf->opened = True;
    return(NULL);
}

InternalReasonPtr
OpenSystemStreamInput(BitStreamPtr mpegBuf,
		      BitStreamPtr firstBuf,
		      char *file,
		      BitstreamType type)
{
    InternalReasonPtr irp;

    ResetBitStreamVaraibles(mpegBuf);

    if ((firstBuf == NULL) ||
        (firstBuf->msys == NULL)) {
	mpegBuf->msys = (MpegSystemInfoPtr)
	    InsecureAlloc(sizeof(MpegSystemInfo));
	if (mpegBuf->msys == NULL)
	    return( &IRBadAllocation );

	memset((char *)mpegBuf->msys, 0, sizeof(MpegSystemInfo));

	mpegBuf->msys->openFunc = mpegBuf->openFunc;
	mpegBuf->msys->readFunc = mpegBuf->readFunc;
	mpegBuf->msys->closeFunc = mpegBuf->closeFunc;
	mpegBuf->msys->seekFunc = mpegBuf->seekFunc;

	if ( (irp = SystemOpen(file, mpegBuf->msys)) )
	    return(irp);

	if (firstBuf != NULL)
	    firstBuf->msys = mpegBuf->msys;
    } else
	mpegBuf->msys = firstBuf->msys;

    if ( (irp = StreamOpen(mpegBuf->msys, type, -1, &mpegBuf->mstr)) )
	return(irp);

    mpegBuf->opened = True;
    mpegBuf->file = file;
    mpegBuf->system = True;

    return(NULL);
}

int
SeekBitStream(BitStreamPtr buf, long offset, SeekWhere sk)
{
    long newOffset;

    if (buf->system)
	newOffset = SystemSeek(buf->msys, offset, sk);
    else
	newOffset = (*buf->seekFunc)(buf->fd, offset, sk);
    return( newOffset );
}

void
CloseBitStream(BitStreamPtr buf)
{
    if (buf->system)
	SystemClose1(buf->msys, &buf->mstr, False);
    else
	(*buf->closeFunc)(buf->fd);
    buf->opened = False;
}

ReadBitsReturn _FReadBits(int _nCacheBits,
			  ULong *_pCacheBuf,
			  BitStream *buf,
			  int nBits,
			  int whatIsIt)
{
    ULong value;

    ReadBits(buf, nBits, value, whatIsIt);
    ReturnFromFReadBits (buf, value);
}

NextBitsReturn _FNextBits(int _nCacheBits,
			  ULong *_pCacheBuf,
			  BitStream *buf,
			  int nBits)
{
    ULong value;

    NextBits(buf, nBits, value);
    ReturnFromFNextBits (buf, value);
}

ConsumeBitsReturn _FConsumeBits(int _nCacheBits,
				ULong *_pCacheBuf,
				BitStream *buf,
				int nConsume,
				int whatIsIt)
{
    ConsumeBits(buf, nConsume, whatIsIt);
    ReturnFromFConsumeBits(buf);
}

/*
 * Look for the first byte-aligned 8-bit zero.  On return the _pCacheBuf
 * and _nCacheBits point just after the zero byte.
 */
ConsumeBitsReturn _FScanForAlignedZeroByte(int _nCacheBits,
					   ULong *_pCacheBuf,
					   BitStream *buf,
					   int whatIsIt)
{
    int nBits;
    ULong bits;
    UChar byte;
    ULong byteOfBits;
    ULong *pEnd;

    /*
     * First, discard the initial non-byte-aligned bits.
     */
    if ( (nBits = _nCacheBits & 0x7) ) {
	ConsumeBits(buf, nBits, whatIsIt);
    }

    for (;;) {
	/*
	 * Then grab the next non-32-bit-aligned bytes and check them
	 * for zeros.
	 */
	if ( (nBits = _nCacheBits & (0x3<<3)) ) {
	    NextBits(buf, nBits, bits);
	    switch (nBits) {
	    case 24:
		ShowBits(buf, bits, 8, whatIsIt);
		_nCacheBits -= 8;
		if (bits >> 16 == 0)
		    goto foundZeroByte;
		bits &= 0xff<<16;
		/* fall through */

	    case 16:
		ShowBits(buf, bits, 8, whatIsIt);
		_nCacheBits -= 8;
		if (bits >> 8 == 0)
		    goto foundZeroByte;
		bits &= 0xff<<8;
		/* fall through */

	    case 8:
		ShowBits(buf, bits, 8, whatIsIt);
		_nCacheBits -= 8;
		if (bits == 0)
		    goto foundZeroByte;
		/* fall through */

	    case 0:
		break;
	    }
	}

	/*
	 * _nCacheBits is now zero, and the stream is 32-bit aligned.
	 * Now just read longwords out of the input buffer.
	 */
	_pCacheBuf++;
	_nCacheBits = 32;
	pEnd = buf->pbufLast + 1;
	while (_pCacheBuf != pEnd) {
	    bits = *_pCacheBuf;
	    byteOfBits = bits & (bits >> 16);
	    byteOfBits &= byteOfBits >> 8;

	    if (byteOfBits) {
		/*
		 * One of the bytes is zero.
		 */
		byte = bits >> 24;
		ShowBits(buf, byte, 8, whatIsIt);
		_nCacheBits -= 8;
		if (byte == 0)
		    goto foundZeroByte;

		byte = (bits >> 16) & 0xff;
		ShowBits(buf, byte, 8, whatIsIt);
		_nCacheBits -= 8;
		if (byte == 0)
		    goto foundZeroByte;

		byte = (bits >> 8) & 0xff;
		ShowBits(buf, byte, 8, whatIsIt);
		_nCacheBits -= 8;
		if (byte == 0)
		    goto foundZeroByte;

		byte = bits & 0xff;
		ShowBits(buf, byte, 8, whatIsIt);
		_nCacheBits -= 8;
		if (byte == 0)
		    goto foundZeroByte;
		_nCacheBits = 32;
	    } else {
		ShowBits(buf, bits, 32, whatIsIt);
	    }
	    _pCacheBuf++;
	}
	_pCacheBuf = FillBitStream(buf, 0, 8);
	_nCacheBits = buf->nCacheBits;
    }

foundZeroByte:
    ReturnFromFScanForAlignedZeroByte(buf);
}

void
FailedAssert(int _nCacheBits,
		  ULong *_pCacheBuf,
		  BitStream *buf,
		  char *mustBeTrue,
		  char *fileName,
		  int line)
{
    BitByteOffset offset;

    buf->nCacheBits = _nCacheBits;
    buf->pbuf = _pCacheBuf;
    offset = BitStreamOffset(buf, 0);
    Error("For bitstream %s\n", buf->file);
    Error("    Assert failed @ byte/bit offset %d/%d:\n",
	offset.bytes, offset.bits);
    Error("    %s, line %d: %s\n", fileName, line, mustBeTrue);

    (*buf->errorFunc)(buf, BitstreamError);
    
    Fatal("Internal error\n");
}

long
BitDifference(BitByteOffsetPtr start, BitByteOffsetPtr end)
{
    long nBytes = end->bytes - start->bytes;
    int nBits;

    if (end->bits >= start->bits) {
	nBits = end->bits - start->bits;
    } else {
	nBits = start->bits - end->bits;
	nBytes--;
    }
    return ((nBytes<<3) + nBits);
}

BitByteOffset
BitStreamOffset(BitStream *pBuf, int newBits)
{
    long wholeBytes;
    int nBits = newBits;
    BitByteOffset offset;

    nBits += pBuf->nCacheBits;
    offset.bits  = (- nBits) & msk2(L_BYTE);
    if (pBuf->system) {
	if ((wholeBytes = ((UChar *)pBuf->pbuf - pBuf->mstr.buf) +
	    ((pwr2(L_LONG) - nBits) >> L_BYTE)) < 0)
	    offset.bytes = pBuf->pos + wholeBytes;
	else
	    offset.bytes = pBuf->mstr.pos + wholeBytes;
    } else
	offset.bytes = pBuf->pos -
	    ((pBuf->pbufLast - pBuf->pbuf) << (L_LONG - L_BYTE)) -
	    ((nBits + msk2(L_BYTE)) >> L_BYTE);
    return(offset);
}
