/*
 * (c) Copyright 1992 North Valley Research, Inc.
 *         All Rights Reserved
 *
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	fileAccess.c
 *
 * DESCRIPTION
 *	OS-specific access to files (permissions, etc).
 *	OS-DEPENDENT
 *
 * REVISION
 *	$Header: /usr/local/src/nvrToMIT/RCS/fileAccess.c,v 1.28 1995/07/06 22:22:17 toddb Exp $
 */

#define NeedFunctionPrototypes 1

#include "fileAccessI.h"
#include "nvrReason.h"

typedef int (*StatFunc)(const char *, struct stat *);

char *
TypeName(int type)
{
    switch (type) {
    case FILE_DIRECTORY:
	return "directory";
    case FILE_CHAR:
	return "character special file";
    case FILE_BLOCK:    
	return "block special file";
    case FILE_REGULAR:  
	return "regular file";
    case FILE_SYMLINK:  
	return "symbolic link pointing nowhere";
    case FILE_FIFO:   
	return "fifo";
    case FILE_SOCKET:   
	return "socket";
    case FILE_SYMLINK | FILE_DIRECTORY:
	return "symbolic link pointing to a directory";
    case FILE_SYMLINK | FILE_CHAR:
	return "symbolic link pointing to a character special file";
    case FILE_SYMLINK | FILE_BLOCK:
	return "symbolic link pointing to a block special file";
    case FILE_SYMLINK | FILE_REGULAR:
	return "symbolic link pointing to a regular file";
    case FILE_SYMLINK | FILE_SOCKET:
	return "symbolic link pointing to a socket";
    case 0:
	return "non-existent file";
    default:
	return "unknown file type";
    }
}

static struct stat *
StatFile(StatFunc func, char *path)
{
    static struct stat lstatBuf, statBuf;
    struct stat *st;

    if (func == (StatFunc)lstat) {
	st = &lstatBuf;
    } else {
	st = &statBuf;
    }

    if ((*func)(path, st) < 0) {
	if (errno == ENOENT)
	    errno = 0;
	return (NULL);
    }
    return (st);
}

ULong
FsBlockSize(int fd)
{
    struct stat    st;
    int    result;

    result = fstat(fd, &st);
    if (result < 0) {
        errno = 0;
        return 0;
    }
    return(st.st_size);
}

ULong
FileSize(char *path)
{
    struct stat    *st;

    st = StatFile((StatFunc)stat, path);
    if (st)
	return(st->st_size);
    return (0);
}

ULong
FdFileSize(int fd)
{
    struct stat    st;
    int    result;

    result = fstat(fd, &st);
    if (result < 0) {
        errno = 0;
        return 0;
    }
    return(st.st_size);
}

ULong
FdModTime(int fd)
{
    struct stat    st;

    if (fstat(fd, &st) < 0) {
        errno = 0;
        return 0;
    }
    return(st.st_mtime);
}

ULong
ModTime(char *path)
{
    struct stat    *st;

    st = StatFile((StatFunc)lstat, path);
    if (st)
	return(st->st_mtime);
    return (0);
}

static long
StatModeToFileType(char *path, mode_t mode)
{
    switch (mode & S_IFMT) {
    case S_IFDIR: return (FILE_DIRECTORY);
    case S_IFCHR: return (FILE_CHAR);
    case S_IFBLK: return (FILE_BLOCK);
    case S_IFREG: return (FILE_REGULAR);
    case S_IFIFO: return (FILE_FIFO);
    case S_IFLNK: return (FILE_SYMLINK);
    case S_IFREG|S_IFSOCK: return (FILE_SOCKET);
    }

    if ((int)path >= 0 && (int)path < 1000)
	ProgMsg("Warning: file descriptor %d: mode 0x%x-->type 0x%x\n",
	    mode, (int)path);
    else
	ProgMsg("Warning: file \"%s\" mode 0x%x-->type 0x%x\n",
	    mode, path);
    return(0);
}

boolean
FdValid(int fd)
{
    struct stat st;

    return( fstat(fd, &st) == 0);
}

long
FdFileType(int fd)
{
    struct stat    st;
    long    type;
    int result;

    result = fstat(fd, &st);
    if (result < 0) {
        errno = 0;
        return 0;
    }

    type = StatModeToFileType((char *)fd, st.st_mode);
    return (type);
}

long
FileType(char *path)
{
    struct stat    *st;
    int    type;

    if ((st = StatFile((StatFunc)lstat, path)) == NULL)
	return(0);

    type = StatModeToFileType(path, st->st_mode);
    if (type & FILE_SYMLINK) {
	if ((st = StatFile((StatFunc)stat, path)) == NULL)
	    return(type);
	type |= StatModeToFileType(path, st->st_mode);
    }
    return (type);
}

ULong
FilePermission(char *path)
{
    struct stat    *st;

    if ((st = StatFile((StatFunc)stat, path)) == NULL)
        Fatal("can't get perm for %s\n", path);
    return(st->st_mode & (~S_IFMT));
}

/*
 * New version
 */
InternalReasonPtr
UnlinkFile(char *path)
{
    if (unlink(path) < 0)
	return( CreateSysReason("Cannot unlink(%s)\n", path) );
    return (NULL);
}

/*
 * Old version
 */
boolean
UnlinkPath(char *path, boolean complain)
{
    if (unlink(path) < 0 && complain) {
	Error("Cannot unlink(%s)\n", path);
	return (False);
    }
    errno = 0;
    return (True);
}
