/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	internalReason.c - descriptive reason support
 *
 * DESCRIPTION
 *	Routines to create and manage descriptive reasons.
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/internalReason.c,v 1.16 1995/11/18 01:30:37 toddb Exp $
 */

#define NeedFunctionPrototypes 1

#include "nvrTypes.h"
#include "nvrReason.h"
#include "nvrAlloc.h"
#include "nvrString.h"
#include "osio.h"

InternalReason IRPipelineEof =
    { "eof\n", IR_Eof };
InternalReason IRCanceled =
    { "canceled\n", IR_Canceled };
InternalReason IRPrematureEof =
    { "premature eof on input\n", IR_PrematureEof };
InternalReason IRBadFileHeader =
    { "bad file header\n", IR_BadIOFile };
InternalReason IRNoSuchFile =
    { "No such file or directory\n", IR_BadIOFile };
InternalReason IRShortRead =
    { "Short read on Input file\n", IR_ShortRead };
InternalReason IRBadIoFileType =
    { "Unrecognized file type\n", IR_UnknowFileType };
InternalReason IRBadTranslation =
    { "image requires translation\n", IR_InternalProblem };
InternalReason IRCantWrite =
    { "can't write output image\n", IR_CantWrite };
InternalReason IRInternalProblem =
    { "oops\n", IR_InternalProblem };
InternalReason IRBadAllocation =
    { "can't allocate memory\n", IR_BadAlloc };
InternalReason IRBadDisplay =
    { "bad display specification\n", IR_BadDisplay };
InternalReason IRBadWindow =
    { "bad window specification\n", IR_BadWindow };
InternalReason IRBadValue =
    { "bad value\n", IR_BadValue };
InternalReason IRBadVisual =
    { "bad visual specification\n", IR_BadVisual };
InternalReason IRBadHardware =
    { "the necessary hardware seem to be broken or missing\n", IR_BadHardware };
InternalReason IRBadInclip =
    { "inclip is out-of-bounds or has a zero dimension\n", IR_BadDimensions };
InternalReason IRBadOutclip =
    { "outclip is out-of-bounds or has a zero dimension\n", IR_BadDimensions };
InternalReason IRBadSample =
    { "bad sample rate\n", IR_BadSampleRate };
InternalReason IRBadColor =
    { "bad color space\n", IR_BadColor };
InternalReason IRBadDimensions =
    { "bad width or height\n", IR_BadDimensions };

/*
 * Hate to include these.... but have to have it for CreateReason()
 */
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>

InternalReasonPtr
CreateSysReason(char *fmt, ...)
{
    char msg[ BUFSIZ ];
    char *pmsg = msg;
    va_list ap;
    InternalReasonPtr reason;
    extern int      errno;
    InternalReasonHint hint = IR_Unknown;

    va_start(ap, fmt);

    if (errno) {
	switch (errno) {
	case ENOENT: hint = IR_BadIOFile; break;
	case ENOSPC: hint = IR_CantWrite; break;
	default: break;
	}

        sprintf(msg, "%s: ", LookupSysError(errno));
	pmsg += strlen(msg);
        errno = 0;
    }

    vsprintf(pmsg, fmt, ap);

    reason = (InternalReasonPtr)
	InsecureAlloc(strlen(msg) + 1 + sizeof(InternalReason));
    if (reason == NULL) {
        reason = &IRBadAllocation;
    } else {
        reason->msg = (char *)(reason + 1);
	strcpy(reason->msg, msg);
	reason->allocated = True;
	reason->hint = hint;
	reason->nodeName = NULL;
	reason->next = NULL;
    }

    va_end(ap);

    return(reason);
}

static InternalReasonPtr
VaCreateReason(char *fmt, va_list ap)
{
    char msg[ BUFSIZ ];
    InternalReasonPtr reason;

    vsprintf(msg, fmt, ap);

    reason = (InternalReasonPtr)
	InsecureAlloc(strlen(msg) + 1 + sizeof(InternalReason));
    if (reason == NULL) {
        reason = &IRBadAllocation;
    } else {
	reason->msg = (char *)(reason + 1);
	strcpy(reason->msg, msg);
	reason->allocated = True;
	reason->hint = IR_Unknown;
	reason->nodeName = NULL;
	reason->next = NULL;
    }

    return(reason);
}

InternalReasonPtr
CreateReason(char *fmt, ...)
{
    va_list ap;
    InternalReasonPtr reason;

    va_start(ap, fmt);
    reason = VaCreateReason(fmt, ap);
    va_end(ap);

    return(reason);
}

InternalReasonPtr
BadFileReason(char *fmt, ...)
{
    va_list ap;
    InternalReasonPtr reason;

    va_start(ap, fmt);
    reason = VaCreateReason(fmt, ap);
    reason->hint = IR_BadIOFile;
    va_end(ap);

    return(reason);
}

InternalReasonPtr
BadSyntaxReason(char *fmt, ...)
{
    va_list ap;
    InternalReasonPtr reason;

    va_start(ap, fmt);
    reason = VaCreateReason(fmt, ap);
    reason->hint = IR_BadSyntax;
    va_end(ap);

    return(reason);
}

InternalReasonPtr
BadValueReason(char *fmt, ...)
{
    va_list ap;
    InternalReasonPtr reason;

    va_start(ap, fmt);
    reason = VaCreateReason(fmt, ap);
    reason->hint = IR_BadValue;
    va_end(ap);

    return(reason);
}

InternalReasonPtr
InternalProblem(char *fmt, ...)
{
    va_list ap;
    InternalReasonPtr reason;

    va_start(ap, fmt);
    reason = VaCreateReason(fmt, ap);
    reason->hint = IR_InternalProblem;
    va_end(ap);

    return(reason);
}

InternalReasonPtr
AllocReason(char *what, int size)
{
    InternalReasonPtr reason;

    reason = CreateReason("Can't allocate %d bytes for %s.\n", size, what);
    reason->hint = IR_BadAlloc;
    return(reason);
}

static InternalReasonPtr
CopyReason(InternalReasonPtr irp)
{
    InternalReasonPtr newIrp;

    newIrp = (InternalReasonPtr)
	InsecureAlloc(strlen(irp->msg) + 1 + sizeof(InternalReason));
    if (newIrp == NULL)
	return(NULL);
    newIrp->msg = (char *)(newIrp + 1);
    strcpy(newIrp->msg, irp->msg);
    newIrp->hint      = irp->hint;
    newIrp->nodeName  = irp->nodeName;
    newIrp->nodeIndex = irp->nodeIndex;
    newIrp->allocated = True;
    newIrp->next      = NULL;

    return(newIrp);
}

void
AppendReason(InternalReasonPtr *list, InternalReasonPtr nextIrp)
{
    InternalReasonPtr last;

    if (! nextIrp->allocated
     && NULL == (nextIrp = CopyReason(nextIrp))) {
	/*
	 * Drop the the list and nextIrp on the floor, and change
	 * the main reason to an allocation reason.
	 */
	*list = &IRBadAllocation;
	return;
    }

    nextIrp->next = NULL;
    if (*list == NULL) {
	*list = nextIrp;
	return;
    }

    /*
     * Find the next-to-last item and stick this one the end
     */
    for (last = *list; last->next != NULL; )
	last = last->next;

    last->next = nextIrp;
}

void
FreeReason(InternalReasonPtr reason)
{
    InternalReasonPtr next;

    while (reason) {
	next = reason->next;
	if (reason->allocated)
	    free((void *)reason);
	reason = next;
    }
}

void
FormatReason(InternalReasonPtr reason, char *msg)
{
    char *p = msg;

    while (reason) {
	if (reason->nodeName) {
	    sprintf(p, "%s, node %d: ", reason->nodeName, reason->nodeIndex+1);
	    p += strlen(p);
	}
	if (reason->hint == IR_InternalProblem) {
	    sprintf(p, "(internal problem) ");
	    p += strlen(p);
	}
	sprintf(p, "%s", reason->msg);
	p += strlen(p);
	reason = reason->next;
    }
}

void
ShowReason(void (*dpyFunc)(char *, ...), InternalReasonPtr reason)
{
    char msg[ 1024 ];

    FormatReason(reason, msg);
    (*dpyFunc)("%s", msg);
}

void
TtyShowReason(InternalReasonPtr reason)
{
    char msg[1024];

    FormatReason(reason, msg);
    ProgMsg("%s", msg);
}
