/*
 *      (c) Copyright 1992 North Valley Research, Inc.
 *              All Rights Reserved
 *
 *      Confidential -- All proprietary rights reserved.
 *      Licensed Material -- Property of North Valley Research, Inc.
 *
 *      This program is the property of North Valley Research, Inc. or others
 *      from whom North Valley Research has obtained a licensing right, and is
 *      considered by North Valley Research to be confidential.  It is protected
 *      by U.S. copyright law as an unpublished work and is
 *      furnished pursuant to a written license agreement.  It may
 *      not be used, copied or otherwise reproduced, or disclosed to
 *      others except in accordance with the terms and conditions of
 *      that agreement.
 *
 *      NAME
 *              nvrClockUnix.c - generic unix clock mechanisms
 *
 *      REVISION
 *              $Header: /usr/local/src/nvrToMIT/RCS/nvrClockUnix.c,v 1.3 1996/01/03 18:51:48 toddb Exp $
 */

#include "nvrClockI.h"
#include "nvrAlloc.h"

/***************************************************************************
 *
 * Unix specific functions
 *
 ***************************************************************************/
#include <sys/types.h>
#include <signal.h>

#ifdef _HPUX_SOURCE
# include <sys/times.h>
# include <sys/param.h>
# define TimeOfDay(tvp,tzp)     gettimeofday(tvp,tzp)
# define SignalFunc signal
#endif

#if defined(sgi) || defined(Solaris1)
# include <sys/time.h>
# include <sys/resource.h>
# include <unistd.h>
# define TimeOfDay(tvp,tzp)     gettimeofday(tvp,tzp)
# define SignalFunc signal
#endif

#ifdef Solaris2
# include <sys/time.h>
# include <sys/resource.h>
# include <sys/procfs.h>
# include <unistd.h>
# include <fcntl.h>
   extern int gettimeofday(struct timeval *);
# define TimeOfDay(tvp,tzp)     gettimeofday(tvp)
#define SignalFunc sigset
#endif

typedef struct _unixPrivClockKeep {
    struct timeval todStart;	/* time of day at (*->init)() call */
    struct timeval tod;		/* last known time of day */
    double microToTicks;	/* pub->resolution / MicroSecPerSec */
    PrivClockKeep super;	/* must be last: public stuff */
} UnixPrivClockKeep, *UnixPrivClockKeepPtr; 

#define UnixClockSize   sizeof(UnixPrivClockKeep)

#define PubClockToUnixClock(pub) \
    (UnixPrivClockKeepPtr)(UCharP(pub) - (UnixClockSize - PubClockSize))

static void
UnixClockUpdate(ClockKeepPtr pubClock, short clock)
{
    ULong changeInSeconds, elapsed, lastElapsed;
    long changeInMicroSecs;
    UnixPrivClockKeepPtr uclk =
	(UnixPrivClockKeepPtr) PubClockToUnixClock(pubClock);

    /*
     * Common code...
     */
    (void) TimeOfDay(&uclk->tod, NULL);
    changeInSeconds   = uclk->tod.tv_sec - uclk->todStart.tv_sec;
    changeInMicroSecs = (long)uclk->tod.tv_usec - (long)uclk->todStart.tv_usec;
    if (changeInMicroSecs < 0) {
	changeInMicroSecs += MicroSecPerSec;
	changeInSeconds--;
    }
    elapsed = changeInSeconds * uclk->super.pub.resolution;
    elapsed += changeInMicroSecs * uclk->microToTicks;

    /*
     * We are only interested in the difference between the current and
     * previous elapsed time.  This amount is whats charged to the current
     * clock.
     */
    lastElapsed = uclk->super.elapsed;
    uclk->super.elapsed = elapsed;
    elapsed -= lastElapsed;

    uclk->super.pub.clocks[ uclk->super.pub.curClock ] += elapsed;
    uclk->super.pub.curClock = clock;
}

static ULong
UnixClockGettime(ClockKeepPtr pubClock)
{
    ULong changeInSeconds, elapsed;
    long changeInMicroSecs;
    UnixPrivClockKeepPtr uclk =
        (UnixPrivClockKeepPtr) PubClockToUnixClock(pubClock);
    struct timeval now;

    /*
     * Common code...
     */
    (void) TimeOfDay(&now, NULL);
    changeInSeconds   = now.tv_sec - uclk->todStart.tv_sec;
    changeInMicroSecs = (long)now.tv_usec - (long)uclk->todStart.tv_usec;
    if (changeInMicroSecs < 0) {
	changeInMicroSecs += MicroSecPerSec;
	changeInSeconds--;
    }
    elapsed = changeInSeconds * uclk->super.pub.resolution;
    elapsed += changeInMicroSecs * uclk->microToTicks;

    return(elapsed);
}

static void
UnixClockFree(ClockKeepPtr pubClock)
{
    UnixPrivClockKeepPtr uclk =
        (UnixPrivClockKeepPtr) PubClockToUnixClock(pubClock);

    free((void *) uclk);
}

static void
UnixClockInit(ClockKeepPtr pubClock)
{
    int i;
    UnixPrivClockKeepPtr uclk =
        (UnixPrivClockKeepPtr) PubClockToUnixClock(pubClock);

    uclk->super.pub.secPerTick = 1.0 / pubClock->resolution;
    uclk->super.pub.init = UnixClockInit;
    uclk->super.pub.gettime = UnixClockGettime;
    uclk->super.pub.update = UnixClockUpdate;
    uclk->super.pub.free = UnixClockFree;

    for (i = uclk->super.nClocks-1; i >= 0; i--) 
	uclk->super.pub.clocks[ i ] = 0;

    uclk->super.nativeResolution = MicroSecPerSec;
    uclk->super.elapsed = 0;
    uclk->super.elapsed = 0;

    (void) TimeOfDay(&uclk->tod, NULL);
    uclk->todStart = uclk->tod;
    uclk->microToTicks =
	(double) uclk->super.pub.resolution / uclk->super.nativeResolution;
}

ClockKeepPtr
CreateClockKeep(short nClocks, long resolution, InternalReasonPtr *pIrp)
{
    UnixPrivClockKeepPtr uclk;
    int allocSize = sizeof(UnixPrivClockKeep)
		  + sizeof(uclk->super.pub.clocks[0]) * nClocks;

    uclk = (UnixPrivClockKeepPtr)
	AllocNamedMem(allocSize, "unix clock storage", pIrp);
    if (uclk == NULL)
	return(NULL);
    uclk->super.pub.resolution = resolution;
    uclk->super.nClocks = nClocks;

    UnixClockInit( &uclk->super.pub );

    return(&uclk->super.pub);
}
