/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	nvrTimer.c - general timing functions
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/nvrTimer.c,v 3.29 1995/06/19 05:01:07 toddb Exp $
 */
#define NeedFunctionPrototypes 1

#include "nvrTypes.h"
#include "longjump.h"
#include "nvrDebug.h"
#include "nvrString.h"
#include "nvrAlloc.h"
#include "osio.h"

#define TIMING_ENABLED	/* the library always defines it */
#include "nvrTimer.h"

void RestoreContext(jmp_buf pContext, int returnValue)
{
    longjmp(pContext, returnValue);
}

long TimeDifference(TimeSpecPtr t1, TimeSpecPtr t2, TimeSpecPtr diff)
{
    long seconds;
    long microSeconds;
    long maxPositive;

    microSeconds = t2->microSeconds - t1->microSeconds;
    seconds = t2->seconds - t1->seconds;

    /*
     * if they're in the same second, the difference of the microseconds
     * is the result
     */
    if (seconds == 0)
    {
	;
    }

    /*
     * If they have the same sign, then we are done
     */
    else if (((ULong)microSeconds >> 31) == ((ULong)seconds >> 31))
    {
	;
    }
    else if (microSeconds >= 0)
    {
	seconds++;	/* negative number */
	microSeconds -= MicroSecPerSec;
    }
    else
    {
	seconds--;	/* positive number */
	microSeconds += MicroSecPerSec;
    }

    if (diff) {
	diff->seconds = seconds;
	diff->microSeconds = microSeconds;
    }

    /*
     * Returns signed value that represents the sign of the result
     * If the number is too big, return the biggest number possible
     */
    maxPositive = 0x7fffffff/MicroSecPerSec;
    if (seconds > maxPositive)
	return (0x7fffffff);			/* largest positive number */
    if (seconds <= -maxPositive)
	return (0x80000000+1);			/* largest negative number */
    return (seconds*MicroSecPerSec + microSeconds);
}

InternalReasonPtr 
InitTimerInc(TimerIncPtr timer, double rate, long unitSize)
{
    double realIncrement;
    InternalReasonPtr reason;

    realIncrement = unitSize / rate;
    timer->increment = realIncrement;
    if (timer->increment == 0) {
	reason = InternalProblem("%s: rate (%f) %s (%d)\n",
	    "InitTimerInc", rate, "must be larger than unit", unitSize);
	return(reason);
    }
    timer->remainder = unitSize * (realIncrement - timer->increment);
    timer->denom = unitSize;
    timer->accumulator = 0;
    return(NULL);
}

/*
 * Return the next increment for the specified timer.  Note that since the
 * accumulater is updated, this function should be called exactly once
 * for each update.
 */
ULong
TimerIncrement(TimerIncPtr inc)
{
    long accumulator, tick;

    tick = inc->increment;
    accumulator = inc->accumulator + inc->remainder;
    if (accumulator >= inc->denom) {
        accumulator -= inc->denom;
        tick++;
    }
    inc->accumulator = accumulator;

    return( tick );
}

long
CalendarToSeconds(CalendarSpecPtr cal)
{
    long second, month;
    long secondsPerDay = 60 * 60 * 24;
    static UChar daysPerMonth[] = {
	0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
    };

    for (second = 0, month = 1; month < cal->month; month++)
	second += secondsPerDay * daysPerMonth[ month ];
    second += secondsPerDay * cal->day;
    second += cal->hour * 60 * 60
	     + cal->minute * 60
	     + cal->second;
    return(second);
}
