/*
 *	(c) Copyright 1992 North Valley Research, Inc.
 *		All Rights Reserved
 *
 *	Confidential -- All proprietary rights reserved.
 *	Licensed Material -- Property of North Valley Research, Inc.
 *
 *	This program is the property of North Valley Research, Inc. or others
 *	from whom North Valley Research has obtained a licensing right, and is
 *	considered by North Valley Research to be confidential.  It is protected
 *	by U.S. copyright law as an unpublished work and is
 *	furnished pursuant to a written license agreement.  It may
 *	not be used, copied or otherwise reproduced, or disclosed to
 *	others except in accordance with the terms and conditions of
 *	that agreement.
 *
 *	NAME
 *		nvrTimer.h - timer constants and definitions
 *
 *	DESCRIPTION
 *		Timer data definition
 *
 *	REVISION
 *		$Header: /usr/local/src/nvrToMIT/RCS/nvrTimer.h,v 1.22 1995/06/19 05:01:07 toddb Exp $
 */
#ifndef _TIMER_H_
#define _TIMER_H_

#include "nvrTypes.h"
#include "nvrReason.h"
#include "longjump.h"

/*
 * If you need something that gets called too often for a subroutine
 * then use this.
 */
#define TimerIncrementMacro(inc)					\
    (									\
	( ((inc).accumulator += (inc).remainder) >= (inc).denom )	\
	    ? ((inc).accumulator -= (inc).denom, (inc).increment+1)	\
	    : (inc).increment						\
    )

/*
 *	TYPEDEFS
 */
typedef struct _timeSpec {
    long seconds;		/* signed, for signed comparisons */
    long microSeconds;		/* signed, for signed comparisons */
} TimeSpec, *TimeSpecPtr;

typedef struct _calendarSpec {
    char month;
    char day;
    char hour;
    char minute;
    char second;
} CalendarSpec, *CalendarSpecPtr;

typedef struct {
    long increment;		/* increment size */
    long remainder;		/* fractional increment */
    long denom;			/* the denominator */
    long accumulator;		/* accumulated remainders */
} TimerInc, *TimerIncPtr;

typedef struct {
    ULong tickUser;		/* user time in ticks */
    ULong tickSystem;		/* system time in ticks */
    ULong tickElapsed;		/* elapsed time in ticks */
    ULong ticksPerSec;		/* ticks per second */
    ULong nFaultsNoIO;		/* number of page faults w/o I/O */
    ULong nFaultsIO;		/* number of page faults w/ I/O */
    ULong nSwaps;		/* number of swap operations */
    ULong nVoluntaryContextSwitches; /* # of involuntary context switches */
    ULong nInvoluntaryContextSwitches; /* # of voluntary context switches */
} CpuUsage, *CpuUsagePtr;

/*
 *	DEFINES
 */
#define MicroSecPerSec		1000000
#define MilliSecPerSecond	   1000


#ifdef TIMING_ENABLED
/*
 * These work like the debug mechanism, but are used for timing tasks.
 * The idea is that you may want timing enabled independent of debugging.
 */
#define TimerTask(flags, whatIsIt, task)				\
    if (GetTimerBit(flags, whatIsIt)) {				\
	task;								\
    }

#define SetTimerBit(flags, bitnum)					\
    (flags)[(bitnum)>>5] |= (1<<((bitnum) & 0x1f))
#define GetTimerBit(flags, bitnum)					\
    ((flags)[(bitnum)>>5] & (1<<((bitnum) & 0x1f)))

/*
 *	EXTERNS
 */

CExternal void StartTimer(void);
CExternal void StopAllTimers(void);
CExternal void ClearFrameTimer(void);
CExternal int BlockTimer(void);

extern char *TimingLabels[];

#else /* TIMING_ENABLED */

#define TimerTask(flags, whatIsIt, task)	/* nothing */
#define SetTimerBit(flags, bitnum)		/* nothing */
#define GetTimerBit(flags, bitnum)		/* nothing */
#define StopAllTimers()				/* nothing */
#define StartTimer()				/* nothing */
#define ClearFrameTimer()			/* nothing */
#define BlockTimer()				/* nothing */


#endif /* TIMING_ENABLED */

CExternal CpuUsagePtr StopTimer(
    void
);

CExternal void ShowTimeSpecAsClock(
    TimeSpecPtr
);

CExternal void PrintTimeUsage(
    CpuUsage *,
    boolean
);

CExternal char *FormatTimeUsage(
    CpuUsage *,
    boolean
);

CExternal char *FormatCalendar(
    char *buf,
    int year,
    CalendarSpecPtr cal
);

CExternal void SetFrameTimer(
    long,		/* microseconds per frame */
    IntFuncPtr		/* callback */
);

CExternal void WaitForTimer(
    void
);

CExternal void SetTimerFunc(
    IntFuncPtr
);

CExternal void AllowTimer(
    int
);

CExternal long CalendarToSeconds(
    CalendarSpecPtr
);

CExternal int MonthNumber(
    char *monthName
);

CExternal void TimeNow(
    TimeSpecPtr
);

CExternal long TimeDifference(
    TimeSpecPtr,
    TimeSpecPtr,
    TimeSpecPtr	/* result */
);

CExternal char *FormatTime(
    TimeSpecPtr
);

CExternal void RestoreContext(
    StackInfo,	/* pointer to context */
    int		/* return value */
);

CExternal InternalReasonPtr InitTimerInc(
    TimerIncPtr timer,
    double rate,		/* number of times per unitSize */
    long unitSize
);

CExternal ULong TimerIncrement(
    TimerIncPtr
);

CExternal ULong * CheckRTClock(
    ULong *newClock
);

CExternal void RegisterTimerSummary(
    void (*func)(PrivPointer data, ULong totalTime),
    PrivPointer arg,
    boolean enable
);

CExternal void CallTimerSummary(
    void
);

#endif /* _TIMER_H_ */
