/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	nvrTimer.c - general timing functions
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/nvrTimerFmt.c,v 1.2 1995/06/19 05:01:07 toddb Exp $
 */
#define NeedFunctionPrototypes 1

#include "nvrTypes.h"
#include "longjump.h"
#include "nvrDebug.h"
#include "nvrString.h"
#include "nvrAlloc.h"
#include "osio.h"

#define TIMING_ENABLED	/* the library always defines it */
#include "nvrTimer.h"

char *
FormatTime(TimeSpecPtr time)
{
    static char timeMsg[ 1024 ];
    char *ptr = timeMsg;
    boolean negative = False;
    int hours, minutes;
    long seconds, microSeconds;

    if ((seconds = time->seconds) < 0) {
	negative = True;
	seconds = - seconds;
    }
    if ((microSeconds = time->microSeconds) < 0) {
	negative = True;
	microSeconds = -microSeconds;
    }

    seconds %= (60*60*24); /* lop of the days */

    hours = seconds / (60*60);
    seconds -= hours * (60*60);
    minutes = seconds / 60;
    seconds -= minutes * 60;
    
    *ptr++ = negative ? '-' : ' ';
    if (hours) {
	sprintf(ptr, "%02d:", hours);
	while (*++ptr)
	    ;
    } else {
	*ptr++ = ' ';
	*ptr++ = ' ';
	*ptr++ = ':';
    }
    if (minutes) {
	sprintf(ptr, "%02d:", minutes);
	while (*++ptr)
	    ;
    } else {
	*ptr++ = ' ';
	*ptr++ = ' ';
	*ptr++ = ':';
    }
    sprintf(ptr, "%02d.%02d", (int)seconds,
	    (int)((microSeconds*100+50) / MicroSecPerSec));
    while (*ptr++)
	;
    return(timeMsg);
}

/*
 * This is provided as a convenient method of printing the timing
 * information.  The returned string is intended to be given to
 * Msg() or NormalOutput().  Like
 *
 *	Msg("%s\n", FormatTimeUsage(usage, brief));
 */
char *FormatTimeUsage(CpuUsagePtr usage, boolean brief)
{
    static char usageMsg[ 1024 ];
    char *p = usageMsg;

    if (brief) {
	sprintf(p, "%5.2fs %5.2fu %5.2fr", 
	    ((double)usage->tickSystem) / usage->ticksPerSec,
	    ((double)usage->tickUser) / usage->ticksPerSec,
	    ((double)usage->tickElapsed) / usage->ticksPerSec);
	return (p);
    }

    sprintf(p, "  %5.3f system %5.3f user %5.3f real", 
	((double)usage->tickSystem) / usage->ticksPerSec,
	((double)usage->tickUser) / usage->ticksPerSec,
	    ((double)usage->tickElapsed) / usage->ticksPerSec);
    p += strlen(p);

    if (usage->nFaultsNoIO) {
	sprintf(p, "\n  %d faults w/o IO", (int)usage->nFaultsNoIO);
	p += strlen(p);
    }
    if (usage->nFaultsIO) {
	sprintf(p, "\n  %d faults w/ IO", (int)usage->nFaultsIO);
	p += strlen(p);
    }
    if (usage->nSwaps) {
	sprintf(p, "\n  %d swaps", (int)usage->nSwaps);
	p += strlen(p);
    }
    if (usage->nInvoluntaryContextSwitches) {
	sprintf(p, "\n  %d involuntary context switches",
	    (int)usage->nInvoluntaryContextSwitches);
	p += strlen(p);
    }
    if (usage->nVoluntaryContextSwitches) {
	sprintf(p, "\n  %d voluntary context switches",
	    (int)usage->nVoluntaryContextSwitches);
	p += strlen(p);
    }
    return (usageMsg);
}

void PrintTimeUsage(CpuUsagePtr usage, boolean brief)
{
    Msg("%s\n", FormatTimeUsage(usage, brief));
}

void ShowTimeSpecAsClock(TimeSpecPtr clock)
{
    long totalSeconds = clock->seconds;
    long hours, minutes, seconds;
    double fseconds;

    seconds = totalSeconds % 60;
    totalSeconds /= 60;
    minutes = totalSeconds % 60;
    totalSeconds /= 60;
    hours = totalSeconds % 24;

    fseconds = seconds + clock->microSeconds / (double)MicroSecPerSec;

    Msg("%2d:%02d:%05.2f", hours, minutes, fseconds);
}

typedef void (*SummaryFunc)(PrivPointer data, ULong totalTime);

typedef struct {
    SummaryFunc func;
    PrivPointer arg;
} SummaryPair, *SummaryPairPtr;

static struct {
    int count;
    boolean initialized;
    TimeSpec startTime; 
    SummaryPairPtr list;
} Summary;

void
RegisterTimerSummary(void (*func)(PrivPointer data, ULong total),
		     PrivPointer arg,
		     boolean enable)
{
    int i;

    if (!Summary.initialized) {
	TimeNow(&Summary.startTime);
    }
    for (i = 0; i < Summary.count; i++) {
	if (enable && Summary.list[ i ].func == NULL) {
	    Summary.list[ i ].func = func;
	    Summary.list[ i ].arg = arg;
	    return;
	}
	if (!enable && Summary.list[ i ].func == func) {
	    Summary.list[ i ].func = NULL;
	    Summary.list[ i ].arg = NULL;
	    return;
	}
    }
    if (! enable)
	return;

    Summary.count++;
    Summary.list = (SummaryPairPtr)
	NullRealloc((char *)Summary.list,
		    sizeof(SummaryPair) * (Summary.count));
    Summary.list[ i ].func = func;
    Summary.list[ i ].arg = arg;
}

void
CallTimerSummary(void)
{
    int i;
    TimeSpec now, diff;
    ULong msDiff;

    TimeNow(&now);
    TimeDifference(&Summary.startTime, &now, &diff);
    msDiff = diff.seconds * MicroSecPerSec + diff.microSeconds;

    for (i = 0; i < Summary.count; i++) {
        if (Summary.list[ i ].func)
            (*Summary.list[ i ].func)(Summary.list[ i ].arg, msDiff);
    }
}

char *
FormatCalendar(char *buf, int year, CalendarSpecPtr cal)
{
    char *p = buf;
    static char *monthNames[] = {
	"??", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
    };

    sprintf(p, "%s %.2d %02d:%02d:%02d",
	monthNames[ cal->month ],
	cal->day, cal->hour, cal->minute, cal->second);
    p += strlen(p);
    if (year)
	sprintf(p, " %d", year);

    return(buf);
}

int
MonthNumber(char *monthName)
{
    static NameEnumPair monthNames[] = {
	"Jan", 1,
	"Feb", 2,
	"Mar", 3,
	"Apr", 4,
	"May", 5,
	"Jun", 6,
	"Jul", 7,
	"Aug", 8,
	"Sep", 9,
	"Oct", 10,
	"Nov", 11,
	"Dec", 12,
	NULL, 0
    };

    return( NameToEnum(monthName, monthNames) );
}
