/*
 * (c) Copyright 1992 North Valley Research, Inc.
 *         All Rights Reserved
 *
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	unixOsio.c
 *
 * DESCRIPTION
 *	I/O functions that hide os dependencies
 *	OS-DEPENDENT
 *
 * REVISION
 *	$Header: /usr/local/src/nvrToMIT/RCS/unixOsio.c,v 1.21 1996/01/03 02:35:31 toddb Exp $
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>

#define GeneralAssert(mustBeTrue) _DoAssert(mustBeTrue)

#ifndef _HPUX_SOURCE
# ifdef SYSV
#  include <net/soioctl.h>
# else
#  include <sys/sockio.h>
#  include <sys/filio.h>
# endif
#endif /* hpux */

#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>

# ifdef Solaris2
#  include <sys/stropts.h>
#  include <string.h>
# endif

#ifdef _HPUX_SOURCE
# define ReinstallSignalHandler
# include <varargs.h>
# include <sys/ioctl.h>
#endif

#ifdef sgi
# include <stdlib.h>
# define ReinstallSignalHandler
#endif

static int GetDTableSize(void);

#define NeedFunctionPrototypes 1

#include "nvrTypes.h"
#include "nvrTimer.h"
#include "osioI.h"
#include "nvrDebug.h"

boolean
TtyAskUser(char *fmt, ...)
{
    va_list ap;
    char buf[ BUFSIZ ];

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    fflush(stderr);

    if (read(fileno(stderr), buf, BUFSIZ) <= 0) {
	SystemError("Can't read %d\n", fileno(stderr));
    }
    va_end(ap);
    return(buf[0] == 'y' || buf[0] == 'Y');
}

boolean
ChangeDirectory(char *path)
{
    if( chdir(path) < 0 ) {
	return(False);
    }
    return(True);
}

FileHandle
OpenFileForAppending(char *file)
{
    int fd;

    if ((fd = open(file, O_RDWR)) < 0)
	return (open(file, O_RDWR|O_CREAT, 0666));
    return (fd);
}

FileHandle
OpenFileForWriting(char *file)
{
    extern int errno;
    int fd;
    
    if (file[0] == '-') {
	if (strcmp(file,"-stdout-") == 0) {
	    return(1);
	}
	if (strcmp(file,"-stderr-") == 0) {
	    return(2);
	}
    }

    fd = open(file, O_WRONLY|O_CREAT, 0666);

    if (fd >= 0) {
	if (ftruncate(fd, 0) < 0)
	    errno = 0;
    }
    return(fd);
}

FileHandle
OpenFileForReading(char *file)
{
    int fd;

    if (file[0] == '-' && strcmp(file,"-stdin-") == 0) {
	return(0);
    }
    return ((FileHandle) open(file, O_RDONLY));
}

FileHandle
FileNo(BufFileHndl fp)
{
    return( fileno((FILE *)fp) );
}

char *
EnvironmentValue(char *name)
{
	return( getenv(name) );
}
BufFileHndl
ReopenBufferedFileForWriting(FileHandle fd)
{
    FILE *fp = fdopen(fd, "w");
    
    return((BufFileHndl)fp);
}

BufFileHndl
ReopenBufferedFileForReading(FileHandle fd)
{
    FILE *fp = fdopen(fd, "r");
    
    return((BufFileHndl)fp);
}

BufFileHndl
OpenBufferedFileForAppending(char *fileName)
{
    FILE *fp = fopen(fileName, "a");
    
    return((BufFileHndl)fp);
}

BufFileHndl
OpenBufferedFileForWriting(char *fileName)
{
    FILE *fp = fopen(fileName, "w");
    
    return((BufFileHndl)fp);
}

BufFileHndl
OpenBufferedFileForReading(char *fileName)
{
    FILE *fp = fopen(fileName, "r");
    
    return((BufFileHndl)fp);
}

FileHandle
GetBufferedFileHandle(BufFileHndl fp)
{
    return((FileHandle) fileno((FILE *)fp));
}

/*
 * Arrange for asynchronous notification of io
 * XXX if this ioctl fails, it may be because gcc isn't getting
 * the -traditional-cpp flag.  Without it, the first argument of
 * the _IOR macro is interpreted wrong.
 */
void
AllowAsyncNotification(FileHandle fd)
{
    int true = True;

    GeneralAssert (IoData.list[fd].handler != NULL);

    if (ioctl(fd, FIOASYNC, &true) < 0) {
	SystemError("can't set async io on client fd %d\n", fd);
	return;
    }
    IoData.list[fd].async = True;
}
    
void
BlockAsyncNotification(FileHandle fd)
{
    int false = False;

    GeneralAssert (IoData.list[fd].handler != NULL);

    if (ioctl(fd, FIOASYNC, &false) < 0) {
	SystemError("can't clear async io on client fd %d\n", fd);
	return;
    }
    IoData.list[fd].async = False;
}

int
BlockAsyncHandler()
{
    return( BlockSignals(SIGURG, SIGIO, 0) );
}

int
AllowAsyncHandler()
{
    return( UnblockSignals(SIGURG, SIGIO, 0) );
}


static void
AsyncHandler(int sig)
{
    fd_set inReady;
    int i, nFdsReady;
    extern int errno;

#ifdef ReinstallSignalHandler
    if (signal(sig, AsyncHandler) == (VoidFuncPtr)-1)
	ProgMsg("can't reset signal handler\n");
#endif /* ReinstallSignalHandler */

    inReady = IoData.inReady;
    nFdsReady = select(IoData.lastFd+1,
		       (SelectDataPtr)&inReady,
		       (SelectDataPtr)NULL,
		       (SelectDataPtr)NULL,
		       &IoData.noWait);

    if (nFdsReady < 0) {
	if (errno != EINTR)
	    SystemError("select failure\n");
    }

    for (i = IoData.lastFd; i >= 0; i--)
    {
	if (! IoData.list[i].async)
	    continue;
	if (FD_ISSET(i, &inReady)
	 || (IoData.list[i].bypassSelect))
	{
	    (*(IoData.list[i].handler))(i, IoData.list[i].pointer);
	}
    }
}

static boolean
WaitForFileIO(int fd, boolean input, TimeSpecPtr maxWait)
{
    fd_set ready;
    SelectDataPtr inReady, outReady;
    int nFdsReady;
    struct timeval osMaxWait, *timeout;
    extern int errno;

    if (input) {
	inReady = (SelectDataPtr)&ready;
	outReady = NULL;
    } else {
	inReady = NULL;
	outReady = (SelectDataPtr)&ready;
    }

    if (maxWait) {
	osMaxWait.tv_sec = maxWait->seconds;
	osMaxWait.tv_usec = maxWait->microSeconds;
	timeout = &osMaxWait;
    } else
	timeout = NULL;

    FD_ZERO(&ready);
    FD_SET(fd, &ready);
    nFdsReady = select(fd+1, inReady, outReady, NULL, timeout);

    if (nFdsReady < 0 && errno != EINTR)
	SystemError("select(%d,%x,%x,{%d,%d}) fails\n",
	    fd+1,
	    inReady ? *ULongP(inReady) : 0,
	    outReady ? *ULongP(outReady) : 0,
	    timeout ? timeout->tv_sec : 0,
	    timeout ? timeout->tv_usec : 0);

    return (nFdsReady > 0);
}

boolean
WaitForFileInput(int fd, TimeSpecPtr maxWait)
{
    return( WaitForFileIO(fd, True, maxWait) );
}

boolean
WaitForFileOutput(int fd, TimeSpecPtr maxWait)
{
    return( WaitForFileIO(fd, False, maxWait) );
}

void
WaitForSomething(TimeSpecPtr maxWait)
{
    fd_set inReady;
    int i, nFdsReady;
    struct timeval osMaxWait, *timeout;
    extern int errno;

    if (maxWait) {
	osMaxWait.tv_sec = maxWait->seconds;
	osMaxWait.tv_usec = maxWait->microSeconds;
	timeout = &osMaxWait;
    } else
	timeout = NULL;

    inReady = IoData.inReady;
    nFdsReady = select(IoData.lastFd+1,
		       (SelectDataPtr)&inReady,
		       (SelectDataPtr)NULL,
		       (SelectDataPtr)NULL,
		       timeout);

    if (nFdsReady < 0 && errno != EINTR) {
	SystemError("select(%d,%x,{%d,%d}) failure\n",
		IoData.lastFd+1,
		&IoData.inReady,
		timeout->tv_sec, timeout->tv_usec);
	return;
    }

    if (nFdsReady > 0) {
	for (i = IoData.lastFd; i >= 0; i--)
	{
	    if (FD_ISSET(i, &inReady))
	    {
		(*(IoData.list[i].handler))(i, IoData.list[i].pointer);
	    }
	}
    }
}

boolean
FileHandleIsTty(int fd)
{
    GeneralAssert(fd < FD_SETSIZE);

    if (IoData.list && IoData.list[fd].handler)
	return (IoData.list[fd].isAtty);
    else
        return (isatty(fd));
}

long
BytesAvailable(FileHandle fd)
{
    long bytesIn;

    if (ioctl(fd, FIONREAD, &bytesIn) < 0) {
	SystemError("can't FIONREAD fd %d\n", fd);
	bytesIn = 0;
    }
    return (bytesIn);
}

int
NonBlockingRead(FileHandle fd, char *pBuf, long spaceAvailable)
{
    long bytesIn;
    extern int errno;

    GeneralAssert(IoData.lastFd >= fd);

    /* XXX special case for stdin */
    if (IoData.list[ fd ].isAtty) {
	bytesIn = BytesAvailable(fd);
        /*
         * You should use the command
         *   dd bs=1 | zip ...
         * to do debugging
         */
	if (bytesIn == 0)
	    return (0);
    }
    bytesIn = read(fd, pBuf, spaceAvailable);
    if (bytesIn < 0) {
        if (errno == EAGAIN)		/* SYS5 non-blocking */
            return (0);
        if (errno == EWOULDBLOCK)	/* 4.2bsd non-blocking */
            return (0);
        SystemError("can't read from fd %d\n",  fd);
	return (-1);
    }
    if (bytesIn == 0) {
	return (-1);
    }

    return (bytesIn);
}

#include "fileAccess.h"

void
ShowOpenFileDescriptors()
{
    char buf[ 100 ];
    int possibleFds = GetDTableSize();
    int i, type;
    boolean readable, writeable;

    for (i = 0; i < possibleFds; i++) {
        type = FdFileType(i);
        if (! type)
            continue;
        writeable = (write(i, buf, 0) == 0);
        readable = (read(i, buf, 0) == 0);
        ProgMsg("fd %d: %s, mode %c%c\n", i,
		TypeName(type), " r"[readable], " w"[writeable]);
    }
}

/*
 * close all open file descriptors except those specified
 * XXX note that we can only specify file descriptors <= 31.
 */
void
CloseOpenFileDescriptors(ULong retained)
{
    int possibleFds = GetDTableSize();
    int i;
    extern int errno;

    for (i = 0; i < possibleFds; i++) {
	if (retained & (1<<i))
	    continue;
	close(i);
    }
    errno = 0;	/* clear side-effects of close() */
}

void
WaitForSignal(void)
{
    pause();
}


/************************************************
 *	*** Start platform specific functions ***
 ************************************************/

/*
 * Set the connection non-blocking... but not if its a tty
 */
#ifdef Solaris1

static int
GetDTableSize()
{
    return(getdtablesize());
}

void
AllocAsyncHandler(int fd,
		  VoidFuncPtr handler,
		  boolean bypassSelect,
		  void *pointer)
{
    AllocHandler(fd, handler, bypassSelect, pointer);

    /*
     * Set the connection non-blocking... but not if its a tty
     */

    SetNonBlocking(fd);

    if( ioctl(fd,FIOSETOWN,&IoData.pid) < 0 ) {
	SystemError("can't set async pid on client fd %d\n", fd);
	return;
    }
    if (! IoData.signalHandlerSet) {
	if (signal(SIGIO, AsyncHandler) == (VoidFuncPtr)-1
	 || signal(SIGURG, AsyncHandler) == (VoidFuncPtr)-1) {
	    SystemError("can't set signal handler\n");
	    return;
	}
	IoData.signalHandlerSet = True;
    }

    AllowAsyncNotification(fd);
}

void
SetNonBlocking(int fd)
{
    int true = True;

    if (IoData.list[fd].handler != NULL && IoData.list[fd].isAtty)
	return;
    if (ioctl(fd, FIONBIO, &true) < 0) {
	SystemError("can't set non-blocking io on client fd %d\n", fd);
	return;
    }
}

#endif /* Solaris1 */

#ifdef Solaris2

static int
GetDTableSize()
{
    return( sysconf(_SC_OPEN_MAX) );
}

void
AllocAsyncHandler(int fd,
		  VoidFuncPtr handler,
		  boolean bypassSelect,
		  void *pointer)
{
    AllocHandler(fd, handler, bypassSelect, pointer);

    /*
     * Set the connection non-blocking... but not if its a tty
     */

    SetNonBlocking(fd);

    if( ioctl(fd,I_SETSIG,S_INPUT|S_OUTPUT|S_MSG) < 0 ) {
	SystemError("can't set async pid on client fd %d\n", fd);
	return;
    }
    if (! IoData.signalHandlerSet) {
	if (sigset(SIGIO, AsyncHandler) == (VoidFuncPtr)-1
	 || sigset(SIGURG, AsyncHandler) == (VoidFuncPtr)-1) {
	    SystemError("can't set signal handler\n");
	    return;
	}
	IoData.signalHandlerSet = True;
    }

    AllowAsyncNotification(fd);
}

void
SetNonBlocking(int fd)
{
    int true = True;

    if (IoData.list[fd].handler != NULL && IoData.list[fd].isAtty)
	return;
    if (ioctl(fd, FIONBIO, &true) < 0) {
	SystemError("can't set non-blocking io on client fd %d\n", fd);
	return;
    }
}
#endif /* Solaris2 */

#ifdef _HPUX_SOURCE

static int
GetDTableSize()
{
    return ( FD_SETSIZE );
}

void
AllocAsyncHandler(int fd,
		  VoidFuncPtr handler,
		  boolean bypassSelect,
		  void *pointer)
{
    AllocHandler(fd, handler, bypassSelect, pointer);

    /*
     * Set the connection non-blocking... but not if its a tty
     */

    SetNonBlocking(fd);

    if( ioctl(fd,FIOSSAIOOWN,&IoData.pid) < 0 ) {
	SystemError("can't set async pid on client fd %d\n", fd);
	return;
    }
    if (! IoData.signalHandlerSet) {
	if (signal(SIGIO, AsyncHandler) == (VoidFuncPtr)-1
	 || signal(SIGURG, AsyncHandler) == (VoidFuncPtr)-1) {
	    SystemError("can't set signal handler\n");
	    return;
	}
	IoData.signalHandlerSet = True;
    }

    AllowAsyncNotification(fd);
}

void
SetNonBlocking(int fd)
{
    if (IoData.list[fd].handler != NULL && IoData.list[fd].isAtty)
	return;

    if (fcntl(fd, F_SETFL, FNBLOCK) < 0) {
	SystemError("Can't fcntl(%d, F_SETFL, FNBLOCK).\n", fd);
	return;
    }
}
#endif /* _HPUX_SOURCE */

#ifdef sgi

static int
GetDTableSize()
{
    return( getdtablesize() );
}

static void
AsyncSignalsForTty(FileHandle fd)
{
    fd_set read_fds;
    fd_set write_fds;
    fd_set except_fds;
    int nReady, pid, ppid;

    Msg("AsyncSignalsForTty: forking for tty %d\n", fd);

    pid = fork();
    switch (pid) {
    case 0:		/* child process */
	break;
    case -1:
	Msg("AsyncSignalsForTty: fork can't fork");
	/* fall through */
    default:		/* parent process */
	return;
    }

    ppid = getppid();
    CloseOpenFileDescriptors(1<<fd);
    while (1) {
	FD_ZERO(&read_fds);
	FD_ZERO(&write_fds);
	FD_ZERO(&except_fds);
	FD_SET(fd, &read_fds);
	nReady = select(1, &read_fds, &write_fds, &except_fds, 0);
	if (nReady == -1) {
	    Error("AsyncSignalsForTty: select returned -1.  Exiting\n");
	    exit(0);
	} else if (nReady == 1) {
	    if (kill(ppid, SIGIO) < 0) {
		Error("AsyncSignalsForTty: parent gone?  Exiting.\n");
		exit(0);
	    }
	    sleep(2);
	}
    }
}

void
AllocAsyncHandler(int fd,
		  VoidFuncPtr handler,
		  boolean bypassSelect,
		  void *pointer)
{
    AllocHandler(fd, handler, bypassSelect, pointer);

    /*
     * Set the connection non-blocking... but not if its a tty
     */

    SetNonBlocking(fd);

    if (! IoData.signalHandlerSet) {
	if (signal(SIGIO, AsyncHandler) == (VoidFuncPtr)-1
	 || signal(SIGURG, AsyncHandler) == (VoidFuncPtr)-1) {
	    SystemError("can't set signal handler\n");
	    return;
	}
	IoData.signalHandlerSet = True;
    }

    if (IoData.list[ fd ].isAtty) {
	AsyncSignalsForTty(fd);
    } else {
	if( fcntl(fd,F_SETOWN,IoData.pid) < 0 ) {
	    SystemError("can't set async pid on client fd %d\n", fd);
	    return;
	}
    }

    AllowAsyncNotification(fd);
}

void
SetNonBlocking(int fd)
{
    if (IoData.list[fd].handler != NULL && IoData.list[fd].isAtty)
	return;

    if (fcntl(fd, F_SETFL, FNONBLK) < 0) {
	SystemError("Can't fcntl(%d, F_SETFL, FNONBLK).\n", fd);
	return;
    }
}
#endif /* sgi */


/*
 * BlockSignals, UnblockSignals - Solaris2 is the only one that's
 *	different
 */

#ifdef Solaris2


int
BlockSignals(int first, ...)       
{
    sigset_t block;
    int signalnum;
    va_list ap;
    
    sigemptyset(&block);
    signalnum = first;
    va_start(ap,first);
    while ( signalnum != 0 ) {
	sigaddset(&block,signalnum);
	signalnum = va_arg(ap,int);
    }
    va_end(ap);

    return(SIG_BLOCK, &block, NULL);
}

int
UnblockSignals(int first, ...)       
{
    sigset_t block;
    int signalnum;
    va_list ap;
    
    sigemptyset(&block);
    signalnum = first;
    va_start(ap,first);
    while ( signalnum != 0 ) {
	sigaddset(&block,signalnum);
	signalnum = va_arg(ap,int);
    }
    va_end(ap);

    return(SIG_UNBLOCK, &block, NULL);
}

#else

int
BlockSignals(int first, ...)       
{
    
    int signalmask = 0;
    int signalnum;
    va_list ap;
    
    signalnum = first;
    va_start(ap,first);
    while ( signalnum != 0 ) {
	signalmask |= sigmask(signalnum);
	signalnum = va_arg(ap,int);
    }
    va_end(ap);


    return( sigblock(signalmask) );
}

int
UnblockSignals(int first, ...)       
{
    
    int signalmask = 0;
    int signalnum;
    va_list ap;
    
    signalnum = first;
    va_start(ap,first);
    while ( signalnum != 0 ) {
	signalmask |= sigmask(signalnum);
	signalnum = va_arg(ap,int);
    }
    va_end(ap);

    return( sigsetmask( sigblock(0) & ~(sigblock(signalmask)) ) );
}

#endif
