/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	unixTimer.c - unix-specific timing functions
 *
 * DESCRIPTION
 *	OS-DEPENDENT
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/unixTimer.c,v 1.9 1995/02/22 16:18:49 russ Exp $
 */
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>

#ifdef _HPUX_SOURCE
# include <sys/times.h>
# include <sys/param.h>
# define TimeOfDay(tvp,tzp)	gettimeofday(tvp,tzp)
# define SignalFunc signal
#endif

#if defined(sgi) || defined(Solaris1)
# include <sys/resource.h>
# include <unistd.h>
# define TimeOfDay(tvp,tzp)	gettimeofday(tvp,tzp)
# define SignalFunc signal
#endif

#ifdef Solaris2
# include <sys/resource.h>
# include <sys/procfs.h>
# include <unistd.h>
# include <fcntl.h>
   extern int gettimeofday(struct timeval *);
# define TimeOfDay(tvp,tzp)	gettimeofday(tvp)
#define SignalFunc sigset
#endif

#define NeedFunctionPrototypes 1
#define TIMING_ENABLED	/* the library always defines it */

#include "nvrTypes.h"
#include "longjump.h"
#include "nvrDebug.h"
#include "nvrString.h"
#include "osio.h"
#include "nvrTimer.h"

#define MAX_TIMER_STACK	5

#define GeneralAssert(mustBeTrue)       _DoAssert(mustBeTrue)

#define GetTimeDifference(result, first, second)			\
    {									\
	long seconds;							\
	if ( (seconds = (second)->tv_sec - (first)->tv_sec) ) {		\
	    result = seconds * MicroSecPerSec;				\
	    if ((second)->tv_usec < (first)->tv_usec)			\
		result -= (first)->tv_usec - (second)->tv_usec;		\
	    else							\
		result += (second)->tv_usec - (first)->tv_usec;		\
	} else {							\
	    result = (second)->tv_usec - (first)->tv_usec;		\
	}								\
    }

void ClearFrameTimer()
{
    struct itimerval clock;

    clock.it_value.tv_sec = clock.it_interval.tv_sec = 0;
    clock.it_value.tv_usec = clock.it_interval.tv_usec = 0;
    if (setitimer (ITIMER_REAL, &clock, (struct itimerval *)0) < 0)
        Fatal("can't clear interval timer\n");
}

#ifdef _HPUX_SOURCE
static VoidFuncPtr SavedTimerFunc;

static void
ResetTimer()
{
    if (SignalFunc(SIGALRM, ResetTimer) == (VoidFuncPtr)-1)
        Fatal("can't reset alarm signal handler\n");
    (*SavedTimerFunc)();
}

static VoidFuncPtr
SaveTimerFunc(VoidFuncPtr timerFunc)
{
    SavedTimerFunc = timerFunc;
    return(ResetTimer);
}

#else
#define SaveTimerFunc(timerFunc)	timerFunc
#endif

void SetTimerFunc(IntFuncPtr timerFunc)
{
    VoidFuncPtr voidTimerFunc = (VoidFuncPtr)timerFunc;

    voidTimerFunc = SaveTimerFunc(voidTimerFunc);
    if (SignalFunc(SIGALRM, voidTimerFunc) == (VoidFuncPtr)-1)
        Fatal("can't set alarm signal handler\n");
}

void SetFrameTimer(long usecPerPic, IntFuncPtr timerFunc)
{
    struct itimerval clock;

    SetTimerFunc(timerFunc);
    if (usecPerPic >= MicroSecPerSec) {
	clock.it_value.tv_sec = clock.it_interval.tv_sec =
	    usecPerPic/MicroSecPerSec;
	clock.it_value.tv_usec = clock.it_interval.tv_usec =
	    usecPerPic%MicroSecPerSec;
    } else {
	clock.it_value.tv_sec = clock.it_interval.tv_sec = 0;
	clock.it_value.tv_usec = clock.it_interval.tv_usec = usecPerPic;
    }
    if (setitimer (ITIMER_REAL, &clock, (struct itimerval *)0) < 0)
        Fatal("can't set interval timer\n");
}

#if 0
void AllowTimer(int key)
{
    sigsetmask(key);
}
#endif /* 0 */

#if 0
int BlockTimer()
{
    return (sigblock(sigmask(SIGALRM)));
}
#endif /* 0 */

int BlockTimer()
{
    return( BlockSignals(SIGALRM,0) );
}

void ClockResolution()
{
    long smallest = 0x7fffffff;		/* not very small */
    long largest = 0;			/* not very big */
    long average = 0;			/* ... to start with */
    long diff, timeRunning, nSamples;
    struct timeval start, last, now;

    NormalOutput("Clock resolution: ");

    if (TimeOfDay(&start, (struct timezone *)NULL) < 0)
	Fatal("can't get time of day\n");

    now = start;
    nSamples = 0;
    do {
	last = now;
	if (TimeOfDay(&now, (struct timezone *)NULL) < 0)
	    Fatal("can't get time of day\n");
	GetTimeDifference(diff, &last, &now);
	if (diff == 0)
	    continue;

	nSamples++;
	average += diff;
	if (diff > largest)
	    largest = diff;
	if (diff < smallest)
	    smallest = diff;

	GetTimeDifference(timeRunning, &start, &now);
    } while (timeRunning < 2 * MicroSecPerSec);

    NormalOutput("ranges from %dus to %dus\n", smallest, largest);
    NormalOutput("    average %fus over %d samples\n",
	((double)average) / nSamples, nSamples);
}

void TimeNow(TimeSpecPtr now)
{
    struct timeval osNow;

    (void) TimeOfDay(&osNow, NULL);
    now->seconds = osNow.tv_sec;
    now->microSeconds = osNow.tv_usec;
}

ULong *
CheckRTClock(ULong *newClock)
{
    ULong seconds, microSeconds;
    long changeInSeconds, changeInMicroSecs;
    ULong *oldClock;

    /*
     * We keep track of the last registered clock and the current
     * time of day.
     */
    static ULong *clock;
    static struct timeval tod;	/* time of day */

    seconds      = tod.tv_sec;
    microSeconds = tod.tv_usec;
    (void) TimeOfDay(&tod, NULL);

    if (clock) {
	changeInSeconds   = ((long)tod.tv_sec  - seconds);
	changeInMicroSecs = ((long)tod.tv_usec - microSeconds);

	/*
	 * Here we assume that the change in time is always small, hence, a loop
	 * instead of a multiply.
	 */
	if (changeInMicroSecs < 0) {
	    changeInMicroSecs += MicroSecPerSec;
	    changeInSeconds--;
	}
	while (--changeInSeconds >= 0) {
	    changeInMicroSecs += MicroSecPerSec;
	}
        *clock += changeInMicroSecs;
    }
    oldClock = clock;
    clock = newClock;
    return(oldClock);
}

static int TimerIndex;
static CpuUsage UsageData[ MAX_TIMER_STACK ];
static CpuUsage Subtract[ MAX_TIMER_STACK ];

static void AccumulateTimer(index)
    long index;
{
    CpuUsage *usage = &UsageData[ index ];
    CpuUsage *subtract = &Subtract[ index-1 ];

    subtract->tickElapsed   += usage->tickElapsed;
    subtract->tickUser      += usage->tickUser;
    subtract->tickSystem    += usage->tickSystem;
    subtract->nFaultsNoIO += usage->nFaultsNoIO;
    subtract->nFaultsIO   += usage->nFaultsIO;
    subtract->nSwaps      += usage->nSwaps;
}

static void AdjustTimer(index)
    long index;
{
    CpuUsage *usage = &UsageData[ index ];
    CpuUsage *subtract = &Subtract[ index ];

    usage->tickElapsed   -= subtract->tickElapsed;
    usage->tickUser      -= subtract->tickUser;
    usage->tickSystem    -= subtract->tickSystem;
    usage->nFaultsNoIO -= subtract->nFaultsNoIO;
    usage->nFaultsIO   -= subtract->nFaultsIO;
    usage->nSwaps      -= subtract->nSwaps;
}

static void NoteOtherTimer(index, pOthers)
    long index;
    boolean *pOthers;
{
    CpuUsage *subtract;

    if (! *pOthers) {
	/*
	 * zero out the accumulators for timers started while this one
	 * is running.
	 */
	subtract = &Subtract[ index ];
	subtract->tickElapsed = 0;
	subtract->tickUser = 0;
	subtract->tickSystem = 0;
	subtract->nFaultsNoIO = 0;
	subtract->nFaultsIO = 0;
	subtract->nSwaps = 0;
	*pOthers = True;
    }
}

#if defined(_HPUX_SOURCE)

typedef struct {
    struct tms StartTime;
    struct tms StopTime;
    struct timeval RealStartTime;
    boolean others;
} TimerData *TimerDataPtr ;
static TimerData Timers[ MAX_TIMER_STACK ];

void StartTimer()
{
    long timerIndex = TimerIndex++;
    TimerDataPtr timer = &Timers[ timerIndex ];
    CpuUsage *subtract = &Subtract[ timerIndex ];

    subtract->ticksPerSec = MicroSecPerSec;

    GeneralAssert(TimerIndex <= MAX_TIMER_STACK);

    (void) times(&timer->StartTime);
    (void) TimeOfDay(&timer->RealStartTime, (struct timezone *)NULL);
    timer->others = False;

    /*
     * If there is more than one timer running, make note of it.
     */
    if (--timerIndex >= 0)
	NoteOtherTimer(timerIndex, &Timers[ timerIndex ].others);
}

CpuUsagePtr StopTimer()
{
    double user, sys;
    struct timeval now;
    long timerIndex = --TimerIndex;
    TimerDataPtr timer = &Timers[ timerIndex ];
    CpuUsage *usage = &UsageData[ timerIndex ];

    GeneralAssert(TimerIndex >= 0);

    (void) times(&timer->StopTime);
    (void) TimeOfDay(&now, (struct timezone *)NULL);
    usage->tickUser =
	(timer->StopTime.tms_utime - timer->StartTime.tms_utime)
	* (MicroSecPerSec/HZ);
    usage->tickSystem =
	(timer->StopTime.tms_stime - timer->StartTime.tms_stime)
	* (MicroSecPerSec/HZ);

    usage->nFaultsNoIO = 0;
    usage->nFaultsIO = 0;
    usage->nSwaps = 0;

    if (timer->others)
	AdjustTimer(timerIndex);
    if (timerIndex > 0)
	AccumulateTimer(timerIndex);

    return (usage);
}

#elif defined(Solaris2)

#define NanoSecPerSec 1000000000
#define GetNTimeDifference(result, first, second)			\
    {									\
	long seconds;							\
	if ( (seconds = (second)->tv_sec - (first)->tv_sec) ) {		\
	    result = seconds * NanoSecPerSec;				\
	    if ((second)->tv_nsec < (first)->tv_nsec)			\
		result -= (first)->tv_nsec - (second)->tv_nsec;   	\
	    else							\
		result += (second)->tv_nsec - (first)->tv_nsec;   	\
	} else {							\
	    result = (second)->tv_nsec - (first)->tv_nsec;		\
	}								\
    }

typedef struct {
    prusage_t StartTime;
    prusage_t StopTime;
    struct timeval RealStartTime;
    boolean others;
} TimerData, *TimerDataPtr;
static TimerData Timers[ MAX_TIMER_STACK ];
static int procFd = -1;

void StartTimer()
{
    long timerIndex = TimerIndex++;
    TimerDataPtr timer = &Timers[ timerIndex ];
    CpuUsage *subtract = &Subtract[ timerIndex ];
    char proc[512];

    subtract->ticksPerSec = NanoSecPerSec;

    GeneralAssert(TimerIndex <= MAX_TIMER_STACK);

    if( procFd < 0 ) {
	sprintf(proc,"/proc/%d",(int)getpid());
	procFd = open(proc,O_RDONLY);
    }

    ioctl(procFd, PIOCUSAGE, &timer->StartTime);
    (void) TimeOfDay(&timer->RealStartTime, (struct timezone *)NULL);
    timer->others = False;

    /*
     * If there is more than one timer running, make note of it.
     */
    if (--timerIndex >= 0)
	NoteOtherTimer(timerIndex, &Timers[ timerIndex ].others);
}

CpuUsage *StopTimer()
{
    long timerIndex = --TimerIndex;
    TimerDataPtr timer = &Timers[ timerIndex ];
    CpuUsage *usage = &UsageData[ timerIndex ];
    struct timeval now;

    GeneralAssert(TimerIndex >= 0);

    ioctl(procFd, PIOCUSAGE, &timer->StopTime);
    (void) TimeOfDay(&now, (struct timezone *)NULL);

    GetTimeDifference(usage->tickElapsed, &timer->RealStartTime, &now);
    GetNTimeDifference(usage->tickUser,
	&timer->StartTime.pr_utime, &timer->StopTime.pr_utime);
    GetNTimeDifference(usage->tickSystem,
	&timer->StartTime.pr_stime, &timer->StopTime.pr_stime);

    usage->nFaultsNoIO = timer->StopTime.pr_minf - timer->StartTime.pr_minf;
    usage->nFaultsIO = timer->StopTime.pr_majf - timer->StartTime.pr_majf;
    usage->nSwaps = timer->StopTime.pr_nswap - timer->StartTime.pr_nswap;
    usage->nInvoluntaryContextSwitches =
	timer->StopTime.pr_ictx - timer->StartTime.pr_ictx;
    usage->nVoluntaryContextSwitches =
	timer->StopTime.pr_vctx - timer->StartTime.pr_vctx;

    if (timer->others)
	AdjustTimer(timerIndex);
    if (timerIndex > 0)
	AccumulateTimer(timerIndex);
    return (usage);
}

#else

typedef struct {
    struct rusage StartTime;
    struct rusage StopTime;
    struct timeval RealStartTime;
    boolean others;
} TimerData, *TimerDataPtr;
static TimerData Timers[ MAX_TIMER_STACK ];

void StartTimer()
{
    long timerIndex = TimerIndex++;
    TimerDataPtr timer = &Timers[ timerIndex ];
    CpuUsage *subtract = &Subtract[ timerIndex ];

    subtract->ticksPerSec = MicroSecPerSec;

    GeneralAssert(TimerIndex <= MAX_TIMER_STACK);

    getrusage(RUSAGE_SELF, &timer->StartTime);
    (void) TimeOfDay(&timer->RealStartTime, (struct timezone *)NULL);
    timer->others = False;

    /*
     * If there is more than one timer running, make note of it.
     */
    if (--timerIndex >= 0)
	NoteOtherTimer(timerIndex, &Timers[ timerIndex ].others);
}

CpuUsage *StopTimer()
{
    long user, sys, real;
    long timerIndex = --TimerIndex;
    TimerDataPtr timer = &Timers[ timerIndex ];
    CpuUsage *usage = &UsageData[ timerIndex ];
    struct timeval now;

    GeneralAssert(TimerIndex >= 0);

    getrusage(RUSAGE_SELF, &timer->StopTime);
    (void) TimeOfDay(&now, (struct timezone *)NULL);

    GetTimeDifference(usage->tickElapsed, &timer->RealStartTime, &now);
    GetTimeDifference(usage->tickUser,
	&timer->StartTime.ru_utime, &timer->StopTime.ru_utime);
    GetTimeDifference(usage->tickSystem,
	&timer->StartTime.ru_stime, &timer->StopTime.ru_stime);

    usage->nFaultsNoIO = timer->StopTime.ru_minflt - timer->StartTime.ru_minflt;
    usage->nFaultsIO = timer->StopTime.ru_majflt - timer->StartTime.ru_majflt;
    usage->nSwaps = timer->StopTime.ru_nswap - timer->StartTime.ru_nswap;
    usage->nInvoluntaryContextSwitches =
	timer->StopTime.ru_nivcsw - timer->StartTime.ru_nivcsw;
    usage->nVoluntaryContextSwitches =
	timer->StopTime.ru_nvcsw - timer->StartTime.ru_nvcsw;

    if (timer->others)
	AdjustTimer(timerIndex);
    if (timerIndex > 0)
	AccumulateTimer(timerIndex);
    return (usage);
}
#endif

void StopAllTimers()
{
    TimerIndex = 0;
}
