/*
 * (c) Copyright 1992 North Valley Research, Inc.
 *         All Rights Reserved
 *
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 * 	vxworksOsio.c
 *
 * DESCRIPTION
 * 	I/O functions for vxworks
 *	OS-DEPENDENT
 *
 * REVISION
 *	$Header: /usr/local/src/nvrToMIT/RCS/vwOsio.c,v 1.2 1994/01/24 16:49:47 toddb Exp $
 */
#define NeedFunctionPrototypes 1

#define GeneralAssert(mustBeTrue) _DoAssert(mustBeTrue)

#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>

#include "nvrTypes.h"
#include "nvrTimer.h"
#include "osioI.h"
#include "nvrDebug.h"

extern char *ProgramName;

static struct {
    int errno;
    char *name;
} SysErrorMap[] = {
#ifdef notdef
    { ENOTDIR,	"not a directory" },
    { EACCES,	"permission denied" },
    { EEXIST,	"already exists" },
    { EBUSY,	"device busy" },
    { EROFS,	"read-only file system" },
    { ENOENT,	"no such file or folder" },	
    { ENFILE,	"no file" },
#endif
    { 0,        "0 errno" },
};

char *
LookupSysError(int errno)
{
    int i;
    static char unknown[64];
    
    for (i = 0; i < NumberOfCells(SysErrorMap); i++) {
        if (SysErrorMap[i].errno == errno)
            return(SysErrorMap[i].name);
    }
    sprintf(unknown, "unknown (%d)", errno);
    return(unknown);
}

boolean TtyAskUser(char *fmt, ...)
{
    FILE *fp;
    va_list ap;
    char buf[ BUFSIZ ];

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    fflush(stderr);

    if (read(fileno(stderr), buf, BUFSIZ) <= 0) {
	SystemError("Can't read %d\n", fileno(stderr));
    }
    va_end(ap);
    return(buf[0] == 'y' || buf[0] == 'Y');
}

FileHandle OpenFileForAppending(file)
    char *file;
{
    int fd;

    if ((fd = open(file, O_RDWR)) < 0)
	return (open(file, O_RDWR|O_CREAT, 0666));
    return (fd);
}

FileHandle OpenFileForWriting(file)
    char *file;
{
    extern int errno;
    int fd = open(file, O_WRONLY|O_CREAT, 0666);

    if (fd >= 0) {
	if (ftruncate(fd, 0) < 0)
	    errno = 0;
    }
    return(fd);
}

FileHandle OpenFileForReading(file)
    char *file;
{
    FILE *fp;
    fp = fopen_local_client_file(file, "r");
    if (fp)
        return (FileHandle) fileno(fp);
    else
        return -1;
}

/*
 * Arrange for asynchronous notification of io
 * XXX if this ioctl fails, it may be because gcc isn't getting
 * the -traditional-cpp flag.  Without it, the first argument of
 * the _IOR macro is interpreted wrong.
 */
void AllowAsyncNotification(fd)
    int fd;
{
    int true = True;

    GeneralAssert (IoData.list[fd].handler != NULL);

    IoData.list[fd].async = True;
    NotImplemented("set async io");
}
    
void BlockAsyncNotification(fd)
    int fd;
{
    int false = False;

    GeneralAssert (IoData.list[fd].handler != NULL);

    IoData.list[fd].async = False;
    NotImplemented("clear async io");
}
    
int BlockAsyncHandler()
{
    return ( sigblock(sigmask(SIGURG) | sigmask(SIGIO)) );
}

int AllowAsyncHandler()
{
    sigsetmask(sigblock(0)
	     & (~( sigblock(sigmask(SIGURG) | sigmask(SIGIO)) ))
    );
}

static void AsyncHandler(int sig)
{
    fd_set inReady;
    int i, nFdsReady;
    extern int errno;

#ifdef SYSV
    if (signal(sig, AsyncHandler) == (VoidFuncPtr)-1)
	Fatal("can't reset signal handler\n");
#endif /* SYSV */

    inReady = IoData.inReady;
    nFdsReady = select(IoData.lastFd+1,
		       (SelectDataPtr)&inReady,
		       (SelectDataPtr)NULL,
		       (SelectDataPtr)NULL,
		       &IoData.noWait);

    if (nFdsReady < 0) {
	if (errno != EINTR)
	    Fatal("select() fails\n");
    }

    for (i = IoData.lastFd; i >= 0; i--)
    {
	if (! IoData.list[i].async)
	    continue;
	if (FD_ISSET(i, &inReady)
	 || (IoData.list[i].bypassSelect))
	{
	    (*(IoData.list[i].handler))(i, IoData.list[i].pointer);
	}
    }
}

boolean
WaitForFileInput(int fd, TimeSpecPtr maxWait)
{
    fd_set inReady;
    int i, nFdsReady;
    struct timeval osMaxWait, *timeout;
    extern int errno;

    if (maxWait) {
	osMaxWait.tv_sec = maxWait->seconds;
	osMaxWait.tv_usec = maxWait->microSeconds;
	timeout = &osMaxWait;
    } else
	timeout = NULL;

    if (fd < 32) {
	inReady.fds_bits[0] = 1<<fd;
	nFdsReady = select(fd+1,
			   (SelectDataPtr)&inReady,
			   (SelectDataPtr)NULL,
			   (SelectDataPtr)NULL,
			   timeout);
    } else {
	FD_ZERO(&inReady);
	FD_SET(fd, &inReady);
	nFdsReady = select(fd+1,
			   (SelectDataPtr)&inReady,
			   (SelectDataPtr)NULL,
			   (SelectDataPtr)NULL,
			   timeout);
    }

    if (nFdsReady < 0 && errno != EINTR)
	Fatal("select(%d,%x,{%d,%d}) fails\n",
	    IoData.lastFd+1,
	    &IoData.inReady,
	    timeout->tv_sec, timeout->tv_usec);

    return (nFdsReady > 0);
}

void WaitForSomething(TimeSpecPtr maxWait)
{
    fd_set inReady;
    int i, nFdsReady;
    struct timeval osMaxWait, *timeout;
    extern int errno;

    if (maxWait) {
	osMaxWait.tv_sec = maxWait->seconds;
	osMaxWait.tv_usec = maxWait->microSeconds;
	timeout = &osMaxWait;
    } else
	timeout = NULL;

    inReady = IoData.inReady;
    nFdsReady = select(IoData.lastFd+1,
		       (SelectDataPtr)&inReady,
		       (SelectDataPtr)NULL,
		       (SelectDataPtr)NULL,
		       timeout);

    if (nFdsReady < 0 && errno != EINTR)
	Fatal("select(%d,%x,{%d,%d}) fails\n",
	    IoData.lastFd+1,
	    &IoData.inReady,
	    timeout->tv_sec, timeout->tv_usec);

    if (nFdsReady > 0) {
	for (i = IoData.lastFd; i >= 0; i--)
	{
	    if (FD_ISSET(i, &inReady))
	    {
		(*(IoData.list[i].handler))(i, IoData.list[i].pointer);
	    }
	}
    }
}

boolean FileHandleIsTty(int fd)
{
    GeneralAssert(fd < FD_SETSIZE);

    if (IoData.list[fd].handler)
	return (IoData.list[fd].isAtty);
    else
        return (isatty(fd));
}

/*
 * Set the connection non-blocking... but not if its a tty
 */
void SetNonBlocking(int fd)
{
    GeneralAssert (IoData.list[fd].handler != NULL);

    if (! IoData.list[fd].isAtty) {
	int true = True;

	if (ioctl(fd, FIONBIO, &true) < 0) {
	    SystemError("can't set non-blocking io on client fd %d\n", fd);
	    return;
	}
    }
}

void AllocAsyncHandler(int fd,
		       VoidFuncPtr handler,
		       boolean bypassSelect,
		       void *pointer)
{
    AllocHandler(fd, handler, bypassSelect, pointer);

    /*
     * Set the connection non-blocking... but not if its a tty
     */

    SetNonBlocking(fd);

    NotImplemented("set async pid");
    if (! IoData.signalHandlerSet) {
	if (signal(SIGIO, AsyncHandler) == (VoidFuncPtr)-1
	 || signal(SIGURG, AsyncHandler) == (VoidFuncPtr)-1)
	    Fatal("can't set signal handler\n");
	IoData.signalHandlerSet = True;
    }

    AllowAsyncNotification(fd);
}

long
BytesAvailable(fd)
{
    long bytesIn;

    if (ioctl(fd, FIONREAD, &bytesIn) < 0) {
	SystemError("can't FIONREAD fd %d\n", fd);
	bytesIn = 0;
    }
    return (bytesIn);
}

int NonBlockingRead(fileHandle, pBuf, spaceAvailable)
    int fileHandle;
    char *pBuf;
    long spaceAvailable;
{
    long bytesIn;
    extern int errno;

    GeneralAssert(IoData.lastFd >= fileHandle);

    /* XXX special case for stdin */
    if (IoData.list[ fileHandle ].isAtty) {
	bytesIn = BytesAvailable(fileHandle);
        /*
         * You should use the command
         *   dd bs=1 | zip ...
         * to do debugging
         */
	if (bytesIn == 0)
	    return (0);
    }
    bytesIn = read(fileHandle, pBuf, spaceAvailable);
    if (bytesIn < 0) {
        if (errno == EAGAIN)		/* SYS5 non-blocking */
            return (0);
        if (errno == EWOULDBLOCK)	/* 4.2bsd non-blocking */
            return (0);
        SystemError("can't read from fd %d\n",  fileHandle);
	return (-1);
    }
    if (bytesIn == 0) {
	return (-1);
    }

    return (bytesIn);
}

#include "fileAccess.h"

#ifdef _HPUX_SOURCE
#define getdtablesize()	FD_SETSIZE
#endif

void
ShowOpenFileDescriptors()
{
    char buf[ 100 ];
    int possibleFds = getdtablesize();
    int i, type;
    extern int errno;
    boolean readable, writeable;

    for (i = 0; i < possibleFds; i++) {
        type = FdFileType(i);
        if (! type)
            continue;
        writeable = (write(i, buf, 0) == 0);
        readable = (read(i, buf, 0) == 0);
        ProgMsg("fd %d: %s, mode %c%c\n", i,
              TypeName(type), " r"[readable], " w"[writeable]);
    }
}

/*
 * close all open file descriptors except those specified
 * XXX note that we can only specify file descriptors <= 31.
 */
void
CloseOpenFileDescriptors(ULong retained)
{
    int possibleFds = getdtablesize();
    int i, type;
    extern int errno;

    for (i = 0; i < possibleFds; i++) {
	if (retained & (1<<i))
	    continue;
	close(i);
    }
    errno = 0;	/* clear side-effects of close() */
}

void
NotImplemented(char *funcName)
{
    ProgMsg("%s not implemented\n", funcName);
}

int
ftruncate()
{
    NotImplemented("ftruncate");
    return 0;
}

int
getpwnam()
{
    NotImplemented("getpwnam");
    return 0;
}

int
regcmp()
{
    NotImplemented("regcmp");
    return 0;
}

int
regex()
{
    NotImplemented("regex");
    return 0;
}

void
WaitForSignal(void)
{
    pause();
}
