/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	unixTimer.c - unix-specific timing functions
 *
 * DESCRIPTION
 *	OS-DEPENDENT
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/vwTimer.c,v 1.1 1994/01/17 15:57:57 toddb Exp $
 */
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/resource.h>
#include <unistd.h>
#include "wdLib.h"

#define NeedFunctionPrototypes 1
#define TIMING_ENABLED	/* the library always defines it */

#include "nvrTypes.h"
#include "longjump.h"
#include "nvrDebug.h"
#include "nvrString.h"
#include "osio.h"
#include "nvrTimer.h"

#define MAX_TIMER_STACK	5

#ifdef TEK_DEBUG
#define TekDebug(args)	ProgMsg args
#endif

WDOG_ID wdogId;
static VoidFuncPtr SavedTimerFunc;

#define GeneralAssert(mustBeTrue)       _DoAssert(mustBeTrue)

#define GetTimeDifference(result, first, second)			\
    {									\
	long seconds;							\
	if (seconds = (second)->tv_sec - (first)->tv_sec) {		\
	    result = seconds * MicroSecPerSec;				\
	    if ((second)->tv_usec < (first)->tv_usec)			\
		result -= (first)->tv_usec - (second)->tv_usec;		\
	    else							\
		result += (second)->tv_usec - (first)->tv_usec;		\
	} else {							\
	    result = (second)->tv_usec - (first)->tv_usec;		\
	}								\
    }

void ClearFrameTimer()
{
    ProgMsg("ClearFrameTimer called\n");
    wdCancel(wdogId);
}

#define SaveTimerFunc(timerFunc)	timerFunc

void SetTimerFunc(IntFuncPtr timerFunc)
{
    SavedTimerFunc = timerFunc;
}

void SetFrameTimer(long usecPerPic, IntFuncPtr timerFunc)
{
    long ticksPerPic;

    TekDebug(("SetFrameTimer called \n"));
    ticksPerPic = (usecPerPic * sysClkRateGet()) / MicroSecPerSec;
    if (wdogId == 0)
        wdogId = wdCreate();
    SetTimerFunc(timerFunc);
    SetFrameHandler();
    setAsyncHandler();
    wdStart(wdogId, ticksPerPic, timerFunc, ticksPerPic);
}

ResetSetFrameTimer(ticksPerPic)
int ticksPerPic;
{
    wdStart(wdogId, ticksPerPic, SavedTimerFunc, ticksPerPic);
}

void AllowTimer(int key)
{
    TekDebug(("AllowTimer called\n"));
    sigsetmask(key);
}

int BlockTimer()
{
    TekDebug(("BlockTimer called\n"));
    return (sigblock(sigmask(SIGALRM)));
}

void ClockResolution()
{
    long smallest = 0x7fffffff;		/* not very small */
    long largest = 0;			/* not very big */
    long average = 0;			/* ... to start with */
    long diff, timeRunning, nSamples;
    struct timeval start, last, now;

    NormalOutput("Clock resolution: ");

    if (gettimeofday(&start, (struct timezone *)NULL) < 0)
	Fatal("can't get time of day\n");

    now = start;
    nSamples = 0;
    do {
	last = now;
	if (gettimeofday(&now, (struct timezone *)NULL) < 0)
	    Fatal("can't get time of day\n");
	GetTimeDifference(diff, &last, &now);
	if (diff == 0)
	    continue;

	nSamples++;
	average += diff;
	if (diff > largest)
	    largest = diff;
	if (diff < smallest)
	    smallest = diff;

	GetTimeDifference(timeRunning, &start, &now);
    } while (timeRunning < 2 * MicroSecPerSec);

    NormalOutput("ranges from %dus to %dus\n", smallest, largest);
    NormalOutput("    average %fus over %d samples\n",
	((double)average) / nSamples, nSamples);
}

void TimeNow(TimeSpecPtr now)
{
    struct timeval osNow;

    (void) gettimeofday(&osNow, NULL);
    now->seconds = osNow.tv_sec;
    now->microSeconds = osNow.tv_usec;
}

static int TimerIndex;
static CpuUsage UsageData[ MAX_TIMER_STACK ];
static CpuUsage Subtract[ MAX_TIMER_STACK ];

static void AccumulateTimer(index)
    long index;
{
    CpuUsage *usage = &UsageData[ index ];
    CpuUsage *subtract = &Subtract[ index-1 ];

    subtract->msElapsed   += usage->msElapsed;
    subtract->msUser      += usage->msUser;
    subtract->msSystem    += usage->msSystem;
    subtract->nFaultsNoIO += usage->nFaultsNoIO;
    subtract->nFaultsIO   += usage->nFaultsIO;
    subtract->nSwaps      += usage->nSwaps;
}

static void AdjustTimer(index)
    long index;
{
    CpuUsage *usage = &UsageData[ index ];
    CpuUsage *subtract = &Subtract[ index ];

    usage->msElapsed   -= subtract->msElapsed;
    usage->msUser      -= subtract->msUser;
    usage->msSystem    -= subtract->msSystem;
    usage->nFaultsNoIO -= subtract->nFaultsNoIO;
    usage->nFaultsIO   -= subtract->nFaultsIO;
    usage->nSwaps      -= subtract->nSwaps;
}

static void NoteOtherTimer(index, pOthers)
    long index;
    boolean *pOthers;
{
    CpuUsage *subtract;

    if (! *pOthers) {
	/*
	 * zero out the accumulators for timers started while this one
	 * is running.
	 */
	subtract = &Subtract[ index ];
	subtract->msElapsed = 0;
	subtract->msUser = 0;
	subtract->msSystem = 0;
	subtract->nFaultsNoIO = 0;
	subtract->nFaultsIO = 0;
	subtract->nSwaps = 0;
	*pOthers = True;
    }
}

typedef struct _timerData {
    struct timeval RealStartTime;
    boolean others;
} TimerData;

static TimerData Timers[ MAX_TIMER_STACK ];


void StartTimer()
{
    long timerIndex = TimerIndex++;
    TimerData *timer = &Timers[ timerIndex ];
    CpuUsage *subtract = &Subtract[ timerIndex ];

    GeneralAssert(TimerIndex <= MAX_TIMER_STACK);

    (void) gettimeofday(&timer->RealStartTime, (struct timezone *)NULL);
    timer->others = False;

    /*
     * If there is more than one timer running, make note of it.
     */
    if (--timerIndex >= 0)
	NoteOtherTimer(timerIndex, &Timers[ timerIndex ].others);
}

CpuUsage *StopTimer()
{
    long user, sys, real;
    long timerIndex = --TimerIndex;
    TimerData *timer = &Timers[ timerIndex ];
    CpuUsage *usage = &UsageData[ timerIndex ];
    struct timeval now;

    GeneralAssert(TimerIndex >= 0);

    (void) gettimeofday(&now, (struct timezone *)NULL);

    GetTimeDifference(usage->msElapsed, &timer->RealStartTime, &now);

    if (timer->others)
	AdjustTimer(timerIndex);
    if (timerIndex > 0)
	AccumulateTimer(timerIndex);
    return (usage);
}

void StopAllTimers()
{
    TimerIndex = 0;
}
