/*
 * Copyright (c) 1992,1993,1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *      floatLeeIdct.c - Lee's floating point inverse dct.
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/floatLeeIdct.c,v 1.26 1995/10/15 23:55:00 toddb Exp $
 */

#include "floatLeeIdctI.h"
#include "nvrAlloc.h"
#include "nvrString.h"

/*
 * These macros are for use with ProcessBlocks.
 */
#define RowOfIndex(_index)	((_index) >> 3)
#define ColOfIndex(_index)	((_index) & 0x7)
#define OneCoefMask(_coef)	(1 << (_coef))

#define SetColMask(_gridColMask, _index)				\
    (_gridColMask[ ColOfIndex(_index) ] |= (1 << RowOfIndex(_index)))

#define SetRowMask(_rowMask, _index)					\
    (_rowMask |= (1 << ColOfIndex(_index)))

#ifdef DEBUG_ENABLED
static void
ShowCoeff(int blockIndex, boolean intra, float *tile, UChar gridColMask[64/8])
{
    int row, col;

    for (row = 0; row < 8; row++) {
	if (row == 0)
	    Msg("%s coef, blk %d: ", intra ? "intra" : "delta", blockIndex);
	else
	    Msg("                   ");
	for (col = 0; col < 8; col++) {
	    if (gridColMask[ col ] & OneCoefMask(row))
		Msg("%4d", (int)(tile[row*8 + col] * 4.0) );
	    else
		Msg("   0");
	}
	Msg("\n");
    }
}

static void
ShowIntraValue(int blockIndex, UChar *pixel, int lineLength)
{
    int row, col;

    for (row = 0; row < 8; row++) {
	if (row == 0)
	    Msg("intra block %d:     ", blockIndex);
	else
	    Msg("                   ");
	for (col = 0; col < 8; col++)
	    Msg("%4d", pixel[ col ]);
	pixel += lineLength;
	Msg("\n");
    }
}

/************************************************************************
 *
 * Standard intra coefficient reconstruction.
 *
 ************************************************************************/
static void
ValidateIntraReconstrution(LeeIdctPrivPtr lee, int level, int minus, int index)
{
    int recon;

    if (GetDebugBit(DCTD_ValidateReconstruction)) {
	recon = (2 * level * lee->quant * lee->super.intraQ[ index ]) / 16;
	recon -= ((~recon) & 1);	/* prevent odd coefficients */
	if (minus)
	    recon = -recon;
	if (recon > 2047) recon = 2047;
	if (recon < -2048) recon = -2048;
	if (recon * lee->coeffScale != lee->initTile[ index ])
	    Msg("bad reconstruction for intra coef #%d, std: %d vs. %f\n",
		index, recon, lee->initTile[ index ]/lee->coeffScale);
    }
}

/************************************************************************
 *
 * Standard delta coefficient reconstruction.
 *
 ************************************************************************/
static void
ValidateDeltaReconstrution(LeeIdctPrivPtr lee, int level, int minus, int index)
{
    int recon;

    if (GetDebugBit(DCTD_ValidateReconstruction)) {
	recon = ((2 * level + 1)
	      * lee->quant * lee->super.nonIntraQ[ index ]) / 16;
	recon -= ((~recon) & 1);	/* prevent odd coefficients */
	if (minus)
	    recon = -recon;
	if (recon > 2047) recon = 2047;
	if (recon < -2048) recon = -2048;
	if (recon * lee->coeffScale != lee->initTile[ index ])
	    Msg("bad reconstruction for delta coef #%d, std: %d vs. %f\n",
		index, recon, lee->initTile[ index ]/lee->coeffScale);
    }
}

/************************************************************************
 *
 * Standard delta coefficient reconstruction.
 *
 ************************************************************************/
static void
ValidateDcDeltaReconstrution(LeeIdctPrivPtr lee, int level,
			     int minus, float coef)
{
    int recon;

    if (GetDebugBit(DCTD_ValidateReconstruction)) {
	recon = ((2 * level + 1) * lee->quant * lee->super.nonIntraQ[ 0 ]) / 16;
	recon -= ((~recon) & 1);	/* prevent odd coefficients */
	if (minus)
	    recon = -recon;
	if (recon > 2047) recon = 2047;
	if (recon < -2048) recon = -2048;
	if (recon != coef)
	    Msg("bad reconstruction for delta dc coef, std: %d vs. %f\n",
		recon, coef);
    }
}

static void
ShowPreDeltaValue(int blockIndex, int row, UChar *pMotion,
                  float p0, float p1, float p2, float p3,
		  float p4, float p5, float p6, float p7)
{
    int col;

    if (row == 0)
	Msg("delta block %d:     ", blockIndex);
    else
	Msg("                   ");
    for (col = 0; col < 8; col++)
	Msg("%4d", pMotion[ col ]);

    Msg("  %c  ", row == 0 ? '+' : ' ');
    Msg("%4d%4d%4d%4d%4d%4d%4d%4d\n",
	(int)p0, (int)p1, (int)p2, (int)p3,
	(int)p4, (int)p5, (int)p6, (int)p7);
}

static void
ShowPostDeltaValue(UChar *pPixel, int lineLength)
{
    int row, col;

    for (row = 0; row < 8; row++) {
	Msg("                 %c ", row == 0 ? '=' : ' ');
	for (col = 0; col < 8; col++)
	    Msg("%4d", pPixel[ col ]);
	pPixel += lineLength;
	Msg("\n");
    }
    Msg("\n");
}

static void
RefIdct(LeeIdctPrivPtr lee,
	int blkIndex,
	float *pTile,
	int index,
	boolean column)
{
    register float f0,f1,f2,f3,f4,f5,f6,f7,ft;
    register float mc14, mc18, mc38, mc116, mc316, mc516, mc716;
    float *input, *output;
    int shift, i;

    if (column) {
	input = lee->vDctInput + index;
	output = lee->vDctOutput + index;
	shift = 3;
	for (i = 0; i < 8; i++)
	    input[ i << shift ] = pTile[ i << shift ];
    } else {
	boolean broken = False;
	float scale = 0.25;

	shift = 0;
	input = lee->vDctOutput + (index << 3);
	for (i = 0; i < 8; i++)
	    broken |= (input[i] != pTile[i]);

	if (broken) {
	    Msg("dct for pic %d, mb %d, blk %d, input row %d in error:\n",
		lee->super.current->frame, lee->super.super.nextMbAddr,
		blkIndex, index);

	    Msg("err input:%8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
		2 * scale * input[0], 2 * scale * input[1],
		2 * scale * input[2], 2 * scale * input[3],
		2 * scale * input[4], 2 * scale * input[5],
		2 * scale * input[6], 2 * scale * input[7]);
	    Msg("input:    %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
		scale * pTile[0], scale * pTile[1],
		scale * pTile[2], scale * pTile[3],
		scale * pTile[4], scale * pTile[5],
		scale * pTile[6], scale * pTile[7]);
	}

	input = lee->vDctInput + (index << 3);
	output = lee->vDctOutput + (index << 3);
	for (i = 0; i < 8; i++)
	    input[ i ] = pTile[ i ];
    }

    mc14  = lee->dctConst[0];      /* Load IDCT constants into registers */
    mc18  = lee->dctConst[1];
    mc38  = lee->dctConst[2];
    mc116 = lee->dctConst[3];
    mc316 = lee->dctConst[4];
    mc516 = lee->dctConst[5];
    mc716 = lee->dctConst[6];

    f7 = pTile[ 7<<shift ];
    f5 = pTile[ 5<<shift ];
    f1 = pTile[ 1<<shift ];
    f3 = pTile[ 3<<shift ];
    f7 += f5;
    f2 = pTile[ 2<<shift ];
    f5 += f3;
    f6 = pTile[ 6<<shift ];
    f5 *= mc14;
    f3 += f1;
    f0 = pTile[ 0<<shift ];
    f7 += f3;
    f4 = pTile[ 4<<shift ];
    f7 *= mc14;
    f6 += f2;
    f4 *= mc14;
    ft = f3 + f7;
    f6 *= mc14;
    f7 = f3 - f7;
    f3 = ft * mc18;
    ft = f2 + f6;
    f7 *= mc38;
    f6 = f2 - f6;
    f2 = ft * mc18;
    ft = f1 + f5;
    f6 *= mc38;
    f5 = f1 - f5;
    f0 *= mc14;
    f1 = ft + f3;
    f3 = ft - f3;
    f1 *= mc116;
    ft = f5 + f7;
    f3 *= mc716;
    f7 = f5 - f7;
    f5 = ft * mc316;
    ft = f0 + f4;
    f7 *= mc516;
    f4 = f0 - f4;
    f0 = ft + f2;
    f2 = ft - f2;
    ft = f4 + f6;
    f6 = f4 - f6;

    output[ 0<<shift ] = f0 + f1;
    output[ 7<<shift ] = f0 - f1;
    output[ 1<<shift ] = ft + f5;
    output[ 6<<shift ] = ft - f5;
    output[ 3<<shift ] = f2 + f3;
    output[ 4<<shift ] = f2 - f3;
    output[ 2<<shift ] = f6 + f7;
    output[ 5<<shift ] = f6 - f7;
}

static void
CheckIdct(LeeIdctPrivPtr lee,
	  float *pTile,
	  int blkIndex,
	  int index,
	  boolean row)
{
    int shift;
    float scale;
    float *output, *input;

    if (row) {
	shift = 0;
	scale = 0.25;
	output = lee->vDctOutput + (index << 3);
	input = lee->vDctInput + (index << 3);
    } else {
	shift = 3;
	scale = 0.5;
	output = lee->vDctOutput + index;
	input = lee->vDctInput + index;
    }

    if (output[0<<shift] == pTile[0<<shift]
     && output[1<<shift] == pTile[1<<shift]
     && output[2<<shift] == pTile[2<<shift]
     && output[3<<shift] == pTile[3<<shift]
     && output[4<<shift] == pTile[4<<shift]
     && output[5<<shift] == pTile[5<<shift]
     && output[6<<shift] == pTile[6<<shift]
     && output[7<<shift] == pTile[7<<shift])
	return;

    Msg("dct for pic %d, mb %d, blk %d, %s %d in error:\n",
	lee->super.current->frame,
	lee->super.super.nextMbAddr,
	blkIndex,
	row ? "row" : "col", index);

    Msg("input:   %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
	2 * scale * input[0<<shift], 2 * scale * input[1<<shift],
	2 * scale * input[2<<shift], 2 * scale * input[3<<shift],
	2 * scale * input[4<<shift], 2 * scale * input[5<<shift],
	2 * scale * input[6<<shift], 2 * scale * input[7<<shift]);
    Msg("ref out: %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
	scale * output[0<<shift], scale * output[1<<shift],
	scale * output[2<<shift], scale * output[3<<shift],
	scale * output[4<<shift], scale * output[5<<shift],
	scale * output[6<<shift], scale * output[7<<shift]);
    Msg("out:     %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
	scale * pTile[0<<shift], scale * pTile[1<<shift],
	scale * pTile[2<<shift], scale * pTile[3<<shift],
	scale * pTile[4<<shift], scale * pTile[5<<shift],
	scale * pTile[6<<shift], scale * pTile[7<<shift]);
}

#if 0
static void
CheckFloatToFixed(float *row)
{
    int i, ival, exp, frac, tmp;
    union {
	struct {
	    ULong sign  : 1;
	    ULong exp   : 8;
	    ULong frac  :23;
	} b;
	ULong composite;
    } fbits;

    for (i = 0; i < 8; i++) {
	ival = row[i];
	fbits.composite = ULongP(row)[i];
	if (fbits.b.exp == 255) {
	    if (fbits.b.frac)
		Msg("NaN\n");
	    else
		Msg("inf\n");
	} else if (fbits.b.exp != 0) {
	    exp = fbits.b.exp - 127;
	    frac = fbits.b.frac;
	    frac |= 0x800000;
	    tmp = frac >> 23-exp;
	    if (fbits.b.sign)
		tmp = -tmp;
	    if (tmp != ival)
		Msg("Bad conversion: %d != %d\n", tmp, ival);
	} else { /* fbits.b.exp == 0 */
	    if (fbits.b.frac)
		Msg("denormalized\n");
	    else {
		if (ival != 0)
		    Msg("Bad conversion: 0 != %d\n", ival);
	    }
	}
    }
}
#endif
	
static void
CheckRowIdct(LeeIdctPrivPtr lee, int blkIndex, int rowCol,
	     float f0, float f1, float f2, float f3,
	     float f4, float f5, float f6, float f7)
{
    float tmp[ 8 ];

    tmp[0] = f0;
    tmp[1] = f1;
    tmp[2] = f2;
    tmp[3] = f3;
    tmp[4] = f4;
    tmp[5] = f5;
    tmp[6] = f6;
    tmp[7] = f7;

    CheckIdct(lee, tmp, blkIndex, rowCol, True);
}

#endif /* DEBUG_ENABLED */

static DctMacroblockPtr
ProcessBlocks(LeeIdctPrivPtr lee, DctMacroblockPtr idctMb)
{
    int i, col, row, index, nCoeff, block, blkIndex, tmp;
    int picWidth, motionPicWidth;
    DctBlockPtr dctBlk;
    UChar *pPixel, *pMotion, picType;
    UChar gridColMask[ 64 / 8 ], colMask;
    ULong rowMask, pixels;
    UShort *pCoeff, coeff;
    QuantCachePtr qLookup;
    float *pTile, *initTile;
    register float f0,f1,f2,f3,f4,f5,f6,f7,ft;

    WhenDebugEnabled(
	lee->quant = idctMb->flags.b.quant;
	if (GetDebugBit(DCTD_IntraValues)
	 || GetDebugBit(DCTD_IntraCoeff)
	 || GetDebugBit(DCTD_DeltaCoeff)
	 || GetDebugBit(DCTD_DeltaValues)
	 || GetDebugBit(DCTD_MotionMbAddr)
	 || GetDebugBit(DCTD_MotionBlkAddr))
	    Msg("pic %d, mb %d:\n",
		lee->super.current->frame, lee->super.super.nextMbAddr);
    )

    if (lee->intra) {
	lee->iqmCur = &lee->iqmCache[ idctMb->flags.b.quant-1 ];
    } else {
        (*lee->super.motionComp)(&lee->super, idctMb);
	if (idctMb->flags.b.blocksMask == 0)
	    return( DctBlockPtrToNextMb(&idctMb->blk[0]) );
	lee->niqmCur = &lee->niqmCache[ idctMb->flags.b.quant-1 ];
    }
    
    dctBlk = &idctMb->blk[0];
    for (block = 0; block < lee->super.geom.nBlocks; block++) {
	if (((1<<block) & idctMb->flags.fastBlocksMask) == 0)
	    continue;

        /*
         * First time through: construct a bitmask of coefficients that
         * are present and a lookup table to get them.  Note that the
         * bitmask is the only thing that needs to be fully initialized;
         * the lookup table needs a valid entry only where it corresponds
         * to a 1 bit in the mask.
         */
        blkIndex = dctBlk->hdr.s.index;
        picType = lee->super.geom.picType[ blkIndex ];
        pPixel = lee->super.pTo[ picType ]
               + lee->super.geom.blkOffset[ blkIndex ];
        picWidth = lee->super.geom.pic[ picType ].width;
        nCoeff = dctBlk->hdr.s.nCoeff;

	/*
	 * short-circuit DC blocks as early as possible.  Note that
	 * we assume that nCoeff is only zero in an intra block.
	 */
	if (nCoeff == 0) {
	    tmp = CalcDCPixel2d(lee, dctBlk->hdr.s.dcCoeff);
	    /*
	     * fill the pipeline with other stuff to wait for the
	     * multiply to complete
	     */
	    dctBlk = DctCoeffPtrToNextBlock(&dctBlk->coefList[0]);

	    DebugTask(DCTD_IntraCoeff,
		lee->initTile[0] = lee->coeffScale * tmp;
		gridColMask[0] = 1;
		ShowCoeff(blkIndex, True, lee->initTile, gridColMask);
		lee->initTile[0] = (float)0.0;
	    )

	    PixelClamp(lee, pixels, tmp);
	    pixels |= pixels << 8;
	    pixels |= pixels << 16;
	    ULongP(pPixel)[0] = ULongP(pPixel)[1] = pixels;
	    pPixel += picWidth;
	    ULongP(pPixel)[0] = ULongP(pPixel)[1] = pixels;
	    pPixel += picWidth;
	    ULongP(pPixel)[0] = ULongP(pPixel)[1] = pixels;
	    pPixel += picWidth;
	    ULongP(pPixel)[0] = ULongP(pPixel)[1] = pixels;
	    pPixel += picWidth;
	    ULongP(pPixel)[0] = ULongP(pPixel)[1] = pixels;
	    pPixel += picWidth;
	    ULongP(pPixel)[0] = ULongP(pPixel)[1] = pixels;
	    pPixel += picWidth;
	    ULongP(pPixel)[0] = ULongP(pPixel)[1] = pixels;
	    pPixel += picWidth;
	    ULongP(pPixel)[0] = ULongP(pPixel)[1] = pixels;

	    DebugTask(DCTD_IntraValues,
		pPixel = lee->super.pTo[ picType ]
		       + lee->super.geom.blkOffset[ blkIndex ];
		ShowIntraValue(blkIndex, pPixel, picWidth);
	    )

	    continue;
	}

        pCoeff = UShortP(&dctBlk->coefList[0]);

	/*
	 * short-circuit delta DC blocks as early as possible
	 * Note that we make the assumption that if there is one
	 * coefficient and its index is 0, this has to be a delta
	 * block, since intra coded dc blocks are handled differently.
	 */
	if (nCoeff == 1) {
	    coeff = pCoeff[ 0 ];
	    index = GetDctCoeffIndex(coeff);
	    if (index == 0) {
		qLookup = lee->niqmCur;
		ReconstructDCDeltaCoeff(lee, tmp, qLookup,
		    GetDctCoeffLevel(coeff), GetDctCoeffSign(coeff));
		DebugTask(DCTD_DeltaCoeff,
		    lee->initTile[0] = lee->coeffScale * tmp;
		    gridColMask[0] = 1;
		    ShowCoeff(blkIndex, False, lee->initTile, gridColMask);
		    lee->initTile[0] = (float)0.0;
		)
		tmp = CalcDCPixel2d(lee, tmp);

		/*
		 * fill the pipeline with other stuff to wait for the
		 * multiply to complete
		 */
		dctBlk = DctCoeffPtrToNextBlock(&dctBlk->coefList[0]);
		motionPicWidth = lee->super.pmc->picWidth[ blkIndex ];
		pMotion = lee->super.pmc->pSrc[ picType ]
			+ lee->super.pmc->offset[ blkIndex ];
		for (row = 0; row < 8; row++) {
		    DebugTask(DCTD_DeltaValues,
			ShowPreDeltaValue(blkIndex, row, pMotion,
			    (float)tmp, (float)tmp, (float)tmp, (float)tmp,
			    (float)tmp, (float)tmp, (float)tmp, (float)tmp);
		    )
		    DCDeltaPixelClamp(lee, tmp, pMotion, pPixel);
		    pPixel += picWidth;
		    pMotion += motionPicWidth;
		}

		DebugTask(DCTD_DeltaValues,
		    pPixel = lee->super.pTo[ picType ]
			   + lee->super.geom.blkOffset[ blkIndex ];
		    ShowPostDeltaValue(pPixel, picWidth);
		)
		continue;
	    }
	}

        ULongP(gridColMask)[ 0 ] = ULongP(gridColMask)[ 1 ] = 0;
        if (lee->intra) {
            /*
             * As elsewhere in the algorithm, a factor of 0.5 has been
             * factored out to save a couple of multiplies.  The dc coeficient
             * is a little special here since it can be odd or even
             */
            f0 = dctBlk->hdr.s.dcCoeff >> 2;
            if (0 != (tmp = dctBlk->hdr.s.dcCoeff & 3))
                f0 += tmp * lee->coeffScale;
            lee->initTile[0] = f0;
            
            qLookup = lee->iqmCur;
            gridColMask[ 0 ] = 1;
	    rowMask = 1;
            for (i = 0; i < nCoeff; i++) {
                coeff = pCoeff[ i ];
                index = GetDctCoeffIndex(coeff);
                SetColMask(gridColMask, index);
                SetRowMask(rowMask, index);
                ReconstructIntraCoeff(lee, index, qLookup,
                                      GetDctCoeffLevel(coeff),
                                      GetDctCoeffSign(coeff))
            }
	    DebugTask(DCTD_IntraCoeff,
		ShowCoeff(blkIndex, True, lee->initTile, gridColMask);
	    )
        } else { /* non-intra */
            qLookup = lee->niqmCur;
	    rowMask = 0;
            for (i = 0; i < nCoeff; i++) {
                coeff = pCoeff[ i ];
                index = GetDctCoeffIndex(coeff);
                SetColMask(gridColMask, index);
                SetRowMask(rowMask, index);
                ReconstructDeltaCoeff(lee, index, qLookup,
				      GetDctCoeffLevel(coeff),
				      GetDctCoeffSign(coeff))
            }
	    pMotion = lee->super.pmc->pSrc[ picType ]
		    + lee->super.pmc->offset[ blkIndex ];
	    motionPicWidth = lee->super.pmc->picWidth[ blkIndex ];
	    DebugTask(DCTD_DeltaCoeff,
		ShowCoeff(blkIndex, False, lee->initTile, gridColMask);
	    )
        }

#ifdef neverDoThisAtHome
	/*
	 * XXX for debugging!!! see what we gain by doing full
	 * DCT matrices elsewhere.
	 */
	rowMask &= 0x1;
        ULongP(gridColMask)[ 0 ] &= 0x01000000;
	ULongP(gridColMask)[ 1 ] = 0;
#endif

        /*
         * Second time through: do the inverse DCT on the columns
         */
        pTile = lee->tile;
        initTile = lee->initTile;

        for (col = 0; col < 8;) {
	    DebugTask(DCTD_ValidateDct,
		RefIdct(lee, blkIndex, initTile, col, True) )

	    colMask = gridColMask[ col ];
	    if (colMask == 0) {
		ULongP(pTile)[0]  = ULongP(pTile)[8]  = 0;
		ULongP(pTile)[16] = ULongP(pTile)[24] = 0;
		ULongP(pTile)[32] = ULongP(pTile)[40] = 0;
		ULongP(pTile)[48] = ULongP(pTile)[56] = 0;
		col++;
		initTile++;
		pTile++;
		DebugTask(DCTD_ValidateDct,
		    CheckIdct(lee, pTile-1, blkIndex, col-1, False)
		)
		continue;
	    }
	    nCoeff = lee->bitCountTable[ colMask ];
	    if (nCoeff == 1) {
		if (colMask == 1) {
		    f1 = lee->dctConst[0];
		    f0 = initTile[ 0<<3 ];
		    f0 = f0 * f1;
		    col++;
		    initTile++;
		    pTile[0<<3] = pTile[1<<3] =
		    pTile[2<<3] = pTile[3<<3] =
		    pTile[4<<3] = pTile[5<<3] =
		    pTile[6<<3] = pTile[7<<3] = f0;
		    pTile++;
		} else if (colMask == (1<<1)) {
		    f1 = initTile[ 1<<3 ];
		    f2 = lee->dctConst[ 0 ];
		    f0 = initTile[ 0<<3 ];
		    f4 = f2 * f1;
		    f2 = lee->dctConst[ 1 ];
		    f5 = f1 + f4;
		    f3 = lee->dctConst[ 2 ];
		    f4 = f1 - f4;
		    f5 = f2 * f5;
		    f2 = lee->dctConst[ 3 ];
		    f4 = f3 * f4;
		    f3 = lee->dctConst[ 6 ];
		    f6 = f1 + f5;
		    f5 = f1 - f5;
		    f6 = f2 * f6;
		    f2 = lee->dctConst[ 4 ];
		    f5 = f3 * f5;
		    f3 = lee->dctConst[ 5 ];
		    f7 = f1 + f4;
		    f4 = f1 - f4;
		    f7 = f2 * f7;
		    f4 = f3 * f4;
		    initTile++;
		    col++;
		    pTile[0<<3] = f6;
		    pTile[7<<3] = -f6;
		    pTile[1<<3] = f7;
		    pTile[6<<3] = -f7;
		    pTile[2<<3] = f4;
		    pTile[5<<3] = -f4;
		    pTile[3<<3] = f5;
		    pTile[4<<3] = -f5;
		    pTile++;
		} else if (colMask == (1<<2)) {
		    f2 = initTile[ 2<<3 ];
		    f0 = lee->dctConst[ 0 ];
		    f3 = f2 * f0;
		    f0 = lee->dctConst[ 1 ];
		    f4 = f2 + f3;
		    f1 = lee->dctConst[ 2 ];
		    f2 = f2 - f3;
		    f4 = f0 * f4;
		    f0 = f1 * f2;
		    initTile++;
		    col++;
		    pTile[1<<3] = f0;
		    pTile[6<<3] = f0;
		    pTile[2<<3] = -f0;
		    pTile[5<<3] = -f0;
		    pTile[0<<3] = f4;
		    pTile[7<<3] = f4;
		    pTile[3<<3] = -f4;
		    pTile[4<<3] = -f4;
		    pTile++;
		} else if (colMask == (1<<3)) {
		    f3 = initTile[ 3<<3 ];
		    f0 = lee->dctConst[ 0 ];
		    f0 = f3 * f0;
		    f1 = lee->dctConst[ 1 ];
		    f2 = lee->dctConst[ 2 ];
		    f4 = f3 + f0;
		    f3 = f3 - f0;
		    f4 = f1 * f4;
		    f5 = f0 + f4;
		    f3 = f2 * f3;
		    f2 = lee->dctConst[ 3 ];
		    f4 = f0 - f4;
		    f1 = lee->dctConst[ 6 ];
		    f2 = f2 * f5;
		    f5 = lee->dctConst[ 4 ];
		    f6 = f3 - f0;
		    f1 = f1 * f4;
		    f4 = lee->dctConst[ 5 ];
		    f0 = f0 + f3;
		    f3 = f5 * f6;
		    initTile++;
		    col++;
		    pTile[0<<3] = f2;
		    pTile[7<<3] = -f2;
		    f4 = f4 * f0;
		    pTile[3<<3] = f1;
		    pTile[4<<3] = -f1;
		    pTile[1<<3] = f3;
		    pTile[6<<3] = -f3;
		    pTile[2<<3] = -f4;
		    pTile[5<<3] = f4;
		    pTile++;
		} else if (colMask == (1<<4)) {
		    f0 = initTile[ 4<<3 ] * lee->dctConst[0];
		    f1 = -f0;
		    initTile++;
		    col++;
		    pTile[0<<3] = f0;
		    pTile[7<<3] = f0;
		    pTile[1<<3] = f1;
		    pTile[6<<3] = f1;
		    pTile[2<<3] = f1;
		    pTile[5<<3] = f1;
		    pTile[3<<3] = f0;
		    pTile[4<<3] = f0;
		    pTile++;
		} else if (colMask == (1<<5)) {
		    f1 = initTile[ 5<<3 ] * lee->dctConst[0];
		    f2 = lee->dctConst[1] * f1;
		    f0 = -f1;
		    f3 = lee->dctConst[2] * f0;
		    f4 = lee->dctConst[3] * (f1 + f2);
		    f5 = lee->dctConst[6] * (f1 - f2);
		    f6 = lee->dctConst[4] * (f0 + f3);
		    f7 = lee->dctConst[5] * (f0 - f3);
		    initTile++;
		    col++;
		    pTile[0<<3] = f4;
		    pTile[7<<3] = -f4;
		    pTile[1<<3] = f6;
		    pTile[6<<3] = -f6;
		    pTile[2<<3] = f7;
		    pTile[5<<3] = -f7;
		    pTile[3<<3] = f5;
		    pTile[4<<3] = -f5;
		    pTile++;
		} else if (colMask == (1<<6)) {
		    f0 = initTile[ 6<<3 ] * lee->dctConst[0];
		    f1 = lee->dctConst[1] * f0;
		    f3 = -f1;
		    f4 = lee->dctConst[2] * f0;
		    f2 = -f4;
		    initTile++;
		    col++;
		    pTile[0<<3] = f1;
		    pTile[7<<3] = f1;
		    pTile[1<<3] = f2;
		    pTile[6<<3] = f2;
		    pTile[2<<3] = f4;
		    pTile[5<<3] = f4;
		    pTile[3<<3] = f3;
		    pTile[4<<3] = f3;
		    pTile++;
		} else /* (colMask == (1<<7)) */ {
		    f1 = initTile[ 7<<3 ] * lee->dctConst[0];
		    f2 = lee->dctConst[1] * f1;
		    f3 = -(lee->dctConst[2] * f1);
		    f4 = lee->dctConst[3] * f2;
		    f5 = lee->dctConst[6] * f2;
		    f6 = lee->dctConst[4] * f3;
		    f7 = lee->dctConst[5] * f3;
		    initTile++;
		    col++;
		    pTile[0<<3] = f4;
		    pTile[7<<3] = -f4;
		    pTile[1<<3] = f6;
		    pTile[6<<3] = -f6;
		    pTile[2<<3] = -f7;
		    pTile[5<<3] = f7;
		    pTile[3<<3] = -f5;
		    pTile[4<<3] = f5;
		    pTile++;
		}
	    } else if (colMask == 0x3) {
		f7 = lee->dctConst[0];
		f0 = initTile[ 0<<3 ];
		f1 = initTile[ 1<<3 ];
		f0 = f7 * f0;
		f2 = f7 * f1;
		f3 = f1 + f2;
		f2 = f1 - f2;
		f3 = lee->dctConst[1] * f3;
		f4 = f1 + f3;
		f2 = lee->dctConst[2] * f2;
		f3 = f1 - f3;
		f4 = lee->dctConst[3] * f4;
		f5 = f1 + f2;
		f3 = lee->dctConst[6] * f3;
		f2 = f1 - f2;
		f5 = lee->dctConst[4] * f5;
		pTile[ 0<<3 ] = f0 + f4;
		f2 = lee->dctConst[5] * f2;
		col++;
		initTile++;
		pTile[ 7<<3 ] = f0 - f4;
		pTile[ 1<<3 ] = f0 + f5;
		pTile[ 6<<3 ] = f0 - f5;
		pTile[ 2<<3 ] = f0 + f2;
		pTile[ 5<<3 ] = f0 - f2;
		pTile[ 3<<3 ] = f0 + f3;
		pTile[ 4<<3 ] = f0 - f3;
		pTile++;
	    } else if (colMask == 0x7) {
		f0 = initTile[ 1<<3 ];
		f1 = initTile[ 0<<3 ] * lee->dctConst[0];
		f2 = initTile[ 2<<3 ];
		f4 = lee->dctConst[0] * f0;
		f3 = f2 * lee->dctConst[0];
		f5 = lee->dctConst[1] * (f2 + f3);
		f6 = lee->dctConst[1] * (f0 + f4);
		f2 = lee->dctConst[2] * (f2 - f3);
		f4 = lee->dctConst[2] * (f0 - f4);
		f3 = f1 + f5;
		f7 = lee->dctConst[3] * (f0 + f6);
		f5 = f1 - f5;
		f6 = lee->dctConst[6] * (f0 - f6);
		ft = f1 + f2;
		f1 = f1 - f2;
		f2 = lee->dctConst[4] * (f0 + f4);
		f0 = lee->dctConst[5] * (f0 - f4);
		col++;
		initTile++;
		pTile[0<<3] = f3 + f7;
		pTile[7<<3] = f3 - f7;
		pTile[1<<3] = ft + f2;
		pTile[6<<3] = ft - f2;
		pTile[2<<3] = f1 + f0;
		pTile[5<<3] = f1 - f0;
		pTile[3<<3] = f5 + f6;
		pTile[4<<3] = f5 - f6;
		pTile++;
	    } else if (colMask == 0x5) {
		float a0 = lee->dctConst[0];

		f0 = initTile[ 0<<3 ];
		f2 = initTile[ 2<<3 ];
		f1 = f0 * a0;
		f3 = f2 * a0;
		f0 = f2 + f3;
		f4 = f2 - f3;
		f5 = lee->dctConst[1] * f0;
		f0 = lee->dctConst[2] * f4;
		f2 = f1 + f5;
		f3 = f1 - f5;
		f4 = f1 + f0;
		f5 = f1 - f0;
		col++;
		initTile++;
		pTile[0<<3] = f2;
		pTile[7<<3] = f2;
		pTile[1<<3] = f4;
		pTile[6<<3] = f4;
		pTile[2<<3] = f5;
		pTile[5<<3] = f5;
		pTile[3<<3] = f3;
		pTile[4<<3] = f3;
		pTile++;
	    } else {
		float a0, a1, a2;

		f7 = initTile[ 7<<3 ];
		f5 = initTile[ 5<<3 ];
		f1 = initTile[ 1<<3 ];
		f3 = initTile[ 3<<3 ];
		f7 += f5;
		f2 = initTile[ 2<<3 ];
		a0 = lee->dctConst[0];
		f5 += f3;
		f6 = initTile[ 6<<3 ];
		f5 *= a0;
		f3 += f1;
		f0 = initTile[ 0<<3 ];
		f7 += f3;
		f4 = initTile[ 4<<3 ];
		f7 *= a0;
		f6 += f2;
		a1 = lee->dctConst[1];
		f4 *= a0;
		ft = f3 + f7;
		a2 = lee->dctConst[2];
		f6 *= a0;
		f7 = f3 - f7;
		f3 = ft * a1;
		ft = f2 + f6;
		f7 *= a2;
		f6 = f2 - f6;
		f2 = ft * a1;
		ft = f1 + f5;
		f6 *= a2;
		f5 = f1 - f5;
		f0 *= a0;
		a0 = lee->dctConst[3];
		f1 = ft + f3;
		a1 = lee->dctConst[6];
		f3 = ft - f3;
		f1 *= a0;
		a2 = lee->dctConst[4];
		ft = f5 + f7;
		f3 *= a1;
		a0 = lee->dctConst[5];
		f7 = f5 - f7;
		f5 = ft * a2;
		ft = f0 + f4;
		f7 *= a0;
		f4 = f0 - f4;
		f0 = ft + f2;
		f2 = ft - f2;
		ft = f4 + f6;
		f6 = f4 - f6;
		col++;
		initTile++;
		pTile[0]  = f0 + f1;
		pTile[56] = f0 - f1;
		pTile[8]  = ft + f5;
		pTile[48] = ft - f5;
		pTile[24] = f2 + f3;
		pTile[32] = f2 - f3;
		pTile[16] = f6 + f7;
		pTile[40] = f6 - f7;
		pTile++;
	    }

	    DebugTask(DCTD_ValidateDct,
		CheckIdct(lee, pTile-1, blkIndex, col-1, False)
	    )
	}

	/*
         * Third time through: do the inverse DCT on the rows
         */
        pTile = lee->tile;

	if (rowMask == 1) {
	    for (row = 0; row < 8;) {
		tmp = pTile[ 0 ] * lee->dctConst[0];
		pTile += 8;
		row++;
		
		if (lee->intra) {
		    PixelClamp(lee, pixels, tmp);
		    pixels |= pixels << 8;
		    pixels |= pixels << 16;
		    ULongP(pPixel)[0] = ULongP(pPixel)[1] = pixels;
		} else {
		    DebugTask(DCTD_DeltaValues,
			ShowPreDeltaValue(blkIndex, row, pMotion,
			    tmp,tmp,tmp,tmp,tmp,tmp,tmp,tmp);
		    )
		    DCDeltaPixelClamp(lee, tmp, pMotion, pPixel);
		    pMotion += motionPicWidth;
		}
		pPixel += picWidth;
	    }
        } else if (rowMask <= 3) {
	    for (row = 0; row < 8;) {
		DebugTask(DCTD_ValidateDct,
		    RefIdct(lee, blkIndex, pTile, row, False) )
		f4 = lee->dctConst[0];
		f0 = pTile[ 0 ];
		f1 = pTile[ 1 ];
		f0 = f4 * f0;
		pTile += 8;
		f5 = lee->dctConst[2];
		f6 = lee->dctConst[1];
		f2 = f4 * f1;
		f3 = f1 + f2;
		f2 = f1 - f2;
		f3 = f6 * f3;
		f7 = lee->dctConst[3];
		f2 = f5 * f2;
		ft = lee->dctConst[6];
		f4 = f1 + f3;
		f3 = f1 - f3;
		f4 = f7 * f4;
		f3 = ft * f3;
		ft = lee->dctConst[5];
		f7 = lee->dctConst[4];
		row++;
		f5 = f1 + f2;
		f2 = f1 - f2;
		f5 = f7 * f5;
		f2 = ft * f2;

		DebugTask(DCTD_ValidateDct,
		    CheckRowIdct(lee, blkIndex, row-1,
			    f0 + f4, f0 + f5, f0 + f2, f0 + f3,
			    f0 - f3, f0 - f2, f0 - f5, f0 - f4);
		)
		if (lee->intra) {
		    IntraPixelRowClamp(lee, pPixel,
			    f0 + f4, f0 + f5, f0 + f2, f0 + f3,
			    f0 - f3, f0 - f2, f0 - f5, f0 - f4);
		} else {
		    DebugTask(DCTD_DeltaValues,
			ShowPreDeltaValue(blkIndex, row, pMotion,
			    f0 + f4, f0 + f5, f0 + f2, f0 + f3,
			    f0 - f3, f0 - f2, f0 - f5, f0 - f4);
		    )

		    DeltaPixelRowClamp(lee, pMotion, pPixel,
			    f0 + f4, f0 + f5, f0 + f2, f0 + f3,
			    f0 - f3, f0 - f2, f0 - f5, f0 - f4);
		    pMotion += motionPicWidth;
		}
		pPixel += picWidth;
	    }
        } else if (rowMask <= 7) {
	    for (row = 0; row < 8;) {
		DebugTask(DCTD_ValidateDct,
		    RefIdct(lee, blkIndex, pTile, row, False) )
		f0 = pTile[ 0 ];
		f1 = pTile[ 1 ];
		f2 = pTile[ 2 ];
		f0 = f0 * lee->dctConst[0];
		f3 = f2 * lee->dctConst[0];
		f4 = lee->dctConst[0] * f1;
		pTile += 8;
		row++;
		ft = f2 + f3;
		f5 = lee->dctConst[1] * ft;
		ft = f1 + f4;
		f6 = lee->dctConst[1] * ft;
		ft = f2 - f3;
		f2 = lee->dctConst[2] * ft;
		ft = f1 - f4;
		f3 = lee->dctConst[2] * ft;
		f4 = f0 + f5;
		f5 = f0 - f5;
		f7 = f1 + f6;
		f6 = f1 - f6;
		f7 = lee->dctConst[3] * f7;
		f6 = lee->dctConst[6] * f6;
		ft = f0 + f2;
		f2 = f0 - f2;
		f0 = f1 + f3;
		f3 = f1 - f3;
		f0 = lee->dctConst[4] * f0;
		f3 = lee->dctConst[5] * f3;

		DebugTask(DCTD_ValidateDct,
		    CheckRowIdct(lee, blkIndex, row-1,
				 f4 + f7, ft + f0, f2 + f3, f5 + f6,
				 f5 - f6, f2 - f3, ft - f0, f4 - f7);
		)
		if (lee->intra) {
		    IntraPixelRowClamp(lee, pPixel,
				       f4 + f7, ft + f0, f2 + f3, f5 + f6,
				       f5 - f6, f2 - f3, ft - f0, f4 - f7);
		} else {
		    DebugTask(DCTD_DeltaValues,
			ShowPreDeltaValue(blkIndex, row, pMotion,
			    f4 + f7, ft + f0, f2 + f3, f5 + f6,
			    f5 - f6, f2 - f3, ft - f0, f4 - f7);
		    )

		    DeltaPixelRowClamp(lee, pMotion, pPixel,
				       f4 + f7, ft + f0, f2 + f3, f5 + f6,
				       f5 - f6, f2 - f3, ft - f0, f4 - f7);
		    pMotion += motionPicWidth;
		}
		pPixel += picWidth;
	    }
        } else if (rowMask <= 15) {
	    for (row = 0; row < 8;) {
		DebugTask(DCTD_ValidateDct,
		    RefIdct(lee, blkIndex, pTile, row, False) )
		f0 = pTile[ 0 ];
		f1 = pTile[ 1 ];
		f2 = pTile[ 2 ];
		f3 = pTile[ 3 ];
		f4 = f1 + f3;
		f0 = f0 * lee->dctConst[0];
		pTile += 8;
		row++;
		f3 = f3 * lee->dctConst[0];
		f5 = f2 * lee->dctConst[0];
		f6 = lee->dctConst[0] * f4;
		f7 = f1 + f3;
		f1 = f1 - f3;
		f3 = f2 + f5;
		f2 = f2 - f5;
		f3 = lee->dctConst[1] * f3;
		f2 = lee->dctConst[2] * f2;
		f5 = f4 + f6;
		f4 = f4 - f6;
		f5 = lee->dctConst[1] * f5;
		f4 = lee->dctConst[2] * f4;
		f6 = f0 + f3;
		f3 = f0 - f3;
		ft = f7 + f5;
		f5 = f7 - f5;
		ft = lee->dctConst[3] * ft;
		f5 = lee->dctConst[6] * f5;
		f7 = f0 + f2;
		f2 = f0 - f2;
		f0 = f1 + f4;
		f1 = f1 - f4;
		f0 = lee->dctConst[4] * f0;
		f1 = lee->dctConst[5] * f1;

		DebugTask(DCTD_ValidateDct,
		    CheckRowIdct(lee, blkIndex, row-1,
				 f6 + ft, f7 + f0, f2 + f1, f3 + f5,
				 f3 - f5, f2 - f1, f7 - f0, f6 - ft);
		)
		if (lee->intra) {
		    IntraPixelRowClamp(lee, pPixel,
				 f6 + ft, f7 + f0, f2 + f1, f3 + f5,
				 f3 - f5, f2 - f1, f7 - f0, f6 - ft);
		} else {
		    DebugTask(DCTD_DeltaValues,
			ShowPreDeltaValue(blkIndex, row, pMotion,
				 f6 + ft, f7 + f0, f2 + f1, f3 + f5,
				 f3 - f5, f2 - f1, f7 - f0, f6 - ft);
		    )

		    DeltaPixelRowClamp(lee, pMotion, pPixel,
				 f6 + ft, f7 + f0, f2 + f1, f3 + f5,
				 f3 - f5, f2 - f1, f7 - f0, f6 - ft);
		    pMotion += motionPicWidth;
		}
		pPixel += picWidth;
	    }
	} else {
	    for (row = 0; row < 8;) {
		DebugTask(DCTD_ValidateDct,
		    RefIdct(lee, blkIndex, pTile, row, False) )
		f7 = pTile[ 7 ];
		f5 = pTile[ 5 ];
		f1 = pTile[ 1 ];
		f3 = pTile[ 3 ];
		f7 += f5;
		f2 = pTile[ 2 ];
		f5 += f3;
		f6 = pTile[ 6 ];
		f5 *= lee->dctConst[0];
		row++;
		f3 += f1;
		f0 = pTile[ 0 ];
		f7 += f3;
		f4 = pTile[ 4 ];
		f7 *= lee->dctConst[0];
		pTile += 8;
		f6 += f2;
		f4 *= lee->dctConst[0];
		ft = f3 + f7;
		f6 *= lee->dctConst[0];
		f7 = f3 - f7;
		f3 = ft * lee->dctConst[1];
		ft = f2 + f6;
		f7 *= lee->dctConst[2];
		f6 = f2 - f6;
		f2 = ft * lee->dctConst[1];
		ft = f1 + f5;
		f6 *= lee->dctConst[2];
		f5 = f1 - f5;
		f0 *= lee->dctConst[0];
		f1 = ft + f3;
		f3 = ft - f3;
		f1 *= lee->dctConst[3];
		ft = f5 + f7;
		f3 *= lee->dctConst[6];
		f7 = f5 - f7;
		f5 = ft * lee->dctConst[4];
		ft = f0 + f4;
		f7 *= lee->dctConst[5];
		f4 = f0 - f4;
		f0 = ft + f2;
		f2 = ft - f2;
		ft = f4 + f6;
		f6 = f4 - f6;

		DebugTask(DCTD_ValidateDct,
		    CheckRowIdct(lee, blkIndex, row-1,
				 f0 + f1, ft + f5, f6 + f7, f2 + f3,
				 f2 - f3, f6 - f7, ft - f5, f0 - f1);
		)
		if (lee->intra) {
		    IntraPixelRowClamp(lee, pPixel,
				       f0 + f1, ft + f5, f6 + f7, f2 + f3,
				       f2 - f3, f6 - f7, ft - f5, f0 - f1);
		} else {
		    DebugTask(DCTD_DeltaValues,
			ShowPreDeltaValue(blkIndex, row, pMotion,
			    f0 + f1, ft + f5, f6 + f7, f2 + f3,
			    f2 - f3, f6 - f7, ft - f5, f0 - f1);
		    )

		    DeltaPixelRowClamp(lee, pMotion, pPixel,
				       f0 + f1, ft + f5, f6 + f7, f2 + f3,
				       f2 - f3, f6 - f7, ft - f5, f0 - f1);
		    pMotion += motionPicWidth;
		}
		pPixel += picWidth;
	    }
	}

	/*
	 * Replace the initial reconstructed coefficients with zeros
	 * so that the block is again filled with zeros.  Unrolling this
	 * loop by hand cannot beat the sgi compiler (which unrolls it
	 * for you).
	 */
        pCoeff = UShortP(&dctBlk->coefList[0]);
        nCoeff = dctBlk->hdr.s.nCoeff;
        ((long *)lee->initTile)[0] = 0;
        for (i = 0; i < nCoeff; i++) {
            coeff = pCoeff[ i ];
            index = GetDctCoeffIndex(coeff);
            ULongP(lee->initTile)[ index ] = 0;
        }

        /*
         * Finally, move to the next block
         */
        nCoeff = dctBlk->hdr.s.nCoeff;
        dctBlk = DctCoeffPtrToNextBlock(&dctBlk->coefList[nCoeff]);

	WhenDebugEnabled(
	    pPixel = lee->super.pTo[ picType ]
		   + lee->super.geom.blkOffset[ blkIndex ];
	    if (lee->intra) {
		if (GetDebugBit(DCTD_IntraValues))
		    ShowIntraValue(blkIndex, pPixel, picWidth);
	    } else {
		if (GetDebugBit(DCTD_DeltaValues))
		    ShowPostDeltaValue(pPixel, picWidth);
	    }
	)
    }
    
    idctMb = DctBlockPtrToNextMb(dctBlk);
    
    return(idctMb);
}

static void
leeInverseDct(IdctParamPtr idctp)
{
    DctMacroblockPtr idctMb;
    DctListInfoPtr idctList;
    LeeIdctPrivPtr lee = (LeeIdctPrivPtr)idctp;
    long lumaInc, cromaInc;
    
    for (idctList = idctp->arenaQ; idctList; idctList = idctp->arenaQ) {
        idctMb = idctList->curIdctMb;
        for (;;) {
	    lee->intra = idctMb->flags.b.intra;

            /*
             * Process the blocks
             */
            idctMb = ProcessBlocks(lee, idctMb);

            /*
             * Decide here whether we have consumed our share of time.
             * If so, return.  If the arenaQ is non-null, then we
             * will return to this function later.
             */
            if (0) {
            	idctList->curIdctMb = idctMb;
            	goto doneForNow;
            }

	    /*
	     * update the pointer to the current macroblock pixels.
	     */
	    lumaInc = 16;
	    cromaInc = 8;
	    if (RowStartMacro(&lee->super, lee->super.super.nextMbAddr)) {
		int lumaWidth = lee->super.geom.pic[ Dct444Pic ].width;
		int cromaWidth = lee->super.geom.pic[ Dct420Pic ].width;
		lumaInc += RowStartAccel(lumaWidth, 4);
		cromaInc += RowStartAccel(cromaWidth, 3);
	    }
	    lee->super.pTo[ Dct444Pic ] += lumaInc;
	    lee->super.pTo[ Dct420Pic ] += cromaInc;
	    lee->super.super.nextMbAddr++;
	    if (idctMb->flags.b.endOfList)
		break;
        }

        /*
         * Clean up after this one and move to the next.
         */
        idctp->arenaQ = idctList->next;
        idctList->next = lee->super.arenaFree;
        lee->super.arenaFree = idctList;
    }

doneForNow:
    if (idctp->arenaQ == NULL)
        lee->super.arenaLast = NULL;
}

static int
leeIdctParamSize(void)
{
    return(sizeof(LeeIdctPriv));
}

static void
leeNewQuantMatrices(IdctParamPrivPtr priv, UChar *iqm, UChar *niqm)
{
    int i, j;
    LeeIdctPrivPtr lee = (LeeIdctPrivPtr)priv;
    QLookupType *lookup;
    
    if (iqm) {
	for (i = 1; i <= NPreScaled; i++) {
	    lookup = lee->iqmCache[ i-1 ].lookup;
	    for (j = 0; j < 64; j++) {
		lookup[ j ] = iqm[ j ] * i * 2;
	    }
	}
    }

    if (niqm) {
	for (i = 1; i <= NPreScaled; i++) {
	    lookup = lee->niqmCache[ i-1 ].lookup;
	    for (j = 0; j < 64; j++) {
		lookup[ j ] = niqm[ j ] * i;
	    }
	}
    }
}

static InternalReasonPtr
leeCreateIdctParam(IdctParamPtr idctp)
{
    LeeIdctPrivPtr lee = (LeeIdctPrivPtr)idctp;
    int i, j, bitCount;
    
    /*
     * XXX very important!  To save on a >>1 (or *0.5) in some places
     * within the inverse dct, the entire algorithm has a *0.5 factored
     * into it.  Thus the coeficients below are twice as big as they
     * would be normally.
     */
    idctp->inverseDct = leeInverseDct;
    lee->super.motionComp = Idct420MotionCompensate;
    lee->super.newQuantMatrices = leeNewQuantMatrices;
    lee->dctConst[0] = 1.0 / sqrt(2.0);
    lee->dctConst[1] = 1.0 / (2.0*cos(      M_PI / 8.0));
    lee->dctConst[2] = 1.0 / (2.0*cos(3.0 * M_PI / 8.0));
    lee->dctConst[3] = 1.0 / (2.0*cos(      M_PI / 16.0));
    lee->dctConst[4] = 1.0 / (2.0*cos(3.0 * M_PI / 16.0));
    lee->dctConst[5] = 1.0 / (2.0*cos(5.0 * M_PI / 16.0));
    lee->dctConst[6] = 1.0 / (2.0*cos(7.0 * M_PI / 16.0));
    lee->coeffScale = 0.25;
    lee->dctDcConst = lee->dctConst[0] * lee->dctConst[0] * lee->coeffScale;
    lee->indexOff[ 0 ] = 64 * sizeof(float);

    /*
     * This creates a fixed point version of dctDcConst with the decimal
     * point to the left of bit 15.
     */
    lee->integer.dctDcConst = lee->dctDcConst * (1<<NFracBits);

    for (i = 0; i < 256; i++) {
	bitCount = 0;
	for (j = 0; j < 8; j++)
	    bitCount += (i >> j) & 1;
	lee->bitCountTable[ i ] = bitCount;
    }

    WhenUsingIndexClamp(
	for (i = 0; i < 255; i++)
	    lee->clamps[i] = i;
	for (; i < MaxPositivePixel; i++)
	    lee->clamps[i] = 255;
    )

    return(NULL);
}

/***************************************************************************
 *
 * public method interfaces.
 *
 ***************************************************************************/

static IdctMethod leeIdctMethod = {
    leeIdctParamSize,
    leeCreateIdctParam
};

IdctMethodPtr FloatLeeIdctMethod = &leeIdctMethod;
