/*
 * (c) Copyright 1992, 1993, 1994 North Valley Research, Inc.
 * 	All Rights Reserved
 *
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	floatLeeIdctI.h - internal definitions for the floating point
 *                      implementation of Lee's inverse dct operations
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/floatLeeIdctI.h,v 1.20 1995/10/15 23:55:00 toddb Exp $
 */
#ifndef FLOATLEEIDCTDCTI_H
#define FLOATLEEIDCTDCTI_H

#include "nvrIdctI.h"

#ifndef DEBUG_ENABLED
#define ValidateIntraReconstrution(_lee, _level, _minus, _index) /* nothing */
#define ValidateDeltaReconstrution(_lee, _level, _minus, _index) /* nothing */
#define ValidateDcDeltaReconstrution(_lee, _level, _minus, _coef) /* nothing */
#endif

/*
 * Since the implementation of this algorithm eliminates a multiply of 0.25
 * across the entire matrix, we must scale the reconstructed coeficients
 * by 0.25; This bit of magic will accomplish the scale by removing the
 * exponent, subtracting 2 (effectively dividing by 2^2=4) and then
 * throw in the sign bit.  This should work for IEEE floats which look
 * like:
 *	s eeeeeeee mmmmmmmmmmmmmmmmmmmmmmm
 *      1|--- 8 --|---------- 23 ---------|
 *
 * On entry, we know the float is positive.
 */
#define ScaleFloat(_coef, _minus)					\
    {									\
        union {								\
	    ULong ival;							\
	    float fval;							\
	} u;								\
	u.fval = _coef;							\
	u.ival -= 2 << 23;		/* divides by 4 */		\
	u.ival |= _minus << 31;						\
	_coef = u.fval;							\
    }

#define ReconstructIntegerMethod	1
#define ReconstructFloatMethod		2

#ifndef ReconstructMethod
#define ReconstructMethod ReconstructIntegerMethod
#endif

#if ReconstructMethod == ReconstructIntegerMethod

typedef short QLookupType;

/*
 * This macro is optimized from the protocol, in that the Sign() function
 * is unnecessary because the sign is applied last.
 */
#define ReconstructIntraCoeff(_lee, _index, _qc, _level, _minus)	\
    {									\
	int _qentry = _qc->lookup[ _index ];				\
	int _recon;							\
	float _frecon;							\
									\
	if      (_level == 1) _recon = _qentry >> 4;			\
	else if (_level == 2) _recon = _qentry >> 3;			\
	else if (_level == 3) _recon = ((_qentry<<1) + _qentry) >> 4;	\
	else if (_level == 4) _recon = _qentry >> 2;			\
	else                  _recon = (_level * _qentry) >> 4;		\
									\
	_frecon = _recon - ((~_recon) & 1); /* force odd */		\
	ScaleFloat(_frecon, _minus);					\
	_lee->initTile[ _index ] = _frecon;				\
	ValidateIntraReconstrution(_lee, _level, _minus, _index);	\
    }

/*
 * This macro is optimized from the protocol, in that the Sign() function
 * is unnecessary the sign is kept separate, so that it will always
 * return 1.  Note that the formula is ((2 * level) + Sign) * qentry) >> 4
 * and works out to be:
 *
 *         level      formula
 *         -----      -------
 *           1        (3 * qentry) >> 4
 *           2        (5 * qentry) >> 4
 *           3        (7 * qentry) >> 4	( == ((8 * quentry) - qentry) >> 4
 *           4        (9 * qentry) >> 4
 */
#define ReconstructDeltaCoeff(_lee, _index, _qc, _lvl, _minus)		\
    {									\
	int _qe = _qc->lookup[ _index ];				\
	int _recon;							\
	float _frecon;							\
									\
	if      (_lvl == 1) _recon = ((_qe<<1) + _qe) >> 4;		\
	else if (_lvl == 2) _recon = ((_qe<<2) + _qe) >> 4;		\
	else if (_lvl == 3) _recon = ((_qe<<3) - _qe) >> 4;		\
	else if (_lvl == 4) _recon = ((_qe<<3) + _qe) >> 4;		\
	else                _recon = (((_lvl<<1)+1) * _qe) >> 4;	\
									\
	_frecon = _recon - ((~_recon) & 1); /* force odd */		\
	ScaleFloat(_frecon, _minus);					\
	_lee->initTile[ _index ] = _frecon;				\
	ValidateDeltaReconstrution(_lee, _lvl, _minus, _index);		\
    }

/*
 * This is a special version of the delta reconstruction for a DC
 * delta block, which seems to be very frequent in 1.5mb streams (~15%)
 * Note that the coefficient is not scaled here because the next
 * factor applied (by CalcDCPixel2d) includes the scale factor.
 * 
 */
#define ReconstructDCDeltaCoeff(_lee, _coef, _qc, _lvl, _minus)		\
    {									\
	int _qe = _qc->lookup[ 0 ];					\
	int _recon;							\
	float _frecon;							\
									\
	if      (_lvl == 1) _recon = ((_qe<<1) + _qe) >> 4;		\
	else if (_lvl == 2) _recon = ((_qe<<2) + _qe) >> 4;		\
	else if (_lvl == 3) _recon = ((_qe<<3) - _qe) >> 4;		\
	else if (_lvl == 4) _recon = ((_qe<<3) + _qe) >> 4;		\
	else                _recon = (((_lvl<<1) + 1) * _qe) >> 4;	\
									\
	_frecon = _recon - ((~_recon) & 1); /* force odd */		\
	if (_minus)							\
	    _frecon = -_frecon;						\
	_coef = _frecon;						\
	ValidateDcDeltaReconstrution(_lee, _lvl, _minus, _coef);	\
    }

#endif /* ReconstructMethod == ReconstructIntegerMethod */

#if ReconstructMethod == ReconstructFloatMethod

typedef float QLookupType;

/*
 * This macro is optimized from the protocol, in that the Sign() function
 * is unnecessary because the sign is applied last.
 */
#define ReconstructIntraCoeff(_lee, _index, _qc, _level, _minus)	\
    {									\
	float _qentry = _qc->lookup[ _index ];				\
	int _recon = ((int)(_level * _qentry)) >> 4;			\
	float _frecon = _recon - ((~_recon) & 1); /* force odd */	\
	ScaleFloat(_frecon, _minus);					\
	_lee->initTile[ _index ] = _frecon;				\
	ValidateIntraReconstrution(_lee, _level, _minus, _index);	\
    }

/*
 * This macro is optimized from the protocol, in that the Sign() function
 * is unnecessary the sign is kept separate, so that it will always
 * return 1.  Note that the formula is ((2 * level) + Sign) * qentry) >> 4
 * and works out to be:
 */
#define ReconstructDeltaCoeff(_lee, _index, _qc, _level, _minus)	\
    {									\
	float _qentry = _qc->lookup[ _index ];				\
	int _recon = ((int)(((_level<<1)+_level) * _qentry)) >> 4;	\
	float _frecon = _recon - ((~_recon) & 1); /* force odd */	\
	ScaleFloat(_frecon, _minus);					\
	_lee->initTile[ _index ] = _frecon;				\
	ValidateDeltaReconstrution(_lee, _level, _minus, _index);	\
    }

/*
 * This is a special version of the delta reconstruction for a DC
 * delta block, which seems to be very frequent in 1.5mb streams (~15%)
 * Note that the coefficient is not scaled here because the next
 * factor applied (by CalcDCPixel2d) includes the scale factor.
 * 
 */
#define ReconstructDCDeltaCoeff(_lee, _coef, _qc, _level, _minus)	\
    {									\
	float _qentry = _qc->lookup[ 0 ];				\
	int _recon = ((int)(((_level<<1)+_level) * _qentry)) >> 4;	\
	float _frecon;							\
									\
	_recon -= ((~_recon) & 1); /* force odd */			\
	if (_minus)							\
	    _recon = -_recon;						\
	_coef = _recon;							\
	ValidateDcDeltaReconstrution(_lee, _level, _minus, _coef);	\
    }

#endif /* ReconstructMethod == ReconstructIntegerMethod */
/*
 * ArithmeticClamp: faster on ppc 601, by perhaps .08 fps
 */
#define ArithmeticClamp	1
#define RangeCheckClamp	2
#define IndexedClamp	3

#ifndef PixelClampStyle
#define PixelClampStyle IndexedClamp
#endif

/*
 * Final clamping for intra and non-intra coded blocks.
 */
#if PixelClampStyle == RangeCheckClamp
#define PixelClamp(lee, storage, value)				\
    {								\
	int _clamped = (int)(value);				\
	if (_clamped >> 8) {					\
	    if (_clamped > 255)					\
		_clamped = 255;					\
	    else						\
		_clamped = 0;					\
	}							\
	storage = _clamped;					\
    }

#define WhenUsingIndexClamp(code)

#endif /* PixelClampStyle == RangeCheckClamp */

#if PixelClampStyle == ArithmeticClamp
#define PixelClamp(lee, storage, value)				\
    {								\
	int _clamped = (int)(value);				\
	/*							\
	 * This clamps negative values to zero			\
	 */							\
	_clamped &= ~(_clamped >> 31);				\
	/*							\
	 * This clamps values > 255 to 255			\
	 */							\
	_clamped |= (~((_clamped - 255) >> 31) & 0xff);		\
	storage = _clamped;					\
    }

#define WhenUsingIndexClamp(code)

#endif /* PixelClampStyle == ArithmeticClamp */

#if PixelClampStyle == IndexedClamp

#define PixelClamp(lee, storage, value)				\
    storage = lee->clamps[ (int)(value) ]

#define WhenUsingIndexClamp(code) code
#define MaxNegativePixel	100
#define MaxPositivePixel	512

#define DCDeltaPixelClamp(_lee, _dc, _pMotion, _pPixel)			\
    {								\
	int _p0,_p1,_p2,_p3,_p4,_p5,_p6,_p7;			\
								\
	_p0 =  _lee->clamps[ _dc + _pMotion[0] ];		\
	_p1 =  _lee->clamps[ _dc + _pMotion[1] ];		\
	_p2 =  _lee->clamps[ _dc + _pMotion[2] ];		\
	_p3 =  _lee->clamps[ _dc + _pMotion[3] ];		\
	_p4 =  _lee->clamps[ _dc + _pMotion[4] ];		\
	_p5 =  _lee->clamps[ _dc + _pMotion[5] ];		\
	_p6 =  _lee->clamps[ _dc + _pMotion[6] ];		\
	_p7 =  _lee->clamps[ _dc + _pMotion[7] ];		\
	_pPixel[0] = _p0;					\
	_pPixel[1] = _p1;					\
	_pPixel[2] = _p2;					\
	_pPixel[3] = _p3;					\
	_pPixel[4] = _p4;					\
	_pPixel[5] = _p5;					\
	_pPixel[6] = _p6;					\
	_pPixel[7] = _p7;					\
    }

#define DeltaPixelRowClamp(_lee,_pMot,_pPix,_f0,_f1,_f2,_f3,_f4,_f5,_f6,_f7) \
    {								\
	int _p0,_p1,_p2,_p3,_p4,_p5,_p6,_p7;			\
								\
	_p0 = _f0;						\
	_p1 = _f1;						\
	_p2 = _f2; _p0 += _pMot[0];				\
	_p3 = _f3; _p1 += _pMot[1];				\
	_p4 = _f4; _p2 += _pMot[2]; _p0 = _lee->clamps[_p0];	\
	_p5 = _f5; _p3 += _pMot[3]; _p1 = _lee->clamps[_p1];	\
	_p6 = _f6; _p4 += _pMot[4]; _p2 = _lee->clamps[_p2]; _pPix[0] = _p0; \
	_p7 = _f7; _p5 += _pMot[5]; _p3 = _lee->clamps[_p3]; _pPix[1] = _p1; \
		   _p6 += _pMot[6]; _p4 = _lee->clamps[_p4]; _pPix[2] = _p2; \
		   _p7 += _pMot[7]; _p5 = _lee->clamps[_p5]; _pPix[3] = _p3; \
				    _p6 = _lee->clamps[_p6]; _pPix[4] = _p4; \
				    _p7 = _lee->clamps[_p7]; _pPix[5] = _p5; \
							     _pPix[6] = _p6; \
							     _pPix[7] = _p7; \
    }

#define IntraPixelRowClamp(_lee,_pPix,_f0,_f1,_f2,_f3,_f4,_f5,_f6,_f7) \
    {								\
	int _p0,_p1,_p2,_p3,_p4,_p5,_p6,_p7;			\
								\
	_p0 = _f0; _p0 = _lee->clamps[_p0]; _pPix[0] = _p0;	\
	_p1 = _f1; _p1 = _lee->clamps[_p1]; _pPix[1] = _p1;	\
	_p2 = _f2; _p2 = _lee->clamps[_p2]; _pPix[2] = _p2;	\
	_p3 = _f3; _p3 = _lee->clamps[_p3]; _pPix[3] = _p3;	\
	_p4 = _f4; _p4 = _lee->clamps[_p4]; _pPix[4] = _p4;	\
	_p5 = _f5; _p5 = _lee->clamps[_p5]; _pPix[5] = _p5;	\
	_p6 = _f6; _p6 = _lee->clamps[_p6]; _pPix[6] = _p6;	\
	_p7 = _f7; _p7 = _lee->clamps[_p7]; _pPix[7] = _p7;	\
    }

#endif /* PixelClampStyle == IndexedClamp */

/***************************************************************************
 *
 * DC pixel calculation.
 *
 ***************************************************************************/
#define NFracBits	16
#define DC2DFloatMethod		1
#define DC2DIntegerMethod	2

#ifndef DC2DMethod
#define DC2DMethod DC2DIntegerMethod
#endif

#if DC2DMethod == DC2DFloatMethod
#define CalcDCPixel2d(_lee, _dcCoef)	((_dcCoef) * _lee->dctDcConst)
#endif

#if DC2DMethod == DC2DIntegerMethod
#define FixedOneHalf	(1<<(NFracBits-1))
#define CalcDCPixel2d(_lee, _dcCoef)	\
    (((_dcCoef) * _lee->integer.dctDcConst + FixedOneHalf) >> NFracBits)
#endif

/***************************************************************************
 *
 * private method interfaces.
 *
 ***************************************************************************/
#define NPreScaled	31	/* All of the possible quantizer values */

typedef struct _quantCache {
    QLookupType	lookup[ 64 ];	/* pre-multiplied quatizers and q matrices */
} QuantCache, *QuantCachePtr;

typedef struct _leeIdctPriv {
    IdctParamPriv super;

    boolean intra;		/* local copy of idctMb->flags.b.intra */
    /*
     * The indexOff array is used as a (byte) offset into 'tile[]' based
     * on whether the indexed coefficient is present.  indexOff[] is
     * initialized as follows:
     *     indexOff[0] = 64 * sizeof(float)
     *     indexOff[1] = 0
     * If the coefficient is present (index 1), then we use the
     * bottom 64 entries; otherwise we use the top 64 entries which
     * are permanently initialized to zero.
     */
    int		indexOff[ 2 ];	
    float	initTile[ 64 ];	/* used to hold initial coefficients */
    float	tile[ 64 ];	/* used to hold intermediat and final results */

    float	coeffScale;	/* prescale for coefficient reconstruction */
    float	dctConst[ 7 ];	/* pre-calculated dct constants */
    float	dctDcConst;	/* used for DC blocks */

    struct {
	long dctDcConst;	/* used for int calculation of DC blocks */
    } integer;

    WhenUsingIndexClamp(
	/*
	 * These two arrays are used to clamp final pixel values to 0-255.
	 * the ->clamps array is dereferenced (meaning that ->negativeClamps
	 * is referenced for negative values).  The assumption is that
	 * final values before clamping fall in the range of
	 * -MaxNegativePixel to MaxPositivePixel-1
	 */
	UChar	negativeClamps[ MaxNegativePixel ];
	UChar	clamps[ MaxPositivePixel ]; 
    )
    UChar bitCountTable[ 256 ];

    QuantCache  iqmCache[ NPreScaled ];
    QuantCache  niqmCache[ NPreScaled ];
    QuantCachePtr iqmCur;
    QuantCachePtr niqmCur;

    /*
     * Validation
     */
    DebugDeclare(
	float vDctInput[ 64 ];
	float vDctOutput[ 64 ];
	int quant		/* current macroblock quantizer */
    )
} LeeIdctPriv, *LeeIdctPrivPtr;

#endif /* FLOATLEEIDCTDCTI_H */
