/*
 * Copyright (c) 1992,1993,1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *      nvrIdct.c - Central functions for dct manipulation
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/nvrIdct.c,v 1.14 1995/09/29 19:49:53 russ Exp $
 */

#include "nvrIdctI.h"
#include "nvrAlloc.h"
#include "nvrString.h"

char *IdctMotionIndexNames[] = {
    "fwd horiz",
    "fwd vert",
    "bak horiz",
    "bak vert"
};

static void
NewQuantMatrices(IdctParamPtr idctp, UChar *iqm, UChar *niqm)
{
    IdctParamPrivPtr priv = (IdctParamPrivPtr)idctp;
    
    if (iqm)
        memcpy((char *)priv->intraQ, (char *)iqm, 64); 
    if (niqm)
        memcpy((char *)priv->nonIntraQ, (char *)niqm, 64);
    if (priv->newQuantMatrices)
        (*priv->newQuantMatrices)(priv, iqm, niqm);
}

#ifdef DEBUG_ENABLED
boolean
CheckPicXY(YcImagePtr yci, UChar *p)
{
    int chromaHeight = yci->lumaHeight >> (yci->sample == Sample4_2_0);
    int chromaSize = yci->chromaWidth * chromaHeight;

    if (p < yci->y
     || p >= yci->cb + chromaSize) {
	return(True);
    }
    return(False);
}

void
ShowPicXY(YcImagePtr yci, UChar *p)
{
    int lumaSize = yci->lumaWidth * yci->lumaHeight;
    int chromaHeight = yci->lumaHeight >> (yci->sample == Sample4_2_0);
    int chromaSize = yci->chromaWidth * chromaHeight;
    int width, offset, x, y, oob;
    char *name;

    oob = 0;
    if (p < yci->y) {
	oob = offset = p - yci->y;
	width = yci->lumaWidth;
	name = "before y";
    } else if (p >= yci->y && p < yci->y + lumaSize) {
	offset = p - yci->y;
	width = yci->lumaWidth;
	name = "y";
    } else if (p >= yci->cr && p < yci->cr + chromaSize) {
	offset = p - yci->cr;
	width = yci->chromaWidth;
	name = "cr";
    } else if (p >= yci->cb && p < yci->cb + chromaSize) {
	offset = p - yci->cb;
	width = yci->chromaWidth;
	name = "cb";
    } else {
	offset = p - yci->cb;
	oob = p - (yci->cb + chromaSize);
	width = yci->chromaWidth;
	name = "after cb";
    }
    y = offset / width;
    x = offset % width;
    Msg("image @0x%x: p=0x%x is %s (%d,%d)", yci, p, name, x, y);
    if (oob)
	Msg(" oob by %d (%d,%d)", oob, oob % width, oob / width);
    Msg("\n");
}
#endif /* DEBUG_ENABLED */

static void
NewPicture(IdctParamPtr idctp,
           UChar fullPel[ 2 ],
           YcImagePtr pTo,
           YcImagePtr pPast,
           YcImagePtr pFuture)
{
    IdctParamPrivPtr priv = (IdctParamPrivPtr)idctp;

    WhenDebugEnabled(
	priv->past = pPast;
	priv->current = pTo;
	priv->future = pFuture;
    )

    /*
     * Initialize the pointers so that we can enter the loop and
     * increment before using.  This allows us to use a do loop with 
     * a single test instead of a while loop that tests twice.
     */
    priv->pTo[ Dct444Pic ]  = pTo->y;
    priv->pTo[ Dct420Pic ] = pTo->cr;
    priv->super.nextMbAddr = 0;
    priv->fwdFullPel = fullPel[ kIndexFwd ];
    priv->bakFullPel = fullPel[ kIndexBak ];
    
    /* XXX debug */
    /* memset((char *)pTo->y, 0, pTo->lumaWidth * pTo->lumaHeight * 3 / 2); */
    
    /*
     * XXX - this assumes that pPast and pFuture is laid out the same way
     * that pTo is (reasonable, but dangerous)
     */
    if (pFuture) {
        priv->futureOffset = pFuture->y - pTo->y;
    }
    
    if (pPast) {
        priv->pastOffset = pPast->y - pTo->y;
    }
}

static InternalReasonPtr
NewPictureSize(IdctParamPtr idctp, DctPictGeomPtr geom)
{
    IdctParamPrivPtr priv = (IdctParamPrivPtr)idctp;
    int mbCount = geom->mbHeight * geom->mbWidth;
    int mbAddr, i, picType, mcAllocSize, rowAllocSize, allocSize;
    UChar *localPicAlloc;
    InternalReasonPtr irp;
    
    mcAllocSize = 8*8*geom->nBlocks;		/* for motion */
    rowAllocSize = 1 + (mbCount + 7) / 8;	/* for row end bitmasks */
    allocSize = mcAllocSize + rowAllocSize;	/* total allocation */

    localPicAlloc = (UChar *)
	AllocNamedMem(allocSize, "dct local picture data", &irp);
    if (localPicAlloc == NULL)
    	return(irp);

    if (priv->localPicAlloc)
    	free((char *)priv->localPicAlloc);
    priv->localPicAlloc = localPicAlloc;

    priv->mcSrc = localPicAlloc;
    localPicAlloc += mcAllocSize;
    priv->pRowStarts = localPicAlloc + 1; /* this makes pRowStarts[-1] valid */
    priv->geom = *geom;

    /*
     * Initialize the motion compensation arenas (multiSrcMc, singleSrcMc).
     * Note that mixedSrcMc must be fully initialized every time.
     */
    priv->multiSrcMc.pSrc[ Dct444Pic ] = priv->mcSrc;
    priv->multiSrcMc.pSrc[ Dct420Pic ] = priv->mcSrc;
    priv->allBlocksMask = (1 << geom->nBlocks) - 1;
    priv->firstChromaBlock = geom->nBlocks;
    priv->lastChromaBlock = 0;
    priv->blocksMask[ Dct444Pic ] = 0;
    priv->blocksMask[ Dct420Pic ] = 0;
    for (i = 0; i < geom->nBlocks; i++) {
	priv->multiSrcMc.picWidth[ i ] = 8;
	priv->multiSrcMc.offset[ i ] = i * 64;

	priv->singleSrcMc.offset[ i ] = geom->blkOffset[ i ];
	picType = geom->picType[ i ];
	priv->singleSrcMc.picWidth[ i ] = geom->pic[ picType ].width;

	priv->blocksMask[ picType ] |= 1<<i;
	if (picType == Dct444Pic) {
	    priv->lumaSingleSrcMc.picWidth[ i ] = geom->pic[ picType ].width;
	    priv->lumaSingleSrcMc.offset[ i ] = geom->blkOffset[ i ];
	} else {
	    if (i < priv->firstChromaBlock)
		priv->firstChromaBlock = i;
	    if (i > priv->lastChromaBlock)
		priv->lastChromaBlock = i;
	    priv->lumaSingleSrcMc.picWidth[ i ] = 8;
	    priv->lumaSingleSrcMc.offset[ i ] = i * 64;
	}

	if (geom->blkOffset[ i ] == 0)
	    priv->lumaY0Block = i;
    }

    for (mbAddr = geom->mbWidth-1; mbAddr < mbCount; mbAddr += geom->mbWidth)
        priv->pRowStarts[ mbAddr >> 3 ] |= (1<<(mbAddr & 0x7));
    return(NULL);
}

IdctParamPtr
CreateIdctParam(IdctMethodPtr method, InternalReasonPtr *pReason)
{
    IdctParamPrivPtr priv;
    int allocSize = sizeof(IdctParamPriv);
    DctListInfoPtr dctList;
    
    if (method == NULL) {
        *pReason = InternalProblem("CreateIdctParam: idct method is NULL.\n");
        return(NULL);
    }
    allocSize += (*method->privIdctParamSize)();

    priv = (IdctParamPrivPtr) InsecureAlloc(allocSize);
    if (priv == NULL) {
        *pReason = AllocReason("IDCT param struct", allocSize);
        return(NULL);
    }
    memset((char *)priv, 0, allocSize);
    
    /*
     * Allocate one arena and put it on the free list.
     */
    dctList = AllocDctArena(&priv->super, pReason);
    if (dctList == NULL) {
        free((char *)priv);
        return(NULL);
    }
    priv->arenaFree = dctList;
    priv->super.newQuantMatrices = NewQuantMatrices;
    priv->super.newPictureSize = NewPictureSize;
    priv->super.newPicture = NewPicture;

    *pReason = (*method->createIdctParam)(&priv->super);
    if (*pReason) {
        FreeIdctParam(&priv->super);
        priv = NULL;
    }
    
    return(&priv->super);
}

void
ClearDctArena(IdctParamPtr idctp)
{
    IdctParamPrivPtr priv = (IdctParamPrivPtr)idctp;

    if (priv->super.arenaQ != NULL) {
	priv->super.arenaQ->next = priv->arenaFree;
	priv->arenaFree = priv->super.arenaQ;
	priv->super.arenaQ = NULL;
    }
    priv->arenaLast = NULL;
}

DctListInfoPtr
AllocDctArena(IdctParamPtr idctp, InternalReasonPtr *pReason)
{
    IdctParamPrivPtr priv = (IdctParamPrivPtr)idctp;
    int allocSize = sizeof(DctListInfo) + DctArenaSize;
    DctListInfoPtr dctList;
    
    if (priv->arenaFree) {
        dctList = priv->arenaFree;
	dctList->curIdctMb = (DctMacroblockPtr)dctList->arena;
        priv->arenaFree = dctList->next;
        return(dctList);
    }
    
    dctList = (DctListInfoPtr) InsecureAlloc(allocSize);
    if (dctList == NULL) {
        *pReason = AllocReason("IDCT arena", allocSize);
        return(NULL);
    }
    memset((char *)dctList, 0, allocSize);
    dctList->arena = (UChar *)(dctList + 1);
    dctList->curIdctMb = (DctMacroblockPtr)dctList->arena;

    return(dctList);
}

/*
 * Note that when the arena is put on the queue, the curIdctMb pointer
 * is reset.  This is so that that the internal dct routines can start at the
 * right place (the beginning) and are re-entrant because the context
 * (the curIdctMb pointer) always holds the current state.
 *
 * Note also that the arenaLast pointer is used to queue the new dctList at
 * at the end.  This is so that the data is queued in macroblock order
 * for the entire picture.  This is important so that the algorithm for
 * updating the pixel pointer can function efficiently.
 */
void
QueueDctArena(IdctParamPtr idctp, DctListInfoPtr dctList)
{
    IdctParamPrivPtr priv = (IdctParamPrivPtr)idctp;

    dctList->next = NULL;
    if (priv->super.arenaQ == NULL) {
        priv->arenaLast = priv->super.arenaQ = dctList;
    } else {
        priv->arenaLast->next = dctList;
        priv->arenaLast = dctList;
    }
    dctList->curIdctMb = (DctMacroblockPtr)(dctList)->arena;
}

void
FreeIdctParam(IdctParamPtr idctp)
{
    DctListInfoPtr dctList, next;
    IdctParamPrivPtr priv = (IdctParamPrivPtr)idctp;
    
    next = NULL;
    for (dctList = priv->arenaFree; dctList; dctList = next) {
        next = dctList->next;
        free((char *)dctList);
    }
    
    next = NULL;
    for (dctList = idctp->arenaQ; dctList; dctList = dctList->next) {
        next = dctList->next;
        free((char *)dctList);
    }
    if (priv->freeIdctParam)
        (*priv->freeIdctParam)(priv);
    if (priv->localPicAlloc)
        free((char *)priv->localPicAlloc);
    
    free((char *)idctp);
}

#if defined(DEBUG_ENABLED)

static char *IDctDebugStrings[] = {
    "motion mb address",
    "motion blk address",
    "intra coeff",
    "intra values",
    "delta coeff",
    "delta values",
    "validate reconstruction",
    "validate dct",
};

FlagMacroList IDctDebugMacros = { NULL, 0 };

FlagNames IDctDebugNames = {
    DBI_IDct,
    IDctDebugStrings,
    NumberOfCells(IDctDebugStrings)
};

#else /* DEBUG_ENABLED */

FlagMacroList IDctDebugMacros = { NULL, 0 };
FlagNames IDctDebugNames = { DBI_IDct, NULL, 0 };

#endif /* DEBUG_ENABLED */
