/*
 * (c) Copyright 1992, 1993, 1994 North Valley Research, Inc.
 * 	All Rights Reserved
 *
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	nvrDct.h - external interface for inverse dct operations
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/nvrIdct.h,v 1.11 1995/11/01 01:37:39 toddb Exp $
 */
#ifndef NVRDCT_H
#define NVRDCT_H

#include "nvrTypes.h"
#include "nvrReason.h"
#include "imageDpy.h"

/*************************************************************************
 *
 * What follows is per-call information for InverseDct()
 *
 *************************************************************************/

/*
 * The following structures are used to store the decoded macroblock and
 * dct-block information.  Note that most things are variable length
 * and packed into the tiniest space possible because we want to avoid
 * expanding encoded data too soon before it goes throught the
 * inverse dct.  With an average of 10 coefficients per block (that's
 * alot), the storage is 4 + 10*2 = 24 bytes versus about 80 bytes for
 * an expanded block where the signs are kept separate from the
 * coefficients; or 128 bytes where the coefficients are expanded to signed
 * 16-bit quantities.
 *
 * The layout of DctMacroblocks are end-to-end in a single memory arena.
 * The idea is that the arena grows to a high water mark and reallocation
 * never happens again.  Also, pointers are unnecessary, saving space.
 *
 *    DctMacroblock
 *       DctBlock
 *          DctCoeff
 *          DctCoeff
 *          ...
 *          DctCoeff
 *       DctBlock
 *          DctCoeff
 *          DctCoeff
 *          ...
 *          DctCoeff
 *       ...
 *    DctMacroblock
 *    ...
 *
 * The first structure, the DctCoeff, stores the level and sign separate
 * so that inverse quantization can be performed on the unsigned value.
 */
typedef union _dctCoeff {
    struct {
	UShort index    :6;	/* 0 - 63 */
	UShort pad      :1;	/* unused */
	UShort negative :1;	/* 0 or 1 */
	UShort level    :8;	/* quantized, unsigned coefficient */
    } b;
    UShort composite;
} DctCoeff, *DctCoeffPtr;

#define GetDctCoeffIndex(_short)	((_short) >> 10)
#define GetDctCoeffSign(_short)		(((_short) >> 8) & 1)
#define GetDctCoeffLevel(_short)	((_short) & 0xff)

#define ComposeDctCoeff(_index, _sign, _level)				\
    (((_index) << 10) | ((_sign) << 8) | (_level))
/*
 * The DctBlock is a variable length structure header for all the coefficients
 * in a single block.  The index of the block is stored so that we don't
 * have to store copied blocks in this list.
 */
typedef union _dctBlockHdr {
    struct {
        short dcCoeff;	/* signed, quantized dc coef */
        UChar nCoeff;	/* number of entries in 'coefList' */
        UChar index;	/* index of this block within macroblock  */
    } s;
    long composite;
} DctBlockHdr, *DctBlockHdrPtr;

typedef struct _dctBlock {
    DctBlockHdr hdr;
    DctCoeff coefList[ 1 ];	/* (longer) quantized, unsigned coefficients */
} DctBlock, *DctBlockPtr;

/*
 * The DctMacroBlock describes all of the context for the macroblock and
 * the blocks within it.  The mbAddr is used to calculate the offset into
 * the Y, Cr and Cb components.  The firstBlk is only the first block
 * in the list.  The length of the structure is discovered by walking
 * through the block list.
 */
typedef union _dctMbFlags {
    struct {
        ULong quant       : 5;	/* quantizer for this macroblock */
        ULong pad         : 11;	/* padding */
        ULong bakMv       : 1;	/* true if backward motion */
        ULong fwdMv       : 1;	/* true if forward motion */
        ULong endOfList   : 1;	/* true just after the last mb in arena */
        ULong intra       : 1;	/* intra-coded: for reconstruction & motion */
	ULong blocksMask  :12;	/* mask of blocks w/ dct in this macroblock */
    } b;
    ULong fastBlocksMask;	/* fast access to b.blocksMask */
    ULong composite;
} DctMbFlags;

/*
 * values that help evaluate the bits above
 */
enum {
    DctMotionBak = 0x8000,
    DctMotionFwd = 0x4000,
    DctMotionFwdBak = DctMotionBak | DctMotionFwd,
    DctMbQuantShift = 32-5
};

/*
 * Indices for DctMacroblock.motion[], below
 */
typedef enum {
    /*
     * These first four index the two pair of x/y values, one forward,
     * one back.  These are used to index DctMacroblockPtr->motion[].
     */
    kIndexFwdX,
    kIndexFwdY,
    kIndexBakX,
    kIndexBakY
} IdctMotionIndex;
    
typedef enum {
    /*
     * In the Mpeg decoder, these two indices correspond with the 
     * four above.
     */
    kIndexFwd = kIndexFwdX>>1,
    kIndexBak = kIndexBakX>>1
} IdctMotionType;

typedef struct _dctMacroblock {
    DctMbFlags flags;
    short motion[ 4 ];		/* motion vectors fwd x, bak y, bak x, bak y */
    DctBlock blk[1];		/* first dct block in this macroblock */
    /* other DctBlocks follow here */
} DctMacroblock, *DctMacroblockPtr;

#define DctCoefSize		2
#define DctBlockHdrSize		(sizeof(DctBlock) - sizeof(DctCoeff))
#define MaxDctCoeffSize		(DctCoefSize*64)
#define AverageDctCoeffSize	(DctCoefSize*8)
#define MaxBlockCount		6		/* Mpeg1-dependent */
#define MacroblockHdrSize	(sizeof(DctMacroblock) - sizeof(DctBlock))

/*
 * The MaxMacroblockSpace defines the maximum possible size required
 * for storing a fully populated set of blocks.  It is used to
 * decide whether there is enough room in the current arena to store
 * the next macroblock.  This also means that at worst, there is
 * a little less than MaxMacroblockSpace wasted space in the arena.
 */
#define MaxMacroblockSpace \
    (MacroblockHdrSize + (DctBlockHdrSize + MaxDctCoeffSize) * MaxBlockCount)
#define AverageMacroblockSpace \
    (MacroblockHdrSize +	\
    (DctBlockHdrSize + AverageDctCoeffSize) * MaxBlockCount)

/*
 * DctArenaSize is the size of each arena.  It should be large enough
 * so that the wasted space at the end of each arena is a small factor.
 * Standard picture sizes are 352x240, 352x288 and 320x240.  The
 * height of these pictures are 15 or 18 macroblocks, and the width are
 * 20 or 22 macroblocks, so an allocation
 * that is about 3 rows of macroblocks seems good.  We use an average
 * macroblock size.
 */
#define DctArenaSize \
    (AverageMacroblockSpace * 3 * 22)

typedef struct _dctListInfo DctListInfo, *DctListInfoPtr;
struct _dctListInfo {
    UChar *arena;
    DctMacroblockPtr curIdctMb;	/* current macroblock storage */
    DctListInfoPtr next;	/* next arena */
};

#define DctCoeffPtrToNextBlock(p) \
    ((DctBlockPtr)(((long)(p) - ((long)(p) & 0x3) + 4)))

#define NextDctCoeff(p) \
    ((DctCoeffPtr)((long)(p) + DctCoefSize))

#define DctBlockPtrToNextMb(p) \
    ((DctMacroblockPtr)(((long)(p) - ((long)(p) & 0x3) + 4)))

#define NeedNewDctArena(dctList, p) \
    ((UChar *)(p) + MaxMacroblockSpace >= (dctList)->arena + DctArenaSize)

#define DctArenaLeft(dctList, p) \
    (((dctList)->arena + DctArenaSize) - ((UChar *)(p) + MaxMacroblockSpace))

/*
 * This structure is specified to the dct methods to describe the picture
 * being decoded.  Each of the blkOffset corresponds to a block address
 * (in MPEG1 there are only 6, in MPEG2 there are 6, 8 or 12).  Each offset
 * is the the number of bytes
 * from the upper left-hand of the picture to the upper left-hand
 * of the target block number.  
 */
enum {
    Dct444Pic    = 0,
    Dct420Pic    = 1,
    Dct422Pic    = 2,
    NDctPicTypes = 3
};

typedef struct _dctPictGeom {
    short mbWidth;
    short mbHeight;
    short nBlocks;		/* number of blocks per macroblock (6-12) */
    UChar picType[ 12 ];	/* which picture type to use; see pic[] below */
    long blkOffset[ 12 ];	/* offset to block pixel address */
    struct {
        short width;		/* width of this picture */
        short height;		/* height of this picture */
    } pic[ NDctPicTypes ];
} DctPictGeom, *DctPictGeomPtr;
    
/*
 * This is the main holding structure for the dct public info.
 */
typedef struct _idctParam IdctParam, *IdctParamPtr;
struct _idctParam {
    DctListInfoPtr arenaQ;	/* queue of dct macroblocks to be processed */
    
    /*
     * this performs the dct.  Do not change! It is set internally.
     */
    void (*inverseDct)(IdctParamPtr);

    /*
     * this function is used to register a new intra- or non-intra
     * quantizer matrix (found in the sequence header).  You must be
     * sure that all of the data in arenaQ has been processed before
     * calling this.
     */
    void (*newQuantMatrices)(IdctParamPtr, UChar *iqm, UChar *niqm);
    
    /*
     * This function is used to set a new width and height for the
     * picture.  It is used at the beginning of a new sequence.
     */
    InternalReasonPtr (*newPictureSize)(IdctParamPtr, DctPictGeomPtr);

    /*
     * This function is used to reinitialize structures for a new picture.
     */
    void (*newPicture)(IdctParamPtr,
                       UChar fullPel[ 2 ],
                       YcImagePtr pTo,
                       YcImagePtr pPast,
                       YcImagePtr pFuture);

    int nextMbAddr;	/* read-only: current mb being transformed */
    /* Note: actual structure is longer */
}; /* typedef: IdctParam, *IdctParamPtr */

/*************************************************************************
 *
 * Opaque type that represent each of the Dct methods
 *
 *************************************************************************/
typedef struct _idctMethod IdctMethod, *IdctMethodPtr;

extern IdctMethodPtr IntegerIdctMethod;
extern IdctMethodPtr FloatLeeIdctMethod;
extern IdctMethodPtr ValidateIdctMethod;
extern IdctMethodPtr NoopIdctMethod;

/*
 * This is used to create the IdctParam structure.
 */
CExternal IdctParamPtr CreateIdctParam(
    IdctMethodPtr,
    InternalReasonPtr *
);

CExternal DctListInfoPtr AllocDctArena(
    IdctParamPtr,
    InternalReasonPtr *
);

CExternal void QueueDctArena(
    IdctParamPtr idctp,
    DctListInfoPtr dctList
);

CExternal void FreeIdctParam(
    IdctParamPtr idctp
);

CExternal InternalReasonPtr EnableIdctValidation(
    IdctParamPtr
);

CExternal InternalReasonPtr DisableIdctValidation(
    IdctParamPtr
);

CExternal void ClearDctArena(
    IdctParamPtr idctp
);

extern char *IdctMotionIndexNames[];

#ifdef DEBUG_ENABLED

void ShowPicXY(YcImagePtr yci, UChar *p);
boolean CheckPicXY(YcImagePtr yci, UChar *p);

#endif

#endif /* NVRDCT_H */

