/*
 * (c) Copyright 1992, 1993, 1994 North Valley Research, Inc.
 * 	All Rights Reserved
 *
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	nvrDctI.h - internal definitions for inverse dct operations
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/nvrIdctI.h,v 1.7 1995/09/17 22:35:39 toddb Exp $
 */
#ifndef NVRDCTI_H
#define NVRDCTI_H

#include "nvrIdct.h"
#include "nvrDebug.h"
#include <math.h>

/*
 * Debug bit indices
 */
typedef enum {
    DCTD_MotionMbAddr = DBI_IDct,
    DCTD_MotionBlkAddr,
    DCTD_IntraCoeff,
    DCTD_IntraValues,
    DCTD_DeltaCoeff,
    DCTD_DeltaValues,
    DCTD_ValidateReconstruction,
    DCTD_ValidateDct
} DebugBitIndices;

#ifdef __MWERKS__
#define M_PI pi
#endif

#define BlocksPer420Mb	6	/* that is, MPEG 1 4:2:0 */
#define BlocksPer422Mb	8	/* that is, MPEG 1 4:2:0 */

/*************************************************************************
 *
 * motion compensation arena.  This is a working structure
 * used to represent the motion compensation for the current macroblock.
 * Once motion compensation is done, this structure represents the
 * pixels that can be added to the dct delta blocks to yield the
 * final pixels.  Note that this can accomodate 420 (6 blocks) or
 * 422 (8 blocks).
 *
 *************************************************************************/
typedef struct _mcSource {
    UChar *pSrc[ NDctPicTypes ]; /* final, resolved motion compensation */
    long offset[ BlocksPer422Mb ];
    long picWidth[ BlocksPer422Mb ];
} McSource, *McSourcePtr;

/*************************************************************************
 *
 * private version of IdctParam
 *
 *************************************************************************/

#define RowStartMacro(_priv, _mb)					\
    ((_priv)->pRowStarts[ (_mb) >> 3 ] & (1 << ((_mb) & 7)))

#define RowStartAccel(_width, _heightShift)				\
    (((_width) << (_heightShift)) - (_width))

typedef struct _idctParamPriv IdctParamPriv, *IdctParamPrivPtr;
struct _idctParamPriv {
    IdctParam	super;		/* structure visible to caller */

    DctListInfoPtr arenaLast;	/* fast ref to last pointer in super.arenaQ */
    DctListInfoPtr arenaFree;	/* free arenas */

    UChar	intraQ[ 64 ];	/* pointer 8x8 block of intra quantizers */
    UChar	nonIntraQ[ 64 ];/* pointer 8x8 block of non-intra quantizers */
    void	(*freeIdctParam)(IdctParamPrivPtr);
    
    /*
     * this performs motion compensation on the specified macroblock.
     */
    void (*motionComp)(IdctParamPrivPtr, DctMacroblockPtr);

    /*
     * the 'pmc' pointer points to the result of the motion compensation
     * (it points to multiSrcMc or singleSrcMc, or mixedSrcMc)
     * and is used by the dct routine for the final step.  There are
     * two possible situations:
     *  1. all of the motion is single source (bak or fwd, not both; and
     *     doesn't have any half-pel motion).  In this case pmc
     *     points to singleSrcMc and all the offsets are permanently
     *     initialized with the correct offsets for within a single
     *     picture.  singleSrcMc.pix must be initialized to the upper-left-
     *     hand corner of the reference picture plus the motion vector offset.
     *  2. all of the motion is multi-source (bak and fwd; or
     *     half-pel motion, or both).  In this case pmc
     *     points to multiSrcMc, multiSrcMc.pix permanently points to
     *     ->mcPix and all values are permanently initialized to the
     *     correct offsets with multiSrcMc.pix.
     */ 
    McSourcePtr pmc;		/* points to the current motion compensation */
    McSource multiSrcMc;	/* space for MC that require "work" */
    McSource singleSrcMc;	/* space for MC that doesn't require "work" */
    McSource lumaSingleSrcMc;	/* space for MC that doesn't require "work" */
    McSource mixedSrcMc;	/* space for MC that doesn't require "work" */
    UChar *mcSrc;		/* used for multiSrcMc and mixedSrcMc */
    
    short lumaY0Block;		/* the block that is upper left of luma */
    short firstChromaBlock;	/* first block of chroma */
    short lastChromaBlock;	/* last block of chroma */

    /*
     * This is a mask of the blocks present in the macroblock.  Those
     * missing must be copied.  It must be valid before the call to
     * IdctMethodPtr->motionComp().
     */
    short	blocksMask[ NDctPicTypes ];
    short	allBlocksMask;	/* value for 'blocksMask' when all present */

    /*
     * If this dct is being validated, then this is a pointer to the data
     * associated with the validation routines.  The functions have already
     * been replaced.
     */
    void	*validate;
    
    /*
     * These pointers and nextMbAddr will change over the course of the
     * IDCT, moving down the picture.
     */
    UChar       *pTo[ NDctPicTypes ]; /* working output picture pointers */
    YcImagePtr	past;		/* for debugging */
    YcImagePtr	current;	/* for debugging */
    YcImagePtr	future;		/* for debugging */

    long        pastOffset;	/* offset to past picture, if defined */
    long        futureOffset;	/* offset to future picture, if defined */
    UChar	*pRowStarts;	/* allocated: 1-bit means an mb row start */
    UChar	*localPicAlloc;	/* pointer to data for rowStarts, mcPix, etc */
    DctPictGeom	geom;		/* picture geometry */
    boolean	fwdFullPel;	/* true if fwd motion is on full-pel */
    boolean	bakFullPel;	/* true if bak motion is on full-pel */
    
    /*
     * This is registered by the subclass
     */
    void	(*newQuantMatrices)(IdctParamPrivPtr, UChar *iqm, UChar *niqm);
}; /* typedef: IdctParamPriv, *IdctParamPrivPtr */

struct _idctMethod {
    int (*privIdctParamSize)(void);
    InternalReasonPtr (*createIdctParam)(IdctParamPtr);
}; /* typedef:IdctMethod, *IdctMethodPtr */

/*************************************************************************
 *
 * private interfaces used for common operations
 *
 *************************************************************************/

CExternal void Idct420MotionCompensate(
    IdctParamPrivPtr,
    DctMacroblockPtr
);

CExternal void Idct422MotionCompensate(
    IdctParamPrivPtr,
    DctMacroblockPtr
);

#endif /* NVRDCTI_H */
