/*
 * Notify routines for applications doing multiple SIGALRM and/or sleeps
 *
 * Global routines in this module are:
 *
 *	add_event(seconds, fnc)
 *		Runs the function "fnc" in seconds. Note - the function is
 *		removed from the event queue after it is run. You must
 *		do another add_event if you wish to schedule it again.
 *		(The add_event can be called from inside fnc if you wish.)
 *	
 *	sleep(seconds)
 *		Pause for seconds.  Identical to the C library version.
 *
 * See sample code at end of this file.
 *
 * Note that sleep is redefined here, replacing the C library version.
 * This sleep along with this notify module work with multiple Alarm events...
 * The C library sleep one doesn't!
 *
 * Implementation note:
 *	The maximum number of concurrent events is controlled 
 *	by the define NEVENTS below.  Feel free to increase it.
 *
 */

#include <stdio.h>
#include <signal.h>

#define DEBUG

#define NIL	0
#define local	static

#define NEVENTS	6

extern char *ctime();

local struct
{
	unsigned long when;
	void (*fnc)();
} callout[NEVENTS];

#ifdef DEBUG
dump_callouts(msg)
char *msg;
{
	register int i;

	printf ("%s:\n", msg);
	for (i = 0; i < NEVENTS; i++)
		if (callout[i].fnc)
			printf ("\t(%2d) %x at %.8s\n",
				i, callout[i].fnc, 
				&((ctime(&callout[i].when))[11]));
}

show_event(msg, num)
char *msg;
{
	char s_when[32], s_now[32];
	long now;

	time (&now);

	sprintf (s_when, "%.8s", &((ctime(&callout[num].when))[11]));
	sprintf (s_now, "%.8s", &((ctime(&now))[11]));
	printf ("%10s: event #%2d %08x in %3ld sec. (%.8s) [now %.8s]\n",
		msg,
		num,
		callout[num].fnc,
		callout[num].when - now,
		s_when,
		s_now);
}
#endif

local void
run_event()
{
	long now;

	alarm (0);

	time (&now);

	while (callout[0].fnc && callout[0].when <= now)
	{
		register int i;
#ifdef DEBUG
		show_event("RUN", 0);
#endif
		(*(callout[0].fnc))();
		for (i = 0; i < (NEVENTS-1); i++)
		{
			callout[i].fnc = callout[i+1].fnc;
			callout[i].when = callout[i+1].when;
		}
		callout[NEVENTS-1].fnc = NIL;
		callout[NEVENTS-1].when = 0;

		time (&now);
	}
#ifdef DEBUG
	dump_callouts("After run_event");
#endif
	sched_event();
}

add_event(seconds, fnc)
unsigned int seconds;
void (*fnc)();
{
	register int i, j;
	int rvalue;
	long now, when;

	alarm (0);

	time (&now);

	when = now + seconds;		/* When event should be executed */

	rvalue = -1;
	for (i = 0; i < NEVENTS; i++)
	{
		if (callout[i].when > when || callout[i].fnc == NIL)
		{
			for (j = NEVENTS-2; j >= i; j--)
			{
				callout[j+1].fnc = callout[j].fnc;
				callout[j+1].when = callout[j].when;
			}
			callout[i].when = when;
			callout[i].fnc = fnc;
#ifdef DEBUG
			show_event("ADD", i);
#endif
			rvalue = 0;
			break;
		}
	}

#ifdef DEBUG
	dump_callouts("After add_event");
#endif

	sched_event();

	return (rvalue);
}

/*
 * Schedule first event (event 0) in callout list.
 */
local
sched_event()
{
	if (callout[0].fnc)
	{
		long remaining, now;

		time (&now);

		remaining = callout[0].when - now;

		if (remaining <= 0)
			remaining = 1;
#ifdef DEBUG
		/*show_event("SCHED", 0);*/
#endif
		signal (SIGALRM, run_event);
		alarm ((unsigned) remaining);
	}
}

/*******************************************
 * Code to implement a sleep() that works. *
 *******************************************/

local int sleeping;

local void
sleep_catch()
{
	sleeping = 0;
}

/*
 * A version of sleep that works.	
 *						SLEEP redefined here.
 */
sleep(seconds)
unsigned int seconds;
{
	sleeping = 1;
	add_event (seconds, sleep_catch);
	while (sleeping)
		pause();
}
/*******************************************
 * End of code to implement sleep()        *
 *******************************************/

#ifdef DEBUG

/*
 * Sample code...
 */

void
dummy() 
{
	printf ("Hello from dummy\n");
	
	/*
	 * Add another add_event here if you want to re-sched.
	 */
	/* add_event(60, dummy);	*/
}

main()
{
	setlinebuf(stdout);
	dump_callouts("Start");
	add_event(20, dummy);
	add_event(30, dummy);
	add_event(5, dummy);
	sleep(10);
	add_event(8, dummy);
	sleep(60);
}
#endif
