/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	nvridx.c
 *
 * DESCRIPTION
 *	Deal with nvr_idx files - raw yc files with
 *		variable length ascii headers
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/nvridx.c,v 1.17 1995/09/05 18:45:54 russ Exp $
 */

#define NeedFunctionPrototypes 1

#include "nvrImageI.h"
#include "nvrString.h"
#include "osio.h"
#include "fileAccess.h"
#include "nvrAlloc.h"
#include "process.h"

#include <ctype.h>

static InternalReason LineTooLong =
    { "nvr_idx line is longer than 1024.", IR_BadValue };
static InternalReason NoMagicNumber =
    { "The first line of an nvridx file must contain only 'nvr_idx'\n",
      IR_BadValue };

static char *MagicNumber = "nvr_idx";

static InternalReasonPtr
ReadNvrIdxFile(char *path, Argvec *argv)
{
    InternalReasonPtr reason;
    int c, commChar;
    char buffer[ 1024 ], *p;
    int line = 1;
    BufFileHndl fp;

    fp = OpenBufferedFileForReading(path);
    if (fp == NULL)
	return( CreateSysReason("can't open %s\n", path) );

    commChar = 0;
    reason = NULL;
    /*
     * Break the file up into an argument vector.
     * All spaces are removed.
     * All words after a '#' and before a '\n' are ignored
     */
    p = buffer;
    while (reason == NULL) {
	c = GetBufferedChar(fp);
        if (c == EndOfFile)
	    goto endOfLine;
	if (p - buffer >= sizeof(buffer) - 1) {
	    if (line == 1)
		reason = &NoMagicNumber;
	    else
	        reason = &LineTooLong;
	    break;
	}
	
	switch(c) {
	case ' ':
	case '\t':
	    break;

	case '\n':
	case '\r':	/* for the mac */
	endOfLine:
	    if(p != buffer) {
		*p = '\0';
		AddArgv(argv, buffer, True);
		if (line++ == 1 && strcmp(buffer, MagicNumber) != 0)
		    reason = &NoMagicNumber;
	    }
	    commChar = 0;
	    p = buffer;
	    if (c == EndOfFile)
		goto out;
	    break;

	case '#':
	    commChar++;
	    break;

	default:
	    if (!commChar)
		*p++ = c;
	    break;
	}
    }

out:
    CloseBufferedFile(fp);
    return(reason);
}


/*
** Node has already been filled out by IoParse().
** We must generate a new StringArgListPtr by paring the
** file associated with node->op.u.io and free the old
** StringArgListPtr before returning the new one.
*/
InternalReasonPtr
NvrIdxParse(PipelineNodePtr node, StringArgListPtr *args)
{
    ImageIOParamPtr ioParam = &node->op.u.io;
    InternalReasonPtr reason = NULL;
    StringArgListPtr idxArgs = NULL;
    StringArgListPtr newArgs = NULL;
    Argvec argv;

    if( ioParam->u.type != FLT_Single ) {
	return(NULL);
    }

    argv.length = 0;
    if (reason = ReadNvrIdxFile(ioParam->u.single.path, &argv)) {
	if (reason == &NoMagicNumber)
	    reason = NULL;
	goto done;
    }

    if( (idxArgs = AllocStringArgList(argv.index, argv.argv)) == NULL ) {
	reason = &IRBadAllocation;
    } else if( (newArgs = CombineStringArgLists(*args,idxArgs)) == NULL ) {
	reason = &IRInternalProblem;
    }

    if( idxArgs ) {
	FreeStringArgList(idxArgs);
    }
    if (reason) {
	if( newArgs ) {
	    FreeStringArgList(newArgs);
	}
	goto done;
    }

    /*
     * Extract the absolute or relative path prefix before
     * it gets overwritten by the next pass of the IO parser...
     * and before it gets freed...
     */
    ioParam->fullPathPrefix = ExtractPathPrefix(ioParam->u.single.path);
	
    if( newArgs ) {
	FreeStringArgList(*args);
	*args = newArgs;
    }

    /*
     * Reset appropriate parts of the ioParam structure so that 
     * parsing can happen a second time.
     */
    ioParam->name = ImageIOUnspecified;
    memset((char *)&ioParam->spec, 0 ,sizeof(ioParam->spec));
    memset((char *)&ioParam->u, 0 ,sizeof(ioParam->u));
    
done:
    FreeArgv(&argv);
    return(reason);
}

InternalReasonPtr
NvrIdxHdr(char *path, ImageFileInfoPtr imagePtr)
{
    int i, width, height, name;
    ImageFileType ift;
    InternalReasonPtr reason = NULL;
    StringArgListPtr args = NULL;
    float fps;
    InterlaceLayoutType iLayout;
    Argvec argv;

    argv.length = 0;
    if (reason = ReadNvrIdxFile(path, &argv))
	goto out;

    if( (args = AllocStringArgList(argv.index, argv.argv)) == NULL ) {
	reason = &IRBadAllocation;
	goto out;
    }
    imagePtr->defined = 0;
    for (i = args->index; reason == NULL && i < args->length; i++) {
	name = NameToEnum(args->list[i].name, FileParamNames);

	switch (name) {
	case IFP_Type:
	case IFP_ColorSpace:
	case IFP_Sample:
	case IFP_Size:
	case IFP_FrameCount:
	case IFP_FramesPerSec:
	case IFP_Interlace:
	    if (args->list[i].value == NULL) {
		reason = CreateReason("missing argument for \"%s\"\n",
		    args->list[i].name);
		continue;
	    }
	    break;
	
	default:
	    continue;	/* ignore all other commands */
	}

	switch (name) {
	case IFP_Type:
	    /*
	     * Note that the imagePtr->type does not change. from IFT_NvrIdx
	     * to the type listed here.  However, it does get checked for
	     * correctness.
	     */
	    ift = NameToEnum(args->list[i].value, ImageFileTypes);
	    if (ift == IFT_Unspecified) {
		reason = CreateReason("unknown file type in nvr_idx file: %s\n",
		    args->list[i].value);
	    }
	    break;

	case IFP_ColorSpace:
	    imagePtr->color = NameToEnum(args->list[i].value, ColorSpaceNames);
	    if (imagePtr->color == ColorUnknown) {
		reason = CreateReason("unknown color space: %s\n",
		    args->list[i].value);
	    }
	    break;
	    
	case IFP_Sample:
	    imagePtr->sample = NameToEnum(args->list[i].value, SampleTypeNames);
	    if (imagePtr->sample == SampleUnknown) {
		reason = CreateReason("unknown sampling: %s\n",
		    args->list[i].value);
	    }
	    break;

	case IFP_Size:
	    if (sscanf(args->list[i].value, "%dx%d", &width, &height) != 2
		&& sscanf(args->list[i].value, "%dX%d", &width, &height) != 2) {
                reason = CreateReason("can't parse image size: %s\n",
                                      args->list[i].value);
		break;
	    }
	    imagePtr->width = width;
	    imagePtr->height = height;
	    break;

	case IFP_FrameCount:
	    if (sscanf(args->list[i].value, "%d", &width) != 1) {
                reason = CreateReason("bad frame count value: %s\n",
                                      args->list[i].value);
		break;
	    }
	    imagePtr->frameCount = width;
	    imagePtr->defined |= IFI_FrameCount;
	    break;

	case IFP_FramesPerSec:
	    if (sscanf(args->list[i].value, "%f", &fps) != 1) {
                reason = CreateReason("bad fps value: %s\n",
                                      args->list[i].value);
		break;
	    }
	    imagePtr->framesPerSecond = fps;
	    imagePtr->defined |= IFI_FrameRate;
	    break;
	
	case IFP_Interlace:
	    iLayout = NameToEnum(args->list[i].value, InterlaceLayoutNames);
            if (iLayout == UnknownInterlace) {
                reason = CreateReason("unknown interlace type: %s\n",
                                      args->list[i].value);
            }
	    imagePtr->interlaced = iLayout;

	    break;

	default:
	    break;
	}
    }
    
out:
    FreeArgv(&argv);
    if( args ) {
	FreeStringArgList(args);
    }
    return(reason);
}
