/*
 *	Copyright (c) 1992, North Valley Research, Inc.
 *	Confidential -- All proprietary rights reserved.
 *	Licensed Material -- Property of North Valley Research, Inc.
 *
 *	This program is the property of North Valley Research, Inc. or others
 *	from whom North Valley Research has obtained a licensing right, and is
 *	considered by North Valley Research to be confidential.  It is protected
 *	by U.S. copyright law as an unpublished work and is
 *	furnished pursuant to a written license agreement.  It may
 *	not be used, copied or otherwise reproduced, or disclosed to
 *	others except in accordance with the terms and conditions of
 *	that agreement.
 *
 *	NAME
 *		viewcmap.c
 *
 *	DESCRIPTION
 *		Routines to allocate colorcells and storing colors into
 *		for support of the Software MPEG.  This package will
 *		create a new colormap when necessary.
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/viewcmap.c,v 1.22 1994/10/02 17:33:02 toddb Exp $
 */

/*
 *      EXTERNAL INCLUDES
 *              Include files that must be exported to any package or
 *              program using this package.
 */
#include "nvrAlloc.h"
#include "nvrString.h"
#include "viewcmap.h"

/*
 *      INTERNAL INCLUDES
 *              Include files for local use only, therefore, NOT exported
 *		to any package or program using this package.
 */
#include <X11/Xutil.h>
#include <X11/Xcms.h>
#include <math.h>


/*
 *      EXTERNS
 *              External declarations required locally to this package
 *              that are not already declared in any of the included header
 *		files (external includes or internal includes).
 */


/*
 *      GLOBALS
 *              Variables declared in this package that are allowed
 *		to be used globally.
 */


/*
 *      LOCAL DEFINES
 *		#define declarations local to this package.
 */

/*
 * the maximimum value for a primary in X
 */
#define XRGB_MAX	0xffff

/*
 * Brightness is defined as a signed offset.
 * The input brightness from the user is an offset from BLACK_LEVEL,
 * scaled to fall proportionally between BLACK_LEVEL and WHITE_LEVEL.
 * The intent is for DEFAULT_BRIGHTNESS to fall on BLACK_LEVEL.
 */
#define BLACK_LEVEL		1000	/* below this, gamma goes wacko! */
#define WHITE_LEVEL		XRGB_MAX

#define REAL_BRIGHTNESS(v)						\
  (									\
    (									\
      (									\
        (								\
	  (v) > MAX_BRIGHTNESS						\
            ? MAX_BRIGHTNESS						\
	    : ((v) < MIN_BRIGHTNESS ? MIN_BRIGHTNESS : (v))		\
        ) * NORMAL_DYNAMIC_RANGE	/* scale it */			\
      ) / 200.0		/* divide by 100 to get the percentage result */\
    ) + BLACK_LEVEL	/* finally add the offset to the base */	\
  )

/*
 * Contrast is defined as a percentage of the dynamic range left over
 * after brightness has been added in, although the percentage may be more
 * than 100%.  The input contrast from the user is an offset from
 * BASE_CONTRAST, scaled to fall proportionally so that MIN_CONTRAST
 * corresponds to BASE_CONTRAST and DEFAULT_CONTRAST corresponds to
 * CENTER_CONTRAST.  The intent is for the DEFAULT_CONTRAST to be 100%.
 * Note that above the default contrast it increases as the square
 * of the value, up to 2*2*MAX_CONTRAST
 */
#define CENTER_CONTRAST		100
#define BASE_CONTRAST		50
#define	CONTRAST_DIV		(MAX_CONTRAST/1.0)
#define TOP_CONTRAST(v)		((1 + (v) / CONTRAST_DIV)		\
			       * (1 + (v) / CONTRAST_DIV)		\
			       * (v))

#define REAL_CONTRAST(v)						\
  (									\
    (									\
      (									\
        (								\
	  (v) < MIN_CONTRAST						\
	    ? MIN_CONTRAST						\
	    : (v) <= DEFAULT_CONTRAST					\
	      ? (v)							\
	      : (v) > MAX_CONTRAST					\
		? TOP_CONTRAST(MAX_CONTRAST)				\
		: TOP_CONTRAST(v)					\
        ) * (CENTER_CONTRAST - BASE_CONTRAST)	/* scale it */		\
      ) / (100.0/2.0)	/* convert to percentage */			\
    ) + CENTER_CONTRAST	/* finally, add to offset to the center */	\
  )

/*
 * After an input level is calculated, it must be clamped to fall between
 * BLACK_LEVEL and WHITE_LEVEL, inclusive.
 */
#define CLAMP_LEVEL(v)							\
  (									\
    (v) > WHITE_LEVEL							\
      ? WHITE_LEVEL							\
      : ((v) < BLACK_LEVEL ? BLACK_LEVEL : (v))				\
  )

/*
 * These two macros takes the brightness and contrast and an input level and
 * calculates what the output level should be, including clamping out-
 * of-range values.  Note that the center of the dynamic range for
 * a given contrast setting should fall in the center of the difference
 * between REAL_BRIGHTNESS(v) and WHITE_LEVEL.
 */
#define NORMAL_DYNAMIC_RANGE	(WHITE_LEVEL - BLACK_LEVEL)

/*
 * The dynamic range is scaled by the contrast setting to a new range,
 * smaller, equal, or larger than the normal dynamic range.
 */
#define DYNAMIC_RANGE(contrast)						\
  ((NORMAL_DYNAMIC_RANGE * (contrast)) / 100.0)

/*
 * The number of steps depends on the number of bits allocated in the
 * colormap to the color (or luma).  This gives the value of each step
 */
#define CONTRAST_STEP_SIZE(contrast, nSteps)				\
  (DYNAMIC_RANGE(contrast) / (double)(nSteps))

/*
 * This gives the base of the dynamic range such that the center of
 * the range is the brightness level plus half the normal dynamic range
 */
#define BASE_OF_DYNAMIC_RANGE(bright, contrast)				\
  ((bright) + (NORMAL_DYNAMIC_RANGE/2.0) - (DYNAMIC_RANGE(contrast)/2.0))

/*
 * This will calculate the level and clamp it.  'base' must be that returned
 * by BASE_OF_DYNAMIC_RANGE().  stepSize must be that returned
 * by CONTRAST_STEP_SIZE().
 */
#define CALC_LEVEL(base, stepSize, inputLevel)				\
    CLAMP_LEVEL((base) + (inputLevel) * stepSize)

#ifndef MIN
#define MIN(x,y) ((x) > (y) ? (y) : (x))
#endif /* MIN */
#ifndef MAX
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#endif /* MAX */

#ifdef USE_X11R4

#define GAMMA_COMPUTE(g,x)					\
    (unsigned short)(pow((double)(x)/65535.0,1.0/(g))*65535)

#define RED_CELL(colorDefs, i)		(colorDefs[i].red)
#define GREEN_CELL(colorDefs, i)	(colorDefs[i].green)
#define BLUE_CELL(colorDefs, i)		(colorDefs[i].blue)

#define MARK_ALL_PRIMARIES(colorDefs, i)				\
    (colorDefs[i].flags = DoRed | DoGreen | DoBlue)

#else

#define GAMMA_COMPUTE(g,x) (pow((double)(x)/65535.0,1.0/(g)))

#define RED_CELL(colorDefs, i)		(colorDefs[i].spec.RGBi.red)
#define GREEN_CELL(colorDefs, i)	(colorDefs[i].spec.RGBi.green)
#define BLUE_CELL(colorDefs, i)		(colorDefs[i].spec.RGBi.blue)

#define MARK_ALL_PRIMARIES(colorDefs, i)				\
    (colorDefs[i].format = XcmsRGBiFormat)

#endif /* USE_X11R4 */

#define GAMMA_CORRECT(g,x) (irint(pow((double)(x)/65535.0,1.0/(g))*65535))
#define GAMMA_SHIFT	(16-GAMMA_TBL_BITS)
#define GAMMA_MASK	(GAMMA_TBL_MAX<<GAMMA_SHIFT)

#define	GAMMA_INDEX(base, inputLevel, stepSize)				\
    (((UInt)CALC_LEVEL(base, stepSize, inputLevel)) >> GAMMA_SHIFT)

#define MAX_COLOR_CELLS		4096

#define MarkAllAllocated(cmapInfo)					\
    {									\
	int i;								\
	for (i = 0; i < MAX_COLOR_CELLS/32; i++)			\
	    cmapInfo->allocated[i] = ~0;				\
    }

#define MarkNoneAllocated(cmapInfo)					\
    {									\
	int i;								\
	for (i = 0; i < MAX_COLOR_CELLS/32; i++)			\
	    cmapInfo->allocated[i] = 0;					\
    }

#define PixelAllocated(cmapInfo, pixel)					\
    (cmapInfo->allocated[pixel >> 5] & (1 << (pixel&0x1f)))

#define MarkAllocated(cmapInfo, pixel)					\
    cmapInfo->allocated[pixel >> 5] |= (1 << (pixel&0x1f))

/*
 *      LOCAL TYPEDEFS
 *              typedefs local to this package (for use with local vars).
 *
 */

typedef Status (*StoreColorFuncType)(Display *,
				     Colormap,
				     ColorDefType *,
				     int nColors,
				     Bool *compArrayReturn);

/*
 *      LOCALS VARIABLES
 *		Variables local to this package.
 */



/************************************************************************
 *									*
 *			PRIVATE ROUTINES				*
 *									*
 *	Routines private to this package therefore declared static	*
 *									*
 ************************************************************************/


#ifdef USE_X11R4

static void
CreateGammaTbl(ViewCmapInfoPtr cmapInfo)
{
    unsigned short *tbl = cmapInfo->gamma_tbl;
    int i;

    for (i = 0; i < GAMMA_TBL_SIZE; i++) {
        *tbl++ = irint(
		    pow(
			((double)i / GAMMA_TBL_MAX * XRGB_MAX)/65535.0,
			1.0 / cmapInfo->dpyGamma)
		    * 65535.0);
    }
}

#else /* USE_X11R4 */

static void
CreateGammaTbl(ViewCmapInfoPtr cmapInfo)
{
    double *tbl = cmapInfo->gamma_tbl;
    int i;

    for (i = 0; i < GAMMA_TBL_SIZE; i++) {
	*tbl++ = pow( ((double)i / GAMMA_TBL_MAX * XRGB_MAX) /65535.0,
		2.2 / cmapInfo->dpyGamma);
    }
}

#ifdef DEBUG_ENABLED
ShowColorCells(ViewCmapInfoPtr cmapInfo, ColorDefType *colorDefs, int nColors)
{
    int i;

    Msg("vis 0x%x, class %d:\n",
	cmapInfo->visual->visualid,
	cmapInfo->visual->class);
    for (i = 0; i < nColors; i++) {
	NormalOutput("%3d: (%.3f,%.3f,%.3f)",
	    colorDefs[i].pixel,
	    colorDefs[i].spec.RGBi.red,
	    colorDefs[i].spec.RGBi.green,
	    colorDefs[i].spec.RGBi.blue);
	if (i & 1)
	    Msg("\n");
	else
	    Msg(" ");
    }
}
#endif /* DEBUG_ENABLED */

#endif /* USE_X11R4 */


static void
CopyColorsToDefaultCmap(ViewCmapInfoPtr cmapInfo,
			ColorDefType *colorDefs,
			int nColors)
{
    Display	*dpy;
    Colormap	defaultCmap;
    int		success, nPixels, pixel, color, copyLength;
    ULong	pixels[ MAX_COLOR_CELLS ], *pPixels;
    ULong	tmpAllocated[ MAX_COLOR_CELLS/32 ]; /* 1 bit/pixel */
    ColorDefType *pFreeColors, *pAllocedColors;
    StoreColorFuncType storeColors = XStoreColorsFunc;

    /*
     * Here we keep a running guess on the number of cells that
     * can be allocated from the default colormap.
     * It grows and shrinks with each call, just as the default colormap
     * changes.
     */
    static int	nDefaultAllocated = 256;

    dpy = cmapInfo->dpy;
    defaultCmap = DefaultColormap(dpy, cmapInfo->scrNumber);
    if (defaultCmap == cmapInfo->cmap)
	return;

    /*
     * Finally, if this system only has one installed colormap,
     * allocate as many cells in the default colormap, and copy
     * the corresponding colormap entries from our private colormap
     * into the default colormap.  This should help make the picture
     * viewable when the colormap is not installed.
     */
    nPixels = nDefaultAllocated;
    pPixels = pixels;
    for (;;) {
	success = XAllocColorCells(dpy, defaultCmap, True,
				   NULL,	/* no masks returned */
				   0,		/* no planes */
				   pPixels,
				   nPixels);
	if (success) {
	    pPixels += nPixels;
	    nPixels = 1;	/* try to allocate one more */
	} else {
	    if (pPixels != pixels)
		break;		/* we have pixels, and can't get any more */
	    if (--nPixels <= 0)
		break;		/* can't seem to get any pixels at all */
	}
    }

    if ((nPixels = pPixels - pixels) == 0) {
	nDefaultAllocated = 1;	/* start somewhere reasonable next time */
	return;
    }
    nDefaultAllocated = nPixels;
    
    cmapInfo->allocated	= tmpAllocated;
    MarkNoneAllocated(cmapInfo);
    for (pixel = 0; pixel < nPixels; pixel++) {
	MarkAllocated(cmapInfo, pixels[pixel]);
    }

    pFreeColors = &colorDefs[ 0 ];
    for (color = 0; color < nColors;) {
	/*
	 * skip over this chunk of unallocated pixels
	 */
	while (color < nColors
	 && ! PixelAllocated(cmapInfo, colorDefs[color].pixel))
	    color++;

	/*
	 * find the end of this chunk of allocated colors
	 */
	pAllocedColors = &colorDefs[ color ];
	while (color < nColors
	 && PixelAllocated(cmapInfo, colorDefs[color].pixel))
	    color++;

	copyLength = UCharP(&colorDefs[ color ]) - UCharP(pAllocedColors);
	if (copyLength) {
	    memcpy((char *)pFreeColors, (char *)pAllocedColors, copyLength);
	    pFreeColors = (ColorDefType *)(UCharP(pFreeColors) + copyLength);
	}
    }

    nColors = pFreeColors - colorDefs;
    if (! (*storeColors)(dpy, defaultCmap, colorDefs, nColors, NULL)) {
	ProgMsg("Cannot store %d colors in colormap 0x%x\n",
	    nColors, cmapInfo->cmap);
    }

    XFreeColors(dpy,
		defaultCmap,
		pixels,		/* list of pixels */
		nPixels,	/* number of pixels to free */
		0);		/* no planes */
}

static void
LoadColors(ViewCmapInfoPtr cmapInfo)
{
    long		i, index;
    long		nSteps, oneBits;
    int			redColors, greenColors, blueColors;
    unsigned long	pixel;
    double 		yStep;
    double		bright = cmapInfo->privateBrightness;
    double		contrast = cmapInfo->privateContrast;
    double		base = BASE_OF_DYNAMIC_RANGE(bright, contrast);
    ColorDefType	colorDefs[ MAX_COLOR_CELLS ];
    StoreColorFuncType	storeColors = XStoreColorsFunc;

    switch (cmapInfo->depthType) {
      case Rgb444:
	if (cmapInfo->visual->class == PseudoColor) {
	    nSteps = 4095;
	    oneBits = 0;
	    redColors = greenColors = blueColors = 16;
	    goto pseudoColor;
	}
	nSteps = 15;
	goto directColor;

      case Rgb888:
	/*
	 * Create a color ramp
	 */
	nSteps = 255;

      directColor:
	yStep = CONTRAST_STEP_SIZE(contrast, nSteps);

	for (i = 0; i <= nSteps; i++) {
	    index = GAMMA_INDEX(base, i, yStep);
	    RED_CELL(colorDefs, i) = 
	    GREEN_CELL(colorDefs, i) = 
	    BLUE_CELL(colorDefs, i) =
		cmapInfo->gamma_tbl[ index ];
	    /*
	     * XXX gross assumption
	     */
	    colorDefs[i].pixel = (i << 16) | (i << 8) | (i << 0);
	    MARK_ALL_PRIMARIES(colorDefs, i);
	}
	break;

      case GrayScale7:
	nSteps = 127;
	oneBits = 0x80;
	goto grayScale;

      case GrayScale8:
	nSteps = 255;
	oneBits = 0;
      grayScale:
	/*
	 * Create a gray ramp
	 */
	yStep = CONTRAST_STEP_SIZE(contrast, nSteps);

	for (i = 0; i <= nSteps; i++) {
	    index = GAMMA_INDEX(base, i, yStep);
	    RED_CELL(colorDefs, i) = 
	    GREEN_CELL(colorDefs, i) = 
	    BLUE_CELL(colorDefs, i) =
		cmapInfo->gamma_tbl[ index ];
	    colorDefs[i].pixel = oneBits | i;
	    MARK_ALL_PRIMARIES(colorDefs, i);
	}
	break;

      case RgbC666:
	oneBits = 0;
	nSteps = 6*6*6 - 1;
	greenColors = redColors = blueColors = 6;
	goto pseudoColor;

      case RgbC555:
	oneBits = 0x80;
	nSteps = 5*5*5 - 1;
	greenColors = redColors = blueColors = 5;
	goto pseudoColor;

      case Rgb232:
	nSteps = 127;
	oneBits = 0x80;
	greenColors = 8;
	redColors = blueColors = 4;
	goto pseudoColor;

      case Rgb332:
	nSteps = 255;
	oneBits = 0;
	redColors = greenColors = 8;
	blueColors = 4;

      pseudoColor:
	{
	    int red, green, blue;
	    float redStep, greenStep, blueStep;

	    redStep   = CONTRAST_STEP_SIZE(contrast, redColors-1);
	    greenStep = CONTRAST_STEP_SIZE(contrast, greenColors-1);
	    blueStep  = CONTRAST_STEP_SIZE(contrast, blueColors-1);

	    /*
	     * A gamma-corrected color ramp for 2:3:2 or 3:3:2
	     */
	    for (red = 0; red < redColors; red++) {
	      for (green = 0; green < greenColors; green++) {
		for (blue = 0; blue < blueColors; blue++) {
		    i = red * (greenColors*blueColors)
		      + green * blueColors
		      + blue;

		    index = GAMMA_INDEX(base, red, redStep);
		    RED_CELL(colorDefs, i)   = cmapInfo->gamma_tbl[ index ];

		    index = GAMMA_INDEX(base, green, greenStep);
		    GREEN_CELL(colorDefs, i) = cmapInfo->gamma_tbl[ index ];

		    index = GAMMA_INDEX(base, blue, blueStep);
		    BLUE_CELL(colorDefs, i)  = cmapInfo->gamma_tbl[ index ];

		    pixel = i | oneBits;
		    colorDefs[i].pixel = pixel;
		    MARK_ALL_PRIMARIES(colorDefs, i);
		}
	      }
	    }
	}
	break;

    default:
	return;
    }

    /*
     * Allocate the X colors for this picture
     */
    if (! (*storeColors)(cmapInfo->dpy, cmapInfo->cmap, colorDefs,
	    nSteps+1, NULL)) {
	ProgMsg("Cannot store %d colors in colormap 0x%x\n",
	    nSteps+1, cmapInfo->cmap);
    }

    CopyColorsToDefaultCmap(cmapInfo, colorDefs, nSteps+1);
}

/*
 * Decide whether to use the default colormap or create a new one.
 */
static boolean
AllocViewColormap(ViewCmapInfoPtr cmapInfo)
{
    Display	*dpy = cmapInfo->dpy;
    Window	root = RootWindow(dpy, cmapInfo->scrNumber);
    Colormap	cmap;

    cmapInfo->cmapWritable = True;	/* this is what we'd like */
    switch (cmapInfo->depthType) {
      case Rgb444:
      case Rgb888:
	if (cmapInfo->visual->class == TrueColor) {
	    cmapInfo->cmapWritable = False;
	    cmap = XCreateColormap(dpy, root, cmapInfo->visual, AllocNone);
	} else if (cmapInfo->visual->class == DirectColor
		|| cmapInfo->visual->class == PseudoColor) {
	    cmap = XCreateColormap(dpy, root, cmapInfo->visual, AllocAll);
	    MarkAllAllocated(cmapInfo);
	} else
	    return (False);	/* don't know how to do any other */
	break;

      case GrayScale1:
	if (cmapInfo->visual->class != StaticGray)
	    return (False);
	cmapInfo->cmapWritable = False;
	cmap = XCreateColormap(dpy, root, cmapInfo->visual, AllocNone);
	break;

      case GrayScale8:
	if (cmapInfo->visual->class != StaticGray)
	    goto writableGray;
	cmapInfo->cmapWritable = False;
	cmap = XCreateColormap(dpy, root, cmapInfo->visual, AllocNone);
	break;

      case GrayScale7:
      writableGray:
	if (cmapInfo->visual->class != PseudoColor
	 && cmapInfo->visual->class != GrayScale)
	    return (False);
	cmap = XCreateColormap(dpy, root, cmapInfo->visual, AllocNone);
	break;

      case Rgb332:
      case Rgb232:
      case RgbC555:
      case RgbC666:
	if (cmapInfo->visual->class != PseudoColor)
	    return (False);
	cmap = XCreateColormap(dpy, root, cmapInfo->visual, AllocNone);
	break;

      default:
	return (False);
    }

    cmapInfo->cmap = cmap;
    return(True);
}

/*
 * Allocate the X colors for this view
 */
static boolean
AllocViewColorcells(ViewCmapInfoPtr cmapInfo)
{
    int		nPlanes, nColors;
    ULong	planeMasks[ 8 ], allPlanes;
    Colormap 	cmap = cmapInfo->cmap;
    Colormap	defaultCmap;
    Display	*dpy = cmapInfo->dpy;
    int		success, i, pixel, nPixels;
    ULong	pixels[ 256 ];

    switch (cmapInfo->depthType) {
      default:
	return(True);	/* its either read-only or alrealy fully allocated */
      case GrayScale8:
      case Rgb332:
	nPlanes = 8;
	nColors = 256;
	break;
      case RgbC666:
	nPlanes = 8;
	nColors = 6*6*6;
	break;
      case RgbC555:
	nPlanes = 7;
	nColors = 5*5*5;
	break;
      case GrayScale7:
      case Rgb232:
	nColors = 128;
	nPlanes = 7;
	break;
    }

    if (cmapInfo->cmapShared) {
	/*
	 * handle 7 bits in an 8-bit colormap
	 */
	memset((char *)planeMasks, 0, nPlanes * sizeof(planeMasks[0]));
	success = XAllocColorCells(dpy, cmap, True,
				   planeMasks, nPlanes, pixels,
				   1 /* one pixel */);
	if (! success || pixels[0] != 0x80)
	    return(False);	/* something really wrong */
	/*
	 * Check the planes for paranoia's sake
	 */
	allPlanes = 0;
	for (i = 0; i < nPlanes; i++)
	    allPlanes |= planeMasks[i];

	if (allPlanes != (1<<nPlanes)-1) {
	    Msg("Plane masks are 0x%x, not 0x%x\n",
		allPlanes, (1<<nPlanes)-1);
	    return(False);	/* something really wrong */
	}
	return(True);
    }
    /*
     * We always allocate the entire map, or at least as much as we can.
     * You'd think that if you allocated a private colormap with 256 entries,
     * you could allocate all the entries.  On the SGI, the second pseudo-color
     * visual only gives you 255.
     */
    for (nPixels = 256; nPixels >= 1; nPixels--) {
	success = XAllocColorCells(dpy, cmap, True,
				   planeMasks,
				   0,	/* no planes */
				   pixels,
				   nPixels);
	if (success)
	    break;
    }

    /*
     * Note the ones allocated
     */
    for (i = 0; i < nPixels; i++) {
	MarkAllocated(cmapInfo, pixels[i]);
    }

    if (nColors < 256) {
	XColor	tmpColor[256];
	/*
	 * NOTE
	 *    We have a non-shared colormap... copy the
	 *    first (256-nColors) colorcells from the default colormap into
	 *    the new one.  Then we'll use the upper cells for this
	 *    program.  Use of any of the upper pixel values
	 *    by other programs will cause the "technicolor" effect
	 *    in that program.  If this program is using any of
	 *    the upper cells for things like window background or
	 *    border colors, or GC foreground and background
	 *    colors, then those things will go "technicolor as
	 *    well!!!
	 */
	for (pixel = nPixels = 0; pixel < 256-nColors; pixel++) {
	    if (! PixelAllocated(cmapInfo, pixel))
		continue;
	    tmpColor[nPixels++].pixel = pixel;
	}
	defaultCmap = DefaultColormap(dpy, cmapInfo->scrNumber);
	XQueryColors(dpy, defaultCmap, tmpColor, nPixels);
	XStoreColors(dpy, cmap, tmpColor, nPixels);
    }

    return (True);
}


/************************************************************************
 *									*
 *			PUBLIC ROUTINES					*
 *									*
 *		Publically accessible routines				*
 *									*
 ************************************************************************/

/*
 *	NAME
 *		InitViewCmapInfo
 *
 *	SYNOPSIS
 */
boolean
InitViewCmapInfo(ViewCmapInfoPtr cmapInfo)
/*
 *	DESCRIPTION
 *		Allocates colors out of the default screen's colormap for use
 *		in a color cube; or if not enough colorcells available, will
 *		create a private colormap.
 *
 *	RETURNS
 *		Pointer to the created ViewCmapInfo data structure. This
 *		structure along with a created colormap will need to be
 *		freed when no longer needed.
 *
 */
{
    Visual		*defaultVisual;
    Display		*dpy = cmapInfo->dpy;
    ULong		tmpAllocated[ MAX_COLOR_CELLS/32 ]; /* 1 bit/pixel */
    boolean		success = False;

    cmapInfo->contrast          = DEFAULT_CONTRAST;
    cmapInfo->brightness        = DEFAULT_BRIGHTNESS;
    cmapInfo->privateContrast   = REAL_CONTRAST(DEFAULT_CONTRAST);
    cmapInfo->privateBrightness = REAL_BRIGHTNESS(DEFAULT_BRIGHTNESS);
    cmapInfo->allocated		= tmpAllocated;

    cmapInfo->cmap = 0;
    MarkNoneAllocated(cmapInfo);
    if (MaxCmapsOfScreen(ScreenOfDisplay(dpy, cmapInfo->scrNumber)) == 1) {
	defaultVisual = DefaultVisual(dpy, cmapInfo->scrNumber);

	/*
	 * The default colormap will only work if we are going
	 * to be allocating a subset of the colors... this
	 * is only the case for Rgb232 and GrayScale7
	 */
	if (cmapInfo->depthType != GrayScale8
	 && cmapInfo->depthType != Rgb332
	 && cmapInfo->depthType != Rgb888
	 && defaultVisual->visualid == cmapInfo->visual->visualid) {
	    cmapInfo->cmapShared = True;
	    cmapInfo->cmap = DefaultColormap(dpy, cmapInfo->scrNumber);
	    switch (defaultVisual->class) {
	    case GrayScale:
	    case PseudoColor:
	    case DirectColor:
		cmapInfo->cmapWritable = True;
		break;
	    case StaticColor:
		goto complete;	/* we can't make this work */
	    case StaticGray:
	    case TrueColor:
		cmapInfo->cmapWritable = False;
		break;
	    }
	}
    }

    if (cmapInfo->cmap == 0) {
	cmapInfo->cmapShared = False;
	if (! AllocViewColormap(cmapInfo))
	    goto complete;	/* failed */
    }

    if (! cmapInfo->cmapWritable) {
	success = True;
	goto complete;
    }

    if (! AllocViewColorcells(cmapInfo)) {
	cmapInfo->cmapShared = False;
	if (! AllocViewColormap(cmapInfo))
	    goto complete;	/* failed */
	if (! AllocViewColorcells(cmapInfo)) {
	    XFreeColormap(cmapInfo->dpy, cmapInfo->cmap);
	    goto complete;	/* failed */
	}
    }

    CreateGammaTbl(cmapInfo);
    /* LoadColors(cmapInfo);  this makes the system thrash a little */
    success = True;

complete:
    cmapInfo->allocated	= NULL;
    return(success);
}

/*
 *	NAME
 *		SetViewAttributes
 *
 *	SYNOPSIS
 */
void
SetViewAttributes(ViewCmapInfoPtr cmapInfo,
		  ViewAttributesPtr viewAttr,
		  UInt mask)
/*
 *	DESCRIPTION
 *		Changes the values in a colormap's color cells that represent
 *		our RGB color cube.  These changes will reflect the requested
 *		changes in hue, saturation, contrast, and brightness.
 *
 *	RETURNS
 *		void
 *
 */
{
    int loadcolors = (mask & LOAD_COLORS_MASK);
    long newBrightness;
    long newContrast;

    if (! cmapInfo->cmapWritable)
	return;
    if (mask & HUE_MASK) {
	/* for now, do nothing */
    }
    if (mask & SATURATION_MASK) {
	/* for now, do nothing */
    }

    /*
     * Set the brightness, and then set the contrast from what's left over.
     */
    if (mask & BRIGHTNESS_MASK) {
	/* Check value */
	cmapInfo->brightness = viewAttr->brightness;
	newBrightness = REAL_BRIGHTNESS(viewAttr->brightness);
	if (newBrightness != cmapInfo->privateBrightness) {
	    cmapInfo->privateBrightness = newBrightness;
	    loadcolors++;
	}
    }
    if (mask & CONTRAST_MASK) {
	/* Check value */
	cmapInfo->contrast = viewAttr->contrast;
	newContrast = REAL_CONTRAST(viewAttr->contrast);
	if (newContrast != cmapInfo->privateContrast) {
	    cmapInfo->privateContrast = newContrast;
	    loadcolors++;
	}
    }
    if (mask & DISPLAY_GAMMA_MASK) {
	/* Check value */
        if ((viewAttr->dpyGamma >= 1.0)
		&& (viewAttr->dpyGamma <= 3.0)
		&& (viewAttr->dpyGamma != cmapInfo->dpyGamma)){
	    cmapInfo->dpyGamma = viewAttr->dpyGamma;
	    CreateGammaTbl(cmapInfo);
	    loadcolors++;
	} else {
	    viewAttr->dpyGamma = cmapInfo->dpyGamma;
	}
    }
    if (loadcolors) {
	LoadColors(cmapInfo);
    }
}
