/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	mpegcomp.c - Interface code to fork off filter/compressor pair.
 *
 * DESCRIPTION
 *	This is the guts of the code to fork off the filter/mpeg compressor
 *	programs. ALL routines are PRIVATE, but they ARE accessible with the
 *	mcs(media control service) subsystem.
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/mpegcomp.c,v 1.62 1995/10/21 00:41:58 russ Exp $
 */
#define	NeedFunctionPrototypes		1
#include "nvrTypes.h"
#include "nvrAlloc.h"
#include "nvrBitTest.h"
#include "fileContent.h"
#include "fileAccess.h"
#include "process.h"
#include "mcs-serverP.h"
#include "mpegcomp.h"
#include "mpegcompP.h"
#include "protoparseP.h"
#include <stdio.h>
#include <stdarg.h>

static void
VaMpegMsg(McsServer server, char *fmt, va_list ap)
{
    char msg[1024];
    McsMpegCompPriv *handle = (McsMpegCompPriv *) (server);
    ParseInfoPtr pParse = handle->base.pParse;
    ParseEventNotifier eventNotify = pParse->config.u.client.eventNotify;
    McsInputNotify inputNotify = handle->params.required.inputNotify;
    ParseValue value;

    vsprintf(msg, fmt, ap);

    value.string = CopyOfString(msg);
    (*eventNotify)(pParse,
		   MCS_BAD_CONFIGURE,
		   MCS_TYPE_STRING,
		   &value);
    (*inputNotify) (server, -1, False, 0);	/* XXX hack! */
}

static void
MpegMsg(McsServer server, char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    VaMpegMsg(server, fmt, ap);
    va_end(ap);
}

static void
MpegBadConfig(McsServer server, char *fmt, ...)
{
    char msg[1024];
    va_list ap;
    McsMpegCompPriv *handle = (McsMpegCompPriv *) (server);
    ParseInfoPtr pParse = handle->base.pParse;
    ParseEventNotifier eventNotify = pParse->config.u.client.eventNotify;
    McsInputNotify inputNotify = handle->params.required.inputNotify;
    ParseValue value;

    va_start(ap, fmt);
    vsprintf(msg, fmt, ap);
    va_end(ap);

    value.string = CopyOfString(msg);
    (*eventNotify)(pParse,
		   MCS_BAD_CONFIGURE,
		   MCS_TYPE_STRING,
		   &value);
    (*inputNotify) (server, -1, False, 0);	/* XXX hack! */
}

/*
 * XXX these should not be in here
 */
static boolean
MpegCompOpenServer(McsServer server)
{

    register McsMpegCompPriv *handle = (McsMpegCompPriv *) (server);
    if (server == (McsServer) 0) {
	return (False);
    }

    /*
     * Now, initialize the REQUIRED things to rational defaults, in case
     * the user does not want to change them.
     */
    handle->videoCompressor.pid = -1;
    handle->params.required.width = -1;
    handle->params.required.height = -1;
    handle->params.required.inputNotify = NULL;
    handle->params.optional.terminationMethod = -1;
    handle->outputFile = NULL;
    return (True);
}

static void
TerminateCompressor(McsMpegCompPriv *handle)
{
    boolean verbose = (BM_ISSET(&handle->paramMask,MPEG_DEBUG) != 0);

   /*
     * Now, decide HOW to terminate the compressor,
     * according to the user app's wishes.
     */
    switch (handle->terminationMethod) {
    case MCVAL_TERM_IMMED:
	if (handle->videoCompressor.pid > 0) {
	    KillProcess(handle->videoCompressor.pid);
	    (void) WaitForPid(handle->videoCompressor.pid, verbose);
	    handle->videoCompressor.pid = -1;
	}
	if (handle->outputFile) {
	    (void) UnlinkPath(handle->outputFile, False);
	    (void) free(handle->outputFile);
	    handle->outputFile = NULL;
	}
	break;

    case MCVAL_TERM_ASAP:
    default:
	if (handle->videoCompressor.pid > 0) {
	    TerminateProcess(handle->videoCompressor.pid);
#if 0
/*
** If we wait for the process to terminate,
**	the UI is locked in a confusing state where the dialog box
**	is up and it looks like things aren't responding
*/
	    (void) WaitForPid(handle->videoCompressor.pid, verbose);
#endif /* 0 */
	    handle->videoCompressor.pid = -1;
	}
	break;
    }
}

static void
MpegCompCloseServer(McsServer handlePtr)
{
    register McsMpegCompPriv *handle = (McsMpegCompPriv *) (handlePtr);

    if (handle == NULL)
	return;
    TerminateCompressor(handle);
}

static boolean
MpegCompSetParameters( McsServer handlePtr,
			PrivPointer paramPtr, ULong mask
		     )
{
    MpegBadConfig(handlePtr, "Must use MpegCompSetWideParameters\n");
    return (False);
}

static boolean
MpegCompSetWideParameters( McsServer handlePtr,
			PrivPointer paramPtr, BitMask *mask
		     )
{
    register McsMpegCompPriv *handle = (McsMpegCompPriv *) (handlePtr);
    register MpegCompParams *newParams = (MpegCompParams *) paramPtr;
    char *errorRoutine = "MpegSetWideParameter";

    if (handle == 0)
	return (False);

    if (handle->videoCompressor.pid >= 0
     && handle->compressorIsStopped == False) {
	MpegMsg(handlePtr, "%s: Must do StopOperation before new SetParam.\n", errorRoutine);
	return (False);
    }

    memcpy((char *)&handle->params.required,(char *)&newParams->required,
	   sizeof(handle->params.required));

/* compressor mask bits */
    
    if (BM_ISSET(mask,MCBitrate))
	handle->params.optional.bitrate = newParams->optional.bitrate;

    if (BM_ISSET(mask,MCBufferSize))
	handle->params.optional.bufferSize = newParams->optional.bufferSize;

    if (BM_ISSET(mask,MCInitFullness))
	handle->params.optional.initFullness = newParams->optional.initFullness;

    if (BM_ISSET(mask,MCInitIcomplex))
	handle->params.optional.initIcomplex = newParams->optional.initIcomplex;

    if (BM_ISSET(mask,MCInitPcomplex))
	handle->params.optional.initPcomplex = newParams->optional.initPcomplex;

    if (BM_ISSET(mask,MCInitBcomplex))
	handle->params.optional.initBcomplex = newParams->optional.initBcomplex;

    if (BM_ISSET(mask,MCSceneEnds))
	handle->params.optional.sceneEnds = newParams->optional.sceneEnds;

    if (BM_ISSET(mask,MCSliceStart))
	handle->params.optional.sliceStart = newParams->optional.sliceStart;

    if (BM_ISSET(mask,MCKbitIcomplex))
	handle->params.optional.kbitIcomplex = newParams->optional.kbitIcomplex;

    if (BM_ISSET(mask,MCKbitPcomplex))
	handle->params.optional.kbitPcomplex = newParams->optional.kbitPcomplex;

    if (BM_ISSET(mask,MCKbitBcomplex))
	handle->params.optional.kbitBcomplex = newParams->optional.kbitBcomplex;

    if (BM_ISSET(mask,MCPpicture))
	handle->params.optional.nppictures = newParams->optional.nppictures;

    if (BM_ISSET(mask,MCBpicture))
	handle->params.optional.nbpictures = newParams->optional.nbpictures;

    if (BM_ISSET(mask,MCMotion))
	handle->params.optional.motionAlgorithm =
				    newParams->optional.motionAlgorithm;

    if (BM_ISSET(mask,MCMotionRadius))
	handle->params.optional.motionRadius = newParams->optional.motionRadius;

    if (BM_ISSET(mask,MCCenteredQscale))
	handle->params.optional.centeredQuantLevel =
		newParams->optional.centeredQuantLevel;

    if (BM_ISSET(mask,MCFixedQscale))
	handle->params.optional.fixedQuantLevel =
		newParams->optional.fixedQuantLevel;

    if (BM_ISSET(mask,MCPictureRate))
	handle->params.optional.pictureRate = newParams->optional.pictureRate;

    if(BM_ISSET(mask,MCNumProcessors))
	handle->params.optional.numProcessors =
		newParams->optional.numProcessors;

    if(BM_ISSET(mask,MCReactionParam))
	handle->params.optional.reactionParam =
		newParams->optional.reactionParam;

    if(BM_ISSET(mask,MCSpatialMaskGain))
	handle->params.optional.spatialMaskGain =
		newParams->optional.spatialMaskGain;

    if (BM_ISSET(mask,MCVbvZero))
	handle->params.optional.vbvZero = newParams->optional.vbvZero;

/* input mask bits */

    if(BM_ISSET(mask,MCVInputName))
	handle->params.optional.vInputName = newParams->optional.vInputName;

    if(BM_ISSET(mask,MCVInputPrefix))
	handle->params.optional.vInputPrefix = newParams->optional.vInputPrefix;

    if(BM_ISSET(mask,MCVInputSuffix))
	handle->params.optional.vInputSuffix = newParams->optional.vInputSuffix;

    if(BM_ISSET(mask,MCVInputNDigits))
	handle->params.optional.vInputNDigits =
		newParams->optional.vInputNDigits;

    if(BM_ISSET(mask,MCVInputStart))
	handle->params.optional.vInputStart = newParams->optional.vInputStart;

    if(BM_ISSET(mask,MCVInputStop))
	handle->params.optional.vInputStop = newParams->optional.vInputStop;

    if(BM_ISSET(mask,MCVResampleFps))
	handle->params.optional.resampleFps = newParams->optional.resampleFps;

/* filter mask bits */

    if (BM_ISSET(mask,MCClipX))
	handle->params.optional.clipX = newParams->optional.clipX;

    if (BM_ISSET(mask,MCClipY))
	handle->params.optional.clipY = newParams->optional.clipY;

    if (BM_ISSET(mask,MCClipWidth))
	handle->params.optional.clipWidth = newParams->optional.clipWidth;

    if (BM_ISSET(mask,MCClipHeight))
	handle->params.optional.clipHeight = newParams->optional.clipHeight;

    if (BM_ISSET(mask,MCResizeWidth))
	handle->params.optional.resizeWidth = newParams->optional.resizeWidth;

    if (BM_ISSET(mask,MCResizeHeight))
	handle->params.optional.resizeHeight = newParams->optional.resizeHeight;

    if (BM_ISSET(mask,MCResizeMethod))
	handle->params.optional.resizeMethod = newParams->optional.resizeMethod;

    if (BM_ISSET(mask,MCDeinterlaceWidth))
	handle->params.optional.deinterlaceWidth =
		newParams->optional.deinterlaceWidth;

    if (BM_ISSET(mask,MCSmoothLevel))
	handle->params.optional.smoothLevel = newParams->optional.smoothLevel;

/* output mask bits */

/* misc mask bits */

    if (BM_ISSET(mask,MCTermMethod)) {
	if (newParams->optional.terminationMethod == MCVAL_TERM_IMMED)
	    handle->outputFile =
			CopyOfString(handle->params.required.outputfile);
	handle->params.optional.terminationMethod =
			newParams->optional.terminationMethod;
	/* XXX */ handle->terminationMethod =
		handle->params.optional.terminationMethod;
    }

    handle->paramMask = *mask;

    return (True);
}

static boolean
MpegCompUnsetParameters(McsServer handlePtr, ULong mask)
{
    MpegMsg(handlePtr, "Must use MpegCompUnsetWideParameters\n");
    return (False);
}

static boolean
MpegCompUnsetWideParameters(McsServer handlePtr, BitMask *bp)
{
    return (False);
}

static void
EofNotify(ParseInfoPtr pParse)
{
    McsMpegCompPriv *handle = (McsMpegCompPriv *)GetParsePrivateData(pParse);
    McsInputNotify inputNotify = handle->params.required.inputNotify;
    _McsEofNotify(pParse);
    
    (*inputNotify)((McsServer)handle, handle->videoCompressor.inFd,
		   False, NULL);
    (void) CloseFile(handle->videoCompressor.inFd);
    (*pParse->config.inputNotify)(pParse);
}

static boolean
BuildVideoArgvec(Argvec *videoCompressArgs, McsMpegCompPriv *handle)
{
    char	*errorRoutine = "BuildVideoArgvec";
    char	*fileName;
    int		i;
    boolean	clip, resize, resample, retype;
    long	fWidth, fHeight;
    long	eWidth, eHeight;
    boolean	interlaced;


    videoCompressArgs->length = 0;
    videoCompressArgs->argv = NULL;

    /*
     * validate Parameters
     */
    /* input file or prefix/suffix/... */
    if (!BM_ISSET(&handle->paramMask,MCVInputName) &&
	!BM_ISSET(&handle->paramMask,MCVInputPrefix))
    {
	MpegBadConfig((McsServer)handle, "%s: must specify input (name or prefix/suffix/...).\n",
	      errorRoutine);
	return (False);
    }

    /* all or nothing for clip{Width,Height,X,Y} */
    if (BM_ISSET(&handle->paramMask,MCClipWidth) &&
	(!BM_ISSET(&handle->paramMask,MCClipHeight) ||
	 !BM_ISSET(&handle->paramMask,MCClipX) ||
	 !BM_ISSET(&handle->paramMask,MCClipY)))
    {
	MpegBadConfig((McsServer)handle, "%s: inappropriate clipping configuration.\n",
	      errorRoutine);
	return (False);
    }
    /* all or nothing of resize{Width,Height,Method},maintainAspect */
    if (BM_ISSET(&handle->paramMask,MCResizeWidth) &&
	(!BM_ISSET(&handle->paramMask,MCResizeHeight) ||
	 !BM_ISSET(&handle->paramMask,MCResizeMethod)))
    {
	MpegBadConfig((McsServer)handle, "%s: inappropriate Resizing configuration.\n",
	      errorRoutine);
	return (False);
    }
	
    /*
     * Now, figure out the command line for the video compressor part of
     * the pipeline.  For now, we default it to "vmpegcomp". In the future,
     * we may want to add another parameter to specify WHICH compressor...
     */
    AddArgv(videoCompressArgs, "vmpegcomp", False);

/*
** Filter Stage 1: read in file
*/
    /*
     * And now, figure out the arguments for the compressor.
     * NOTE: Some of the optional arguments of the compressor have
     * a weird syntax. I'd like to fix that, but for now we'll
     * leave it alone. If you fix the compressor, you MUST change
     * the arguments below. For help, see the comments in the file
     * <compressor>/main.c for a (somewhat) good description of what
     * the various arguments really do.
     */
    AddArgv(videoCompressArgs, EnumToName(ImgOpIO,ImageOpNames), False);
    if (BM_ISSET(&handle->paramMask,MCVInputName)) {
	AddFmtArgv(videoCompressArgs, "read=%s",
		   handle->params.optional.vInputName);
    } else {
	AddArgv(videoCompressArgs, "read", False);
	AddFmtArgv(videoCompressArgs, "prefix=%s",
		   handle->params.optional.vInputPrefix);
	if (BM_ISSET(&handle->paramMask,MCVInputSuffix)) {
	    AddFmtArgv(videoCompressArgs, "suffix=%s",
		   handle->params.optional.vInputSuffix);
	}
	if (BM_ISSET(&handle->paramMask,MCVInputNDigits)) {
	    AddFmtArgv(videoCompressArgs, "ndigits=%s",
		   handle->params.optional.vInputNDigits);
	}
	if (BM_ISSET(&handle->paramMask,MCVInputStart)) {
	    AddFmtArgv(videoCompressArgs, "start=%ld",
		   handle->params.optional.vInputStart);
	}
	if (BM_ISSET(&handle->paramMask,MCVInputStop)) {
	    AddFmtArgv(videoCompressArgs, "stop=%ld",
		   handle->params.optional.vInputStop);
	}
    }
    if (BM_ISSET(&handle->paramMask,MCVResampleFps)) {
	AddFmtArgv(videoCompressArgs, "changefps=%.3f",
	    handle->params.optional.resampleFps);
	AddFmtArgv(videoCompressArgs, "framerate=%.3f",
	    handle->params.optional.pictureRate);
    }
/*
** Filter Stage 2: deinterlace
*/
    if( BM_ISSET(&handle->paramMask,MCDeinterlaceWidth) ) {
	AddArgv(videoCompressArgs,EnumToName(ImgOpDeinterlace,ImageOpNames),
		False);
	AddArgv(videoCompressArgs,"method=spatial",False);
	AddFmtArgv(videoCompressArgs,"width=%ld",
		   handle->params.optional.deinterlaceWidth);
    }

/*
** Filter Stage 2+: Color
*/
    resample = (handle->params.required.inputSampleType != Sample4_2_0);
    retype = True;

/* piggybacked onto a color or resample OP */
#define DORESAMPLE \
    if( resample ) { \
        AddFmtArgv(videoCompressArgs,"sample=%s", \
		   EnumToName(Sample4_2_0,SampleTypeNames), \
		    False); \
	resample = 0; \
    }
#define DOTYPE \
    if( retype ) { \
    	AddArgv(videoCompressArgs,"type=rawycrcb",False); \
        retype = 0; \
    }

    if (handle->params.required.inputColorSpace != ColorYC) {
	AddArgv(videoCompressArgs,EnumToName(ImgOpColor,ImageOpNames),
		    False);
	DOTYPE;
	DORESAMPLE;
    }
/*
** Filter Stage 3: Resize
*/
    interlaced = handle->params.required.interlaced;
    clip = BM_ISSET(&handle->paramMask,MCClipWidth);
    resize = BM_ISSET(&handle->paramMask,MCResizeWidth);
    fWidth = resize ? handle->params.optional.resizeWidth :
    		      handle->params.required.width;
    fHeight = resize ? handle->params.optional.resizeHeight :
    		       (interlaced ? handle->params.required.height << 1
			           : handle->params.required.height);
    eWidth = clip ? handle->params.optional.clipWidth :
	            handle->params.required.width;
    eHeight = clip ? handle->params.optional.clipHeight :
	             (interlaced ? handle->params.required.height << 1
			           : handle->params.required.height);

/* piggybacked onto a resample or filter OP */
#define DOCLIP \
    if( clip ) { \
        AddFmtArgv(videoCompressArgs,"inclip=%ldx%ld+%ld+%ld", \
		   handle->params.optional.clipWidth, \
		   handle->params.optional.clipHeight, \
		   handle->params.optional.clipX, \
		   handle->params.optional.clipY); \
	clip = 0; \
    }
    
    

    if (resize) {
	switch((int) handle->params.optional.resizeMethod) {
	default:
	    break;	/* handled at an earlier point */
	case ResampleNearest:
	    AddArgv(videoCompressArgs,EnumToName(ImgOpResample,ImageOpNames),
		    False);
	    AddArgv(videoCompressArgs,"method=nearest",False);
	    AddFmtArgv(videoCompressArgs,"size=%ldx%ld",
		       handle->params.optional.resizeWidth,
		       handle->params.optional.resizeHeight);
	    DOCLIP;
	    DORESAMPLE;
	    DOTYPE;
	    eWidth = fWidth; eHeight = fHeight;
	    break;
	case ResampleBiLinear:
	    AddArgv(videoCompressArgs,EnumToName(ImgOpResample,ImageOpNames),
		    False);
	    AddArgv(videoCompressArgs,"method=bilinear",False);
	    AddFmtArgv(videoCompressArgs,"size=%ldx%ld",
		       handle->params.optional.resizeWidth,
		       handle->params.optional.resizeHeight);
	    DOCLIP;
	    DORESAMPLE;
	    DOTYPE;
	    eWidth = fWidth; eHeight = fHeight;
	    break;
	case 0x77:	/* FIR filtering */

	    do {
		i = 0;
		if( (eWidth << 1) <= fWidth ) {
		    AddArgv(videoCompressArgs,
			    EnumToName(ImgOpFilter,ImageOpNames),False);
		    AddArgv(videoCompressArgs,"1x7=-12,0,140,256,140,0,-12/256",
			    False);
		    AddArgv(videoCompressArgs,"interpolate=horizontal",False);
		    eWidth <<= 1;
		    i++;
		} else if( (eWidth >> 1) >= fWidth ) {
		    AddArgv(videoCompressArgs,
			    EnumToName(ImgOpFilter,ImageOpNames),False);
		    AddArgv(videoCompressArgs,"7x1=-29,0,88,138,88,0,-29/256",
			    False);
		    AddArgv(videoCompressArgs,"decimate=horizontal",False);
		    eWidth >>= 1;
		    i++;
		}
		if( i > 0 ) {
		    DOCLIP;
		}
		if( (eHeight << 1) <= fHeight ) {
		    AddArgv(videoCompressArgs,
			    EnumToName(ImgOpFilter,ImageOpNames),False);
		    AddArgv(videoCompressArgs,"1x7=-12,0,140,256,140,0,-12/256",
			    False);
		    AddArgv(videoCompressArgs,"interpolate=vertical",False);
		    eHeight <<= 1;
		    i++;
		} else if( (eHeight >> 1) >= fHeight ) {
		    AddArgv(videoCompressArgs,
			    EnumToName(ImgOpFilter,ImageOpNames),False);
		    AddArgv(videoCompressArgs,"1x7=-29,0,88,138,88,0,-29/256",
			    False);
		    AddArgv(videoCompressArgs,"decimate=vertical",False);
		    eHeight >>= 1;
		    i++;
		}
		if( i > 0 ) {
		    DOCLIP;
		}
	    } while(i > 0);

	    break;
	}
    }
    if( eWidth != fWidth || eHeight != fHeight ) {
	AddArgv(videoCompressArgs,
		EnumToName(ImgOpResample,ImageOpNames),False);
	AddArgv(videoCompressArgs,"method=bilinear",False);
	AddFmtArgv(videoCompressArgs,"size=%ldx%ld",fWidth,fHeight);
	DOCLIP;
	DORESAMPLE;
	DOTYPE;
    }
    if( resample ) {
	AddArgv(videoCompressArgs,
		EnumToName(ImgOpResample,ImageOpNames),False);
	AddArgv(videoCompressArgs,"method=bilinear",False);
	DOCLIP;
	DORESAMPLE;
	DOTYPE;
    }
/*
** Filter Stage 4: Smooth image
*/
    if (BM_ISSET(&handle->paramMask,MCSmoothLevel)) {
	AddArgv(videoCompressArgs,EnumToName(ImgOpFilter,ImageOpNames),
		False);
	switch(handle->params.optional.smoothLevel) {
	case 1:
	    AddArgv(videoCompressArgs,"3x3=1,2,1,2,52,2,1,2,1/64",False);
	    break;
	case 2:
	    AddArgv(videoCompressArgs,"3x3=1,2,1,2,20,2,1,2,1/32",False);
	    break;
	case 3:
	    AddArgv(videoCompressArgs,"3x3=1,2,1,2,4,2,1,2,1/16",False);
	    break;
	}
    }
/*
** Filter Stage 5: write out file
*/
    /*
     * output file name, which is what the output of the compressor will be.
     * NOTE: this is the mpeg output, not the event output.
     */
    AddArgv(videoCompressArgs, EnumToName(ImgOpIO,ImageOpNames), False);
    fileName = handle->params.required.outputfile;
    AddFmtArgv(videoCompressArgs, "write=%s", fileName);
    AddArgv(videoCompressArgs, "type=mpeg", False);

    if (BM_ISSET(&handle->paramMask,MCPictureRate))
	AddFmtArgv(videoCompressArgs, "-p%.3f",
	    handle->params.optional.pictureRate);

    if (BM_ISSET(&handle->paramMask,MCBitrate))
	AddFmtArgv(videoCompressArgs, "-b%ld", handle->params.optional.bitrate);

    if (BM_ISSET(&handle->paramMask,MCFixedQscale))
	AddFmtArgv(videoCompressArgs, "-q%ld",
		   handle->params.optional.fixedQuantLevel);

    if (BM_ISSET(&handle->paramMask,MCNumProcessors))
	AddFmtArgv(videoCompressArgs, "-P%ld",
		   handle->params.optional.numProcessors);

    i = 0;
    if (BM_ISSET(&handle->paramMask,MCMotion)) {
	i += 1;
    }
    if (BM_ISSET(&handle->paramMask,MCMotionRadius)) {
	i += 2;
    }
    switch (i) {
    case 3:
	AddFmtArgv(videoCompressArgs, "-M%ld,%ld",
	    handle->params.optional.motionAlgorithm,
	    handle->params.optional.motionRadius);
	break;

    case 1:
	AddFmtArgv(videoCompressArgs, "-M%ld",
	    handle->params.optional.motionAlgorithm);
	break;

    case 2:
	AddFmtArgv(videoCompressArgs, "-M1,%ld",
	    handle->params.optional.motionRadius);
	break;

    case 0:
	AddArgv(videoCompressArgs, "-M1", False);
	break;
    }

    i = 0;
    if (BM_ISSET(&handle->paramMask,MCPpicture)) {
	i += 1;
    }
    if (BM_ISSET(&handle->paramMask,MCBpicture)) {
	i += 2;
    }
    switch (i) {
    case 1:
	AddFmtArgv(videoCompressArgs, "-g%ld,2",
	    handle->params.optional.nppictures);
	break;

    case 2:
	AddFmtArgv(videoCompressArgs, "-g4,%ld",
	    handle->params.optional.nbpictures);
	break;

    case 3:
	AddFmtArgv(videoCompressArgs, "-g%ld,%ld",
	    handle->params.optional.nppictures,
	    handle->params.optional.nbpictures);
	break;

    case 0:
	break;

    }
    return(True);
}

static boolean
MpegCompStartOperation(McsServer handlePtr)
{
    register McsMpegCompPriv *handle = (McsMpegCompPriv *) (handlePtr);
    Argvec videoCompressARGS, *videoCompressArgs = &videoCompressARGS;
    ParseConfig config;
    McsInputNotify inputNotify = handle->params.required.inputNotify;
    char *errorRoutine = "MpegStartOperation";

    if (handle == 0)
	return (False);

    if (handle->videoCompressor.pid > 0) {
        if (handle->compressorIsStopped) {
            ResumeProcess(handle->videoCompressor.pid);
            handle->compressorIsStopped = False;
	    if (BM_ISSET(&handle->paramMask,MPEG_DEBUG) ||
		BM_ISSET(&handle->paramMask,MPEG_FULL_DEBUG))
		    Msg("%s: Restarted compressor PID %ld.\n",
			errorRoutine, handle->videoCompressor.pid);
	    return (True);
	}
	return (False);
    }
    
    /*
     * init the connection data
     */
    config.inFileHandle = -1;
    config.outFileHandle = -1;
    config.inputNotify = _McsInputNotify;
    config.outputNotify = NULL;
    config.eofNotify = EofNotify;
    config.pointer = (PrivPointer)handle;
    config.server = False;
    config.asynchronous = False;
    config.u.client.eventNotify = _McsEventNotify;
    handle->base.pParse = InitParsedConnection(&config);

    /*
     * Check to see if the application has specified both an input and an
     * output file name. Both MUST be specified.
     */
    if (handle->params.required.outputfile == (char *)0 ||
	(!BM_SET(&handle->paramMask,MCVInputName) && 
	 !BM_SET(&handle->paramMask,MCVInputPrefix)) ) {

	MpegBadConfig((McsServer)handle, "%s: must specify both input/output.\n", errorRoutine);
	return (False);
    }

    /*
     * If the user application has specified an output width & height,
     * we need to tell the filter AND the compressor.
     */
    if (handle->params.optional.resizeWidth > 0
	  || handle->params.optional.resizeHeight > 0) {
	if (handle->params.optional.resizeWidth <= 0
	     || handle->params.optional.resizeHeight <= 0) {
	    MpegBadConfig((McsServer)handle, "%s: output width/height must be greater than 0.\n",
		  errorRoutine);
	    return (False);
	}
    }

    /*
     * And check to see if the user specified an input notify routine
     * for MCS events.
     */
    if (inputNotify == NULL) {
	MpegBadConfig((McsServer)handle, "%s: must specify an input routine.\n", errorRoutine);
	return (False);
    }

    if( !BuildVideoArgvec(videoCompressArgs,handle) ) {
	return(False);
    }

    /*
     * Now, all of the arguments are set up. fork()/exec() both the
     * mpeg filter and the compressor, and in THAT order (so the pipes
     * get set up correctly).
     * The processes that get created will look like this:
     * ---------       ---------------
     * |       |       |             |
     * | USER  |<------|-o           |
     * | APP   |------>|-i   Proc 2  |
     * | ----  |       |     ------  |
     * |       |   +-->|(fd 0)       |
     * |       |   |   |             |
     * ---------   |   -------------
     * ---------   |
     * | (fd 1)|   |
     * |       |   |
     * | Proc1 |---+
     * | ----- |
     * |       |
     * | fd 0  |
     * | closed|
     * ---------
     *
     * The user app is the mpeg compressor GUI, proc2 is the mpeg compressor, 
     * and proc1 is a filter for translating movie files into yuv and writing
     * them on stdout.  This compressor recieves commands on the file descriptor
     * specified by the -i arg, and gives back events on the fd specified by -o.
     * It is most convenient for the compressor to recieve yuv input on stdin
     * (file descriptor 0).
     */
    videoCompressArgs->path = videoCompressArgs->argv[0];


    handle->compressorIsStopped = False;

    handle->videoCompressor =
        PipeToAndFrom(videoCompressArgs, DuplexPipeRewriterToStdin);

PrintArgv(videoCompressArgs);

    if (BM_ISSET(&handle->paramMask,MPEG_DEBUG)) {
	PrintArgv(videoCompressArgs);
	Msg("%s: COMPRESS pid %ld inFD <fd%ld> outFD <fd%ld>\n", errorRoutine,
	handle->videoCompressor.pid, handle->videoCompressor.inFd,
	handle->videoCompressor.outFd);
    }

    /*
     * Now that everything is forked off, initialize the event mechanism.
     */
    config.inFileHandle = handle->videoCompressor.inFd;
    config.outFileHandle = handle->videoCompressor.outFd;
    AddConnectionIO(handle->base.pParse, &config);
    (*inputNotify) (handlePtr, handle->videoCompressor.inFd, True, 0);

    /*
     * Now that we have set up the compressor, we (i.e. the user application
     * that called us) still have an open file descriptor FOR INPUT to the com-
     * pressor, and we need to close that.
     */
    CloseFile(handle->videoCompressor.outFd);

    /*
     * Free up the Argv stuff for the compressor.
     */
    FreeArgv(videoCompressArgs);

    return (True);
}

static boolean
MpegCompStopOperation(McsServer handlePtr)
{
    register McsMpegCompPriv *handle = (McsMpegCompPriv *) (handlePtr);
    char *errorRoutine = "MpegStopOperation";
    ProcessState state;

    if (handle == 0 || handle->videoCompressor.pid <= 0)
	return (False);

    /*
     * See if the compressor is already stopped. If so, return success.
     */
    if (handle->compressorIsStopped)
	return (True);

    if (BM_ISSET(&handle->paramMask,MPEG_DEBUG) ||
	BM_ISSET(&handle->paramMask,MPEG_FULL_DEBUG))
	Msg("%s: stop(%ld).\n", errorRoutine, handle->videoCompressor.pid);

    if (handle->videoCompressor.pid > 0) {
        SuspendProcess(handle->videoCompressor.pid);
        state = WaitForPid(handle->videoCompressor.pid,
        		   BM_ISSET(&handle->paramMask,MPEG_DEBUG) ||
        		   BM_ISSET(&handle->paramMask,MPEG_FULL_DEBUG));
        switch(state) {
        case ProcIsDead:
	    handle->videoCompressor.pid = -1;
	    break;
        case ProcIsSuspended:
	    handle->compressorIsStopped = True;
	    break;
        case ProcIsRunning:
            ProgMsg("proc %ld running after stop?\n", handle->videoCompressor.pid);
            break;	/* hmmmm.... */
        }
    }

    return (True);
}

static McsEventPtr
MpegCompGetNextEvent(McsServer handlePtr)
{
    return (False);
}

static
McsServerClass _MpegCompClass = {
    McsMpegCompressor,
    sizeof(McsMpegCompPriv),
    MpegCompOpenServer,
    MpegCompSetParameters,
    MpegCompUnsetParameters,
    MpegCompGetNextEvent,
    MpegCompStartOperation,
    MpegCompStopOperation,
    MpegCompCloseServer,
    MpegCompSetWideParameters,
    MpegCompUnsetWideParameters
};

McsServerClassPtr MpegCompClass = &_MpegCompClass;
