/* $Header: /usr/local/src/nvrToMIT/RCS/Pixmap.c,v 1.4 1995/01/10 01:57:50 toddb Exp $ */

#include "nvrTypes.h"
#include "nvrAlloc.h"
#include "nvrString.h"

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/extensions/shape.h>
#include <Xm/Xm.h>
#include "xpm.h"
#include "WidgetPixmap.h"

#define MAX_PATH_LEN	256

typedef struct _xpmData XpmData, *XpmDataPtr;

struct _xpmData {
    char *name;
    Pixmap pixmap;
    Pixmap mask;
    Pixel background;
    XpmAttributes attributes;
    XpmDataPtr next;
}; /* typedef: XpmData, *XpmDataPtr */

struct _pixmapResource {
    XpmDataPtr list;
    Display *dpy;
    Screen *screen;
    GC gc;
    Window root;
    int dragx, dragy;
    Colormap colormap;
    char path[ MAX_PATH_LEN ];
    char *pPath;
}; /* typedef: PixmapResource, *PixmapResourcePtr */

static void
FreeXpmData(PixmapResourcePtr pixdb, XpmDataPtr data)
{
    if (data->pixmap)
	XFreePixmap(pixdb->dpy, data->pixmap);
    /*
     * Note: we never do an XFreePixmap(pixdb->dpy, data->mask) because
     *   - it is too expensive to copy it around
     *   - there may be multiple references to it via CopyPixmap
     *   - the mask never needs to change
     */

    XFreeColors(pixdb->dpy, pixdb->colormap,
		data->attributes.pixels, data->attributes.npixels, 0);

    XpmFreeAttributes(&data->attributes);
}

static XpmDataPtr
NewXpmData(char *name, Pixel background)
{
    XpmDataPtr newData = (XpmDataPtr)InsecureAlloc(sizeof(XpmData));
    long allocSize = sizeof(XpmData) + strlen(name) + 1;

    newData = (XpmDataPtr)InsecureAlloc(allocSize);
    if (newData)
	memset((char *)newData, 0, sizeof(XpmData));
    if (newData == NULL) {
	printf("can't allocate %d bytes for \"%s\" pixmap\n", allocSize, name);
	return (NULL);
    }
    newData->name = (char *)(newData + 1);
    strcpy(newData->name, name);
    newData->background = background;
    return (newData);
}

static int
LoadNamedPixmap(PixmapResourcePtr pixdb, char *name, XpmDataPtr data)
{
    int status;
    char pathbuf[ 256 ];
    int i;
    static char *imageSuffixes[] = {
	".image",
	".image.xpm",
	"",
	".xpm",
    };
#if 0
    static char *maskSuffixes[] = {
	".mask",
	".mask.xpm",
    };

    /*
     * first read in the mask, if there is one.  Force the depth to one
     * for this.
     */
    data->attributes.depth = 1;
    data->attributes.valuemask = XpmDepth;
    for (i = 0; i < XtNumber(maskSuffixes); i++) {
	sprintf(pathbuf, "%s%s", name, maskSuffixes[i]);
	status = XpmReadFileToPixmap(pixdb->dpy,
				     pixdb->root,
				     pixdb->path,
				     &data->mask,
				     NULL,
				     &data->attributes);
	if (status == XpmSuccess)
	    break;
	if (status != XpmOpenFailed)
	    return(status);
    }
#endif

    /*
     * Now get the image; allow the depth to be set by the software.
     */
    data->attributes.colormap = pixdb->colormap;
    data->attributes.valuemask = XpmColormap;
    for (i = 0; i < XtNumber(imageSuffixes); i++) {
	sprintf(pathbuf, "%s%s", name, imageSuffixes[i]);
	status = XpmReadFileToPixmap(pixdb->dpy,
				     pixdb->root,
				     pathbuf,
				     &data->pixmap,
				     &data->mask,
				     &data->attributes);
	if (status == XpmSuccess)
	    break;
	if (status != XpmOpenFailed)
	    return(status);
    }

    return(status);
}

static Boolean
XmpErrorStatus(int status, char *name)
{
    char *error = NULL;
    char *warning = NULL;

    switch (status) {
    case XpmSuccess:
	break;
    case XpmColorError:
        warning = "Could not parse or alloc requested color for";
        break;
    case XpmOpenFailed:
        error = "Cannot open";
        break;
    case XpmFileInvalid:
        error = "Invalid XPM";
        break;
    case XpmNoMemory:
        error = "Not enough memory for";
        break;
    case XpmColorFailed:
        error = "Failed to parse or alloc some color for";
        break;
    }

    if (warning) {
        printf("%s Xpm Warning: %s pixmap \"%s\".\n",
	    name, warning, name);
    } else if (error) {
        printf("%s Xpm Error: %s pixmap \"%s\".\n", name, error, name);
	return (True);
    }
    return (False);
}

static void
CopyPixmap(PixmapResourcePtr pixdb, XpmDataPtr src, XpmDataPtr data)
{
    Display *dpy = pixdb->dpy;
    XGCValues values;
    GC gc = pixdb->gc;

    data->attributes = src->attributes;
    data->pixmap = XCreatePixmap (dpy,
				  pixdb->root,
				  data->attributes.width,
				  data->attributes.height,
				  DefaultDepthOfScreen(pixdb->screen));
    values.clip_mask = None;
    XChangeGC (dpy, gc, GCClipMask, &values);
    XCopyArea(dpy, src->pixmap, data->pixmap, gc,
	       0, 0,				/* src x, y */
	       data->attributes.width, data->attributes.height,
	       0, 0);				/* dst x, y */

    data->mask = src->mask; /* see comment in FreeXpmData() */
}

static void
ReconstructImage(PixmapResourcePtr pixdb, XpmDataPtr data, Pixel background)
{
    Pixmap newImage, image, mask;
    Display *dpy = pixdb->dpy;
    GC gc;
    XGCValues values;

    /*
     * If what we loaded turns out to be an image/mask pair, then we
     * we must construct the real image as follows:
     *   1. allocate another pixmap, call it "newImage"
     *   2. FillRectangle the pixmap with the background pixel from the widget
     *   3. CopyArea() from image to newImage using mask as the clip mask.
     *   4. free image and replace it with newImage.
     */
    image = data->pixmap;
    mask = data->mask;

    /* #1 */
    newImage = XCreatePixmap (pixdb->dpy,
			      pixdb->root,
			      data->attributes.width,
			      data->attributes.height,
			      DefaultDepthOfScreen(pixdb->screen));
    if ((gc = pixdb->gc) == NULL) {
	gc = pixdb->gc = XCreateGC (pixdb->dpy,
				     pixdb->root,
				     0,			/* valuemask */
				     NULL);		/* values */
    }

    /* #2 */
    values.foreground = background;
    values.fill_style = FillSolid;
    values.clip_mask = None;
    XChangeGC (dpy, gc, GCClipMask | GCForeground, &values);
    XFillRectangle(dpy, newImage, gc,
		   0, 0,			/* width, height */
		   data->attributes.width, data->attributes.height);

    /* #3 */
    values.clip_mask = mask;
    XChangeGC (dpy, gc, GCClipMask, &values);
    XCopyArea(dpy, image, newImage, gc,
	       0, 0,				/* src x, y */
	       data->attributes.width, data->attributes.height,
	       0, 0);				/* dst x, y */

    /* #4 */
    XFreePixmap(dpy, image);
    data->pixmap = newImage;
    data->background = background;
}

static XpmDataPtr
LookupXpmPixmap(PixmapResourcePtr pixdb, char *name)
{
    XpmDataPtr p;

    for (p = pixdb->list; p; p = p->next) {
	if (strcmp(name, p->name) == 0)
	    return(p);
    }
    return (NULL);
}

static XpmDataPtr
LoadXpmPixmap(PixmapResourcePtr pixdb, char *name, Pixel background)
{
    XpmDataPtr data;
    XpmDataPtr src = NULL;
    int status;

    if (data = LookupXpmPixmap(pixdb, name)) {
	if (data->mask && background != data->background)
	    src = data;
	else
	    return(data);
    }

    if ((data = NewXpmData(name, background)) == NULL)
	return(NULL);

    if (src) {
	CopyPixmap(pixdb, src, data);
    } else { 
	if ((status = LoadNamedPixmap(pixdb, name, data)) != XpmSuccess
	 && XmpErrorStatus(status, name)) {
	    FreeXpmData(pixdb, data);
	    return (NULL);
	}
    }

    data->next = pixdb->list;
    pixdb->list = data;

    if (data->mask)
	ReconstructImage(pixdb, data, background);

    return(data);
}

PixmapResourcePtr
InitPixmapDB(Widget w, char *path)
{
    char *ptr;
    PixmapResourcePtr pixdb;
    
    pixdb = (PixmapResourcePtr) InsecureAlloc(sizeof(PixmapResource));
    if (pixdb == NULL)
	return(NULL);
    memset((char *)pixdb, 0, sizeof(PixmapResource));

    pixdb->screen = XtScreen(w);
    pixdb->dpy = XtDisplay(w);
    pixdb->root = XRootWindowOfScreen(pixdb->screen);
    pixdb->colormap = XDefaultColormapOfScreen(pixdb->screen);
    if (path) {
	strcpy(pixdb->path, path);
	ptr = &pixdb->path[ strlen(pixdb->path) ];
	*ptr++ = '/';
	*ptr = '\0';
	pixdb->pPath = ptr;
    } else {
	pixdb->pPath = pixdb->path;
    }
	
    return(pixdb);
}

void
FreePixmapDB(PixmapResourcePtr pixdb)
{
    XpmDataPtr p, next;

    for (p = pixdb->list; p; p = next) {
	next = p->next;
	FreeXpmData(pixdb, p);
    }
    free((char *)pixdb);
}

#ifdef sgi
#    define OnSgi	True
#else
#    define OnSgi	True
#    define SgNsgiMode	"nothing"
#endif

void
SetLabelPixmap(PixmapResourcePtr pixdb,
	       Widget w,
	       char *name,
	       char *insName)
{
    XpmDataPtr insensitive, normal;
    Pixel background;

    XtVaGetValues(w, XmNbackground, &background, NULL);

    if ((normal = LoadXpmPixmap(pixdb, name, background)) == NULL)
	return;
    XtVaSetValues(w,
	XmNlabelPixmap, normal->pixmap,
	XmNlabelType, XmPIXMAP,
	XmNmarginBottom, 0,
	XmNmarginTop, 0,
	XmNmarginLeft, 0,
	XmNmarginRight, 0,
	XmNmarginWidth, 0,
	XmNmarginHeight, 0,
	XmNhighlightThickness, 0,
	NULL);

    if (insName == NULL)
	return;
    if ((insensitive = LoadXpmPixmap(pixdb, insName, background)) == NULL)
	return;
    XtVaSetValues(w, XmNlabelInsensitivePixmap, insensitive->pixmap, NULL);

#   if 0
	if (OnSgi) {
	    Boolean sgiMode = False;

	    XtVaGetValues(w,
		SgNsgiMode, &sgiMode,
		NULL);
	    if (sgiMode) {
		if ((armed = LoadXpmPixmap(pixdb, name, armColor)) == NULL)
		    return;
		XtVaSetValues(w,
		    XmNarmPixmap, armed->pixmap,
			XmNhighlightOnEnter, False,
			XmNfillOnArm, True,
			XmNhighlightColor, background,
			XmNforeground, background,
			XmNarmColor, background,
			XmNtopShadowColor, background,
			XmNbottomShadowColor, background,
			XmNselectColor, background,
			XmNtroughColor, background,
			XmNborderColor, background,
			SgNlampColor, background,
		    NULL);
	    }
	}
#   endif

}
