/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *      SelectionBox
 *
 * DESCRIPTION
 *      Used for File Selection Box activities
 *
 * $Header: /usr/local/src/nvr/library/RCS/SelectionBox.c,v 1.7 1994/10/02 23:58
:07 toddb Exp $
 */
#include "fileAccess.h"
#include "nvrAlloc.h"
#include "nvrString.h"
#include "WidgetDebug.h"
#include "WidgetDialog.h"
#include "WidgetUtil.h"
#include "WidgetFile.h"

#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <Xm/ToggleB.h>

#define DirMagicNumber	0x19283746
#define ComponentLength 256

typedef struct _pathName PathName, *PathNamePtr;
struct _pathName {
    PathNamePtr next;	/* must be first */
    char *name;
}; /* typedef: PathName, *PathNamePtr */

typedef struct _dirInfo {
    PathNamePtr nameList;	/* must be first */
    long        magic;		/* ... so that we can be sure its ours */
    XmString    *xmNames;
    int         nNames;
} DirInfo, *DirInfoPtr;

static Boolean
InsertPath(DirInfoPtr dirData, char *name)
{
    PathNamePtr pName, pPrev, pNew;
    int length = strlen(name) + 1;

    pName = dirData->nameList;
    pPrev = (PathNamePtr)&dirData->nameList;
    while (pName) {
	if (strcmp(name, pName->name) <= 0)
	    break;
	pPrev = pName;
	pName = pName->next;
    }

    pNew = (PathNamePtr) InsecureAlloc(sizeof(PathName) + length);
    if (pNew == NULL)
	return(False);
    pNew->name = (char *)(pNew + 1);
    strcpy(pNew->name, name);
    pNew->next = pName;
    pPrev->next = pNew;
    return(True);
}

static int PatternMatchAfterStar (char *pattern, char *text);

/* Match the pattern PATTERN against the string TEXT;
   return 1 if it matches, 0 otherwise.

   A match means the entire string TEXT is used up in matching.

   In the pattern string, `*' matches any sequence of characters,
   `?' matches any character, [SET] matches any character in the specified set,
   [!SET] matches any character not in the specified set.

   A set is composed of characters or ranges; a range looks like
   character hyphen character (as in 0-9 or A-Z).
   [0-9a-zA-Z_] is the set of characters allowed in C identifiers.
   Any other character in the pattern must be matched exactly.

   To suppress the special syntactic significance of any of `[]*?!-\',
   and match the character exactly, precede it with a `\'.

   If DOT_SPECIAL is nonzero,
   `*' and `?' do not match `.' at the beginning of TEXT.  */

static int
PatternMatch (char *pattern, char *text, int dotSpecial)
{
  register char *p = pattern, *t = text;
  register char c;

  while ((c = *p++) != '\0')
    switch (c)
      {
      case '?':
        if (*t == '\0' || (dotSpecial && t == text && *t == '.'))
          return 0;
        else
          ++t;
        break;

      case '\\':
        if (*p++ != *t++)
          return 0;
        break;

      case '*':
        if (dotSpecial && t == text && *t == '.')
          return 0;
        return PatternMatchAfterStar (p, t);

      case '[':
        {
          register char c1 = *t++;
          int invert;

          if (!c1)
            return (0);

          invert = ((*p == '!') || (*p == '^'));
          if (invert)
            p++;

          c = *p++;
          while (1)
            {
              register char cstart = c, cend = c;

              if (c == '\\')
                {
                  cstart = *p++;
                  cend = cstart;
                }

              if (c == '\0')
                return 0;

              c = *p++;
              if (c == '-')
                {
                  cend = *p++;
                  if (cend == '\\')
                    cend = *p++;
                  if (cend == '\0')
                    return 0;
                  c = *p++;
                }
              if (c1 >= cstart && c1 <= cend)
                goto match;
              if (c == ']')
                break;
            }
          if (!invert)
            return 0;
          break;

        match:
          /* Skip the rest of the [...] construct that already matched.  */
          while (c != ']')
            {
              if (c == '\0')
                return 0;
              c = *p++;
              if (c == '\0')
                return 0;
              else if (c == '\\')
                ++p;
            }
          if (invert)
            return 0;
          break;
        }

      default:
        if (c != *t++)
          return 0;
      }

  return *t == '\0';
}

/* Like glob_match, but match PATTERN against any final segment of TEXT.  */

static int
PatternMatchAfterStar (char *pattern, char *text)
{
  register char *p = pattern, *t = text;
  register char c, c1;

  while ((c = *p++) == '?' || c == '*')
    if (c == '?' && *t++ == '\0')
      return 0;

  if (c == '\0')
    return 1;

  if (c == '\\')
    c1 = *p;
  else
    c1 = c;

  while (1)
    {
      if ((c == '[' || *t == c1) && PatternMatch (p - 1, t, 0))
        return 1;
      if (*t++ == '\0')
        return 0;
    }
}

static void
FreeDirData(Widget fs, DirInfoPtr dirData)
{
    int nNames;
    PathNamePtr pName, pNext;

    if (dirData->magic != DirMagicNumber) {
	char wpath[ ComponentLength ];
	WarningDialog(fs, NULL,
	    "%s.%s resource has a bad magic number!",
	    GetWidgetPath(fs, wpath, sizeof(wpath)),
	    XmNuserData);
	return;
    }

    /* Free the XmStrings */
    for (nNames = 0; nNames < dirData->nNames; nNames++) {
	XtFree((char *)dirData->xmNames[nNames]);
	dirData->xmNames[nNames] = NULL;
    }

    /* Free the XmString list */
    if (dirData->xmNames) {
	free((void *)dirData->xmNames);
	dirData->xmNames = NULL;
    }

    /* Free the char * names */
    pNext = NULL;	/* for lint */
    for (pName = dirData->nameList; pName; pName = pNext) {
	pNext = pName->next;
	free((void *)pName);
    }
    free((void *)dirData);
}

/*
 * This is not intended to be the XmNfileSearchProc callback.  Instead,
 * The user calls this function from his own XmNfileSearchProc, specifying
 * a file filter function and a filter-specific pointer, 'userCriterion'.
 * For each file in the directory, fileFilter() will be called to
 * see whether to include it in the list.
 *
 * fileFilter() can be null, and will be ignored.
 */
void
StdFileSearchForFsb(Widget fs,
		    XmFileSelectionBoxCallbackStruct *cbs, 
		    boolean (*fileFilter)(char *path, void *userCriterion),
		    void *userCriterion)
{
    void *vp;
    char *directory, *path, *file, *component, *pattern;
    int baseLen, nNames, allocSize;
    DirInfoPtr dirData;
    int nAllocFailures, nTooLong, i, nameLength;
    PathNamePtr pName;
    XmStringCharSet charSet = (XmStringCharSet)XmSTRING_DEFAULT_CHARSET;
    UChar fileTypeMask = XmFILE_REGULAR;

    /*
     * prepare pointers that are freed on exit.
     */
    dirData = NULL;
    pattern = directory = NULL;

    if (! XmStringGetLtoR(cbs->dir, charSet, &directory)
     || ! XmStringGetLtoR(cbs->pattern, charSet, &pattern)) {
	WarningDialog(fs, NULL, "Can't get file selection info!");
        goto complete; /* can't do a thing without these */
    }

    /*
     * Go get the user data that we tacked onto
     * the widget, free it, then recreate-the directory content list.
     */
    XtVaGetValues(fs,
		  XmNuserData, (XtPointer)&dirData,
		  XmNfileTypeMask, (XtPointer)&fileTypeMask,
		  NULL);
    if (dirData)
	FreeDirData(fs, dirData);

    /*
     * First, clear the value in the widget.  If we get very far
     * along, we'll fill it in with a new value.
     */
    XtVaSetValues(fs, XmNuserData, (XtPointer)NULL, NULL);

    baseLen = strlen(directory);
    allocSize = sizeof(DirInfo) + baseLen + ComponentLength;
    dirData = (DirInfoPtr) InsecureAlloc(allocSize);
    if (dirData == NULL) {
	AllocFailDialog(fs, NULL, "directory data", allocSize);
	goto complete;
    }
    memset((char *)dirData, 0, allocSize);
    dirData->magic = DirMagicNumber;
    path = (char *)(dirData + 1);
    strcpy(path, directory);
    component = path + baseLen;
    if (component[-1] != '/')
	*component++ = '/';

    vp = OpenDir(directory);
    if( vp == NULL ) {
	WarningDialog(fs, NULL, "Can't open directory: \"%s\"", directory);
	goto complete;
    }

    /*
     * This is the new value for the directory data.
     */
    XtVaSetValues(fs, XmNuserData, (XtPointer)dirData, NULL);

    nAllocFailures = nTooLong = nNames = 0;
    while( (file = ReadDir(vp)) ) {
	nameLength = strlen(file);
	if (nameLength >= ComponentLength) {
	    nTooLong++;
	    continue;
	}

        if (! PatternMatch (pattern, file, True))
	    continue;
        strcpy(component, file);
        if (fileFilter && ! (*fileFilter)(path, userCriterion))
	    continue;

	if (fileTypeMask != XmFILE_ANY_TYPE) {
	    switch (FileType(path) & (FILE_DIRECTORY | FILE_REGULAR)) {
	    case FILE_REGULAR:
		if (fileTypeMask != XmFILE_REGULAR)
		    continue;
		break;
	    case FILE_DIRECTORY:
		if (fileTypeMask != XmFILE_DIRECTORY)
		    continue;
		break;
	    default:
		continue;
	    }
	}

	if (! InsertPath(dirData, path))	/* XXX was 'file' */
	    nAllocFailures++;
	else
	    nNames++;
    }
    if (nAllocFailures)
	WarningDialog(fs, NULL, "storage for %d file names failed!",
	    nAllocFailures);
    if (nTooLong)
	WarningDialog(fs, NULL, "%d file names were too long (>%d)!",
	    nTooLong, ComponentLength);
    CloseDir(vp);

    /*
     * Now create the XmStrings.
     */
    allocSize = sizeof(XmString) * nNames;
    dirData->xmNames = (XmString *)InsecureAlloc(allocSize);
    if (dirData->xmNames == NULL)
        AllocFailDialog(fs, NULL, "file name XmStrings", allocSize);

    i = nAllocFailures = 0;
    for (pName = dirData->nameList; pName; pName = pName->next) {
        dirData->xmNames[i] = XmStringCreateSimple(pName->name);
        if( dirData->xmNames[i] == NULL )
            nAllocFailures++;
	else
	    dirData->nNames = ++i;
    }
    if (nAllocFailures)
	WarningDialog(fs, NULL, "storage for %d XmString file names failed!",
	    nAllocFailures);

    /*
     * Finally, stick it into the widget
     */
    if (dirData->nNames) {
        XtVaSetValues(fs,
                XmNfileListItems,       dirData->xmNames,
                XmNfileListItemCount,   dirData->nNames,
                XmNdirSpec,             dirData->xmNames[0],
                XmNlistUpdated,         True,
                NULL);
    } else {
        XtVaSetValues(fs,
                XmNfileListItems,       NULL,
                XmNfileListItemCount,   0,
                XmNlistUpdated,         True,
                NULL);
    }
    dirData = NULL;	/* keep it from being freed, below */

complete:
    if (pattern)
	free((void *)pattern);
    if (directory)
	free((void *)directory);
    if (dirData)
	FreeDirData(fs, dirData);
}

/*
 * This can be used to extract a selected file from a file
 * dialog box.
 */
char *
StdGetFsbFile(Widget fs, int targetTypeMask, Boolean newFile)
{
    XmString pathString = NULL;
    char *path, *file;
    int type;
    InternalReasonPtr irp;
    XmStringCharSet charSet = (XmStringCharSet)XmSTRING_DEFAULT_CHARSET;

    XtVaGetValues(fs, XmNtextString, (XtPointer)&pathString, NULL);
    if (pathString == NULL) {
	WarningDialog(fs, NULL, "No path name.");
	return(NULL);
    }

    if (! XmStringGetLtoR(pathString, charSet, &path)) {
	WarningDialog(fs, NULL, "Can't get path string.");
	return(NULL);
    }

    type = FileType(path);
    file = ExtractPathSuffix(path);
    if (*file == '\0')
	file = path;
    if (! newFile) {
	if (type == 0) {
	    WarningDialog(fs, NULL,
		"Sorry, this path doesn't exist:\n   %s", path);
	    free((void *)path);
	    return(NULL);
	}

	if ((type & targetTypeMask) == 0) {
	    WarningDialog(fs, NULL, "%s is a %s, but it should be a %s",
		file, TypeName(type), TypeName(targetTypeMask));
	    free((void *)path);
	    return(NULL);
	}
    } else {
	MessageBoxInfo mbi;
	DialogResponse response;

	if (type) {
	    memset((char *)&mbi, 0, sizeof(mbi));
	    mbi.showOkButton = True;
	    mbi.showCancelButton = True;
	    response = WarningDialog(fs, &mbi,
		"%s already exists as a %s.\n\n%s",
                file, TypeName(type), "Create it anyway?");
	    if (response == kResponseCancel) {
		free((void *)path);
		return(NULL);
	    }
	    if (NULL != (irp = UnlinkFile(path))) {
		ReasonDialog(fs, NULL, irp);
		return(NULL);
	    }
	}
    }
    return(path);
}

static void
FsbProc(Widget fs, XtPointer fsbStruct)
{
    XmFileSelectionBoxCallbackStruct *fsb =
        (XmFileSelectionBoxCallbackStruct *)fsbStruct;

    /*
     * Eventually, we should specify a filter function to limit
     * the search to just the files we are interested in.
     */
    StdFileSearchForFsb(fs, fsb,
                        NULL,           /* filter function */
                        NULL);          /* user criterion */
}

#define FsbMagicNumber	0x12983476

typedef struct _fsbInfo {
    long        magic;		/* ... so that we can be sure its ours */
    boolean	createNewFile;
    Widget	fsb;
    Widget	toggles;
    Widget	newToggle;
    Widget	existing;
    Widget	fileList;
    Widget	okButton;
    char	*selectedFile;
} FsbInfo, *FsbInfoPtr;

static void
UpdateFsbState(FsbInfoPtr fsbData)
{
    XtSetSensitive(fsbData->fileList, ! fsbData->createNewFile);
    SetWidgetLabel(fsbData->okButton,
		   fsbData->createNewFile ? "Create" : "Open");
}

static void
FsbToggleCB(Widget w, XtPointer appArg, XtPointer toggleStruct)
{
    XmToggleButtonCallbackStruct *tb =
        (XmToggleButtonCallbackStruct *)toggleStruct;
    FsbInfoPtr fsbData = (FsbInfoPtr)appArg;

    fsbData->createNewFile = tb->set;
    UpdateFsbState(fsbData);
}

static void
FsbCB(Widget w, XtPointer appArg, XtPointer sbStruct)
{
    XmSelectionBoxCallbackStruct *sb =
        (XmSelectionBoxCallbackStruct *)sbStruct;
    FsbInfoPtr fsbData = (FsbInfoPtr)appArg;

    XtUnmanageChild(w);

    if (sb->reason == XmCR_OK) {
        fsbData->selectedFile =
	    StdGetFsbFile(w, FILE_REGULAR, fsbData->createNewFile);
    }
}

/*
 * This interface provides synchronous access to the file dialog box.
 * Essentially, the user calls the routine, and when it returns,
 * either the string returned points to a valid file name or it returns
 * NULL because the user pressed cancel.
 * 
 * Note that we hang user data off of the specified widget so that
 * we can get back to the file dialog box.  The reason is if 'w' is
 * a composite widget, the XtNameToWidget() will not return popup
 * children, which is what a file dialog box is.  Since all primitive
 * and manager widgets have the XmNuserData resource, this should not be
 * a problem.
 *
 * XXX this kind of 'user data' use should be extracted into some
 * common subroutines.
 */

char *
StdSelectFile(Widget w, char *name, char *winTitle, int selectMask)
{
    char fsbName[ 128 ];
    int nameLen = strlen(name);
    FsbInfoPtr fsbData;

    fsbData = (FsbInfoPtr)1;	/* and invalid value */
    XtVaGetValues(w, XmNuserData, (XtPointer)&fsbData, NULL);
    if (fsbData == (FsbInfoPtr)1) {
	WarningDialog(w, NULL,
	    "StdSelectFile: %s doesn't have a %s resource.",
	    GetWidgetPath(w, fsbName, sizeof(fsbName)));
	return(NULL);
    }
    if (fsbData && fsbData->magic != FsbMagicNumber) {
	WarningDialog(w, NULL,
	    "%s.%s resource has a bad magic number... already in use?",
	    GetWidgetPath(w, fsbName, sizeof(fsbName)),
	    XmNuserData);
	return(NULL);
    }

    if (fsbData == NULL) {
        Arg args[2];

	fsbData = (FsbInfoPtr)InsecureAlloc(sizeof(FsbInfo));
	if (fsbData == NULL) {
	    AllocFailDialog(w, NULL, "fsb data", sizeof(FsbInfo));
	    return(NULL);
	}

	if (nameLen > sizeof(fsbName) + 4)
	    nameLen = sizeof(fsbName) + 4;
	sprintf(fsbName, "fsb%.*s", nameLen, name);

        XtSetArg(args[0], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
        XtSetArg(args[1], XmNchildPlacement, XmPLACE_BELOW_SELECTION);
        fsbData->fsb = XmCreateFileSelectionDialog(w, fsbName, args, 2);
	fsbData->magic = FsbMagicNumber;
	fsbData->okButton =
	    XmFileSelectionBoxGetChild(fsbData->fsb, XmDIALOG_OK_BUTTON);
	fsbData->fileList =
	    XmFileSelectionBoxGetChild(fsbData->fsb, XmDIALOG_LIST);

        XtSetArg(args[0], XmNorientation, XmHORIZONTAL);
	fsbData->toggles = XmCreateRadioBox(fsbData->fsb, "toggles", args, 1);
	fsbData->newToggle =
	    XtVaCreateManagedWidget("New file", xmToggleButtonWidgetClass,
				    fsbData->toggles, NULL);
	fsbData->existing =
	    XtVaCreateManagedWidget("Existing file", xmToggleButtonWidgetClass,
				    fsbData->toggles, NULL);

        if (fsbData->fsb == NULL) {
            WarningDialog(w, NULL, "Can't create file dialog for %s\n", name);
	    free((char *)fsbData);
            return(NULL);
        }
	XtVaSetValues(w, XmNuserData, (XtPointer)fsbData, NULL);
	XtVaSetValues(fsbData->fsb, XmNfileSearchProc, FsbProc, NULL);
	XtAddCallback(fsbData->fsb, XmNokCallback, FsbCB, (XtPointer)fsbData);
	XtAddCallback(fsbData->fsb, XmNcancelCallback, FsbCB,
		      (XtPointer)fsbData);
	/*
	 * Note that we only have to set the callback on one of the
	 * toggles, because if one button is false, the other is true.
	 */
	XtAddCallback(fsbData->newToggle, XmNvalueChangedCallback,
		      FsbToggleCB, (XtPointer)fsbData);
    }

    fsbData->selectedFile = NULL;
    if (selectMask == (kSelectExistingFile | kSelectNewFile)) {
	XtManageChild(fsbData->toggles);
	fsbData->createNewFile = False;
    } else {
	XtUnmanageChild(fsbData->toggles);
	fsbData->createNewFile = (selectMask & kSelectNewFile) ? True : False;
    }
    XmToggleButtonSetState(fsbData->newToggle, fsbData->createNewFile, False);
    XmToggleButtonSetState(fsbData->existing, ! fsbData->createNewFile, False);

    UpdateFsbState(fsbData);
    XtVaSetValues(XtParent(fsbData->fsb), XmNtitle, winTitle, NULL);
    XtManageChild(fsbData->fsb);

    /*
     * Since the fsb is full application modal, it doesn't matter that
     * we are completely synchronous here.
     */
    while (XtIsManaged(fsbData->fsb)) {
	ProcessEvents(w);
    }

    return (fsbData->selectedFile);
}
