/*
 * Copyright (c) 1992, 1993, 1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	deckWidget.c - UI interface for generic play/record deck
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/deckPlayer.c,v 1.17 1995/08/22 20:08:07 toddb Exp $
 */

#include "nvrAlloc.h"
#include "fileAccess.h"
#include "PlayerCntl.h"
#include "WidgetDialog.h"
#include "WidgetUtil.h"
#include "WidgetFile.h"
#include "NvrDisplay.h"
#include <Xm/DrawingA.h>
#include <Xm/Label.h>

#include "deckPlayerI.h"

#define	DeckButtons \
    (kFastRevButton \
   | kPlayRevButton \
   | kStepRevButton \
   | kPauseButton \
   | kPlayButton \
   | kStepFwdButton \
   | kFastFwdButton)

static InternalReasonPtr DeckFastRev  (PlayerClass);
static InternalReasonPtr DeckStepRev  (PlayerClass);
static InternalReasonPtr DeckPlayRev  (PlayerClass);
static InternalReasonPtr DeckPlay     (PlayerClass, boolean record);
static InternalReasonPtr DeckPause    (PlayerClass);
static InternalReasonPtr DeckStepFwd  (PlayerClass);
static InternalReasonPtr DeckFastFwd  (PlayerClass);
static InternalReasonPtr DeckSeek     (PlayerClass, long frame);
static InternalReasonPtr DeckConfigure(PlayerClass, PlayerRunConfigPtr);

static void DeckFree(PlayerClass);

static PlayerClassRec DeckTemplate = {
    0,			/* magic number */
    DeckFastRev,	/* fast reverse */
    DeckStepRev,	/* step reverse */
    DeckPlayRev,	/* play reverse */
    DeckPlay,		/* play */
    DeckPause,		/* pause/stop */
    DeckStepFwd,	/* step forward */
    DeckFastFwd,	/* fast forward */
    DeckSeek,
    DeckConfigure,
    DeckFree
};

char *DefaultDeckFrameLabel = "Frame:";

char *DeckMagicName = DeckMagicNameDef;

static long
UpdatePlayerFromDeck(PlayerClass player)
{
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    ULong changed = 0;

    player->prodName = deckUI->deck->prodName;
    
    if (player->ability != deckUI->deck->ability) {
	player->ability = deckUI->deck->ability;
	changed |= kPlayerAbilityChanged;
    }
    if (player->width != deckUI->deck->width) {
	player->width = deckUI->deck->width;
	changed |= kPlayerWidthChanged;
    }
    if (player->height != deckUI->deck->height) {
	player->height = deckUI->deck->height;
	changed |= kPlayerHeightChanged;
    }
    if (player->frameLabel != deckUI->deck->frameLabel) {
	player->frameLabel = deckUI->deck->frameLabel;
	changed |= kPlayerFrameLabelChanged;
    }
    if (player->frameFormat != deckUI->deck->frameFormat) {
	player->frameFormat = deckUI->deck->frameFormat;
	changed |= kPlayerFrameFormatChanged;
    }
    if (player->frame != deckUI->deck->frame) {
	player->frame = deckUI->deck->frame;
	changed |= kPlayerFrameChanged;
    }
    if (player->minFrame != deckUI->deck->minFrame) {
	player->minFrame = deckUI->deck->minFrame;
	changed |= kPlayerMinFrameChanged;
    }
    if (player->maxFrame != deckUI->deck->maxFrame) {
	player->maxFrame = deckUI->deck->maxFrame;
	changed |= kPlayerMaxFrameChanged;
    }
    if (player->mediaAvailable != deckUI->deck->mediaAvailable) {
	player->mediaAvailable = deckUI->deck->mediaAvailable;
	changed |= kPlayerMediaAvailableChanged;
    }
    if (player->speed != deckUI->deck->speed) {
	player->speed = deckUI->deck->speed;
	changed |= kPlayerSpeedChanged;
    }
    if (player->minSpeed != deckUI->deck->minSpeed) {
	player->minSpeed = deckUI->deck->minSpeed;
	changed |= kPlayerMinSpeedChanged;
    }
    if (player->maxSpeed != deckUI->deck->maxSpeed) {
	player->maxSpeed = deckUI->deck->maxSpeed;
	changed |= kPlayerMaxSpeedChanged;
    }
    if (player->running != deckUI->deck->running) {
	player->running = deckUI->deck->running;
	changed |= kPlayerRunningChanged;
    }

    return(changed);
}

void
DeckPlayerChanged(DeckCntlPtr deck, void *cbData)
{
    PlayerClass player = (PlayerClass)cbData;
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    ULong changed;

    if (0 != (changed = UpdatePlayerFromDeck(player))) {
	(*deckUI->playerNotify)(player, changed);
	if (deckUI->controlNotify)
	    (*deckUI->controlNotify)(player, changed);
	if (deckUI->stripeNotify)
	    (*deckUI->stripeNotify)(player, changed);
    }
}

PlayerClass
AllocDeckPlayer(PlayerConfigPtr conf)
{
    PlayerClass player = NULL;
    int allocSize;
    DeckPlayerInfoPtr deckUI;
    
    allocSize = sizeof(PlayerClassRec) + sizeof(DeckPlayerInfo);

    player = (PlayerClass) InsecureAlloc(allocSize);
    if (player == NULL) {
	AllocFailDialog(conf->dpyParent, NULL, "Deck player class", allocSize);
	return(NULL);
    }
    memset((char *)player, 0, allocSize);
    *player = DeckTemplate;
    player->magic = StringToMagic(DeckMagicName);
    deckUI = (DeckPlayerInfoPtr)(player + 1);
    player->priv = (void *)deckUI;
    deckUI->playerNotify = conf->changedCB;

    return(player);
}

PlayerClass
CreateDeckPlayer(PlayerConfigPtr conf)
{
    PlayerClass player = NULL;
    DeckPlayerInfoPtr deckUI;
    DeckConfigPtr deckConf;
    DeckDescPtr deckDesc;
    InternalReasonPtr reason;
    XnvrDisplayType dpyType;

    deckConf = PopupDeckSelectConf(conf->dpyParent, conf->medium,
				   conf->ability, &reason);
    if (deckConf == NULL)
	return(NULL);
    deckDesc = deckConf->desc;

    player = AllocDeckPlayer(conf);
    if (player == NULL)
	return(NULL);

    deckUI = (DeckPlayerInfoPtr)player->priv;

    switch(deckDesc->medium) {
    case kHardwareDeck:
	dpyType = XnvrHwVideo;
	break;

    case kJpegDeck:
	if (conf->ability & kDeckIsDevice)
	    dpyType = XnvrHwJpeg;
	else
	    dpyType = XnvrSoftDisplay;
	break;

    case kMpegDeck:
	if (conf->ability & kDeckIsDevice)
	    dpyType = XnvrHwMpeg;
	else
	    dpyType = XnvrSoftDisplay;
	break;

    case kFileSetDeck:
	dpyType = XnvrSoftDisplay;
	break;
    }

    /*
     * Give it a default to start with.
     */
    player->width = 640;
    player->height = 480;

    {
	Arg args[ 2 ];
	int nArgs = 0;

	XtSetArg(args[ nArgs ], XmNwidth, player->width);   nArgs++;
	XtSetArg(args[ nArgs ], XmNheight, player->height); nArgs++;
	player->wDpy = CreateDisplayWidget(deckDesc->prodName, conf->dpyParent,
				     dpyType, args, nArgs, &reason);
	if (player->wDpy == NULL) {
	showFailureReason:
	    if (player)
		DeckFree(player);
	    ReasonDialog(conf->dpyParent, NULL, reason);
	    return(NULL);
	}
    }

    /*
     * now that we have the widget, set up the deck controller
     */
    deckConf->changedCB = DeckPlayerChanged;
    deckConf->cbData = (void *)player;
    deckConf->player = player;
    deckUI->deck = (*deckDesc->create)(deckConf, &reason);
    if (deckUI->deck == NULL)
	goto showFailureReason;

    (void) UpdatePlayerFromDeck(player);

    return(player);
}

static InternalReasonPtr
DeckStepRev(PlayerClass player)
{
    InternalReasonPtr reason;
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    DeckCntlPtr deck = deckUI->deck;

    XtVaSetValues(player->wDpy, XnvrNisActive, True, NULL);
    reason = (*deck->stepRev)(deck);
    XtVaSetValues(player->wDpy, XnvrNisActive, False, NULL);

    return(reason);
}

static InternalReasonPtr
DeckFastRev(PlayerClass player)
{
    InternalReasonPtr reason;
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    DeckCntlPtr deck = deckUI->deck;

    reason = (*deck->fastRev)(deck);
    XtVaSetValues(player->wDpy, XnvrNisActive, True, NULL);

    return(reason);
}

static InternalReasonPtr
DeckPlayRev(PlayerClass player)
{
    InternalReasonPtr reason;
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    DeckCntlPtr deck = deckUI->deck;

    reason = (*deck->playRev)(deck);
    XtVaSetValues(player->wDpy, XnvrNisActive, True, NULL);

    return(reason);
}

static InternalReasonPtr
DeckPlay(PlayerClass player, boolean record)
{
    InternalReasonPtr reason;
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    DeckCntlPtr pDeck = deckUI->deck;
    RecordType recordType;

    if (record) {
	DeckRunConfig conf;
	DeckCntlPtr rDeck = pDeck->rDeck;

	if (rDeck == NULL)
	    return(InternalProblem("DeckPlay: no record deck set.\n"));
	conf.configType = kChangeRecordType;
	recordType = conf.recordType = deckUI->recordType;
	if (NULL != (reason = (*rDeck->confDeck) (rDeck, &conf)))
	    return(reason);
    } else {
	recordType = kNoRecording;
    }
    reason = (*pDeck->play)(pDeck, recordType);
    XtVaSetValues(player->wDpy, XnvrNisActive, True, NULL);

    return(reason);
}

static InternalReasonPtr
DeckFastFwd(PlayerClass player)
{
    InternalReasonPtr reason;
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    DeckCntlPtr deck = deckUI->deck;

    reason = (*deck->fastFwd)(deck);
    XtVaSetValues(player->wDpy, XnvrNisActive, True, NULL);

    return(reason);
}

static InternalReasonPtr
DeckPause(PlayerClass player)
{
    InternalReasonPtr reason;
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    DeckCntlPtr deck = deckUI->deck;

    XtVaSetValues(player->wDpy, XnvrNisActive, False, NULL);
    reason = (*deck->pause)(deck);

    return(reason);
}

static InternalReasonPtr
DeckSeek(PlayerClass player, long frame)
{
    InternalReasonPtr reason;
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    DeckCntlPtr deck = deckUI->deck;

    reason = (*deck->seek)(deck, frame);

    return(reason);
}

static InternalReasonPtr
DeckStepFwd(PlayerClass player)
{
    InternalReasonPtr reason;
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    DeckCntlPtr deck = deckUI->deck;

    XtVaSetValues(player->wDpy, XnvrNisActive, True, NULL);
    reason = (*deck->stepFwd)(deck);
    XtVaSetValues(player->wDpy, XnvrNisActive, False, NULL);

    return(reason);
}

static InternalReasonPtr
DeckConfigure(PlayerClass player, PlayerRunConfigPtr playerConf)
{
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    DeckCntlPtr pDeck = deckUI->deck;
    DeckRunConfig conf;
    InternalReasonPtr reason = NULL;

    if (playerConf->configType & kChangePlayerMedia) {
	if (pDeck->ability & kDeckMediaIsFile) {
	    char winTitle[ 256 ];

	    sprintf(winTitle, "%s file selection", deckUI->deck->prodName);
	    /*
	     * Note that we use the parent, because the display widget
	     * may have a wild depth/visual that breaks Xt.
	     */
	    conf.path = StdSelectFile(XtParent(player->wDpy),
				      deckUI->deck->prodName,
				      winTitle,
				      kSelectNewFile | kSelectExistingFile);
	    if (conf.path) {
		conf.newMedia = (FileType(conf.path) == 0);
		conf.configType = kChangeMedia;
		reason = (*pDeck->confDeck) (pDeck, &conf);
		free(conf.path);
	    }
	} else {
	    PopupDeckControl(player);
	}
    }

    if (reason == NULL &&(playerConf->configType & kChangePlayerSelect)) {
	XtVaSetValues(player->wDpy, XnvrNisEnabled, False, NULL);
	conf.select = playerConf->select;
	conf.configType = kChangeSelect;
	if (NULL == (reason = (*pDeck->confDeck) (pDeck, &conf))) {
	    pDeck->selected = playerConf->select;

	    /*
	     * notify the recorder deck that we are no longer current.
	     * We do this by configuring the player as null, but we
	     * don't change the value of the player in the deck structure
	     * so that when we are reselected, we can renotify the recorder.
	     */
	    if (pDeck->rDeck) {
		conf.configType = kChangePlayer;
		conf.player = pDeck->selected ? pDeck->rDeck : NULL;
		reason = (*pDeck->rDeck->confDeck) (pDeck->rDeck, &conf);
	    }
	}
    }

    if (reason == NULL &&(playerConf->configType & kChangePlayerRecorder)) {
	DeckCntlPtr rDeck;

	if (playerConf->recorder) {
	    rDeck = ((DeckPlayerInfoPtr)playerConf->recorder->priv)->deck;
	    if (pDeck->ability & kDeckCanRealTime) {
		if (! ((pDeck->ability | rDeck->ability) & kDeckCanSlave)) {
		    reason = CreateReason("Sorry, neither %s nor %s %s.",
			pDeck->prodName,
			rDeck->prodName,
			"can slave to the other");
		} else if (pDeck->ability & kDeckCanSlave) {
		    deckUI->recordType = kRecording;
		} else {
		    deckUI->recordType = kSlaveRecording;
		}
	    } else {
		if ((rDeck->ability & kDeckFrameRecord) == 0) {
		    reason = CreateReason("Sorry, %s %s\nand %s %s.",
			pDeck->prodName,
			"produces frames in non-real-time",
			rDeck->prodName,
			"cannot do single frame recording");
		}
		deckUI->recordType = kFrameRecording;
	    }
	} else {
	    rDeck = NULL;
	}

	/*
	 * notify the old recorder
	 */
	if (pDeck->rDeck && pDeck->rDeck != rDeck && reason == NULL) {
	    conf.configType = kChangePlayer;
	    conf.player = NULL;
	    reason = (*pDeck->rDeck->confDeck) (pDeck->rDeck, &conf);
	    pDeck->rDeck->parent->player = NULL;
	    pDeck->rDeck->pDeck = NULL;
	}

	/*
	 * notify the player about the new recorder
	 */
	if (reason == NULL) {
	    conf.recorder = rDeck;
	    conf.configType = kChangeRecorder;
	    if (NULL == (reason = (*pDeck->confDeck) (pDeck, &conf))) {
		player->recorder = playerConf->recorder;
		pDeck->rDeck = rDeck;
	    }
	}

	/*
	 * notify the new recorder about the player
	 */
	if (rDeck && reason == NULL) {
	    conf.configType = kChangePlayer;
	    conf.player = pDeck;
	    if (NULL == (reason = (*rDeck->confDeck) (rDeck, &conf))) {
		playerConf->recorder->player = player;
		rDeck->pDeck = pDeck;
	    }
	}
    }

    if (reason == NULL &&(playerConf->configType & kChangePlayerSpeed)) {
	conf.configType = kChangeSpeed;
	conf.speed = playerConf->speed;
	if (NULL == (reason = (*pDeck->confDeck) (pDeck, &conf)))
	    player->speed = pDeck->speed;
    }

    if (reason == NULL &&(playerConf->configType & kChangePlayerSegment)) {
	conf.configType = kChangeSegment;
	conf.seg = playerConf->seg;
	if (NULL == (reason = (*pDeck->confDeck) (pDeck, &conf)))
	    player->edit = pDeck->edit;
    }

    return(reason);
}

static void
DeckFree(PlayerClass player)
{
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;

    if (player->wDpy) {
	XtDestroyWidget(player->wDpy);
	player->wDpy = NULL;	/* paranoid */
    }
    if (deckUI->deck) {
	(*deckUI->deck->close)(deckUI->deck);
	deckUI->deck = NULL;	/* paranoid */
    }
    if (deckUI->control) {
	XtDestroyWidget(deckUI->control);
	deckUI->control = NULL;	/* paranoid */
    }
    free((char *)player);
}

DeckCntlPtr
GetDeckFromPlayer(PlayerClass player)
{
    DeckPlayerInfoPtr deckUI = (DeckPlayerInfoPtr)player->priv;
    return( deckUI->deck );
}

void
CancelDeckRecord(DeckCntlPtr rDeck)
{
    PlayerClass player;
    InternalReasonPtr irp;

    player = rDeck->pDeck->parent;
    if (NULL != (irp = (*player->pause)(player)))
	ReasonDialog(player->wDpy, NULL, irp);
}
