/*
 * Copyright (c) 1992-1995 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	fastLabel.c - support for drawing labels very fast
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/fastLabel.c,v 1.1 1995/05/19 22:05:32 toddb Exp $
 */
#include "WidgetUtil.h"
#include "WidgetDebug.h"
#include "WidgetDialog.h"
#include "nvrAlloc.h"
#include <Xm/Xm.h>
#include <Xm/XmStrDefs.h>

typedef struct _fastFontInfo FastFontInfo, *FastFontInfoPtr;
struct _fastFontInfo {
    int refCount;		/* reference count */
    Display *dpy;		/* display pointer */
    XFontStruct font;		/* font extracted from widget */
    Pixmap glyphs;		/* pixmap of pre-rendered glyphs */
    short left[ 128 ];		/* x coord of each glyph */
    UChar widths[ 128 ];	/* width of each glyph */
    int pixWidth, pixHeight;	/* dimension of current pixmap */
    FastFontInfoPtr next;	/* next in list */
}; /* typedef: FastFontInfo, *FastFontInfoPtr; */

struct _fastLabelInfo {
    FastFontInfoPtr ff;		/* pointer to fast font data */
    GC gc;			/* gc for drawing into widget */
    Widget w;			/* this widget */
    Window win;			/* window of this widget */
    int x, y;			/* upper left of label in window */
    int labelWidth;		/* right-most edge of label */
    char *label;		/* last known label */
    boolean changed;		/* marked true before widget is updated */
    UChar alignment;		/* widget label alignment resource */
};

static FastFontInfoPtr FFList;

/****************************************************************************
 *
 * private interfaces
 *
 ****************************************************************************/

static void
AddNewGlyph(FastLabelInfoPtr flip, char glyph)
{
    FastFontInfoPtr ff = flip->ff;
    XGCValues values;
    Pixmap newGlyphs;
    XCharStruct glyphDim;
    int direction, ascent, descent;
    int width;
    GC bitgc;

    XTextExtents(&ff->font, &glyph, 1,
		 &direction, &ascent, &descent, &glyphDim);
    if (glyphDim.width == 0)
	return;
    if (ff->pixHeight == 0)
	ff->pixHeight = ascent + descent;

    width = ff->pixWidth + glyphDim.width;
    newGlyphs = XCreatePixmap(ff->dpy, flip->win,
			      width, ff->pixHeight,
			      1 /* depth 1 */);
    values.font = ff->font.fid;
    values.foreground = 1;
    values.background = 0;
    bitgc = XCreateGC(ff->dpy, newGlyphs,
		      GCFont | GCForeground | GCBackground,
		      &values);
    if (ff->pixWidth) {
	XCopyArea(ff->dpy, ff->glyphs, newGlyphs, bitgc,
		  0, 0,
		  ff->pixWidth, ff->pixHeight,
		  0, 0);
	XFreePixmap(ff->dpy, ff->glyphs);
    }

    XDrawImageString(ff->dpy, newGlyphs, bitgc,
		     ff->pixWidth, ff->font.ascent,	/* x, y */
		     &glyph,
		     1);			/* 1 glyph */
    XFreeGC(ff->dpy, bitgc);
    ff->glyphs = newGlyphs;
    ff->left[ (int)glyph ] = ff->pixWidth;
    ff->pixWidth += glyphDim.width;
    ff->widths[ (int)glyph ] = glyphDim.width;
}

static XFontStruct *
GetFontStruct(FastLabelInfoPtr flip, XmFontList fontList)
{
    XFontStruct *font = NULL;
    XmFontContext context;
    XmFontListEntry entry;
    XtPointer data;
    XmFontType type;

    if (! XmFontListInitFontContext(&context, fontList)) {
        AllocFailDialog(flip->w, NULL, "Font context", 1);
	return(False);
    }

    while (NULL != (entry = XmFontListNextEntry(context))) {
	data = XmFontListEntryGetFont(entry, &type);
	if (type == XmFONT_IS_FONT) {
	    font = (XFontStruct *)data;
	    break;
	} else {
	    InternalProblemDialog(flip->w, NULL,
		"Can't handle XFontSet structures.\n");
	    break;
	}
    }
    XmFontListFreeFontContext(context);

    return(font);
}

static void
CalculateCoord(FastLabelInfoPtr flip)
{
    Dimension width, height;
    XCharStruct overall;
    int direction, ascent, descent;

    XtVaGetValues(flip->w,
		  XmNalignment, &flip->alignment,
		  XmNwidth, &width,
		  XmNheight, &height,
		  NULL);

    XTextExtents(&flip->ff->font, flip->label, strlen(flip->label),
		 &direction, &ascent, &descent, &overall);

    flip->y = (height - (ascent + descent)) / 2;
    switch (flip->alignment) {
    case XmALIGNMENT_BEGINNING:
	flip->x = 0;
	break;

    case XmALIGNMENT_CENTER:
	flip->x = (width - overall.width) / 2;
	break;

    case XmALIGNMENT_END:
	flip->x = width - overall.width;
	break;
    }
}

static void
FastLabelEvent(Widget w,
               XtPointer closure,
               XEvent *event,
               Boolean *continueToDispatch)
{
    FastLabelInfoPtr flip = (FastLabelInfoPtr) closure;

    switch (event->type) {
    case Expose:
	if (flip->changed) {
	    SetWidgetLabel(w, flip->label);
	    flip->changed = False;
	}
        break;

    case ConfigureNotify:
	CalculateCoord(flip);
	break;
    }
}

static FastFontInfoPtr
AllocFastFont(FastLabelInfoPtr flip, XmFontList fontList)
{
    FastFontInfoPtr ff;
    XFontStruct *font;
    Display *dpy;
    int allocSize = sizeof(FastFontInfo);

    dpy = XtDisplay(flip->w);
    if (NULL == (font = GetFontStruct(flip, fontList)))
	return(NULL);

    for (ff = FFList; ff; ff = ff->next) {
	if (ff->dpy == dpy && font->fid == ff->font.fid) {
	    ff->refCount++;
	    return(ff);
	}
    }

    ff = (FastFontInfoPtr) InsecureAlloc(allocSize);
    if (ff == NULL) {
	AllocFailDialog(flip->w, NULL, "fast font storage", allocSize);
	return(NULL);
    }
    memset((char *)ff, 0, allocSize);
    ff->refCount = 1;
    ff->font = *font;
    ff->dpy = dpy;
    ff->next = FFList;
    FFList = ff;

    return(ff);
}

/****************************************************************************
 *
 * public interfaces
 *
 ****************************************************************************/

void
UpdateFastLabel(FastLabelInfoPtr flip, char *label)
{
    char *old, *new;
    int labelWidth, oldLabelWidth;
    int glyph, oldGlyph;
    boolean changed = flip->changed;
    FastFontInfoPtr ff = flip->ff;

    old = flip->label;
    new = label;
    labelWidth = 0;
    while (*old && *old == *new) {
	labelWidth += ff->widths[ (int)*new ];
	old++;
	new++;
    }

    oldLabelWidth = labelWidth;
    while (0 != (glyph = *new++)) {
	glyph &= 0x7f;
	/*
	 * Try to track what we overwrite... if we simply change a single
	 * character in the middle of the string, then we may not have
	 * to change anything else
	 */
	if (0 != (oldGlyph = *old)
	 && glyph == oldGlyph
	 && labelWidth == oldLabelWidth) {
	    oldLabelWidth += ff->widths[ glyph ];
	    labelWidth += ff->widths[ glyph ];
	    old++;
	    continue;
	}

	/*
	 * If the width is greater than zero, then we already have the glyph
	 */
	changed = True;
	if (ff->widths[ glyph ] == 0)
	    AddNewGlyph(flip, glyph);

	XCopyPlane(ff->dpy, ff->glyphs, flip->win, flip->gc,
		   ff->left[ glyph ], 0,		/* src x, y */
		   ff->widths[ glyph ],		/* width */
		   ff->pixHeight,			/* height */
		   flip->x + labelWidth, flip->y,	/* dst x, y */
		   1);					/* bit plane */
	labelWidth += ff->widths[ glyph ];
    }
    flip->changed = changed;

    if (flip->label)
	free(flip->label);
    flip->label = CopyOfString(label);

    if (labelWidth != flip->labelWidth) {
	SetWidgetLabel(flip->w, flip->label);
	flip->changed = False;
	flip->labelWidth = labelWidth;
    }
}

void
FreeFastLabel(FastLabelInfoPtr flip)
{
    FastFontInfoPtr ff = flip->ff;

    if (--ff->refCount == 0) {
	if (ff->glyphs)
	    XFreePixmap(ff->dpy, ff->glyphs);
	if (ff == FFList) {
	    FFList = ff->next;
	} else {
	    FastFontInfoPtr f;
	    for (f = FFList; f->next != ff; f = f->next)
		;
	    f->next = ff->next;
	}
	free((char *)ff);
    }

    if (flip->gc)
	XFreeGC(ff->dpy, flip->gc);
    if (flip->label)
	free(flip->label);
    free(flip);
}

FastLabelInfoPtr
AllocFastLabel(Widget w)
{
    FastLabelInfoPtr flip;
    int allocSize;
    XmString string = NULL;
    char *cString, *nullString;
    XGCValues values;
    XmFontList fontList;

    if (w == NULL)
	return(NULL);	/* FindWidget() probably failed */
    XtVaGetValues(w,
		  XmNlabelString, &string,
		  XmNforeground, &values.foreground,
		  XmNbackground, &values.background,
		  XmNfontList, &fontList,
		  NULL);

    if (string == NULL) {
	char widgetPath[ 256 ];
	InternalProblemDialog(w, NULL,
	    "AllocFastLabel: no label resource in \"%s\".",
	    GetWidgetPath(w, widgetPath, sizeof(widgetPath)));
	return(NULL);
    }
    XmStringGetLtoR(string, XmFONTLIST_DEFAULT_TAG, &cString);
    if (cString == NULL)
	cString = CopyOfString("");

    nullString = CopyOfString("");
    allocSize = sizeof(FastLabelInfo);
    flip = (FastLabelInfoPtr)InsecureAlloc(allocSize);
    if (nullString == NULL || cString == NULL || flip == NULL) {
	AllocFailDialog(w, NULL, "fast label storage", allocSize);
    abandon:
	if (nullString)
	    free(nullString);
	if (cString)
	    free(cString);
	if (flip)
	    free((char *)flip);
	return(NULL);
    }
    memset((char *)flip, 0, allocSize);
    flip->label = nullString;
    flip->w = w;
    flip->win = XtWindow(w);
    if (flip->win == 0) {
	XtRealizeWidget(w);
	flip->win = XtWindow(w);
    }
    
    if (NULL == (flip->ff = AllocFastFont(flip, fontList)))
	goto abandon;

    flip->gc = XCreateGC(flip->ff->dpy, flip->win,
			 GCForeground | GCBackground,
			 &values);

    XtAddEventHandler(w, StructureNotifyMask | ExposureMask,
                      FALSE, FastLabelEvent, (XtPointer)flip);
    CalculateCoord(flip);

    UpdateFastLabel(flip, cString);

    free(cString);
    return(flip);
}
