/*
 * Copyright (c) 1992,1993,1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	mpeg1vInit.c - various initialization for mpeg video decoding
 *
 * DESCRIPTION
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/mpeg1vinit.c,v 1.2 1995/04/04 22:36:54 toddb Exp $
 */

#define NeedFunctionPrototypes 1

#include "mpeg1vdecI.h"
#include "mpeg1huffman.h"
#include <stdlib.h>	/* for qsort */

typedef union {
    UShort *shortValue;
    UChar *charValue;
} UValue, *UValuePtr;

static int
HuffCodeMapCompare(const void *void1, const void *void2)
{
    HuffCodeMap *p1 = (HuffCodeMap *)void1;
    HuffCodeMap *p2 = (HuffCodeMap *)void2;
    return(p1->code - p2->code);
}

/*
 * prepare a table that can be addressed with MBAI_LOOKUP_BITS bits
 * to determine the bits consumed and the address increment.
 */
static InternalReasonPtr
InitLookupTable(HuffCodeMap initData[],
		int nInit,
		char *ptr,
		int nLookupCells,
		int nLookupBits,
		int lookupSize)
{
    int i, j, next, start;
    ULong value;
    UValue lookupTable;

    qsort((void *)initData, nInit, sizeof(HuffCodeMap), HuffCodeMapCompare);

    switch (lookupSize) {
    case sizeof(UChar):	lookupTable.charValue  = (UChar *) ptr; break;
    case sizeof(UShort):lookupTable.shortValue = (UShort *)ptr; break;
    }

    start = 0;
    value = 0;
    for (i = 0; i < nInit; i++) {
	next = initData[ i ].code;
	
	if (start < 0
	 || (next != 0 && start >= next)
	 || next >= nLookupCells)
	     return( InternalProblem("InitLookupTable: table init broken\n") );

	for (j = start; j < next; j++) {
	    switch (lookupSize) {
	    case sizeof(UChar):	lookupTable.charValue[ j ]  = value; break;
	    case sizeof(UShort):lookupTable.shortValue[ j ] = value; break;
	    }
	}
	start = next;
	value = initData[ i ].value;
    }

    for (j = start; j < nLookupCells; j++) {
	switch (lookupSize) {
	case sizeof(UChar): lookupTable.charValue[ j ]  = value; break;
	case sizeof(UShort):lookupTable.shortValue[ j ] = value; break;
	}
    }
    return(NULL);
}

/*
 * prepare a table that can be addressed with MBAI_LOOKUP_BITS bits
 * to determine the bits consumed and the address increment.
 */
static InternalReasonPtr
InitMBAddressInc(void)
{
    InternalReasonPtr reason;
    
    reason = InitLookupTable(MBAI_InitData, MBAI_INIT_CELLS,
		    (char *)HuffLookup.mbAddressInc,
		    MBAI_LOOKUP_CELLS, MBAI_LOOKUP_BITS,
		    sizeof(HuffLookup.mbAddressInc[0]));
    return(reason);
}

/*
 * prepare a table that can be addressed with MBT_B_LOOKUP_BITS bits
 * to determine the bits consumed and the address increment.
 */
static InternalReasonPtr
InitMBType(void)
{
    InternalReasonPtr reason;
    
    /*
     * ... for B pictures...
     */
    reason = InitLookupTable(MBT_B_InitData, MBT_B_INIT_CELLS,
		    (char *)HuffLookup.mb_B_Type,
		    MBT_B_LOOKUP_CELLS, MBT_B_LOOKUP_BITS,
		    sizeof(HuffLookup.mb_B_Type[0]));
    if (reason)
        return(reason);

    /*
     * ... for P pictures...
     */
    reason = InitLookupTable(MBT_P_InitData, MBT_P_INIT_CELLS,
		    (char *)HuffLookup.mb_P_Type,
		    MBT_P_LOOKUP_CELLS, MBT_P_LOOKUP_BITS,
		    sizeof(HuffLookup.mb_P_Type[0]));
    return(reason);
}

/*
 * prepare a table that can be addressed with MOTION_LOOKUP_BITS bits
 * to determine the bits consumed and the motion vector.
 */
static InternalReasonPtr
InitMotionVector(void)
{
    InternalReasonPtr reason;
    
    reason = InitLookupTable(MotionVectorInitData, MOTION_INIT_CELLS,
		    (char *)HuffLookup.motionVector,
		    MOTION_LOOKUP_CELLS, MOTION_LOOKUP_BITS,
		    sizeof(HuffLookup.motionVector[0]));
    return(reason);
}

/*
 * prepare a table that can be addressed with CODED_BLOCK_LOOKUP_BITS bits
 * to determine the bits consumed and the coded block value
 */
static InternalReasonPtr
InitCodedBlock(void)
{
    InternalReasonPtr reason;
    
    reason = InitLookupTable(CodedBlockInitData, CODED_BLOCK_INIT_CELLS,
		    (char *)HuffLookup.codedBlock,
		    CODED_BLOCK_LOOKUP_CELLS, CODED_BLOCK_LOOKUP_BITS,
		    sizeof(HuffLookup.codedBlock[0]));
    return(reason);
}

/*
 * prepare a tables that can be addressed with SIZE_LUMA_LOOKUP_CELLS bits
 * and SIZE_CHROMA_LOOKUP_CELLS bits to determine the bits consumed
 * and the size luma and chroma.
 */
static InternalReasonPtr
InitSizeYC(void)
{
    InternalReasonPtr reason;
    
    reason = InitLookupTable(SizeLuminanceInitData, SIZE_YC_INIT_CELLS,
		    (char *)HuffLookup.sizeLuminance,
		    SIZE_LUMA_LOOKUP_CELLS, SIZE_LUMA_LOOKUP_BITS,
		    sizeof(HuffLookup.sizeLuminance[0]));
    if (reason)
        return(reason);

    reason = InitLookupTable(SizeChrominanceInitData, SIZE_YC_INIT_CELLS,
		    (char *)HuffLookup.sizeChrominance,
		    SIZE_CHROMA_LOOKUP_CELLS, SIZE_CHROMA_LOOKUP_BITS,
		    sizeof(HuffLookup.sizeChrominance[0]));
    return(reason);
}

/*
 * prepare a tables that can be addressed with DCT_COEFF1_LOOKUP_CELLS bits
 * and DCT_COEFF2_LOOKUP_CELLS bits to determine the bits consumed
 * and the dct coefficients.
 */
static InternalReasonPtr
InitDCTCoeff(void)
{
    InternalReasonPtr reason;
    
    reason = InitLookupTable(DCTCoeff1InitData, DCT_COEFF1_INIT_CELLS,
		    (char *)HuffLookup.dctCoeff1,
		    DCT_COEFF1_LOOKUP_CELLS, DCT_COEFF1_LOOKUP_BITS,
		    sizeof(HuffLookup.dctCoeff1[0]));
    if (reason)
        return(reason);

    reason = InitLookupTable(DCTCoeff2InitData, DCT_COEFF2_INIT_CELLS,
		    (char *)HuffLookup.dctCoeff2,
		    DCT_COEFF2_LOOKUP_CELLS, DCT_COEFF2_LOOKUP_BITS,
		    sizeof(HuffLookup.dctCoeff2[0]));
    return(reason);
}

InternalReasonPtr
InitMpeg1VData(void)
{
    InternalReasonPtr reason;
    
    if (NULL != (reason = InitMBAddressInc()))   return (reason);
    if (NULL != (reason = InitMBType()))         return (reason);
    if (NULL != (reason = InitMotionVector()))   return (reason);
    if (NULL != (reason = InitCodedBlock()))     return (reason);
    if (NULL != (reason = InitSizeYC()))         return (reason);
    if (NULL != (reason = InitDCTCoeff()))       return (reason);
    
    return(NULL);
}
