/*
 * Copyright (c) 1992,1993,1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	mpeg1vpacket.c - main control for mpeg decoder
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/mpeg1vpacket.c,v 1.19 1996/01/03 02:42:51 toddb Exp $
 */

#include "mpeg1vdecI.h"
#include "nvrAlloc.h"

static InternalReason CantChangeMedia =
    { "Mpeg1VConfigure: cannot change media\n", IR_InternalProblem };

static InternalReason BadMediaValue =
    { "OpenMpeg1VideoStream: media is NULL\n", IR_InternalProblem };

static InternalReason BadOpsInitValue =
    { "OpenMpeg1VideoStream: opsInit is NULL\n", IR_InternalProblem };

static InternalReasonPtr
Mpeg1VideoDeliver(PacketLayerPtr pLayer, PacketDataPtr packet)
{
    QueuePacket(&pLayer->queue, packet);
    return(NULL);
}

/*
 * Save information about the beginning of a token in a given packet
 */
static void
SavePacketSeekState(Mpeg1VDecPrivPtr vdec, PacketDataPtr packet, long offset)
{
    PacketDataPtr pkt;

    if (vdec->seek.found) {
	return;
    }

    vdec->seek.startCode = vdec->parse.startCode;
    /* discard seekQueue packets */
    while( NULL != (pkt = vdec->seek.queue) ) {
	RemovePacket(&vdec->seek.queue, pkt);
	(*pkt->pLayer->ops.free)(pkt);
    }

    /* save packet offset - include startcode*/
    vdec->packet.offset = offset;
}
		    
static int
FindNextStart(Mpeg1VDecPrivPtr vdec, PacketDataPtr packet, UChar *pSrc)
{
    ULong startCode = vdec->parse.startCode;
    UChar *pSrcEnd = packet->buffer + packet->dataBytes;
    UChar *pTokenEnd;
    ULong prefix;

    pTokenEnd = pSrc;
    while (pTokenEnd < pSrcEnd) {
	prefix = startCode << 8;
	startCode = prefix | *pTokenEnd++;
	if (prefix == (StartCodePrefix << 8)) {
	    vdec->parse.nextStartCode = (MpegSmuxCodes) startCode;
	    break;
	}
    }

    vdec->parse.startCode = startCode;
    return( pTokenEnd - pSrc );
}

static InternalReasonPtr
DispatchNull(Mpeg1VDecPrivPtr vdec, UChar *pSrc)
{
    /* this should never happen */
    ProgMsg("Bad Seek State\n");
    return(NULL);
}

/*
  Mpeg1VideoDeliverInSeek()

	Parse the packets coming up from the layer below saving them on the
  seekQueue looking for a valid start point:
  1) sequence header
  2) Gop (after we've seen a seqHdr)
  3) I picture (after we've seen a seqHdr)
	When found put packet containing start point on data queue and
  reset deliver routine. Keep track of the pts of start point.

   Display Order (timeRef):
   ... B0 B1 I2 B2 B3 P4 B5 B6 P7 B0 B1 I2 B2 B3 P4 B5 B6 P7 B0 B1 I2 ...


   Transmit Order (timeRef):
   ... I2 B0 B1 P4 B2 B3 P7 B5 B6 I2 B0 B1 P4 B2 B3 P7 B5 B6 I2 B0 B1 ...

   The PTS of a given packet refers to the first access unit (picture)
   in the packet. If that picture is not an I picture, then the PTS of
   the desired I picture must be calculated. The formula to use is:

   PTS(packet) + pictureTime * (timeRef(max) - timeRef(first) + 1 + timeRef(I))
*/
static InternalReasonPtr
Mpeg1VideoDeliverInSeek(PacketLayerPtr pLayer, PacketDataPtr packet)
{
    Mpeg1VDecPrivPtr vdec = (Mpeg1VDecPrivPtr)pLayer;

    UChar *pSrc, *pSrcEnd;
    int tokenLength, available;
    PacketDataPtr pkt;
    boolean complete = False;
    InternalReasonPtr reason = NULL;
    static InternalReason BadSeekState = 
        { "Mpeg1VideoDeliverInSeek: Bad Seek State\n" };

    pSrc = packet->buffer;
    pSrcEnd = packet->buffer + packet->dataBytes;
    if (vdec->parse.timeCodes && vdec->parse.pktTimecode != packet->timecode) {
	vdec->parse.pktTimecode = packet->timecode;
	vdec->parse.newTimeRef = True;
    }

    while (pSrc < pSrcEnd) {

	switch (vdec->parse.token & kTokenTypeMask) {
	case kTokenStartCode:

	    tokenLength = FindNextStart(vdec, packet, pSrc);
	    pSrc += tokenLength;
	    if ((vdec->parse.startCode >> 8) != StartCodePrefix) {
		continue;
	    }
	    SavePacketSeekState(vdec, packet, pSrc - packet->buffer - 4);
	    break;
 	    
	case kTokenLenCopy:
	    tokenLength = vdec->parse.length - vdec->parse.offset;
	    available = pSrcEnd - pSrc;
	    if (available > tokenLength) {
		available = tokenLength;
	    }
	    memcpy((char *)vdec->parse.pToken + vdec->parse.offset,
		   (char *)pSrc,
		   available);
	    vdec->parse.offset += available;
	    pSrc += available;
	    if (vdec->parse.offset >= vdec->parse.length) {
		break;
	    }
	    continue;
 
	default:    /* not possible */
	    reason = &BadSeekState;
	    goto parsingComplete;
	}

	switch(vdec->parse.token) {
	case kSeqSearch:
	    switch(vdec->parse.nextStartCode) {
	    case PictureStartCode:
		SetTokenMacro(vdec, kPicHeader, DispatchNull);
		vdec->parse.length = Mpeg1PictureHdrLen;
		vdec->parse.offset = 4;
		vdec->parse.pToken = (UChar *)&vdec->tokens.picHdr;
		break;

	    case GroupStartCode:
		SetTokenMacro(vdec, kGopHeader, DispatchNull);
		vdec->parse.length = Mpeg1GroupHdrLen;
		vdec->parse.offset = 4;
		vdec->parse.pToken = (UChar *)&vdec->tokens.gopHdr;
		if (vdec->parse.seqHeaderSeen) {
		    vdec->seek.found = True;
		}
		break;

	    case SeqHeaderStartCode:
		vdec->seek.found = True;
		break;

	    default:
		break;
	    }

	    vdec->parse.startCode = ~0;
	    break;

	case kGopHeader:
	    if (! vdec->parse.timeCodes) {
		Mpeg1GopHdrPtr gop = &vdec->tokens.gopHdr;
		long seconds = 3600 * gop->hours
			     + 60 * gop->minutes
			     + gop->seconds;

		vdec->parse.pktTimecode = seconds * MilliSecPerSecond
                            + gop->pictures * vdec->parse.msPerPicture;
	    }
	    vdec->parse.startCode = ~0;
	    SetTokenMacro(vdec, kSeqSearch, DispatchNull);
	    break;
	    
	    
	case kPicHeader:
	    if (vdec->parse.timeCodes) {
		if (vdec->parse.newTimeRef) {
		    vdec->parse.newTimeRef = False;
		    vdec->parse.firstPic = vdec->tokens.picHdr.timeRef;
		}
	    }
	    if (    vdec->parse.seqHeaderSeen
		&& vdec->tokens.picHdr.type == Mpeg1_I_Pic ) {
		
		complete = True;
		goto parsingComplete;
	    }
	    vdec->parse.startCode = ~0;
	    SetTokenMacro(vdec, kSeqSearch, DispatchNull);
	    break;

	default:    /* not possible */
	    reason = &BadSeekState;
	    goto parsingComplete;
	}
    }

 parsingComplete:

    if (reason) {
	while( NULL != (pkt = vdec->seek.queue) ) {     /* free seek packets */
	    RemovePacket(&vdec->seek.queue, pkt);
	    (*pkt->pLayer->ops.free)(pkt);
	}
	pLayer->ops.deliver = Mpeg1VideoDeliver;
	SetTokenMacro(vdec, kSeqSearch, (  vdec->parse.seqHeaderSeen
					 ? vdec->ops.picDispatch
					 : vdec->ops.seqDispatch));
	return(reason);
    }

    if (complete) {

	/* seek pkts to data queue */
	while( NULL != (pkt = vdec->seek.queue) ) {
	    RemovePacket(&vdec->seek.queue, pkt);
	    QueuePacket(&pLayer->queue, pkt);
	}

	/* queue this packet */
	QueuePacket(&pLayer->queue, packet);

	/* reset deliver state */
	pLayer->ops.deliver = Mpeg1VideoDeliver;

	/* reset parsing/decoding state */
	SetTokenMacro(vdec, (vdec->seek.startCode | kTokenLenCopy),
		      vdec->ops.picDispatch);
	vdec->parse.startCode = vdec->seek.startCode;
	vdec->parse.nextStartCode = (MpegSmuxCodes) vdec->seek.startCode;
	vdec->parse.length = vdec->parse.offset = 0;

	if (vdec->parse.timeCodes) {
	    long tcDiff = vdec->tokens.picHdr.timeRef - vdec->parse.firstPic;
            vdec->parse.timecode = vdec->parse.pktTimecode
				 + tcDiff * vdec->parse.msPerPicture;
	} else {
	    vdec->parse.timecode = vdec->parse.pktTimecode
                     + vdec->tokens.picHdr.timeRef * vdec->parse.msPerPicture;
	}
	if (vdec->cb.select & Mpeg1VPtsSeekMask) {
	    vdec->cb.arg.seekTimecode = vdec->parse.timecode;
            (*vdec->cb.tokenCb)(&vdec->pub, Mpeg1VPtsSeekCb, &vdec->cb.arg);
	}
    } else {
	QueuePacket(&vdec->seek.queue, packet);  /* queue to seek queue */
    }
    
    return(NULL);
}

static InternalReasonPtr
Mpeg1VideoFlush(PacketLayerPtr pLayer)
{
    Mpeg1VDecPrivPtr vdec = (Mpeg1VDecPrivPtr)pLayer;
    YcMpegImagePtr image;

    if (NULL != (image = vdec->currentPict->image)) {
	Mpeg1VDeliverImage(vdec, image);
    }
    if (NULL != (image = vdec->futurePict->image)) {
	Mpeg1VDeliverImage(vdec, image);
   }

    return( NULL );
}

/*
 * Note: 'absOffset' parameter is always ignored because the *real* offset is
 *	determined by the media layer which hasn't been called yet
 *	and which will almost certainly change the offset value
 */
static InternalReasonPtr
Mpeg1VideoSeek(PacketLayerPtr pLayer, ULong absOffset)
{
    Mpeg1VDecPrivPtr vdec = (Mpeg1VDecPrivPtr)pLayer;
    YcMpegImagePtr yci;

    vdec->parse.startCode = ~0;
    pLayer->ops.process = vdec->ops.startCodeProcess;
    pLayer->ops.deliver = Mpeg1VideoDeliverInSeek;
    SetTokenMacro(vdec, kSeqSearch, (  vdec->parse.seqHeaderSeen
					 ? vdec->ops.picDispatch
					 : vdec->ops.seqDispatch));
    ClearDctArena(vdec->slice.idct);
    vdec->parse.pktTimecode = 0;
    vdec->parse.newTimeRef = True;
    vdec->seek.found = False;
    if (vdec->pastPict && (NULL != (yci = vdec->pastPict->image))) {
	(*yci->pub.free)(&yci->pub);
	vdec->pastPict->image = NULL;
    }
    if (vdec->futurePict && (NULL != (yci = vdec->futurePict->image))) {
	(*yci->pub.free)(&yci->pub);
	vdec->futurePict->image = NULL;
    }
    if (vdec->currentPict && (NULL != (yci = vdec->currentPict->image))) {
	(*yci->pub.free)(&yci->pub);
	vdec->currentPict->image = NULL;
    }

    return( NULL );
}

static void
Mpeg1VideoClose(PacketLayerPtr pLayer)
{
    Mpeg1VDecPrivPtr vdec = (Mpeg1VDecPrivPtr)pLayer;
    YcMpegImagePtr yci, next;

    FreePacketList(&pLayer->freeList);
    pLayer->freeList = NULL;
    FreePacketList(&pLayer->queue);
    pLayer->queue = NULL;
    if (vdec->parse.copyBuf)
	free((char *)vdec->parse.copyBuf);
    if (vdec->slice.curDctList) {
        QueueDctArena(vdec->slice.idct, vdec->slice.curDctList);
        vdec->slice.curDctList = NULL;
    }
    FreeIdctParam(vdec->slice.idct);

    /*
     * Free pictures on the free list.
     */
    for (yci = vdec->ycFree; yci; yci = next) {
	next = (YcMpegImagePtr) yci->pub.next;
	free((char *)yci);
    }
    if (vdec->currentPict->image)
	free((char *)vdec->currentPict->image);
    if (vdec->futurePict->image)
	free((char *)vdec->futurePict->image);
    if (vdec->pastPict->image)
	free((char *)vdec->pastPict->image);

    (*pLayer->down->ops.close)(pLayer->down);
    free((char *)pLayer);
}

static void
Mpeg1VideoFree(PacketDataPtr packet)
{
    /* we don't allocate any packets */
}

InternalReasonPtr
Mpeg1VRunConfigure(PacketLayerPtr pLayer,
		   Mpeg1VideoRunConfPtr conf,
		   long confMasks)
{
    Mpeg1VDecPrivPtr vdec = (Mpeg1VDecPrivPtr) pLayer;

    if (confMasks & Mpeg1VConfMediaMask)
	return( &CantChangeMedia );
    if (confMasks & Mpeg1VConfWarningMask)
	vdec->warning = conf->core.warning;
    if (confMasks & Mpeg1VConfTokenCbMask)
	vdec->cb.tokenCb = conf->core.tokenCb;
    if (confMasks & Mpeg1VConfCbDataMask)
	vdec->cb.arg.cbData = conf->core.cbData;
    if (confMasks & Mpeg1VConfOpsInitMask)
	(*conf->core.opsInit->init)((Mpeg1VDecPrivPtr) pLayer);
    if (confMasks & Mpeg1VConfCbSelectMask)
	vdec->cb.select = conf->core.cbSelect;
    if (confMasks & Mpeg1VConfFormatMask)
	vdec->cb.maxLevel = conf->core.formatLevel;
    if (confMasks & Mpeg1VConfWeightMask) {
	vdec->decode.weight = conf->weight;
	vdec->decode.dropFrames = (vdec->decode.weight < 1.0);
	vdec->decode.dropRate = (float)1.0 / vdec->decode.weight;
	/*
	 * if the next picture is a B picture this will cause it to be
	 * dropped; if a P or I picture, it will be kept.
	 */
	vdec->decode.keepCounter = vdec->decode.dropRate * (float)0.5;
    }
    return(NULL);
}

PacketLayerPtr
OpenMpeg1VideoStream(Mpeg1VideoConfPtr conf, InternalReasonPtr *pReason)
{
    Mpeg1VDecPrivPtr vdec;
    PacketLayerPtr pLayer;
    int allocSize;
    PictureInfoPtr pic;
    IdctParamPtr idct;

    if (conf->opsInit == NULL) {
        *pReason = &BadOpsInitValue;
        return(NULL);
    }
    if (conf->media == NULL) {
        *pReason = &BadMediaValue;
        return(NULL);
    }
    if (NULL != (*pReason = InitMpeg1VData()))
        return(NULL);

    idct = CreateIdctParam(conf->idctMethod, pReason);
    if (idct == NULL)
        return(NULL);
    (*idct->newQuantMatrices)(idct,
                              DefaultIntraQMatrix,
                              DefaultNonIntraQMatrix);
    
    allocSize = NFreePictures * sizeof(PictureInfo) + sizeof(Mpeg1VDecPriv);
    pLayer = AllocPacketLayer(PacketMpeg1Video, conf->media, allocSize);
    if (pLayer == NULL) {
	*pReason = AllocReason("Mpeg 1 Video packet layer",
	                       sizeof(PacketLayer) + allocSize);
	free((char *)idct);
	return(NULL);
    }
    vdec = (Mpeg1VDecPrivPtr)pLayer;
    pic = vdec->picArray = (PictureInfoPtr)(vdec + 1);
    vdec->currentPict = pic++;
    vdec->futurePict = pic++;
    vdec->pastPict = pic;

    /*
     * Assign the defaults and then run the configuration
     */
    vdec->decode.dropRate = (float)1.0;
    vdec->decode.weight = (float)1.0;
    (void) Mpeg1VRunConfigure(pLayer, (Mpeg1VideoRunConfPtr) conf,
			      Mpeg1VConfOpenMask & (~Mpeg1VConfMediaMask));

    vdec->slice.idct = idct;
    vdec->parse.startCode = ~0;
    vdec->cb.arg.pTokens = &vdec->tokens;
    vdec->cb.arg.idctp = idct;

    vdec->tokens.seqHdr.seqHdrCode = SeqHeaderStartCode;
    vdec->tokens.picHdr.picStartCode = PictureStartCode;
    vdec->tokens.gopHdr.gopStartCode = GroupStartCode;
    pLayer->ops.deliver = Mpeg1VideoDeliver;
    pLayer->ops.process = vdec->ops.startCodeProcess;
    pLayer->ops.flush   = Mpeg1VideoFlush;
    pLayer->ops.close   = Mpeg1VideoClose;
    pLayer->ops.seek    = Mpeg1VideoSeek;
    pLayer->ops.free    = Mpeg1VideoFree;
    
    SetTokenMacro(vdec, kSeqSearch, vdec->ops.seqDispatch); /* where we start */

    vdec->parse.timeCodes = pLayer->down->timeCodes;
    if (vdec->parse.timeCodes) {
	vdec->parse.pktTimecode = -1.0;
    } else {
	vdec->parse.pktTimecode = 0.0;
    }

    return(pLayer);
}

static void
Mpeg1VFreeYcImage(YcImagePtr image)
{
    YcMpegImagePtr yci = (YcMpegImagePtr)image;
    Mpeg1VDecPrivPtr vdec = yci->vdec;

    if (--yci->refCount == 0) {
	if (yci->validate != vdec->imageValidate) {
	    free((char *)yci);
	    DebugTask(VMD_PicAlloc,
		Msg("image %d @0x%x freed, ref=%d\n",
		    image->frame, image, yci->refCount);
	    )
	} else {
	    image->next = & vdec->ycFree->pub;
	    vdec->ycFree = (YcMpegImagePtr) image;
	    DebugTask(VMD_PicAlloc,
		Msg("image %d @0x%x put on free list, ref=%d\n",
		    image->frame, image, yci->refCount);
	    )
	}
    } else {
	DebugTask(VMD_PicAlloc,
	    Msg("image %d @0x%x ref=%d\n", image->frame, image, yci->refCount);
	)
    }
}

YcMpegImagePtr
Mpeg1VAllocYcImage(Mpeg1VDecPrivPtr vdec, InternalReasonPtr *pIrp)
{
    YcMpegImagePtr yci;
    int mbArea, lumaSize, chromaSize, allocSize;
    UChar *pixels;

    if (vdec->ycFree) {
        yci = vdec->ycFree;
        vdec->ycFree = (YcMpegImagePtr)yci->pub.next;
        yci->refCount = 1;
        DebugTask(VMD_PicAlloc, Msg("image @0x%x from free list\n", yci))
        return(yci);
    }

    mbArea = vdec->decode.lastMacroBlock + 1;
    lumaSize = mbArea * 16 * 16;
    chromaSize = mbArea * 8 * 8;
    allocSize = sizeof(YcMpegImage) + lumaSize + 2*chromaSize;

    yci = (YcMpegImagePtr) AllocNamedMem(allocSize, "mpeg picture data", pIrp);
    if (yci) {
        pixels = (UChar *)(yci + 1);
        yci->pub.y = pixels; pixels += lumaSize;
        yci->pub.cr = pixels; pixels += chromaSize;
        yci->pub.cb = pixels;
        yci->pub.sample = Sample4_2_0;
        yci->pub.lumaWidth = vdec->decode.mbWidth * 16;
        yci->pub.lumaHeight = vdec->decode.mbHeight * 16;
        yci->pub.chromaWidth = vdec->decode.mbWidth * 8;
	yci->pub.free = Mpeg1VFreeYcImage;
        yci->vdec = vdec;
        yci->refCount = 1;
        yci->validate = vdec->imageValidate;
    }
    DebugTask(VMD_PicAlloc, Msg("image @0x%x allocated\n", yci))

    return(yci);
}

void
Mpeg1VDeliverImage(Mpeg1VDecPrivPtr vdec, YcMpegImagePtr image)
{
    if (! image->delivered) {

        if (vdec->cb.select & Mpeg1VDisplayMask) {
            DebugTask(VMD_DisplayCb, Msg("dsplay %2d\n", image->pub.frame))
            image->refCount++;
            vdec->cb.arg.image = &image->pub;
            (*vdec->cb.tokenCb)(&vdec->pub, Mpeg1VDisplayCb, &vdec->cb.arg);
        }
        image->delivered = True;
    }
}
