/*
 *      Copyright (c) 1992, North Valley Research, Inc.
 *      Confidential -- All proprietary rights reserved.
 *      Licensed Material -- Property of North Valley Research, Inc.
 *
 *      This program is the property of North Valley Research, Inc. or others
 *      from whom North Valley Research has obtained a licensing right, and is
 *      considered by North Valley Research to be confidential.  It is protected
 *      by U.S. copyright law as an unpublished work and is
 *      furnished pursuant to a written license agreement.  It may
 *      not be used, copied or otherwise reproduced, or disclosed to
 *      others except in accordance with the terms and conditions of
 *      that agreement.
 *
 * NAME
 *   sysdec.c		MPEG-1 system stream de-multiplexor
 *
 * DESCRIPTION
 *
 *    The procedures in the file implement the MPEG-1 system demultiplexor
 *    with two modes of operation: (1) temporally correct and (2) data driven.
 *    In the first case, a call back routine is used to schedule the arrival
 *    of the incomming data such that the STD buffers are implicitly simulated.
 *
 *    In the second case data is read on demand from the elementary stream
 *    read functions.  Temporal skew between the streams is ignored.
 *
 *    Data copying is minimized by passing pointers to actual read buffers
 *    back from the callers of the elementary stream read functions.
 * $Header: /usr/local/src/nvrToMIT/RCS/mpegDemux.c,v 2.3 1995/10/23 15:56:15 toddb Exp $
 */

/*
 * NOTE: This is a shared module. It is shared between the video
 *	decompressor, and the audio decompress library.
 *	If any conditional compilation directives are added to this file,
 *	the Imakefiles for those two items must be updated synchronously
 */

#define NeedFunctionPrototypes	1

#include "osio.h"
#include "nvrReason.h"
#include "nvrDebug.h"
#include "nvrString.h"
#include "nvrAlloc.h"
#include "mpegSysDecode.h"

#define	FDBSIZE	   	4096		/* Size of FileDataBlk buffer */
#define	NFDBALLOC	5		/* Initial number of FileDataBlk */
#define	NSDRALLOC	10		/* Initial number of StreamDataRange */
#define	IDBASE		0xC0		/* lowest ID for packet start codes */
#define	MAXAUDIO	32		/* Max number audio streams */
#define	MAXVIDEO	16		/* Max number video streams */
#define	MAXDATA		16		/* Max number data streams */
#define	NSTREAM		68		/* Maximum number of streams */
#define	SYSCLKF		90000.0		/* System clock frequency (Hz) */
#define	POLLTIME	0.0001		/* Poll time if muxrate is not defined*/
#define	MAXBAK		6		/* Max bytes trashed prior to payload */

#define	L_BYTE		3		/* Log 2 of bits per byte */

#define	B_STARTCODE	0x000001	/* Startcode prefix */
#define	B_SC_SYSEND	0xB9		/* System end code */
#define	B_SC_PACK	0xBA		/* last byte of pack startcode */
#define	B_SC_SYSHDR	0xBB		/* last byte of sys header startcode */
#define	B_SC_RESERVED	0xBC		/* Reserved stream */
#define	B_SC_PRIVATE1	0xBD		/* Private stream 1 */
#define	B_SC_PADDING	0xBE		/* Padding stream */
#define	B_SC_PRIVATE2	0xBF		/* Private stream 2 */
#define	B_SC_AUDIO	0xC0		/* Base audio stream */
#define	B_SC_VIDEO	0xE0		/* Base video stream */
#define	B_SC_DATA	0xF0		/* Base data stream */

typedef struct _fileDataBlk * FileDataBlkPtr;
typedef struct _streamDataRange * StreamDataRangePtr;
typedef struct _stream * StreamPtr;
typedef	struct _system * SystemPtr;

typedef struct _fileDataBlk {
    FileDataBlkPtr	next;		/* Next pointer */
    short		refcnt;		/* Reference count */
    short		count;		/* Number of bytes in file buffer */
    UChar		backup[MAXBAK];	/* Backup region for splicing streams */
    UChar		data[FDBSIZE];	/* File buffer */
} FileDataBlk;


typedef struct _streamDataRange {
    StreamDataRangePtr	 next;		/* Next StreamDataRange */
    FileDataBlkPtr	 fdb;		/* This FileDataBlock contains range */
    UChar		*start;		/* Starting byte */
    short		 length;	/* Number of bytes in range */
    long		 pos;		/* Byte offset to .start[0] */
    TimeSpec		 pts;		/* Presentation timestamp */
    TimeSpec		 dts;		/* Decode timestamp */
} StreamDataRange;

typedef struct _stream {
    SystemPtr		 sys;		/* System state structure */
    StreamPtr		 next;		/* Next stream */
    BitstreamType	 type;		/* Video, Audio or Data stream */
    UChar		 id;		/* Packet startcode LSB */
    UChar		 flags;		/* Flags: (STR_* constants) */
    StreamDataRangePtr	 sdrin;		/* In ptr for buffered data */
    StreamDataRangePtr	 sdrout;	/* Out ptr for buffered data */
    StreamDataRangePtr	 reading;	/* Currently been consumed */
    int			 bufocc;	/* Stream buffer occupancy (bytes) */
    int			 bufsize;	/* Stream STD buffer size (bytes) */
} Stream;

#define	STR_UFLO	0x01		/* Buffer underflow in progress */
#define	STR_OFLO	0x02		/* Buffer overflow is in progress */

typedef struct _system {
    char		*name;		/* System stream input filename */
    FileHandle		 handle;	/* File handle of system input stream */
    char		 flags;		/* system flags (SYS_* constants) */
    char		 state;		/* Input state (a SystemStateCode) */
    UChar		 id;		/* last byte of last system startcode */
    char		 audiobound;	/* Max audio channels */
    char		 videobound;	/* Max video channels */
    char		 hdrflags;	/* System header flags (HDR_* consts) */
    char		 pktflags;	/* Packet flags (PKT_* consts) */
    int			 bufbound[NSTREAM]; /* Maximum buffer sizes (bytes) */
    long		 pos;		/* Bytes of system stream read sofar */
    FileDataBlkPtr	 fdbfree;	/* FileDataBlk free list */
    StreamDataRangePtr	 sdrfree;	/* StreamDataRange free list */
    StreamPtr		 streams[NSTREAM]; /* Stream pointers */
    StreamPtr		 unspec;	/* Stream with no specific ID */
    FileDataBlkPtr	 fdb;		/* File data block being parsed */
    UChar		*nextbyte;	/* Next packet in FSB */
    SystemReadFunc	 readFunc;	/* System read function */
    void	       (*scheduler)(	/* Temporal/data-ready callback sched */
		void (*)(),		/* Callback function pointer */
		PrivPointer,		/* Argument to the callback function */
		TimeSpecPtr,		/* Callback at this time or earlier */
		FileHandle);		/* Callback when this file is readable*/
    void	    (*startAccessUnit)( /* Start of access unit callback */
		MpegStreamInfoPtr);	/* Stream of start of access unit */
    double		 muxrate;	/* Mux rate (bits/second) */
    double		 timescale;	/* Real time multiplier */
    double		 scr;		/* Most recent SCR */
    double		 muxbound;	/* Muxrate bound */
    double		 pts;		/* Temporary PTS stg */
    double		 dts;		/* Temporary DTS stg */
    double		 maxbehind;	/* limit on how far behind real time */
    long		 scrpos;	/* Position of last byte of SCR */
    long		 pktloc;	/* Loc of pkt startcode for cur pkt */
    int			 tmp;		/* Tmp use to assemble fields */
    int			 bufsize;	/* Buffer size of current packet */
    UShort		 packetsize;	/* Size of current pkt (decreasing) */
    UShort		 fullsize;	/* Orignial packet size */
    UShort		 padding;	/* Nbr bytes of padding */
    short		 openstreams;	/* Number of streams currently open */
    short		 nhdrbufspec;	/* Nbr buffer size bounds in header */
} System;

#define	SYS_PACK	0x01		/* Pack header encountered */
#define	SYS_HEADER	0x02		/* System header encountered */
#define	SYS_EOF		0x04		/* System stream EOF encountered */
#define	SYS_ERROR	0x08		/* Seaching for next start after error*/
#define	SYS_STARTUP	0x10		/* Searchinf for pack & sys header */

#define	HDR_FIXED	0x01		/* Fixed bitrate flag */
#define	HDR_CSPS	0x02		/* Constrained system parameter flag */
#define	HDR_AUDIOLOCK	0x04		/* Audio clock locked to system clock */
#define	HDR_VIDEOLOCK	0x08		/* Video clock locked to system clock */

#define	PKT_NEWPKT	0x01		/* New packet (vs leftover partial) */
#define	PKT_BUFSIZE	0x02		/* Buffer size specified */
#define	PKT_PTS		0x04		/* PTS specified */
#define	PKT_DTS		0x08		/* DTS specified */

enum {
    SYS_search,			/*  0 Seaching for 1st byte of startcode */
    SYS_pkt0,			/*  1 Looking at 2nd byte of startcode */
    SYS_pkt1,			/*  2 Looking at 3rd byte of startcode */
    SYS_pkt2,			/*  3 Looking at 4th byte of startcode */
    SYS_pkt3,			/*  4 Looking at 1st byte of packet size */
    SYS_pkt4,			/*  5 Looking at 2nd byte of packet size */
    SYS_pkt5,			/*  6 packet state 5 */
    SYS_pkt6,			/*  7 packet state 6 */
    SYS_pkt7,			/*  8 packet state 7 */
    SYS_pkt8,			/*  9 packet state 8 */
    SYS_pts1,			/* 10 PTS only byte 1 */
    SYS_pts2,			/* 11 PTS only byte 2 */
    SYS_pts3,			/* 12 PTS only byte 3 */
    SYS_pts4,			/* 13 PTS only byte 4 */
    SYS_ptsdts1,		/* 14 PTS followed by DTS byte 1 */
    SYS_ptsdts2,		/* 15 PTS followed by DTS byte 2 */
    SYS_ptsdts3,		/* 16 PTS followed by DTS byte 3 */
    SYS_ptsdts4,		/* 17 PTS followed by DTS byte 4 */
    SYS_dts0,			/* 19 DTS only byte 0 */
    SYS_dts1,			/* 20 DTS byte 1 */
    SYS_dts2,			/* 21 DTS byte 2 */
    SYS_dts3,			/* 22 DTS byte 3 */
    SYS_dts4,			/* 23 DTS byte 4 */
    SYS_pack0,			/* 24 Looking at 5th byte of pack */
    SYS_pack1,			/* 25 Looking at 6th byte of pack */
    SYS_pack2,			/* 26 Looking at 7th byte of pack */
    SYS_pack3,			/* 27 Looking at 8th byte of pack */
    SYS_pack4,			/* 28 Looking at 9th byte of pack */
    SYS_pack5,			/* 29 Looking at 10th byte of pack */
    SYS_pack6,			/* 30 Looking at 11th byte of pack */
    SYS_pack7,			/* 31 Looking at 12th byte of pack */
    SYS_hdr0,			/* 32 System header 0 */
    SYS_hdr1,			/* 33 System header 1 */
    SYS_hdr2,			/* 34 System header 2 */
    SYS_hdr3,			/* 35 System header 3 */
    SYS_hdr4,			/* 36 System header 4 */
    SYS_hdr5,			/* 37 System header 5 */
    SYS_hdr6,			/* 38 System header 6 */
    SYS_hdr7,			/* 39 System header 7 */
    SYS_hdr8			/* 30 System header 8 */
} SystemStateCode;


static void StreamClose1(StreamPtr, boolean);
static void SystemReader(SystemPtr);
static void SystemFreeFDB(SystemPtr, FileDataBlkPtr);

enum {
    N_SysSyntax = DBI_MpegSdemux,
    N_SysWarn,
    N_SysRealtime,
    N_SysBuffer,
    N_SysLimits,
    N_SysIgnore,
    N_SysTrace
};

static char *DebugNames[] = {
    "system syntax",
    "system warnings",
    "system real time",
    "system buffers",
    "system limits",
    "system ignored packets",
    "system trace",
};

FlagNames MpegSdemuxNames = {
    DBI_MpegSdemux,
    DebugNames,
    NumberOfCells(DebugNames)
};

FlagMacroList MpegSdemuxMacros = { NULL, 0 };

/*
 * NAME
 *   SystemOpen()   Open system stream and initialize system demux
 *
 * INPUTS
 *   nm   - System stream file name.
 *   msys - Ptr to structure filled in with system stream "table of contents";
 *	    also contains hook to hold all system demux data structures and
 *	    some parameters.
 *   tc	  - temporally correct flag
 *
 * DESCRIPTION:
 *   Opens system stream file, finds first pack and system header, and
 *   intializes system stream layer.
 *
 * RETURNS:
 *   An InternalReasonPtr if an error occurs else NULL.  The MpegSystemInfo
 *   structure is filled in with the "table of contents" of the system stream
 *   and a pointer to the system demultiplexor state data structures.
 */

InternalReasonPtr
SystemOpen(char *nm, MpegSystemInfoPtr msys)
{
    SystemPtr sys;
    FileHandle handle;

    if ((handle = (*msys->openFunc)(nm)) < 0)
	return(CreateSysReason("Cannot open %s\n", nm));
    if (msys->systemState == NULL) {
	sys = (SystemPtr)SecureAlloc(sizeof(System));
	memset((char *)sys, 0, sizeof(System));
    } else {
	SystemClose(msys, False);
	sys = (SystemPtr)msys->systemState;
    }
    sys->name = CopyOfString(nm);
    sys->handle = handle;
    sys->scr = -1.0;
    sys->muxbound = -1.0;
    /*
     * Copy parameters from MpegSystemInfo
     */
    sys->timescale = msys->timescale;
    sys->scheduler = msys->scheduler;
    sys->maxbehind = msys->maxbehind;
    sys->startAccessUnit = msys->startAccessUnit;
    sys->readFunc = msys->readFunc;

    sys->state = SYS_search;
    sys->flags = SYS_STARTUP;
    while ((sys->flags & (SYS_PACK | SYS_HEADER)) != (SYS_PACK | SYS_HEADER))
	SystemReader(sys);
    sys->flags &= ~SYS_STARTUP;

    /*
     * Return parameters
     */
    msys->nVideo = sys->videobound;
    msys->nAudio = sys->audiobound;
    msys->systemState = (PrivPointer)sys;
    return(NULL);
}

/*
 * NAME
 *   SystemSeek()   Seek to a specified offset
 *
 * INPUTS
 *   msys - Ptr to structure filled in with system stream "table of contents";
 *	    also contains hook to hold all system demux data structures and
 *	    some parameters.
 *   off  - offset in stream
 *   sk   - relative positioning flag
 *
 * DESCRIPTION:
 *   Seeks to specified offset + or - some amount to make the desired
 *   result happen.
 *
 * RETURNS:
 *   The actual offset or -1 on failure.
 */

int
SystemSeek(MpegSystemInfoPtr msys, long off, SeekWhere sk)
{
    return(off);
}

/*
 * NAME
 *   SystemClose()   Closes system layer, frees all memory
 *
 * INPUTS
 *   msys - a MpegSystemInfoPtr
 *
 * DESCRIPTION
 *
 * RETURNS
 */

void
SystemClose(MpegSystemInfoPtr msys, boolean freeStg)
{
    SystemPtr sys = (SystemPtr)msys->systemState;
    StreamDataRangePtr sdr;
    FileDataBlkPtr fdb;
    StreamPtr str, *strp;

    (*msys->closeFunc)(sys->handle);
    for (strp = sys->streams; strp < sys->streams + NSTREAM; strp++) {
	if (*strp != NULL)
	    StreamClose1(*strp, freeStg);
    }
    while ((str = sys->unspec) != NULL)
	StreamClose1(str, freeStg);
    if (sys->fdb != NULL) {
	SystemFreeFDB(sys, sys->fdb);
	sys->fdb = NULL;
    }
    free(sys->name);
    if (freeStg) {
	while ((sdr = sys->sdrfree) != NULL) {
	    sys->sdrfree = sdr->next;
	    free((char *)sdr);
	}
	while ((fdb = sys->fdbfree) != NULL) {
	    sys->fdbfree = fdb->next;
	    free((char *)fdb);
	}
	free((char *)sys);
	msys->systemState = NULL;
    } else {
	sdr = sys->sdrfree;
	fdb = sys->fdbfree;
	memset((char *)sys, 0, sizeof(System));
	sys->sdrfree = sdr;
	sys->fdbfree = fdb;
    }
}

void
SystemClose1(MpegSystemInfoPtr msys, MpegStreamInfoPtr mstr, boolean freeStg)
{
    SystemPtr sys = (SystemPtr)msys->systemState;
    StreamPtr *strp;
    int i;
    UChar matchId = ((StreamPtr)mstr->streamState)->id;

    i = 0;
    for (strp = sys->streams; strp < sys->streams + NSTREAM; strp++) {
	if (*strp != NULL) {
	    i++;
	    if( (*strp)->id == matchId ) {
		StreamClose1(*strp, freeStg);
		*strp = NULL;
		i--;
	    }
	}
    }
    if( i == 0 ) {
	SystemClose(msys, freeStg);
    }
}

/*
 * NAME
 *   StreamOpen	- Open a stream
 *
 * INPUTS
 *   msys  - a MpegSystemInfoPtr
 *   type  - bit stream type
 *   ndx   - Bitstream index or -1 for any non-opened stream of a type;
 *   mstrp - a MpegStreamInfoPtr
 *
 * DESCRIPTION
 *   Opens a stream.
 *
 * RETURNS
 *   InternalReasonPtr if error, NULL otherwise; *mstrp is set to the address of
 *   new MpegStreamInfoPtr
 */

InternalReasonPtr
StreamOpen(MpegSystemInfoPtr msys,
	   BitstreamType type,
	   int ndx,
	   MpegStreamInfoPtr mstr)
{
    SystemPtr sys = (SystemPtr)msys->systemState;
    StreamPtr str, *strp;
    int id, max;

    if (ndx < 0) {
	for (strp = &sys->unspec; *strp != NULL; strp = &((*strp)->next))
	    ;
	id = 0;
    } else {
	switch (type) {
	    case VideoBitstream:
		id = B_SC_VIDEO + ndx;
		max = B_SC_DATA;
		break;

	    case AudioBitstream:
		id = B_SC_AUDIO + ndx;
		max = B_SC_VIDEO;
		break;

	    case DataBitstream:
		id = B_SC_DATA + ndx;
		max = 0x100;
		break;

	    case PrivateBitstream:
		id = B_SC_PRIVATE1 + 2*ndx;
		max = B_SC_AUDIO;
		break;

	    default:
		return(CreateReason("Bad bitsream type: %d", type));
	}
	if (id >= max)
	    return(CreateReason("Stream index (%d) out-of-range", id));
	strp = &sys->streams[id - B_SC_RESERVED];
	if (*strp != NULL)
	    return(CreateReason("Stream (id=0x%02x) being opened twice", id));
    }
    *strp =  str  = (StreamPtr)SecureAlloc(sizeof(Stream));
    memset((char *)str, 0, sizeof(Stream));
    mstr->streamState = (PrivPointer)str;
    mstr->pos = 0;
    mstr->length = 0;
    str->id = id;
    str->type = type;
    str->sys = sys;
    sys->openstreams++;
    return(NULL);
}

/*
 * NAME
 *   StreamClose()	Close a stream
 *
 * INPUTS
 *   mstr - MpegStreamInfoPtr referencing stream to be closed
 *
 * DESCRIPTION
 *   Disconnects from the specified stream, free data structures
 *
 * RETURNS
 *   Nothing
 */

void
StreamClose(MpegStreamInfoPtr mstr, boolean freeStg)
{
    if (mstr->streamState != NULL)
	StreamClose1((StreamPtr)mstr->streamState, freeStg);
    if (freeStg)
	mstr->streamState = NULL;
}

static void
StreamClose1(StreamPtr str, boolean freeStg)
{
    SystemPtr sys;
    StreamPtr *strp;
    StreamDataRangePtr sdr;

    sys = str->sys;
    while ((sdr = str->sdrout) != NULL) {
	str->sdrout = sdr->next;
	SystemFreeFDB(sys, sdr->fdb);
	sdr->next = sys->sdrfree;
	sys->sdrfree = sdr;
    }
    if ((sdr = str->reading) != NULL) {
	sdr->next = sys->sdrfree;
	sys->sdrfree = sdr;
    }
    if (str->id) {
	if (sys->streams[str->id - B_SC_RESERVED] != str)
	    Fatal("System stream close error 1 on ID=0x%02x\n", str->id);
	sys->streams[str->id - B_SC_RESERVED] = NULL;
    } else {
	for (strp = &sys->unspec; ; strp = &((*strp)->next)) {
	    if (*strp == NULL)
		Fatal("System stream close error 2\n");
	    if (*strp == str) {
		*strp = str->next;
		break;
	    }
	}
    }
    memset((char *)str, 0, sizeof(Stream));
    sys->openstreams--;
    if (freeStg)
	free((char *)str);
}

/*
 * NAME
 *   StreamRead()	Return a block of data from a stream
 *
 * INPUTS
 *   mstr - MpegStreamInfoPtr for source stream
 *
 * DESCRIPTION
 *   Reads data from the stream
 *
 * RETURNS
 *   Indirectly returns the data via pointers in MpegStreamInfoPtr.
 *   EOF on the system stream is indicated by mstr->length == 0 on return.
 */

void
StreamRead(MpegStreamInfoPtr mstr)
{
    StreamPtr str = (StreamPtr)mstr->streamState;
    SystemPtr sys = str->sys;
    StreamDataRangePtr sdr;

    if ((sdr = str->reading) != NULL) {
	SystemFreeFDB(sys, sdr->fdb);
	sdr->next = sys->sdrfree;
	sys->sdrfree = sdr;
    }
    mstr->length = 0;
    if (str->id == 0) {
	/*
	 * Connect an unspecified stream
	 */
	while (str->id == 0) {
	    if (sys->flags & SYS_EOF)
		return;
	    SystemReader(sys);
	}
    }
    str->flags &= ~STR_UFLO;
    if ((sdr = str->sdrout) == NULL) {
	if (sys->timescale > 0.0) {
	    DebugTask(N_SysBuffer, {
		if (!(str->flags & STR_UFLO))
		    Msg("Buffer underflow on stream 0x%02X\n", str->id);
	    })
	}
	str->flags |= STR_UFLO;
	while ((sdr = str->sdrout) == NULL) {
	    if (sys->flags & SYS_EOF)
		return;
	    SystemReader(sys);
	}
    }
    str->sdrout = sdr->next;
    str->reading = sdr;
    str->bufocc -= sdr->length;
    mstr->buf = sdr->start;
    mstr->length = sdr->length;
    mstr->pos = sdr->pos;
    mstr->currentPts = sdr->pts;
    mstr->currentDts = sdr->dts;
    if ((sdr->dts.seconds > 0) && (sys->startAccessUnit != NULL))
	(*sys->startAccessUnit)(mstr);
    /*
     * If a DTS exists, call SystemReader() until sys->scr >= sdr->dts ???
     */
}

/*
 * NAME
 *   SystemReader()  Reads and parses the system stream.
 *
 * INPUTS:
 *   sys - Ptr to the System structure
 *
 * DESCRIPTION:
 *   In the temporally correct mode, this procedure is called from the
 *   temporal event dispatcher, otherwise from the system stream consumers.
 *   The routine parses system packets, packs and system headers from the
 *   system stream and reschedules itself after each FileDataBlock or when
 *   a pack header with an SCR denoting a temporal gap is seen.
 *
 * RETURNS:
 *   Indirectly returns parsed packed bodies for each elementary stream.
 */

static void
SystemReader(SystemPtr sys)
{
    int c, n, startpos, state;
    StreamDataRangePtr sdr;
    StreamPtr str, *strp;
    FileDataBlkPtr fdb;
    UChar *p, *plast;
    double t, t1;
    TimeSpec ts;

    if ((fdb = sys->fdb) == NULL) {
	/*
	 * Allocate new FSB
	 */
	if ((fdb = sys->fdbfree) == NULL)
	    fdb = (FileDataBlkPtr)SecureAlloc(sizeof(FileDataBlk));
	else
	    sys->fdbfree = fdb->next;
	fdb->refcnt = 1;
	fdb->count = 0;
	sys->fdb = fdb;
	sys->nextbyte = fdb->data;
    }
    if (fdb->count < FDBSIZE) {
	/*
	 * Read system stream into FSB
	 */
	if (sys->flags & SYS_EOF)
	    return;
	if ((n = (*sys->readFunc)(sys->handle, (char *)(fdb->data + fdb->count),
	    FDBSIZE - fdb->count)) == 0) {
	    sys->flags |= SYS_EOF;
	} else if (n < 0) {	
	    extern int errno;
#if 0
	    if (errno == EWOULDBLOCK) {		/* Need to make portable!! */
		if (sys->scheduler == NULL)
		    Fatal(
		    "System read would block in non-temporally correct mode\n");
		(*sys->scheduler)(SystemReader,
		    (PrivPointer)sys, NULL, sys->handle);
		return;
	    }
#endif
	    Fatal("Read error on %s: %s\n", sys->name, LookupSysError(errno));
	}
	fdb->count += n;
	sys->pos += n;
    }
    p = sys->nextbyte;
    plast = fdb->data + fdb->count;
    startpos = sys->pos + p - plast;
    state = sys->state;
    while (p < plast) {
	c = *p++;
	switch (state++) {

	case SYS_search:		/* Expecting first byte of startcode */
	    if (c != ((B_STARTCODE >> 16) & 0xFF)) {
		DebugTask(N_SysSyntax, {
		    if (!(sys->flags & SYS_ERROR))
			Msg("%d: Expected byte 0 of system start code\n",
			    sys->pos + p - plast - 1);
		})
		break;
	    }
	    continue;

	case SYS_pkt0:			/* Expecting 2nd byte of startcode */
	    if (c != ((B_STARTCODE >> 8) & 0xFF)) {
		DebugTask(N_SysSyntax, {
		    if (!(sys->flags & SYS_ERROR))
			Msg("%d: Expected byte 1 of system start code\n",
			    sys->pos + p - plast - 1);
		})
		break;
	    }
	    continue;

	case SYS_pkt1:			/* Expecting third byte of statcode */
	    if (c != (B_STARTCODE & 0xFF)) {
		if (c == ((B_STARTCODE >> 8) & 0xFF)) {
		    state = SYS_pkt1;
		    continue;
		}
		DebugTask(N_SysSyntax, {
		    if (!(sys->flags & SYS_ERROR))
			Msg("%d: Expected byte 2 of system start code\n",
			    sys->pos + p - plast - 1);
		})
		break;
	    }
	    continue;

	case SYS_pkt2:			/* Expecting 4th byte of startcode */
	    if (c < B_SC_SYSEND) {
		DebugTask(N_SysSyntax, {
		    if (!(sys->flags & SYS_ERROR))
			Msg("%d: Expected byte 3 of system start code\n",
			    sys->pos + p - plast - 1);
		})
		break;
	    }
	    sys->flags &= ~SYS_ERROR;
	    sys->pktloc = sys->pos + p - plast - 4;
	    switch (sys->id = c) {
		case B_SC_PACK: state = SYS_pack0; continue;
		case B_SC_SYSEND: sys->flags |= SYS_EOF; goto out;
	    }
	    continue;

	case SYS_pkt3:			/* Expecting 1st byte of packet length*/
	    sys->packetsize = c << 8;
	    state = SYS_pkt4;
	    continue;

	case SYS_pkt4:			/* Expecting 2nd byte of packet length*/
	    sys->packetsize |= c;
	    if (sys->id != B_SC_SYSHDR) {
		sys->pktflags = PKT_NEWPKT;
		sys->fullsize = sys->packetsize;
		sys->padding = 0;
		if ((sys->id < B_SC_RESERVED) ||
		    (sys->streams[sys->id - B_SC_RESERVED] != NULL) ||
		    (sys->id < B_SC_AUDIO))
		    continue;
		/*
		 * Check for unspecified streams to attach to this ID
		 */
		n = (sys->id < B_SC_VIDEO) ? AudioBitstream :
		    ((sys->id < B_SC_DATA) ? VideoBitstream : DataBitstream);
		for (strp = &sys->unspec; (str = *strp) != NULL;
		     strp = &str->next) {
		    if (str->type == n) {
			*strp = str->next;
			sys->streams[sys->id - B_SC_RESERVED] = str;
			str->id = sys->id;
			break;
		    }
		}
		continue;
	    }
	    if ((sys->packetsize > 198) ||
		(sys->packetsize < 6) ||
		((sys->packetsize - 6) % 3)) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Bad system header size (%d)\n",
			sys->pktloc, sys->packetsize));
		break;
	    }
	    sys->nhdrbufspec = (sys->packetsize - 6) / 3;
	    state = SYS_hdr0;
	    continue;

	case SYS_pkt5:			/* Byte after packet length */
	    sys->pts = -1.0;
	    sys->dts = -1.0;
	    switch (c & 0xC0) {

	    case 0xC0:
	    case 0x80:
		if (c != 0xFF) {
		    DebugTask(N_SysSyntax,
			Msg("%d: Padding byte not 0xFF (0x%02x) at %d\n",
			    sys->pktloc, c, sys->pos + p - plast - 1));
		    break;
		}
		sys->padding++;
		if (sys->padding == 17) {
		    DebugTask(N_SysSyntax, Msg(
	    "%d: Packet (ID=0x%x) contains more than 16 bytes of stuffing\n",
			sys->pktloc, sys->id));
		}
		state = SYS_pkt5;
		if ((--sys->packetsize <= 0) && (sys->id != B_SC_PADDING)) {
		    DebugTask(N_SysSyntax,
			Msg("%d: Packet is entirely stuffing\n", sys->pktloc));
		    break;
		}
		continue;

	    case 0x40:
		sys->tmp = c << 8;
		if (--sys->packetsize <= 0) {
		    DebugTask(N_SysSyntax,
			Msg("%d: Premature end-of-packet (ID=0x%02x) at %d\n",
			    sys->pktloc, sys->id, sys->pos + p - plast - 1));
		    break;
		}
		continue;

	    case 0x00:
		state = SYS_pkt7;
		p--;
		continue;
	    }
	    break;

	case SYS_pkt6:			/* 2nd byte of STD buffer size */
	    sys->tmp |= c;
	    if (--sys->packetsize <= 0) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Premature end-of-packet (ID=0x%02x) at %d\n",
			sys->pktloc, sys->id, sys->pos + p - plast - 1));
		break;
	    }
	    sys->bufsize = (sys->tmp & 0x1FFF) <<
		((sys->tmp & 0x2000) ? 10 : 7);
	    sys->pktflags |= PKT_BUFSIZE;
	    if ((str = sys->streams[sys->id - B_SC_RESERVED]) != NULL) {
		str->bufsize = sys->bufsize;
		n = sys->bufbound[sys->id - B_SC_RESERVED];
		if ((n > 0) && (str->bufsize > n)) {
		    DebugTask(N_SysLimits,
			Msg("%d (ID=0x%02x) Buffer size (%d) > bound (%d)\n",
			    sys->pktloc, sys->id, str->bufsize, n));
		}
	    }
	    continue;
	
	case SYS_pkt7:		/* 1st byte of PTS/DTS or fixed 0x0F byte */
	    if (--sys->packetsize <= 0) {
		DebugTask(N_SysSyntax,
		    if (sys->id != B_SC_PADDING)
			Msg("%d: Premature end-of-packet (ID=0x%02x,pos=%d)\n",
			    sys->pktloc, sys->id, sys->pos + p - plast - 1));
		break;
	    }
	    if ((n = c >> 4) == 0) {
		if (c != 0x0F) {
		    DebugTask(N_SysSyntax,
			Msg("%d: ID=0x%02x, Expected 0x0F, got 0x%02x\n",
			    sys->pktloc, sys->id, c));
		    break;
		}
		continue;
	    }
	    DebugTask(N_SysSyntax,
		if ((c & 1) == 0)
		    Msg("%d: (ID=0x%02x) LSB=0 in first packet ts at %d\n",
			sys->pktloc, sys->id, sys->pos + p - plast - 1);
	    )

	    sys->tmp |= (c & 0x0E) << 23;
	    sys->pktflags |= PKT_PTS;
	    switch (n) {
		case 2: state = SYS_pts1; continue;
		case 3: sys->pktflags |= PKT_DTS; state = SYS_ptsdts1; continue;
	    }
	    DebugTask(N_SysSyntax,
		Msg("%d: Invalid code (SYS_pkt7) 0x%02x at %d\n",
		    sys->pktloc, c, sys->pos + p - plast - 1));
	    break;

	case SYS_pkt8:			/* First data byte of packet */
	    p--;
	    if (sys->packetsize <= 0) {
		DebugTask(N_SysSyntax,
		    if (sys->id != B_SC_PADDING)
			Msg("%d: No data in packet (ID=0x%02x)\n",
			    sys->pktloc, sys->id));
		state = SYS_search;
		break;
	    }
	    if ((n = plast - p) > sys->packetsize)
		n = sys->packetsize;
	    DebugTask(N_SysTrace, {
		if (sys->pktflags & PKT_NEWPKT) {
		    Msg("%d: Packet ID=0x%02X size=%d",
			sys->pktloc, sys->id, sys->fullsize);
		    if (sys->padding != 0)
			Msg(" pad=%d", sys->padding);
		    if (sys->pktflags & PKT_PTS)
			Msg(" PTS=%.5f", sys->pts);
		    if (sys->pktflags & PKT_DTS)
			Msg(" DTS=%.5f", sys->dts);
		    if (sys->pktflags & PKT_BUFSIZE)
			Msg(" buf=%d", sys->bufsize);
		} else
		    Msg("%d: Continuing Packet", sys->pktloc);
		Msg(" ret=%d\n", n);
	    })
	    sys->pktflags = 0;
	    if ((str = sys->streams[sys->id - B_SC_RESERVED]) != NULL) {
		if ((sdr = sys->sdrfree) == NULL)
		    sdr = (StreamDataRangePtr)
			SecureAlloc(sizeof(StreamDataRange));
		else
		    sys->sdrfree = sdr->next;
		sdr->next = NULL;
		sdr->start = p;
		sdr->fdb = sys->fdb;
		sdr->length = n;
		sdr->pos = sys->pos + p - plast;
		if ((sys->timescale > 0.0) && (sys->pts >= 0)) {
		    TimeNow(&ts);
		    t1 = ((sdr->pos - sys->scrpos) << L_BYTE) /
			sys->muxrate + sys->scr;
		    if ((t = sys->timescale * (sys->pts - t1)) < 0.0) {
			DebugTask(N_SysRealtime,
			    Msg("%d: (ID=0x%02X) PTS (%.5f) < STC (%.5f)\n",
				sys->pktloc, sys->id, sys->pts, t1));
			sdr->pts.seconds = -1;
		    } else {
			t += ts.seconds + ts.microSeconds * 1.0e-6;
			sdr->pts.seconds = (int)t;
			sdr->pts.microSeconds = MicroSecPerSec *
			    (t - sdr->pts.seconds);		
		    }
		    if ((t = sys->timescale * (sys->dts - t1)) < 0.0) {
			DebugTask(N_SysRealtime,
			    Msg("%d: (ID=0x%02X) DTS (%.5f) < STC (%.5f)\n",
				sys->pktloc, sys->id, sys->dts, t1));
			sdr->pts.seconds = -1;
		    } else {
			t += ts.seconds + ts.microSeconds * 1.0e-6;
			sdr->dts.seconds = (int)t;
			sdr->dts.microSeconds = MicroSecPerSec *
			    (t - sdr->dts.seconds);		
		    }
		} else {
		    sdr->pts.seconds = -1;
		    sdr->dts.seconds = -1;
		}
		if (str->sdrout == NULL)
		    str->sdrout = sdr;
		else
		    str->sdrin->next = sdr;
		str->sdrin = sdr;
		str->bufocc += n;
		sys->fdb->refcnt++;
		sys->pts = -1.0;
		sys->dts = -1.0;
		if ((str->bufocc > str->bufsize) &&
		    (sys->timescale > 0.0)) {		/* Buffer overflow */
		    DebugTask(N_SysBuffer, {
			if (!(str->flags & STR_OFLO))
			    Msg("%d: (ID=0x%02X) Buffer overflow\n",
				sys->pktloc, sys->id);
		    })
		    str->flags |= STR_OFLO;
		} else
		    str->flags &= ~STR_OFLO;
	    } else {
		DebugTask(N_SysIgnore,
		    Msg("%d: Ignored packet (ID=0x%x data=%d bytes)\n",
			sys->pktloc, sys->id, n));
	    }
	    p += n;
	    state = ((sys->packetsize -= n) <= 0) ? SYS_search : SYS_pkt8;
	    continue;

	case SYS_dts0:		/* first byte of DTS */
	    if (--sys->packetsize <= 0) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Premature end-of-packet (ID=0x%02x) at %d\n",
			sys->pktloc, sys->id, sys->pos + p - plast - 1));
		break;
	    }
	    if ((c & 0xF1) != 0x11) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Expected DTS '0001xxx1' at %d after PTS\n",
			sys->pktloc, sys->pos + p - plast - 1));
		break;
	    }
	    sys->tmp = (c & 0x0E) << 23;
	    continue;

	case SYS_pack0:				/* First byte of SCR */
	    if ((c & 0xF1) != 0x21) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Expected '0010xxx1' at %d after pack start code\n",
			sys->pktloc, sys->pos + p - plast - 1));
		break;
	    }
	    sys->tmp = (c & 0x0E) << 23;
	    continue;

	case SYS_pts1:			/* 2nd byte of PTS only */
	case SYS_dts1:			/* 2nd byte of DTS */
	case SYS_ptsdts1:		/* 2nd byte of PTS followed by DTS */
	    if (--sys->packetsize <= 0) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Premature end-of-packet (ID=0x%02x) at %d\n",
			sys->pktloc, sys->id, sys->pos + p - plast - 1));
		break;
	    }
	    /*FALLINTO*/

	case SYS_pack1:			/* 2nd byte of SCR */
	    sys->tmp |= c << 15;
	    continue;

	case SYS_pts2:			/* 3rd byte of PTS only */
	case SYS_dts2:			/* 3rd byte of DTS */
	case SYS_ptsdts2:		/* 3rd byte of PTS followed by DTS */
	    if (--sys->packetsize <= 0) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Premature end-of-packet (ID=0x%02x) at %d\n",
			sys->pktloc, sys->id, sys->pos + p - plast - 1));
		break;
	    }
	    /*FALLINTO*/

	case SYS_pack2:			/* 3rd byte of SCR */
	    if (!(c & 1)) {
		DebugTask(N_SysSyntax,
		    Msg(
		"%d: missing marker=1 in timestamp byte 2 (state=%d) at %d\n",
			sys->pktloc, state - 1, sys->pos + p - plast - 1));
		break;
	    }
	    sys->tmp |= (c & 0xFE) << 7;
	    continue;

	case SYS_pts3:			/* 4th byte of PTS only */
	case SYS_dts3:			/* 4th byte of DTS */
	case SYS_ptsdts3:		/* 4th byte of PTS followed by DTS */
	    if (--sys->packetsize <= 0) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Premature end-of-packet (ID=0x%02x) at %d\n",
			sys->pktloc, sys->id, sys->pos + p - plast - 1));
		break;
	    }
	    /*FALLINTO*/

	case SYS_pack3:			/* 4th byte of SCR */
	    sys->tmp |= c;
	    continue;

	case SYS_pts4:			/* 5th byte of PTS only */
	case SYS_dts4:			/* 5th byte of DTS */
	case SYS_ptsdts4:		/* 5th byte of PTS followed by DTS */
	    if (--sys->packetsize <= 0) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Premature end-of-packet (ID=0x%02x) at %d\n",
			sys->pktloc, sys->id, sys->pos + p - plast - 1));
		break;
	    }
	    /*FALLINTO*/

	case SYS_pack4:			/* 5th byte of SCR */
	    if (!(c & 1)) {
		DebugTask(N_SysSyntax,
		    Msg(
		"%d: missing marker=1 in timstamp byte 4 (state=%d) at %d\n",
			sys->pktloc, state - 1, sys->pos + p - plast - 1));
		break;
	    }
	    t = (128.0 * sys->tmp + (c >> 1)) / SYSCLKF;
	    switch (state - 1) {

		case SYS_pts4:
		    sys->dts = sys->pts = t;
		    state = SYS_pkt8;
		    continue;

		case SYS_ptsdts4:
		    sys->pts = t;
		    state = SYS_dts0;
		    continue;

		case SYS_dts4:
		    sys->dts = t;
		    state = SYS_pkt8;
		    continue;
	    }
	    t1 = sys->scr;
	    n = sys->scrpos;
	    sys->scr = t;
	    sys->scrpos = sys->pos + p - plast;
	    if (t1 < 0.0)
		continue;					/* First time */
	    t1 += ((sys->scrpos - n) << L_BYTE) / sys->muxrate;
	    if (sys->timescale > 0) {
		if ((t1 - t) > sys->maxbehind) {
		    DebugTask(N_SysRealtime,
			Msg("%d: SCR (%.5f sec) is non causal (>= %.5f sec)\n",
			    sys->scrpos, sys->scr, t));
		}
		if ((sys->scheduler != NULL) && (t1 < t)) {
		    t = sys->timescale * (t - t1);
		    ts.seconds = (int)t;
		    ts.microSeconds = (int)(MicroSecPerSec *
			(t - ts.seconds) + 0.5);
		    (sys->scheduler)(SystemReader, (PrivPointer)sys, &ts, -1);
		    return;
		}
	    }
	    continue;

	case SYS_pack5:				/* 1st byte of muxrate */
	case SYS_hdr0:				/* 1st byte of muxbound */
	    if (!(c & 0x80)) {
		DebugTask(N_SysSyntax,
		    Msg("%d: marker=0 in first byte of mux%s at %d\n",
			sys->pktloc, (state == (SYS_pack5 + 1)) ? "rate" :
			"bound", sys->pos + p - plast - 1));
		break;
	    }
	    sys->tmp = (c & 0x7F) << 15;
	    continue;

	case SYS_pack6:				/* 2nd byte of muxrate */
	case SYS_hdr1:				/* 2nd byte of muxbound */
	    sys->tmp |= c << 7;
	    continue;

	case SYS_pack7:				/* 3rd byte of muxrate */
	case SYS_hdr2:				/* 3rd byte of muxbound */
	    if (!(c & 1)) {
		DebugTask(N_SysSyntax,
		    Msg("%d: marker=0 in third byte of mux%s at %d\n",
			sys->pktloc, (state == (SYS_pack7 + 1)) ? "rate" :
			"bound", sys->pos + p - plast - 1));
		break;
	    }
	    sys->tmp |= c >> 1;
	    if (state == (SYS_hdr2 + 1)) {
		sys->muxbound = 400.0 * sys->tmp;
		continue;
	    }
	    sys->muxrate = 400.0 * sys->tmp;
	    DebugTask(N_SysLimits, {
		if ((sys->muxbound >= 0) && (sys->muxrate > sys->muxbound))
		    Msg("%d: Muxrate (%.0f) exceeds muxrate bound (%.0f)\n",
			sys->pktloc, sys->muxrate, sys->muxbound);
	    })
	    DebugTask(N_SysTrace, Msg(
		"%d: Pack (SCR=%.5f sec, MuxRate=%.0f bit/sec)\n",
		sys->pktloc, sys->scr, sys->muxrate));
	    sys->flags |= SYS_PACK;
	    state = SYS_search;
	    if (sys->flags & SYS_STARTUP)
		goto out;
	    continue;

	case SYS_hdr3:				/* audiobound, fixed, CSPS */
	    sys->audiobound = c >> 2;
	    if (c & 2) sys->hdrflags |= HDR_FIXED;
	    if (c & 1) sys->hdrflags |= HDR_CSPS;
	    continue;

	case SYS_hdr4:			/* audiolock, videolock, videobound */
	    if ((c & 0x20) == 0) {
		DebugTask(N_SysSyntax,
		    Msg(
		"%d: missing marker bit in byte 10 of system header at %d\n",
			sys->pktloc, sys->pos + p - plast - 1));
		break;
	    }
	    if (c & 0x80) sys->hdrflags |= HDR_AUDIOLOCK;
	    if (c & 0x40) sys->hdrflags |= HDR_VIDEOLOCK;
	    sys->videobound = c & 0x1F;
	    continue;

	case SYS_hdr5:				/* System header reserved byte*/
	    if (c != 0xFF) {
		DebugTask(N_SysSyntax,
		    Msg("%d: reserved byte of system header not 0xFF at %d\n",
			sys->pktloc, sys->pos + p - plast - 1));
		break;
	    }
	    DebugTask(N_SysTrace,
		Msg("%d: System Header mxb=%.0f nAud=%d nVid=%d%s%s%s%s\n",
		    sys->pktloc, sys->muxbound, sys->audiobound,
		    sys->videobound,
		    (sys->hdrflags & HDR_FIXED)     ? " fixed"   : "",
		    (sys->hdrflags & HDR_CSPS)      ? " csps"    : "",
		    (sys->hdrflags & HDR_AUDIOLOCK) ? " audlock" : "",
		    (sys->hdrflags & HDR_VIDEOLOCK) ? " vidlock" : ""));
	    if (sys->nhdrbufspec == 0)
		state = SYS_search;
	    continue;

	case SYS_hdr6:				/* 1st byte of buffer bound */
	    if ((c < 0xB8) || (c == B_SC_PACK) || (c == B_SC_SYSHDR)) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Invalid buffer bounds ID byte (0x%02X) at %d\n",
			sys->pktloc, c, sys->pos + p - plast - 1));
		break;
	    }
	    sys->id = c;
	    continue;

	case SYS_hdr7:				/* 2nd byte of buffer bound */
	    if ((c & 0xC0) != 0xC0) {
		DebugTask(N_SysSyntax,
		    Msg("%d: Expected 11xxxxxx, ", sys->pktloc);
		    Msg("got 0x%02x for buffer bounds at pos=%d\n",
			c, sys->pos + p - plast - 1);
		)
		break;
	    }
	    sys->tmp = c << 8;
	    continue;
	
	case SYS_hdr8:				/* 3rd byte of buffer bound */
	    c |= sys->tmp;
	    switch (sys->id) {

	    case 0xB8:				/* Code for all audio streams */
		DebugTask(N_SysSyntax, {
		    if (c & 0x2000)
			Msg("%d: All audio buffer bound with 1024 byte scale\n",
			    sys->pktloc);
		})
		c = (c & 0x1FFF) << 7;
		for (n = B_SC_AUDIO; n < B_SC_VIDEO; n++)
		    sys->bufbound[n - B_SC_RESERVED] = c;
		DebugTask(N_SysTrace,
		    Msg("%d: All-audio buf bound: %d\n", sys->pktloc, c));
		break;

	    case 0xB9:				/* Code for all video streams */
		DebugTask(N_SysSyntax, {
		    if (!(sys->tmp & 0x2000))
			Msg("%d: All video buffer bound with 128 byte scale\n",
			    sys->pktloc);
		})
		c = (c & 0x1FFF) << 10;
		for (n = B_SC_VIDEO; n < B_SC_DATA; n++)
		    sys->bufbound[n - B_SC_RESERVED] = c;
		DebugTask(N_SysTrace,
		    Msg("%d: All-video buf bound: %d\n", sys->pktloc, c));
		break;

	    default:
		DebugTask(N_SysSyntax, {
		    if ((sys->id >= B_SC_AUDIO) &&
			(sys->id < B_SC_VIDEO) &&
			(sys->tmp & 0x2000))
			Msg("%d: Audio buffer bound with 1024 byte scale\n",
			    sys->pktloc);
		    if ((sys->id >= B_SC_VIDEO) &&
			(sys->id < B_SC_DATA) &&
			!(sys->tmp & 0x2000))
			Msg("%d: Video buffer bound with 128 byte scale\n",
			    sys->pktloc);
		})
		sys->bufbound[sys->id - B_SC_RESERVED] = c =
		    (c & 0x1FFF) << ((c & 0x2000) ? 10 : 7);
		DebugTask(N_SysTrace,
		    Msg("%d: Stream 0x%02X buffer bound: %d\n",
			sys->pktloc, sys->id, c));
		break;
	    }
	    if (--sys->nhdrbufspec > 0) {
		state = SYS_hdr6;
		continue;
	    }
	    state = SYS_search;
	    sys->flags |= SYS_HEADER;
	    if (sys->flags & SYS_STARTUP)
		goto out;
	    continue;

	default:
	    Fatal("Bad system state: %d\n", state - 1);
	}
	state = SYS_search;
	sys->flags |= SYS_ERROR;
    }
out:
    sys->nextbyte = p;
    sys->state = state;
    if (p >= fdb->data + FDBSIZE) {
	SystemFreeFDB(sys, fdb);
	sys->fdb = NULL;
    }
    if (sys->scheduler != NULL) {
	if (sys->muxrate <= 0)
	    t = POLLTIME;
	else
	    t = sys->timescale * ((sys->pos - startpos) << L_BYTE) /
		sys->muxrate;
	ts.seconds = (int)t;
	ts.microSeconds = (int)(MicroSecPerSec * (t - ts.seconds) + 0.5);
	(sys->scheduler)(SystemReader, (PrivPointer)sys, &ts, -1);
    }
}

/*
 * NAME
 *   SystemFreeFDB	Frees a file data block
 *
 * INPUTS
 *   sys - System state structure pointer
 *   fdb - file data block to be freed
 *
 * DESCRIPTION
 *   Decrements a reference count.  When it reaches zero, the fdb is placed
 *   on the free list
 *
 * RETURNS
 *   Nothing
 */

static void
SystemFreeFDB(SystemPtr sys, FileDataBlkPtr fdb)
{
    if (--fdb->refcnt > 0)
	return;
    fdb->next = sys->fdbfree;
    sys->fdbfree = fdb;
}
