/*
 * Copyright (c) 1992, 1993, 1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	fsPacket.c - physical fs packet interface
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/fsPacket.c,v 1.18 1995/10/02 04:56:29 toddb Exp $
 */

#include "nvrPacketI.h"
#include "nvrAlloc.h"
#include "nvrString.h"
#include "fsPacket.h"
#include "fileAccess.h"

static InternalReason CantProcess =
    { "Process called for fs packet layer.\n", IR_InternalProblem };
static InternalReason BadDeliverCall =
    { "FsDeliver called with non-null pointer\n", IR_InternalProblem };

#define DefaultPacketSize	8192

typedef struct _fsPriv {
    PacketLayer pub;		/* the public packet layer struct */
    FileHandle fd;		/* file descriptor */
    char *file;			/* input file name */
    PacketDataPtr alignPacket;	/* holds data for next packet */
    boolean align;		/* true if we align packets */
    long pos;			/* file offset */
    long packetSize;		/* preferred read size */
    void (*warning)(char *fmt, ...);
} FsPriv, *FsPrivPtr;

static void
FsForceAligment(PacketLayerPtr pLayer)
{
    FsPrivPtr priv = (FsPrivPtr) pLayer;
    priv->align = True;
}

static void
FsFree(PacketDataPtr packet)
{
    PacketLayerPtr pLayer = packet->pLayer;
    int refCount = packet->refCount - 1;

    /*
     * Special! since we are the lowest layer, then the
     * last free will reduce the refCount to 1 (our own reference);
     */
    if (refCount > 1) {
	packet->refCount = refCount;
	return;
    }
    packet->refCount = 0;

    /*
     * we are (always) at the lowest layer, so we know it is unnecessary
     * to call the free function for the layer below
     */
    QueuePacket(&pLayer->freeList, packet);
}

static InternalReasonPtr
FsDeliver(PacketLayerPtr pLayer, PacketDataPtr mustBeNull)
{
    long nBytes, need, have;
    PacketDataPtr packet;
    FsPrivPtr priv = (FsPrivPtr) pLayer;

    if (mustBeNull)
	return( &BadDeliverCall );
    if (NULL != (packet = pLayer->queue)) {
	pLayer->queue = packet->next;
	DeliverLayersAbove(pLayer, packet);
	return(NULL);
    }
    if (pLayer->eof)
	return( &IRPipelineEof );

    if (! priv->align || NULL == (packet = priv->alignPacket)) {
	packet = AllocPacket(&pLayer->freeList, priv->packetSize);
	if (packet == NULL)
	    return( AllocReason("FS packet", priv->packetSize) );
	need = priv->packetSize;
	have = 0;
    } else {
	have = packet->dataBytes;
	need = priv->packetSize - have;
    }

    packet->pos = priv->pos;
    packet->packetType = FsAllPackets;
    packet->pLayer = pLayer;
    packet->down = NULL;

    nBytes = ReadData(priv->fd, (char *)(packet->buffer + have), need);
    if (nBytes < 0) {
	QueuePacket(&pLayer->freeList, packet);
	return( CreateSysReason("Can't read from %s\n", priv->file) );
    }
    nBytes += have;
    packet->dataBytes = nBytes;
    priv->pos += nBytes;
    
    if (priv->align) {
	InternalReasonPtr irp;

	priv->alignPacket = AlignPacket(packet, &pLayer->freeList, &irp);
	if (priv->alignPacket == NULL) {
	    if (irp != NULL) {
		QueuePacket(&pLayer->freeList, packet);
		return(irp);
	    }
	} else {
	    priv->pos -= priv->alignPacket->dataBytes;
	}
    }

    if (nBytes == 0) {
	(*packet->pLayer->ops.free)(packet);
	MarkEofInLayersAbove(pLayer, True);
	return( NULL );
    }

    if (pLayer->pkiList)
	return( DeliverLayersAbove(pLayer, packet) );
    QueuePacket(&pLayer->queue, packet);
    return(NULL);
}

static InternalReasonPtr
FsProcess(PacketLayerPtr pLayer)
{
    return( &CantProcess );
}

static InternalReasonPtr
FsFlush(PacketLayerPtr pLayer)
{
    return( NULL );
}

static InternalReasonPtr
FsSeek(PacketLayerPtr pLayer, ULong absOffset)
{
    FsPrivPtr priv = (FsPrivPtr) pLayer;
    ULong curOffset;
    InternalReasonPtr reason;
    long where;
    ULong packetNum;

    MarkEofInLayersAbove(pLayer, False);

    packetNum = absOffset / priv->packetSize;
    if (packetNum * priv->packetSize != absOffset) {
	/*
	 * Truncate to point *before* requested offset
	 * so we get the byte requested
	 */
	absOffset = packetNum * priv->packetSize;
    }

    /*
     * By the time we are called, all packets in the system have
     * been released and should be on our free queue (unless some
     * packet is being, illegally held by some unknown software layer)
     * Make an attempt to use packets already in existence
     */
    curOffset = InternalPacketSeek(pLayer, absOffset);
    if (curOffset == priv->pos)
	return(NULL);		/* sucess! */
    
    /*
     * For the fs layer, the requested offset maps directly to
     * the file offset
     */
    where = SeekFile(priv->fd, absOffset, SK_Start);
    if (where < 0)
	return( CreateSysReason("Can't seek on %s\n", priv->file) );
    if (where != absOffset) {
	reason = CreateReason("Requested seek(%d) on %s, got %d\n",
			      absOffset, priv->file, where);
	return( reason );
    }
    priv->pos = where;
    DiscardQueuedPackets(pLayer);
    return(NULL);
}

static void
FsClose(PacketLayerPtr pLayer)
{
    FsPrivPtr priv = (FsPrivPtr) pLayer;
    int refCount = pLayer->refCount - 1;

    /*
     * Special! since we are the lowest layer, then the
     * last close will reduce the refCount to 1 (our own reference);
     */
    if (refCount > 1) {
	pLayer->refCount = refCount;
	return;
    }
    pLayer->refCount = 0;

    CloseFile(priv->fd);

    FreePacketList(&pLayer->freeList);
    FreePacketList(&pLayer->queue);
    free((char *)priv->file);
    free((char *)pLayer);
}

PacketLayerPtr
OpenFsPacket(FsPacketConfPtr conf, InternalReasonPtr *pReason)
{
    FileHandle fd;
    FsPrivPtr priv;
    PacketLayerPtr pLayer;

    fd = OpenFileForReading(conf->file);
    if (fd < 0) {
	*pReason = CreateSysReason("Can't open %s\n", conf->file);
	return(NULL);
    }
    pLayer = AllocPacketLayer(PacketFs, NULL, sizeof(FsPriv));
    if (pLayer == NULL) {
	CloseFile(fd);
	*pReason = &IRBadAllocation;
	return(NULL);
    }
    priv = (FsPrivPtr)pLayer;
    priv->fd = fd;
    priv->file = CopyOfString(conf->file);
    priv->warning = conf->warning;
    priv->packetSize = conf->packetSize;
    if (priv->packetSize == 0)
        priv->packetSize = DefaultPacketSize;

    pLayer->ops.deliver      = FsDeliver;
    pLayer->ops.process      = FsProcess;
    pLayer->ops.flush        = FsFlush;
    pLayer->ops.close        = FsClose;
    pLayer->ops.seek         = FsSeek;
    pLayer->ops.free         = FsFree;
    pLayer->ops.alignPackets = FsForceAligment;

    return(pLayer);
}
