/*
 * Copyright (c) 1992, 1993, 1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	netPacket.c - physical net packet interface
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/netPacket.c,v 1.6 1995/10/02 00:39:59 toddb Exp $
 */

#include "nvrPacketI.h"
#include "nvrAlloc.h"
#include "nvrString.h"
#include "netPacket.h"
#include "fileAccess.h"

static InternalReason CantProcess =
    { "Process called for net packet layer.\n", IR_InternalProblem };
static InternalReason CantSeek =
    { "Seek called on network socket.\n", IR_InternalProblem };
static InternalReason BadDeliverCall =
    { "NetDeliver called with non-null pointer\n", IR_InternalProblem };

#define DefaultPacketSize	8192

typedef struct _netPriv {
    PacketLayer pub;
    FileHandle fd;
    char *source;
    int port;
    CastType castType;
    long pos;
    long packetSize;
    boolean holdPackets;
    void (*warning)(char *fmt, ...);
} NetPriv, *NetPrivPtr;

static void
NetFree(PacketDataPtr packet)
{
    PacketLayerPtr pLayer = packet->pLayer;
    int refCount = packet->refCount - 1;

    /*
     * Special! since we are the lowest layer, then the
     * last free will reduce the refCount to 1 (our own reference);
     */
    if (refCount > 1) {
	packet->refCount = refCount;
	return;
    }
    packet->refCount = 0;

    /*
     * we are (always) at the lowest layer, so we know it is unnecessary
     * to call the free function for the layer below
     */
    QueuePacket(&pLayer->freeList, packet);
}

static InternalReasonPtr
NetDeliver(PacketLayerPtr pLayer, PacketDataPtr mustBeNull)
{
    long nBytes;
    PacketDataPtr packet;
    NetPrivPtr priv = (NetPrivPtr) pLayer;

    if (mustBeNull)
	return( &BadDeliverCall );
    if (NULL != (packet = pLayer->queue)) {
	pLayer->queue = packet->next;
	DeliverLayersAbove(pLayer, packet);
	return(NULL);
    }
    if (pLayer->eof)
	return( &IRPipelineEof );

    if ((packet = AllocPacket(&pLayer->freeList, priv->packetSize)) == NULL)
        return( &IRBadAllocation );
    packet->pos = priv->pos;
    packet->packetType = NetAllPackets;
    packet->pLayer = pLayer;
    packet->down = NULL;

    nBytes = ReadData(priv->fd, (char *)packet->buffer, priv->packetSize);
    if (nBytes < 0) {
        QueuePacket(&pLayer->freeList, packet);
        return( CreateSysReason("Can't read from %s\n", priv->source) );
    }
    if ((packet->dataBytes = nBytes) == 0) {
	(*packet->pLayer->ops.free)(packet);
	MarkEofInLayersAbove(pLayer, True);
	return( NULL );
    }

    priv->pos += packet->dataBytes;

    if (pLayer->pkiList)
	return( DeliverLayersAbove(pLayer, packet) );
    QueuePacket(&pLayer->queue, packet);
    return(NULL);
}

static InternalReasonPtr
NetProcess(PacketLayerPtr pLayer)
{
    return( &CantProcess );
}

static InternalReasonPtr
NetFlush(PacketLayerPtr pLayer)
{
    return( NULL );
}

static InternalReasonPtr
NetSeek(PacketLayerPtr pLayer, ULong absOffset)
{
    NetPrivPtr priv = (NetPrivPtr) pLayer;
    ULong curOffset;
    ULong packetNum;

    MarkEofInLayersAbove(pLayer, False);

    packetNum = absOffset / priv->packetSize;
    if (packetNum * priv->packetSize != absOffset) {
	/*
	 * Truncate to point *before* requested offset
	 * so we get the byte requested
	 */
	absOffset = packetNum * priv->packetSize;
    }

    /*
     * By the time we are called, all packets in the system have
     * been released and should be on our free queue (unless some
     * packet is being, illegally held by some unknown software layer)
     * Make an attempt to use packets already in existence
     */
    curOffset = InternalPacketSeek(pLayer, absOffset);
    if (curOffset == priv->pos)
	return(NULL);		/* sucess! */
    
    return(&CantSeek);
#if	   0
    /*
     * For the net layer, the requested offset maps directly to
     * the file offset
     */
    where = SeekFile(priv->fd, absOffset, SK_Start);
    if (where < 0)
	return( CreateSysReason("Can't seek on %s\n", priv->file) );
    if (where != absOffset) {
	reason = CreateReason("Requested seek(%d) on %s, got %d\n",
			      absOffset, priv->file, where);
	return( reason );
    }
    priv->pos = where;
    return(NULL);
#endif	/* 0 */
}

static void
NetClose(PacketLayerPtr pLayer)
{
    NetPrivPtr priv = (NetPrivPtr) pLayer;
    int refCount = pLayer->refCount - 1;

    if (refCount > 1) {
	pLayer->refCount = refCount;
	return;
    }

    CloseFile(priv->fd);

    FreePacketList(&pLayer->freeList);
    FreePacketList(&pLayer->queue);
    free((char *)priv->source);
    free((char *)pLayer);
}

/*
 * Host specifications come in 3 flavors:
 *	host:<host>:<port>	make a TCP connection to source/port
 *		                and await  mpeg bits
 *	net:<net>:<port>	        listen to UDP port for packetes from anyone
 *	group:<group>:<port>	join multicast group and listen to UDP port
 *		                for packets from anyone
 */
static boolean
parseHostSpec(char *source, PacketLayerPtr pLayer)
{
    char **stringList = AllocStringList(source,':');
    NetPrivPtr priv = (NetPrivPtr)pLayer;
    boolean success = True;

    if (stringList == NULL || stringList[0] == NULL ||
	stringList[1] == NULL) {
	return(False);
    }

    if (strcmp(stringList[0],"host") == 0) {
	if (stringList[2] == NULL) {
	    return(False);
	}
	priv->source = CopyOfString(stringList[1]);
	priv->castType = uniCast;
	priv->port = StringToInt(stringList[2]);
    } else if (strcmp(stringList[0],"net") == 0) {
	priv->source = CopyOfString(stringList[1]);
	priv->castType = broadCast;
    } else if (strcmp(stringList[0],"group") == 0) {
	priv->source = CopyOfString(stringList[1]);
	priv->castType = multiCast;
    } else {
	success = False;
    }


    free(stringList);
    return(success);
}

PacketLayerPtr
OpenNetPacket(NetPacketConfPtr conf, InternalReasonPtr *pReason)
{
    FileHandle fd;
    NetPrivPtr priv;
    PacketLayerPtr pLayer;
    InternalReasonPtr reason;

    pLayer = AllocPacketLayer(PacketNet, NULL, sizeof(NetPriv));
    if (pLayer == NULL) {
	*pReason = &IRBadAllocation;
	return(NULL);
    }
    if (parseHostSpec(conf->source, pLayer) != True) {
	*pReason = CreateSysReason("Can't open %s\n", conf->source);
	return(NULL);
    }

    priv = (NetPrivPtr)pLayer;

    switch(priv->castType) {
    case uniCast:
	fd = MakeTcpConnection(priv->source, priv->port, &reason);
	break;
    case multiCast:
    case broadCast:
	fd = MakeUdpListenSocket(priv->port, priv->castType, priv->source,
				 &reason);
    }
    if (fd < 0) {
	*pReason = reason;
	return(NULL);
    }
    priv->fd = fd;
    priv->warning = conf->warning;
    priv->packetSize = conf->packetSize;
    if (priv->packetSize == 0)
        priv->packetSize = DefaultPacketSize;

    pLayer->ops.deliver = NetDeliver;
    pLayer->ops.process = NetProcess;
    pLayer->ops.flush   = NetFlush;
    pLayer->ops.close   = NetClose;
    pLayer->ops.seek    = NetSeek;
    pLayer->ops.free    = NetFree;

    return(pLayer);
}
