/*
 * Copyright (c) 1992, 1993, 1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	nvrPacket.c - central packet support routines.
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/nvrPacket.c,v 1.20 1995/10/02 04:56:29 toddb Exp $
 */

#include "nvrPacketI.h"
#include "nvrAlloc.h"
#include "nvrString.h"
#include "mpeg1Proto.h"

NameEnumPair PacketLayerNames[] = {
    { "mpeg1 sys",   PacketMpeg1Sys },
    { "mpeg1 video", PacketMpeg1Video },
    { "mpeg1 audio", PacketMpeg1Audio },
    { "mpeg1 data",  PacketMpeg1Data },
    { "mpeg1 other", PacketMpeg1Other },
    { "CD-i",        PacketCDI },
    { "cdi",         PacketCDI },
    { "Video CD",    PacketVCD },
    { "vcd",         PacketVCD },
    { "CD",          PacketCD },
    { "cd",          PacketCD },
    { "file sys",    PacketFs },
    { "fs",          PacketFs },
    { "net",         PacketNet },
    { NULL,          PacketLayerUnknown }
};

static char *DebugNames[] = {
    "packet discard",
    "packet queue",
};

FlagNames PacketDebugNames = {
    DBI_Packet,
    DebugNames,
    NumberOfCells(DebugNames)
};

FlagMacroList PacketDebugMacros = { NULL, 0 };

static void
ShowPacket(PacketDataPtr packet, int indent)
{
    Msg("%*s%s,addr=0x%x,sz=%d,pos=%d,ref=%d,type=%x\n",
	indent, "",
	EnumToName(packet->pLayer->type, PacketLayerNames),
	packet,
	packet->dataBytes,
	packet->pos,
	packet->refCount,
	packet->packetType);
}

void
ShowPacketQueue(PacketLayerPtr layer)
{
    PacketDataPtr pkt;
    char *name;
    int indent = 0;

    while (layer) {
	name = EnumToName(layer->type, PacketLayerNames);
	if (indent)
	    Msg("%*s", indent, "");
	Msg("%s queue: %s\n", name, layer->queue ? "" : "empty");
	for (pkt = layer->queue; pkt; pkt = pkt->next) {
	    ShowPacket(pkt, indent+4);
	}
	if (indent)
	    Msg("%*s", indent, "");
	Msg("%s free: %s\n", name, layer->freeList ? "" : "empty");
	for (pkt = layer->freeList; pkt; pkt = pkt->next) {
	    ShowPacket(pkt, indent+4);
	}
	layer = layer->down;
	indent += 8;
    }
}

#ifdef DEBUG_ENABLED

static PacketDataPtr AllPackets;

static void
RemoveDebugPacket(PacketDataPtr target)
{
    PacketDataPtr pkt, prevPacket;

    prevPacket = NULL;
    for (pkt = AllPackets; pkt; prevPacket = pkt, pkt = pkt->dbgNext) {
	if (pkt != target)
	    continue;
	if (prevPacket == NULL)
	    AllPackets = pkt->dbgNext;
	else
	    prevPacket->dbgNext = pkt->dbgNext;
	pkt->next = NULL;
	return;
    }
    Msg("can't remove packet from debug queue!\n");
    return;
}

static void
QueueDebugPacket(PacketDataPtr new)
{
    PacketDataPtr packet = AllPackets;

    if (packet == NULL) {
	AllPackets = new;
	return;
    }
    while (packet->dbgNext)
        packet = packet->dbgNext;
    packet->dbgNext = new;
    new->next = NULL;
}

static void
ShowPacketHierarchy(PacketDataPtr packet, int indent)
{
    for (; packet; packet = packet->sibling) {
	ShowPacket(packet, indent);
	if (packet->dbgUpList)
	    ShowPacketHierarchy(packet->dbgUpList, indent+4);
    }
}

static void
ShowAllPackets(char *msg)
{
    PacketDataPtr packet;
    int nPackets = 0;

    for (packet = AllPackets; packet; packet = packet->dbgNext) {
	packet->dbgUpList = packet->sibling = NULL;
	nPackets++;
    }

    Msg("%s: %d packets\n", msg, nPackets);
    for (packet = AllPackets; packet; packet = packet->dbgNext) {
	if (packet->down) {
	    packet->sibling = packet->down->dbgUpList;
	    packet->down->dbgUpList = packet;
	}
    }

    for (packet = AllPackets; packet; packet = packet->dbgNext) {
	if (packet->down == NULL)
	    ShowPacketHierarchy(packet, 0);
    }
}

#endif /* DEBUG_ENABLED */
    
PacketDataPtr
AllocPacket(PacketDataPtr *freeList, long dataSize)
{
    PacketDataPtr packet = NULL;
    PacketDataPtr free = *freeList;

    if (free) {
	/*
	 * Note that we pull a fast one here!  freeList is a pointer to
	 * a pointer.  But since the next pointer is first in the structure
	 * we can pretend it is the previous pointer in the list.
	 */
	PacketDataPtr prevFree = (PacketDataPtr)freeList;

	while (free && packet == NULL) {
	    /*
	     * reuse a packet if
	     *   - the requested size and packet size are both zero
	     *   - the requested size is non-zero and fits in packet size
	     */
	    if ((dataSize == 0 && free->allocBytes == 0)
	     || (dataSize > 0 && dataSize <= free->allocBytes)) {
		prevFree->next = free->next;	/* remove from linked list */
		packet = free;
	    }
	    prevFree = free;
	    free = free->next;
	}
    }

    if (packet == NULL) {
	if (sizeof(PacketData) + dataSize > MaxPacketSize)
	    return(NULL);
	packet = (PacketDataPtr) InsecureAlloc(sizeof(PacketData) + dataSize);
	if (packet == NULL)
	    return(NULL);
	memset((char *)packet, 0, sizeof(PacketData) + dataSize);
	DebugTask(PD_queue, QueueDebugPacket(packet));
	packet->allocBytes = dataSize;
    }

    packet->next = NULL;
    packet->refCount = 1;
    if (packet->allocBytes) /* reset buffer pointer */
	packet->buffer = (UChar *)(packet + 1);
    else
	packet->buffer = NULL;

    return(packet);
}

void
FreePacketList(PacketDataPtr *list)
{
    PacketDataPtr packet = *list;
    PacketDataPtr next;

    packet = *list;
    while (packet) {
	next = packet->next;
	free((char *)packet);
	DebugTask(PD_queue, RemoveDebugPacket(packet));
	packet = next;
    }
    *list = NULL;
}

void
QueuePacket(PacketDataPtr *queueList, PacketDataPtr new)
{
    PacketDataPtr packet = *queueList;

    if (packet == NULL) {
	*queueList = new;
    } else {
	while (packet->next)
	    packet = packet->next;
	packet->next = new;
    }
    new->next = NULL;
}

PacketLayerPtr
AllocPacketLayer(PacketLayerType type,
                 PacketLayerPtr nextLayer,
                 long privSize)
{
    PacketLayerPtr pLayer;
    long totalSize = privSize + sizeof(PacketLayer);

    pLayer = (PacketLayerPtr) InsecureAlloc(totalSize);
    if (pLayer == NULL)
        return( NULL );
    memset((char *)pLayer, 0, totalSize);
    pLayer->type = type;
    pLayer->refCount = 1;
    if ((pLayer->down = nextLayer))
	nextLayer->refCount++;

    return(pLayer);
}

InternalReasonPtr
RegisterPacketInterest(PacketLayerPtr pLayer, BitMaskPtr mask)
{
    PacketInterestPtr pki;
    PacketLayerPtr layerDown = pLayer->down;
    PacketInterestPtr new;
    InternalReasonPtr reason;
    PacketDataPtr packet, prevPacket, nextPacket;

    if (layerDown == NULL) {
	reason = InternalProblem("RegisterPacketInterest() no down layer\n");
	return(reason);
    }

    /*
     * First, check that the requested mask does not intersect with
     * existing masks.
     */
    for (pki = layerDown->pkiList; pki; pki = pki->next) {
	if ((mask && AnyBitsCommon(mask, &pki->mask))
	 || (mask == NULL && AnyBitsSet(&pki->mask))) {
	    reason = InternalProblem(
		"two sibling %s layers want packets from %s layer\n",
		EnumToName(pki->pLayer->type, PacketLayerNames),
		EnumToName(pLayer->type, PacketLayerNames));
	    return(reason);
	}
    }

    new = (PacketInterestPtr) InsecureAlloc(sizeof(PacketInterest));
    if (new == NULL)
	return( &IRBadAllocation );
    new->next = NULL;
    new->pLayer = pLayer;
    if (mask) {
	BM_COPY(&new->mask, mask);
    } else {
	BM_SETALL(&new->mask);
    }

    /*
     * link it in at the head of the list
     */
    new->next = layerDown->pkiList;
    layerDown->pkiList = new;

    /*
     * Look through the list of packets queued in the layer
     * below, and deliver them if they match the interest just
     * expressed.  Note that we have to be careful about removing
     * packets from the middle of the list.
     */
    prevPacket = (PacketDataPtr) &layerDown->queue;
    nextPacket = NULL; /* for lint */
    for (packet = layerDown->queue; packet; packet = nextPacket) {
	nextPacket = packet->next;
	if (BM_ISSET(&new->mask, packet->packetType)) {
	    prevPacket->next = nextPacket;
	    packet->next = NULL;
	    DeliverLayersAbove(layerDown, packet);
	} else {
	    prevPacket = packet;
	}
    }
    return (NULL);
}

/*
 * Discard any queued packets for this layer
 */
void
DiscardQueuedPackets(PacketLayerPtr pLayer)
{
    PacketDataPtr packet;

    while (packet = GetNextPacket(pLayer)) {
	(*packet->pLayer->ops.free)(packet);
    }
}

PacketDataPtr
GetNextPacket(PacketLayerPtr pLayer)
{
    PacketDataPtr packet;

    if (pLayer->queue) {
	packet = pLayer->queue;
	pLayer->queue = packet->next;
	packet->next = NULL;
	return(packet);
    }
    return(NULL);
}

InternalReasonPtr
ReadNextPacket(PacketLayerPtr pLayer)
{
    while (pLayer->down)
        pLayer = pLayer->down;

    return( (*pLayer->ops.deliver)(pLayer, NULL) ); /* this starts it rolling */
}

/*
 * pBuf is assumed to be a pointer into packet->buffer.
 * To get the offset of the beginning of the packet data,
 * call:
 *    GetFileOffsetOfPacket(packet, NULL)
 * Which is equivalent to
 *    GetFileOffsetOfPacket(packet, packet->buffer)
 */
ULong
GetFileOffsetOfPacket(PacketDataPtr packet, UChar *pBuf)
{
    PacketDataPtr bottomPacket = packet;

    if (pBuf == NULL)
	pBuf = packet->buffer;
    while (bottomPacket->down)
        bottomPacket = bottomPacket->down;
    return(bottomPacket->pos + pBuf - bottomPacket->buffer);
}

ULong
GetFileOffsetOfPacketOff(PacketDataPtr packet, long packetOff)
{
    ULong offset = GetFileOffsetOfPacket(packet, packet->buffer);

    return(offset + (packetOff - packet->pos));
}

char *
FormatOffset(char *buf, PacketDataPtr packet, UChar *pData)
{
    PacketDataPtr bottomPacket = packet;

    while (bottomPacket->down)
        bottomPacket = bottomPacket->down;
    sprintf(buf, "at byte %d",
	bottomPacket->pos + pData - bottomPacket->buffer);
    return(buf);
}

InternalReasonPtr
DeliverLayersAbove(PacketLayerPtr pLayer, PacketDataPtr packet)
{
    PacketInterestPtr pki = pLayer->pkiList;
    UShort type = packet->packetType;

    while (pki) {
	if (BM_ISSET(&pki->mask, type)) {
	    return( (*pki->pLayer->ops.deliver)(pki->pLayer, packet) );
	}
	pki = pki->next;
    }

    DebugTask(PD_discard,
	Msg("Drop %s packet, type %d\n",
	     EnumToName(packet->pLayer->type, PacketLayerNames),
	     packet->packetType);
    )

    /*
     * Nobody interested, discard it.
     */
    (*packet->pLayer->ops.free)(packet);
    return(NULL);
}

void
MarkEofInLayersAbove(PacketLayerPtr pLayer, boolean eof)
{
    PacketInterestPtr pki = pLayer->pkiList;

    pLayer->eof = eof;
    while (pki) {
	pki->pLayer->eof = eof;
	MarkEofInLayersAbove(pki->pLayer, eof);
	pki = pki->next;
    }
}

static InternalReasonPtr
CallFlushInLayersAbove(PacketLayerPtr pLayer)
{
    InternalReasonPtr reason;
    PacketInterestPtr pki = pLayer->pkiList;

    while (pki) {
        if (NULL != (reason = (*pki->pLayer->ops.flush)(pki->pLayer))
	 || NULL != (reason = CallFlushInLayersAbove(pki->pLayer)))
	    break;
        pki = pki->next;
    }
    return(reason);
}

InternalReasonPtr
FlushPacket(PacketLayerPtr pLayer)
{
    InternalReasonPtr reason;

    while (pLayer->down)
        pLayer = pLayer->down;
    if (NULL != (reason = (*pLayer->ops.flush)(pLayer)))
	return(reason);
    return(CallFlushInLayersAbove(pLayer));
}

static InternalReasonPtr
CallSeekInLayersAbove(PacketLayerPtr pLayer, ULong absOffset)
{
    InternalReasonPtr reason;
    PacketInterestPtr pki = pLayer->pkiList;

    while (pki) {
	reason = (*pki->pLayer->ops.seek)(pki->pLayer, absOffset);
	if (reason)
	    return(reason);
	DiscardQueuedPackets(pki->pLayer);
	reason = CallSeekInLayersAbove(pki->pLayer, absOffset);
	if (reason)
	    return(reason);
	pki = pki->next;
    }
    return(NULL);
}

InternalReasonPtr
SeekPacketOffset(PacketLayerPtr pLayer, ULong absOffset)
{
    InternalReasonPtr reason;

    while (pLayer->down)
        pLayer = pLayer->down;
    DebugTask(PD_queue, ShowAllPackets("before seek"));
    reason = CallSeekInLayersAbove(pLayer, absOffset);
    if (reason)
	return(reason);
    DebugTask(PD_queue, ShowAllPackets("after seek"));

    /*
     * all packets should be free now.
     * Call the bottom layer and it will do the real seek
     */
    return( (*pLayer->ops.seek)(pLayer, absOffset) );
}

void
RemovePacket(PacketDataPtr *pQueue, PacketDataPtr target)
{
    PacketDataPtr packet, prevPacket;

    packet = *pQueue;

    if (packet == target) {
	*pQueue = packet->next;
	packet->next = NULL;
	return;
    }

    prevPacket = packet;
    packet = packet->next;
    for ( ; packet; prevPacket = packet, packet = packet->next) {
	if (packet != target)
	    continue;
	prevPacket->next = packet->next;
	packet->next = NULL;
	return;
    }

    Msg("can't remove packet from queue!\n");
    return;
}

static PacketDataPtr
ExtractPacketContaining(PacketDataPtr *pQueue, ULong absOffset, boolean exact)
{
    PacketDataPtr packet;

    for (packet = *pQueue; packet; packet = packet->next) {
	if (exact) {
	    if (absOffset != packet->pos)
		continue;
	} else {
	    if (absOffset < packet->pos
	     || absOffset >= packet->pos + packet->dataBytes)
		continue;
	}
	RemovePacket(pQueue, packet);
	return(packet);
    }
    return(NULL);
}

ULong
InternalPacketSeek(PacketLayerPtr pLayer, ULong absOffset)
{
    PacketDataPtr packet;
    boolean exact = False;

    /*
     * Look through the list of packets queued in the layer's free list.
     * recycling as many packets as possible.
     */
    for (;;) {
	packet = ExtractPacketContaining(&pLayer->freeList, absOffset, exact);
	if (packet == NULL)
	    break;
	exact = True;
	absOffset = packet->pos + packet->dataBytes;
	packet->refCount = 1;
        QueuePacket(&pLayer->queue, packet);
    }

    return(absOffset);
}

/*
 * Given a packet just composed by a media layer, trim data
 * off the end of the packet that will go on the next packet.
 * Performace is not paramount here... delivering packets in
 * one piece is so that syntax can be checked correctly.
 */
PacketDataPtr
AlignPacket(PacketDataPtr packet, PacketDataPtr *free, InternalReasonPtr *pIrp)
{
    ULong startCode = ~0;
    UChar *pSrc;
    long dataBytes;
    PacketDataPtr aPacket;

    /*
     * We look backwards through the buffer looking for a system
     * start code.
     */
    *pIrp = NULL;
    pSrc = packet->buffer + packet->dataBytes;
    while (pSrc > packet->buffer) {
	startCode = (startCode >> 8) | (*--pSrc << 24);
	if ((startCode >> 8) == StartCodePrefix) {
	    switch (startCode) {
	    default:
		switch (startCode & 0xf0) {
		default:           continue;	/* not a packet */
		case MSId_Audio00: break;	/* audio packet */
		case MSId_Video00: break;	/* video packet */
		case MSId_Data00:  break;	/* data packet */
		case 0xb0:
		    switch (startCode & 0xff) {
		    case MSId_Reserved: break;	/* reserved packet */
		    case MSId_Private1: break;	/* private packet */
		    case MSId_Padding:  break;	/* padding packet */
		    case MSId_Private2: break;	/* private packet */
		    default: continue;		/* not a packet */
		    }
		    break;
		}
		/* fall through... */

	    case PackStartCode:
	    case SysStartCode:
	    case SysEndCode:
		/*
		 * If we scanned all the way back to the beginning of
		 * the buffer, then assume the entire thing is aligned.
		 */
		if (pSrc == packet->buffer)
		    return(NULL);

		/*
		 * everything from this point forward can be copied to the
		 * a new packet.
		 */
		aPacket = AllocPacket(free, packet->allocBytes);
		if (aPacket == NULL) {
		    *pIrp = AllocReason("alignment packet", packet->allocBytes);
		    return(NULL);
		}

		dataBytes = pSrc - packet->buffer;
		aPacket->dataBytes = packet->dataBytes - dataBytes;
		memcpy((char *)aPacket->buffer, pSrc, aPacket->dataBytes);
		packet->dataBytes = dataBytes;

		return(aPacket);
	    }
	}
    }

    /*
     * If we scanned all the way back to the beginning of
     * the buffer, then assume the entire thing is aligned.
     */
    return(NULL);
}
