/*
 * (c) Copyright 1992, 1993, 1994 North Valley Research, Inc.
 * 	All Rights Reserved
 *
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	nvrPacket.h - public interface for packet-level transport
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/nvrPacket.h,v 1.19 1995/12/29 18:54:47 toddb Exp $
 */
#ifndef _NVRPACKET_H_
#define _NVRPACKET_H_

#include "nvrTypes.h"
#include "nvrReason.h"
#include "nvrString.h"
#include "osio.h"

typedef struct _packetLayer PacketLayer, *PacketLayerPtr;
typedef struct _packetData PacketData, *PacketDataPtr;
typedef struct _packetInterest *PacketInterestPtr;

typedef InternalReasonPtr (*PacketDeliverFunc)(PacketLayerPtr, PacketDataPtr);
typedef InternalReasonPtr (*PacketSeekFunc)   (PacketLayerPtr, ULong absOffset);
typedef InternalReasonPtr (*PacketReasonFunc) (PacketLayerPtr);
typedef void              (*PacketVoidFunc)   (PacketDataPtr);
typedef void              (*PktLayerVoidFunc) (PacketLayerPtr);
typedef PacketDataPtr     (*PacketReadFunc)   (PacketLayerPtr,
					       long nBytes,
					       InternalReasonPtr *);

/*
 * These enums mark several things: packets (PacketData), packet layer
 * descriptions (PacketLayer).
 */
typedef enum {
    PacketLayerUnknown = 0,
    PacketMpeg1Sys,	/* mpeg 1 system packet layer */
    PacketMpeg1Video,	/* mpeg 1 video packet layer */
    PacketMpeg1Audio,	/* mpeg 1 audio packet layer */
    PacketMpeg1Data,	/* mpeg 1 data packet layer */
    PacketMpeg1Other,	/* mpeg 1 padding (and others) packet layer */
    PacketCDI,		/* cd-i logical packet layer */
    PacketVCD,		/* Video CD logical packet layer */

    PacketCD,		/* the physical cd */
    PacketFs,		/* the physical file system */
    PacketNet,		/* the physical network */
    NPacketLayerTypes	/* count of all the types above */
} PacketLayerType;

/*
 * This is the basic structure of a packet.
 */
struct _packetData {
    PacketDataPtr next;		/* must be first: used for free list or queue */
    PacketDataPtr down;		/* link to packet in next layer */
    UChar *buffer;		/* packet payload */
    UShort dataBytes;		/* length of data */
    UShort allocBytes;		/* number of data bytes allocated */ 
    UShort refCount;		/* reference count */
    UShort packetType;		/* layer-specific packet type */
    PacketLayerPtr pLayer;	/* packet layer for this packet */
    void *layerPriv;		/* layer-specific info, like packet header */
    ULong pos;			/* stream-relative position */
    long timecode;		/* millisecond timecode, if available */

    /*
     * For debugging: must be last
     */
#   ifdef DEBUG_ENABLED
    boolean shown;
    PacketDataPtr dbgUpList;	/* list of packets above this */
    PacketDataPtr sibling;	/* list of packets at this same level */
    PacketDataPtr dbgNext;
#   endif /* DEBUG_ENABLED */
}; /* typedef: PacketData, *PacketDataPtr */

#define MaxPacketSize	0xffff		/* a UShort with all bits set */

/*
 * This is the list of functions used to perform operations on the
 * packet stream.
 */
typedef struct _packetOps {
    PacketDeliverFunc deliver;	/* called internally: use ReadNextPacket() */
    PacketReasonFunc process;	/* top layer only: called to process packets */
    PacketReasonFunc flush;
    PktLayerVoidFunc close;
    PacketSeekFunc seek;	/* called internally, use SeekPacketOffset() */
    PacketVoidFunc free;
    PktLayerVoidFunc alignPackets;/* called internally, if non-null, it */
				/* forces packets to be delieverd on */
				/* start-code boundaries */
} PacketOps, *PacketOpsPtr;

/*
 * This is the basic structure of a (software) layer for packet
 * parsing.  When a packet mechanism is configured, there will be a colloction
 * of layers forming a tree.  The tree is doubly-linked with the
 * single 'down' and the upward to zero or more layers with 'pkiList'.
 */
struct _packetLayer {
    PacketLayerType type;
    PacketOps ops;
    PacketInterestPtr pkiList;	/* list of interested layers above this one */
    PacketLayerPtr down;	/* next packet layer down */
    PacketLayerPtr sibling;	/* list of other packet layers at this level */
    long packetSize;   		/* packet size layers above want */
    long refCount;		/* number of PacketLayer from here up */
    PacketDataPtr freeList;	/* queue of unused packets */
    PacketDataPtr queue;	/* queue containing ready packets */
    boolean eof;		/* true if eof detected */
    boolean timeCodes;		/* true if packet timeCode is valid */
}; /* typedef: PacketLayer, *PacketLayerPtr */

/*
 * During syntax checking, all tokens are formated for output in a linked
 * list of the items below.  See mpeg1vdec and mpeg1adec sources for
 * usage.
 */
typedef struct _formattedItem FormattedItem, *FormattedItemPtr;
struct _formattedItem {
    char *name;
    char *value;
    char *bits;
    long streamOffset;
    long fileOffset;
    long tokenOffset;
    UChar bitOffset;
    UChar level;
    short size;
    FormattedItemPtr next;
}; /* typedef: FormattedItem, *FormattedItemPtr */

extern NameEnumPair PacketLayerNames[];

CExternal PacketDataPtr GetNextPacket(
    PacketLayerPtr pLayer
);

CExternal InternalReasonPtr ReadNextPacket(
    PacketLayerPtr pLayer
);

#define OffsetBufSize		32	/* bytes to hold a byte offset string */

CExternal char *FormatOffset(
    char *buf,
    PacketDataPtr packet,
    UChar *pData
);

CExternal ULong GetFileOffsetOfPacketOff(
    PacketDataPtr packet,
    long packetOff	/* relative to packet->off */
);

CExternal ULong GetFileOffsetOfPacket(
    PacketDataPtr packet,
    UChar *pBuf			/* NULL or pointer into packet->buffer */
);

CExternal InternalReasonPtr FlushPackets(
    PacketLayerPtr
);

CExternal InternalReasonPtr SeekPacketOffset(
    PacketLayerPtr,
    ULong absOffset
);

#define GetPacketOffset(packet, pBuf)				\
    ((packet)->pos + ((pBuf) - (packet)->buffer))

/*************************************************************************
 *
 * public debug tools
 *
 *************************************************************************/
CExternal void ShowPacketQueue(PacketLayerPtr layer);

#endif /* _NVRPACKET_H_ */
