/*
 * Copyright (c) 1992, 1993, 1994 North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	vcdPacket.c - Video CD packet interface
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/vcdPacket.c,v 1.13 1995/04/08 19:06:24 toddb Exp $
 */

#include "nvrPacketI.h"
#include "nvrAlloc.h"
#include "vcdPacket.h"

static InternalReason CantProcess =
    { "Process called for vcd packet layer.\n", IR_InternalProblem };
static InternalReason BadMediaValue =
    { "OpenCdPacket: media is NULL\n", IR_InternalProblem };

typedef struct _vcdPriv {
    PacketLayer pub;
    void (*sectorCb)(PacketLayerPtr, PacketDataPtr, void *cdxaCbData);
    void *sectorCbData;
    void (*warning)(char *fmt, ...);
    int errorNum;
    long sectorNum;
} CdPriv, *CdPrivPtr;

static InternalReasonPtr
CdDeliver(PacketLayerPtr pLayer, PacketDataPtr packet)
{
    PacketDataPtr new;
    InternalReasonPtr reason = NULL;
    CdSectorType type;
    long bytesAvailable = packet->dataBytes;
    CdPrivPtr priv = (CdPrivPtr) pLayer;
    UChar *pBuf = packet->buffer;
    CdSubSectorPtr sh;
    long sectorNum;

    /*
     * This prevents the packet from being freed by callbacks called by
     * us before we are through parsing it.
     */
    packet->refCount++;

    while (reason == NULL && bytesAvailable >= CdSectorDataSize) {
	bytesAvailable -= CdSectorDataSize;
	sh = (CdSubSectorPtr)pBuf;
	pBuf += CdSectorDataSize;
	sectorNum = priv->sectorNum++;
	switch(sh->subHeader.s.subMode & (CDS_Data | CDS_Audio | CDS_Video)) {
	case CDS_Data:	type = CdDataSector; break;
	case CDS_Audio:	type = CdAudioSector; break;
	case CDS_Video:	type = CdVideoSector; break;
	case 0:		type = CdPadSector; break;
	default:
	    (*priv->warning)("%s #%d: sector %d: invalid submode: 0x%x\n"
		"VCD Syntax error", priv->sectorNum,
		++priv->errorNum, sh->subHeader.s.subMode);
	    (*packet->pLayer->ops.free)(packet);
	    return(NULL);
	}

	if ((new = AllocPacket(&pLayer->freeList, 0)) == NULL) {
	    (*packet->pLayer->ops.free)(packet);
	    return( &IRBadAllocation );
	}

	if (new->pLayer == NULL) {
	    new->dataBytes = CdPacketDataSize;
	    new->pLayer = pLayer;
	}
	new->buffer = (UChar *)sh->sector;
	new->pLayer = pLayer;
	new->packetType = type;
	new->down = packet;
	new->pos = sectorNum * CdPacketDataSize;
	new->layerPriv = (void *)sh;

	if (priv->sectorCb)
	    (*priv->sectorCb)(pLayer, new, priv->sectorCbData);

	if (pLayer->pkiList) {
	    reason = DeliverLayersAbove(pLayer, new);
	    if (reason) {
		(*new->pLayer->ops.free)(new);
		(*packet->pLayer->ops.free)(packet);
	    } else
		packet->refCount++;
	} else {
	    QueuePacket(&pLayer->queue, new);
	    packet->refCount++;
	}
    }

    if (reason == NULL && bytesAvailable) {
	char buf[ 128 ];
	reason = BadValueReason("incomplete VCD packet starting %s\n",
		    FormatOffset(buf, packet, pBuf));
    }

    /*
     * We are through parsing the packet... this will alow it to be freed.
     */
    if (--packet->refCount == 1)
        (*packet->pLayer->ops.free)(packet);

    return(reason);
}

static InternalReasonPtr
CdProcess(PacketLayerPtr pLayer)
{
    return( &CantProcess );
}

static InternalReasonPtr
CdFlush(PacketLayerPtr pLayer)
{
    return( NULL );
}

static InternalReasonPtr
CdSeek(PacketLayerPtr pLayer, ULong absOffset)
{
    CdPrivPtr priv = (CdPrivPtr) pLayer;

    priv->sectorNum = absOffset / CdSectorDataSize;
    priv->errorNum = 0;

    return( NULL );
}

static void
CdClose(PacketLayerPtr pLayer)
{
    FreePacketList(&pLayer->freeList);
    FreePacketList(&pLayer->queue);
}

static void
CdFree(PacketDataPtr packet)
{
    PacketLayerPtr pLayer = packet->pLayer;
    PacketDataPtr nextPacket = packet->down;

    if (--packet->refCount)
	return;
    (*nextPacket->pLayer->ops.free)(nextPacket);
    QueuePacket(&pLayer->freeList, packet);
}

PacketLayerPtr
OpenCdPacket(CdPacketConfPtr conf,
              PacketLayerPtr nextLayer,
              InternalReasonPtr *pReason)
{
    PacketLayerPtr pLayer;
    CdPrivPtr priv;

    if (nextLayer == NULL) {
        *pReason = &BadMediaValue;
        return(NULL);
    }
    if (nextLayer->packetSize % CdSectorDataSize != 0) {
	*pReason = BadValueReason("%s: %d\n",
			"CD packet layer: configured packet size is wrong",
			nextLayer->packetSize);
	return(NULL);
    }

    pLayer = AllocPacketLayer(PacketVCD, nextLayer, sizeof(CdPriv));
    if (pLayer == NULL) {
	*pReason = &IRBadAllocation;
	return(NULL);
    }
    priv = (CdPrivPtr) pLayer;
    priv->sectorCb = conf->sectorCb;
    priv->sectorCbData = conf->sectorCbData;
    priv->warning = conf->warning;

    pLayer->ops.deliver = CdDeliver;
    pLayer->ops.process = CdProcess;
    pLayer->ops.flush   = CdFlush;
    pLayer->ops.close   = CdClose;
    pLayer->ops.seek    = CdSeek;
    pLayer->ops.free    = CdFree;

    return(pLayer);
}
