// $Header: /usr/local/src/nvrToMIT/RCS/Ring.C,v 1.1 1994/03/10 23:45:58 russ Exp $

/*
 * $Id: Ring.C,v 1.1 1994/03/10 23:45:58 russ Exp $
 *
 * Ring.C
 *
 * Copyright Pencom Software, 1994
 */

#include <Ring.h>
#include <Mutex.h>

const size_t MinSize = 1024*1024;

Ring::
Ring( speed_t s, size_t netbufsz, int empty_full_ratio, Logger& log, 
      Ring::type rtype )
 : rtype( rtype ), key( -1 ), shmid( -1 ), sem( 1 )
{
	size_t chunks_per_empty = empty_full_ratio;
	size_t chunks_per_full = 1;

	count_t best_size = s * 4;
	if( best_size < MinSize )
	{	best_size = MinSize;
	}

	size_t num_chunks = chunks_per_empty;
	size_t total_size;
	while( (total_size = num_chunks * netbufsz ) < best_size )
	{	num_chunks += chunks_per_empty;
	}
	if( num_chunks % 2 != 0 ) num_chunks++;
	total_size = num_chunks * netbufsz;

CBR_DEBUG( CBR_DEBUG_RING_2, 
	cout << log << "netbufsz        =" << netbufsz << endl
	     << log << "total_size      =" << total_size << endl
	     << log << "num_chunks      =" << num_chunks << endl
	     << log << "chunks_per_empty=" << chunks_per_empty << endl
	     << log << "chunks_per_full =" << chunks_per_full << endl;
);

	alloc( total_size, num_chunks, chunks_per_empty, chunks_per_full );
}

Ring::
Ring( 
	size_t size,    // the buffer size in Kbytes
	count_t chunks,  // divide the buffer into this many chunks
	count_t in_bpe, // Number of buffers that constitute an empty
			// buffer.
	count_t in_bpf, // Number of buffers that constitute a full
			// buffer.
	Ring::type rtype // type of ring to create
) : rtype( rtype ), key( -1 ), shmid( -1 ), sem( 1 )
{
	alloc( size, chunks, in_bpe, in_bpf );
}

void
Ring::
alloc( 
	size_t size,    // the buffer size in Kbytes
	count_t chunks,  // divide the buffer into this many chunks
	count_t in_bpe, // Number of buffers that constitute an empty
			// buffer.
	count_t in_bpf  // Number of buffers that constitute a full
			// buffer.
)
{
	assert( size > 0 );
	assert( chunks > 0 && ( chunks % 2 ) == 0 );
	assert( in_bpe > 0 && in_bpe <= chunks && ( chunks % in_bpe ) == 0 );
	assert( in_bpf > 0 && in_bpf <= chunks && ( chunks % in_bpf ) == 0 );

	image_size = sizeof( RingInfo ) + sizeof(Buffer)*(size_t)chunks + size;

	image = alloc_image();

	ri = (RingInfo*)image;
	ri->bufs = (Buffer*)(image + sizeof( RingInfo ));
	ri->parent_buffer = image + sizeof( RingInfo ) + sizeof(Buffer)*chunks;

	ri->pbsize = size;
	ri->nbufs = chunks;
	ri->bpe = in_bpe;
	ri->bpf = in_bpf;
	ri->cbsize = ri->pbsize / (size_t)ri->nbufs;
	ri->max_empty = ri->nbufs / ri->bpe;
	ri->max_filling = ri->nbufs;
	ri->max_full = ri->nbufs / ri->bpf;
	ri->max_emptying = 0;			// zero means no auto consume

	size_t s = ri->cbsize;
	for( int i = 0; i < ri->nbufs; i++ )
	{	ri->bufs[i].init( ri->parent_buffer + ( i * s ), s );
	}

	clear();

}

Ring::
~Ring()
{	Logger log( "Ring::~Ring()" );

	sem.remove();

	if( shmid == -1 )
	{	delete[] image;
	}
	else
	if( shmid != -1 )
	{	shmid_ds buf;
		if( shmctl( shmid, IPC_STAT, &buf ) != -1 )
		{	if( buf.shm_cpid == getpid() )
			{	shmctl( shmid, IPC_RMID, &buf );
			}
			else
			if( buf.shm_nattch <= 1  )
			{	shmctl( shmid, IPC_RMID, &buf );
			}
			else
			{	shmdt( image );
			}
		}
		else
		{	perror( "shmctl()" );
			shmctl( shmid, IPC_RMID, &buf );
		}
	}
}

char*
Ring::
alloc_image()
{
	char* tp;

	switch( rtype )
	{
	case Private:
		tp = new char[image_size];
		assert( tp );
		break;
	case Shared:
		shmid = shmget( IPC_PRIVATE, image_size, 0666|IPC_CREAT );
		if( shmid == -1 )
		{	perror( "shmget" );
		}
		assert( shmid != -1 );

		tp = (char*)shmat( shmid, 0, 0 );
		if( tp == (char*)-1 )
		{	perror( "shmat" );
		}
		assert( tp != (char*)-1 );
		break;
	}

	return tp;

}

void
Ring::
set_max_emptying( count_t s )
{	Mutex m( sem );
	ri->max_emptying = s < ri->max_full ? s : ri->max_full;
	ri->max_emptying *= ri->bpf;
}

void
Ring::
set_max_filling( count_t s )
{	Mutex m( sem );
	ri->max_filling = s < ri->max_empty ? s : ri->max_full;
	ri->max_filling *= ri->bpe;
}

bufid_t
Ring::
get_empty(
	char** ptr,     // [out] pointer to an empty buffer
	size_t *size    // [out] size of the empty buffer
)
{	
	if( ri->nfilling >= ri->max_filling )
	{	ri->error = EBUSY;
		return -1;
	}

	size_t filled = 0;
	for( count_t i = ri->neb; i < ri->neb + ri->bpe; i++ )
	{	if( !ri->bufs[i].isempty() && !ri->bufs[i].ispartial() )
		{	ri->error = EBUSY;
			return -1;
		}
		if( ri->bufs[i].ispartial() )
		{	filled += ri->bufs[i].getused();
		}
	}

	*ptr = ri->bufs[ri->neb].getptr();
	*size = ri->bufs[ri->neb].getsize() * (size_t)ri->bpe;
	
	if( ri->bufs[ri->neb].ispartial() )
	{	*ptr += filled;
		*size -= filled;
	}

	for( i = ri->neb; i < ri->neb + ri->bpe; i++ )
	{	update_bufstate( ri->bufs[i], FILLING );
	}

	ri->leb = ri->neb;
	ri->neb = ( ri->neb + ri->bpe ) % ri->nbufs;

	if( !ri->psw.started() )
	{	ri->psw.start();
	}

	return ri->leb;

} // get_empty()

Buffer*
Ring::
get_empty()
{	Mutex m(sem );

	char* p;
	size_t s;
	Buffer* b = NULL;

	bufid_t bufid = get_empty( &p, &s );
	if( bufid != -1 )
	{	b = &ri->bufs[bufid];
		b->set_total( s );
	}

	return b;

} // get_empty()

Buffer*
Ring::
get_filling()
{	Mutex m( sem );

	Buffer* b = &ri->bufs[ri->nFb];
	return ( b->get_state() == FILLING ) ? b : NULL;
}

Buffer*
Ring::
get_emptying()
{	Mutex m( sem );

	Buffer* b = &ri->bufs[ri->nEb];
	return ( b->get_state() == EMPTYING ) ? b : NULL;
}

void
Ring::
set_emptying_to_full()
{	Mutex m( sem );

	for( bufid_t i = ri->nEb;
	     ri->bufs[i].get_state() == EMPTYING;
	     i = ( i + 1 ) % ri->nbufs )
	{	update_bufstate( ri->bufs[i], FULL );
	}
	
	ri->nfb = ri->nEb;
	
}

void
Ring::
set_filling_to_empty()
{	Mutex m( sem );

	for( bufid_t i = ri->nFb; 
	     ri->bufs[i].get_state() == FILLING;
	     i = ( i + 1 ) % ri->nbufs )
	{	update_bufstate( ri->bufs[i], EMPTY );
	}
	
	ri->neb = ri->nFb;
	
}

int
Ring::
now_empty( 
	bufid_t bufid 
)
{
	if( bufid < 0 || bufid + ri->bpf > ri->nbufs )
	{	ri->error = EINVAL;
		return -1;
	}

	for( bufid_t i = bufid; i < bufid + ri->bpf; i++ )
	{	update_bufstate( ri->bufs[i], EMPTY );
	}

	return 0; 

} // now_empty()

void
Ring::
now_empty()
{	Mutex m( sem );
	do_now_empty();
}

void
Ring::
do_now_empty()
{
	now_empty( ri->nEb );
	ri->nEb = ( ri->nEb + ri->bpf ) % ri->nbufs;
}

int
Ring::
back_to_empty( 
	bufid_t bufid
)
{	size_t u = 0;
	if( bufid < 0 || bufid + ri->bpe > ri->nbufs )
	{	ri->error = EINVAL;
		return -1;
	}

	for( bufid_t i = bufid; i < bufid + ri->bpe; i++ )
	{	update_bufstate( ri->bufs[i], EMPTY );
	}

	ri->neb = (count_t)bufid;

	return 0;

} // back_to_empty()

void
Ring::
back_to_empty()
{	Mutex m( sem );

	if( ri->leb != -1 )
	{	back_to_empty( ri->leb );
		ri->leb = -1;
	}
}

bufid_t
Ring::
get_full(
	char** ptr,     // [out] pointer to an empty buffer
	size_t *size    // [out] size of the empty buffer
)
{	
/*
 * Auto consume
 */
	if( ri->max_emptying && ri->nemptying == ri->max_emptying )
	{	do_now_empty();
	}

	for( bufid_t i = ri->nfb; i < ri->nfb + ri->bpf; i++ )
	{	if( !ri->bufs[i].isfull() )
		{	ri->error = EBUSY;
			return -1;
		}
	}

	size_t s = 0;

	for( i = ri->nfb; i < ri->nfb + ri->bpf; i++ )
	{	update_bufstate( ri->bufs[i], EMPTYING );
		s += ri->bufs[i].getused();
	}

	*ptr = ri->bufs[ri->nfb].getptr();
	*size = s;
	ri->consumed += s;

	ri->lfb = ri->nfb;
	ri->nfb = ( ri->nfb + ri->bpf ) % ri->nbufs;

	if( !ri->csw.started() )
	{	ri->csw.start();
	}

	return ri->lfb;

} // get_full()

Buffer*
Ring::
get_full()
{	Mutex m( sem );

	char* p;
	size_t s;
	Buffer* b = NULL;

	bufid_t bufid = get_full( &p, &s );
	if( bufid != -1 )
	{	b = &ri->bufs[bufid];
		b->set_total( s );
	}

	return b;

} // get_empty()

bufid_t
Ring::
next_full()
{	Mutex m( sem );

	for( bufid_t i = ri->nfb; i < ri->nfb + ri->bpf; i++ )
	{	if( !ri->bufs[i].isfull() )
		{	ri->error = EBUSY;
			return -1;
		}
	}

	return ri->nfb;

}

int
Ring::
now_full( 
	bufid_t bufid,
	size_t used
)
{
	if( bufid < 0 || bufid + ri->bpe > ri->nbufs )
	{	ri->error = EINVAL;
		return -1;
	}

	size_t u = 0;
	size_t filled = 0;

	for( bufid_t i = bufid; i < bufid + ri->bpe; i++ )
	{	filled += ri->bufs[i].getused();
	}

	used += filled;
	ri->produced += used;

	for( i = bufid; i < bufid + ri->bpe; i++ )
	{	if( ri->bufs[i].getsize() < used )
		{	u = ri->bufs[i].getsize();
		}
		else
		{	u = used;
		}
		used -= u;
		update_bufstate( ri->bufs[i], FULL, u );
	}

	return 0;

} // now_full()

void
Ring::
now_full( size_t used )
{	Mutex m( sem );

	now_full( ri->nFb, used );
	ri->nFb = ( ri->nFb + ri->bpe ) % ri->nbufs;
}

int
Ring::
back_to_full( 
	bufid_t bufid 
)
{
	if( bufid < 0 || bufid + ri->bpf > ri->nbufs )
	{	ri->error = EINVAL;
		return -1;
	}

	for( bufid_t i = bufid; i < bufid + ri->bpf; i++ )
	{	update_bufstate( ri->bufs[i], FULL );
	}

	ri->nfb = (size_t)bufid;

	return 0; 

} // back_to_full()

void
Ring::
back_to_full()
{	Mutex m( sem );

	if( ri->lfb != -1 )
	{	back_to_full( ri->lfb );
		ri->lfb = -1;
	}
}

#if 0	// not in use at this time
int
Ring::
now_partial( 
	bufid_t bufid,
	size_t used
)
{	size_t u = 0;
	size_t filled = 0;
	if( bufid < 0 || bufid + ri->bpe > ri->nbufs )
	{	ri->error = EINVAL;
		return -1;
	}

	for( bufid_t i = bufid; i < bufid + ri->bpe; i++ )
	{	filled += ri->bufs[i].getused();
	}

	used += filled;

	for( i = bufid; i < bufid + ri->bpe; i++ )
	{	if( ri->bufs[i].getsize() < used )
		{	u = ri->bufs[i].getsize();
		}
		else
		{	u = used;
		}
		used -= u;
		update_bufstate( ri->bufs[i], PARTIAL, u );
	}

	ri->neb = (size_t)bufid;

	return 0;

} // now_partial()
#endif

#if 0 // not in use at this time
int 
Ring::
will_fill( 
	bufid_t bufid, 
	size_t used
)
{	size_t u = 0;
	if( bufid < 0 || bufid + ri->bpe > ri->nbufs )
	{	ri->error = EINVAL;
		return -1;
	}

	for( bufid_t i = bufid; i < bufid + ri->bpe; i++ )
	{	u += ri->bufs[i].getused();
	}

	assert( used + u <= empty_size() );

	return( used + u == empty_size() );

}
#endif

void 
Ring::
clear()
{	Mutex m( sem );
	Logger log( "Ring::clear" );

	ri->error = 0;

	ri->neb = 0;
	ri->nfb = 0;
	ri->nEb = 0;
	ri->nFb = 0;
	
	ri->leb = -1;
	ri->lfb = -1;

	ri->nempty = ri->nbufs;
	ri->nemptying = 0;
	ri->nfull = 0;
	ri->nfilling = 0;
	ri->npartial = 0;

	ri->produced = 0;
	ri->consumed = 0;

	ri->psw.clear();
	ri->csw.clear();

	for( int i = 0; i < ri->nbufs; i++ )
	{	//update_bufstate( ri->bufs[i], EMPTY );
		ri->bufs[i].now_empty();
	}

}

void 
Ring::
update_bufstate( Buffer& b, bufstat_t next_state, size_t s )
{
	Logger log( "Ring::update_bufstate()" );

	switch( b.get_state() )
	{
	case EMPTY:    ri->nempty--; break;
	case EMPTYING: ri->nemptying--; break;
	case PARTIAL:  ri->npartial--; break;
	case FULL:     ri->nfull--; break;
	case FILLING:  ri->nfilling--; break;
	}

	if( ri->nfilling < 0 || ri->nemptying < 0 ||
	    ri->nfull < 0 || ri->nempty < 0 )
	{	cout << log << sem << endl;
		pause();
	}

	switch( next_state )
	{
	case EMPTY:    ri->nempty++; b.now_empty(); break;
	case EMPTYING: ri->nemptying++; b.now_emptying(); break;
	case PARTIAL:  ri->npartial++; b.now_partial( s ); break;
	case FULL:     ri->nfull++; (s==-1)?b.now_full():b.now_full( s );break;
	case FILLING:  ri->nfilling++; b.now_filling(); break;
	}

}

ostream&
operator<<( ostream& o, Ring& r )
{
	o << "rtype             = " << r.rtype << endl
	  << "key               = " << r.key << endl
	  << "shmid             = " << r.shmid << endl
	  << "image             = 0x" << hex << (long)r.image << endl
	  << "image_size        = " << r.image_size << endl
	  << "ri                = 0x" << hex << (long)r.ri << endl 
	  << endl
	  << "ri->parent_buffer = 0x" << hex << (long)r.ri->parent_buffer << endl
	  << "ri->pbsize        = " << dec << r.ri->pbsize << endl
	  << "ri->cbsize        = " << r.ri->cbsize << endl
	  << "ri->nbufs         = " << r.ri->nbufs << endl
	  << "ri->bpf           = " << r.ri->bpf << endl
	  << "ri->bpe           = " << r.ri->bpe << endl
	  << "ri->neb           = " << r.ri->neb << endl
	  << "ri->nFb           = " << r.ri->nFb << endl
	  << "ri->nfb           = " << r.ri->nfb << endl
	  << "ri->nEb           = " << r.ri->nEb << endl
	  << "ri->leb           = " << r.ri->leb << endl
	  << "ri->lfb           = " << r.ri->lfb << endl
	  << "ri->max_empty     = " << r.ri->max_empty << endl
	  << "ri->max_empty     = " << r.ri->max_empty << endl
	  << "ri->max_full      = " << r.ri->max_full << endl
	  << "ri->max_filling   = " << r.ri->max_filling << endl
	  << "ri->max_emptying  = " << r.ri->max_emptying << endl
	  << "ri->nempty        = " << r.ri->nempty << endl
	  << "ri->nemptying     = " << r.ri->nemptying << endl
	  << "ri->nfull         = " << r.ri->nfull << endl
	  << "ri->nfilling      = " << r.ri->nfilling << endl
	  << "ri->npartial      = " << r.ri->npartial << endl
	  << "ri->produced      = " << r.ri->produced << endl
	  << "ri->consumed      = " << r.ri->consumed << endl
	  << "ri->psw.diff()    = " << r.ri->psw.diff() << endl
	  << "ri->csw.diff()    = " << r.ri->csw.diff() << endl
	  << "ri->error         = " << r.ri->error << endl
	  << endl;

	for( int i = 0; i < r.ri->nbufs; i++ )
	{	o << "ri->bufs[" << i << "]:" << endl
		  << r.ri->bufs[i];
	}

	return o << endl;
}

#include <NetHead.h>

Ring*
alloc_ring( speed_t s, size_t netbufsz, int fd, int empty_full_ratio )
{
	Logger log( "alloc_ring()" );
	NetHead nh;

	int l = netbufsz + nh.size();
	int rc = setsockopt( fd, SOL_SOCKET, SO_SNDBUF, (char*)&l, sizeof(l));
	if( rc == -1 )
	{	cout << log << "setsockopt( SO_SNDBUF ) failed (errno=" 
		     << ::errno << ")" << endl;
		return NULL;
	}

CBR_DEBUG( CBR_DEBUG_RING_2, 
	cout << log << "SO_SNDBUF=" << l << endl;
);

	l = (netbufsz + nh.size())*4;
	rc = setsockopt( fd, SOL_SOCKET, SO_RCVBUF, (char*)&l, sizeof(l));
	if( rc == -1 )
	{	cout << log << "setsockopt( SO_SNDBUF ) failed (errno=" 
		     << ::errno << ")" << endl;
		return NULL;
	}

CBR_DEBUG( CBR_DEBUG_RING_2, 
	cout << log << "SO_RCVBUF=" << l << endl;
);

	Ring* ring = new Ring( s, netbufsz, empty_full_ratio, log );

	return ring;
	
}
