// $Header: /usr/local/src/nvrToMIT/RCS/SchedCons.C,v 1.1 1994/03/10 23:27:58 russ Exp $

/*
 * $Id: SchedCons.C,v 1.1 1994/03/10 23:27:58 russ Exp $
 *
 * SchedCons.C
 *
 * Copyright Pencom Software, 1994
 */

#include <SchedCons.h>
#include <NetHead.h>
#include <VTDprot.h>

static Logger glog( "/dev/null" );

SchedCons::
SchedCons( char* cntr_dir, speed_t max_speed, Feed* feed, count_t nfeeds, 
	   Logger& log )
 : cntr_dir( cntr_dir ), max_speed( max_speed ), Scheduler( feed, nfeeds ), 
   cntr( NULL )
{
        pid = sproc( SchedCons_entry, PR_SALL, this );
        if( pid == -1 )
        {       cout << log << "unable to create consumer (errno=" << ::errno
		     << ")" << endl;
        }
	glog = log;
}

SchedCons::
~SchedCons()
{	delete cntr;
}

static
void
cleanup( int signo, ... )
{
CBR_DEBUG( CBR_DEBUG_CONS_1,
	cout << glog << "consumer trapped signal " << signo 
	     << ", exiting..." << endl;
);
	exit( 1 );
}

void
SchedCons_entry( void* arg )
{
	signal( SIGHUP, SIG_IGN );
	signal( SIGINT, cleanup );
	signal( SIGQUIT, cleanup );
	signal( SIGTERM, cleanup );

	SchedCons* p = (SchedCons*)arg;
	p->run();
}

perr_t
SchedCons::
run()
{	
	Logger log( "SchedCons::run()" );

	cntr = new Counter( cntr_dir );

CBR_DEBUG( CBR_DEBUG_CONS_1,
	cout << log << "start scheduling..." << endl;
);

	for( bool_t stop = FALSE; !stop; )
	{	for( count_t i = 0; i < nfeeds; i++ )
		{
			switch( feed[i].get_state() )
			{
			case Feed::idle:
			case Feed::loaded:
				break;
			case Feed::paused:
			case Feed::playing:
				schedule( feed[i] );
				break;
			default:
				cout << log << "bad state, bye bye" << endl;
				stop = TRUE;
				break;
			}
		}
		sleep(1);
	}

	cout << log << "SchedCons::run(): leaving...should never get here" 
	     << endl;

	return pInval;		// should never return
}

void
SchedCons::
schedule( Feed& f )
{
	Logger log( "SchedCons::schedule()" );

	Ring* ring = f.get_ring();

	assert( ring );
	
	ring->set_max_emptying( ring->max_full() / 2 );
	Stopwatch csw;
	Stopwatch wsw;
	wsw.start();

	for(;;)
	{
		wsw.stop();
		switch( f.get_state() )
		{
		case Feed::idle:
		case Feed::paused:
		case Feed::loaded:
			release();
			continue;
		case Feed::playing:
			break;
		}

		csw.start();
		if( adjust_speed( f, ring ) )
		{	consume_buffer( f, ring );
		}
		csw.stop();
CBR_DEBUG( CBR_DEBUG_CONS_2,
		cout << log << "compute time is " << csw.diff() << endl
		     << log << "wait time is " << wsw.diff() << ", nbufs="
		     << f.get_nbufs() << endl;
);
		wsw.start();
		release();
	}

}

int
SchedCons::
adjust_speed( Feed& f, Ring* ring )
{
	Logger log( "SchedCons::adjust_speed()" );
 
//	speed_t new_speed = max( f.get_speed(), 8192 );
	speed_t new_speed = f.get_speed();

	int rc = TRUE;

	count_t n = f.get_nbufs();
	if( n < 0 )	// too fast, so slow down 
	{	f.inc_nbufs( 1 );
		rc = FALSE;
	}
	else
	if( n > 0 )	// too slow, so speed up
	{	new_speed = min( max_speed / cntr->count(), new_speed * 2 );
//	{	new_speed = max( new_speed, max_speed / cntr->count() );
CBR_DEBUG( CBR_DEBUG_CONS_2,
		cout << log << "f.get_speed=" << f.get_speed() << ",new_speed=" << new_speed << endl;
);
	}

	set_speed( new_speed, ring );

	return rc;
}

int
SchedCons::
consume_buffer( Feed& f, Ring* ring )
{
	Logger log( "SchedCons::consume_buffer()" );

	Buffer* b;
	while( f.get_rseq() != -1 &&
	       ( b = ring->get_emptying() ) != NULL )
	{	
		if( b->get_seq() < f.get_rseq() )
		{	// old data, just skip it
CBR_DEBUG( CBR_DEBUG_CONS_2,
			cout << log << "skipping old data. seq="
			     << b->get_seq() << endl;
);
			ring->now_empty();
			continue;
		}
		else
		if( b->get_seq() > f.get_rseq() )
		{
CBR_DEBUG( CBR_DEBUG_CONS_1,
			cout << log << "unable to reset, cached data has "
			     << "already expired! rseq=" << f.get_rseq() 
			     << ",seq=" << b->get_seq() << endl;
);
		}
		else
		{	// found the proper sequence
CBR_DEBUG( CBR_DEBUG_CONS_1,
			cout << log << "seq found!" << endl;
);
		}
		f.set_seq( (seq_t)f.get_rseq() );
		f.set_rseq( -1 );
		f.set_c_eom( FALSE );
		ring->set_emptying_to_full();
	}

	if( f.get_rseq() != -1 )
	{	cout << log << "unable to reset, no more cached data! rseq="
		     << f.get_rseq() << endl;
		f.set_seq( (seq_t)f.get_rseq() );
		f.set_rseq( -1 );
	}
	
	b = ring->get_full();
	if( b == NULL )
	{	if( f.get_p_eom() && !f.get_c_eom() )
		{	send_eom( f );
		}
		else
		{	// consumer starved
		}
		return TRUE;
	}

	if( b->get_total() == 0 && f.get_p_eom() && !f.get_c_eom() )
	{	send_eom( f );
#ifdef TCPEOM
		return TRUE;
#endif
	}
	else
	if( b->get_total() == 0 )
	{	return TRUE;
	}
	
	int rc = consume( f, b );
	if( rc < 0 )
	{	cout << log << "consumer failed (errno=" << ::errno
		     << ")" << endl;
		ring->back_to_full();
#if 0
		break;
#endif
	}
	else
	{
CBR_DEBUG( CBR_DEBUG_CONS_2,
		cout <<log << "consumed " << b->get_total() << " bytes" <<endl;
);
	}

	return TRUE;
}

int
SchedCons::
consume( Feed& f, Buffer* b )
{
	Logger log( "SchedCons::consume()" );

	int rc;

	b->set_seq( f.get_seq() );

	NetHead nh;
	nh.set_seq( f.get_seq() );
#ifdef TCPEOM
	nh.set_type( NetHead::Data );
#else
	nh.set_type( b->get_total() ? NetHead::Data : NetHead::EOM );
	if( !b->get_total() )
	{
CBR_DEBUG( CBR_DEBUG_CONS_1,
		cout << log << "sending eom..." << endl;
);
	}
#endif
	nh.set_size( b->get_total() );
	nh.set_offset( b->get_offset() );
	nh.set_checksum( 0 );

	ulong csum = comp_checksum( nh(), nh.size() );
	csum += comp_checksum( b->getptr(), b->get_total() );

	nh.set_checksum( csum );

CBR_DEBUG( CBR_DEBUG_CONS_2,
	cout << log << "checksum=" << nh.get_checksum() << endl;
);

#if 0
	rc = send( f.get_destin_fd(), nh(), nh.size(), 0 );
	if( rc == -1 )
	{	return -1;
	}
	if( rc != nh.size() ) cout << log << "rc=" << rc << " != nh.size()=" 
		<< nh.size() << endl;
	rc = send( f.get_destin_fd(), b->getptr(), f.get_netbufsz(), 0 );
	if( rc == -1 )
	{	return -1;	
	}
	if( rc != f.get_netbufsz() )
	{	cout << log << "rc=" << rc << " != f.netbufsz=" 
		     << f.get_netbufsz() << endl;
	}
#else
	
	static char lbuf[MaxUDP];
	memcpy( lbuf, nh(), nh.size() );
	memcpy( &lbuf[nh.size()], b->getptr(), b->get_total() );
	rc = send( f.get_destin_fd(), lbuf, nh.size() + f.get_netbufsz(), 0 );
	if( rc == -1 )
	{	return -1;	
	}
	if( rc != nh.size() + f.get_netbufsz() )
	{	cout << log << "rc=" << rc << " != size=" << 
		     nh.size() + f.get_netbufsz() << endl;
	}
#endif

	
	f.inc_seq();

	count_t n = f.get_nbufs();
	if( n < 0 )	// too fast, so slow down 
	{	f.inc_nbufs( 1 );
	}
	else
	if( n > 0 )	// too slow, so speed up
	{	f.inc_nbufs( -1 );
	}

	return rc;
}

int
SchedCons::
send_eom( Feed& f )
{
	Logger log( "SchedCons::send_eom()" );

	if( f.get_ring() )
	{	f.get_ring()->cstop();
		Time t;
		size_t s;
		double tp = f.get_ring()->ptp( t, s );
CBR_DEBUG( CBR_DEBUG_CONS_1,
		cout << log << "Produced " << s << " bytes in " << t 
		     << " seconds (" << tp << " bytes/sec" << endl;
);
		tp = f.get_ring()->ctp( t, s );
CBR_DEBUG( CBR_DEBUG_CONS_1,
		cout << log << "Consumed " << s << " bytes in " << t 
		     << " seconds (" << tp << " bytes/sec" << endl;
);
	}
	f.set_nbufs( 0 );
	f.set_c_eom( TRUE );

	int rc = 0;

#if TCPEOM
        VTDprot cmd( VS_PROT_EOM );

CBR_DEBUG( CBR_DEBUG_CONS_1,
        cout << log << "sending " << cmd;
);

        vtdprot_t* buf = cmd.send();
        rc = send( f.get_ctrl_if(), (char*)buf, vtdprot_sizeof( buf ), 0 );
        if( rc < 0 )
        {       cout << log << "failed sending eom (errno="
                     << ::errno << ")" << endl;
        }
#endif

	return rc;
}

