/*
 *      (c) Copyright 1992,1993,1994 North Valley Research, Inc.
 *              All Rights Reserved
 *
 *      Confidential -- All proprietary rights reserved.
 *      Licensed Material -- Property of North Valley Research, Inc.
 *
 *      This program is the property of North Valley Research, Inc. or others
 *      from whom North Valley Research has obtained a licensing right, and is
 *      considered by North Valley Research to be confidential.  It is protected
 *      by U.S. copyright law as an unpublished work and is
 *      furnished pursuant to a written license agreement.  It may
 *      not be used, copied or otherwise reproduced, or disclosed to
 *      others except in accordance with the terms and conditions of
 *      that agreement.
 *
 *      NAME
 *              calaudio.c - calibrate the difference in time between 
 *                           writing audio samples and hearing them.
 *
 *      REVISION
 *              $Header: /nvr/share/src/nvr/prod/library/mpegsched/RCS/mpegsched
.c,v 1.15 1996/01/14 20:06:39 toddb Exp toddb $
 */
#include "osio.h"
#include "nvrAudio.h"
#include "nvrClock.h"
#include "nvrAlloc.h"
#include "mu-law.h"

#include <math.h>

enum {
    DefaultSampleDuration = 200,/* 200 ms */
    DefaultDuration = 20,	/* 100 ms */
    DefaultAmplitude = 65,	/* 65% of total */
    DefaultFrequency = 880,	/* A-880 */
    DefaultPulseGap = 500,	/* 500 ms */
    DefaultAvPeriod = 10	/* each new period weighs 1/10th */
};

typedef union _sampleSet {
    UChar *ubyte;
    char *byte;
    short *word;
} SampleSet, *SampleSetPtr;

typedef struct _pulseParam {
    long amplitude;
    long frequency;
    double sampleDuration;
    double duration;
    float avPeriod;
    AudioForm dst;
    SampleSet buf;
    long bufSize;
    long gap;
    long nSamples;
    AudioFilePtr output;
} PulseParam, *PulseParamPtr;

static long
labs(long value)
{
    if (value < 0)
	value = -value;
    return(value);
}
static void
InitPulseParam(PulseParamPtr pulse)
{
    memset((char *)pulse, 0, sizeof(PulseParam));
    pulse->amplitude = DefaultAmplitude;
    pulse->frequency = DefaultFrequency;
    pulse->sampleDuration = DefaultSampleDuration / (float)MilliSecPerSecond;
    pulse->duration = DefaultDuration / (float)MilliSecPerSecond;
    pulse->avPeriod = 1.0 / DefaultAvPeriod;
    pulse->gap = DefaultPulseGap;
}

static InternalReasonPtr
Calibrate(PulseParamPtr pulse)
{
    ClockKeepPtr clock;
    InternalReasonPtr irp;
    long writeTarget, writeTime, readTime;
    long delay;
    char msgBuf[ 1024 ];
    BufFileHndl fp = GetBufferedIn();
    FileHandle fd = GetBufferedFileHandle(fp);
    double avgDelay;
    TimeSpec maxWait;
    TimeSpec zeroWait;

    clock = CreateClockKeep(1, MilliSecPerSecond, &irp);
    if (clock == NULL)
        return(irp);

    NormalOutput("configured hw delay: %d ms\n\n",
	AudioDelayMs(pulse->output));
    NormalOutput(" delta  avg\n");
    NormalOutput(" time   delta\n");
    NormalOutput("-------------\n");
    NormalOutput("--------------");
    (*clock->init)(clock);
    writeTarget = 0;
    avgDelay = 0;
    zeroWait.seconds = zeroWait.microSeconds = 0;
    maxWait.seconds = 0;
    maxWait.microSeconds = 0.75 * (MicroSecPerSec / MilliSecPerSecond)
		         * pulse->gap;
    for (;;) {
	do {
	    writeTime = (*clock->gettime)(clock);
	} while (writeTime < writeTarget);
	writeTime = writeTarget;
	writeTarget += pulse->gap;

	(void) WriteAudioSamples(pulse->output,
				 pulse->buf.ubyte,
				 pulse->nSamples);

	if (WaitForFileInput(fd, &maxWait)) {
	    readTime = (*clock->gettime)(clock);
	    do {
		GetBufferedString(msgBuf, sizeof(msgBuf), fp);
	    } while (WaitForFileInput(fd, &zeroWait));
	    delay = readTime - writeTime;
	    if (delay > MilliSecPerSecond)
		delay = MilliSecPerSecond;

	    avgDelay = delay * pulse->avPeriod
		     + avgDelay * (1.0 - pulse->avPeriod);
	    NormalOutput("%7d%7.2f", delay, avgDelay);
	} else {
	    NormalOutput("\n--------------");
	    continue;
	}
    }

    return(NULL);
}

static InternalReasonPtr
CreatePulse(PulseParamPtr pulse)
{
    long i, maxPcm, pcm, newPcm;
    long pulseWidth;
    double inputInc, input;
    InternalReasonPtr irp;
    char msgBuf[ 1024 ];

    pulse->nSamples = pulse->dst.rate
		    * pulse->dst.channels
		    * pulse->sampleDuration;
    if (pulse->nSamples & 1)
	pulse->nSamples++;
    pulse->bufSize = pulse->nSamples * DataFormToBytes[ pulse->dst.dataForm ];
    pulse->buf.byte = AllocNamedMem(pulse->bufSize, "sample buffer", &irp);
    if (pulse->buf.byte == NULL)
	return(irp);

    pulseWidth = pulse->dst.rate * pulse->dst.channels * pulse->duration;
    if (pulseWidth > pulse->nSamples)
	pulseWidth = pulse->nSamples;
    if (pulseWidth & 1)
	pulseWidth++;	/* force even for phase of pulse */

    /*
     * Create the pulse
     */
    inputInc = 2 * M_PI * pulse->frequency / pulse->dst.rate;
    input = 0.0;
    switch (pulse->dst.dataForm) {
    default:
	return(BadValueReason("can't handle %s output\n",
			      AudioAttrName(&pulse->dst, msgBuf)));
	break;

    case Audio8BitULaw:
	i = 0;
	maxPcm = 0x7fff * (pulse->amplitude / 100.0);
	while (i < pulseWidth) {
	    pcm = sin(input) * maxPcm;
	    input += inputInc;
	    pulse->buf.byte[ i++ ] = PcmShortToMulaw(input);
	    if (pulse->dst.channels == AudioStereo) {
		pulse->buf.byte[ i++ ] = PcmShortToMulaw(input);
	    }
	}
	/*
	 * Now attenuate the signal back to zero
	 */
	for (;;) {
	    newPcm = sin(input) * maxPcm;
	    if (labs(newPcm) >= labs(pcm))
		break;
	    pcm = newPcm;
	    input += inputInc;
	    pulse->buf.byte[ i++ ] = pcm;
	    if (pulse->dst.channels == AudioStereo)
		pulse->buf.byte[ i++ ] = pcm;
	}
	break;

    case Audio8BitPcm:
	i = 0;
	maxPcm = 0x7f * (pulse->amplitude / 100.0);
	while (i < pulseWidth) {
	    pcm = sin(input) * maxPcm;
	    input += inputInc;
	    pulse->buf.byte[ i++ ] = pcm;
	    if (pulse->dst.channels == AudioStereo)
		pulse->buf.byte[ i++ ] = pcm;
	}
	/*
	 * Now attenuate the signal back to zero
	 */
	for (;;) {
	    newPcm = sin(input) * maxPcm;
	    if (labs(newPcm) >= labs(pcm))
		break;
	    pcm = newPcm;
	    input += inputInc;
	    pulse->buf.byte[ i++ ] = pcm;
	    if (pulse->dst.channels == AudioStereo)
		pulse->buf.byte[ i++ ] = pcm;
	}
	break;

    case Audio16BitPcm:
	i = 0;
	maxPcm = 0x7fff * (pulse->amplitude / 100.0);
	while (i < pulseWidth) {
	    pcm = sin(input) * maxPcm;
	    input += inputInc;
	    pulse->buf.word[ i++ ] = pcm;
	    if (pulse->dst.channels == AudioStereo)
		pulse->buf.word[ i++ ] = pcm;
	}
	/*
	 * Now attenuate the signal back to zero
	 */
	for (;;) {
	    newPcm = sin(input) * maxPcm;
	    if (labs(newPcm) >= labs(pcm))
		break;
	    pcm = newPcm;
	    input += inputInc;
	    pulse->buf.word[ i++ ] = pcm;
	    if (pulse->dst.channels == AudioStereo)
		pulse->buf.word[ i++ ] = pcm;
	}
	break;
    }
    return(NULL);
}

int
main(int argc, char **argv)
{
    PulseParam pulse;
    InternalReasonPtr irp;
    AudioForm src;
    char msgBuf[ 1024 ];

    InitPulseParam(&pulse);

    pulse.output = OpenAudioDevice(NULL, &irp);
    if (pulse.output == NULL) {
	ShowReason(Msg, irp);
        return(0);
    }

    src.rate     = Audio44_1Khz;
    src.channels = AudioStereo;
    src.dataForm = Audio16BitPcm;

    irp = ChooseAudioForm(&src, pulse.output->u.devFormats, &pulse.dst);
    if (irp) {
	ShowReason(Msg, irp);
        return(0);
    }

    if (!InitAudioDevice(pulse.output, &pulse.dst)) {
	ShowReason(Msg, pulse.output->reason);
	return(0);
    }

    irp = CreatePulse(&pulse);
    if (irp) {
	ShowReason(Msg, irp);
        return(0);
    }
    NormalOutput("%.2fms pulse, %.2fms sample @ %s\n\n",
	pulse.duration, pulse.sampleDuration,
	AudioAttrName(&pulse.dst, msgBuf));

    irp = Calibrate(&pulse);
    if (irp) {
	ShowReason(Msg, irp);
        return(0);
    }
    return(0);
}
