%{
/* $Header: /usr/local/src/nvrToMIT/RCS/parseconfig.y,v 2.13 1992/09/15 00:05:57 toddb Exp $ */

/*
 * The yacc grammar below accepts something like:
 *
 *    "Macros" "{" <assignments> "}"
 *    "RcsBin" <abspath>
 *    "Tree" <label> "{" <treespec> "}"
 *    "Omission" <label> "{" <omitspec> "}"
 *    "Restriction" <label> "{" <restrictspec> "}"
 *
 * assignments is defined as
 *     <label> = <path> [ "," <label> = <path> ]
 * treespec is defined as
 *     <abspath> "-->" <relpath> [ "," <abspath> "-->" <relpath> ] ...
 * or "Omit" <relpath> [ "," <relpath> ] ...
 * or "Omit" <label> [ "," <label> ] ...
 * or "Only" <relpath> [ "," <relpath> ] ...
 * or "Only" <label> [ "," <label> ] ...
 * or "Include" <label> [ "," <label> ] ...
 * or "TreeRoot" <abspath>
 * or "RcsLabel" <string>
 * or "RcsVersion" <version>
 * or "DefaultRcsRevision"
 * or "UseRcs"
 * or "UseOnlyRcs"
 * or "DropRcs"
 * or "KeepRcsLink"
 * or "Template"
 * or "SubTree" <relpath> { <treespec> }
 *
 * onlyspec is defined as
 *    "Restriction" <label> "{" <label> [ "," <label> ] ... "}"
 * or "Restriction" <label> "{" <relpath> [ "," <relpath> ] ... "}"
 * or "SubTree" <relpath> { <onlyspec> }
 *
 * omitspec is defined as
 *    "Omission" <label> "{" <label> [ "," <label> ] ... "}"
 * or "Omission" <label> "{" <relpath> [ "," <relpath> ] ... "}"
 * or "SubTree" <relpath> { <omitspec> }
 *
 * label is a unique string w/ letters, numbers and underscore
 * abspath is an absolute pathname (leading '/')
 * relpath is a relative pathname (no leading '/')
 * path is a relative or absolute pathname
 *
 * For example, the following spec is legal
 *
 * Omission Demos { demo/one, demo/two }
 * Ommision Extra { doc, Demos, other }
 * Restriction Minimal { program/lib1 }
 * Tree Library { /src/library --> lib }
 * Tree Project1 { /src/foo --> . ,
 *		      Template,
 *                    /internal-src/fum --> fum
 *                    Omit Extra, lib/graphics, lib/math
 *                    /usr/lib/libc.a -> lib/libc.a
 *                  }
 * Tree Project2 { /src/foo --> . ,
 *		      Template,
 *                    /external-src/fum --> fum
 *                    Include Library
 *                  }
 *
 * Tree Special { TreeRoot /private/special
 *		     Include Project1
 *		     Only Minimal, program/lib2
 *		   }
 *
 */

#include "struct.h"
#include <fcntl.h>

#ifdef YYDEBUG
#define DebugToken(args) (yydebug>0) ? ShowToken args : yydebug
extern int yydebug;
#else
#define DebugToken(args)
#endif

int CurrentLabel = 0;
#define Expect(string)	TokenExpected = string
#define GotIt	TokenExpected = NULL
char *TokenExpected = NULL;
%}

%union {
    int intvalue;
    char *string;
}

%token            POINTER TREE RESTRICTION ONLY OMISSION OMIT DEFAULTRCSREVISION
%token            INCLUDE TREEROOT SUBTREE USERCS DROPRCS RCSLABEL KEEPRCSLINK
%token		  RCSBIN USEONLYRCS RCSVERSION TEMPLATE MACROS
%token <intvalue> OMITLABEL
%token <intvalue> TREELABEL
%token <intvalue> ONLYLABEL
%token <string>   ABSOLUTE_PATH
%token <string>   RELATIVE_PATH
%token <string>   STRING
%token <string>   VERSION
%type  <string>   RelativePath RelativePathToken
%type  <string>   AbsolutePath AbsolutePathToken
%type  <string>   String NewLabel
%type  <intvalue> TreeLabel OmitLabel OnlyLabel

%%

SpecList:
    SpecList Spec
  | Spec
    ;

Spec:
    ShowError | Macros | RcsBin | Tree | Omission | Restriction
    ;

Macros:
    MACROS LCurly MacroList RCurly
    ;

RcsBin:
    RCSBIN AbsolutePath
    { if (RcsBin)
	ConfigError("Warning: RcsBin respecified\n");
      RcsBin = $2;
    }
    ;

Tree:
    TREE NewLabel
    { CurrentLabel = NewToken($2, TREELABEL);
      NewTree(CurrentLabel);
    }
    LCurly TreeSpecs RCurly
    ;

Restriction:
    RESTRICTION NewLabel
    { CurrentLabel = NewToken($2, ONLYLABEL);
      NewTree(CurrentLabel);
    }
    LCurly OnlyList RCurly
    ;

Omission:
    OMISSION NewLabel
    { CurrentLabel = NewToken($2, OMITLABEL);
      NewTree(CurrentLabel);
    }
    LCurly OmitList RCurly
    ;

TreeSpecs:
    TreeSpecs Comma TreeSpecToken
  | TreeSpecToken
    ;

TreeSpec:
    ShowError
  | OMIT LCurly OmitList RCurly
  | OMIT OmitItem
  | ONLY LCurly OnlyList RCurly
  | ONLY OnlyItem
  | INCLUDE LCurly IncludeList RCurly
  | INCLUDE TreeLabel { CombineTreeLabels(CurrentLabel, $2); }
  | RCSLABEL String
    { SetRcsRevision(CurrentLabel,
		    ExpandTreeStack(NullString),
		    $2,
		    NamedLabel);
    }
  | RCSVERSION VERSION
    { SetRcsRevision(CurrentLabel,
		    ExpandTreeStack(NullString),
		    $2,
		    VersionNumber);
    }
  | DEFAULTRCSREVISION
    { SetRcsRevision(CurrentLabel,
		    ExpandTreeStack(NullString),
		    NullString,
		    DefaultRevision);
    }
  | TEMPLATE { LabelToTree(CurrentLabel)->dontConsider = TRUE; }
  | USERCS { UseRcs(CurrentLabel, ExpandTreeStack(NullString)); }
  | USEONLYRCS { UseOnlyRcs(CurrentLabel, ExpandTreeStack(NullString)); }
  | DROPRCS { DropRcs(CurrentLabel, ExpandTreeStack(NullString)); }
  | KEEPRCSLINK { KeepRcsLink(CurrentLabel, ExpandTreeStack(NullString)); }
  | SUBTREE RelativePath { PushSubTree($2); } SubtreeSpec { PopSubTree(); }
  | TREEROOT AbsolutePath { SetTreeRootToString(CurrentLabel, $2); }
  | AbsolutePath Pointer RelativePath
    { (void) AddPathMapToLabel(CurrentLabel, $1, $3); }
    ;

SubtreeSpec: '{' TreeSpecs '}' | TreeSpec;

MacroList:
    MacroList Comma MacroDeclaration
  | MacroDeclaration
    ;

MacroDeclaration:
    String Equals RelativePathToken { NewMacro($1, $3); }
  | String Equals AbsolutePathToken { NewMacro($1, $3); }
  | String Equals String            { NewMacro($1, $3); }
    ;

OnlyList:
    SUBTREE RelativePath { PushSubTree($2); } '{' OnlyList '}' { PopSubTree(); }
  | OnlyList Comma OnlyItem
  | OnlyItem
    ;

OnlyItem:
    ShowError
  | RelativePath
    { AddPathToLabel(CurrentLabel, $1, NODE_ONLY); }
  | OnlyLabel
    { AddFlagsFromLabel(CurrentLabel, $1, NODE_ONLY | NODE_HASONLY); }
    ;

OmitList:
    SUBTREE RelativePath { PushSubTree($2); } '{' OmitList '}' { PopSubTree(); }
  | OmitList Comma OmitItem
  | OmitItem
    ;

OmitItem:
    ShowError
  | RelativePath
    { AddPathToLabel(CurrentLabel, $1, NODE_OMIT); }
  | OmitLabel
    { AddFlagsFromLabel(CurrentLabel, $1, NODE_OMIT | NODE_HASOMIT); }
    ;

IncludeList:
    ShowError
  | IncludeList Comma TreeLabel
    { CombineTreeLabels(CurrentLabel, $3); }
  | TreeLabel
    { CombineTreeLabels(CurrentLabel, $1); }
    ;

RelativePath:
    String { $$ = ExpandTreeStack($1); }
  | RelativePathToken   { $$ = ExpandTreeStack($1); }
    ;

AbsolutePath:
    AbsolutePathToken { $$ = $1; }
    ;

/*
 * Errors
 */
Equals: { Expect("'='"); } '=' { GotIt; } ;
LCurly: { Expect("'{'"); } '{' { GotIt; } ;
RCurly: { Expect("'}'"); } '}' { GotIt; } ;
Comma: { Expect("','"); } ',' { GotIt; } ;
TreeLabel: { Expect("a tree label"); } TREELABEL { $$ = $2; GotIt; } ;
OmitLabel: { Expect("an omission label"); } OMITLABEL { $$ = $2; GotIt; } ;
OnlyLabel: { Expect("a restriction label"); } ONLYLABEL { $$ = $2; GotIt; } ;
Pointer: { Expect("-->"); } POINTER { GotIt; } ;
TreeSpecToken: { Expect("a tree spec directive or an absolute path"); }
	TreeSpec { Expect("',' or another tree spec directive"); GotIt; } ;
String: { Expect("a string"); } STRING { $$ = $2; GotIt; } ;
NewLabel: { Expect("a new label"); } STRING { $$ = $2; GotIt; } ;
AbsolutePathToken: { Expect("an absolute path"); }
			ABSOLUTE_PATH { $$ = $2; GotIt; } ;
RelativePathToken: { Expect("a relative path"); }
			RELATIVE_PATH { $$ = $2; GotIt; } ;

ShowError: error { Expected(TokenExpected); } ;
%%

typedef struct _InputFile {
    int fileSize;       /* number of characters in the file */
    char *fileName;     /* file name */
    char *contents;     /* stack of descriptors for nested include files */
    char *p;     	/* pointer to current point of file */
    char *previous;	/* pointer to point just before last token */
    char *macro;	/* pointer to point just before last token */
    int lineNumber;     /* line number in file */
    char *stringBuf;	/* storage for captured strings */
    int stringLength;   /* length of stringBuf */
    boolean insideMacro;
} InputFile;

typedef struct _macro {
    char *name;
    char *contents;
} Macro;

typedef struct _token {
    int type;
    char *name;
} Token;

#define NextChar(cReturned)						\
    for (;;) {								\
	if (ConfigFile.insideMacro) {					\
	    if ((cReturned = *ConfigFile.macro) == '\0') {		\
		ConfigFile.insideMacro = FALSE;				\
		continue;						\
	    }								\
	    ConfigFile.macro++;						\
	    break;							\
	} else {							\
	    if ((cReturned = *ConfigFile.p) == '$') {			\
		char _c, _savedc;					\
		char *_p = ++ConfigFile.p;				\
									\
		while (_c = *ConfigFile.p) {				\
		    if (!isalnum(_c) && _c != '_' && _c != '-')		\
			break;						\
		    ConfigFile.p++;					\
		}							\
		_savedc = _c;						\
		*ConfigFile.p = '\0';					\
		ConfigFile.macro = MacroToContents(_p);			\
		ConfigFile.insideMacro = TRUE;				\
		*ConfigFile.p = _savedc;				\
	    } else if (cReturned == '\0') {				\
		if (! OpenNextFile())					\
		    break;						\
	    } else {							\
		ConfigFile.p++;						\
		break;							\
	    }								\
	}								\
    }

#define ReplaceChar(cReturned)						\
    {									\
	if (cReturned) {	/* don't replace null characters */	\
	    if (ConfigFile.insideMacro)					\
		ConfigFile.macro--;					\
	    else							\
		ConfigFile.p--;						\
	}								\
    }

#define TokenIndex(token)	((token) - Tokens)

static Macro *Macros;
static Token *Tokens;
static Token Internal[] = {
    TREEROOT,	        "TreeRoot",
    TREE,	        "Tree",
    TEMPLATE,	        "Template",
    RESTRICTION,        "Restriction",
    OMISSION,	        "Omission",
    ONLY,	        "Only",
    OMIT,	        "Omit",
    INCLUDE,	        "Include",
    RCSLABEL,	        "RcsLabel",
    RCSVERSION,	        "RcsVersion",
    SUBTREE,	        "SubTree",
    USERCS,	        "UseRcs",
    USEONLYRCS,	        "UseOnlyRcs",
    DROPRCS,	        "DropRcs",
    KEEPRCSLINK,        "KeepRcsLink",
    DEFAULTRCSREVISION, "DefaultRcsRevision",
    RCSBIN,	        "RcsBin",
    MACROS,	        "Macros",
    0,			0
};

static InputFile ConfigFile;
static int ErrorCount;

static char *SingleTreeSpec[] = {CONFIGFILE, NULL };
static char **TreeSpecFiles;
static long TreeSpecIndex;

/*
 * Open the file, read it in, initialize the pointers.
 */
static boolean OpenNextFile()
{
    int fd;
    int size, i;

    if (ConfigFile.fileName = TreeSpecFiles[ TreeSpecIndex ])
	fd = open(ConfigFile.fileName, O_RDONLY);
    else
	return(FALSE);
    TreeSpecIndex++;

    if (fd < 0)
	return(FALSE);

    size = ConfigFile.fileSize = FileSize(fd);
    if (ConfigFile.contents)
	free(ConfigFile.contents);
    ConfigFile.contents = SecureAlloc(size + 1);
    ConfigFile.stringLength = BUFSIZ;
    ConfigFile.stringBuf = SecureAlloc(BUFSIZ);
    if (ConfigFile.contents == NullString
     || ConfigFile.stringBuf == NullString)
	Fatal("can't allocate storage\n");

    if (read(fd, ConfigFile.contents, size) < size)
	Fatal("can't read all of config file\n");
    (void) close(fd);
    ConfigFile.p = ConfigFile.contents;
    ConfigFile.contents[ size ] = '\0';
    ConfigFile.lineNumber = 1;
    return(TRUE);
}

boolean ReadTreeSpec()
{
    static boolean initialized;
    char *specList, *p;
    long nSpecs, i;

    /*
     * The specification of the source tree comes from three sources:
     *     1. the command line
     *     2. a configuration file (CONFIGFILE)
     *     3. optional local configuration files.
     * Parse the configuration file now so we are ready for command-line
     * labels.
     */

    if (!initialized) {
	if (specList = getenv("MKLINKS_TREESPECS")) {
	    for (nSpecs = 1, p = specList; *p ; ) {
		if (*p)
		    nSpecs++;
		while (*p && *p != ':')
		    p++;
		if (*p == ':')
		    p++;
	    }

	    TreeSpecFiles = (char **)malloc(sizeof(char *) * (nSpecs+1));
	    TreeSpecFiles[ 0 ] = CONFIGFILE;
	
	    p = specList;
	    for (i = 1; i < nSpecs; i++) {
		TreeSpecFiles[ i ] = p;
		while (*p && *p != ':')
		    p++;
		if (*p == ':')
		    *p++ = '\0';
	    }
	    TreeSpecFiles[ i ] = NULL;
	} else
	    TreeSpecFiles = SingleTreeSpec;

	if (!OpenNextFile())
	    Fatal("No configuration file (%s)\n", TreeSpecFiles[ 0 ]);

	for (i=0; Internal[ i ].name; i++)
	    (void) NewToken(Internal[i].name, Internal[i].type);
    } else
	return(FALSE);

    yyparse();
    if (ErrorCount) {
	if (VerboseLevel > 3)
	    ShowTrees(FALSE);
	else if (ShowTreeChoice)
	    ShowTrees(TRUE);
	Fatal("No action taken.\n");
    }
    return(TRUE);
}

static char *MacroToContents(name)
    char *name;
{
    static char noMacro[ 2 ] = { 0, 0 };
    long i;

    if (Macros) {
	for (i = 0; Macros[ i ].name; i++) {
	    if (strcmp(name, Macros[ i ].name) == 0)
		return(Macros[ i ].contents);
	}
    }
    return (noMacro);
}

static Token *StringToToken(string)
    char *string;
{
    long i;

    if (Tokens) {
	for (i = 0; Tokens[ i ].name; i++) {
	    if (strcmp(string, Tokens[ i ].name) == 0)
		return(&Tokens[ i ]);
	}
    }
    return(NULL);
}

int StringToTokenIndex(string)
    char *string;
{
    Token *token;

    if ((token = StringToToken(string)) == NULL)
	return(-1);
    return (TokenIndex(token));
}

int TokenIndexToType(index)
    int index;
{
    return (Tokens[ index ].type);
}

char *TokenIndexToString(index)
    int index;
{
    return (Tokens[ index ].name);
}

int NewToken(ptr, type)
    char *ptr;
    int type;
{
    int i;

    if (Tokens) {
	for (i = 0; Tokens[ i ].name; i++) {
	    if (strcmp(ptr, Tokens[ i ].name) == 0) {
		ConfigError("Token \"%s\" reused\n", ptr);
		return(i);
	    }
	}
    } else
	i = 0;

    Tokens = (Token *)NullRealloc((char *) Tokens,
				  (unsigned) (sizeof(Token) * (i + 2)));
    Tokens[i].name = ptr; /* space is already allocated */
    Tokens[i].type = type;

    Tokens[i+1].name = NULL;
    return(i);
}

void NewMacro(name, contents)
    char *name;
    char *contents;
{
    int i;

    if (Macros) {
	for (i = 0; Macros[ i ].name; i++) {
	    if (strcmp(name, Macros[ i ].name) == 0) {
		ConfigError("Macro \"%s\" redefined\n", name);
		return;
	    }
	}
    } else
	i = 0;

    Macros = (Macro *)NullRealloc((char *) Macros,
				  (unsigned) (sizeof(Macro) * (i + 2)));
    Macros[i].name = name; /* space is already allocated */
    Macros[i].contents = contents;	/* space is already allocated */

    Macros[i+1].name = NULL;
}

static void AbsorbCComment()
{
    while (! (ConfigFile.p[0] == '*' && ConfigFile.p[1] == '/')) {
	if (ConfigFile.p[0] == '\0')
	    Fatal("EOF while in C-style comment @ line %d\n",
		ConfigFile.lineNumber);
	if (ConfigFile.p[0] == '\n')
	    ConfigFile.lineNumber++;
	ConfigFile.p++;
    }
    ConfigFile.p += 2;
}

static void AbsorbShellComment()
{
    while (ConfigFile.p[0] != '\n') {
	if (ConfigFile.p[0] == '\0')
	    Fatal("EOF while in shell style comment @ line %d\n",
		ConfigFile.lineNumber);
	ConfigFile.p++;
    }
    ConfigFile.lineNumber++;
}

#ifdef YYDEBUG
/*VARARGS1*/
static int ShowToken(fmt, x0,x1,x2,x3,x4,x5,x6,x7,x8,x9)
    char *fmt;
{
    Msg("Token: ");
    Msg(fmt, x0,x1,x2,x3,x4,x5,x6,x7,x8,x9);
    Msg(", ");
}
#endif

/*
 * This gets called when we see a number as the first character.  scan the
 * input looking for a version number of the forms
 *	n.n
 *	n.n.n.n
 *	n.n.n.n.n.n
 *	etc
 * If it matches, return a copy of it.
 */
static char *GetRcsVersionNumber()
{
    char *pto, *pfrom, *pend, c;
    int dots = 0;
    int needNumber = 0;

    pto = ConfigFile.stringBuf;
    pend = pto + BUFSIZ;
    pfrom = ConfigFile.p;
    for (;;) {
	NextChar(c);

	switch (c) {
	case '.':
	    if (needNumber++)
		return(NullString);	/* two consecutive '.' */
	    break;

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	    needNumber = 0;
	    break;

	default:
	    goto parseComplete;
	}

	*pto++ = c;
	if (pto >= pend) {
	    ConfigError("Version number overflow\n");
	    return (NullString);
	}
    }

parseComplete:
    /*
     * must be an odd number of dots
     */
    if ((dots & 1) && ! needNumber) {
	*pto = '\0';
	return (CopyOfString(ConfigFile.stringBuf));
    }
    return (NullString);
}

/*
 * These defines mark the kind of string token being parsed.
 */
#define StringToken	0x00
#define AbsPathToken	0x01
#define RelPathToken	0x02

yylex()
{
    char *ptr, **pptr;
    char c, quote;
    int stringType;
    Token *token;

    /*
     * Start reading tokens.
     */
    stringType = StringToken;
    ConfigFile.previous = ConfigFile.p;

    for (;;) {
	NextChar(c);
	if (!c)
	    break;

	switch (c) {
	case '#':
	    AbsorbShellComment();
	    break;

	case '=':
	case ':':
	case '{':
	case '}':
	case ',':
	    DebugToken(("%c", c));
	    return (c);

	case '\n':	/* newlines don't appear inside of macros */
	    ConfigFile.previous = ConfigFile.p;
	    ConfigFile.lineNumber++;
	    break;

	case ' ':
	case '\t':
	    ConfigFile.previous = ConfigFile.p;
	    continue;

	case '-':
	    if (ConfigFile.insideMacro)
		pptr = &ConfigFile.macro;
	    else
		pptr = &ConfigFile.p;
	    if ((*pptr)[0] == '-' && (*pptr)[1] == '>') {
		*pptr += 2;
		return(POINTER);
	    }
	    goto DefaultCase;

	case '/':
	    /*
	     * This could be a C-style comment.
	     */
	    if (ConfigFile.insideMacro)
		pptr = &ConfigFile.macro;
	    else
		pptr = &ConfigFile.p;

	    if (**pptr == '*') {
		(*pptr)++;
		AbsorbCComment();
		break;
	    }

	    /*
	     * Nope... its an absolute path
	     */
	    stringType = AbsPathToken;
	    goto DefaultCase;

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	    /* this could be an RCS version number */
	    if (yylval.string = GetRcsVersionNumber()) {
		DebugToken(("%s", yylval.string));
		return(VERSION);
	    }
	    goto DefaultCase;

	default:
    DefaultCase:
	    ptr = ConfigFile.stringBuf;
	    ReplaceChar(c);
	    for (;;) {
		NextChar(c);
		if (!c)
		    break;
		switch (c) {
		case '=':
		case ':':
		case ',':
		case '{':
		case '}':
		case ' ':
		case '\n':
		case '\t':
		    ReplaceChar(c);
		    goto ParsedString;

		case '/':
		    /* should we allow C-style comments to be embedded ? */

		    stringType |= RelPathToken;
		    *ptr++ = c;
		    break;

		case '\\':
		    stringType |= RelPathToken;
		    if ((*ptr++ = *ConfigFile.p++) == '\0')
			Fatal("premature EOF after escape, line %d\n",
			    ConfigFile.lineNumber);
		    break;

		case '\'':
		case '"':
		    stringType |= RelPathToken;
		    quote = c;
		    for (;;) {
			NextChar(c);
			if (c == quote)
			    break;
			if (c == '\0')
			    Fatal("premature EOF after quote %c, line %d\n",
				quote, ConfigFile.lineNumber);
			*ptr++ = c;
		    }
		    break;

		default:
		    *ptr++ = c;
		    if (isalnum(c) || c == '_' || c == '-')
			break;
		    if (isgraph(c)) {
			stringType |= RelPathToken;
			break;
		    }
		    Fatal("unknown char 0%o in config file, line %d\n",
			c, ConfigFile.lineNumber);
		}
	    }
	ParsedString:
	    *ptr++ = '\0';
	    if (stringType) {
		yylval.string = CopyOfString(ConfigFile.stringBuf);
		DebugToken(("%s", yylval.string));
		if (stringType & AbsPathToken)
		    return(ABSOLUTE_PATH);
		else
		    return(RELATIVE_PATH);
	    }
	    if (token = StringToToken(ConfigFile.stringBuf)) {
		yylval.intvalue = TokenIndex(token);
		return(token->type);
	    }
	    yylval.string = CopyOfString(ConfigFile.stringBuf);
	    DebugToken(("%s", yylval.string));
	    return(STRING);

	/* all non-printables except tab, newline and space */
	case '\01': case '\02': case '\03': case '\04': case '\05':
	case '\06': case '\07': case '\10': case '\13':
	case '\14': case '\15': case '\16': case '\17': case '\20':
	case '\21': case '\22': case '\23': case '\24': case '\25':
	case '\26': case '\27': case '\30': case '\31': case '\32':
	case '\33': case '\34': case '\35': case '\36': case '\37':
	case '\177':
	    Fatal("unknown char 0%o in config file, line %d\n",
		c, ConfigFile.lineNumber);
	    break;
	}
    }

    if (OpenNextFile())
	return(yylex());
    else
	return (EOF);
}

void ShowLine(input)
    InputFile *input;
{
    char *p, *start, *end;

    if (input->previous == NULL)
	input->previous = input->p;
    for (p = input->previous; p >= input->contents; p--)
        if (*p == '\n')
            break;
    start = p+1;

    for (p = input->previous; *p; p++)
        if (*p == '\n')
            break;
    end = p;

    for (p = start; p < end; p++)
        Msg("%c", *p);
    Msg("\n");
    for (p = start; p < end; p++) {
	if (p == input->previous)
	    break;
        if (*p == '\t')
            Msg("\t");
        else
            Msg(" ");
    }

    Msg("^-----\n");
}

yyerror(msg)
    char *msg;
{
    ErrorCount++;
    ConfigError("yacc %s\n", msg);
    ShowLine(&ConfigFile);
}

Expected(what)
    char *what;
{
    if (what)
	ConfigError("expected %s\n", what);
    else {
	switch(yychar) {
	case ABSOLUTE_PATH:	what = "absolute path"; break;
	case RELATIVE_PATH:	what = "relative path"; break;
	case POINTER:		what = "'-->'"; break;
	case STRING:		what = "string"; break;
	case TREEROOT:		what = "'TreeRoot' string"; break;
	case TREE:		what = "'Tree' string"; break;
	case OMISSION:		what = "'Omission' string"; break;
	case OMIT:		what = "'Omit' string"; break;
	case INCLUDE:		what = "'Include' string"; break;
	case ':':		what = "':'"; break;
	case '{':		what = "'{'"; break;
	case '}':		what = "'}'"; break;
	case ',':		what = "','"; break;
	default:		what = "label?"; break;
	}

	ConfigError("unexpected %s\n", what);
    }
}

/*VARARGS1*/
void ConfigError(fmt, x0,x1,x2,x3,x4,x5,x6,x7,x8,x9)
    char *fmt;
{
    ErrorCount++;
    Msg("%s, line %d: ", ConfigFile.fileName, ConfigFile.lineNumber);
    Msg(fmt, x0,x1,x2,x3,x4,x5,x6,x7,x8,x9);
}

static char **TreeStack;
static int TopOfStack;

void PushSubTree(path)
    char *path;
{
    TreeStack = (char **)
	NullRealloc((char *) TreeStack,
		    (unsigned) (sizeof(char *) * ++TopOfStack));

    TreeStack[ TopOfStack - 1 ] = CopyOfString(path);
}

void PopSubTree()
{
    if (--TopOfStack < 0)
	Fatal("Stack underflow\n");
    free(TreeStack[ TopOfStack ]);
}

char *ExpandTreeStack(endOfPath)
    char *endOfPath;
{
    static char fullPath[ BUFSIZ ];

    if (TopOfStack == 0)
	if (endOfPath)
	    return (endOfPath);
	else
	    return (".");
    fullPath[ 0 ] = '\0';
    ConcatPath(fullPath, fullPath, TreeStack[ TopOfStack - 1 ]);
    if (endOfPath)
	ConcatPath(fullPath, fullPath, endOfPath);

    return (fullPath);
}
