/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *      transform.c  -  perform dct transforms
 *
 * DESCRIPTION
 *      Forward and inverse cosine transform calculations are performed
 *      here.
 *
 * $Header: /nvr/share/src/nvr/prod/applications/vmpegcomp/RCS/transform.c,v 7.0
 1995/07/26 19:20:30 toddb Exp $
 */
#include "osio.h"
#include "process.h"
#include "nvrString.h"
#include "nvrAlloc.h"

char *ProgramName;

#define UsageMacro(args)        { Msg args; Usage(); }

enum {
    Y0 = 0,
    Y1,
    Y2,
    Y3,
    CR,
    CB
};

static NameEnumPair YcNames[] = {
    { "y0", Y0 },
    { "y1", Y1 },
    { "y2", Y2 },
    { "y3", Y3 },
    { "cr", CR },
    { "cb", CB },
    { NULL, -1 },
};

typedef struct _pycContext {
    FileHandle fdIn;           /* input file */
    BufFileHndl fpOut;          /* output file */
    char *fileName;
    int frame;
    int width;
    int height;
    int mbAddr;
    int mbWidth;
    int mbHeight;
    int block;
    int ycSize;
    UChar *y;
    UChar *cr;
    UChar *cb;
} PycContext, *PycContextPtr;

static void
Usage(void)
{
    Msg("Usage: %s size=WxH mb=N blk=y0|y1|y2|y3|cr|cb [frame=%d] ycfile\n",
	ProgramName);
    Msg("For example:\n\n    %s size=352x240 mb=0 blk=y0 movie.yc\n",
	ProgramName);
    exit(1);
}

main(int argc, char **argv)
{
    PycContext ctx;
    UChar *line;
    int width, row, frame, bytesIn;

    ProgramName = *argv;

    memset((char *)&ctx, 0, sizeof(ctx));

    ctx.fpOut = GetBufferedOut();
    for (argc--, argv++; argc; argc--, argv++) {
        switch (**argv) {
        case '-':
            UsageMacro(("Unknown flag: \"%s\"\n", *argv));
            break;

        default:
            if (strncmp(*argv, "size=", 5) == 0) {
		if (sscanf(*argv+5, "%dx%d", &ctx.width, &ctx.height) != 2) {
		    UsageMacro(("Bad size spec: \"%s\"\n", *argv));
		}
            } else if (strncmp(*argv, "frame=", 6) == 0) {
                ctx.frame = atoi(*argv+6);
            } else if (strncmp(*argv, "mb=", 3) == 0) {
                ctx.mbAddr = atoi(*argv+3);
            } else if (strncmp(*argv, "blk=", 4) == 0) {
                ctx.block = NameToEnum(*argv, YcNames);
		if (ctx.block < 0) {
		    UsageMacro(("Bad block spec: \"%s\"\n", *argv));
		}
            } else if (strcmp(*argv, "usage") == 0) {
                Usage();
            } else if (ctx.fileName == NULL) {
		ctx.fileName = *argv;
	    } else {
                UsageMacro(("Unknown argument: \"%s\"\n", *argv));
            }
            break;
        }
    }

    if (ctx.fileName == NULL) {
	ctx.fdIn = 0;
    } else {
	ctx.fdIn = OpenFileForReading(ctx.fileName);
	if (ctx.fdIn < 0) {
	    SystemError("can't open %s for reading\n", ctx.fileName);
	    exit(1);
	}
    }

    if (ctx.width <= 0 || ctx.height <= 0) {
	UsageMacro(("width or height is <= 0 (%dx%d)\n",
	    ctx.width, ctx.height));
    }
    if ((ctx.width & 0xf) || (ctx.height & 0xf)) {
	UsageMacro(("width and height must be multiples of 16 (%dx%d)\n",
	    ctx.width, ctx.height));
    }
    ctx.mbWidth = ctx.width >> 4;
    ctx.mbHeight = ctx.height >> 4;

    ctx.ycSize = (ctx.width * ctx.height * 3) >> 1;
    ctx.y = (UChar *)InsecureAlloc(ctx.ycSize);
    if (ctx.y == NULL)
	Fatal("Can't allocate %d bytes for picture", ctx.ycSize);

    ctx.cr = ctx.y + ctx.width * ctx.height;
    ctx.cb = ctx.cr + (ctx.width * ctx.height / 4);

    for (frame = 0; frame <= ctx.frame; frame++) {
	bytesIn = ReadData(ctx.fdIn, ctx.y, ctx.ycSize);
	if (bytesIn != ctx.ycSize)
	    Fatal("frame %d: wanted %d bytes, got %d\n",
		frame, ctx.ycSize, bytesIn);
    }

    switch (ctx.block) { 
    case Y0: case Y1: case Y2: case Y3:
	width = ctx.width;
	line = ctx.y
	     + width * 16 * (ctx.mbAddr / ctx.mbWidth)
	     + 16 * (ctx.mbAddr % ctx.mbWidth);
	switch (ctx.block) { 
	case Y0: line += 0; break;
	case Y1: line += 8; break;
	case Y2: line += width * 8; break;
	case Y3: line += width * 8 + 8; break;
	}
	break;

    case CR:
	width = ctx.width >> 1;
	line = ctx.cr
	     + width * 8 * (ctx.mbAddr / ctx.mbWidth)
	     + 8 * (ctx.mbAddr % ctx.mbWidth);
	break;

    case CB:
	width = ctx.width >> 1;
	line = ctx.cb
	     + width * 8 * (ctx.mbAddr / ctx.mbWidth)
	     + 8 * (ctx.mbAddr % ctx.mbWidth);
	break;
    }

    for (row = 0; row < 8; row++) {
	NormalOutput("%4d%4d%4d%4d%4d%4d%4d%4d\n",
	    line[0], line[1], line[2], line[3],
	    line[4], line[5], line[6], line[7]);
	line += width;
    }
    exit (0);
}

