/*
 * Copyright (c) 1992,1993, North Valley Research, Inc.
 * Confidential -- All proprietary rights reserved.
 * Licensed Material -- Property of North Valley Research, Inc.
 *
 * This program is the property of North Valley Research, Inc. or others
 * from whom North Valley Research has obtained a licensing right, and is
 * considered by North Valley Research to be confidential.  It is protected
 * by U.S. copyright law as an unpublished work and is
 * furnished pursuant to a written license agreement.  It may
 * not be used, copied or otherwise reproduced, or disclosed to
 * others except in accordance with the terms and conditions of
 * that agreement.
 *
 * NAME
 *	transform.c  -  perform dct transforms
 *
 * DESCRIPTION
 *	Forward and inverse cosine transform calculations are performed
 *	here.
 *
 * $Header: /usr/local/src/nvrToMIT/RCS/transform.c,v 1.4 1995/09/07 02:19:43 toddb Exp $
 */
#include "osio.h"
#include "process.h"
#include "nvrString.h"

#include <math.h>

char *ProgramName;

#define UsageMacro(args)	{ Msg args; Usage(); }

/*
 * The follwing factors must be multipled by the IDCT and FDCT data
 * (either before or after the transform) to get the mathematically
 * correct results.
 */
#define FDCTSCALE	1.0
#define IDCTSCALE	1.0

#define M_PI            3.14159265358979323846

typedef struct _xformContext XformContext, *XformContextPtr;
struct _xformContext {
    float sqrtOf2Reciprocal;	/* constant for dct transform */
    float refDctCoeffTable[64];	/* constants for dct transform */
    float refIDctCoeffTable[64];/* constants for idct transform */
    float srcMatrix[ 64 ];	/* source matrix */
    float dstMatrix[ 64 ];	/* destination matrix */
    BufFileHndl fpIn;		/* input file */
    BufFileHndl fpOut;		/* output file */
    boolean eof;		/* whether eof seen yet */
    boolean showBoth;		/* print both input and output blocks */
    boolean oneDimension;	/* do only a 1-D transform */
    boolean exact;		/* print both input and output blocks */
    void (*transform)(XformContextPtr);
};

static void
Usage(void)
{
    Msg("Usage: %s [1d|forward|showboth|exact] < coefficients\n", ProgramName);
    Msg("%s usage\nWill produce this message\n");
    exit(1);
}

static void
InitDctReference(XformContextPtr ctx)
{
    int ij, uv, i, u;
    double si = 1.0 / (2.0 * sqrt(IDCTSCALE));
    double sf = 1.0 / (2.0 * sqrt(FDCTSCALE));

    ctx->sqrtOf2Reciprocal = 1.0/sqrt(2.0);
    for (ij = 0; ij < 8; ij++) {
	for (uv = 0; uv < 8; uv++) {
	    ctx->refIDctCoeffTable[ij*8+uv] = si * cos(M_PI*(2*ij+1)*uv/16.0);
	    ctx->refDctCoeffTable[ij*8+uv]  = sf * cos(M_PI*(2*ij+1)*uv/16.0);
	}
	ctx->refIDctCoeffTable[ij*8+0] *= ctx->sqrtOf2Reciprocal;
    }
}

#define FDctCoeff(ctx, ij, uv)	ctx->refDctCoeffTable[ij*8 + uv]
#define IDctCoeff(ctx, ij, uv)	ctx->refIDctCoeffTable[ij*8 + uv]

/*
 * Reference forward Dct.
 *                         7   7             (2i+1)u*PI      (2j+1)v*PI
 * X(u,v) = (1/4)C(u)C(v) Sum Sum x(i,j) cos(----------) cos(-----------)
 *                        i=0 j=0                16              16
 *
 * Where x(i,j) is the pixel
 *       X(u,v) is the is the transformed coefficient
 *       C(0) = 1/sqrt(2), and C(u) = C(v) = 1 for u,v=1,...,7
 *
 * The (1/4) factor above is done externally (for compatibility with fast ver)
 */
static void
ReferenceDct2d8x8(XformContextPtr ctx)
{
    int i, j, u, v;
    double coeff;
    float *x = ctx->srcMatrix;
    float *X = ctx->dstMatrix;

    for(u = 0; u < 8; u++) {
	for(v = 0; v < 8; v++) {
	    coeff = 0.0;
	    for(i = 0; i < 8; i++) {
		for(j = 0; j < 8; j++) {
		    coeff += x[8*i+j]
			  * FDctCoeff(ctx, i,u) * FDctCoeff(ctx, j,v);
		}
	    }
	    if (u == 0)
		coeff *= ctx->sqrtOf2Reciprocal;
	    if (v == 0)
		coeff *= ctx->sqrtOf2Reciprocal;
	    X[8*u+v] = coeff;
	}
    }
}

/*
 * Reference forward Dct.
 *                   7           (2i+1)u*PI
 * X(u) = (1/2)C(u) Sum x(i) cos(----------)
 *                  i=0              16
 *
 * Where x(i) is the pixel
 *       X(u) is the is the transformed coefficient
 *       C(0) = 1/sqrt(2), and C(u) = C(v) = 1 for u,v=1,...,7
 *
 * The (1/2) factor above is done externally (for compatibility with fast ver)
 */
static void
ReferenceDct1d8x8(XformContextPtr ctx)
{
    int i, u;
    double coeff;
    float *x = ctx->srcMatrix;
    float *X = ctx->dstMatrix;

    for(u = 0; u < 8; u++) {
	coeff = 0.0;
	for(i = 0; i < 8; i++) {
	    coeff += x[i] * FDctCoeff(ctx, i,u);
	}
	if (u == 0)
	    coeff *= ctx->sqrtOf2Reciprocal;
	X[u] = coeff;
    }
}

/*
 * Reference inverse Dct.
 *                 7   7                      (2i+1)u*PI      (2j+1)v*PI
 * x(i,j) = (1/4) Sum Sum X(u,v) C(u)C(v) cos(----------) cos(-----------)
 *                u=0 v=0                          16              16
 *
 * The (1/4) factor above is done externally (for compatibility with fast ver)
 */
static void
ReferenceIDct2d8x8(XformContextPtr ctx)
{
    int i, j, u, v;
    double pixel;
    float *X = ctx->srcMatrix;
    float *x = ctx->dstMatrix;

    for(i = 0; i < 8; i++) {
	for(j = 0; j < 8; j++) {
	    pixel = 0.0;
	    for(u = 0; u < 8; u++) {
		for(v = 0; v < 8; v++) {
		    pixel += X[8*u+v]
			  * IDctCoeff(ctx, i,u) * IDctCoeff(ctx, j,v);
		}
	    }
	    x[8*i+j] = pixel;
	}
    }
}

/*
 * Reference inverse Dct.
 *               7                (2i+1)u*PI
 * x(i) = (1/2) Sum X(u) C(u) cos(----------)
 *              u=0                   16
 *
 * The (1/2) factor above is done externally (for compatibility with fast ver)
 */
static void
ReferenceIDct1d8x8(XformContextPtr ctx)
{
    int i, u;
    double pixel;
    float *X = ctx->srcMatrix;
    float *x = ctx->dstMatrix;

    for(i = 0; i < 8; i++) {
	pixel = 0.0;
	for(u = 0; u < 8; u++) {
	    pixel += X[u] * IDctCoeff(ctx, i,u);
	}
	x[i] = pixel;
    }
}

static boolean
GetNumber(XformContextPtr ctx, float *pNumber)
{
    int c;
    BufFileHndl fp = ctx->fpIn;
    char buf[ 1024 ], *p = NULL;

    if (ctx->eof)
	return(False);
    while (p == NULL) {
	c = GetBufferedChar(fp);
	switch (c) {
	case EndOfFile:
	    ctx->eof = True;
	    return(False);

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	case '.': case '-':
	    p = buf;
	    *p++ = c;
	    break;

	default:
	    break;	/* skip over all of this */
	}
    }

    for (;;) {
	c = GetBufferedChar(fp);
	switch (c) {
	default:
	    UngetBufferedChar(c, fp);
	    /* fall through... */

	case EndOfFile:
	    if (c == EndOfFile)
		ctx->eof = True;
	    *p = '\0';
	    *pNumber = atof(buf);
	    return(True);

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	case '.':
	    *p++ = c;
	    break;
	}
    }
}

static boolean
Get1DMatrix(XformContextPtr ctx)
{
    float *matrix = ctx->srcMatrix;
    int i;

    memset((char *)matrix, 0, 8 * sizeof(float));

    for (i = 0; i < 8; i++) {
	if (! GetNumber(ctx, &matrix[ i ]))
	    break;
    }

    return(i > 0);
}

static void
Transform1DMatrix(XformContextPtr ctx)
{
    int i;
    float *src = ctx->srcMatrix;
    float *dst = ctx->dstMatrix;

    (*ctx->transform)(ctx);
    if (ctx->exact) {
	if (ctx->showBoth) {
	    NormalOutput("%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f", 
		src[0], src[1], src[2], src[3],
		src[4], src[5], src[6], src[7]);
	    NormalOutput("  %s", (i == 0) ? "-->" : "   ");
	}
	NormalOutput("%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f\n", 
	    dst[0], dst[1], dst[2], dst[3],
	    dst[4], dst[5], dst[6], dst[7]);
    } else {
	if (ctx->showBoth) {
	    NormalOutput("%4d%4d%4d%4d%4d%4d%4d%4d", 
		(int)(src[0]+0.5), (int)(src[1]+0.5),
		(int)(src[2]+0.5), (int)(src[3]+0.5),
		(int)(src[4]+0.5), (int)(src[5]+0.5),
		(int)(src[6]+0.5), (int)(src[7]+0.5));
	    NormalOutput("  %s", (i == 0) ? "-->" : "   ");
	}
	NormalOutput("%4d%4d%4d%4d%4d%4d%4d%4d\n", 
	    (int)(dst[0]+0.5), (int)(dst[1]+0.5),
	    (int)(dst[2]+0.5), (int)(dst[3]+0.5),
	    (int)(dst[4]+0.5), (int)(dst[5]+0.5),
	    (int)(dst[6]+0.5), (int)(dst[7]+0.5));
    }
}

static boolean
Get2DMatrix(XformContextPtr ctx)
{
    float *matrix = ctx->srcMatrix;
    int i;

    memset((char *)matrix, 0, 64 * sizeof(float));

    for (i = 0; i < 64; i++) {
	if (! GetNumber(ctx, &matrix[ i ]))
	    break;
    }

    return(i > 0);
}

static void
Transform2DMatrix(XformContextPtr ctx)
{
    int i;
    float *src = ctx->srcMatrix;
    float *dst = ctx->dstMatrix;

    (*ctx->transform)(ctx);
    if (ctx->exact) {
	for (i = 0; i < 8; i++) {
	    if (ctx->showBoth) {
		NormalOutput("%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f", 
		    src[0], src[1], src[2], src[3],
		    src[4], src[5], src[6], src[7]);
		NormalOutput("  %s", (i == 0) ? "-->" : "   ");
	    }
	    NormalOutput("%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f\n", 
		dst[0], dst[1], dst[2], dst[3],
		dst[4], dst[5], dst[6], dst[7]);
	    src += 8;
	    dst += 8;
	}
    } else {
	for (i = 0; i < 8; i++) {
	    if (ctx->showBoth) {
		NormalOutput("%4d%4d%4d%4d%4d%4d%4d%4d", 
		    (int)(src[0]+0.5), (int)(src[1]+0.5),
		    (int)(src[2]+0.5), (int)(src[3]+0.5),
		    (int)(src[4]+0.5), (int)(src[5]+0.5),
		    (int)(src[6]+0.5), (int)(src[7]+0.5));
		NormalOutput("  %s", (i == 0) ? "-->" : "   ");
	    }
	    NormalOutput("%4d%4d%4d%4d%4d%4d%4d%4d\n", 
		(int)(dst[0]+0.5), (int)(dst[1]+0.5),
		(int)(dst[2]+0.5), (int)(dst[3]+0.5),
		(int)(dst[4]+0.5), (int)(dst[5]+0.5),
		(int)(dst[6]+0.5), (int)(dst[7]+0.5));
	    src += 8;
	    dst += 8;
	}
    }
}

main(int argc, char **argv)
{
    XformContext ctx;
    boolean forward = False;

    ProgramName = *argv;

    memset((char *)&ctx, 0, sizeof(ctx));

    ctx.fpIn = GetBufferedIn();
    ctx.fpOut = GetBufferedOut();
    for (argc--, argv++; argc; argc--, argv++) {
	switch (**argv) {
	case '-':
	    UsageMacro(("Unknown flag: \"%s\"\n", *argv));
	    break;

	default:
	    if (strcmp(*argv, "showboth") == 0) {
		ctx.showBoth = True;
	    } else if (strcmp(*argv, "exact") == 0) {
		ctx.exact = True;
	    } else if (strcmp(*argv, "1d") == 0) {
		ctx.oneDimension = True;
	    } else if (strcmp(*argv, "forward") == 0) {
		forward = True;
	    } else if (strcmp(*argv, "usage") == 0) {
		Usage();
	    } else {
		UsageMacro(("Unknown argument: \"%s\"\n", *argv));
	    }
	    break;
	}
    }

    InitDctReference(&ctx);
    if (ctx.oneDimension) {
	if (forward)
	    ctx.transform = ReferenceDct1d8x8;
	else
	    ctx.transform = ReferenceIDct1d8x8;
	while (Get1DMatrix(&ctx))
	    Transform1DMatrix(&ctx);
    } else {
	if (forward)
	    ctx.transform = ReferenceDct2d8x8;
	else
	    ctx.transform = ReferenceIDct2d8x8;
	while (Get2DMatrix(&ctx))
	    Transform2DMatrix(&ctx);
    }
    exit (0);
}
