/*
 * Interface routines for controlling the Abekas A60 videodisk.  These routines
 * are not involved with the production or communication of the video image;
 * it is assumed that the output of the video frame buffer is connected to
 * Abekas' input.
 *
 * Most of the parameters are set via a string passed to abekasOpen(str):
 *
 * v	   Activates the debug (verbose) output.
 * i<i>    Select input (1 or 2).
 * n<n>    Nbr of abekas frames to record per call to abekasRecord()
 * s<n>	   Set the starting frame number.
 * h<host> Specifies the Abekas' network hostname.
 *
 * The protocol for communicating with the A60 is to call
 *
 *	abekasFSend(fmt, arg, ... ); abekasReceive();
 *
 * which works like printf() to compose an outgoing command to the A60.  Each
 * command ends with a \n.  abekasReceive() waits for the prompt following
 * the completion of the command.  The following functions handle the most
 * common commands:
 *
 * abekasSeek(frame_number)
 * abekasRecord(0 or number_of_frames)		0==>Use 'r' parameter above
 * abekasInput(in)
 * abekasClose();
 */

static char rcsid[] = " $Header: /usr/local/src/nvrToMIT/RCS/abekas.c,v 1.4 1994/12/17 12:55:39 russ Exp $ ";

#include <stdio.h>
#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/telnet.h>

#define TimeoutSeconds	8
#define	False		0
#define True		1

char	abekasHost[50];		/* Abekas host name */
int	abekasFid;		/* Socket file descriptor for Abekas */
char	abekasSendOK = False;	/* OK to send */
char	abekasReceiveOK = False;/* OK to receive */
int	abekasNFrames;		/* Number of frames per call to abekasRecord */
int	abekasDebug;		/* Abekas debug */
int	abekasStart;		/* Abekas starting frame */
int	abekasIn;		/* selected A60 input */

extern	int	 errno;
extern	char	*sys_errlist[];

abekasFatalError(fmt, x0,x1,x2,x3,x4,x5,x6,x7,x8,x9)
    char *fmt;
{
    fflush(stdout);
    fprintf(stderr, "\n");
    if (errno)
	fprintf(stderr, "error %d %s\n", errno, sys_errlist[errno]);
    fprintf(stderr, fmt, x0,x1,x2,x3,x4,x5,x6,x7,x8,x9);
    abekasShutdown(0);
    close(abekasFid);
    abekasFid = -1;
    exit(1);
}

abekasShutdown(sig)
{
    if (sig)
	signal(sig, SIG_IGN);
    if (abekasSendOK) {
	abekasSend("enable\nexit\n");
    }
    if (sig) {
	close(abekasFid);
	abekasFid = -1;
	exit(0);
    }
}

abekasConnectionGone()
{
	signal(SIGPIPE, SIG_IGN);
	/*
	 * In some cases we want to recover....
	 *
	abekasFatalError("Connection to %s dropped\n", abekasHost);
	 */
	fprintf(stderr, "Connection to %s dropped\n", abekasHost);
}

abekasShowString(announce, string, length)
    char *announce;
    char *string;
    int length;
{
    if (!abekasDebug)
	return;
    fprintf(stderr, "%s \"", announce);
    while (length-- > 0) {
	if (*string >= ' ' && *string <= '~')
	    putc(*string, stderr);
	else {
	    switch (*string) {
	    case '\n': putc('\\', stderr); putc('n', stderr); break;
	    case '\r': putc('\\', stderr); putc('r', stderr); break;
	    case '\b': putc('\\', stderr); putc('b', stderr); break;
	    case '\0': putc('\\', stderr); putc('0', stderr); break;
	    default:   fprintf(stderr, "\\%03o", *string); break;
	    }
	}
	string++;
    }
    putc('"', stderr);
    putc('\n', stderr);
}

abekasFSend(fmt, x0, x1,x2,x3,x4,x5,x6,x7,x8,x9)
    char *fmt;
{
    char buf[ BUFSIZ ];

    sprintf(buf, fmt, x0, x1,x2,x3,x4,x5,x6,x7,x8,x9);
    abekasSend(buf);
}

abekasSend(msg)
    char *msg;
{
    int len = strlen(msg);
    int wrote;
    struct timeval timeout;
    int nFound;
    u_long writeFds;

    if (!abekasSendOK)
	return;

    timeout.tv_sec = TimeoutSeconds;
    timeout.tv_usec = 0;
    writeFds = 1<<abekasFid;

    nFound = select(abekasFid+1, NULL, &writeFds, NULL, &timeout);
    if (nFound <= 0) {
	abekasSendOK = False;
	if (errno == EINTR) {
	    if (abekasDebug)
		fprintf(stderr, "select on write interrupted\n");
	    return;
	}
	abekasFatalError("Cannot send message to %s after %d seconds: \"%s\"\n",
	    abekasHost, TimeoutSeconds, msg);
    }

    if ((wrote = write(abekasFid, msg, len)) != len) {
	abekasSendOK = False;
	if (errno == EINTR) {
	    if (abekasDebug)
	    	fprintf(stderr, "write interrupted\n");
	    return;
	}
	abekasFatalError("Only able to send %d out of %d bytes to %s: \"%s\"\n",
	    wrote, len, msg);
    }
    abekasShowString("sent", msg, len);
}

abekasReceive() {
    char buf[ BUFSIZ ];
    char *pPrompt;
    struct timeval timeout;
    int nFound;
    int red;
    char redSome = False;
    char gotPrompt = False;
    u_long readFds;

    if (!abekasReceiveOK)
	return;

    while (! gotPrompt) {
	timeout.tv_sec = TimeoutSeconds;
	timeout.tv_usec = 0;
	readFds = 1<<abekasFid;

	nFound = select(abekasFid+1, &readFds, NULL, NULL, &timeout);
	if (nFound <= 0) {
	    abekasReceiveOK = False;
	    if (errno == EINTR)
		break;
	    if (redSome)
		abekasFatalError("Select failure after reading some from %s\n",
		    abekasHost);
	    else
		abekasFatalError("Cannot read anything from %s after %d seconds\n",
		    abekasHost, TimeoutSeconds);
	}

	if ((red = read(abekasFid, buf, BUFSIZ)) < 0) {
	    abekasReceiveOK = False;
	    if (errno == EINTR)
		break;
	    abekasFatalError("Read failure from %s\n", abekasHost);
	}

	if (red == 0) {
	    abekasReceiveOK = False;
	    if (abekasDebug)
		fprintf(stderr, "got EOF\n");
	    break;
	}

	redSome = True;
	buf[ red ] = '\0';
        pPrompt = buf+red-2;
	if (pPrompt[0] == '>' && pPrompt[1] == ' ') {
	    gotPrompt = True;
	    abekasShowString("got prompt=", buf, red);
	} else
	    abekasShowString("got", buf, red);
    }
}

abekasOpen(str)
char *str;
{
    char *p, *host, *getenv();
    struct hostent *haddr;
    struct sockaddr_in sock;
    struct servent *sp;
    int uid;

    if (str == NULL) {
	close(abekasFid);
    } else {
	abekasDebug = 0;
	abekasNFrames = 1;
	abekasStart = 0;
	abekasIn = 1;
	if ((host = getenv("ABEKAS")) == NULL)
	    host = "a60";
	while (*str) switch (*str++) {

	case 'v':			/* Verbose */
	    abekasDebug = 1;
	    break;

	case 'n':			/* Number of frames to record */
	    abekasNFrames = strtol(str, &str, 10);
	    break;

	case 's':			/* Starting frame number */
	    abekasStart = strtol(str, &str, 10);
	    break;

	case 'i':			/* Input select; must be 1 or 2 */
	    abekasIn = strtol(str, &str, 10);
	    break;

	case 'h':			/* Abekas host name */
	    host = str;
	    str += strlen(str);
	    break;

	default:
	    fprintf(stderr, "abekasOpen: Unknown parameter: %s\n", str);
	    return(1);
	}
	strncpy(abekasHost, host, sizeof abekasHost);
    }
    /*
     * Obtain the service info for telnet.
     */
    if ((sp = getservbyname("telnet", "tcp")) == NULL) {
	fprintf(stderr, "Cannot find 'telnet' service\n");
	return(1);
    }
    /*
     * First, try interpreting the address as an internet standard "dot"
     * form.  If that fails, look up the address as if it were a host name.
     */
    if ((sock.sin_addr.s_addr = inet_addr(abekasHost)) != -1) {
	sock.sin_family = AF_INET;
    } else {
        if (haddr = gethostbyname(abekasHost)) {
            sock.sin_family = haddr->h_addrtype;
            memcpy((caddr_t)&sock.sin_addr, haddr->h_addr,
                   haddr->h_length);
        } else {
	    fprintf(stderr, "Unknown host: %s\n", abekasHost);
	    return(1);
	}
    }
    sock.sin_port = sp->s_port;
    /*
     * Create a socket to fiddle with.
     */
    if ((abekasFid = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	fprintf(stderr, "Cannot create socket\n");
	return(1);
    }
    /*
     * Try to connect to the address in question...
     */
    if (connect(abekasFid, (struct sockaddr *)&sock, sizeof(sock)) < 0) {
	fprintf(stderr, "Connect() error: %s\n", sys_errlist[errno]);
	return(1);
    }
    signal(SIGPIPE, abekasConnectionGone);
    signal(SIGINT, abekasShutdown);
    signal(SIGQUIT, abekasShutdown);
    abekasSendOK = True;
    abekasReceiveOK = True;
    abekasReceive();
    abekasFSend("disable\n");
    abekasReceive();
#ifdef FIELDMODE
    abekasFSend("mode field\n");
    abekasReceive();
#else
    abekasFSend("mode frame\n");
    abekasReceive();
#endif
    abekasSeek(abekasStart);
    abekasInput(abekasIn);
    return(0);
}


abekasRecord(nframe) {
#ifdef FIELDMODE
    abekasFSend("recrel %d\n", 2 * ((nframe > 0) ? nframe: abekasNFrames));
#else
    abekasFSend("recrel %d\n", (nframe > 0) ? nframe: abekasNFrames);
#endif
    abekasReceive();
}

abekasSeek(frame) {
    abekasFSend("goto %d\n", frame);
    abekasReceive();
}

abekasClose() {
    if (abekasSendOK)
	abekasSend("enable\nexit\n");
    close(abekasFid);
    abekasFid = -1;
}

abekasInput(in) {
    if (in != 1 && in != 2) {
	fprintf(stderr, "AbekasInput(%d): Must be 1 or 2\n", in);
	return;
    }
    abekasFSend("input %d\n", in);
    abekasReceive();
}
