/* $Header: /usr/local/src/nvrToMIT/RCS/vcopy.c,v 1.5 1994/12/17 12:55:42 russ Exp $ */
/*
 * vcopy [-b] [-r] [-c<count>] <input> [<output>]
 *
 * Vcopy copies <count> (defaults to infinity) frames from <input> to <output>.
 * If <output> is missing, it assumes that a monitor or unsupported video
 * device is receiving the output.  <input> can be one of the following:
 *
 * <filename>[:[s<start>][m<nrec>][b<bufsize>][p<pad>]]
 *	<filename> is a computer pathname used to read/write the video.  The
 *	modifiers are s<start> to skip <start> frames on input before copying,
 *      m<nrec> to use MPEG format with a buffer size of <nrec> MPEG records
 *	(3600 bytes each), b<bufsize> sets the buffer size to <bufsize>, and
 *	p<pad> pads and output sequence to multiple of <pad> bytes with 0.
 *
 * a60:[n<n>][s<s>][v][i<i>][h<h>]
 *	This source/sink is an Abekas A60 video disk.  The arguments after
 *	have the following meaning:
 *
 *   n<n>  Replicate the input frame <n> time.  The applies only when the A60
 *	   is the output and the input is provided 1 frame at a time.
 *
 *   s<s>  Specifies the starting frame number.
 *
 *   i<i>  Specifies the A60 input, 1 or 2.
 *
 *   h<h>  Specifies the A60's host name.  Must be last parameter.
 *
 *   v     Enables A60 debug output.
 *
 * vtr:[n<n>][s<s>][v][i][b][t][l<tty>]
 *	This source/sink specifies a video tape recorder controlled by a Lyon
 *	Lamb minivas VTR controller.  Currently, the VTR is fixed configured
 *	to be the Sony BVW-40.  The options are:
 *
 *    n<n>  Replicate the input frame <n> time; only applies to 1 frame at a
 *	    time recording.
 *
 *    s<s>  Specifies the starting frame.  NOTE: the minivas number frames from
 *	    1, not 0.
 *
 *    v     Activate debug output.
 *
 *    i     Initialize the minivas.
 *
 *    b	    Preblack the tape; that is write time codes and VIFC codes on
 *	    the tape, erasing all previous video/audio information.  This
 *	    command will cause the program to exit when done.
 *
 *    t     Position tape using SMPTE timecodes (as opposed to VIFC codes
 *	    which is the default).
 *
 *    l<tty> Specifies the tty special file connected to the minivas.
 *	    All chars after the 'l' are assumed to be part of the name.
 *
 * -b	Pre-blacks the tapes and record timecodes.
 *
 * -r	Rewinds tape before the copy.
 */

/*
 * Include file for compression/decompression software
 */

#include <sys/types.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <setjmp.h>
#include <errno.h>

#ifndef A60ONLY
#include "minivas.h"
#include <sys/io/videoout.h>
#endif

#define	ABUFSIZ	    (128*1024)	/* Abekas format read buffer size */

#define	DVISIZE		0x00300000	/* Size of DVI address space (bytes) */
#define DVISWREG	0x200301	/* Switch register */	
#define	DVIFBEN		0x10		/* frame buffer enable */
#define	DVIENABLE	0x01		/* Enable */
#define	DVILINE(l)	(((((l) & 1) ? 0 : D1HEIGHT/2) + ((l) >> 1)) * D1WIDTH)
#define DVIBLACK        ((D1CBIAS << 24) | (D1YMIN << 8))

#define	D1WIDTH		720	/* Total pixels per line in 422 D1 video */
#define	D1HEIGHT	486	/* Total lines in 422 D1 video per frame */

#define	D1FRAME		(2 * D1WIDTH * D1HEIGHT)

#define	D1YMIN		16	/* Minimum luminance */
#define	D1YMAX		235	/* Maximum luminance */
#define	D1CMIN		16	/* Minimum chromanance */
#define	D1CMAX		240	/* Maximum chromanance */
#define	D1YBIAS		125	/* Luminance offset to make 0 centered  range */
#define	D1CBIAS		128	/* Chomanance offset to make 0 centered range */

#define	MINIVASERRORMAX	 3	/* Maximum number of Minivas record retries */
#define	ABEKAS_MAX_FRAME 750	/* Nbr of frame on 25sec A60 */
#define	RCPTIMEOUT	 30	/* rcp timeout (seconds) */
#define PREROLLTIME	 (5*30)/* Pre-roll time in frames */

#define	A60_INPUT_DVI	 1	/* A60 input from DVI */
#define	A60_INPUT_VTR	 2	/* A60 input from VTR */

/*
 * AVEERROR is the average number of pixel units of error allowable between
 * fields which are supposed to be identical in undo'ing 2-3 pulldown.
 */

#define	AVEERROR	0.1

u_char	 fb1[2*D1WIDTH*D1HEIGHT];	/* Input frame buffer */
u_char	 fb2[2*D1WIDTH*D1HEIGHT];	/* Output frame buffer */
u_char	 fieldbuf[D1WIDTH*D1HEIGHT];	/* Tmp for 2-3 pulldown */
int	 dvimap[D1HEIGHT];	/* Offset to various DVI fb lines */
int	*dvi_fb;		/* Digital video interface YUV frame buffer */
int	 minivas;		/* VTR control thru Lyon-Lamb Minivas */
int	 abekas;		/* Abekas video disk open */
int	 count = 0x70000000;	/* Frame count (initially infinite) */
int	 ftpinprogress;		/* FTP to A60 is in progress */
int	 state23 = -1;		/* 2-3 pulldown state */
int	 video_to_movie;	/* Video to movie frames (undo 2-3 pulldown) */
int	 movie_to_video;	/* Movie to video (implement 2-3 pulldown) */
int	 inframe;		/* input frame counter */
int	 nodvi;			/* Do not use DVI framebuffer */
int	 offset;		/* Offset from VIFC to timecode on VTR */

struct	fileinfo {
	char	*f_name;	/* UNIX file name */
	char	 f_fid;		/* UNIX file descriptor */
	char	 f_format;	/* File format code (F_* constants) */
	u_char	*f_buf;		/* I/O buffer base address */
	u_char	*f_ptr;		/* Current buffer pointer */
	int	 f_bufsize;	/* Buffer size */
	int	 f_nbuf;	/* Number of bytes in buffer */
	int	 f_pad;		/* Zero pad output to a multiple of .f_pad */
} fin, fout;

#define	F_ABEKAS	1	/* Abekas Frame format */
#define	F_MPEG		2	/* MPEG format */

jmp_buf	 unwind;		/* Timeout return to top level */

extern	int	 errno;		/* Syscall error code */
extern	char	*sys_errlist[];	/* Syscall error strings */

#define	fatal(p)	{ fprintf p ; quit(1); }

#ifndef A60ONLY
char	*usage[] = {
"Usage: vcopy [-v<s>] [-m<s>] [-c<count>] <input> [<output>]",
    "",
    "The input/output identifiers are one of:",
    "",
    "<filename>[:[s<start>][b<bufsize>][m<nrec>][p<padsize>]]",
    "a60:[n<n>][s<s>][v][i<i>][h<h>]",
    NULL
};
#else
char	*usage[] = {
"Usage: vcopy [-b] [-r] [-v<s>] [-m<s>] [-c<count>] <input> [<output>]",
    "",
    "The input/output identifiers are one of:",
    "",
    "<filename>[:[s<start>][b<bufsize>][m<nrec>][p<padsize>]]",
    "a60:[n<n>][s<s>][v][i<i>][h<h>]",
    "vtr:[n<n>][s<s>][v][i][t][l<tty>]",
    NULL
};
#endif

#define MRECSIZE	3600	/* MPEG CCIR 601 record size */

/*
 * Source / sink codes
 */

#define	T_NULL		0	/* Unspecified output */
#define	T_FILE		1	/* Unix file or pipe */
#define	T_ABEKAS	2	/* Abekas A60 */
#define	T_VTR		3	/* VTR controlled by Lyon-Lamb minivas */

char	*types[] = {
    "<nothing>",  "<file>", "<A60>", "<VTR>",
};

struct	vtrtab {
	char	*vt_name;	/* VTR name on command line */
	char	*vt_mininame;	/* name of VTR in minivas driver tables */
	char	 vt_recdelay;	/* Frame delay when recording from A60 trigger */
	char	 vt_playdelay;	/* Frame delay when playing via A60 trigger */
} *vt, vtrtab[] = {
	{ "vtr:", NULL, 2, 2 },
	{ "bvw40:", "SonyBVW40", 2, 2 },
	{ "ag7500:", "PanAG7500", 2, 2 },
	{ NULL },

};

char	*malloc();
char	*strchr();

/*
 * Main program:  parse parameters and call other to do the real work
 */

main(argc, argv)
char **argv;
{
    int startframe, itype, otype, n, x, *q, rewind = 0;
    char *p;

    if (argc < 2) {
usg:
	for (argv = usage; *argv; argv++)
	    fprintf(stderr, "%s\n", *argv);
	quit(1);
    }
    while (((p = *++argv) != NULL) && (*p++ == '-')) {
	switch (*p++) {

#ifndef A60ONLY
	case 'b':		/* Pre black the tape */
	    fprintf(stderr, "CAUTION:  Input 1 must be selected on the A60\n");
	    sleep(5);
	    open_dvi();
	    for (n = 0; n < D1HEIGHT; n++) {
		q = &dvi_fb[dvimap[n]];
		for (x = 0; x < D1WIDTH; x++)
		    *q++ = DVIBLACK;
	    }
	    p[-2] = 'i';
	    quit(minivasOpen(p - 2));

	case 'r':		/* rewind VTR */
	    rewind = 1;
	    continue;
#endif

	case 'c':		/* Specify a frame count */
	    if ((*p == 0) && ((p = *++argv) == NULL))
		break;
	    count = atoi(p);
	    continue;
	
	case 'm':		/* Convert Video frame to Movie frames */
	    video_to_movie = 1;
	    movie_to_video = 0;
	    state23 = atoi(p) % 5;
	    continue;

	case 'v':		/* Convert Movie frames to Video frames */
	    movie_to_video = 1;
	    video_to_movie = 0;
	    state23 = atoi(p) & 3;;
	    continue;
	
	case 'd':		/* No DVI framebuffer available */
	    nodvi = 1;
	    continue;

	case 'o':		/* Offset from VIFC to timecode */
	    offset = atoi(p);
	    continue;

	case ':':		/* assume part of input spec */
	case 0:
	    p--;
	    goto input;

	default:
	    fatal((stderr, "Unknown parameter: %s\n", argv[0]));
	}
	fatal((stderr, "Argument expected after %s\n", argv[-1]));
    }
    if (p == NULL) {
	if (rewind == 0)
	    goto usg;
#ifndef A60ONLY
	if (minivasOpen("-"))
	    quit(1);
	minivasSendChar(MINIVAS_REWIND);
#endif
	quit(0);
    }
input:
    p--;
    if (!strcmp(p, ".")) {
	itype = T_NULL;
    } else if (!strncmp(p, "a60:", 4)) {
	itype = T_ABEKAS;
	if (abekasOpen(p + 4))
	    quit(1);
	abekas = 1;
#ifndef A60ONLY
    } else if (openvtr(p, rewind, "-")) {
	itype = T_VTR;
#endif
    } else {
	itype = T_FILE;
	fin.f_name = p;
	fin.f_format = F_ABEKAS;
	fin.f_bufsize = ABUFSIZ;
	startframe = 0;
	if ((p = strchr(p, ':')) != NULL) {
	    *p++ = 0;
	    while (*p) {
	        switch (*p++) {

	        case 's':
		   startframe = strtol(p, &p, 10);
		   continue;

		case 'b':
		    fin.f_bufsize = strtol(p, &p, 10);
		    continue;

		case 'm':
		    fin.f_format = F_MPEG;
		    if ((fin.f_bufsize = strtol(p, &p, 10)) == 0)
			fin.f_bufsize = 1;
		    fin.f_bufsize *= MRECSIZE;
		    continue;
		}
		fatal((stderr, "Unknown input flag '%c' following %s\n",
		    p[-1], fin.f_name));
	    }
	}
	if (strcmp(fin.f_name, "-")) {
	    if ((fin.f_fid = open(fin.f_name, 0)) < 0)
		fatal((stderr, "Cannot open %s: %s\n",
		    fin.f_name, sys_errlist[errno]));
	} else {
	    fin.f_fid = 0;
	    fin.f_name = "<stdin>";
	}
	if ((fin.f_buf = (u_char *)malloc(fin.f_bufsize)) == NULL)
	    fatal((stderr, "Cannot malloc input buffer (%d bytes)\n",
		fin.f_bufsize));
	fin.f_nbuf = 0;
	if (fin.f_format == F_MPEG)
	    tread(&fin, fb1, MRECSIZE);
	if (startframe < 0) {
	    startframe = -startframe;
	    for (inframe = 0; --startframe >= 0; inframe++)
		getframe(&fin, fb1);
	} else {
	    if (startframe && (lseek(fin.f_fid, startframe * D1FRAME, 0) == -1))
		fatal((stderr, "Lseek error on %s: %s\n",
		    fin.f_name, sys_errlist[errno]));
	}
    }
    if ((p = *++argv) == NULL) {
	otype = T_NULL;
    } else if (!strncmp(p, "a60:", 4)) {
	otype = T_ABEKAS;
	if (!abekas && abekasOpen(p + 4))
	    quit(1);
	abekas = 1;
#ifndef A60ONLY
    } else if (openvtr(p, rewind, "")) {
	otype = T_VTR;
#endif
    } else {
	otype = T_FILE;
	fout.f_name = p;
	fout.f_format = F_ABEKAS;
	fout.f_bufsize = ABUFSIZ;
	startframe = 0;
	if ((p = strchr(p, ':')) != NULL) {
	    *p++ = 0;
	    while (*p) {
	        switch (*p++) {

	        case 's':
		   startframe = strtol(p, &p, 10);
		   continue;

		case 'p':
		    if ((fout.f_pad = strtol(p, &p, 10)) == 0)
			fout.f_pad = ABUFSIZ;
		    continue;
		
		case 'b':
		    fout.f_bufsize = strtol(p, &p, 10);
		    continue;

		case 'm':
		    fout.f_format = F_MPEG;
		    if ((fout.f_bufsize = strtol(p, &p, 10)) == 0)
			fout.f_bufsize = 1;
		    fout.f_bufsize *= MRECSIZE;
		    continue;
		}
		fatal((stderr, "Unknown input flag '%c' following %s\n",
		    p[-1], fout.f_name));
	    }
	}
	if (strcmp(fout.f_name, "-")) {
	    if ((fout.f_fid = creat(fout.f_name, 0666)) < 0)
		fatal((stderr, "Cannot open %s: %s\n",
		    fout.f_name, sys_errlist[errno]));
	} else {
	    fout.f_fid = 0;
	    fout.f_name = "<stdout>";
	}
	if (fout.f_pad)
	    fout.f_bufsize = ((fout.f_bufsize + fout.f_pad - 1) /
		fout.f_pad) * fout.f_pad;
	if ((fout.f_buf = (u_char *)malloc(fout.f_bufsize)) == NULL)
	    fatal((stderr, "Cannot malloc output buffer (%d bytes)\n",
		fout.f_bufsize));
	fout.f_ptr = fout.f_buf;
	if (fout.f_format == F_MPEG) {
            memset(fout.f_ptr, 0, MRECSIZE);
	    fout.f_ptr += MRECSIZE;
	}
	if (startframe < 0) {
	    startframe = -startframe;
	    for (inframe = 0; --startframe >= 0; inframe++)
		getframe(&fout, fb1);
	} else {
	    if (startframe && (lseek(fout.f_fid, startframe * D1FRAME, 0) == -1))
		fatal((stderr, "Lseek error on %s: %s\n",
		    fout.f_name, sys_errlist[errno]));
	}
    }
    if ((movie_to_video || video_to_movie) &&
	(itype != T_FILE) && (otype != T_FILE))
	fatal((stderr,
	    "Cannot do 2-3 pulldown unless input or output is a file\n"));
    /*
     * Find the illegal combinations
     */
    switch (8*itype + otype) {

    case 8*T_FILE+T_ABEKAS:
	abekasInput(A60_INPUT_DVI);
#ifndef A60ONLY
	if (nodvi)
#endif
	{
	    while ((--count >= 0) && getframe(&fin, p = (char *)fb1)) {
		/*
		 * Do/undo 2-3 pulldown from movie format
		 */
		if (video_to_movie && !undo23pulldown(fb1, p = (char *)fb2))
		    continue;
		if (movie_to_video) {
		    if (do23pulldown(fb1, p = (char *)fb2)) {
			abekas_put(p);
			p = (char *)fb1;
		    }
		}
		abekas_put(p);
		inframe++;
	    }
	    alarm(0);
	    break;
	}
	/*FALLINTO*/

#ifndef A60ONLY
    case 8*T_FILE+T_NULL:
    case 8*T_FILE+T_VTR:
	if (nodvi)
	    fatal((stderr, "DVI framebuffer required\n"));
	open_dvi();
	while ((--count >= 0) && getframe(&fin, p = (char *)fb1)) {
	    /*
	     * Do/undo 2-3 pulldown from movie format
	     */
	    if (video_to_movie && !undo23pulldown(fb1, p = (char *)fb2))
		continue;
	    if (movie_to_video) {
		if (do23pulldown(fb1, p = (char *)fb2)) {
		    display(p);
		    p = (char *)fb1;
		}
	    }
	    display(p);
	    inframe++;
	}
	break;

    case 8*T_ABEKAS+T_VTR:
	abekas_vtr();
	break;

    case 8*T_VTR+T_ABEKAS:
	vtr_abekas();
	break;

    case 8*T_NULL+T_VTR:
	vtr_record();
#endif
	break;

    case 8*T_FILE+T_FILE:
	while ((--count >= 0) && getframe(&fin, p = (char *)fb1)) {
	    /*
	     * Do/undo 2-3 pulldown from movie format
	     */
	    if (video_to_movie && !undo23pulldown(fb1, p = (char *)fb2))
		continue;
	    if (movie_to_video) {
		if (do23pulldown(fb1, p = (char *)fb2)) {
		    putframe(p);
		    p = (char *)fb1;
		}
	    }
	    putframe(&fout, p);
	    inframe++;
	}
	putframe(&fout, NULL);
	break;
	    
    case 8*T_ABEKAS+T_NULL:
	abekas_play();
	break;

    case 8*T_ABEKAS+T_FILE:
	abekas_file();
	break;

    default:
	fatal((stderr, "Copy from %s to %s is not implemented\n",
	    types[itype], types[otype]));
    }
    quit(0);
}

#ifndef A60ONLY
/*
 * Openvtr() tests the input string for a valid VTR name.  If one is
 * found, the minivas is opened and initialized.  The global variable
 * vt points to the vtrtab table entry on successful exit.
 */

openvtr(str, rewind, arg)
char *str, *arg;
{
    char *p, *q, minivasopen[40];
    int len;

    for (vt = vtrtab; vt->vt_name; vt++) {
	if (strncmp(str, vt->vt_name, len = strlen(vt->vt_name)))
	    continue;
	str += len;
	p = minivasopen;
	if (rewind)
	    *p++ = 'r';
	while (*arg)
	    *p++ = *arg++;
	while (*str)
	    *p++ = *str++;
	if (q = vt->vt_mininame) {
	    *p++ = 'd';
	    while (*q)
		*p++ = *q++;
	}
	*p = 0;
	if (minivasOpen(minivasopen))
	    quit(1);
	minivas = 1;
	return(1);
    }
    return(0);
}
#endif

/*
 * Write a frame to the A60 via the FTP protocol (No RP128 (D1) frame buffer
 * is available).
 */

abekas_put(fbp)
u_char *fbp;
{
    extern int abekasStart, abekasFtpCmd, abekasFtpData;
    extern int abekasFtpDebug, abekasDebug;
    extern char abekasFtpPut[], abekasHost[];
    static int pid;
    char name[20];
    int afid;
    register i, n, m;
    register u_char *p;

    if (setjmp(unwind)) {
again:
	fprintf(stderr, "Reset A60, then type 'kill -CONT %d' to resume\n", pid);
	kill(pid, SIGTSTP);
	fprintf(stderr, "Continuing ...\n");
	/*
	 * Resume form an aborted transfer; A60 is presumably reset
	 */
	close(abekasFtpCmd);
	if (abekasFtpData >= 0)
	    close(abekasFtpData);
	abekasFtpCmd = -1;
	abekasFtpData = -1;
	if (abekasOpen(NULL))		/* To disable A60 keyboard only */
	    quit(1);
    }
    if (abekasFtpCmd < 0) {
	abekasFtpDebug = abekasDebug;
	pid = getpid();
	if (abekasFtpOpen(abekasHost))
	    goto again;
	ftptimer(0);
    }
    sprintf(name, "%d.yuv", abekasStart);
    ftpinprogress = 1;
    if ((afid = abekasXferSetup(abekasFtpPut, name)) < 0)
	goto again;
    for (n = D1FRAME, p = fbp; n > 0; n -= m, p += m) {
	if ((m = write(afid, p, n)) < 0) {
	    if (errno == EINTR) {
		m = 0;
		continue;
	    }
	    fatal((stderr, "Write error %s data socket: %s\n",
		abekasHost, sys_errlist[errno]));
	}
	if (m == 0)
	    fatal((stderr, "Zero byte write to %s in frame %d\n",
		abekasHost, abekasStart));
	if (abekasFtpDebug)
	    fprintf(stderr, "Write %d bytes\n", m);
    }
    if (abekasFtpDone(0))
	goto again;
    ftpinprogress = 0;
    abekasStart++;
}

#ifndef A60ONLY

/*
 * Open DVI frame buffer
 */

open_dvi() {
    char *p;
    int n, key, physaddr, space;
    static char dev[] = "/dev/videoout-0";

    if ((n = open(dev, O_RDONLY, 0)) < 0)
	fatal((stderr, "Cannot open %s: %s\n",
	    dev, sys_errlist[errno]));
    if (ioctl(n, VIDEOOUTADDR, &physaddr) < 0)
	fatal((stderr, "VIDEOOUTADDR ioctl error: %s\n",
	    sys_errlist[errno]));
    if (ioctl(n, VIDEOOUTSPACE, &space) < 0)
	fatal((stderr, "VIDEOOUTSPACE ioctl error: %s\n",
	    sys_errlist[errno]));
    close(n);
    key = ftok(dev, 'P');
    if ((n = shmget(key, DVISIZE, IPC_CREAT | IPC_CI | IPC_NOCLEAR |
	IPC_PHYS | 0666, physaddr, space)) < 0)
	fatal((stderr, "shmget(%d, 0x%x, 0x%x, 0x%x, 0x%x) error: %s\n",
	    key, DVISIZE, IPC_CREAT | IPC_CI | IPC_NOCLEAR |
	    IPC_PHYS | 0666, physaddr, space));
    if ((int)(dvi_fb = (int *)shmat(n, 0, 0)) == -1)
	fatal((stderr, "shmat(%d, 0, 0) error: %s\n",
	    n, sys_errlist[errno]));
    if (shmctl(n, IPC_RMID, 0) == -1)
	fatal((stderr, "shmctl(%d, IPC_RMID, 0) error: %s\n",
	    n, sys_errlist[errno]));
    p = (char *)dvi_fb;
    p[DVISWREG] |= DVIENABLE;
    if ((p[DVISWREG] & DVIFBEN) == 0) {
	p[DVISWREG] |= DVIFBEN;
	sleep(1);
    }
    for (n = 0; n < D1HEIGHT; n++)
	dvimap[n] = DVILINE(n);
}

/*
 * Write a D1 video line on video output board, opt-9 (NTSC):
 *
 *  line array	    line array
 *  ---- -----      ---- -----
 *   0    243	    482   484
 *   1     0	    483   241
 *   2    244	    484   485
 *   3     1	    485   242
 */

display(fbp)
u_char *fbp;
{
    register u_char *q;
    register long *p;
    register int x, y;
    int n, i;

    if (minivas) {
	for (i = 0; n = minivasPollStatus(); i++) {
	    if (i >= MINIVASERRORMAX)
		fatal((stderr,
		    "minivas retry count (%d) exceeded\n", i));
	    minivasInit(n, 0);
	    if (minivasPollStatus())
		fatal((stderr,
		    "minivasPollStatus() error after minivasInit()\n"));
	    minivasRecord();
	}
    }
    q = fbp;
    for (y = 0; y < D1HEIGHT; y++) {
	p = &dvi_fb[dvimap[y]];
	for (x = 0; x < D1WIDTH; x++, q += 2)
	    *p++ = (q[0] << 24) | (q[1] << 8);
    }
    if (minivas)
	minivasRecord();
    if (abekas)
	abekasRecord(0);
}
#endif /* A60ONLY */

/*
 * Getframe() read a frame of D1 8-bit video from file.  This procedure will
 * read Abekas format or MPEG format.  Zero is returned for frame aligned EOF.
 */

getframe(f, fbp)
register struct fileinfo *f;
u_char *fbp;
{
    static int nbuf;
    register u_char *p, *y0, *y1, *cr0, *cr1, *cb0, *cb1;
    register int n, m;
    u_char *plast, inbuf[198 * MRECSIZE];

    switch (f->f_format) {

    case F_ABEKAS:			/* Abekas format */
	for (p = fbp, n = D1FRAME; n > 0; n -= m, p += m) {
	    if (f->f_nbuf <= 0) {
		if ((f->f_nbuf = read(f->f_fid, f->f_buf, f->f_bufsize)) < 0)
		    fatal((stderr, "Read error on %s: %s\n",
			f->f_name, sys_errlist[errno]));
		if (f->f_nbuf == 0) {
		    if (p != fbp)
			fatal((stderr, "Unexpected EOF on %s\n", f->f_name));
		    return(0);
		}
		f->f_ptr = f->f_buf;
	    }
	    if ((m = f->f_nbuf) > n)
		m = n;
            memcpy(p, f->f_ptr, m);
	    f->f_ptr += m;
	    f->f_nbuf -= m;
	}
	break;
    
    case F_MPEG:
	/*
	 * Read MPEG format and convert to Abekas format in memory
	 */
	if (!tread(f, inbuf, 198 * MRECSIZE))
	    return(0);
	y0  = inbuf;
	cb0 = &inbuf[49  * MRECSIZE];
	cr0 = &inbuf[74  * MRECSIZE];
	y1  = &inbuf[99  * MRECSIZE];
	cb1 = &inbuf[148 * MRECSIZE];
	cr1 = &inbuf[173 * MRECSIZE];
	p   = fbp;
	for (n = 0; n < D1HEIGHT/2; n++) {
	    /*
	     * Even field line (2nd field chronologically)
	     */
	    for (plast = p + 2*D1WIDTH; p < plast; p += 4) {
		p[0] = *cb1++;
		p[1] = *y1++;
		p[2] = *cr1++;
		p[3] = *y1++;
	    }
	    /*
	     * Odd field line (1st field chronologically)
	     */
	    for (plast = p + 2*D1WIDTH; p < plast; p += 4) {
		p[0] = *cb0++;
		p[1] = *y0++;
		p[2] = *cr0++;
		p[3] = *y0++;
	    }
	}
    }
    return(1);
}

/*
 * tread() read a block of input (presumably from a tape) and returns
 * contiguous chunks of it.
 */

tread(f, rec, len)
register struct fileinfo *f;
char *rec;
{
    int first = 1;
    register int n;

    while (len > 0) {
	if (f->f_nbuf <= 0) {
	    if ((f->f_nbuf = read(f->f_fid, f->f_buf, f->f_bufsize)) < 0) {
		fprintf(stderr, "Read error %s: %s\n",
		    f->f_name, sys_errlist[errno]);
		exit(1);
	    }
	    if (f->f_nbuf == 0) {
		if (first)
		    return(0);
		fprintf(stderr, "Unexpected EOF on %s\n", f->f_name);
		exit(1);
	    }
	    f->f_ptr = f->f_buf;
	}
	if ((n = len) > f->f_nbuf)
	    n = f->f_nbuf;
        memcpy(rec, f->f_ptr, n);
	f->f_ptr += n;
	f->f_nbuf -= n;
	rec += n;
	len -= n;
	first = 0;
    }
    return(1);
}

/*
 * flushbuf() empties the output buffer
 */

u_char *
flushbuf(last, f)
u_char *last;
register struct fileinfo *f;
{
    register u_char *p;
    register int n, m;

    for (p = f->f_buf; p < last; p += n) {
	if ((n = write(f->f_fid, p, last - p)) < 0) {
	    if (errno == EINTR) {
		n = 0;
		continue;
	    }
	    fatal((stderr, "Write error on %s: %s\n", f->f_name,
		sys_errlist[errno]));
	}
	if (n == 0)
	    fatal((stderr, "Zero-byte write on %s\n", f->f_name));
    }
    return(f->f_ptr = f->f_buf);
}

/*
 * putframe() outputs a frame on the specified stream.  If the frame pointer
 * is NULL, the output buffer is flushed.
 */

putframe(f, fbp)
register struct fileinfo *f;
u_char *fbp;
{
    register u_char *last;
    register int n, m;

    if (fbp == NULL) {
	if (f->f_pad) {
	    n = ((f->f_ptr - f->f_buf + f->f_pad - 1) / f->f_pad) * f->f_pad;
            memset(f->f_ptr, 0, &f->f_buf[n] - f->f_ptr);
	    f->f_ptr = &f->f_buf[f->f_bufsize];
	}
	flushbuf(f->f_ptr, f);
	return;
    }
    switch (f->f_format) {

    case F_MPEG:			/* Write MPEG format video */
	write_mpeg(fbp + 2 * D1WIDTH, f);
	write_mpeg(fbp, f);
	break;

    case F_ABEKAS:			/* Abekas frame format video */
	last = &f->f_buf[f->f_bufsize];
	for (n = D1FRAME; n > 0; n -= m, fbp += m) {
	    if (f->f_ptr >= last)
		(void) flushbuf(last, f);
	    if ((m = last - f->f_ptr) > n)
		m = n;
            memcpy(f->f_ptr, fbp, m);
	    f->f_ptr += m;
	}
	break;
    }
}

/*
 * write_mpeg() encodes a field in 'mpeg' format.
 */

write_mpeg(field, f)
u_char *field;
struct fileinfo *f;
{
    register u_char *a, *b, *blast, *alast;
    register int i;

    blast = f->f_buf + f->f_bufsize;
    b = f->f_ptr;
    a = field + 1;
    for (i = 0; i < D1HEIGHT/2; i++) {
	if (b >= blast)
	    b = flushbuf(blast, f);
	for (alast = a + 2*D1WIDTH; a < alast; a += 2)
	    *b++ = *a;
	a += 2*D1WIDTH;
    }
    memset(b, 0, 2*D1WIDTH);
    b += 2*D1WIDTH;
    a = field;
    for (i = 0; i < D1HEIGHT/2; i++) {
	if (b >= blast)
	    b = flushbuf(blast, f);
	for (alast = a + 2*D1WIDTH; a < alast; a += 4)
	    *b++ = *a;
	a += 2*D1WIDTH;
    }
    memset(b, 0,  7*D1WIDTH/2);
    b += 7*D1WIDTH/2;
    a = field + 2;
    for (i = 0; i < D1HEIGHT/2; i++) {
	if (b >= blast)
	    b = flushbuf(blast, f);
	for (alast = a + 2*D1WIDTH; a < alast; a += 4)
	    *b++ = *a;
	a += 2*D1WIDTH;
    }
    memset(b, 0,  7*D1WIDTH/2);
    f->f_ptr = b + 7*D1WIDTH/2;
}

#ifndef A60ONLY
/*
 * abekas_vtr() implements a real time video transfer from the A60 to
 * the VTR synchronized by time codes
 */

abekas_vtr() {
    int max, hour, min, sec, frame;
    extern minivas_num_frames, minivas_start_frame, abekasStart;

    frame_tc(offset + minivas_start_frame - vt->vt_playdelay,
	&hour, &min, &sec, &frame);
    if ((max = abekasStart + count) > ABEKAS_MAX_FRAME) {
	max = ABEKAS_MAX_FRAME;
	count = ABEKAS_MAX_FRAME - abekasStart;
    }
    if (count != minivas_num_frames) {
	minivasSendChar(MINIVAS_FRAME_CHANGE);
	minivasSendNum(count);
	minivasSendChar(MINIVAS_ENTER);
    }
    if (minivasPollStatus())
	fatal((stderr, "Minivas tape cueing error\n"));
    abekasFSend("enable\n");
    abekasReceive();
    abekasFSend("trigger %02d:%02d:%02d play %d %d\n",
	min, sec, frame, abekasStart, max);
    abekasReceive();
    minivasRecord();
    if (minivasPollStatus())
	fatal((stderr, "Minivas record error\n"));
};

/*
 * vtr_abekas() transfer the selected frames from the VTR to the A60 using
 * time codes to synchronize the transfer.
 */

vtr_abekas() {
    int max, hour, min, sec, frame, frame0;
    extern minivasFGrab, minivas_start_frame, minivasTC, abekasIn, abekasStart;

    if ((frame0 = minivas_start_frame - PREROLLTIME) < 0) {
	minivasSendCmd(MINIVAS_REWIND, NULL);
	/*
	 * Should wait for completion here?
	 */
    } else {
	if (minivasTC) {
	    minivasSendCmd(MINIVAS_SEARCH_TCODE, NULL);
	    minivasSendTC(frame0);
	} else {
	    minivasSendCmd(MINIVAS_SEARCH_FRAME, NULL);
	    minivasSendNum(frame0);
	}
	if (minivasSendCmd(MINIVAS_ENTER, MINIVAS_A_F))
	    fatal((stderr, "No acknowlege after search\n"));
	minivasSendChar(MINIVAS_STOP);
    }
    /*
     * Program the A60 to record starting at the specified time code
     */
    if ((max = abekasStart + count) > ABEKAS_MAX_FRAME) {
	count = ABEKAS_MAX_FRAME - abekasStart;
	max = ABEKAS_MAX_FRAME;
    }
    frame_tc(offset + minivas_start_frame - vt->vt_recdelay,
	&hour, &min, &sec, &frame);
    abekasFSend("enable\n");
    abekasReceive();
    abekasFSend("trigger %02d:%02d:%02d record input %d %d %d\n",
	min, sec, frame, A60_INPUT_VTR, abekasStart, max);
    abekasReceive();
    /*
     * Ask VTR to play thru the desired number of frames
     */
    frame = minivas_start_frame + count + PREROLLTIME;
    if (minivasTC) {
	minivasSendChar(MINIVAS_PLAY_TCODES);
	minivasSendTC((frame0 < 0) ? (24*60*60*30 + frame0) : frame0);
	minivasSendChar(MINIVAS_ENTER);
	minivasSendTC(frame);
    } else {
	minivasSendChar(MINIVAS_PLAY_FRAMES);
	minivasSendNum((frame0 < 0) ? 1 : frame0);
	minivasSendChar(MINIVAS_ENTER);
	minivasSendNum(frame);
    }
    minivasSendChar(MINIVAS_ENTER);
    if (minivasSendCmd(MINIVAS_EE, MINIVAS_A_Y))
	fatal((stderr, "No end of play sequence acknowledge\n"));
}
#endif

/*
 * ftptimer() polls the progress of the FTP operations from the A60.  If
 * it is stuck, the program suspends itself such that it can be restarted
 * after the A60 is reset
 */

ftptimer(sig) {
    extern int abekasFtpData;
    static lastcount;
    int p;

    if (sig == 0)
	lastcount = -1;
    signal(SIGALRM, ftptimer);
    if ((count != lastcount) || (ftpinprogress == 0)) {
	lastcount = count;
	alarm(30);
	return;
    }
    longjmp(unwind, 1);
}

/*
 * abekas_file() copies frames from the A60 to a file using the builtin
 * FTP client to retrieve frames from the A60.
 */

abekas_file() {
    extern int abekasStart, abekasFtpCmd, abekasFtpData;
    extern int abekasFtpDebug, abekasDebug, abekasFtpState;
    extern char abekasFtpGet[], abekasHost[];
    int pid, afid, max, i, cmax, n, m;
    u_char *p;

    if ((max = abekasStart + count) > ABEKAS_MAX_FRAME) {
	max = ABEKAS_MAX_FRAME - 1;
	count = ABEKAS_MAX_FRAME - abekasStart;
    }
    pid = getpid();
    cmax = count;
    abekasFtpDebug = abekasDebug;
    if (setjmp(unwind)) {
again:
	fprintf(stderr,
	    "Reset A60, then type 'kill -CONT %d' to resume(Ftp state %d)\n",
	    pid, abekasFtpState);
	kill(pid, SIGTSTP);
	fprintf(stderr, "Continuing ...\n");
	/*
	 * Resume form an aborted transfer; A60 is presumably reset
	 */
	close(abekasFtpCmd);
	if (abekasFtpData >= 0)
	    close(abekasFtpData);
	abekasFtpCmd = -1;
	abekasFtpData = -1;
	if (abekasOpen(NULL))		/* To disable A60 keyboard only */
	    quit(1);
    }
    if (abekasFtpOpen(abekasHost))
	goto again;
    ftptimer(0);
    do {
	sprintf(fb1, "%d.yuv", abekasStart);
	ftpinprogress = 1;
	if ((afid = abekasXferSetup(abekasFtpGet, fb1)) < 0)
	    goto again;
	for (n = D1FRAME, p = fb1; n > 0; n -= m, p += m) {
abekasFtpState = 17;
	    if ((m = read(afid, p, n)) < 0) {
abekasFtpState = 16;
		if (errno == EINTR) {
		    m = 0;
		    continue;
		}
		fatal((stderr, "Read error %s data socket: %s\n",
		    abekasHost, sys_errlist[errno]));
	    }
abekasFtpState = 16;
	    if (m == 0)
		fatal((stderr, "Unexpected EOF from %s in frame %d\n",
		    abekasHost, abekasStart));
	    if (abekasFtpDebug)
		fprintf(stderr, "Read %d bytes\n", m);
	}
	if (abekasFtpDone(0))
	    goto again;
	ftpinprogress = 0;
	abekasStart++;
	inframe++;
	p = fb1;
	if (video_to_movie && !undo23pulldown(fb1, p = fb2))
	    continue;
	if (movie_to_video) {
	    if (do23pulldown(fb1, p = fb2)) {
		putframe(&fout, p);
		p = fb1;
	    }
	}
	putframe(&fout, p);
    } while (--count > 0);
    alarm(0);
    putframe(&fout, NULL);
}

/*
 * abekas_play() plays the specified sequence of frames frome the A60.
 */

abekas_play() {
    int max;

    if ((max = abekasStart + count) > ABEKAS_MAX_FRAME) {
	max = ABEKAS_MAX_FRAME;
	count = ABEKAS_MAX_FRAME - abekasStart;
    }
    abekasFSend("play\n");
    abekasReceive();
}

#ifndef A60ONLY
/*
 * vtr_record() records the specified number of frames at the specified
 * position without specifying the video source.
 */

vtr_record() {
    if (count != minivas_num_frames) {
	minivasSendChar(MINIVAS_FRAME_CHANGE);
	minivasSendNum(count);
	minivasSendChar(MINIVAS_ENTER);
    }
    if (minivasPollStatus())
	fatal((stderr, "Minivas tape cueing error\n"));
    minivasRecord();
    if (minivasPollStatus())
	fatal((stderr, "Minivas record error\n"));
}
#endif


/*
 * Exit routine, making sure to close the minivas and the A60
 */

quit(code) {
    extern abekasFtpCmd;

    if (abekasFtpCmd >= 0)
	abekasFtpClose();
#ifndef A60ONLY
    if (minivas)
	minivasClose();
#endif
    if (abekas)
	abekasClose();
    exit(code);
}


/*
 * undo23pulldown() tries to reverse the effects of converting film to video
 * using a telecine with the 2-3 pulldown algorithm:  In the 2-3 pulldown
 * scheme, 2 film frames become 5 video fields (or for every 5 video frames
 * there are 2 of them which contain fields out of different film frames):
 *
 *		     /------ Video field 0, video frame 0, state 0, stable frame
 *   Film frame 0 --*
 *                   \------ Video field 1, video frame 0
 *                   /------ Video field 2, video frame 1, state 1, stable frame
 *   Film frame 1 --*------- Video field 3, video frame 1
 *                   \------ Video field 4, video frame 2, state 2, mixed fields
 *                   /------ Video field 8, video frame 2
 *   Film frame 2 --*
 *                   \------ Video field 9, video frame 3, state 3, mixed fields
 *                   /------ Video field 5, video frame 3
 *   film frame 3 --*------- Video field 6, video frame 3, state 4, stable frame
 *                   \------ Video field 7, video frame 3
 *
 * If the return value is 1, then the there is an output frame, otherwise
 * there is no output from this state.
 */

undo23pulldown(in, out)
u_char *in, *out;
{
    register u_char *a, *b, *alast;
    register int i, n, error;
    static int first = 1;

    switch (state23++) {
    
    case 0:
        memcpy(out, in, 2 * D1WIDTH * D1HEIGHT);
	break;

    case 1:
        memcpy(out, in, 2 * D1WIDTH * D1HEIGHT);
	/*
	 * Save 1st field chronologically (odd lines) in fieldbuf[]
	 */
	a = in + 2*D1WIDTH;
	b = fieldbuf;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(b, a, 2*D1WIDTH);
	    a += 4*D1WIDTH;
	    b += 2*D1WIDTH;
	}
	break;

    case 2:
	if (first == 0) {
	    /*
	     * Compare first field (odd lines) of this video frame to the first
	     * field of the last video frame.  They should be "identical" if
	     * this is 2-3 pulldown of a movie original.
	     */
	    error = 0;
	    a = fieldbuf;
	    b = in + 2*D1WIDTH;
	    for (i = 0; i < D1HEIGHT/2; i++) {
		for (alast = a + 2*D1WIDTH; a < alast; ) {
		    if ((n = (int)*a++ - (int)*b++) < 0) n = - n;
		    error += n;
		}
		b += 2*D1WIDTH;
	    }
	    if (error > (D1WIDTH * D1HEIGHT * AVEERROR))
		fatal((stderr,
		"Field 0 of frames %d and %d are not the same: ave error=%f\n",
		    inframe - 1, inframe, (double)error/(D1WIDTH * D1HEIGHT)));
	}
	/*
	 * Save the second field (even lines) in fieldbuf[]
	 */
	a = fieldbuf;
	b = in;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(a, b, 2*D1WIDTH);
	    b += 4*D1WIDTH;
	    a += 2*D1WIDTH;
	}
	first = 0;
	return(0);	/* No output frame avaliable */

    case 3:
	/*
	 * First field (odd lines) of output (movie) frame is taken from the
	 * first field (odd lines) of the input (video) frame.
	 */
	a = in + 2*D1WIDTH;
	b = out + 2*D1WIDTH;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(b, a, 2*D1WIDTH);
	    b += 4*D1WIDTH;
	    a += 4*D1WIDTH;
	}
	/*
	 * Second field (even lines) of output (movie) frame is taken from
	 * fieldbuf[] which is the second field (even lines) of the previous
	 * input (video) frame.
	 */
	a = fieldbuf;
	b = out;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(b, a, 2*D1WIDTH);
	    b += 4*D1WIDTH;
	    a += 2*D1WIDTH;
	}
	/*
	 * Save the second field (even lines) of input in fieldbuf[]
	 */
	a = in;
	b = fieldbuf;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(b, a, 2*D1WIDTH);
	    b += 2*D1WIDTH;
	    a += 4*D1WIDTH;
	}
	break;

    case 4:
	state23 = 0;
	if (first == 0) {
	    a = fieldbuf;
	    b = in;
	    error = 0;
	    for (i = 0; i < D1HEIGHT/2; i++) {
		for (alast = a + 2*D1WIDTH; a < alast; ) {
		    if ((n = (int)*a++ - (int)*b++) < 0) n = - n;
		    error += n;
		}
		b += 2*D1WIDTH;
	    }
	    if (error > (D1WIDTH * D1HEIGHT * AVEERROR))
		fatal((stderr,
		"Field 1 of frames %d and %d are not the same: ave error=%f\n",
		    inframe - 1, inframe, (double)error/(D1WIDTH * D1HEIGHT)));
	}
        memcpy(out, in, 2 * D1WIDTH * D1HEIGHT);
	break;
    }
    first = 0;
    return(1);
}

/*
 * do23pulldown() converts a 24-frame per second (movie frame rate) video to
 * 30 frame per second video by mapping one input video frame into first
 * 2 output fields then 3 output fields.  If the return value is non-zero
 * (or state23 == 4), then the input frame is to be copies to the output
 * stream AFTER the returned output frame is processed.  The global state,
 * state23, must be reset to zero at this time.
 */

do23pulldown(in, out)
u_char *in, *out;
{
    register u_char *a, *b;
    register int i;

    switch (state23++) {

    case 0:
        memcpy(out, in, 2 * D1WIDTH * D1HEIGHT);
	return(0);

    case 1:
        memcpy(out, in, 2 * D1WIDTH * D1HEIGHT);
	/*
	 * Save 1st field of movie frame in fieldbuf[]
	 */
	a = fieldbuf;
	b = in;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(a, b, 2*D1WIDTH);
	    b += 2*D1WIDTH;
	    a += 4*D1WIDTH;
	}
	return(0);

    case 2:
	/*
	 * Construct the output video frame from the saved first field (odd
	 * lines) of the previous movie frame and the second field (even lines)
	 * of the current movie frame.
	 */
	a = fieldbuf;
	b = out + 2*D1WIDTH;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(b, a, 2*D1WIDTH);
	    b += 4*D1WIDTH;
	    a += 2*D1WIDTH;
	}
	a = in;
	b = out;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(b, a, 2*D1WIDTH);
	    b += 4*D1WIDTH;
	    a += 4*D1WIDTH;
	}
	/*
	 * Save the first field (odd lines) of the movie frame in fieldbuf[]
	 */
	a = fieldbuf;
	b = in;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(a, b, 2*D1WIDTH);
	    b += 2*D1WIDTH;
	    a += 4*D1WIDTH;
	}
	return(0);

    case 3:
	/*
	 * Construct the output video frame from the saved first field (odd
	 * lines) of the previous movie frame and the second field (even lines)
	 * of the current movie frame.
	 */
	a = fieldbuf;
	b = out + 2*D1WIDTH;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(b, a, 2*D1WIDTH);
	    b += 4*D1WIDTH;
	    a += 2*D1WIDTH;
	}
	a = in;
	b = out;
	for (i = 0; i < D1HEIGHT/2; i++) {
            memcpy(b, a, 2*D1WIDTH);
	    b += 4*D1WIDTH;
	    a += 4*D1WIDTH;
	}
	/*
	 * Following 'out' in the output sequence, make a copy of 'in'.
	 */
	state23 = 0;
	return(1);
    }
}

/*
 * frame_tc() converts a sequential frame number to a time code using the
 * dropped frame convention for NTSC.
 */

frame_tc(numb, hourp, minp, secp, framep)
int *hourp, *minp, *secp, *framep;
register int numb;
{
    *hourp = numb / ((2997 * 60 * 60) / 100);
    numb %= ((2997 * 60 * 60) / 100);
    *minp = (numb / 17982) * 10;
    numb %= 17982;			/* 17982 is nbr frames in 10 min */
    if (numb >= 1800) {			/* frames in 1min at 30 Hz */
	/*
	 * Drop 2 frames per minute, except every 10th minute.  Codes
	 * hh:mm:00:00 & hh:mm:00:01 never occur if mm is not a multiple
	 * of 10.  This will result in 108 frames per hour less than 30 Hz.
	 */
	numb -= 2;
	*minp += numb / (1800 - 2);
	numb %= (1800 - 2);
	numb += 2;
    }
    *framep = numb % 30;
    *secp = numb / 30;
}
